// -*-c++-*-
/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* 
 * Sources
 * 
 *   [list sources here]
 */

#include "ecc.h"

b163_curve_s *
b163_curve_s::b163_curve ()
{
  static bool done = false;    /* guard preventing recursive calls */
  static b163_curve_s *_curve = NULL;
  
  if (!done) {
    done = true;
    _curve = new b163_curve_s ();
  }
  
  return _curve;
}

inline
b163_curve_s::b163_curve_s ()
{
  m =  B163_MAX_DEGREE;
    
  // non-zero coeffs of the irredux
  p[0] =  163;
  p[1] =    7;
  p[2] =    6;
  p[3] =    3;
  p[4] =    1;    
  
  // Curve equation (over GF_2m/(p(.))): y^2 + x*y = x^3 + a*x + b 
  a = 0x1;
  
  /* b is 0x020A601907B8C953CA1481EB10512F78744A3205FD
   *  b[5] = 0x00000002;
   *  b[4] = 0x0A601907;
   *  b[3] = 0xB8C953CA;
   *  b[2] = 0x1481EB10;
   *  b[1] = 0x512F7874;
   *  b[0] = 0x4A3205FD;
   */

  const char *b_as_string = "0x020A601907B8C953CA1481EB10512F78744A3205FD";
  b163_set_str (b, b_as_string);

  // curve order is 0x040000000000000000000292FE77E70C12A4234C33 (in Z)
  const char *q_as_string = "0x040000000000000000000292FE77E70C12A4234C33";
  order = bigint(q_as_string);
  // mpz_init_set_str (order, q_as_string, 0);
  
  co_factor = 2;
  
  _base_point = NULL;    /* will be set later via base_point () */

#ifdef _USE_MIRACL_
  init_miracl();
#endif

}

#ifdef _USE_MIRACL_
void 
b163_curve_s::init_miracl () {
  _mip = mirsys (10, 0); /* up to 320-bit integers */
  const char *str = "020A601907B8C953CA1481EB10512F78744A3205FD";
  _mip->IOBASE = 16;
  ecurve2(p[0], p[1], p[2], p[3], Big(1), 
	  Big ((char *)str),
	  false, MR_PROJECTIVE);
}
#endif

b163_point *
b163_curve_s::base_point () {
  if (!_base_point) {
    
    // XXX: should multiply the base point G by a random Zq element r_net 
    //      to ensure cryptographic separation of networks
    
    // base point x-coord is 0x03F0EBA16286A2D57EA0991168D4994637E8343E36
    b163_elem xG = { 0xE8343E36,
		     0xD4994637,
		     0xA0991168,
		     0x86A2D57E,
		     0xF0EBA162,
		     0x00000003 };
    
    // base point y-coord is 0x00D51FBC6C71A0094FA2CDD545B11C5C0C797324F1
    b163_elem yG = { 0x797324F1,
		     0xB11C5C0C,
		     0xA2CDD545,
		     0x71A0094F, 
		     0xD51FBC6C,
		     0x00000000 };
    
    _base_point = new b163_point(xG, yG);
  }
  
  return _base_point;
}

/* Factory function for point decompression (y-coordinate only) */
b163_point *
b163_point::make_point (b163_elem _x, u_int32_t ybit) {
  b163_point *res = NULL;
  b163_elem _y;
  
  if (B163_IS_ZERO(_x))
    b163_sqrt (_y, b163_curve()->b);
  else {
    b163_elem _xx, beta, h;
    
    /* 
     * Solving the quadratic equation: y^2 +x*y = x^3 + a*x^2 + b
     *
     * Dividing by x^2: (y/x)^2 +(y/x) = x + a + b/x^2
     * 
     * This has the form h^2 + h = beta: solution is Half-trace(beta)
     * 
     */
    
    /* beta = b/x^2 + a + x */
    b163_square(_xx, _x);
    b163_div (beta, b163_curve()->b, _xx);
    b163_add4g_into (beta, b163_curve()->a);
    b163_add_into (beta, _x);
    
    /* h = Half-Trace (beta) */
    b163_halftrace (h, beta);
    
    /* Let's make sure we got a good h, checking that (h+1)*h = beta  */
    b163_elem h1;
    
    b163_copy (h1, h);
    b163_add4g_into (h1, 0x1);
    b163_mul (h1, h1, h);
    
    /* Compare by xoring together and checking against hero */
    b163_add_into (h1, beta);
    
    if (!B163_IS_ZERO(h1)) {
      res = NULL;
    } else {
      /* Our two solutions are h*x and h*x + x.  ybit encodes which of
	 the two solutions to pick as follows:
	 
	 if y_bit equals the lsb of h,        pick h*x, 
	 otherwise                            pick h*x + x
      */
      b163_mul (_y, h, _x);
      
      if (B163_CONSTANT_TERM(h) != ybit)
	b163_add_into (_y, _x);
    }
    res = new b163_point (_x, _y);
  }

  return res;
}

/* Factory function for point decompression (both coordinates)    */
/* Note: ybit (LSB of y/x) is stored as LSB (constant) term of _x */
b163_point *
b163_point::make_point (b163_elem _x) {
  u_int32_t ybit;
  b163_elem x_;

  uncompress_x(x_, &ybit, _x);
 
  /* Now Call the y-decompressing constructor */
  return make_point (x_, ybit);
}


/* Conveniency function to decompress xc into x and ybit     */
/* Note: does not check whether xc is a valid compressed x-coordinate */
void
b163_point::uncompress_x (b163_elem _x, u_int32_t *ybit, b163_elem _xc) {
  b163_copy (_x, _xc);

  /* retrieve the ybit */
  *ybit = B163_CONSTANT_TERM(_x);

  b163_clear_constant_term (_x);

  /* Set the constant term of _x if clearing it yields trace 0 */
  /* XXX: Note that this is specific to B-163 */
  if (0 == B163_TRACE(_x))
    b163_add4g_into (_x, 0x1);
}

/* Check whether (xx, yy) lies on the curve */
bool
b163_point::is_on_curve (b163_elem xx, b163_elem yy)
{
  if (B163_IS_ZERO(xx) && B163_IS_ZERO(yy)) return true;
  else {  
    b163_elem lhs, rhs;
    
    /* LHS: y^2 + x*y, computed as (y+x)*y */
    b163_copy (lhs, yy);
    b163_add_into (lhs, xx);
    b163_mul (lhs, lhs, yy);
    
    /* RHS: x^3 + a*x^2 + b, computed as (x+a)*x*x + b */
    b163_copy (rhs, xx);
    b163_add4g_into (rhs, b163_curve()->a);
    b163_mul (rhs, rhs, xx);
    b163_mul (rhs, rhs, xx);
    b163_add_into (rhs, b163_curve()->b);
    
    /* LHS == RHS iff their sum is zero */
    b163_add_into (lhs, rhs);
    
    return (B163_IS_ZERO(lhs));
  }
}

void
b163_point::add_p (b163_point &R, b163_point &P, b163_point &Q, bool proj) {
  if (P.is_zero ()) R.set_to (Q);
  else if (Q.is_zero ()) R.set_to (P);
  else if (proj) {
    // Following code uses three temps instead of 7, and works even if P and R same
    b163_elem A, B, C;

    // Second point ought to be affine
    Q.affine_eq ();

    b163_mul (B, Q.x, P.z);     // B   = X2.Z1
    b163_add_into (B, P.x);     // B  += X1
    b163_mul (C, P.z, B);       // C   = Z1.B
    b163_square (R.x, P.z);     // X3  = Z1^2  (D = Z1^2)
    b163_mul (A, Q.y, R.x);     // A   = Y2.X3 (= Y2.Z1^2)
    b163_add_into (A, P.y);     // A  += Y1    (= Y2.Z1^2 + Y1)

    /* R.x *= a */              // X3 *= a     (D = a.Z1^2)
    switch (b163_curve()->a) {
    case 0:
      b163_set_to_zero(R.x);
      break;
    case 1:
      break;
    default: assert (false);    // the a-param can only be 0 or 1 in NIST binary curves
    }  

    b163_add_into (R.x, C);     // X3 += C     (D = a.Z1^2 + C)
    b163_square (B, B);
    b163_mul (R.x, R.x, B);     // X3 *= B^2   (D = (a.Z1^2 + C).B^2)
    b163_square (R.z, C);       // Z3  = C^2
    b163_mul (C, C, A);         // C  *= A     (E = A.C)
    b163_add_into (R.x, C);     // X3 += C     (= D + E)
    b163_square (A, A);
    b163_add_into (R.x, A);     // X3 += A^2   (= A^2 + D + E)
    b163_mul (A, Q.x, R.z);     // A   = X2.Z3 (F = X2.Z3)
    b163_add_into (A, R.x);     // A  += X3    (F = X3 + X2.Z3)
    b163_mul (B, Q.y, R.z);     // B   = X2.Z3 (G = Y2.Z3)
    b163_add_into (B, R.x);     // B  += X3    (G = X3 + Y2.Z3)
    b163_mul (B, B, R.z);       // B  *= Z3    (= Z3.G)
    b163_mul (R.y, C, A);       // Y3  = C*A   (= E.F)
    b163_add_into (R.y, B);     // Y3 += B     (Y3 = E.F + Z3.G)
  }
  else {
    P.affine_eq ();   Q.affine_eq ();

    /* check if same x-coordinate */
    b163_elem xdiff, ydiff;
    
    b163_copy (xdiff, P.x);
    b163_add_into (xdiff, Q.x);  /* + and - are same op over GF_2^m */
    
    b163_copy (ydiff, P.y);
    b163_add_into (ydiff, Q.y);
    
    if (B163_IS_ZERO(xdiff)) {
      /* same x-coordinate; now check y-coordinates */	
      if (B163_IS_ZERO(ydiff)) {
	/* P and Qare the same point; do point doubling */
	double_p (R, P, proj);
      }
      else {
	/* same x-coordinate, but different y-coordinate: sum is zero */
	/* before returning, though, check that xor of y's equals x */
	b163_add_into (ydiff, P.x);
	
	assert (B163_IS_ZERO(ydiff));
	
	R.set_to_zero();
      } 
    } else { 
      /* adding distinct non-zero points that aren't negation of each other */
      b163_elem slope, xs, ys;
      
      /* slope = (P.y + Q.y) / (P.x + Q.x) = ydiff / xdiff */
      b163_div (slope, ydiff, xdiff);
      
      /* xs = slope*slope + slope + P.x + Q.x + a */
      b163_square(xs, slope);
      b163_add_into (xs, slope);
      b163_add_into (xs, P.x);
      b163_add_into (xs, Q.x);
      b163_add4g_into (xs, b163_curve()->a);
      
      /* ys = (P.x+xs)*slope + xs + P.y */
      b163_copy (ys, P.x);
      b163_add_into (ys, xs);
      b163_mul(ys, ys, slope);
      b163_add_into (ys, xs);
      b163_add_into (ys, P.y);
      
      R.set_to (xs, ys);
    }
  }
}

void 
b163_point::double_p (b163_point &R, b163_point &P, bool proj) {
  if (B163_IS_ZERO(P.x)) 
    R.set_to_zero ();  /* point at infinity */
  else if (proj) {
    // Following code works when P and R are same
    b163_elem B;

    b163_square (B, P.z);       // B   = Z1^2
    b163_square (R.x, P.x);     // X3  = X1^2
    b163_mul (R.z, R.x, B);     // Z3  = X3.B  (= X1^2.Z1^2)
    b163_square (R.x, R.x);     // X3 *= X3    (= X1^4)

    b163_square (B, B);         // B  *= B     (= Z1^4)
    b163_mul(B, B, b163_curve()->b); // B *= b (= b.Z1^4)         
    b163_add_into (R.x, B);     // X3 += B     (= X1^4 + b.Z1^4)
    b163_square (R.y, P.y);     // Y3  = Y1^2
    b163_add_into (R.y, B);     // Y3 += B     (= Y1^2 + b.Z1^4)

    /* R.y += a*R.z */          // Y3 += a.Z3  (= Y1^2 + b.Z1^4 + a.Z3)
    switch (b163_curve()->a) {
    case 0:
      break;
    case 1:
      b163_add_into (R.y, R.z);
      break;
    default: assert (false);    // the a-param can only be 0 or 1 in NIST binary curves
    }  

    b163_mul (R.y, R.y, R.x);   // Y3 *= X3    (= (Y1^2 + b.Z1^4 + a.Z3).X3)
    b163_mul (B, B, R.z);       // B  *= Z3    (= b.Z1^4.Z3)b.Z1^4.Z3)
    b163_add_into (R.y, B);     // Y3 += B     (= (Y1^2 + b.Z1^4 + a.Z3).X3 + b.Z1^4.Z3)
  }
  else {
    b163_elem slope, xx, yy;

    P.affine_eq ();
    
    /* slope = x + y/x */
    b163_div (slope, P.y, P.x);
    b163_add_into (slope, P.x);
    
    /* xx = slope*slope + slope + a */
    b163_square(xx, slope);
    b163_add_into (xx, slope);
    b163_add4g_into (xx, b163_curve()->a);
    
    /* yy = (x+xx)*slope + xx + y */
    b163_copy (yy, P.x);
    b163_add_into (yy, xx);
    b163_mul (yy, yy, slope);
    b163_add_into (yy, xx);
    b163_add_into (yy, P.y);
    
    R.set_to (xx, yy);
  }
}

/* 
 * cost of scalar multiplication goes with the number of one bits in scalar;
 * since negation is cheap, check if that would help
 *
 */
void
b163_point::mul_s (b163_point &R, b163_point &P, bigint &scalar, bool proj) {
#ifdef _USE_MIRACL_
  EC2 RR, PP;
  Big ss;

  miracl_from_b163 (PP, P);
  miracl_from_sfs (ss, scalar);

  RR = ss * PP;

  b163_from_miracl (R, RR);

#else
  bigint s = scalar;

  /* make sure the scalar is between 0 and q-1 */
  if ((mpz_sgn (&s) <= 0) || (mpz_cmp (&s, &b163_curve()->order) >= 0))
    mpz_mod (&s,  &s, &b163_curve()->order);
    
  /* check if scalar multiplier is zero */
  if (0 == mpz_sgn (&s)) {
    R.set_to_zero ();
  } else {
    /* compute q - s */
    bigint neg_s = b163_curve()->order - s;

    /* 
     * check which one among s and neg_s has the fewest 1-bit in it,
     * make s_ point to that, and set P_ to point to P or negP accordingly
     */
    bigint *s_;
    b163_point negP;
    b163_point *P_;
    
    if (mpz_popcount (&neg_s) < mpz_popcount (&s)) {
      s_ = &neg_s;
      
      set_to_neg_p (negP, P);
      P_ = &negP;
    } else {
      s_ = &s;

      P_ = &P;
    }

    /* now do vanilla double-and-add */

    size_t i = mpz_sizeinbase2 (s_) - 1;

    R.set_to (*P_);

    while (i-- > 0) {
      /* double ... */
      double_p (R, R, proj);
      
      if (mpz_getbit (s_, i)) {
	/* ... add */
	add_p (R, R, *P_, proj);
      }
    }

    // if we used projective coordinate, flip back
    if (proj)  R.affine_eq();
  }
#endif
}

/* 
 * multi-scalar multiplication 
 *
 * D = a.A + b.B,       where 0 <= a, b < q
 *
 * Idea:
 *
 *       D =   a . A + b . B
 *         = (a-b).A + b.(A+B)
 *         = a.(A+B) + (b-a).B
 *
 * Check which pair among [a, b], [1||(a-b), b], [a, 1||(b-a)] has the least # of 1's
 * (where the 1 in the second and last pair accounts for the extra point addition)
 */
void
b163_point::mul_ms (b163_point &D, b163_point &A, bigint &a, b163_point &B, bigint &b, 
		    bool proj) {
#ifdef _USE_MIRACL_
  EC2 DD, AA, BB;
  Big aa, bb;

  miracl_from_b163 (AA, A);
  miracl_from_b163 (BB, B);
  miracl_from_sfs (aa, a);
  miracl_from_sfs (bb, b);

  DD = mul (aa, AA, bb, BB);

  b163_from_miracl (D, DD);

#else
#if 0 
  /* make sure both scalars are between 0 and q-1 */
  if ((mpz_sgn (&a) <= 0) || (mpz_cmp (&a, &b163_curve()->order) >= 0))
    mpz_mod (&a,  &a, &b163_curve()->order);

  if ((mpz_sgn (&b) <= 0) || (mpz_cmp (&b, &b163_curve()->order) >= 0))
    mpz_mod (&b,  &b, &b163_curve()->order);

  bigint neg_a = b163_curve()->order - a;
  bigint neg_b = b163_curve()->order - b;
  bigint a_minus_b;  
  bigint b_minus_a;  

  if ((mpz_cmp (&a, &b)) < 0) {
    mpz_sub (&b_minus_a,  &b, &a);
    mpz_sub (&a_minus_b,  &b163_curve()->order, &b_minus_a);
  } else {
    mpz_sub (&a_minus_b,  &a, &b);
    mpz_sub (&b_minus_a,  &b163_curve()->order, &a_minus_b);
  }
  /* 
   * check which one among a/neg_a, b/neg_b, a_minus_b/b_minus_a has the most
   * number of ones.  The computations will then continue using the other two vals
   */
  size_t aux;
  size_t cnt_a = mpz_popcount (&a); 
  if ((aux = mpz_popcount (&neg_a)) < cnt_a)
    cnt_a = aux;

  size_t cnt_b = mpz_popcount (&b); 
  if ((aux = mpz_popcount (&neg_b)) < cnt_b)
    cnt_b = aux;

  size_t cnt_m = mpz_popcount (&a_minus_b); 
  if ((aux = mpz_popcount (&b_minus_a)) < cnt_m)
    cnt_m = aux;

  /* set c to 'a','b', or 'm' to keep track of which case to avoid */
  char c = (cnt_a >= cnt_b) ? 
             ((cnt_a >= 1 + cnt_m) ? 'a' : 'm') :
             ((cnt_b >= 1 + cnt_m) ? 'b' : 'm');

  b163_point C;

  switch (c) {
  case 'm': /* just do a.A + b.B */
    mul_s (D, A, a, proj);
    mul_s (C, B, b, proj);
    add_p (D, D, C, proj);
    
    break;
  case 'a': /* do (a-b).A + b.(A+B) */
    mul_s (D, A, a_minus_b, proj);
    add_p (C, A, B, proj);
    mul_s (C, C, b, proj);
    add_p (D, D, C, proj);

    break;
  case 'b': /* do a.(A+B) + (b-a).B */
    add_p (D, A, B, proj);
    mul_s (D, D, a, proj);
    mul_s (C, B, b_minus_a, proj);
    add_p (D, D, C, proj);

    break;

  default: assert (false); /* should never get here */
  }
  
#else  
  b163_point C;

  mul_s (D, A, a, proj);
  mul_s (C, B, b, proj);
  add_p (D, D, C, proj);
#endif

  // if we used projective coordinate, flip back
  if (proj)  D.affine_eq();
#endif  
}
