// -*-c++-*-
/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* 
 * Sources
 * 
 *   [list sources here]
 */

#ifndef _ECC_H_
#define _ECC_H_

#include "b163.h"

#define _USE_MIRACL_


typedef u_int32_t chopped_ecpk_raw[5];
/* from ecdh.C */
extern const char *ecdh_suffix_str;       // included in all hashes

class b163_point;

#ifdef _USE_MIRACL_
#include "mr_wrap.h"
#endif

class b163_curve_s {
public:
  u_int16_t m;         // log of the underlying binary field 
  u_int16_t p[5];      // Support only tri- and penta-nomial irredux
  
  // Curve equation (over GF_2m/(p(.)): y^2 + x*y = x^3 + a*x + b 
  u_int8_t a;          // Either 0 or 1
  b163_elem b;
  bigint order;            // (prime) order of base_point
  u_int16_t co_factor; // order of curve is order*co_factor
 
private: 
  b163_point *_base_point;
  b163_curve_s (); 
  
public:
  b163_point *base_point (); 
  static b163_curve_s *b163_curve ();

#ifdef _USE_MIRACL_
public:
  static miracl *mip () { return b163_curve ()->_mip; }

private:
  miracl *_mip;
  void init_miracl ();
#endif

};

class b163_point {
public:
  static b163_curve_s *b163_curve () {
    return b163_curve_s::b163_curve ();
  }
  
protected:
  b163_elem x, y, z;
  
public:
  // Default no-arg constructor gives back the point at infinity
  b163_point () {
    b163_set_to_zero (x);
    b163_set_to_zero (y);
    b163_set_to_zero (z);
  }
  
  // XXX No constructor for projective coordinates
  b163_point (b163_elem _x, b163_elem _y) {
    set_to (_x, _y); 
  }

  // XXX this does not check for equality after rescaling of the z-coordinate
  bool operator== (const b163_point &other) const { 
    return (B163_ARE_EQUAL(x,other.x)
	    && B163_ARE_EQUAL(y,other.y)
	    && B163_ARE_EQUAL(z,other.z));
  }

  void set_to (b163_elem _x, b163_elem _y) {
    assert (is_on_curve (_x, _y));
    b163_copy (x, _x); 
    b163_copy (y, _y); 
    b163_set_to_one (z); 
  }

  void set_to (b163_elem _x, b163_elem _y, b163_elem _z) {
    // XXX No equation checking for projective coordinates
    //    assert (is_on_curve (_x, _y, _z));
    b163_copy (x, _x); 
    b163_copy (y, _y); 
    b163_copy (z, _z); 
  }

  b163_point (b163_point &other) {
    set_to (other); 
  }

  void set_to (b163_point &other) {     
    set_to (other.x, other.y, other.z); 
  }
    
  void clone_x (b163_elem x_) { b163_copy (x_, x); }
  
  void clone_y (b163_elem y_) { b163_copy (y_, y); }
  
  void clone_z (b163_elem z_) { b163_copy (z_, z); }
  
  static bool is_zero_p (b163_point &P) {
    return B163_IS_ZERO(P.z) || (B163_IS_ZERO(P.x) && B163_IS_ZERO(P.y)); 
  }

  static bool is_infinity_p (b163_point &P) { return is_zero_p (P); }
  
  bool is_zero () { return is_zero_p (*this); }
  
  bool is_infinity () { return is_zero (); }
  
  static bool is_on_curve (b163_elem xx, b163_elem yy);
  
  bool is_on_curve () { return is_on_curve (x, y); }
  
  static void set_to_zero_p (b163_point &R) {
    b163_set_to_zero (R.x);
    b163_set_to_zero (R.y);
    b163_set_to_zero (R.z);  // XXX or should we set z to 1 for a canonical zero?
  }

  void set_to_zero () { set_to_zero_p (*this); }

  static bool is_affine_p (b163_point &P) {
    if (P.is_infinity ()) return true;

    b163_elem _one;  
    b163_copy (_one, P.z); 
    b163_add4g_into (_one, 0x1);

    return B163_IS_ZERO (_one);
  }

  bool is_affine () { return is_affine_p (*this); }

  /* Projecive-to-Affine conversion:
   * 
   *   x' = x / z
   *   y' = y / z^2
   *   z' = 1
   */
  static void set_to_affine_p (b163_point &R, b163_point &P) {
    if (P.is_affine ()) {
      R.set_to(P);
    } else {
      b163_invert (R.z, P.z);
      b163_mul (R.x, P.x, R.z);
      b163_square (R.z, R.z);
      b163_mul (R.y, P.y, R.z);
      b163_set_to_one (R.z);
    }
  }  

  static void set_to_neg_p (b163_point &R, b163_point &P) {
    set_to_affine_p (R, P);
    b163_add_into (R.y, P.x);
  }

  static void add_p (b163_point &R, b163_point &P, b163_point &Q, bool proj=false);
  
  static void double_p (b163_point &R, b163_point &P, bool proj=false);
  
  static void mul_s (b163_point &R, b163_point &P, bigint &k, bool proj=true);

/* 
 * multi-scalar multiplication 
 *
 * D = a.A + b.B
 *
 */
  static void mul_ms (b163_point &D, b163_point &A, bigint &a, b163_point &B, bigint &b, 
		      bool proj=true);

  static void affinize_p (b163_point &P) {
    set_to_affine_p (P, P);
  }

  static void flip_sign_p (b163_point &P){
    set_to_neg_p (P, P);
  }

  static void add_into_p (b163_point &P, b163_point &Q) {
    add_p (P, P, Q);
  }

  static void double_into_p (b163_point &P){
    double_p (P, P);
  }
  
  static void mul_into_p (b163_point &P, bigint &k) {
    mul_s (P, P, k);
  }

  b163_point *affinize () {
    b163_point *res = new b163_point();
    set_to_affine_p (*res, *this);
    return res;
  }

  b163_point *negate () {
    b163_point *res = new b163_point();
    set_to_neg_p (*res, *this);
    return res;
  }
  
  b163_point *plus (b163_point &other) {
    b163_point *res = new b163_point();
    add_p (*res, *this, other);
    return res;
  }
  
  b163_point *twice () {
    b163_point *res = new b163_point();
    double_p (*res, *this);
    return res;
  }

  b163_point *times (bigint &scalar){
    b163_point *res = new b163_point();
    mul_s (*res, *this, scalar);
    return res;
  }

  void affine_eq () {
    if (!is_affine())  set_to_affine_p (*this, *this);
  }

  void flip_sign () {
    set_to_neg_p (*this, *this);
  }

  void plus_eq (b163_point &other) {
    add_p (*this, *this, other);
  }
  
  void twice_eq () {
    double_p (*this, *this);
  }

  void times_eq (bigint &scalar){
    mul_s (*this, *this, scalar);
  }

  void cat () {
    b163_cat (x);  b163_cat (y); b163_cat (z);
  }
  

  
  /* Factory function for point decompression (y-coordinate only) */
  /* Note: ybit (LSB of y/x) stored as LSB (constant) term of _x */
  static b163_point *make_point (b163_elem y, u_int32_t ybit);

  /* Factory function for point decompression (both coordinates) */
  static b163_point *make_point (b163_elem x);

  /* Conveniency function to decompress xc into x and ybit     */
  /* Note: does not check whether xc is a valid compressed x-coordinate */
  static void uncompress_x (b163_elem _x, u_int32_t *ybit, b163_elem _xc);

  /* Conveniency function to compress P into xc  */
  static void compress_x (b163_elem _xc, b163_point P) {

    P.affine_eq();
    /* compressed coordinate: all but lsb of x, followed by ybit */
    b163_copy (_xc, P.x);
    
    /* y-bit is lsb of z=y/x, unless x == 0, but then it doesn't matter  */
    b163_elem z;
    if (B163_IS_ZERO(_xc)) {
      b163_set_to_zero (z);
    } else {
      b163_div (z, P.y, P.x);
    }
    
    /* store the y-bit in the lsb of the raw chopped form */
    b163_clear_constant_term (_xc);
    if (1 == B163_CONSTANT_TERM(z)) {
      b163_add4g_into (_xc, 0x1);
    }
  }

  void compress_x (b163_elem _xc) { compress_x (_xc, *this); }
  
  /*
   * Make a public key from a chopped raw representation.
   *
   * Chopped format consists of a compressed x-coordinate where the 
   * top 3 bits have been removed.  This bits are recovered as the
   * three bottom bits of the SHA-1 of the chopped raw representation.
   * In general, it is not guaranteed that one would get a valid point 
   * on the curve, but assuming that SHA-1 behaves as a random oracle
   * the chances are roughly 50/50.  In terms of security, the loss in
   * the random oracle model equals the number of chopped bits, i.e., 3
   */
  
  static b163_point *make_point_from_chopped(chopped_ecpk_raw raw) {
    SFS_SHA1CTX(sc);
    u_char buf[SFS_SHA1HASHSIZE];
    
	// AS/MW NOTE: NOTE: NOTE:
	//    used to be sizeof(raw) but that is bug because
	//    compiler treats raw as a pointer, so (a) we weren't
	//    hashing the whole string and (b) we were getting
	//    different answers on different architectures

    SFS_SHA1UPDATE(sc,raw,sizeof(chopped_ecpk_raw));
    size_t l = strlen(ecdh_suffix_str);
    SFS_SHA1UPDATE(sc,ecdh_suffix_str,l);
    SFS_SHA1FINAL(sc,buf);
    
    b163_elem x;
    
    /* XXX this exposes the internal structure of b163_elem */
    x[0] = raw[0];
    x[1] = raw[1];
    x[2] = raw[2];
    x[3] = raw[3];
    x[4] = raw[4];
    x[5] = (u_int32_t)(buf[0] & 0x07);
    
    return make_point (x);
  }
  
  static bool chop (chopped_ecpk_raw raw, b163_point &P) {
    b163_elem xc;
    
    /* Get P in compressed coordinates into xc */
    P.compress_x (xc);
    
    /* XXX this exposes the internal structure of b163_elem */
    raw[0] =  xc[0];
    raw[1] =  xc[1];
    raw[2] =  xc[2];
    raw[3] =  xc[3];
    raw[4] =  xc[4];
    /* Purposely omit highest byte, as it only contains the 3 bit to chop */
    
    SFS_SHA1CTX(sc);
    u_char buf[SFS_SHA1HASHSIZE];
    
    SFS_SHA1UPDATE(sc,raw,sizeof(chopped_ecpk_raw));
    size_t l = strlen(ecdh_suffix_str);
    SFS_SHA1UPDATE(sc,ecdh_suffix_str,l);
    SFS_SHA1FINAL(sc,buf);

    /* now check if bottom 3 bits of the SHA-1 equals top 3 of xc */
    return ((buf[0] & 0x07) == (u_char)(xc[5] & 0x00000007));
  }

  bool chop_into (chopped_ecpk_raw raw) { return chop (raw, *this); }

  static bool choppable (b163_point &P) { 
    chopped_ecpk_raw raw;
    return chop (raw, P);
  }

  bool choppable () { return choppable (*this); }
};

#endif /* !_ECC_H_ */
