/// -*-c++-*-
/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* 
 * Sources
 * 
 *   [list sources here]
 */

#include "ecdh.h"

const char *ecdh_suffix_str="ECDHv0.0 (ICING)";


b163_point *
ec_nidh_premaster (ecdh_priv *priv, ecdh_pub *pub) {
  if (!priv || !pub) return NULL;
  else return pub->P->times (priv->x);
}

/* SHA-1 ([label,]G.x,P.x,ecdh_suffix_str) */
bool 
ec_point_label_kdf (u_char *buf, size_t len, b163_point *P, 
		    u_char *label, size_t llen) 
{
  if (!buf || !P) return false;
  else {
    SFS_SHA1CTX(sc);
    if (label)     SFS_SHA1UPDATE(sc,label,llen);
    
    b163_elem x;  

    b163_point::b163_curve()->base_point()->clone_x(x);
    SFS_SHA1UPDATE(sc,x,sizeof(u_int32_t)*B163_MAX_LEN);

    P->clone_x(x);
    SFS_SHA1UPDATE(sc,x,sizeof(u_int32_t)*B163_MAX_LEN);

    size_t l = strlen(ecdh_suffix_str);
    SFS_SHA1UPDATE(sc,ecdh_suffix_str,l);
    
    if (len < SFS_SHA1HASHSIZE) {
      u_char kdf[SFS_SHA1HASHSIZE];
      SFS_SHA1FINAL(sc,kdf);
      memcpy (buf, kdf, len);
    } else if (len > SFS_SHA1HASHSIZE) {
      u_char kdf[SFS_SHA1HASHSIZE];
      SFS_SHA1FINAL(sc,kdf);
      memcpy (buf, kdf, SFS_SHA1HASHSIZE);
      /* XXX: Should extend via a prng; for now, just zero-pad */
      bzero((char *)buf + SFS_SHA1HASHSIZE, len - SFS_SHA1HASHSIZE);
    } else SFS_SHA1FINAL(sc,buf);
    
    return true;    
  }
}
  
bool 
ec_nidh (u_char *buf, size_t len, ecdh_priv *priv, ecdh_pub *pub) 
{
  return ec_point_label_kdf (buf, len, ec_nidh_premaster (priv, pub), NULL, 0);
}

bool 
ecdh_pub::verify (u_char *msg, size_t len, bigint &e, bigint &s) 
{
  b163_point R, xorRP, negP;
  negP.set_to_neg_p (negP, *P);

  /*  R = s.G - e.P = s.G + e.negP */
  b163_point::mul_ms (R, *generator(), s, negP, e);
  
  /* Now we compute the xor of R and P just so we can stuff them both
     in the computation of raw_e */
  b163_point::add_p (xorRP, R, *P);
  
  bigint _e;     
  u_char raw_e[SFS_SHA1HASHSIZE];

  assert(ec_point_label_kdf (raw_e, SFS_SHA1HASHSIZE, &xorRP, msg, len));

  mpz_set_rawmag_le(&_e, (char *) raw_e, sizeof (raw_e));

  bool res = (mpz_cmp (&_e, &e) == 0);  
  return res;  
}

bool 
ecdh_priv::sign (bigint &e, bigint &s, u_char *msg, size_t len)
{
  b163_point *R, xorRP;

  u_char raw_e[SFS_SHA1HASHSIZE];

  make_ekp ();     if (!ekp) return false;
  
  R = ekp->public_half ();

  /* Now we compute the xor of R and P just so we can stuff them both in the 
     computation of raw_e */
  b163_point::add_p (xorRP, *R, *P);
  
  assert(ec_point_label_kdf (raw_e, SFS_SHA1HASHSIZE, &xorRP, msg, len));

  mpz_set_rawmag_le(&e, (char *) raw_e, sizeof (raw_e));

  /* s = e*x + r */ 
  mpz_mul (&s, &e, &x);
  mpz_mod (&s, &s, &order ());
  mpz_add (&s, &s, ekp->private_half ());
  mpz_mod (&s, &s, &order ());
  delete ekp;    ekp = NULL;
  // debug 
  //  assert (verify (msg, len, e, s)); 

  delaycb (0, wrap (this, &ecdh_priv::make_ekp));
  // Use the delaycb when switching to sfslite
  make_ekp ();

  return true;
}
