/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "gf2m.h"

/* degree of a polynomial in GF_2^32, ie, index of leftmost bit set to 1 */
u_int16_t 
gf4g_degree (u_int32_t A) {
  u_int8_t a3 = (A & 0xff000000) >> 24;
  u_int8_t a2 = (A & 0x00ff0000) >> 16;
  u_int8_t a1 = (A & 0x0000ff00) >>  8;
  u_int8_t a0 =  A & 0x000000ff;

  return (a3 ? 24 + GF256_DEGREE(a3) :
	  a2 ? 16 + GF256_DEGREE(a2) :
	  a1 ?  8 + GF256_DEGREE(a1) :
	  GF256_DEGREE(a0));
}

/* degree of a polynomial in GF_2m, ie, index of leftmost bit set to 1 */
u_int16_t 
gf_degree (size_t U_len, u_int32_t U[]) {
  size_t blk = U_len - 1;

  while ((blk > 0)  && (0 == U[blk]))
    blk--;

  return gf4g_degree(U[blk]) + 8*sizeof(u_int32_t) * blk;
}

/* Polynomial multiplication (monomials, i.e., left shifts)  */

/* Special case: GF_2^32                        */
/* Input:  32-bit O, A, size_t k (0<=k<=32)     */
/* Output: (O||0) ^ (A << k), stored in (O, A)  */
/*         (Note: Use a zero O to get A << k    */

u_int32_t gf64k_shl_tbl[32] =
  { 0x00000000,    0x80000000,    0xc0000000,    0xe0000000,
    0xf0000000,    0xf8000000,    0xfc000000,    0xfe000000,
    0xff000000,    0xff800000,    0xffc00000,    0xffe00000,
    0xfff00000,    0xfff80000,    0xfffc0000,    0xfffe0000,
    0xffff0000,    0xffff8000,    0xffffc000,    0xffffe000,
    0xfffff000,    0xfffff800,    0xfffffc00,    0xfffffe00,
    0xffffff00,    0xffffff80,    0xffffffc0,    0xffffffe0,
    0xfffffff0,    0xfffffff8,    0xfffffffc,    0xfffffffe };

void
gf64k_shl (u_int32_t *OO, u_int32_t *AA, size_t k) {

  if (32 == k) {
    *OO ^= *AA;
    *AA = 0;
  } else if (0 == k) return;
  else {
    assert ((0 < k) && (k < 32));

    /* XOR in the top k bits from A into the bottom k bits of O */
    (*OO) ^= (((*AA) & (gf64k_shl_tbl[k])) >> (32 - k));
    (*AA) <<= k;
  }
}

#include <stdio.h>

/* Check that a 32-byte array has got enough space for a 
   left shift to go through without bits falling on the floor */
int
gf_can_do_shift(size_t U_len, u_int32_t U[], size_t k) {
  size_t bitcount  = 1+gf_degree (U_len, U) + k ;

  return (32*U_len >= bitcount);
}

/* General case (note: assumes space is enough) */
/* Input:  32-bit array U, size_t k >=0         */ 
/* Output: (U << k), stored in U                */
void
gf_shl (size_t U_len, u_int32_t U[], size_t k) {

  if (k <= 0) return;

  assert (gf_can_do_shift(U_len, U, k));

/* first shift by entire blocks ((k / 32) of them, left to right) */
  size_t blk_cnt = k / 32;
  size_t i;

  if (blk_cnt > 0) {
    for (i=U_len-1; i >= blk_cnt; i--)
      U[i] = U[i-blk_cnt];
    do {
      U[i] = 0;
    } while (i-- > 0);
  }

/* then shift by leftover (k % 32): first the last entry by itself, then
  in pairs, left to right */
  size_t lft_ovr = k % 32;

  if (lft_ovr > 0) {
    
    U[U_len-1] <<= lft_ovr;     /* assertion above ensure no bit-dropping */
    for (i=U_len-1; i>=1; i--) 
      gf64k_shl (&(U[i]), &(U[i-1]), lft_ovr);
  }
}

/* Polynomial multiplication  */

/* Special case: Product in GF_2^16 (deg-15 polynmials) */
/* Input:  16-bit a, b */
/* Output: 32-bit C += a*b */

#if 0 /* suppressing the table lookup sped gf64k_mul considerably */
u_int32_t gf64k_mul_tbl[16] =
  { 0x0001,
    0x0002,
    0x0004,
    0x0008,
    0x0010,
    0x0020,
    0x0040,
    0x0080,
    0x0100,
    0x0200,
    0x0400,
    0x0800,
    0x1000,
    0x2000,
    0x4000,
    0x8000 };
#endif


/* #if 0 */ /* Replaced my MACRO/ASM */

u_int32_t
gf64k_mul (u_int16_t a, u_int16_t b) {

  u_int32_t C = 0;

#if 0 /* unrolling this and suppressing the table sped it up considerably */
  int i;
  u_int32_t D = a;
  
  for (i=0; i<16; i++) {
    if (b & gf64k_mul_tbl[i])  
      (C ^= D);          
    D <<= 1;
  }
#endif

  u_int32_t D = a;

  if (b & 0x0001) (C ^= D);
  D <<= 1;
  if (b & 0x0002) (C ^= D);
  D <<= 1;
  if (b & 0x0004) (C ^= D);
  D <<= 1;
  if (b & 0x0008) (C ^= D);
  D <<= 1;
  if (b & 0x0010) (C ^= D);
  D <<= 1;
  if (b & 0x0020) (C ^= D);
  D <<= 1;
  if (b & 0x0040) (C ^= D);
  D <<= 1;
  if (b & 0x0080) (C ^= D);
  D <<= 1;
  if (b & 0x0100) (C ^= D);
  D <<= 1;
  if (b & 0x0200) (C ^= D);
  D <<= 1;
  if (b & 0x0400) (C ^= D);
  D <<= 1;
  if (b & 0x0800) (C ^= D);
  D <<= 1;
  if (b & 0x1000) (C ^= D);
  D <<= 1;
  if (b & 0x2000) (C ^= D);
  D <<= 1;
  if (b & 0x4000) (C ^= D);
  D <<= 1;
  if (b & 0x8000) (C ^= D);

  return C;
}

#define GF64K_MUL(C,a,b)      C = gf64k_mul (a, b)

/* #endif */

  
/* Need 3 registers, and AND (andw) XOR (xorl) SHL ops */
/* 
 * 1. bind %0 I/O to C (long), %1 with D (long), %2 with b (word)
 *     (note: eax is input/output register)
 * 2. testw each bit of %2 in turn; on zero, jump over the xorl %1 %0
 * 3. left shift %1
 * 4. do this 15 more times
 * Note: 'q' encodes "eax, ebx, ecx, edx"
 */

#if 0
#define GF64K_MUL(C,a,b) \
  __asm__ __volatile__ (           \
           "xorl %0, %0\n\t"       \
           "testw $0x0001, %2\n\t" \
	   "je 1\n\t"              \
	   "xorl %1, %0\n"         \
       "1:\n\t"                    \
	   "sall %1\n\t"           \
	   "testw $0x0002, %2\n\t" \
	   "je 2\n\t"              \
	   "xorl %1, %0\n"         \
       "2:\n\t"                    \
	   "sall $1, %1\n\t"       \
	   "testw $0x0004, %2\n\t" \
	   "je 3\n\t"              \
	   "xorl %1, %0\n"         \
       "3:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0008, %2\n\t" \
	   "je 4\n\t"		   \
	   "xorl %1, %0\n"	   \
       "4:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0010, %2\n\t" \
	   "je 5\n\t"		   \
	   "xorl %1, %0\n"	   \
       "5:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0020, %2\n\t" \
	   "je 6\n\t"              \
	   "xorl %1, %0\n"	   \
       "6:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0040, %2\n\t" \
	   "je 7\n\t"		   \
	   "xorl %1, %0\n"	   \
       "7:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0080, %2\n\t" \
	   "je 8\n\t"		   \
	   "xorl %1, %0\n"	   \
       "8:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0100, %2\n\t" \
	   "je 9\n\t"		   \
	   "xorl %1, %0\n"	   \
       "9:\n\t"			   \
	   "sall %1\n\t"	   \
	   "testw $0x0200, %2\n\t" \
	   "je 10\n\t"		   \
	   "xorl %1, %0\n"	   \
       "10:\n\t"		   \
	   "sall %1\n\t"	   \
	   "testw $0x0400, %2\n\t" \
	   "je 11\n\t"		   \
	   "xorl %1, %0\n"	   \
       "11:\n\t"		   \
	   "sall %1\n\t"           \
	   "testw $0x0800, %2\n\t" \
	   "je 12\n\t"		   \
	   "xorl %1, %0\n"	   \
       "12:\n\t"		   \
	   "sall %1\n\t"	   \
	   "testw $0x1000, %2\n\t" \
	   "je 13\n\t"		   \
	   "xorl %1, %0\n"	   \
       "13:\n\t"		   \
	   "sall %1\n\t"	   \
	   "testw $0x2000, %2\n\t" \
	   "je 14\n\t"		   \
	   "xorl %1, %0\n"	   \
       "14:\n\t"                   \
	   "sall %1\n\t"	   \
	   "testw $0x4000, %2\n\t" \
	   "je 15\n\t"		   \
	   "xorl %1, %0\n"	   \
       "15:\n\t"		   \
	   "sall %1\n\t"	   \
	   "testw $0x8000, %2\n\t" \
	   "je 16\n\t"		   \
	   "xorl %1, %0\n"	   \
       "16:" 			   \
	      : "=&q" ((u_int32_t) C)	   \
              : "q" ((u_int32_t) (a)), "q" ((u_int16_t)b) \
	      : "cc")
#endif


/* Special case: Product in GF_2^32 (deg-31 polynomials)  */
/* Input:  32-bit A, B                                    */
/* Output: 32-bit X, Y (X high, Y low),    (X||Y) ^= A*B  */
/*                                                        */
/*                       A ( |----a_h----||----a_l----| ) */
/*                     * B ( |----b_h----||----b_l----| ) */
/*                                                        */
/*                       ===============================  */
/*                                                        */
/*                           |(a_l.b_l)_h||(a_l.b_l)_l|   */
/*                                 +                      */
/*              |(a_h.b_l)_h||(a_h.b_l)_l|                */
/*                   +             +                      */
/*              |(b_h.a_l)_h||(b_h.a_l)_l|                */
/*                   +             +                      */
/* |(a_h.b_h)_h||(a_h.b_h)_l|                             */
/*                                                        */
/* ====================================================   */
/*                                                        */
/* |----x_h----||----x_l----||----y_h----||----y_l----|   */
/*                                                        */
/* |-----------X------------||------------Y-----------|   */
/*                                                        */

void 
gf4g_mul (u_int32_t *XX, u_int32_t *YY, u_int32_t A, u_int32_t B) {
  u_int16_t a_h, a_l; 
  u_int16_t b_h, b_l; 

  u_int32_t C=0;

  u_int32_t x_h, x_l; 
  u_int32_t y_h, y_l;

  u_int32_t h, l;

  a_l = (A & 0x0000ffff);
  a_h = (A & 0xffff0000) >> 16;
  b_l = (B & 0x0000ffff);
  b_h = (B & 0xffff0000) >> 16;

  GF64K_MUL(C,a_l,b_l);   
  l = (C & 0x0000ffff); 
  h = (C & 0xffff0000) >> 16;

  y_l = l;
  y_h = h;

  GF64K_MUL(C,a_h,b_l);
  l = (C & 0x0000ffff); 
  h = (C & 0xffff0000) >> 16;

  y_h ^= l;
  x_l = h;

  GF64K_MUL(C,b_h,a_l);   
  l = (C & 0x0000ffff); 
  h = (C & 0xffff0000) >> 16;

  y_h ^= l;
  x_l ^= h;

  GF64K_MUL(C,a_h,b_h);   
  l = (C & 0x0000ffff); 
  h = (C & 0xffff0000) >> 16;

  x_l ^= l;
  x_h = h;

  (*YY) ^= ((y_h << 16) ^ (y_l));
  (*XX) ^= ((x_h << 16) ^ (x_l));
}

/* Special case: (U||A)*B                                 */
/* Input:  32-bit array U; 32-bit A, B                    */
/* Output  32-bit array W; 32-bit Z, Y (Z high, Y low)    */
/*                                                        */
/*              (W||Z||Y) ^= (U||A)*B       (note: ^=)    */
/*                                                        */
/*         (Note: Use a zero O to get A << k                         */
/*                                  |-----U------||--A--| */
/*                                                |--B--| */
/*                                                        */
/*                       ================================ */
/*                                                        */
/*                                         |--X--||--Y--| */
/*                                            +           */
/*                         (|-----U------|*|--B--|)       */  
/*                                                        */
/* [ above recurses to:    (|--U_--||-A_-|*|--B--|) ,     */
/*   stored in:             |---W_--||-Z_-||--Y_-|  ]     */
/*                                                        */
/* ====================================================== */
/*                                                        */
/*                 |-----------W----------||--Z--||--Y--| */
/*                                                        */
#if 0 /* optimized by unrolling the tail recursion */
void 
gf_mul_by4g (u_int32_t W[], u_int32_t *ZZ, u_int32_t *YY, 
	     size_t U_len, u_int32_t U[], u_int32_t A, u_int32_t B) {

  gf4g_mul (ZZ, YY, A, B);           /* no need to zero anything out */

  if (U_len > 0) {
    u_int32_t A_ = U[0];             /* least significant u_int32_t */ 
    u_int32_t *U_ = U+1;           
    u_int32_t *ZZ_ = &(W[0]);
    u_int32_t *W_ = W+1;           
    gf_mul_by4g (W_, ZZ_, ZZ, U_len-1, U_, A_, B);   
  }
}
#endif 

void 
gf_mul_by4g (u_int32_t W[], u_int32_t *ZZ, u_int32_t *YY, 
	     size_t U_len, u_int32_t U[], u_int32_t A, u_int32_t B) {

  /* gf4g_mul (ZZ, YY, A, B); */           /* no need to zero anything out */
  /* unrolling tail recursion moves it to:
     - first line of while body
     - first line after while
  */


  /* if (U_len > 0) {     */
  while (U_len-- > 0) {
    gf4g_mul (ZZ, YY, A, B);           /* no need to zero anything out */

    /* u_int32_t A_ = U[0]; */             /* least significant u_int32_t */ 
    A = *U;
 
   /* u_int32_t *U_ = U+1; */
    U++;

    /* recursive call uses ZZ in the arg-position of YY */
    YY = ZZ;

    /* u_int32_t *ZZ_ = &(W[0]); */
    ZZ = W;

    /* u_int32_t *W_ = W+1; */
    W++;

    /* suppressed after unrolling tail recursion */
    /* gf_mul_by4g (W_, ZZ_, ZZ, U_len-1, U_, A_, B);  */
  }

  gf4g_mul (ZZ, YY, A, B);           /* no need to zero anything out */

}

/* General case: (U||A)*(V||B), U as long as V or longer  */
/* Input:  32-bit arrays U, V; 32-bit A, B                */
/* Output  32-bit array S; 32-bit T, Y (T high, Y low)    */
/*                                                        */
/*              (S||T||Y) ^= (U||A)*(V||B)  (note: ^=)    */
/*                                                        */
/*                                  |-----U------||--A--| */
/*                                      |---V----||--B--| */
/*                                                        */
/*                       ================================ */
/*                                                        */
/*                                         |--X--||--Y--| */
/*                                            +           */
/*                         (|-----U------|*|--B--|)       */  
/*                                        +               */
/*                             (|---V----|*|--A--|)       */  
/*                             +                          */
/*              (|-----U------|*|---V----|)               */  
/*                                                        */
/* recurses to: (|--U_--||-A_-|*|-V_||-B-|)               */
/*   stored in:  |----S_---||--T_-||--Y_-|]               */
/*                                                        */
/* ====================================================== */
/*                                                        */
/*                 |-----------S----------||--T--||--Y--| */
/*                                                        */
#if 0 /* optimized by switching to tail recursion and unrolling */
void 
gf_mul (u_int32_t S[], u_int32_t *TT, u_int32_t *YY, 
	size_t U_len, u_int32_t U[], u_int32_t A, 
	size_t V_len, u_int32_t V[], u_int32_t B) {
  
  assert (U_len >= V_len);

  if (0 == V_len)    gf_mul_by4g (S, TT, YY, U_len, U, A, B);
  else {
    u_int32_t A_ = U[0];        
    u_int32_t *U_ = U+1;
    u_int32_t B_ = V[0];
    u_int32_t *V_ = V+1;
    u_int32_t *YY_ = &(S[0]);
    u_int32_t *TT_ = &(S[1]);
    u_int32_t *S_ = S+2;               /* S will always have length >=2 */
    u_int32_t *ST_ = S+1;              /* i.e., S_||T_ */

    /* S_and T_ starts off respectively as 
       high and middle of U*V; after that,
       ST_ (consisting of S_||T_) accumumulates 
       first high of V*A and then high of U*B */

    /* Y_ starts off as low of U*V
       then accumulates middle of V*A,
       then accumulates middle of U*B */

    /* T starts off as low of V*A, 
       then accumulates low of U*B, 
       then accumulates high of A*B */

    my_gf_mul (S_, TT_, YY_, U_len-1, U_, A_, V_len-1, V_, B_);

    gf_mul_by4g (ST_, YY_, TT, V_len-1, V_, B_, A);

    gf_mul_by4g (ST_, YY_, TT, U_len-1, U_, A_, B);

    gf4g_mul (TT, YY, A, B);            
  }
}
#endif

void 
gf_mul (u_int32_t S[], u_int32_t *TT, u_int32_t *YY, 
	size_t U_len, u_int32_t U[], u_int32_t A, 
	size_t V_len, u_int32_t V[], u_int32_t B) {
  
  assert (U_len >= V_len);

  /* unrolling tail recursion on gf_mul */

  /* if (0 == V_len)    gf_mul_by4g (S, TT, YY, U_len, U, A, B); 
     else { */
  while  (V_len > 0) {
    /* decrement both lengths */
    U_len--, V_len--;

    /* keep this for fst call to gf_mul_by4g */
    u_int32_t A_ = U[0];            

    /* advance U */
    /* u_int32_t *U_ = U+1; */
    U = U+1;

    /* keep this for snd call to gf_mul_by4g */
    u_int32_t B_ = V[0];

    /* advance V */
    /* u_int32_t *V_ = V+1; */
    V = V+1;

    /* keep the following for the two calls to gf_mul_by4g */
    u_int32_t *ST_ = S+1;              /* i.e., S_||T_ */

    gf4g_mul (TT, YY, A, B);            

    /* now advance YY */
    /* u_int32_t *YY_ = &(S[0]); */
    YY = S;

    /* gf_mul_by4g (ST_, YY_, TT, U_len-1, U_, A_, B); */
    gf_mul_by4g (ST_, YY, TT, U_len, U, A_, B);

    /* gf_mul_by4g (ST_, YY_, TT, V_len-1, V_, B_, A); */
    gf_mul_by4g (ST_, YY, TT, V_len, V, B_, A);

    /* finally advance S, TT, A, and B */
    
    /* u_int32_t *TT_ = &(S[1]); */
    TT = S+1;

    /*	u_int32_t *S_ = S+2; */            /* S will always have length >=2 */
    S = S+2;

    A = A_;
    B = B_;

    /* gf_mul (S_, TT_, YY_, U_len-1, U_, A_, V_len-1, V_, B_); */
  }

  gf_mul_by4g (S, TT, YY, U_len, U, A, B);
}

/* Convenience method for computing S=U*V */
void 
gf_multiply (u_int32_t S[], size_t U_len, u_int32_t U[],
		  size_t V_len, u_int32_t V[]) {
  assert ((U_len >= V_len) && (V_len > 0));
  gf_mul (S+2, &(S[1]), &(S[0]), 
	  U_len-1, U+1, U[0], V_len-1, V+1, V[0]);
}


