#include <click/config.h>
#include "icingcommon.hh"
#include "checknexthop.hh"
#include "icingconfig.hh"
#include "realmtable.hh"

CLICK_DECLS

CheckNextHop::CheckNextHop() : _con(NULL), _realmtable(NULL)
{
    _errh = router()->chatter_channel("default");
}

CheckNextHop::~CheckNextHop()
{
}

void
CheckNextHop::push(int /* port */, Packet* p)
{
    int out;
    if ( (p = simple_action(p, &out)) )
	output(out).push(p);
}

// only returns a packet if "pull" is successful, i.e., if the "pull"
// caused a packet to be emitted on the same port as pulled the packet.
// in other cases, either the packet is pushed out a *different* port,
// or else the packet is killed. in both of those cases, this function
// returns NULL.
Packet*
CheckNextHop::pull(int pullingport)
{
    assert(pullingport == 0 || pullingport == 2);

    Packet* p;
    if ( (p = input(0).pull())) {

	int outport;
	if ( (p = simple_action(p, &outport)) ) {
	    if (outport == pullingport) {
		return p;
	    } else {
		output(outport).push(p);
	    }
	}
    }

    return NULL; 
}

Packet*
CheckNextHop::handle_at_destination(Packet* p, int* outport)
{
    if (noutputs() > 2) {
	*outport = 2;
    } else {
	p->kill();
	p = NULL;
    }
    return p;
}

Packet*
CheckNextHop::handle_unknown_next_hop(Packet* p, int* outport)
{
    if (noutputs() > 1) {
	*outport = 1;
    } else {
	p->kill();
	p = NULL;
    }
    return p;
}

// returns either a packet, which can then be pushed to *outport, or
// else NULL, in which case the packet's memory has been freed and no
// further action on the packet is possible.
Packet* 
CheckNextHop::simple_action(Packet* p, int* outport)
{
    const IcingHdr1* ice_hdr = 
	    reinterpret_cast<const IcingHdr1*>(p->data());

    uint8_t idx = ice_hdr->path_idx;
    uint8_t nrealms = ice_hdr->path_len;

    assert(idx < nrealms);

    if (++idx == nrealms) 
	return handle_at_destination(p, outport);

    const Path* path = get_path(ice_hdr);
    const RealmVoq* next_realmvoq = &path->realmvoqs[idx];

    RealmID packet_rid;
    packet_rid.parse(next_realmvoq->realmid);

    smart_ptr<const RealmInfo> RI = _realmtable->lookup_realm(packet_rid);
    if (RI == NULL || RI->entry_kind == RealmInfo::NON_NEIGH)
	return handle_unknown_next_hop(p, outport);

    *outport = 0;
    return p;
}

int
CheckNextHop::configure(Vector<String> &conf, ErrorHandler *errh)
{
    int before = errh->nerrors();

    Element* elem;

    if (cp_va_kparse(conf, this, errh,
		     "CONFIG", cpkP+cpkM, cpElement, &elem,
		     cpEnd) < 0)
	return -1;

    if ( (_con = static_cast<IcingConfig*>(elem->cast("IcingConfig"))) == NULL)
	errh->error("CONFIG needs to be an IcingConfig object");

    if ( (_realmtable = _con->realmtable()) == NULL)
	errh->error("This element requires CONFIG to specify a REALMTABLE");

    return (before == errh->nerrors() ? 0 : -1);
}


CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(CheckNextHop)
