/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <click/config.h>
#include "sha1.h"
#include "async.h"
#include "crypt.h"
#include "cryptkey.hh"

CLICK_DECLS

ptr<rabin_privkey> 
gen_rabin_privkey(int nbits)
{
    rabin_priv privk = rabin_keygen(nbits);
    ptr<rabin_priv> privk_copy = New refcounted<rabin_priv>(privk.p, privk.q);
    return New refcounted<rabin_privkey> (privk_copy);
}

ptr<rabin_pub>
rabin_pubkey::buf2rabinpub(const char* p, int len, cryptkeyformat_t f)
{
    str s = (f == CKF_B64_T) 
	    ? dearmor64(p, len) 
	    : str(p, len);

    if (s.cstr() == NULL) return NULL;

    bigint n;
    n.setraw(s);

    return New refcounted<rabin_pub>(n);
}

// Rabin public keys are stored as base64-encoded binary strings. the
// binary looks like:
// <n-in-raw-form>
ptr<rabin_pubkey>
rabin_pubkey::from_str(const str& s, cryptkeyformat_t f)
{
    ptr<rabin_pub> p;

    if ( (p = buf2rabinpub(s.cstr(), s.len(), f)) == NULL)
	return NULL;

    return New refcounted<rabin_pubkey>(p);
}

bool
rabin_pubkey::verify(const str& msg, const str& sig) const
{
    assert(pubk != NULL);

    bigint sig_bi;

    sig_bi.setraw(sig);

    return pubk->verify(msg, sig_bi);
}

bool
rabin_pubkey::parse_from_buf(const char* buf, int len, cryptkeyformat_t f)
{
    ptr<rabin_pub> p;

    if ( (p = buf2rabinpub(buf, len, f)) == NULL)
	return false;

    pubk = p;

    return true;
}

bool
rabin_pubkey::parse_from_buf(const str& s, cryptkeyformat_t f)
{
    return parse_from_buf(s.cstr(), s.len(), f);
}

str
rabin_pubkey::to_b64() const
{
    assert(pubk != NULL);
    return armor64(pubk->n.getraw()); 
}

str
rabin_pubkey::to_bytes() const
{
   assert(pubk != NULL);
   return pubk->n.getraw();
}

ptr<rabin_priv>
rabin_privkey::buf2rabinpriv(const char* buf, int len, cryptkeyformat_t f)
{
    str s = (f == CKF_B64_T) 
	    ? dearmor64(buf, len)
	    : str(buf, len);

    if (s.cstr() == NULL) return NULL;

    bigint p, q;

    const char* pq_str = s.cstr();
    int len_p = getint(pq_str);
    pq_str = &pq_str[4];
    p.setraw(str(pq_str, len_p));

    pq_str = &pq_str[len_p];
    int len_q = getint(pq_str);
    pq_str = &pq_str[4];
    q.setraw(str(pq_str, len_q));

    return New refcounted<rabin_priv>(p,q);
}

// Rabin private keys are stored as base-64 encoded binary strings. The
// binary format is:
// <4-byte-big-endian-length-of-p><raw p><4-byte-big-endian-length-of-q><raw q>
ptr<rabin_privkey>
rabin_privkey::from_str(const str& s, cryptkeyformat_t f)
{
    ptr<rabin_priv> p;
    
    if ( (p = buf2rabinpriv(s.cstr(), s.len(), f)) == NULL)
	return NULL;
    
    return New refcounted<rabin_privkey>(p);
}

str
rabin_privkey::sign(const str& msg) const
{
    assert(privk != NULL);

    bigint sig = privk->sign(msg);
    return sig.getraw();
}

bool
rabin_privkey::parse_from_buf(const char* buf, int len, cryptkeyformat_t f)
{
    ptr<rabin_priv> p;
    if ( (p = buf2rabinpriv(buf, len, f)) == NULL)
	return false;

    privk = p;
    return true;
}

bool
rabin_privkey::parse_from_buf(const str& s, cryptkeyformat_t f)
{
    return parse_from_buf(s.cstr(), s.len(), f);
}

str
rabin_privkey::to_bytes() const
{
    assert(privk != NULL);

    str p_str = privk->p.getraw();
    str q_str = privk->q.getraw();
    
    int p_len, q_len;
    
    p_len = p_str.len();
    q_len = q_str.len();

    mstr m(4 + p_len + 4 + q_len);

    char* mp = m.cstr();

    putint(mp, p_len);
    mp = &mp[4];

    memcpy(mp, p_str.cstr(), p_len);
    mp = &mp[p_len];

    putint(mp, q_len);
    mp = &mp[4];

    memcpy(mp, q_str.cstr(), q_len);

    return m;
}

str
rabin_privkey::to_b64() const
{
    assert(privk != NULL);
    return armor64(to_bytes());
}

CLICK_ENDDECLS
ELEMENT_PROVIDES(CryptKeyFuncs)

