#include <click/config.h>
#include "icingcommon.hh"
#include "hwinterface.hh"
#include "icingconfig.hh"
#include "realmtable.hh"

CLICK_DECLS

HardwareInterface::HardwareInterface() : 
    _con(NULL), _errh(NULL), _nf2(NULL)
{
    _errh = router()->chatter_channel("default");
}

HardwareInterface::~HardwareInterface()
{
    
}

void
HardwareInterface::reset()
{
    nf_icing3_close_dev(_nf2);
    delete _nf2;
    _nf2 = NULL;

    init_nf2(NULL);
}

int
HardwareInterface::configure(Vector<String> &conf, ErrorHandler *errh)
{
    int before = errh->nerrors();

    Element* elem;

    if (cp_va_kparse(conf, this, errh,
		     "CONFIG", cpkP+cpkM, cpElement, &elem,
		     "FIRST_IFACE", cpkP+cpkM, cpString, &_iface_anchor,
		     "ETHADDR0", cpkP+cpkM, cpEtherAddress, &_eth_addrs[0],
		     "ETHADDR1", cpkP+cpkM, cpEtherAddress, &_eth_addrs[1],
		     "ETHADDR2", cpkP+cpkM, cpEtherAddress, &_eth_addrs[2],
		     "ETHADDR3", cpkP+cpkM, cpEtherAddress, &_eth_addrs[3],
		     cpEnd) < 0)
	return -1;

    if ( (_con = static_cast<IcingConfig*>(elem->cast("IcingConfig"))) == NULL)
	errh->error("CONFIG needs to be an IcingConfig object");

    return (before == errh->nerrors() ? 0 : -1);
}

int
HardwareInterface::initialize(ErrorHandler* errh)
{
    int before = errh->nerrors();
    int rc;

    rc = init_nf2(errh);
    if (rc) return -1;

    //clear();

    return (errh->nerrors() != before ? -1 : 0);
}

void
HardwareInterface::cleanup(CleanupStage)
{
    nf_icing3_close_dev(_nf2);
}

bool
HardwareInterface::set_entry(smart_ptr<const RealmInfo> ri)
{
    nf_icing_realm_info_t hri = toHWInfo(ri);

    // if there is going to be a collision
    if(nf_icing3_check_realm(&hri, _nf2) != 0)
		return false;

    nf_icing3_write_realm(&hri, _nf2);
	return true;
}

bool
sanity_check(SharedKey &checked,
		const RealmIDWire &r1,
		ecdh_priv &priv)
{
	SharedKey correct;
	set_sharedkey(correct, r1, priv);
	return !memcmp(correct, checked, sizeof(SharedKey));
}

smart_ptr<const RealmInfo>
HardwareInterface::get_entry(const RealmID &rid)
{
    smart_ptr<RealmInfo> rinfo = NULL;

    nf_icing_realm_info_t hri;
    memcpy(hri.realm_id.id, rid.toWire()->bytes, RealmIDWire::SIZE);
    nf_icing3_read_realm(&hri, _nf2);

    rinfo = fromHWInfo(hri);

    if(rinfo->rid != rid)
	return NULL;

    return rinfo;
}

nf_icing_realm_info_t
HardwareInterface::toHWInfo(smart_ptr<const RealmInfo> ri)
{
    EtherAddress ea;
    nf_icing_realm_info_t hri;
    bzero(&hri, sizeof(nf_icing_realm_info));

    if (ri->entry_kind != RealmInfo::NON_NEIGH) {
	// first static cast below is an ugly hack resulting from the fact that 
	// my refcounted pointer class sucks and won't do
	// the right thing without being explicitly coerced.
	IcingLinkAddr* temp = const_cast<IcingLinkAddr*>(
				static_cast<const IcingLinkAddr*>(ri->linkaddr));
	L2LinkAddr* la_l2 = static_cast<L2LinkAddr*>(temp->cast("L2LinkAddr"));
	assert(la_l2 != NULL);

	if(ri->entry_kind == RealmInfo::PHYS_NEIGH)
	{
		ea = la_l2->ea();
		hri.output_mask = encode_mac_fwd(la_l2->outputport());
	}
	if(ri->entry_kind == RealmInfo::PCI_NEIGH)
		hri.output_mask = encode_cpu_fwd(la_l2->outputport());
    }
    memcpy(hri.realm_id.id, ri->rid.toWire()->bytes, RealmIDWire::SIZE);
    memcpy(hri.shared_key, ri->sharedkey, sizeof(SharedKey));
    memcpy(hri.next_hop_mac, ea.data(), sizeof(hri.next_hop_mac));

    return hri;
}

smart_ptr<RealmInfo>
HardwareInterface::fromHWInfo(nf_icing_realm_info_t &hri)
{
    smart_ptr<RealmInfo> rinfo = new RealmInfo;
    RealmIDWire rid;

    memcpy(rid.bytes, hri.realm_id.id, RealmIDWire::SIZE);
    memcpy(rinfo->sharedkey, hri.shared_key, sizeof(SharedKey));

    assert(sanity_check(rinfo->sharedkey, rid, _con->privkey()));

    rinfo->rid = rid;

    uint8_t outmask = hri.output_mask;
    int8_t decode;

    if (!outmask) {

	rinfo->entry_kind = RealmInfo::NON_NEIGH;
	rinfo->linkaddr = NULL;

    } else {

	L2LinkAddr* l2addr = new L2LinkAddr;
	
	EtherAddress ea(hri.next_hop_mac);
	l2addr->set_ea(ea);
	if((decode = decode_cpu_fwd(outmask)) != -1)
	{
	    rinfo->entry_kind = RealmInfo::PCI_NEIGH;
	    l2addr->set_outputport(decode);
	}
	else if((decode = decode_mac_fwd(outmask)) != -1)
	{
	    rinfo->entry_kind = RealmInfo::PHYS_NEIGH;
	    l2addr->set_outputport(decode);
	}
	rinfo->linkaddr = l2addr;

    }

    return rinfo;
}

void
HardwareInterface::rehash(Vector<smart_ptr<const RealmInfo> > &entries)
{
	Vector<smart_ptr<const RealmInfo> >::iterator iter;
	nf_icing_realm_info_t **hris;
	int i;

	hris = new nf_icing_realm_info_t*[entries.size()];
	iter = entries.begin();

	for(iter = entries.begin(), i = 0; iter != entries.end(); iter++, i++)
	{
		hris[i] = new nf_icing_realm_info;
		*hris[i] = toHWInfo(*iter);
	}
	
	nf_icing3_change_uhf(hris, entries.size(), 3, _nf2);

	for(i = 0; i < entries.size(); i++)
		delete hris[i];
	delete [] hris;
}

// No longer works. Need to do a hard reset of the hardware.
void
HardwareInterface::clear()
{
	warn("HardwareInterface::clear(): Operation not supported by ICING3");
}

#if 0
String
HardwareInterface::dump() const
{
    StringAccum sa;

    sa << "Hardware Table:\n";
    dump_table(sa);
   
    return sa.take_string();
}
#endif

int
HardwareInterface::init_nf2(ErrorHandler* errh)
{
    _nf2 = new nf_icing_dev;
    nf_icing_dev_config nf2config;

    nf2config.dev_name = new char[_iface_anchor.length() + 1];
    strcpy(nf2config.dev_name, _iface_anchor.c_str());

    RealmIDWire* rid_wire = _con->realmid().toWire();
    for (uint8_t i = 0; i < NPORTS; i++) 
	memcpy(nf2config.macs + 6*i, _eth_addrs[i].data(), 6);

    memcpy(nf2config.realm_id.id, rid_wire->bytes, RealmIDWire::SIZE);

    int error;
    if((error = nf_icing3_init_dev(_nf2, &nf2config, 1, 1, 1)) != 0)
    {
	errh->error("Could not initialize NF2 device, error code: %d", error);
	return -1;
    };

	click_chatter("filling poc key cache");
	const RealmSecret *cache = _con->derivekey()->cache();
	for(uint32_t i = 0; i < (1 << 16); i++)
	{
		vnode_t vnode;
		poc_key_t key;

		memcpy(key.key, cache[i].bytes, RealmSecret::SIZE);
		*(uint32_t *)vnode.vnode = htonl(i << 16);
		
		nf_icing3_write_poc_key(&vnode, &key, _nf2);
	}

    click_chatter("Initialized NF2");
    return 0;
}

void
HardwareInterface::clear_entry(const RealmID &rid)
{
    nf_icing_realm_info_t hri;
    memcpy(hri.realm_id.id, rid.toWire()->bytes, RealmIDWire::SIZE);
    nf_icing3_delete_realm(&hri, _nf2);
}

bool
HardwareInterface::is_zeroed(const RealmIDWire& rid, const SharedKey& sharedkey, 
			     const EtherAddress& ea,
			     uint8_t outmask, uint32_t num_hits) const
{
    RealmIDWire null_rid;
    EtherAddress null_ea;
    SharedKey null_sharedkey;

    memset(null_sharedkey, 0, sizeof(null_sharedkey));

    bool answer = !memcmp(rid.bytes, null_rid.bytes, RealmIDWire::SIZE);

    if (answer) {

	bool double_check = 
	    !memcmp(sharedkey, null_sharedkey, sizeof(null_sharedkey)) &&
	    (ea == null_ea) && !outmask && !num_hits;

	assert(double_check);
    } 

    return answer;
}

#if 0
void
HardwareInterface::dump_table(StringAccum& sa) const
{
    RealmIDWire rid;
    SharedKey sharedkey;
    EtherAddress nexthop;
    uint8_t outmask;
    uint32_t num_hits;

    for (uint32_t i = 0; i < REALM_LOOKUP_TABLE_SIZE; i++) {

	nf_icing_read_entry(i, rid.bytes, sharedkey, nexthop.data(),
			    &outmask, &num_hits, _nf2);
	if (!is_zeroed(rid, sharedkey, nexthop, outmask, num_hits)) 
	    dump_entry(sa, i, rid, sharedkey, nexthop, outmask, num_hits);
    }
}

void
HardwareInterface::dump_entry(StringAccum& sa, uint32_t index, 
			      const RealmIDWire& rid, 
			      const SharedKey& sharedkey,
			      const EtherAddress& ea, uint8_t outmask, 
			      uint32_t num_hits)  const
{
    sa << "\tindex: " << index << "\n"
       << "\trealm_id: " << myhexdump(rid.bytes, RealmIDWire::SIZE) << "\n"
       << "\tshared_key: " << myhexdump(sharedkey, sizeof(sharedkey)) << "\n"
       << "\tnexthopmac: " << ea << "\n"
       << "\toutmask: ";
    sa.snprintf(1, "%hhx", outmask) << "\n";
    sa << "\tnum_hits: " << num_hits << "\n";
}
#endif

// precondition: i is between 0 and 3.
uint8_t
HardwareInterface::encode_mac_fwd(uint8_t i) const
{
    assert((i & 0xfc) == 0);
    assert(i < NPORTS);
    return 1 << (2*i);
}


// precondition: i is between 0 and 3.
uint8_t
HardwareInterface::encode_cpu_fwd(uint8_t i) const
{
    assert((i & 0xfc) == 0);
    assert(i < NPORTS);
    return 2 << (2*i);
}

// return -1 if 'bits' doesn't encode exactly one MAC port
int8_t
HardwareInterface::decode_mac_fwd(uint8_t bits) const
{
    for (uint8_t i = 0; i < NPORTS; i++) 
	if (bits == (1 << (2*i)))
	    return i;
    
    return -1;
}

// return -1 if 'bits' doesn't encode exactly one CPU port
int8_t
HardwareInterface::decode_cpu_fwd(uint8_t bits) const
{
    for (uint8_t i = 0; i < NPORTS; i++) 
	if (bits == (2 << (2*i)))
	    return i;

    return -1;
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(icing linux userlevel)
EXPORT_ELEMENT(HardwareInterface)
