#include <click/config.h>
#include "icingcommon.hh"
#include "realmtable.hh"
#include "icingconfig.hh"

CLICK_DECLS

IcingConfig::IcingConfig() : _mode(MODE_UNSET), _realmtable(NULL), 
			     _hwinterface(NULL), _derivekey(NULL)
{
    _daemon_ip.s_addr = 0;
    _daemon_port = 0;
    _errh = router()->chatter_channel("default");
}

IcingConfig::~IcingConfig()
{
}


int
IcingConfig::configure(Vector<String> &conf, ErrorHandler *errh)
{
    int before = errh->nerrors();

    String k_str, realmid_str, mode_str, privk_str;
    bool k_given = false;

    Element* realmtable  = NULL;
    Element* hwinterface = NULL;
    Element* derivekey   = NULL;
    RealmSecret k;

    if (cp_va_kparse(conf, this, errh,
	     "OUR_REALMID", cpkP+cpkM, cpString, &realmid_str,
	     "OUR_PRIV_K", cpkP+cpkM, cpString, &privk_str,
	     "K", cpkP+cpkC, &k_given, cpString, &k_str,
	     "DERIVEKEY",cpkP, cpElement, &derivekey,
	     "OUR_LINK_IP", cpkP, cpIPAddress, _daemon_ip,
	     "OUR_LINK_PORT", cpkP, cpUDPPort, _daemon_port,
	     "REALMTABLE", cpkP, cpElement, &realmtable,
	     "HWINTERFACE", cpkP, cpElement, &hwinterface,
	     "MODE", cpkP, cpString, &mode_str,
	     "USE_POCOMON", cpkP, cpBool, _use_pocomon,
	     "POCOMON_PORT", cpkP, cpUDPPort, _pocomon_port,
	     cpEnd) < 0)
	return -1;
    _pocomon_port = htons(_pocomon_port);
    
     if (!derivekey && !k_given) {
	errh->error("Either a DERIVEKEY or a K must be specified");
	return -1;
     }
    // MW: XXX: if k_given == true and derivekey != NULL, then should
    // you check that the k inside derivekey matches k_given, or no?
    // i.e., should you have some code like
    // 
    // if (derivekey && k_given) {
    //	   <sanity check>
    //	   if (sanity_fails)  {
    //		errh->error(...);
    //		return -1;
    //	    }
    //	}

    if (k_given && !cp_icing_realmsecret(k_str, &k))
	errh->error("Error processing k");

    if (!cp_icing_realmid(realmid_str, &_realmid)) {
	errh->error("Error processing realmid");
	return -1;
    }

    if(! cp_icing_privkey(privk_str, &_privkey, &_realmid))
	errh->error("Configured with wrong private key");

    if (realmtable && 
	  (_realmtable = static_cast<RealmTable*>(
			    realmtable->cast("RealmTable"))) == NULL)
	errh->error("REALMTABLE argument needs to be "
		    "of type RealmTable");

    if (hwinterface && 
	  (_hwinterface = static_cast<HardwareInterface*>(
			    hwinterface->cast("HardwareInterface"))) == NULL)
	errh->error("HWINTERFACE argument needs to be "
		    "of type HardwareInterface");

    if(derivekey &&
		(_derivekey = static_cast<DeriveKey *>(
		     derivekey->cast("DeriveKey"))) == NULL)
	errh->error("DERIVEKEY argument must be of type DeriveKey");

    if(!derivekey)
    {
	assert(k_given);

	// next line needed because this function call may be
	// re-configuration, and there may be a previous _derivekey
	// sitting around. 
	//
	//  MW: XXX: cleaner way to do it is with refcounted pointers.
	//  with refcounted pointers, an assignment operation like
	//  _derivekey = new DeriveKey(k) would automatically deref and
	//  possibly deallocate the old memory pointed to by _derivekey.
	//  for now, though, what you have is fine *provided* you can be
	//  sure that *other* elements are not maintaining pointers to
	//  the old _derivekey. can you guarantee that? 
	delete _derivekey; 

	_derivekey = new DeriveKey(k);
    }

    process_mode(mode_str, errh);
	
    return (errh->nerrors() == before ? 0 : -1);
}

void
IcingConfig::process_mode(const String& mode_str, ErrorHandler* errh)
{
    if (!_realmtable)
	return;

    if (!mode_str) {
	errh->error("Need to specify a MODE if specifying REALMTABLE ");
	return;
    }

    if (mode_str == "Overlay") {
	_mode = MODE_OVERLAY;
    } else if (mode_str == "L3") {
	_mode = MODE_L3;
    } else
	errh->error("Unrecognized value for MODE");
}

void
IcingConfig::add_handlers()
{
//    add_write_handler("k", reconfigure_keyword_handler, "1 K");
//    add_read_handler("k",  read_handler, 1, Handler::CALM);

    add_write_handler("privkey", reconfigure_keyword_handler, "2 PRIVATE_KEY");
    add_read_handler("privkey",  read_handler, 2, Handler::CALM);

    add_write_handler("ip", reconfigure_keyword_handler, "3 OUR_LINK_IP");
    add_read_handler("ip", read_handler, 3, Handler::CALM);

    add_write_handler("port", reconfigure_keyword_handler, "4 OUR_LINK_PORT");
    add_read_handler("port", read_handler, 4, Handler::CALM);

}

String
IcingConfig::read_handler(Element *e, void* thunk)
{
    IcingConfig* d = static_cast<IcingConfig*>(e);
    switch((intptr_t)thunk) {
	case 1:
	    return myhexdump(d->_derivekey->k().bytes, d->_derivekey->k().SIZE);
	case 2: {
	      //str s = d->_privkey.to_b64();
	      //return String(s.cstr(), s.len());
	}
	case 3:
	    return  IPAddress(d->_daemon_ip).unparse();
	case 4:
	    return String(d->_daemon_port);
	default:
	    return "<error>";
    }
}

const RealmSecret &
IcingConfig::k()
{
    return _derivekey->k();
}


CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(IcingConfig)
