#include <click/config.h>
#include "icingdencap.hh"
#include "translationtable.hh"
#include <clicknet/tcp.h>
#include <clicknet/udp.h>

CLICK_DECLS

int
IcingDencap::configure(Vector<String> &conf, ErrorHandler *errh)
{
	int before = errh->nerrors();
	Element *table, *config;

	if(cp_va_kparse(conf, this, errh,
				"CONFIG", cpkP+cpkM, cpElement, &config,
				"TRANSLATIONTABLE", cpkP+cpkM, cpElement, &table,
				"DEFAULT_IP", cpkP+cpkM, cpIPAddress, &_defaultAddress,
				cpEnd) < 0)
		return -1;

	if(config && (_config = static_cast<IcingConfig *>(config->cast("IcingConfig"))) == NULL)
		errh->error("Error processing IcingConfig");

	if(table && (_table = static_cast<TranslationTable *>(table->cast("TranslationTable"))) == NULL)
		errh->error("Error processing TranslationTable");

	return (errh->nerrors() == before ? 0 : -1);
}

Packet *
IcingDencap::simple_action(Packet *p)
{
	TranslationInfo *info = infoFromPacket(p);
	if(!info)
	{
		_errh->debug("Table out of memory; not good!");
		errorPacket(p);
	}

	WritablePacket *wp = p->uniqueify();
	if(!wp)
		return NULL;
	p = NULL;

	buildPacket(wp, info);
	return wp;
}

TranslationInfo *
IcingDencap::infoFromPacket(Packet *p)
{
	const IcingHdr1 *hdr1 = reinterpret_cast<const IcingHdr1 *>(p->data());
	const Path *path = get_path(hdr1);
	RealmID sender(path->realmvoqs[0].realmid);

	TranslationInfo *info = _table->lookup(sender);

	if(!info)
	{
		info = new TranslationInfo();
		info->infotype = TranslationInfo::IN;
		info->rid = sender;
		info->pathlen = hdr1->path_len;

		//reverse path
		RealmVoq *dst = new RealmVoq[info->pathlen];
		const RealmVoq *src = reinterpret_cast<const RealmVoq *>(path);
		for(int i = hdr1->path_len - 1, j = 0; i >= 0; i--, j++)
			memcpy(&dst[j], &src[i], sizeof(RealmVoq));
		info->path = reinterpret_cast<Path *>(dst);

		info->rpocs = NULL;
		info->remoteIP = _table->addEntry(info);
	}
	
	const ReturnPocs *rpocs = reinterpret_cast<const ReturnPocs *>(get_payload(hdr1));
	if(rpocs->num_rpocs)
	{
		if(info->pocs)
			delete [] info->pocs;
		info->pocs = new POC[rpocs->num_rpocs];
		memcpy(info->pocs, rpocs->rpocs, sizeof(POC) * rpocs->num_rpocs);
	}

	if(info->localIP == IPAddress())
		info->localIP = _defaultAddress;

	return info;
}

void
IcingDencap::buildPacket(WritablePacket *wp, TranslationInfo *info)
{
	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(wp->data());
	uint32_t hdrlen = icing_hdrlen(hdr1);

	uint8_t *payload = get_payload(hdr1);

	hdrlen += 1; //for the length of return pocs
	ReturnPocs *rpocs = reinterpret_cast<ReturnPocs *>(payload);
	if(rpocs->num_rpocs)
		hdrlen += sizeof(POC) * rpocs->num_rpocs;

	wp->pull(hdrlen);

	uint8_t ips[3];
	uint16_t ipl[2];
	memcpy(ips, wp->data(), sizeof(ips));
	wp->pull(sizeof(ips));
	memcpy(ipl, wp->data(), sizeof(ipl));
	wp->pull(sizeof(ipl));

	wp = wp->push(sizeof(click_ip));
	click_ip *ip = reinterpret_cast<click_ip *>(wp->data());
	wp->set_ip_header(ip, sizeof(click_ip));

	// MW: XXX: similar comment to icingencap: can't you copy the ip
	// header all at once?
	ip->ip_v = 4;
	ip->ip_hl = sizeof(click_ip) >> 2;
	ip->ip_tos = ips[0];
	ip->ip_ttl = ips[1] - 1;
	ip->ip_p = ips[2];
	ip->ip_id = ipl[0];
	ip->ip_off = ipl[1];

	ip->ip_len = htons(wp->length());
	ip->ip_src = info->remoteIP;
	ip->ip_dst = info->localIP;

	wp->set_network_header(wp->data(), sizeof(click_ip));

	ip->ip_sum = 0;
	ip->ip_sum = click_in_cksum((uint8_t *)ip, sizeof(click_ip));

	if(ip->ip_p == IP_PROTO_TCP)
	{
		click_tcp *tcp = wp->tcp_header();

		unsigned off = tcp->th_off << 2;
		unsigned plen = wp->length() - sizeof(click_ip);

		if(off < sizeof(click_tcp))
			tcp->th_off = sizeof(click_tcp) >> 2;
		else if(off  > plen && !IP_ISFRAG(ip))
			tcp->th_off = plen >> 2;

		tcp->th_sum = 0;
		unsigned csum = click_in_cksum((uint8_t*)tcp, plen);
		tcp->th_sum = click_in_cksum_pseudohdr(csum, ip, plen);
	}
	else if(ip->ip_p == IP_PROTO_UDP)
	{
		click_udp *udp = wp->udp_header();
		unsigned len = wp->length() - sizeof(click_ip);
		udp->uh_ulen = htons(len);
		udp->uh_sum = 0;
		unsigned csum = click_in_cksum((uint8_t *)udp, len);
		udp->uh_sum = click_in_cksum_pseudohdr(csum, ip, len);
	}
}

void
IcingDencap::errorPacket(Packet *p)
{
	if(noutputs() > 1)
		output(1).push(p);
	else
		p->kill();
}

CLICK_ENDDECLS

ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(IcingDencap)
