#include <click/config.h>
#include "icingencap.hh"
#include "translationtable.hh"

CLICK_DECLS

int
IcingEncap::configure(Vector<String> &conf, ErrorHandler *errh)
{
	int before = errh->nerrors();
	Element *table, *config;

	if(cp_va_kparse(conf, this, errh,
				"CONFIG", cpkP+cpkM, cpElement, &config,
				"TRANSLATIONTABLE", cpkP+cpkM, cpElement, &table,
				cpEnd) < 0)
		return -1;

	if(config && (_config = static_cast<IcingConfig *>(config->cast("IcingConfig"))) == NULL)
		errh->error("Error processing IcingConfig");

	if(table && (_table = static_cast<TranslationTable *>(table->cast("TranslationTable"))) == NULL)
		errh->error("Error processing TranslationTable");

	return (errh->nerrors() == before ? 0 : -1);
}

void
IcingEncap::buildPacket(WritablePacket *wp, TranslationInfo *info)
{
	click_ip *ip = wp->ip_header();
	uint8_t ips[3];
	uint16_t ipl[2];

	// MW: XXX: can't you copy the ip header all at once?
	ips[0] = ip->ip_tos;
	ips[1] = ip->ip_ttl;
	ips[2] = ip->ip_p;
	ipl[0] = ip->ip_id;
	ipl[1] = ip->ip_off;

	wp->pull(wp->ip_header_length());

	wp = wp->push(sizeof(ipl) + sizeof(ips));
	memcpy(wp->data(), ips, sizeof(ips));
	memcpy(wp->data() + sizeof(ips), ipl, sizeof(ipl));

	if(info->rpocs)
	{
		uint32_t rpoclen = (sizeof(POC) * info->pathlen) + 1;
		wp = wp->push(rpoclen);

		ReturnPocs *rpocs = reinterpret_cast<ReturnPocs *>(wp->data());
		rpocs->num_rpocs = info->pathlen;
		memcpy(rpocs->rpocs, info->rpocs, sizeof(POC) * rpocs->num_rpocs);
	}
	else
	{
		wp = wp->push(1);
		memset(wp->data(), 0, 1);
	}

	uint32_t hdrlen = icing_hdrlen(info->pathlen);
	wp = wp->push(hdrlen);

	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(wp->data());
	hdr1->icing_ver = ICING_CURRENT_VERSION;
	hdr1->path_len = info->pathlen;
	hdr1->path_idx = 0;

	IcingHdr2 *hdr2 = get_icinghdr2(hdr1);
	hdr2->proto = 8;
	hdr2->pkt_len = htons(wp->length());

	Path *path = get_path(hdr1);
	memcpy(path, info->path, sizeof(RealmVoq) * info->pathlen);
	init_vtags(hdr1, info->pocs, _config->privkey()); 
	hdr1->path_idx = 1;

	// MW: XXX: how come you don't have to recalculate the TCP
	// checksum if the Icing packet is carrying TCP? If the answer
	// is, "we don't care about the TCP checksum any more because
	// we'll calculate it at dencap time and it's not needed for
	// packet carriage as long as the TCP pkt is inside Icing", that
	// is fine, but it might be good to leave a comment to that
	// effect somewhere, either here or in the comments that explain
	// what the element does in the .hh file.
}

Packet *
IcingEncap::simple_action(Packet *p)
{
	TranslationInfo *info = infoFromPacket(p);
	if(!info)
	{
		_errh->debug("Error: No info found for this IP address");
		errorPacket(p);
		return NULL;
	}

	WritablePacket *wp = p->uniqueify();
	if(!wp)
		return NULL;
	p = NULL;

	buildPacket(wp, info);

	return wp;
}

TranslationInfo *
IcingEncap::infoFromPacket(Packet *p)
{
	IPAddress dest = p->dst_ip_anno();
	TranslationInfo *info;

	if( !(info = _table->lookup(dest)))
	{
		errorPacket(p);
		return NULL;
	}

	if(info->localIP == IPAddress())
	{
		// Get the source IP address
		const click_ip *ip = p->ip_header();
		IPAddress source(ip->ip_src);
		info->localIP = source;
	}

	return info;
}

void
IcingEncap::errorPacket(Packet *p)
{
	if(noutputs() > 1)
		output(1).push(p);
	else
		p->kill();
}

CLICK_ENDDECLS

ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(IcingEncap)
