// Following the pattern of InfiniteSource
#include <click/config.h>
#include "icingcommon.hh"
#include "icingsource.hh"
#include "initvector.hh"

CLICK_DECLS

IcingSource::IcingSource() 
    : _packet(0),
      _timer(this),
      _n(0),
      _interval(INTERVAL_DEFAULT)
{
}

IcingSource::~IcingSource()
{
}

int
IcingSource::configure(Vector<String> &conf, ErrorHandler *errh)
{
    int before = errh->nerrors();
    assert(_nrealms >= 3);
    String realm_str[_nrealms], k_str[_nrealms], privk_str;

    if (cp_va_kparse(conf, this, errh,
	     "SECRET1", cpkP+cpkM, cpString, &k_str[1],
	     "SECRET2", cpkP+cpkM, cpString, &k_str[2],
	     "REALM0", cpkP+cpkM, cpString, &realm_str[0],
	     "REALM1", cpkP+cpkM, cpString, &realm_str[1],
	     "REALM2", cpkP+cpkM, cpString, &realm_str[2],
		 "PRIVK", cpkP+cpkM, cpString, &privk_str,
		 "G1", cpkP+cpkM, cpInteger, &_g[1],
		 "G2", cpkP+cpkM, cpInteger, &_g[2],
		 "V1", cpkP+cpkM, cpInteger, &_v[1],
		 "V2", cpkP+cpkM, cpInteger, &_v[2],
	     "INTERVAL", cpkP, cpInteger, &_interval,
	     cpEnd) < 0)
	return -1;

    if (!init_secrets(k_str)) {
	errh->error("Error processing secrets");
	return -1;
    }

    if (!init_realmids(realm_str)) {
	errh->error("Error processing realm id");
	return -1;
    }

	if(!cp_icing_privkey(privk_str, &_privk, &_realmid[0]))
		errh->error("Configured with wrong private key");
 
    if (!setup_packet()) 
	errh->error("Couldn't set up packet");

    return (errh->nerrors() == before ? 0 : -1);
}

bool
IcingSource::init_secrets(const String* k_str)
{
    // Note that we start indexing from 1 because the source does not
    // have a k_0 for the present purposes
    for (int i = 1; i < _nrealms; i++) {
	if (!cp_icing_realmsecret(k_str[i], &_k[i])) 
	    return false;
    }

    return true;
}

bool
IcingSource::init_realmids(const String* realm_str)
{
    // NOTE: first realm id is our own
    for (int i = 0; i < _nrealms; i++) {
	if (!cp_icing_realmid(realm_str[i], &_realmid[i]))
	    return false;
    }
    return true;
}


int
IcingSource::initialize(ErrorHandler *errh)
{
    _timer.initialize(this);
    _timer.schedule_now();

    return 0;
}

void
IcingSource::cleanup(CleanupStage)
{
    cleanup_packet(&_packet);
}

void
IcingSource::cleanup_packet(Packet** p)
{
    if (*p)
	(*p)->kill();
    *p = 0; 
}

bool
IcingSource::setup_packet()
{
    uint32_t hdrlen = icing_hdrlen(_nrealms);
    WritablePacket* wp = Packet::make(hdrlen + _payloadlen);
    if (!wp) 
	return false;

    IcingHdr1* hdr = reinterpret_cast<IcingHdr1*>(wp->data());
    hdr->icing_ver = ICING_CURRENT_VERSION;
    hdr->path_len = _nrealms;
    hdr->path_idx = 0;
    set_dummy_initvec(hdr->init_vec);

    IcingHdr2* hdr2 = get_icinghdr2(hdr);
    hdr2->proto = 8;
    hdr2->pkt_len = htons(hdrlen + _payloadlen);

    Path* path = get_path(hdr);

    for (uint32_t i = 0; i < _nrealms; i++) {
	RealmVoq* rv = &path->realmvoqs[i];
	memcpy(rv->realmid.bytes, _realmid[i].toWire()->bytes, 
	       RealmIDWire::SIZE);
	rv->voq = htonl(0x76627265);
    }

    uint8_t* payload = wp->data() + hdrlen + 1;
//    for(uint32_t i = 0; i < _payloadlen; i++)
//	payload[i] = i & 0xFF;
    memset(payload, 0x6f, _payloadlen);

    init_vtags(hdr, _k, _g, _v, _privk);
    hdr->path_idx = 1;
		 
    _packet = wp;
    return true;
}

void
IcingSource::run_timer(Timer* t)
{
    Packet *out_pkt = _packet->clone();
   
    out_pkt->timestamp_anno().set_now();

    output(0).push(out_pkt);

    _timer.reschedule_after_sec(1);

}




CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(IcingSource)
