// Following the pattern of InfiniteSource
#include <click/config.h>
#include "icingcommon.hh"
#include "icingsource2.hh"
#include "initvector.hh"

CLICK_DECLS

IcingSource2::IcingSource2() 
    : _packet(0),
      _timer(this)
{
}

IcingSource2::~IcingSource2()
{
}

int
IcingSource2::configure(Vector<String> &conf, ErrorHandler *errh)
{
	String priv_str;
    if (cp_va_kparse(conf, this, errh,
	     "REALMLIST", cpkP+cpkM, cpString, &_path_str,
	     "POCLIST", cpkP+cpkM, cpString, &_poc_str,
		 "SOURCE_PRIVK", cpkP+cpkM, cpString, &priv_str,
	     cpEnd) < 0)
	return -1;

	cp_icing_privkey(priv_str, &_privk);

    if (!validate_params(errh))
	return false;

    if (!setup_packet()) {
	errh->error("Couldn't set up packet");
	return -1;
    }

    return 0; 
}

bool
IcingSource2::validate_params(ErrorHandler* errh)
{
    if (_path_str.length() % sizeof(RealmVoq)) {
	errh->error("Length of realm list, %d, needs to be a multiple "
		    "of RealmVoq length, %d\n",
		    _path_str.length(),
		    sizeof(RealmVoq));
	return false;
    }

    if (_poc_str.length() % CMACSIZE) {
	errh->error("Length of POC/token/turd list, %d, needs to be a "
		    "multiple of CMAC length, %d\n",
		    _poc_str.length(),
		    CMACSIZE);
	return false;
    }

    uint8_t nrealms, npocs;
    nrealms = _path_str.length() / sizeof(RealmVoq);
    npocs = _poc_str.length() / CMACSIZE;

    uint32_t pathlenmax = (1 << sizeof(((IcingHdr1*)0)->path_len));

    if (nrealms != npocs || nrealms >= pathlenmax) {
	errh->error("Length of realm list, %d, needs to match length of POC "
		    "list, %d, and both must be less than %d\n",
		    nrealms,
		    npocs,
		    pathlenmax);
	return false;
    }

    return true;
}

int
IcingSource2::initialize(ErrorHandler *errh)
{
    _timer.initialize(this);
    _timer.schedule_now();

    return 0;
}

void
IcingSource2::cleanup(CleanupStage)
{
    if (_packet)
	_packet->kill();
    _packet = 0;
}

bool
IcingSource2::setup_packet()
{
    uint8_t nrealms = _path_str.length() / sizeof(RealmVoq);
    uint32_t hdrlen = icing_hdrlen(nrealms);

    WritablePacket* wp = Packet::make(hdrlen + _payloadlen);
    if (!wp) 
	return false;

    IcingHdr1* hdr1 = reinterpret_cast<IcingHdr1*>(wp->data());
    hdr1->icing_ver = ICING_CURRENT_VERSION;
    hdr1->path_len  = nrealms;
    hdr1->path_idx = 0;
    set_dummy_initvec(hdr1->init_vec);

    IcingHdr2* hdr2 = get_icinghdr2(hdr1);
    hdr2->proto = 8;
    hdr2->pkt_len = htons(hdrlen + _payloadlen);

    Path* path = get_path(hdr1);

    memcpy(&path->realmvoqs, _path_str.data(), _path_str.length());

    uint8_t* payload = wp->data() + hdrlen;
    memset(payload, 0x61, _payloadlen);

    init_vtags(hdr1, reinterpret_cast<const POC*>(_poc_str.data()), _privk);
    hdr1->path_idx = 1;

    _packet = wp;
    return true;
}

void
IcingSource2::run_timer(Timer* t)
{
    Packet *p = _packet->clone();
   
    p->timestamp_anno().set_now();

    output(0).push(p);

    _timer.reschedule_after_sec(1);

}


CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(IcingSource2)
