#ifndef _ICINGLINKADDR_HH_
#define _ICINGLINKADDR_HH_

CLICK_DECLS

class IcingLinkAddr { public:

    virtual void *cast(const char* name);
    virtual void annotate(Packet*) const = 0;
    virtual bool parse(Vector<String>& words, Element* context) = 0;
    virtual const char* describe_args() const = 0;

    virtual String unparse() const;
    friend StringAccum& operator<<(StringAccum&, const IcingLinkAddr &);

    virtual ~IcingLinkAddr() {}

protected:
    virtual void dump(StringAccum&) const = 0;
};

class OverlayLinkAddr : public IcingLinkAddr { public:

    virtual void *cast(const char* name);
    void annotate(Packet*) const;
    bool parse(Vector<String>& words, Element* context);
    const char* describe_args() const { return "IP PORT"; }

  OverlayLinkAddr() {};

    OverlayLinkAddr( IPAddress ipaddr, uint16_t udpport )
    : _ipaddr(ipaddr), _udpport(udpport) { } //udpport has to be in network order
private:
    void dump(StringAccum&) const;

    IPAddress _ipaddr;
    uint16_t  _udpport;
};

class L2LinkAddr : public IcingLinkAddr { public:

    virtual void *cast(const char* name);
    void annotate(Packet*) const;
    bool parse(Vector<String>& words, Element* context);
    const char* describe_args() const { return "ETHADDR PORT"; }

    const EtherAddress& ea() const { return _ea; }
    uint32_t outputport() const { return _outputport; }

    void set_ea(EtherAddress& ea) { _ea = ea; }
    void set_outputport(uint32_t p) { _outputport = p; }

    bool operator==(L2LinkAddr& b) {
	return (_ea == b._ea && _outputport == b._outputport);
    }

private:
    void dump(StringAccum&) const;

    EtherAddress  _ea;
    uint32_t	  _outputport;

};

typedef smart_ptr<const IcingLinkAddr> LinkAddr_constp;
typedef smart_ptr<IcingLinkAddr>       LinkAddr_p;
typedef smart_ptr<const OverlayLinkAddr> OverlayLinkAddr_constp;
typedef smart_ptr<OverlayLinkAddr>       OverlayLinkAddr_p;
typedef smart_ptr<const L2LinkAddr> L2LinkAddr_constp;
typedef smart_ptr<L2LinkAddr>       L2LinkAddr_p;

// useful to serve as a "marker", i.e., to take up space in a flow
// table without carrying any information.
class NullLinkAddr : public IcingLinkAddr { public:

    void annotate(Packet*) const {}
    bool parse(Vector<String>& words, Element* context) {return true;}
    const char* describe_args() { return "<no args>"; }

private:
    void dump(StringAccum&) const {}

};

inline String
IcingLinkAddr::unparse() const
{
    StringAccum sa;
    sa << this;
    return sa.take_string();
}

inline StringAccum&
operator<<(StringAccum& sa, const IcingLinkAddr& linkaddr)
{
    linkaddr.dump(sa);
    return sa;
}

inline void
OverlayLinkAddr::dump(StringAccum& sa) const
{
    struct in_addr in;
    in.s_addr = _ipaddr.addr();
    sa.snprintf(32, "(%s,%d)", inet_ntoa(in), ntohs(_udpport));
}

inline void
L2LinkAddr::dump(StringAccum& sa) const
{
    sa << "ethernet: " << _ea << " port: " << _outputport;
}

CLICK_ENDDECLS

#endif
