#define ENABLE_HARDWARE 0

#include <click/config.h>
#include "icingcommon.hh"
#include "realmtable.hh"
#include "icingconfig.hh"

#if ENABLE_HARDWARE
#include "hwinterface.hh"
#endif

CLICK_DECLS

//testing values
//const uint32_t RealmTable::TABLESIZE = 2;
//const uint32_t RealmTable::TABLESIZE = 4;
const uint32_t RealmTable::TABLESIZE = 64;

LinkAddr_p
new_linkaddr(IcingMode mode)
{
    if (mode == MODE_OVERLAY)
	return new OverlayLinkAddr;
    else if (mode == MODE_L3)
	return new L2LinkAddr;
    else
	assert(false);
}

smart_ptr<RealmInfo>
new_realminfo(IcingMode mode)
{
    smart_ptr<RealmInfo> ri = new RealmInfo;
    ri->linkaddr = new_linkaddr(mode);
    return ri;
}

RealmTable::RealmTable() : 
    _mode(MODE_UNSET), _con(NULL)
#if ENABLE_HARDWARE
    , _hwinterface(NULL)
#endif
{

   _realmtable = new ht_entry[TABLESIZE];

   clear();
}

RealmTable::~RealmTable()
{
    clear();

    delete [] _realmtable;
}

int
RealmTable::configure(Vector<String> &conf, ErrorHandler *errh)
{
    int before = errh->nerrors();

    Element *elem;

    // XXX: come back and use cp_register_argtype so that we register an
    // argtype of "neighbor" and don't need to hack in the IcingConfig
    // thingy. worth doing this in the rest of the code that does
    // "manual" parsing. 
    if (conf.size() < 1 || 
        (elem = cp_element(conf[0], this, errh)) == NULL ||
	(_con = static_cast<IcingConfig*>(elem->cast("IcingConfig"))) == NULL) {
	    errh->error("First argument needs to be an IcingConfig object");
	    return  -1;
    }

    _rid = _con->realmid();
    _mode = _con->mode();
#if ENABLE_HARDWARE
    _hwinterface = _con->hwinterface();
#endif

    // start indexing from 1, because of the required config object
    for (int i = 1; i < conf.size(); i++) {

	smart_ptr<RealmInfo> ri = new_realminfo(_mode);

	if (!parse_realm(conf[i], ri, this)) {
	    errh->error("argument %d should be 'PUBKEY 0|1 0|1 %s'", i+1, 
		        ri->linkaddr->describe_args());
	} else {
	    add_realm(ri, true, NULL, errh);
	}
    }

    return (errh->nerrors() != before ? -1 : 0);
}

bool
RealmTable::parse_realm(const String& s, smart_ptr<RealmInfo> ri, RealmTable* t)
{
    Vector<String> words;
    cp_spacevec(s, words);

    if (words.size() < 1 || !confstr2realmid(words[0], &ri->rid))
	return false;

    words.pop_front();

	if(!confstr2entrytype(words[0], ri->entry_kind))
		return false;
    words.pop_front();

#ifdef WITH_SANITY
	// XXX: MW: I don't quite understand this. Why would a different
	// realm's private key be in our realm table? And, even if we
	// had access to that realm's private key, shouldn't we check to
	// make sure that the public key that corresponds to ri->privkey
	// equals ri->rid? Please see my comment below
	/*
	 * AS: This code will only be executed when the sanity check is enabled.
	 * At that point, conf strings will require an additional field: the
	 * private key for each added realm.
	 */

	String privk_str;
	cp_string(words[0], &privk_str);
	words.pop_front();

	cp_icing_privkey(s, &ri->privkey, & ri->rid);
#endif

    set_sharedkey(ri->sharedkey, *(ri->rid.toWire()), t->_con->privkey());

#ifdef WITH_SANITY
	SharedKey k;

	set_sharedkey(k, *(t->_con->realmid().toWire()), ri->privkey);
	if(! memcmp(k, ri->sharedkey, sizeof(SharedKey)))
		click_chatter("Sanity check failed!");
#endif

    return ri->linkaddr->parse(words, t);
}


void
RealmTable::add_handlers()
{
    add_write_handler("add_realm", add_realm_handler, 0);
    add_write_handler("remove_realm", remove_realm_handler, 0);
    add_read_handler("dump",  dump_handler, 0);
    add_write_handler("clear",  clear_handler, 0);
	add_write_handler("remove_non_neigh", remove_non_neigh_handler, 0);
    set_handler("lookup", Handler::OP_READ | Handler::READ_PARAM, 
		 lookup_handler);
}

int
RealmTable::add_realm_handler(const String& conf, 
			    Element* e, void *, ErrorHandler* eh)
{
    RealmTable* t = static_cast<RealmTable*>(e); 

    smart_ptr<RealmInfo> ri = new_realminfo(t->get_mode());

    if (!parse_realm(conf, ri, t)) {
	return -1;
    }

    if (!t->add_realm(ri, true, NULL, eh)) {
	return -1;
    }

    return 0;
}

int
RealmTable::remove_realm_handler(const String& conf,
			       Element* e, void *, ErrorHandler* eh)
{
    RealmTable* t = static_cast<RealmTable*>(e);
    RealmID rid;

    if (!confstr2realmid(conf, &rid))
	return -1;

    if (!t->remove_realm(rid, NULL, eh))
	return -1;

    return 0;
}

String
RealmTable::dump_handler(Element* e, void*)
{
    RealmTable* t = static_cast<RealmTable*>(e);
    return t->dump();
}

String
RealmTable::dump() const
{
    StringAccum sa;

    sa << "Realm Table:\n";
    for (uint32_t i = 0; i < TABLESIZE; i++) 
	if (_realmtable[i].status == ICING_HT_INUSE)
	    sa << "entry " << i << " " << *_realmtable[i].rinfo << "\n";

    return sa.take_string();
}

int
RealmTable::clear_handler(const String&, Element* e, void*, ErrorHandler*)
{
    RealmTable* t = static_cast<RealmTable*>(e);
    t->clear();
#if ENABLE_HARDWARE
	if(t->_hwinterface)
		t->_hwinterface->clear();
#endif

    return 0;
}

int
RealmTable::remove_non_neigh_handler(const String&, Element *e, void *, ErrorHandler *)
{
	click_chatter("Removing non-neighbor entries");
	RealmTable *t = static_cast<RealmTable *>(e);
	t->remove_non_neigh();
	return 0;
}

int
RealmTable::lookup_handler(int paranoia, String& s, 
			      Element* e, const Handler*, ErrorHandler* errh)
{
    assert(paranoia == Handler::OP_READ);
    RealmTable* t = static_cast<RealmTable*>(e);

    RealmID rid;

    if (!confstr2realmid(s, &rid))
	return errh->error("Expected realm identifier. Was given %s\n", 
			   s.c_str());

    smart_ptr<const RealmInfo> ri;
    
    if ( (ri = t->lookup_realm(rid)) == NULL) 
	s = "";
    else
	s = ri->unparse(); 
    
    return 0;
}

smart_ptr<const RealmInfo> 
RealmTable::lookup_realm(const RealmID& realm_id) const
{
    int32_t entry = find_where_stored(realm_id);
    if (entry == -1) return NULL;

    assert(entry >= 0 && entry < static_cast<int32_t>(TABLESIZE));

    return _realmtable[entry].rinfo;
}

int32_t
RealmTable::get_loc(const RealmID& realm_id) 
{
    return (realm_id.hashcode() % TABLESIZE);
}

// XXX: lots of repeated code in the functions below. should encapsulate,
// perhaps with an iterator. also, check out google's spare hash_map.

// should be called only if whole table is full.
// returns -1 if can't find a suitable eviction target.
int32_t
RealmTable::choose_eviction(const RealmID& realm_id) const
{
    uint32_t loc = get_loc(realm_id);

    for (uint32_t i = 1; i <= TABLESIZE; i++) {

	assert(_realmtable[loc].status == ICING_HT_INUSE);

	if (_realmtable[loc].rinfo->entry_kind == RealmInfo::NON_NEIGH)
	    return loc;

	loc = ((loc + i) % TABLESIZE);
    }

    return -1;
}

// returns -1 if table is full
int32_t
RealmTable::find_free_entry(const RealmID& realm_id) const
{
    int32_t loc = get_loc(realm_id);

    for (uint32_t i = 1; i <= TABLESIZE; i++) {

	ht_status status = _realmtable[loc].status;
	if (status == ICING_HT_FREE || status == ICING_HT_TOMB)
	    return loc;

	loc = ((loc + i) % TABLESIZE);
    }

    return -1;

}

// returns -1 if it can't find the realm_id
int32_t
RealmTable::find_where_stored(const RealmID& realm_id) const
{
    int32_t loc = get_loc(realm_id);

    for (uint32_t i = 1; i <= TABLESIZE; i++) {

	ht_status status = _realmtable[loc].status;

	if (status == ICING_HT_FREE) 
	    return -1;

	if (status == ICING_HT_INUSE && 
	    _realmtable[loc].rinfo->rid == realm_id)
	    return loc;

	// quadratic probing
	loc = ((loc + i) % TABLESIZE);
    }

    return -1;
}

void
RealmTable::clear()
{
    for (uint32_t i = 0; i < TABLESIZE; i++) {
	_realmtable[i].status = ICING_HT_FREE;
	_realmtable[i].rinfo = NULL;
    }
}

// returns false if it fails. it fails iff 'overwrite == false' and
// there was a realm with n's id already in the table. 
bool
RealmTable::add_realm(smart_ptr<const RealmInfo> ri,
		      bool overwrite, 
		      smart_ptr<const RealmInfo>* old,
		      ErrorHandler* errh)
{
    // XXX: this is less b.s. than it was, but it still sucks. for one
    // thing, it doesn't periodically rehash, so eventually, all entries
    // will be tombs or occupied, which will lead to bad performance.
    // for another thing, we ought, if possible, to be abstracting these
    // functions with templates and so forth. we'll need to abstract the
    // notion of a "sticky" realm that doesn't get evicted
    int32_t entry;

    if ( (entry = find_where_stored(ri->rid)) == -1) {
	if ( (entry = find_free_entry(ri->rid)) == -1) {
	    if ( (entry = choose_eviction(ri->rid)) == -1) {
		errh->error("Uh-oh, table is full.\n");
		return false;
	    } else if (old) {
		*old = _realmtable[entry].rinfo;
	    }
	}
    } else {

	if (old) {
	    *old = _realmtable[entry].rinfo;
	}

	if (!overwrite) {
	    errh->error("Existing realm with requested id %s\n", 
			ri->rid.unparse().c_str());
	    return false;
	}
    }

    assert(entry >= 0 && entry < static_cast<int32_t>(TABLESIZE));

    _realmtable[entry].status = ICING_HT_INUSE;
    _realmtable[entry].rinfo  = ri;

    //XXX: HERE
#if ENABLE_HARDWARE
    if (_hwinterface && !_hwinterface->set_entry(ri)) 
    {
		/*
		 * If a collision occurs, walk software realm table and add all entries
		 * from it a vector. Call hardware interface
		 */

		Vector<smart_ptr<const RealmInfo> > entries;
		for(uint32_t i = 0; i < TABLESIZE; i++)
			if(_realmtable[i].status == ICING_HT_INUSE)
				entries.push_back(_realmtable[i].rinfo);
		_hwinterface->rehash(entries);
    }
#endif
    return true;
}

bool
RealmTable::remove_realm(const RealmID& rid, 
		         smart_ptr<const RealmInfo>* old,
			 ErrorHandler*)
{
    int32_t entry = find_where_stored(rid);
    if (entry == -1)
	return false;

    assert(entry >= 0 && entry < static_cast<int32_t>(TABLESIZE));

    if (old)
	*old = _realmtable[entry].rinfo;

    _realmtable[entry].status = ICING_HT_TOMB;
    _realmtable[entry].rinfo = NULL;

#if ENABLE_HARDWARE
    if (_hwinterface)
	_hwinterface->clear_entry(rid);
#endif

    return true;
}

void
RealmTable::remove_non_neigh()
{
	for (uint32_t i = 0; i < TABLESIZE; i++)
	{
		if(_realmtable[i].status == ICING_HT_INUSE &&
				_realmtable[i].rinfo->entry_kind == RealmInfo::NON_NEIGH)
		{
			_realmtable[i].status = ICING_HT_TOMB;
#if ENABLE_HARDWARE
			if(_hwinterface)
				_hwinterface->clear_entry(_realmtable[i].rinfo->rid);
#endif
			_realmtable[i].rinfo = NULL;
			
		}
	}
}

String
entrytype2str(const RealmInfo::entry_type &e)
{
	switch(e)
	{
		case RealmInfo::PCI_NEIGH:
			return PCI_NEIGH_STR;
		case RealmInfo::PHYS_NEIGH:
			return PHYS_NEIGH_STR;
		case RealmInfo::NON_NEIGH:
			return NON_NEIGH_STR;
		default:
			assert(false);
	}
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(RealmTable)
