#ifndef _REALMTABLE_HH_
#define _REALMTABLE_HH_

#include <click/element.hh>
#include "icingcommon.hh"
CLICK_DECLS

class IcingConfig;
class HardwareInterface;

// this is info about *other* realms that we know of
struct RealmInfo {
	enum entry_type { NON_NEIGH, PHYS_NEIGH, PCI_NEIGH };
    RealmID rid;
	entry_type entry_kind;
    SharedKey sharedkey;    
    LinkAddr_p linkaddr;
    String unparse() const;
    friend StringAccum& operator<<(StringAccum&, const RealmInfo &);
#ifdef WITH_SANITY
	ecdh_priv privkey;
#endif
};

String entrytype2str(const RealmInfo::entry_type &);
static const char *PCI_NEIGH_STR = "pci_neigh";
static const char *PHYS_NEIGH_STR = "phys_neigh";
static const char *NON_NEIGH_STR = "non_neigh";

class RealmTable : public Element { public:

    RealmTable();
    virtual ~RealmTable();

    const char *class_name() const	{ return "RealmTable"; }
    const char *port_count() const      { return PORTS_0_0;  }

    int configure(Vector<String> &, ErrorHandler *);
    void add_handlers();
    int configure_phase() const        { return CONFIGURE_PHASE_INFO + 2; }

    smart_ptr<const RealmInfo> lookup_realm(const RealmID&) const;
    bool add_realm(smart_ptr<const RealmInfo>,
		   bool overwrite,
		   smart_ptr<const RealmInfo>*,
		   ErrorHandler*);
    bool remove_realm(const RealmID&,
		      smart_ptr<const RealmInfo>*,
		      ErrorHandler*);

	void remove_non_neigh();

    void clear();
    String dump() const;

private:

    static bool parse_realm(const String& s, smart_ptr<RealmInfo>, RealmTable*);

    void set_mode(IcingMode mode)      { _mode = mode; }
    IcingMode get_mode() const         { return _mode; }

    DECLARE_WRITE_HANDLER(add_realm_handler);
    DECLARE_WRITE_HANDLER(remove_realm_handler);
    DECLARE_READ_HANDLER(dump_handler);
    DECLARE_WRITE_HANDLER(clear_handler);
    DECLARE_COMPREHENSIVE_HANDLER(lookup_handler);
	DECLARE_WRITE_HANDLER(remove_non_neigh_handler);

    typedef enum {ICING_HT_FREE, ICING_HT_INUSE, ICING_HT_TOMB} ht_status;
    struct ht_entry {
	ht_status status;
	smart_ptr<const RealmInfo> rinfo;
    };

    int32_t find_free_entry(const RealmID&) const;
    int32_t choose_eviction(const RealmID&) const;
    int32_t find_where_stored(const RealmID&) const;

    static int32_t get_loc(const RealmID&);

    // mirror Jad's hardware table in software. choose entries based on
    // open addressing and quadratic probing
    static const uint32_t TABLESIZE;
    ht_entry* _realmtable;

    IcingMode _mode;
    RealmID _rid;
    IcingConfig* _con;
    HardwareInterface* _hwinterface;
};

inline StringAccum&
operator<<(StringAccum& sa, const RealmInfo& ri)
{
    sa << "realm_id: "    << ri.rid
       << " neighbor: "   << entrytype2str(ri.entry_kind)
       << " shared_key: " << myhexdump(ri.sharedkey, sizeof(ri.sharedkey))
       << " link_addr: "  << *(ri.linkaddr);
    return sa;
}

inline String
RealmInfo::unparse() const
{
    StringAccum sa;
    sa << *(this);
    return sa.take_string();
}

inline bool
confstr2entrytype(const String& str, RealmInfo::entry_type &e)
{
	String unquoted;
	if(cp_string(str, &unquoted))
	{
		if(unquoted == NON_NEIGH_STR)
		{
			e = RealmInfo::NON_NEIGH;
			return true;
		}
		if(unquoted == PHYS_NEIGH_STR)
		{
			e = RealmInfo::PHYS_NEIGH;
			return true;
		}
		if(unquoted == PCI_NEIGH_STR)
		{
			e = RealmInfo::PCI_NEIGH;
			return true;
		}
	}
	return false;
}

CLICK_ENDDECLS
#endif
