#include <click/config.h>
#include "icingcommon.hh"
#include "translationtable.hh"
extern "C"
{
//#include <arpa/inet.h>
#include <math.h>
}

CLICK_DECLS
TranslationTable::TranslationTable()
{
	_tablesize = -1;
	_iptable = NULL;
	_nextIP = -1;
}

TranslationTable::~TranslationTable()
{
	if(_iptable)
		delete [] _iptable;
}

int
TranslationTable::configure(Vector<String> &conf, ErrorHandler *errh)
{
	int before = errh->nerrors();

	if(conf.size() < 1)
	{
			errh->error("Requires an IP prefix");
			return -1;
	}

	if(cp_va_kparse(conf, this, errh,
							"PREFIX", cpkP+cpkM, cpIPPrefix, &_baseaddr, &_mask,
							cpEnd) < 0)
			return -1;


	_baseint = ntohl(_baseaddr.addr());
	_masklen = _mask.mask_to_prefix_len();
	_tablesize = 1 << (32 - _masklen);

	_iptable = new ht_entry[_tablesize];
//	for(int i = 0; i < _tablesize; i++)
//		_iptable[i].info = NULL;

	clear();

	click_chatter("Initialized a table with %d entries.", _tablesize);

	return (errh->nerrors() != before ? -1 : 0);
}

void
TranslationTable::clear()
{
	if(_tablesize < 0)
	{
		assert(! _iptable);
		return;
	}

	assert(_iptable);
	for(int32_t i = 0; i < _tablesize; i++)
	{
		ht_entry &cur = _iptable[i];
		cur.status = ht_entry::FREE; 
		if(cur.info)
		{
			delete cur.info;
			cur.info = NULL;
		}
	}

	_ridtable.clear();
	_nextIP = 1;
}

int32_t
TranslationTable::ipLoc(const IPAddress &addr) const
{
	int32_t loc = ntohl(addr.addr()) - _baseint;
	StringAccum s("loc = ");
	s << loc << "\n";
	assert(loc >= 0 && loc < _tablesize);
	return loc;
}

IPAddress
TranslationTable::locIP(const int32_t loc) const
{
	IPAddress test(htonl(_baseint));
	IPAddress addr(htonl(_baseint + loc));
	assert(addr.matches_prefix(_baseaddr, _mask));
	return addr;
}

void
TranslationTable::add_handlers()
{
	add_write_handler("add_realm", add_realm_handler, 0);
	// needs lookup handler too
}

/*
 * Format is:
 * add_realm domain realmid pathlen path pocs [rpocs]
 * pathlen is simply the 4 byte uint encoded as an 8 byte hex string
 */
int
TranslationTable::add_realm_handler(const String &conf, Element *e, void *, ErrorHandler *eh)
{
	TranslationTable *t = static_cast<TranslationTable *>(e);
	Vector<String> words;
	String unquote;

	cp_spacevec(conf, words);
	TranslationInfo *info = new TranslationInfo();
	info->infotype = TranslationInfo::OUT;

	if(words.size() < 1)
	{
		delete info;
		return -1;
	}
	cp_string(words[0], &info->domain);
	words.pop_front();

	if(words.size() < 1 || !confstr2realmid(words[0], &info->rid))
	{
		delete info;
		return -1;
	}
	words.pop_front();

	if(words.size() < 1)
	{
		delete info;
		return -1;
	}
	cp_string(words[0], &unquote);
	memcpy(&info->pathlen, unquote.data(), sizeof(uint8_t));
	words.pop_front();

	if(words.size() < 1)
	{
		delete info;
		return -1;
	}
	cp_string(words[0], &unquote);
	info->initPathPocs();
	memcpy(info->path, unquote.data(), sizeof(RealmVoq) * info->pathlen);
	words.pop_front();

	if(words.size() < 1)
	{
		delete info;
		return -1;
	}
	cp_string(words[0], &unquote);
	memcpy(info->pocs, unquote.data(), sizeof(POC) * (info->pathlen - 1));
	words.pop_front();

	const IPAddress &ip = t->addEntry(info);

	if(words.size() < 1)
		return 0;

	cp_string(words[0], &unquote);
	info->initRPocs();
	memcpy(info->rpocs, unquote.data(), sizeof(POC) * (info->pathlen - 1));

	return 0;
}

void
TranslationTable::getNextIP(const int32_t loc)
{
	if(loc > 0 && _iptable[loc].status != ht_entry::INUSE && loc < _nextIP)
		_nextIP = loc;
	else
		while(_nextIP < _tablesize && _iptable[_nextIP].status == ht_entry::INUSE)
			_nextIP++;
	if(_nextIP == _tablesize)
	{
		_nextIP = 1;	 // MW: XXX: how come _nextIP isn't set to 0 here?
		while(_iptable[_nextIP].status == ht_entry::INUSE)
		{
			ht_entry &c = _iptable[_nextIP];
			// MW: XXX: this is the only place that GLUE
			// appears, which leads to two questions: (1)
			// why are GLUE entries getting deleted? (2) how
			// are GLUE entries getting created in the first
			// place?
			if(c.info->infotype == TranslationInfo::GLUE)
			{
				delete c.info;
				c.info = NULL;
				c.status = ht_entry::FREE;
			}
			_nextIP++;
		}
		_nextIP = 1;
		while(_nextIP < _tablesize && _iptable[_nextIP].status == ht_entry::INUSE)
			_nextIP++;
		if(_nextIP == _tablesize)
			//XXX:
			// We've run out of space. bad bad bad.
			assert(false);
	}
}

const IPAddress &
TranslationTable::addEntry(TranslationInfo *info)
{
	ht_entry &c = _iptable[_nextIP];
	assert(c.status == ht_entry::FREE);

	c.info = info;
	c.info->remoteIP = locIP(_nextIP);
	c.status = ht_entry::INUSE;

	_ridtable.set(info->rid, info);
	getNextIP();

	return c.info->remoteIP;
}

TranslationInfo *
TranslationTable::lookup(const RealmID &rid)
{
	HashTable_t::iterator i;
	if((i = _ridtable.find(rid)) == _ridtable.end())
		return NULL;
	return i.value();
}

TranslationInfo *
TranslationTable::lookup(const IPAddress &addr)
{
	if(!_iptable)
		return NULL;
	assert(_tablesize >= 0);

	ht_entry &e = _iptable[ipLoc(addr)];
	if(e.status == ht_entry::INUSE)
		return e.info;
	return NULL;
}

void
TranslationTable::deleteEntry(int32_t loc, HashTable_t::iterator &iter)
{
	delete _iptable[loc].info;
	_iptable[loc].info = NULL;
	_iptable[loc].status = ht_entry::FREE;
	_ridtable.erase(iter);
}

bool
TranslationTable::removeEntry(const RealmID &rid)
{
	HashTable_t::iterator i;
	if((i = _ridtable.find(rid)) == _ridtable.end())
		return false;

	smart_ptr<TranslationInfo> info = i.value();
	int32_t loc = ipLoc(info->remoteIP);
	assert(_iptable[loc].status == ht_entry::INUSE);

	deleteEntry(loc, i);
	
	getNextIP(loc);
	return true;
}

bool
TranslationTable::removeEntry(const IPAddress &ip)
{
	int32_t loc = ipLoc(ip);
	if(_iptable[loc].status != ht_entry::INUSE)
		return false;
	smart_ptr<TranslationInfo> info = _iptable[loc].info;

	HashTable_t::iterator i;
	i = _ridtable.find(info->rid);
	assert(i != _ridtable.end());

	deleteEntry(loc, i);

	getNextIP(loc);
	return true;
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(icing)
EXPORT_ELEMENT(TranslationTable)
