#include "util.h"
#include "icinghdr.h"
#include "icingsec.h"
#include "ecdh.h"
#include <cstdio>
#include <cstring>
#include <ctime>
extern "C"
{
#include <arpa/inet.h>
}

ptr<asrv>
init_tcp_server(int fd, const rpc_program &prog)
{
	ptr<axprt_stream> tcpx = axprt_stream::alloc(fd);
	if(!tcpx)
	{
		warn << "Failed to allocate tcp transport object\n";
		return NULL;
	}

	ptr<asrv> srv = asrv::alloc(tcpx, prog);

	return srv;
}

ptr<asrv>
init_udp_server(uint16_t port, const rpc_program &prog)
{
	int32_t dgram_fd = inetsocket(SOCK_DGRAM, port);
	if(dgram_fd < 0)
		fatal("Unable to initialize socket on port %d.\n", port);
	ptr<axprt_dgram> dx = axprt_dgram::alloc(dgram_fd);
	if(!dx)
		fatal << "Failed to allocate dgram xprt\n";
	ptr<asrv> srv = asrv::alloc(dx, prog);
	if(!srv)
		fatal << "Failed to allocate serv object\n";
	return srv;
}

ptr<aclnt>
init_tcp_client(int32_t fd, const rpc_program &prog)
{
	ptr<axprt_stream> tcpx = axprt_stream::alloc(fd);
	if(!tcpx)
	{
		warn << "Failed to allocate tcp transport object\n";
		return NULL;
	}
	ptr<aclnt > clnt = aclnt::alloc(tcpx, prog);

	return clnt;
}

void print_paths(const icp_path_vec &vec, keydb_t *keydb)
{
	for(uint32_t i = 0; i < vec.size(); i++)
		print_path(vec[i], keydb);
}

void
print_path(const icp_path &path, keydb_t *keydb)
{
	warn << "Path:\n";
	for(uint32_t j = 0; j < path.waypoints.size(); j++)
	{
		str rawkey(path.waypoints[j].pubkey.base(), path.waypoints[j].pubkey.size());
		if(keydb)
			warn << (*(*keydb)[rawkey])->id;
		else
		{
			const char *raw = rawkey.cstr();
			for(int i = 0; i < 20; i++)
				warn(" %02X", raw[i] & 0xFF);
		}
		warn(":0x%X\n", ntohl(path.waypoints[j].vq));
	}
	warn << "\n";

	icp_pubkey pattern[2];

	memset(pattern[0].base(), 0, pattern[0].size());
	memset(pattern[1].base(), 0xFF, pattern[0].size());

	for(uint32_t j = 0; j < path.certs.size(); j++)
	{
		const icp_certificate &cert = path.certs[j];
		warn << "Certificate " << j+1 << ". ";

		if(cert.prev.pubkey == pattern[0])
			warn << "*";
		else if (cert.prev.pubkey== pattern[1])
			warn << "N/A";
		else
		{
			str rawkey(cert.prev.pubkey.base(), cert.prev.pubkey.size());
			if(keydb)
				warn << (*(*keydb)[rawkey])->id;
			else
				warn << armor64(rawkey);
			warn(":0x%X", ntohl(cert.prev.vq));
		}

		str rawkey(cert.realm.pubkey.base(), cert.realm.pubkey.size());
		if(keydb)
			warn << " -> " << (*(*keydb)[rawkey])->id;
		else
			warn << " -> " << armor64(rawkey);
		warn(":0x%X -> ", ntohl(cert.realm.vq));

		if(cert.next.pubkey == pattern[0])
			warn << "*";
		else if(cert.next.pubkey == pattern[1])
			warn << "N/A";
		else
		{
			str rawkey(cert.next.pubkey.base(), cert.next.pubkey.size());
			if(keydb)
				warn << (*(*keydb)[rawkey])->id;
			else
				warn << armor64(rawkey);
			warn(":0x%X", ntohl(cert.next.vq));
		}
		warn << "\n";
	}

	for(uint32_t j = 0; j < path.memcerts.size(); j++)
	{
		const icp_memcert &memcert = path.memcerts[j];
		warn << "Memcert " << j+1 << ". ";

		str rawkey(memcert.realm.base(), memcert.realm.size());
		if(keydb)
			warn << (*(*keydb)[rawkey])->id;
		else
			warn << armor64(rawkey);
		warn << " : ";

		rawkey = str(memcert.member.base(), memcert.member.size());
		if(keydb)
			warn << (*(*keydb)[rawkey])->id;
		else
			warn << armor64(rawkey);
		warn << " : ";

		warn("0x%X/%X\n", ntohl(memcert.vq), ntohl(memcert.mask));
	}

#if 0
	if(keydb && path.pocs.size() == path.waypoints.size())
	{
		uint32_t pathlen = path.waypoints.size();

		RealmVoq *realms = New RealmVoq[pathlen];
		bzero(realms, sizeof(RealmVoq) * pathlen);

		for(uint32_t i = 0; i < pathlen; i++)
		{
			const icp_realm &cur = path.waypoints[i];
			realms[i].voq = htonl(cur.vq);

			memcpy(realms[i].realmid.bytes, cur.pubkey.base(), RealmIDWire::SIZE);
		}
		Path *wirepath = reinterpret_cast<Path *>(realms);

		for(uint32_t i = 0; i< pathlen; i++)
		{
			str rawpub(path.waypoints[i].pubkey.base(), path.waypoints[i].pubkey.size());
			ptr<keydb_rec> *rec = (*keydb)[rawpub];
			if(!rec)
				fatal << "no aes key\n";
			
			RealmSecret k;
			memcpy(k.bytes, (*rec)->k.cstr(), RealmSecret::SIZE);
			POC wirepoc;
			poc_from_path(wirepoc, &k, wirepath, pathlen);

			icp_poc localpoc;
			localpoc.set(reinterpret_cast<char *>(wirepoc), RealmSecret::SIZE);

			if(localpoc != path.pocs[i])
			{
				warn << "CMAC FAILED\n";
				return;
			}
		}
		warn << "CMAC verified.\n";
	}
#endif
}

void print_realmid(const icp_pubkey &realmid)
{
	for(int i = 0; i < realmid.size(); i++)
		warn("%02X", realmid[i] & 0xFF);
	warn << "\n";
}

bool
icp_verify(const icp_certificate &cert)
{
	time_t now = time(NULL);
	time_t expire = cert.expiry;

	// XXX: temporarily commented out to allow for testing
#if 0
	if(now >= expire)
		return false;
#endif

	ptr<strbuf> buf = build_buf(cert);

	bool result = icp_sig_verify(*buf, cert.sig, cert.realm.pubkey);

	return result;
	return true;
}

bool
icp_verify(const icp_memcert &memcert)
{
	time_t now = time(NULL);
	time_t expire = memcert.expiry;

	// XXX: temporarily commented out to allow for testing
#if 0
	if(now >= expire)
		return false;
#endif

	ptr<strbuf> buf = build_buf(memcert);

	bool result = icp_sig_verify(*buf, memcert.sig, memcert.realm);

	return result;
	return true;
}

ptr<strbuf>
build_buf(const icp_certificate &cert)
{
	ptr<strbuf> buffer = New refcounted<strbuf>();
	strbuf &buf = *buffer;

	str prev(cert.prev.pubkey.base(), cert.prev.pubkey.size());
	buf.cat(prev);
	buf.cat(cert.prev.vq);

	str realm(cert.realm.pubkey.base(), cert.realm.pubkey.size());
	buf.cat(realm);
	buf.cat(cert.realm.vq);

	str next(cert.next.pubkey.base(), cert.next.pubkey.size());
	buf.cat(next);
	buf.cat(cert.next.vq);

	buf.cat(cert.metric);
	buf.cat(cert.creation);
	buf.cat(cert.expiry);

	return buffer;
}

ptr<strbuf>
build_buf(const icp_memcert &memcert)
{
	ptr<strbuf> buffer = New refcounted<strbuf>();
	strbuf &buf = *buffer;

	buf.cat(str(memcert.realm.base(), memcert.realm.size()));
	buf.cat(str(memcert.member.base(), memcert.member.size()));

	buf.cat(memcert.vq);
	buf.cat(memcert.mask);

	buf.cat(memcert.creation);
	buf.cat(memcert.expiry);

	return buffer;
}

bool
icp_sig_verify(const str &msg, const icp_sig &sig, const icp_pubkey &realm)
{
	return true;

	chopped_ecpk_raw raw;
	memcpy(raw, realm.base(), realm.size());
	ecdh_pub pub(raw);

	bigint e, s;

	// sig is actually e || s, where both e and s are raw bigints (21 bytes each)
	int sigsize = sig.size() >> 1;
	
	// split sig
	str e_str(sig.base(), sigsize);
	str s_str(sig.base() + sigsize, sigsize);
	
	// convert to bigint
	e.setraw(e_str);
	s.setraw(s_str);

	// XXX: uncomment when ready to use sigs properly!
//	return pub.verify((uint8_t *)msg.cstr(), msg.len(), e, s);

	return true;
}

void
icp_sig_gen(icp_sig &sig, const str &msg, ecdh_priv &priv)
{
	bigint e, s;
	priv.sign(e, s, (uint8_t *)msg.cstr(), msg.len());

	strbuf buf;
	buf.cat(e.getraw());
	buf.cat(s.getraw());
	str sbuf;
	sbuf = buf;

	memcpy(sig.base(), sbuf.cstr(), sbuf.len());
}
