/* $Id: dnsfmt.C,v 1.2 2004/04/21 07:06:34 max Exp $ */

/*
 *
 * Copyright (C) 2003 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "dnsfmt.h"
#include "root_servers.h"

#ifdef NEED_DN_COMP_DECL
extern "C" int dn_comp (char *exp_dn, u_char *comp_dn, int length,
			u_char **dnptrs, u_char **lastdnptr);
#endif /* NEED_DN_COMP_DECL */

void
dnsfmt::reset ()
{
  len = sizeof (h);
  bzero (&h, sizeof (h));
  h.qr = 1;
  h.opcode = QUERY;
  h.aa = 1;
  dnptrs[0] = buf;
  dnptrs[1] = NULL;
}

bool
dnsfmt::init (question *qp, const u_char *qbuf, u_int size)
{
  const HEADER *qhp = reinterpret_cast<const HEADER *> (qbuf);
  question q;
  if (!qp)
    qp = &q;
  if (size < sizeof (*qhp)) {
    h.rcode = FORMERR;
    return false;
  }
  h.id = qhp->id;
  h.rd = qhp->rd;
  dnsparse rp (qbuf, size, false);
  if (!rp.qparse (qp)) {
    h.rcode = FORMERR;
    return false;
  }
  //reset ();
  if (!putquestion (*qp)) {
    h.rcode = FORMERR;
    return false;
  }
  if (qhp->opcode != QUERY) {
    h.opcode = qhp->opcode;
    h.rcode = NOTIMP;
    return false;
  }
  return true;
}

bool
dnsfmt::putname (u_char *&cp, const char *name)
{
  if (h.tc)
    return false;
  if (cp >= buf + sizeof (buf)) {
    h.tc = 1;
    return false;
  }
  if (!name) {			// Use the query name
    assert (ntohs (h.qdcount) > 0);
    assert (len > sizeof (HEADER));
    return putshort (cp, 0xc000 | sizeof (HEADER));
  }
  int n = dn_comp (name, cp, buf + sizeof (buf) - cp,
		   dnptrs, dnptrs + num_dnptrs);
  if (n < 0) {
    h.tc = 1;
    return false;
  }
  cp += n;
  return true;
}

bool
dnsfmt::putbytes (u_char *&cp, const void *data, size_t len)
{
  if (cp + len > buf + sizeof (buf)) {
    h.tc = 1;
    return false;
  }
  memcpy (cp, data, len);
  cp += len;
  return true;
}

bool
dnsfmt::putrr (const resrec &rr)
{
  const u_char *const eom = buf + sizeof (buf);
  u_char *cp = buf + len;

  if (!putname (cp, rr.rr_name) || cp + 10 > eom) {
    h.tc = 1;
    return false;
  }

  PUTSHORT (rr.rr_type, cp);
  PUTSHORT (rr.rr_class, cp);
  PUTLONG (rr.rr_ttl, cp);
  u_char *rdlenp = cp;
  cp += 2;

  assert (rr.rr_class == C_IN);
  switch (rr.rr_type) {
  case T_A:
    if (!putbytes (cp, &rr.rr_a, 4))
      return false;
    break;
  case T_NS:
  case T_CNAME:
  case T_DNAME:
  case T_PTR:
    if (!putname (cp, rr.rr_cname))
      return false;
    break;
  case T_MX:
  case T_AFSDB:
    if (!putshort (cp, rr.rr_mx.mx_pref)
	|| !putname (cp, rr.rr_mx.mx_exch))
      return false;
    break;
  case T_TXT:
    {
      int n = strlen (rr.rr_txt);
      assert (n < 256);
      *cp++ = n;
      if (!putbytes (cp, rr.rr_txt, n))
	return false;
      break;
    }
  case T_SOA:
    if (!putname (cp, rr.rr_soa.soa_mname)
	|| !putname (cp, rr.rr_soa.soa_rname)
	|| !putint (cp, rr.rr_soa.soa_serial)
	|| !putint (cp, rr.rr_soa.soa_refresh)
	|| !putint (cp, rr.rr_soa.soa_retry)
	|| !putint (cp, rr.rr_soa.soa_expire)
	|| !putint (cp, rr.rr_soa.soa_minimum))
      return false;
    break;
  default:
    panic ("dnsfmt::putrr: unknown rr_type %d\n", rr.rr_type);
    break;
  };

  putshort (rdlenp, cp - rdlenp - 2);
  len = cp - buf;
  return true;
}

bool
dnsfmt::putquestion (const question &q)
{
  assert (h.qdcount == htons (0)
	  && h.ancount == htons (0)
	  && h.nscount == htons (0)
	  && h.arcount == htons (0));
  u_char *cp = buf + len;
  if (!putname (cp, q.q_name)
      || !putshort (cp, q.q_type)
      || !putshort (cp, q.q_class))
    return false;
  len = cp - buf;
  h.qdcount = htons (ntohs (h.qdcount) + 1);
  return true;
}

void
dnsfmt::sectioncheck (section s)
{
  switch (s) {
  case QD:
    assert (h.qdcount == htons (0)
	    && h.ancount == htons (0)
	    && h.nscount == htons (0)
	    && h.arcount == htons (0));
    break;
  case AN:
    assert (ntohs (h.qdcount) > 0
	    && h.nscount == htons (0)
	    && h.arcount == htons (0));
    break;
  case NS:
    assert (ntohs (h.qdcount) > 0 && h.arcount == htons (0));
    break;
  case AR:
    assert (ntohs (h.qdcount) > 0);
    break;
  default:
    panic ("unknown DNS packet section %d\n", s);
    break;
  }
}

void
dnsfmt::bumpcount (section s)
{
  switch (s) {
  case QD:
    h.qdcount = htons (ntohs (h.qdcount) + 1);
    break;
  case AN:
    h.ancount = htons (ntohs (h.ancount) + 1);
    break;
  case NS:
    h.nscount = htons (ntohs (h.nscount) + 1);
    break;
  case AR:
    h.arcount = htons (ntohs (h.arcount) + 1);
    break;
  default:
    panic ("unknown DNS packet section %d\n", s);
    break;
  }
}

bool
dnsfmt::puta (section s, str domain, in_addr addr, u_int32_t ttl)
{
  assert (s != QD);
  sectioncheck (s);
  u_char *cp = buf + len;

  if (!putname (cp, domain)
      || !putshort (cp, T_A)
      || !putshort (cp, C_IN)
      || !putint (cp, ttl)
      || !putshort (cp, 4)
      || !putbytes (cp, &addr, 4))
    return false;

  bumpcount (s);
  len = cp - buf;
  return true;
}

bool
dnsfmt::putns (section s, str domain, str server, u_int32_t ttl,
	       const in_addr *addrp, u_int32_t addrttl)
{
  assert (s != QD);
  sectioncheck (s);
  u_char *cp = buf + len;

  if (!putname (cp, domain)
      || !putshort (cp, T_NS)
      || !putshort (cp, C_IN)
      || !putint (cp, ttl))
    return false;

  u_char *rdlenp = cp;
  cp += 2;

  if (!putname (cp, server))
    return false;
  if (addrp) {
    arvec.push_back ();
    arvec.back ().name = server;
    arvec.back ().ip4addr = *addrp;
    arvec.back ().ttl = addrttl ? addrttl : ttl;
  }

  bumpcount (s);
  putshort (rdlenp, cp - rdlenp - 2);
  len = cp - buf;
  return true;
}

bool
dnsfmt::putcname (str name, str cname, u_int32_t ttl)
{
  sectioncheck (AN);
  u_char *cp = buf + len;

  if (!putname (cp, name)
      || !putshort (cp, T_CNAME)
      || !putshort (cp, C_IN)
      || !putint (cp, ttl))
    return false;

  u_char *rdlenp = cp;
  cp += 2;

  if (!putname (cp, cname))
    return false;
  bumpcount (AN);
  putshort (rdlenp, cp - rdlenp - 2);
  len = cp - buf;
  return true;
}

bool
dnsfmt::putdname (str name, str dname, u_int32_t ttl)
{
  sectioncheck (AN);
  u_char *cp = buf + len;

  if (!putname (cp, name)
      || !putshort (cp, T_DNAME)
      || !putshort (cp, C_IN)
      || !putint (cp, ttl))
    return false;

  u_char *rdlenp = cp;
  cp += 2;

  if (!putname (cp, dname))
    return false;
  bumpcount (AN);
  putshort (rdlenp, cp - rdlenp - 2);
  len = cp - buf;
  return true;
}


bool
dnsfmt::putadrecs ()
{
  while (!arvec.empty ()) {
    if (!puta (AR, arvec[0].name, arvec[0].ip4addr, arvec[0].ttl))
      return false;
    arvec.pop_front ();
  }
  return true;
}

bool
dnsfmt::putsoa (section s, const nsinfo &ni, str domain)
{
  assert (s != QD);
  sectioncheck (s);

  u_char *cp = buf + len;
  if (domain) {
    if (!putname (cp, domain))
      return false;
  }
  else {
    /* Cook up some appropriate domain name based on the question */
    u_int16_t qdomain = sizeof (HEADER);
    assert (qdomain < len);
    assert (!(buf[qdomain] & 0xc0)); // First name can't be compressed
    if (buf[qdomain])
      qdomain += buf[qdomain] + 1;
    assert (qdomain <= len);
    if (!putshort (cp, qdomain | 0xc000))
      return false;
  }

  if (!putshort (cp, T_SOA)
      || !putshort (cp, C_IN)
      || !putint (cp, ni.ttl))
    return false;

  u_char *rdlenp = cp;
  cp += 2;

  if (!putname (cp, ni.mname)
      || !putname (cp, ni.rname)
      || !putint (cp, 0)	// serial
      || !putint (cp, 1800)	// refresh
      || !putint (cp, 3600)	// retry
      || !putint (cp, 7200)	// expire
      || !putint (cp, ni.ttl))	// minimum
    return false;
		   
  bumpcount (s);
  putshort (rdlenp, cp - rdlenp - 2);
  len = cp - buf;
  return true;
}

bool
dnsfmt::fwd (const str &h)
{
#if 0
  in_addr a = { htonl (rsam.ip4addr) } ;
  if (!putns (NS, h, rsam.name, 3600, &a)) {
    return false;
  }
#endif

  return true;
}

bool
dnsfmt::dontknow ()
{
  switch (0) case 0: case num_root_servers:;
  h.aa = 0;
  static u_int start;
  u_int stop = start++ % num_root_servers;
  u_int i = stop;
#if 1
  do {
    in_addr a = { htonl (root_servers[i].ip4addr) };
    if (!putns (NS, ".", root_servers[i].name, 3600, &a))
      return false;
    if (++i >= num_root_servers)
      i = 0;
  } while (i != stop);
  return putadrecs ();
#else
  resrec rr;
  strcpy (rr.rr_name, ".");
  rr.rr_class = C_IN;
  rr.rr_type = T_NS;
  rr.rr_ttl = 3600;
  do {
    strcpy (rr.rr_ns, root_servers[i].name);
    if (!putresponse (NS, rr))
      return false;
    if (++i >= num_root_servers)
      i = 0;
  } while (i != stop);

  rr.rr_type = T_A;
  do {
    strcpy (rr.rr_name, root_servers[i].name);
    rr.rr_a.s_addr = htonl (root_servers[i].ip4addr);
    if (!putresponse (AR, rr))
      return false;
    if (++i >= num_root_servers)
      i = 0;
  } while (i != stop);
  return true;
#endif
}
