#include "icingclient.h"
#include "click_interface.h"

void
IcingClient::setup(callback<void>::ptr cb)
{
	tcpconnect(_server, _port,
			wrap(this, &IcingClient::connected, cb));
}

void
IcingClient::connected(callback<void>::ptr cb, int fd)
{
	if(fd < 0)
		fatal("Could not open socket. Errno %d - %m\n", errno);
	_fd = fd;
	_client = init_tcp_client(_fd, icing_pathserver_1);
	(*cb)();
}

void
IcingClient::call(str domain, lookupcb cb)
{
	IcingQuestion q;
	q.query = domain;
	q.client = _client;
	q.type = IcingQuestion::HOST;
	call(q, cb);
}

void
IcingClient::call(IcingQuestion q, lookupcb cb)
{
	ptr<ps_reply> reply = New refcounted<ps_reply>();

	ps_getargs args;
	args.host = str(q.query); // copy
	args.partialpaths = _partialpaths;

	switch(q.type)
	{
		case IcingQuestion::HOST:
			q.client->call(
					FIND_PATHTO_HOST,
					&args,
					reply,
					wrap(this, &IcingClient::getReply, reply, q, cb));
			break;
		case IcingQuestion::PS:
			q.client->call(
					FIND_PATHTO_PS,
					&args,
					reply,
					wrap(this, &IcingClient::getReply, reply, q, cb));
			break;
	}
}

void
IcingClient::getReply(ptr<ps_reply> reply, IcingQuestion q, lookupcb cb, clnt_stat err)
{
	switch(err)
	{
		case RPC_SUCCESS:
			handleReply(reply, q, cb);
			break;
		default:
			fatal << "ERROR!\n";
	}
}

void
IcingClient::handleReply(ptr<ps_reply> reply, IcingQuestion q, lookupcb cb)
{
	switch(reply->status)
	{
		case PS_PATH_FOUND:
			handlePaths(reply, cb);
			break;
		case PS_SERVER_FOUND:
			handlePS(reply, q, cb);
			break;
		case PS_UNKNOWN_HOST:
			q.type = IcingQuestion::PS;
			call(q, cb);
			break;
		case PS_INVALID_ARG:
			(*cb)(0);
			break;
		case PS_NO_PATH:
			(*cb)(0);
			break;
		case PS_UNKNOWN_DOMAIN:
			(*cb)(0);
			break;
		default:
			assert(false);
	}
}

void
IcingClient::handlePaths(ptr<ps_reply> reply, lookupcb cb)
{
	icp_path_vec paths = *reply->paths;
	addPaths(paths);

	uint32_t address = click_add_destination(_ttfd, paths[0]);
	(*cb)(address);
}

void
IcingClient::handlePS(ptr<ps_reply> reply, IcingQuestion q, lookupcb cb)
{
	in_addr address;
	icp_path_vec paths = reply->nextps->paths;
	addPaths(paths);

	address.s_addr = click_add_glue(_ttfd, reply->nextps->host, paths[0]);
	tcpconnect(address, _port, wrap(this, &IcingClient::connectedToGluePS, q, cb));
	
	q.type = IcingQuestion::HOST;
}

void
IcingClient::connectedToGluePS(IcingQuestion q, lookupcb cb, int fd)
{
	q.client = init_tcp_client(fd, icing_pathserver_1);
	call(q, cb);
}

void
IcingClient::addPaths(icp_path_vec paths)
{
	icp_pubkey realm = paths[0].waypoints.back().pubkey;

	uint32_t i = 0;
	bool found = false;
	while(!found && i  < _partialpaths.size())
	{
		if(_partialpaths[i].realm == realm)
		{
			rpc_vec_append(_partialpaths[i].paths, paths);
			found = true;
		}
		i++;
	}
	if(!found)
	{
		ps_partialpath ppath;
		ppath.realm = realm;
		ppath.paths = paths;
		_partialpaths.push_back(ppath);
	}
}

void
IcingClient::addPartials(ps_partialpath_vec partials)
{
	for(int i = 0; i < partials.size(); i++)
	{
		bool found = false;
		for(int j = 0; j < _partialpaths.size() && !found; j++)
		{
			if(_partialpaths[j].realm == partials[i].realm)
			{
				rpc_vec_append(_partialpaths[j].paths, partials[i].paths);
				found = true;
			}
		}

		if(!found)
			_partialpaths.push_back(partials[i]);
	}
}
