#include "pathserver.h"
#include "tt.h"

#include <cstdlib>

enum { TIMES_TO_RUN = 1000, NUM_LENS = 4, NUM_DEST = 750, NUM_COMMON = 500};
uint32_t slens[] = {3, 7, 13, 20};

class Test
{
public:
	FILE *out;
	ptr<Pathserver> ps;
	icp_realm client, destinations[NUM_DEST], common[NUM_COMMON];
	ps_partialpath_vec clientPartials;

	void
	random_pubkey(icp_pubkey &r)
	{
		uint32_t *pk = reinterpret_cast<uint32_t *>(r.base());
		for(int i = 0; i < 5; i++)
			pk[i] = rand();
	}

	void
	random_realm(icp_realm &r)
	{
		r.vq = 0;
		random_pubkey(r.pubkey);
	}

	void
	fill_certs(icp_path &path)
	{
		for(uint32_t i = 1; i < path.waypoints.size() - 1; i++)
		{
			icp_certificate c;

			c.prev = path.waypoints[i-1];
			c.realm = path.waypoints[i];
			//c.next = path.waypoints[i+1];
			bzero(c.next.pubkey.base(), c.next.pubkey.size());
			path.certs.push_back(c);

			//c.prev= path.waypoints[i+1];
			bzero(c.prev.pubkey.base(), c.prev.pubkey.size());
			c.realm = path.waypoints[i];
			c.next = path.waypoints[i-1];
			path.certs.push_back(c);
		}
	}

	icp_path
	buildPath(icp_realm &from, icp_realm &to, int len)
	{
		icp_path path;
		path.waypoints.push_back(from);

		for(int i = 0; i < len; i++)
		{
			icp_realm realm;
			random_realm(realm);
			path.waypoints.push_back(realm);
		}
		path.waypoints.push_back(to);

		fill_certs(path);
		return path;
	}

	ps_partialpath
	buildPartial(icp_realm &from, icp_realm &to, int len)
	{
		ps_partialpath partial;
		partial.realm = to.pubkey;
		icp_path path = buildPath(from, to, len);
		partial.paths.push_back(path);

		return partial;
	}

	void
	buildClientPartials(int len)
	{
		for(int c = 0, i = rand() % NUM_COMMON; c < 10; c++, i = (++i % NUM_COMMON))
		{
			ps_partialpath partial = buildPartial(client, common[i], len);
			clientPartials.push_back(partial);
		}
	}

	void
	buildPathserverDB(int len)
	{
		for(int i = 0; i < NUM_DEST; i++)
			for(int j = 0; j < NUM_COMMON; j++)
			{
				ps_partialpath partial;
				icp_pathdb_rec *rec = new icp_pathdb_rec();

				rec->dest = destinations[i].pubkey;

				partial = buildPartial(destinations[i], common[j], len);
				rec->partialpaths.push_back(partial);

				ps->pathdb.insert(rec);
				assert(ps->pathdb[destinations[i].pubkey]);
			}
	}

	void
	initRealms()
	{
		for(int i = 0; i < NUM_DEST; i++)
			random_realm(destinations[i]);
		for(int i = 0; i < NUM_COMMON; i++)
			random_realm(common[i]);
	}

	uint64_t
	runTrial(int len)
	{
		ptr<known_waypoint_db_t> db;
		ptr<icp_path_vec> paths;

		uint64_t begin, delta, sum = 0;

		for(int i = 0; i < TIMES_TO_RUN; i++)
		{
			icp_pubkey destination = destinations[rand() % NUM_DEST].pubkey;
			begin = get_now();
			db = ps->build_known_waypoint_db(clientPartials);
			paths = ps->build_paths(destination, db);
			assert(paths);
			delta = get_now() - begin;
			db->deleteall();

			if(out)
			{
				fprintf(out, "%d, %llu\n", len, delta);
				fflush(out);
			}
			sum += delta;
		}

		return sum;
	}

	void
	runExperiment(int len)
	{
		int clen, slen;
		uint64_t sum;

		clen = len / 2;
		slen = len /2;
		slen += len % 2;

		ps = New refcounted<Pathserver>();
		clientPartials.clear();

		printf("Generating path server database ... ");
		fflush(stdout);
		buildPathserverDB(slen);
		printf("DONE\n");
		fflush(stdout);

		printf("Generating client partials ... ");
		fflush(stdout);
		buildClientPartials(clen);
		printf("DONE\n\n");
		fflush(stdout);
		buildClientPartials(clen);

		printf("Running experiment of length %d ... ", len);
		fflush(stdout);
		sum = runTrial(len);
		printf("AVG = %f\n", (sum * 1.0)/TIMES_TO_RUN);

	}

	Test(FILE *o) : out(o)
	{
		srand((unsigned)time(NULL));
		ps = NULL;

		random_realm(client);
		initRealms();
		
		if(out)
		{
			fprintf(out, "# Pathserver, build_path\n");
			fprintf(out, "# path length, proc. time (us)\n");
		}
	}
};

int
main(int argc, char **argv)
{
	FILE *file = NULL;
	if(argc > 1)
		file = fopen(argv[1], "w+");
	Test t(file);

	//for(uint32_t i = 0; i < NUM_LENS; i++)
	
	for(int i = 0; i < NUM_LENS; i++)
		t.runExperiment(slens[i]);
}
