#include "pathserver.h"
#include "tt.h"

#include <cstdlib>

enum { TIMES_TO_RUN = 20, NUM_LENS = 5};
uint32_t slens[] = {1, 5, 10, 15, 20};

class Test
{
public:
	FILE *out;
	ptr<Pathserver> ps;
	icp_realm src, dst, intersect;
	vec<ps_partialpath_vec> pos_partials;

	ps_partialpath cl_partial;
	icp_path cl_path, srv_path;

	void
	random_pubkey(icp_pubkey &r)
	{
		uint32_t *pk = reinterpret_cast<uint32_t *>(r.base());
		for(int i = 0; i < 5; i++)
			pk[i] = rand();
	}

	void
	random_realm(icp_realm &r)
	{
		r.vq = 0;
		random_pubkey(r.pubkey);
	}

	void
	fill_certs(icp_path &path)
	{
		for(uint32_t i = 1; i < path.waypoints.size() - 1; i++)
		{
			icp_certificate c;

			c.prev = path.waypoints[i-1];
			c.realm = path.waypoints[i];
			c.next = path.waypoints[i+1];
			path.certs.push_back(c);

			c.prev= path.waypoints[i+1];
			c.realm = path.waypoints[i];
			c.next = path.waypoints[i-1];
			path.certs.push_back(c);
		}
	}

	void
	build_real_path(icp_path &cl_path, icp_path &srv_path)
	{
		cl_path.waypoints.push_back(src);
		for(uint32_t i = 0; i < 2; i++)
		{
			icp_realm r;
			random_realm(r);
			cl_path.waypoints.push_back(r);
		}
		cl_path.waypoints.push_back(intersect);
		fill_certs(cl_path);

		srv_path.waypoints.push_back(dst);
		for(uint32_t i = 0; i < 2; i++)
		{
			icp_realm r;
			random_realm(r);
			srv_path.waypoints.push_back(r);
		}
		srv_path.waypoints.push_back(intersect);
		fill_certs(srv_path);

		uint32_t last = srv_path.waypoints.size() - 1;
		icp_realm r;
		bzero(r.pubkey.base(), 20);
		r.vq = 0;

		icp_certificate c;

		c.prev = srv_path.waypoints[last - 1];
		c.realm = srv_path.waypoints[last];
		c.next = r;
		srv_path.certs.push_back(c);

		c.next = srv_path.waypoints[last - 1];
		c.prev = r;
		srv_path.certs.push_back(c);
	}

	void
	fill_pathdb(icp_path &srv_path, uint32_t len)
	{
		ps->pathdb.deleteall();
		icp_pathdb_rec *rec = New icp_pathdb_rec();
		rec->dest = dst.pubkey;
		ps_partialpath ppath;
		ppath.realm = intersect.pubkey;
		ppath.paths.push_back(srv_path);
		rec->partialpaths.push_back(ppath);

		ps->pathdb.insert(rec);

		for(uint32_t i = 0; i < len; i++)
		{
			rec = New icp_pathdb_rec();
			random_pubkey(rec->dest);
			ps->pathdb.insert(rec);
		}
	}

	void
	fill_partials()
	{
		for(uint32_t i = 0; i < 1000; i++)
		{
			ps_partialpath_vec vec;
			for(uint32_t i = 0; i < (unsigned)(rand()) % 20; i++)
			{
				ps_partialpath p;
				random_pubkey(p.realm);
				vec.push_back(p);
			}
			vec.push_back(cl_partial);
			pos_partials.push_back(vec);
		}
	}

#if 0
	void
	reset_partials()
	{
		cl_partials.clear();
		ps_partialpath ppath;
		ppath.realm = intersect.pubkey;
		ppath.paths.push_back(cl_path);
		cl_partials.push_back(ppath);
	}
#endif

	void
	run_experiment(uint32_t len)
	{
		printf("Filling pathdb (%u) ... ", len);
		fflush(stdout);
		fill_pathdb(srv_path, len);
		printf("Running experiment (%u) ... ", len);
		fflush(stdout);
		double total = 0;
		for(uint32_t i = 0; i < TIMES_TO_RUN; i++)
		{

//			reset_partials();
//			fill_partials(size);

			ptr<known_waypoint_db_t> db;
			ptr<icp_path_vec> paths;

			double count = 0;
			uint64_t end = get_now() + 30000000;
			while(get_now() < end)
			{
				db = ps->build_known_waypoint_db(pos_partials[rand() % 1000]);
				paths = ps->build_paths(dst.pubkey, db);
				count++;
			}

			count /= 30.0;

			total += count;

			if(out)
			{
				fprintf(out, "%u, %f\n",len, count);
				fflush(out);
			}
			assert(paths);
			assert(paths->size() > 0);
		}
		double avg = total / TIMES_TO_RUN;
		printf("avg: %f\n", avg);
		fflush(stdout);
	}

	Test(FILE *o) : out(o)
	{
		srand((unsigned)time(NULL));
		ps = New refcounted<Pathserver>();

		random_realm(src);
		random_realm(dst);
		random_realm(intersect);

		build_real_path(cl_path, srv_path);
		cl_partial.realm = intersect.pubkey;
		cl_partial.paths.push_back(cl_path);
//		fill_pathdb(srv_path);
		printf("Filling partial path vecs ... ");
		fflush(stdout);
		fill_partials();
		printf("DONE\n");
		fflush(stdout);
		
		if(out)
		{
			fprintf(out, "# Pathserver, build_path\n");
			fprintf(out, "# no. client paths, throughput (req/s)\n");
		}
	}
};

int
main(int argc, char **argv)
{
	FILE *file = NULL;
	if(argc > 1)
		file = fopen(argv[1], "w+");
	Test t(file);

	//for(uint32_t i = 0; i < NUM_LENS; i++)
	for(uint32_t i = 0; i < NUM_LENS; i++)
		t.run_experiment(slens[i]);
}
