#include "psclient.h"
#include <cstring>
#include <cstdio>
extern "C"
{
#include <arpa/inet.h>
}

Path *
convert_to_path(icp_path &path)
{
	RealmVoq *realms = New RealmVoq[path.waypoints.size()];

	for(uint32_t i = 0; i < path.waypoints.size(); i++)
	{
		memcpy(realms[i].realmid.bytes, path.waypoints[i].pubkey.base(), RealmIDWire::SIZE);
		realms[i].voq = htonl(path.waypoints[i].vq);
	}
	Path *wirepath = reinterpret_cast<Path *>(realms);
	return wirepath;
} 

POC *
get_pocs(icp_path &path)
{
	uint32_t poclen = path.waypoints.size() -1;
	POC *pocs = New POC[poclen];
	bzero(pocs, sizeof(POC) * poclen);

	for(uint32_t i = 0; i < poclen; i++)
		memcpy(pocs[i], path.pocs[i].base(), sizeof(POC));
	return pocs;
}

void
build_packet(icp_path &path, uint8_t **buf, uint32_t *size)
{
	uint32_t pathlen = path.waypoints.size();
	uint32_t hdrlen = icing_hdrlen(pathlen);

	*size = hdrlen + ((pathlen-1) * sizeof(POC));
	*buf = new uint8_t[*size];

	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(*buf);
	hdr1->icing_ver = ICING_CURRENT_VERSION;
	hdr1->path_len = static_cast<uint8_t>(pathlen);
	hdr1->path_idx = 0;
	for (uint32_t i = 0; i < IV_SIZE; i++)
		hdr1->init_vec[i] = (i+1) & 0xff;

	IcingHdr2 *hdr2 = get_icinghdr2(hdr1);
	hdr2->proto = 8;
	hdr2->pkt_len = htons(*size);

	Path *wirepath = get_path(hdr1);
	Path *temppath = convert_to_path(path);
	memcpy(wirepath, temppath, pathlen * sizeof(RealmVoq));
	delete[] temppath;

	uint8_t *payload = *buf + hdrlen;
	for(uint32_t i = pathlen - 1; i < path.pocs.size(); i++)
	{
		memcpy(payload, path.pocs[i].base(), sizeof(POC));
		payload += sizeof(POC);
	}

	POC *pocs = get_pocs(path);

	// XXX:
	ecdh_priv p;
	init_vtags(hdr1, pocs, p);
	delete[] pocs;

	hdr1->path_idx = 1;
}

uint64_t
receive_response()
{
	uint8_t buf[1500];
	bzero(buf, sizeof(buf));

	int32_t packetsize;
	if((packetsize = recvfrom(_recvfd, buf, sizeof(buf), 0, NULL, NULL)) < 0)
		fatal("ERROR: %d - %m", errno);
	uint64_t now = get_now();
	pingt.end();
	e2et.end();

	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(buf);
	IcingHdr2 *hdr2 = get_icinghdr2(hdr1);
	if(ntohs(hdr2->pkt_len) != packetsize)
		fatal("Packet size mismatch: %d vs. %d\n", ntohs(hdr2->pkt_len), packetsize);
	return now;
}

uint64_t
send_packet(icp_path path)
{
	uint8_t *packet;
	uint32_t packetsize;
	build_packet(path, &packet, &packetsize);

	size_t s = sendto(_fd, packet, packetsize, 0, 
			reinterpret_cast<const sockaddr *>(&_addr), sizeof(sockaddr_in));
	uint64_t now = get_now();
	pingt.begin();

	if(s < packetsize)
		fatal("could not send, %d - %m", errno);
	delete[] packet;
	return now;
}

void
handle_result(ptr<ps_reply> reply, uint64_t reqt)
{
	switch(reply->status)
	{
		case PS_PATH_FOUND:
		{
			//warn << "Path(s) found: \n\n";
			icp_path_vec &paths = *reply->paths;
			//print_paths(paths, &keydb);
			uint64_t sendt = send_packet(paths[0]);
			uint64_t recvt = receive_response();
			
			fprintf(out, "%lld, %lld\n", recvt - sendt, recvt - reqt);

			if(--times_to_send > 0)
			{
				sleep(delay);
				make_call();
			}
			else
			{
				fprintf(out, "\n\n");
				fprintf(out, "Min Ping:\t%f\n", pingt.min_us());
				fprintf(out, "Avg Ping:\t%f\n", pingt.avg_us());
				fprintf(out, "Min Path Req:\t%f\n", patht.min_us());
				fprintf(out, "Avg Path Req:\t%f\n", patht.avg_us());
				fprintf(out, "Min E2E:\t%f\n", e2et.min_us());
				fprintf(out, "Avg E2E:\t%f\n", e2et.avg_us());
				fclose(out);
				exit(0);
			}
			break;
		}
		case PS_SERVER_FOUND:
		{
			ps_nextps &res = *reply->nextps;

			warn << "Pathserver: " << res.host << ":\n\n";

			//print_paths(res.paths, &keydb);
			break;
		}
		case PS_UNKNOWN_HOST:
			warn << "Unknown host\n";
			break;
		case PS_INVALID_ARG:
			warn << "Invalid arguments\n";
			break;
		case PS_NO_PATH:
			warn << "No path\n";
			break;
		case PS_UNKNOWN_DOMAIN:
			warn << "Unknown domain\n";
			break;
		default:
			warn << "Unknown error\n";
	}
}

void
get_reply(ptr<ps_reply> reply, uint64_t reqt, clnt_stat err)
{
	patht.end();
	uint64_t endt = get_now();
	fprintf(out, "%lld,", endt - reqt);
	switch(err)
	{
		case RPC_SUCCESS:
			handle_result(reply, reqt);
			break;
		default:
			fatal << "ERROR\n";
	}
}

void
make_call()
{
	ptr<ps_reply> reply = New refcounted<ps_reply>();

	ps_getargs args;
	args.host = str(hostname);
	args.partialpaths = partialpaths;
	uint64_t reqt = get_now();
	e2et.begin();
	patht.begin();

	if(ps_lookup)
	{
		clnt->call(
				FIND_PATHTO_PS,
				&args,
				reply,
				wrap(get_reply, reply, reqt));
	}
	else
	{
		clnt->call(
				FIND_PATHTO_HOST,
				&args,
				reply,
				wrap(get_reply, reply, reqt));
	}

}

void
connected(int32_t fd)
{
	if(fd < 0)
		fatal("Could not open socket.\nerrno: %d - %m\n", errno);
	clnt = init_tcp_client(fd, icing_pathserver_1);
	make_call();
}

void
add_partial(keydb_t *keydb, vec<str> args, str loc, bool *err)
{
	if(args.size() != 2)
	{
		warn << loc << ": Usage: partial <filename>\n";
	}
	
	str end;
	ps_partialpath partial;

	conftab ct;
	ct.ignore("start");
	ct.add("end",		&end);
	ct.add("path",		wrap(add_path, keydb, &partial));
	ct.run(args[1]);

	ptr<keydb_rec> *rec;
	rec = (*keydb)[end];
	if(!rec)
		return;

	//ptr<rabin_priv> priv = (*rec)->priv;
	partial.realm = (*rec)->key; //priv->n.getraw();
	partialpaths.push_back(partial);
}

void
config(str filename, keydb_t *keydb)
{
	str daemon;
	int32_t daemon_port;

	conftab ct;
	ct.add("server",	&server);
	ct.add("port",  	&port, 0, 65535);
	ct.add("key",		wrap(&add_key, keydb));
	ct.add("partial", 	wrap(&add_partial, keydb));
	ct.add("payload_len",	&payload_len, 0, 65535);
	ct.add("delay",		&delay, 0, 65535);
	ct.add("times_to_send", &times_to_send, 0, 65535);
	ct.add("daemon",	&daemon);
	ct.add("daemon-port", &daemon_port, 0 , 65535);
	ct.add("hostname", &hostname);
	ct.run(filename);

	_addr.sin_family = AF_INET;
	_addr.sin_port = htons(daemon_port);
	_addr.sin_addr.s_addr = inet_addr(daemon.cstr());
	bzero(_addr.sin_zero, sizeof(_addr.sin_zero));
}

int32_t
main(int32_t argc, char **argv)
{
	if(argc < 2)
	{
		warn << "Usage: psclient <path to configfile>\n";
		return 1;
	}

	out = fopen(argv[1], "w+");

	filename = "config/client.config";

	config(filename, &keydb);

	//print_path(partialpaths[0].paths[0], &keydb);
	if((_fd = inetsocket(SOCK_DGRAM)) < 0)
		fatal << "Could not open UDP socket.\n";
	if((_recvfd = inetsocket(SOCK_DGRAM, 0x2345, ntohl(inet_addr("192.168.39.1")))) < 0)
		fatal("Could not open UDP socket to listen: %d - %m", errno);

	tcpconnect(server, port, wrap(&connected));
	amain();
	return 0;
}
