#include "pathserver.h"
#include "verify_path.h"
#include "icp_config.h"
#include "parseopt.h"
#include "icinghdr.h"
#include "icingsec.h"
#include <cstdio>
#include <cstring>
extern "C"
{
#include <arpa/inet.h>
}

void
Pathserver::acpt_con()
{
	int32_t acpt_fd = accept(fd, NULL, NULL);
	if(acpt_fd < 0)
	{
		warn("Could not accept request:\n%m\n");
		return;
	}

	ptr<axprt_stream> sx = axprt_stream::alloc(acpt_fd);
	if(!sx)
	{
		warn << "Unable to allocate a xprt object\n";
		return;
	}

	ptr<asrv> srv = asrv::alloc(sx, icing_pathserver_1);
	srv->setcb(wrap(this, &Pathserver::dispatch, srv));
}

void
Pathserver::handle_find_tops(svccb *rpc)
{
	ps_reply reply;
	ps_getargs *args = rpc->getarg<ps_getargs>();
	if(!args)
	{
		reply.set_status(PS_INVALID_ARG);
		rpc->reply(&reply);
		return;
	}

	psdb_rec* ps_rec = get_known_host(args->host);
	hostname_rec *host_rec = NULL;
	if(!ps_rec)
	{
		if(root)
		{
			reply.set_status(PS_UNKNOWN_DOMAIN);
			rpc->reply(&reply);
			return;
		}
		else
			host_rec = hostdb[root_server];
	}
	else
		host_rec = hostdb[ps_rec->server];
	
	if(!host_rec)
	{
		reply.set_status(PS_UNKNOWN_DOMAIN);
		rpc->reply(&reply);
		return;
	}

	ps_partialpath_vec &cl_partialpaths = args->partialpaths;
	ptr<icp_path_vec> paths = get_paths(host_rec->realm, cl_partialpaths); 

	if(!paths || paths->size() == 0)
	{
		reply.set_status(PS_NO_PATH);
	}
	else
	{
		reply.set_status(PS_SERVER_FOUND);

		ps_nextps nextps;
		nextps.host = host_rec->name;
		nextps.paths = *paths;

		*(reply.nextps) = nextps; //XXX:
	}
	rpc->reply(&reply);
}

void 
Pathserver::handle_find_tohost(svccb *rpc)
{
	ps_reply reply;

	ps_getargs *args = rpc->getarg<ps_getargs>();
	if(!args)
	{
		reply.set_status(PS_INVALID_ARG);
		rpc->reply(&reply);
		return;
	}

	hostname_rec *host_rec = hostdb[args->host];
	if(!host_rec)
	{
		reply.set_status(PS_UNKNOWN_HOST);
		rpc->reply(&reply);
		return;
	}

	ptr<icp_path_vec> paths = get_paths(host_rec->realm, args->partialpaths);
	if(!paths || paths->size() == 0)
	{
		reply.set_status(PS_NO_PATH);
		rpc->reply(&reply);
		return;
	}

	reply.set_status(PS_PATH_FOUND);
	*(reply.paths) = *paths; //XXX
	rpc->reply(&reply);
}

void
Pathserver::dispatch(ptr<asrv> srv, svccb *rpc)
{
	if(!rpc)
	{
		srv->setcb(NULL);
		return;
	}

	switch(rpc->proc())
	{
		case FIND_PATHTO_PS:
			handle_find_tops(rpc);
			break;
		case FIND_PATHTO_HOST:
			handle_find_tohost(rpc);
			break;
		default:
			fatal("Procedure %d not handled!\n", rpc->proc());
	}
}

/*
 * Try to find a host which we know about:
 * given master.cs.utexas.edu and a vector holding utexas.edu
 * it should return utexas.edu
 * if cs.utexas.edu was in it, return cs.utexas.edu
*/
psdb_rec *
Pathserver::get_known_host(const icp_hostname &host)
{
	char *ph = New char[host.len() + 1], *todel = ph;
	strncpy(ph, host.cstr(), host.len() + 1);

	psdb_rec *rec;
	rec = psdb[icp_hostname(ph)];
	if(rec)
	{
		delete[] todel;
		return rec;
	}

	while(true)
	{
		strsep(&ph, ".");
		if(!ph)
			break;
		rec = psdb[icp_hostname(ph)];
		if(rec)
		{
			delete[] todel;
			return rec;
		}
	}

	delete[] todel;
	return NULL;
}

ptr<icp_path_vec>
Pathserver::get_paths(const icp_pubkey &realm, const ps_partialpath_vec &cl_partialpaths)
{
	// (1) build known waypoint db
	ptr<known_waypoint_db_t> waypoint_db = build_known_waypoint_db(cl_partialpaths);
	if(!waypoint_db)
	{
#ifdef DEBUG
		warn << "Empty waypoint_db\n";
#endif
		return NULL;
	}
	// (2) find common waypoints
	// (3) build paths
	// (4) check paths
	ptr<icp_path_vec> complete_paths = build_paths(realm, waypoint_db);
	return complete_paths;
}

ptr<known_waypoint_db_t>
Pathserver::build_known_waypoint_db(const ps_partialpath_vec &partialpaths)
{
	if(partialpaths.size() == 0)
		return NULL;

	ptr<known_waypoint_db_t> tempdb = New refcounted<known_waypoint_db_t>();
	for(uint32_t i = 0; i < partialpaths.size(); i++)
	{
		const ps_partialpath &path_rec = partialpaths[i];
		known_waypoint_rec *rec = New known_waypoint_rec(path_rec.realm, &(path_rec.paths));
		tempdb->insert(rec);
	}

	return tempdb;
}

void
Pathserver::get_pocs(icp_path &path)
{
	uint32_t pathlen = path.waypoints.size();
	
	RealmVoq *realms = New RealmVoq[pathlen];
	bzero(realms, sizeof(RealmVoq) * pathlen);

	// Pack the path
	for(uint32_t i = 0; i < pathlen; i++)
	{
		icp_realm &cur = path.waypoints[i];
		realms[i].voq = cur.vq;

		memcpy(realms[i].realmid.bytes, cur.pubkey.base(), RealmIDWire::SIZE);
	}

	Path *wirepath = reinterpret_cast<Path *>(realms);

	// XXX: may need to derive the poc key correctly. (maybe)
	// Now, we need to get the AES key for each realm and MAC the path
	// each time.
	for(uint32_t i = 1; i < pathlen; i++)
	{
		str rawpubkey(path.waypoints[i].pubkey.base(), path.waypoints[i].pubkey.size());
		ptr<keydb_rec> *rec = keydb[rawpubkey];
		if(! rec)
		{
//			warn("NO AES KEY (k_i) for a realm on the path\n");
			return;
		}
		//str &kstr = (*rec)->k;

		assert(*rec);
		assert( (*rec)->k.base() );
		RealmSecret k, derived;
		bzero(k.bytes, RealmSecret::SIZE);
		memcpy(k.bytes, (*rec)->k.base(), RealmSecret::SIZE);

		//XXX: in future, do NOT hard code these with magic values
		derive_poc_key(k, 1, 16, path.waypoints[i].vq, derived);

		POC wirepoc;
		poc_from_path(wirepoc, &derived, wirepath, pathlen);

		icp_poc localpoc;
		memcpy(localpoc.base(), wirepoc, sizeof(POC));
		path.pocs.push_back(localpoc);
	}

	for(int32_t i = pathlen - 1, j = 0; i >= 0; i--, j++)
	{
		icp_realm &cur = path.waypoints[i];
		realms[j].voq = cur.vq;
		memcpy(realms[j].realmid.bytes, cur.pubkey.base(), RealmIDWire::SIZE);
	}
	for(int32_t i = pathlen - 2; i >= 0; i--)
	{
		str rawpubkey(path.waypoints[i].pubkey.base(), path.waypoints[i].pubkey.size());
		ptr<keydb_rec> *rec = keydb[rawpubkey];
		if(! rec)
			fatal("NO AES KEY (k_i) for a realm on the path\n");
		//str &kstr = (*rec)->k;

		RealmSecret k, derived;
		bzero(k.bytes, RealmSecret::SIZE);
		memcpy(k.bytes, (*rec)->k.base(), RealmSecret::SIZE);

		derive_poc_key(k, 1, 16, path.waypoints[i].vq, derived);

		POC wirepoc;
		poc_from_path(wirepoc, &k, wirepath, pathlen);

		icp_poc localpoc;
		memcpy(localpoc.base(), wirepoc, sizeof(POC));
		path.pocs.push_back(localpoc);
	}

	delete[] realms;
}

ptr<icp_path_vec>
Pathserver::build_paths(const icp_pubkey &realm, ptr<known_waypoint_db_t> waypoint_db)
{
	icp_pathdb_rec *pathdb_rec = pathdb[realm];
	if(!pathdb_rec)
	{
#ifdef DEBUG
		warn << "No partials to this realm\n";
#endif
		return NULL;
	}

	ptr<icp_path_vec> complete_paths = New refcounted<icp_path_vec>();
	ps_partialpath_vec &partials = pathdb_rec->partialpaths;
	for(uint32_t i = 0; i < partials.size(); i++)
	{
		icp_pubkey &current_waypoint = partials[i].realm;
		known_waypoint_rec *known = (*waypoint_db)[current_waypoint];
		if(known)
		{
			const icp_path_vec &cl_path = *(known->paths);
			const icp_path_vec &srv_path = partials[i].paths;

			for(uint32_t j = 0; j < cl_path.size(); j++)
				for(uint32_t k = 0; k < srv_path.size(); k++)
				{
#ifdef DEBUG
					warn << "client:\n";
					print_path(cl_path[j]);
					warn << "server:\n";
					print_path(srv_path[k]);
#endif
					ptr<icp_path> new_path = generate_path(cl_path[j], srv_path[k]);
#ifdef DEBUG
					warn << "new:\n";
					print_path(*new_path, &keydb);
#endif
					if(new_path && icp_verify_path(*new_path))
					{
						get_pocs(*new_path);
//						assert(new_path->pocs.size());
						complete_paths->push_back(*new_path);
					}
				}
		}
	}

	return complete_paths;
}

ptr<icp_path>
Pathserver::generate_path(const icp_path &cl_path, const icp_path &srv_path)
{
	ptr<icp_path> path = New refcounted<icp_path>();
	// (1) build path
	path->waypoints = cl_path.waypoints;
	rpc_vec_append_reverse(path->waypoints, srv_path.waypoints, 1);
	// (2) append certs
	path->certs = cl_path.certs;
	rpc_vec_append_reverse(path->certs, srv_path.certs);

	path->memcerts = cl_path.memcerts;
	rpc_vec_append_reverse(path->memcerts, srv_path.memcerts);
	return path;
}

#if 0
void
Pathserver::init_from_file(const char *file, const char *key_loc)
{
	FILE *fp = fopen(file, "r");
	if(!fp)
		fatal("Failed to open file %s\nerrno: %d - %m", file, errno);

	char *line = New char[MAX_LINE_LENGTH];
	
	if(fgets(line, MAX_LINE_LENGTH, fp) && line)
	{
		strip_newline(line);
		ps_realm = pub_from_file(line, key_loc);
	}

	if(fgets(line, MAX_LINE_LENGTH, fp) && line)
	{
		strip_newline(line);
		root = (line[0] == '1');
	}

	if(!root && fgets(line, MAX_LINE_LENGTH,fp) && line)
	{
		strip_newline(line);
		root_server = str(line);
	}

	while(fgets(line, MAX_LINE_LENGTH, fp) && line && strncmp("END: DOMAINS", line, 12) != 0)
	{
		strip_newline(line);
		ptr<psdb_rec> rec = New refcounted<psdb_rec>();

		char *pointstr = strsep(&line, ",");
		rec->domain = str(pointstr);

		pointstr = strsep(&line, ",");
		rec->server = str(pointstr);

		psdb.insert(rec);
		delete line;
		line = New char[MAX_LINE_LENGTH];
	}

	while(fgets(line, MAX_LINE_LENGTH, fp) && line && strncmp("END: HOSTS", line, 10) != 0)
	{
		strip_newline(line);

		char *name = strsep(&line, ",");

		char *realm = strsep(&line, ",");
		icp_pubkey key;
		key = pub_from_file(realm,key_loc);
		
		ptr<hostname_rec> hnrec = New refcounted<hostname_rec>(str(name), key);
		hostdb.insert(hnrec);
		delete line;
		line = New char[MAX_LINE_LENGTH];

	}

	while(fgets(line, MAX_LINE_LENGTH, fp) && line)
	{
		icp_pubkey dest;
		ps_partialpath partialpath;

		strip_newline(line);
		dest = pub_from_file(line, key_loc);
		fgets(line, MAX_LINE_LENGTH, fp);
		strip_newline(line);
		partialpath.realm = pub_from_file(line, key_loc);

		ptr<icp_path> path;
		do
		{
			path = build_path_from_file(fp, key_loc);
			if(path)
				partialpath.paths.push_back(*path);
		}
		while(path);

		if(pathdb[dest])
			pathdb[dest]->partialpaths.push_back(partialpath);
		else
		{
			ptr<icp_pathdb_rec> rec = New refcounted<icp_pathdb_rec>();
			rec->dest = dest;
			rec->partialpaths.push_back(partialpath);
			pathdb.insert(rec);
		}
	}
	delete line;
	fclose(fp);
}
#endif

void
Pathserver::add_domain_ps(vec<str> args, str loc, bool *err)
{
	if(args.size() != 3)
	{
		warn << loc << ": Usage: domain_ps domain ps_hostname\n";
		*err = true;
		return;
	}

	psdb_rec *rec = New psdb_rec();
	rec->domain = args[1];
	rec->server = args[2];
	psdb.insert(rec);
}

void
Pathserver::add_host(vec<str> args, str loc, bool *err)
{
	if(args.size() != 3)
	{
		warn << loc << ": Usage: host hostname realm\n";
		*err = true;
		return;
	}

	ptr<keydb_rec> *rec = keydb[args[2]];
	if(!rec)
		return;
	//ptr<rabin_priv> priv = (*rec)->priv;

	icp_pubkey realm;
	realm = (*rec)->key; //priv->n.getraw();

	hostname_rec *hnrec = New hostname_rec(args[1], realm);
	hostdb.insert(hnrec);
}

void
Pathserver::add_partial(vec<str> args, str loc, bool *err)
{
	if(args.size() != 2)
	{
		warn << loc << ": Usage: partial filename\n";
		*err = true;
		return;
	}
	
	ps_partialpath partial;
	conftab ct;
	str start, end;
	ct.add("start",		&start);
	ct.add("end",		&end);
	ct.add("path",		wrap(add_path, &keydb, &partial));
	ct.run(args[1]);

	ptr<keydb_rec> *rec = keydb[end];
	if(!rec)
		return;
	//ptr<rabin_priv> priv = (*rec)->priv;
	icp_realm realm;
	realm.pubkey = (*rec)->key; //priv->n.getraw();
	realm.vq = 0;
	partial.realm = (*rec)->key; //priv->n.getraw();

	rec = keydb[start];
	if(!rec)
		return;
	//priv = (*rec)->priv;
	icp_pubkey dest;
	dest = (*rec)->key; //priv->n.getraw();

	icp_pathdb_rec *dbrec = pathdb[dest];
	if(!dbrec)
	{
		dbrec = New icp_pathdb_rec();
		dbrec->dest = dest;
		pathdb.insert(dbrec);
	}
	dbrec->partialpaths.push_back(partial);
}

void Pathserver::SIRPRecieverDidReceiveUpdate(SIRPReceiver &receiver, ps_partialpath_vec& pathList){
  warn << "Did recieve update! Size: " << pathList.size() << "\n";
  for(unsigned i=0; i<pathList.size(); i++){
    ps_partialpath &partial = pathList[i];
    icp_pubkey &dest = partial.realm;
    icp_pathdb_rec *dbrec = pathdb[dest];
  	if(!dbrec)
  	{
  		dbrec = New icp_pathdb_rec();
  		dbrec->dest = dest;
  		pathdb.insert(dbrec);
  	}
  	// XXX: MM: Need to check if path already exists before inserting it.
  	// XXX: MM: Should dbrec be a hash? This will give us O(1) insertion time, vs O(n) when the correct implementation is added here
  	dbrec->partialpaths.push_back(partial);
  }
}

void
Pathserver::config(str filename)
{
	str server_realm, root_ps;
	conftab ct;
	ct.add("port",			&port, 0, 65535);
	ct.add("root",			&root);
	ct.add("root_ps",		&root_ps);
	ct.add("key",			wrap(&add_key, &keydb));
	ct.add("domain_ps",		wrap(this, &Pathserver::add_domain_ps));
	ct.add("host",			wrap(this, &Pathserver::add_host));
	ct.add("partial",		wrap(this, &Pathserver::add_partial));
	ct.run(filename);

	if(!root)
		root_server = root_ps;
		
	// Once everything is configured, let the SIRPReciever know about the pre-configured list of certificates/membership certificates. These certificates will be sent out to any client who tries to connect to us. MM: Is this a security hazard?
  icp_cert_vec cert_vec;
  icp_memcert_vec memcert_vec;
  for(icp_pathdb_rec *rec = pathdb.first(); rec != NULL; rec = pathdb.next(rec)){
    ps_partialpath_vec partialVec = rec->partialpaths;
    for(unsigned i=0; i<partialVec.size(); i++){
      ps_partialpath &partial = partialVec[i];
      icp_path_vec &pathVec = partial.paths;
      
      for(unsigned j=0; j<pathVec.size(); j++){
	rpc_vec<icp_certificate, RPC_INFINITY> certVec = pathVec[j].certs;
	rpc_vec<icp_memcert, RPC_INFINITY> memcertVec = pathVec[j].memcerts;
        
        //Add certificates XXX: MM: Need to take care of dupes
        for(unsigned k=0; k<certVec.size(); k++){
          cert_vec.push_back(certVec[k]);
        }
        //Add membership certificates XXX: MM: Need to take care of dupes.
        for(unsigned k=0; k<memcertVec.size(); k++){
          memcert_vec.push_back(memcertVec[k]);
        }
      }       
    }
  }
  _receiver.set_builtin_certs(cert_vec);
  _receiver.set_builtin_memcerts(memcert_vec);
}

void
Pathserver::start()
{
	fd = inetsocket(SOCK_STREAM, port);
	if(fd < 0)
		fatal("Unable to initialize socket on port %d: %m\n", port);

	int32_t listen_res = listen(fd, 100); // Allow up to 1000 in queue
	if(listen_res < 0)
		fatal("Was not able to listen on socket:\n%m\n");
	fdcb(fd, selread, wrap(this, &Pathserver::acpt_con));
	_receiver.start();
}
