#ifndef PATHSERVER_H_
#define PATHSERVER_H_

#include "async.h"
#include "arpc.h"
#include "crypt.h"
#include "aes.h"
#include "ihash.h"
#include "icp_ps.h"
#include "util.h"
#include "sirpreceiver.h"
#include "sirpreceiverdelegate.h"

class Test;

struct hostname_rec
{
	hostname_rec(icp_hostname n, icp_pubkey r) : name(n), realm(r) {};

	icp_hostname name;
	icp_pubkey realm;
	ihash_entry<hostname_rec> hlink;
};
typedef ihash<icp_hostname, hostname_rec, &hostname_rec::name, &hostname_rec::hlink> hostdb_t;

struct psdb_rec
{
	icp_hostname domain;
	icp_hostname server;
	ihash_entry<psdb_rec> hlink;
};
typedef ihash<icp_hostname, psdb_rec, &psdb_rec::domain, &psdb_rec::hlink> psdb_t;

struct icp_pathdb_rec
{
	icp_pubkey dest;
	ps_partialpath_vec partialpaths;
	ihash_entry<icp_pathdb_rec> hlink;
};
typedef ihash<icp_pubkey, icp_pathdb_rec, &icp_pathdb_rec::dest, &icp_pathdb_rec::hlink> icp_pathdb;

struct known_waypoint_rec
{
	known_waypoint_rec() {};
	known_waypoint_rec(icp_pubkey w, const icp_path_vec *p) : waypoint(w), paths(p)
	{};
	
	icp_pubkey waypoint;
	const icp_path_vec *paths;
	ihash_entry<known_waypoint_rec> hlink;
};
typedef ihash<icp_pubkey, known_waypoint_rec, &known_waypoint_rec::waypoint,
		&known_waypoint_rec::hlink> known_waypoint_db_t;

class Pathserver : public virtual refcount, SIRPReceiverDelegate
{
	friend class Test;

	private:
		int32_t fd;
		bool root;
		uint16_t port;
		icp_realm ps_realm;
		psdb_t psdb;
		hostdb_t hostdb;
		icp_hostname root_server;
		icp_pathdb pathdb;
		keydb_t keydb;

	public:
		Pathserver() : root(false), _receiver(*this) {}
    ~Pathserver() { }
    
		ptr<icp_path_vec> get_paths(const icp_pubkey &, const ps_partialpath_vec &);
		void init_from_file(const char *, const char * = "key");
		void config(str);
		void start();

	private:
		void acpt_con();

		void dispatch(ptr<asrv> srv, svccb *);
		void handle_find_tops(svccb *);
		void handle_find_tohost(svccb *);

		psdb_rec *get_known_host(const icp_hostname &);
		ptr<known_waypoint_db_t> build_known_waypoint_db(const ps_partialpath_vec &);

		void initialize_keys(const char *key_loc);
		ptr<icp_path_vec> build_paths(const icp_pubkey &, ptr<known_waypoint_db_t>);
		ptr<icp_path> generate_path(const icp_path &, const icp_path &);
		void get_pocs(icp_path &);

		void add_domain_ps(vec<str>, str, bool *);
		void add_host(vec<str>, str, bool *);
		void add_partial(vec<str>, str, bool *);
		
    /** 
		  @brief Handles delegate callbacks from the SIRPReciever
		  
		  Upon recipt of update, adds a partial path to the current db of paths.
		  
		  XXX: MM: Need callback to *remove* paths.
		  
		  @param receiver The receiver which sent the update
		  
		  @param partial_vec A list of partial paths to add.
		*/
		virtual void SIRPRecieverDidReceiveUpdate(SIRPReceiver &receiver, ps_partialpath_vec& pathList);
		
    /** The SIRPReceiver object, which listens for sIRP updates and issues a callback to add_partialpath_vec when an IRP_UPDATE is received */
    SIRPReceiver _receiver;
};

#endif /* PATHSERVER_H_ */
