#include "async.h"
#include "arpc.h"
#include "ihash.h"
#include "verify_path.h"
#include "util.h"

ptr<memcert_db>
build_memcert_db(const rpc_vec<icp_memcert, RPC_INFINITY> &memcerts)
{
	ptr<memcert_db> db_ptr = New refcounted<memcert_db>();
	memcert_db &db = *db_ptr;

	for(uint32_t i = 0; i < memcerts.size(); i++)
	{
		const icp_memcert &cur = memcerts[i];

		memcert_db_rec *rec;
		rec = db[cur.realm];

		if(!rec)
		{
			rec = New memcert_db_rec();
			rec->realm = cur.realm;
			db.insert(rec);
		}

		rec->memcerts.push_back(&cur);
	}

	return db_ptr;
}

/*
 * Checks whether <realm> is listed in the <SEL> field of <cert>, or if it
 * is a member of the vnode denoted class listed in the certificate.
 */
bool
icp_cert_includes(const icp_certificate &cert, const icp_cert_field_selector sel, 
		const icp_realm &realm, ptr<memcert_db> db)
{
	icp_realm field;

	if(sel == ICP_CERT_TO)
		field = cert.next;
	else if(sel == ICP_CERT_FROM)
		field = cert.prev;
	else
		return false;

	// If they're actually equal, then just go for it!
	if(field == realm)
		return true;

	icp_pubkey pattern;

	// Check against global wildcard, all zeros
	memset(pattern.base(), 0, pattern.size());
	if(field.pubkey == pattern)
		return true;

	// Check to make sure that "use class/vnode" pattern is present
	memset(pattern.base(), 0xFF, pattern.size());
	if(field.pubkey != pattern)
		return false;

	memcert_db_rec *rec = (*db)[cert.realm.pubkey];
	if(!rec)
		return false;
	memcert_vec &memcerts = rec->memcerts;

	// Check whether there is a memcert which states realm is a member of
	// the vnode denoted class
	for(uint32_t i = 0; i < memcerts.size(); i++)
	{
		const icp_memcert *memcert = memcerts[i];

		uint32_t realm_masked = ntohl(realm.vq) & ntohl(memcert->mask);
		uint32_t mem_vq_masked = ntohl(memcert->vq) & ntohl(memcert->mask);

		if(realm_masked == mem_vq_masked && memcert->member == realm.pubkey && 
				icp_verify(*memcert))
			return true;
	}

	return false;
}

bool
verify_forward(const icp_path &path, ptr<memcert_db> db)
{
	// Copy so that we can 'pop' as we read through certs and points
	rpc_vec<icp_realm, RPC_INFINITY> waypoints = path.waypoints;
	rpc_vec<icp_certificate, RPC_INFINITY> certs = path.certs;

	if( (certs[0].realm != waypoints[1] || 
			!icp_cert_includes(certs[0], ICP_CERT_FROM, waypoints[0], db)) ||
			!icp_verify(certs[0]))
	{
#ifdef DEBUG
		warn << "verify_forward: here1\n";
#endif
		return false;
	}
	waypoints.pop_front();
	icp_realm prev_point = waypoints.pop_front();

	while(waypoints.size() != 0)
	{
		icp_realm cur_point;

		if(icp_cert_includes(certs.front(), ICP_CERT_TO, waypoints.front(), db))
			cur_point = waypoints.front();

		// TODO: try to bring this to the top.
		if(cur_point == waypoints.back())
			return true;

		certs.pop_front();

		// this one is actually fine
		while(certs.size() != 0 && certs.front().realm != cur_point)
			certs.pop_front();
		if(certs.size() == 0)
		{
#ifdef DEBUG
			warn << "verify_forward: here2\n";
#endif
			return false;
		}

		if(!icp_cert_includes(certs.front(), ICP_CERT_FROM, prev_point, db))
		{
			certs.pop_front();
			if(certs.size() == 0 || certs.front().realm != cur_point || 
					!icp_cert_includes(certs.front(), ICP_CERT_FROM, prev_point, db))
			{
#ifdef DEBUG
				warn << "verify_forward: here3\n";
#endif
				return false;
			}
		}

		if(!icp_verify(certs.front()))
		{
#ifdef DEBUG
			warn << "verify_forward: here4\n";
#endif
			return false;
		}

		prev_point = waypoints.pop_front();
	}

	return true;
}

bool
verify_backward(const icp_path &path, ptr<memcert_db> db)
{
	rpc_vec<icp_realm, RPC_INFINITY> waypoints = path.waypoints;
	rpc_vec<icp_certificate, RPC_INFINITY> certs = path.certs;

	if( !icp_cert_includes(certs.back(), ICP_CERT_FROM, waypoints.back(), db) ||
			certs.back().realm != waypoints[waypoints.size() - 2] || !icp_verify(certs.back()))
		return false;
	waypoints.pop_back();
	icp_realm prev_point = waypoints.pop_back();

	while(waypoints.size() != 0)
	{
		icp_realm cur_point;

		if(icp_cert_includes(certs.back(), ICP_CERT_TO, waypoints.back(), db))
			cur_point = waypoints.back();

		if(cur_point == waypoints.front())
			return true;

		certs.pop_back();

		while(certs.size() != 0 && certs.back().realm != cur_point)
			certs.pop_back();
		if(certs.size() == 0)
			return false;

		if(!icp_cert_includes(certs.back(), ICP_CERT_FROM, prev_point, db))
		{
			certs.pop_back();
			if(certs.size() == 0 || certs.back().realm != cur_point ||
					!icp_cert_includes(certs.back(), ICP_CERT_FROM, prev_point, db))
			{
				return false;
			}
		}
		if(!icp_verify(certs.back()))
			return false;
		prev_point = waypoints.pop_back();
	}
	
	return true;

}

bool
icp_verify_path(const icp_path &path)
{
	ptr<memcert_db> memcertdb = build_memcert_db(path.memcerts);
	bool forward = verify_forward(path, memcertdb);
	bool backward = verify_backward(path, memcertdb);
#ifdef DEBUG
	if(!forward)
		warn << "forward didn't verify\n";
	if(!backward)
		warn << "backward didn't verify\n";
#endif
	return (forward && backward);
}
