/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIRPNEIGHBOR_H_
#define SIRPNEIGHBOR_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "async.h"
#include "arpc.h"
#include "icp_sirp.h"
#include <time.h>

/**
  @struct SIRPNeighbor
  
  This struct represents a SIRP neighbor, from which we can recieve updates to build a network topology.
*/
struct SIRPNeighbor
{

  /** A neighbor is:
  CONNECTING if we're trying to issue a connection to the neighbor,
  UP if we recieve a IRP_OPEN_ACK from that host
  and DOWN if it times out responding to an RPC,*/
	enum neighborStatus { DOWN, UP, CONNECTING };
	
	/** The public key of the neighbor */
	icp_pubkey realmid;
	
	/** The IP address of the neighbor, in string format */
	str ipStr;
	
	/** The port of the neighbor */
	uint16_t port;
	
	/** The last time the neighbor sent us an update */
	time_t lastUpdate;
	
	/** A list of "I constant to carry traffic from <X> to <Y>", where either <X> or <Y> can be 0 for wildcard */
	icp_cert_vec certs;
	/** A list of "membership certificates" sirp.tex  makes it sound like these are an optimization for certs to group contiguous vnodes together. Not too sure about the details. */
	icp_memcert_vec memcerts;

  /** Neighbor type(peer, provider, customer)*/
	irp_neighbor_type type;
	
	/** Neighbor status: either DOWN, UP, or CONNECTING*/
	neighborStatus status;

  /** RPC object to communicate with the neighbor. (IRP_OPEN, ... etc.)*/
	ptr<aclnt> client;
	/** RPC object to handle incomming communications from neighbors. */
	ptr<asrv> server;

  /** The last time which we heard from the neighbor. After keepAlive + KEEPALIVE_TIMEOUT_S, the neighbor is dropped if we don't get any traffic from it. */
	time_t keepAlive;
	/** Callback which keeps sending keep alive messages to the neighbor. */
	timecb_t *sendkacb;
	/** Callback which keeps checking if we have gotten a keep alive message from the neighbor, and drops it if not. */
	timecb_t *checkkacb;

  //Struct members to allow SIRPNeighbor to be used in an ihash. Nothing too interesting.
	ihash_entry<SIRPNeighbor> neighborentry;
	ihash_entry<SIRPNeighbor> customerentry;

	SIRPNeighbor() : lastUpdate(0) { }
};
/** A struct which maps public key -> SIRPNeighbor */
typedef ihash<icp_pubkey, SIRPNeighbor, &SIRPNeighbor::realmid, 
		&SIRPNeighbor::neighborentry> neighborHash_t;
/** A struct which maps public key -> SIRPNeighbor for customers. MM: Not sure what the use of this is, versus the above. I guess it's because customers are stored in two hashes: _neighbors and _customers(SIRPRouter.h). Seems sloppy to be doing this. */
typedef ihash<icp_pubkey, SIRPNeighbor, &SIRPNeighbor::realmid, 
		&SIRPNeighbor::customerentry> customerHash_t;

/**
  @brief Returns the "opposite" kind of neighbor.
  
  For example, a "Customer" is the opposite of a "Provider", and a "Peer" is the opposite of a "Peer".
  
  @param type A neighbor type
  
  @returns The opposite neighbor type of 'type'
*/
inline irp_neighbor_type
oppositeNeighborType(irp_neighbor_type type)
{
	switch(type)
	{
		case IRP_CUSTOMER:
			return IRP_PROVIDER;
		case IRP_PEER:
			return IRP_PEER;
		case IRP_PROVIDER:
			return IRP_CUSTOMER;
		default:
			assert(false);
	}
}

#endif /* SIRPNEIGHBOR_H_ */
