/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sirpreceiver.h"
#include "updatemanager.h"
#include "ihash.h"

void
SIRPReceiver::start()
{
	_listenFD = inetsocket(SOCK_STREAM, _port);
	if(_listenFD < 0)
		fatal("Unable to initialize listen socket on port %d: %m\n", _port);

	int listenResult = listen(_listenFD, 100);
	if(listenResult < 0)
		fatal("Was not able to listen on socket: %m\n");

	fdcb(
			_listenFD,
			selread,
			wrap(this, &SIRPReceiver::acceptConnection)
		);
}

void
SIRPReceiver::acceptConnection()
{
	int acceptFD = accept(_listenFD, NULL, NULL);
	if(acceptFD < 0)
	{
		warn("Could not accept request: %m\n");
		return;
	}

	ptr<asrv> srv = init_tcp_server(acceptFD, icing_sirp_interface_1);
	if(!srv)
	{
		warn << "Failed to initialize a TCP server object\n";
		return;
	}

	srv->setcb(
			wrap(this, &SIRPReceiver::dispatch, srv));
}

void
SIRPReceiver::dispatch(ptr<asrv> srv, svccb *rpc)
{
	if(!rpc)
	{
		srv->setcb(NULL);
		return;
	}

	switch(rpc->proc())
	{
		case IRP_UPDATE:
			handleUpdate(rpc);
			break;
		case IRP_KEEPALIVE:
			handleKeepAlive(rpc);
			break;
		default:
#ifdef DEBUG
			warn("Procedure no. %d not handled by SIRP v1\n", rpc->proc());
#endif
			rpc->reject(PROC_UNAVAIL);
	}
}

void
SIRPReceiver::handleKeepAlive(svccb *rpc){
  irp_reply reply;
	reply.set_status(IRP_ACK);
	rpc->reply(&reply);
}

void
SIRPReceiver::handleOpen(svccb *rpc){
  irp_open *openargs = rpc->getarg<irp_open>();
	irp_neighbor_type opposite = oppositeNeighborType(openargs->type);
  assert(opposite == IRP_PROVIDER); // XXX: MM: OK to assume that we're always a customer in the pathserver case? Made this assumption throughout the code.
  
  // Reply with our constant list of paths to let SIRP know we're up and ready to recieve
  irp_reply reply;
  reply.set_status(IRP_PS_OPEN_ACK);

	reply.ps_open->realmid = _realmid;
	reply.ps_open->type = IRP_PROVIDER;
  reply.ps_open->certs = _builtin_certs;
  reply.ps_open->memcerts = _builtin_memcerts;
  
  rpc->reply(&reply);
}
void
SIRPReceiver::handleUpdate(svccb *rpc)
{
	irp_update *update = rpc->getarg<irp_update>();
	assert(update);

#ifdef DEBUG
	warn << "Update!\n";
#endif

	_certs = update->certs;
	_memcerts = update->memcerts;

	IMMEDIATE_CB(
			wrap(this, &SIRPReceiver::convertUpdate, rpc));
}

void
SIRPReceiver::convertUpdate(svccb *rpc)
{
	conversionHash_t chash;
	partialHash_t phash;
	buildHash(chash);

	vec<icp_pubkey> neighbors = findImmediateNeighbors(_realmid, chash);
	getDestinations(neighbors, phash, chash);

  // Prepare data so we have a list of partialpaths(what the pathserver expects)
  ps_partialpath_vec vec;
  warn << "Got update from sirp router! Size: " << phash.size() << "\n";
  for(PartialEntry *pe = phash.first(); pe != NULL; pe = phash.next(pe)){
    ps_partialpath p = pe->paths;
    vec.push_back(p);
  }

  // Execute the callback so the SIRPReceiver delegate(most likely the pathserver) gets access to the new partialpaths after the IPR_UPDATE
  _delegate.SIRPRecieverDidReceiveUpdate(*this, vec);
  
  

	// Destroy chash (and members)
	chash.deleteall();
	phash.deleteall();
}

void
SIRPReceiver::buildHash(conversionHash_t &chash)
{
	for(int i = 0; i < _certs.size(); i++)
	{
		icp_certificate &cert = _certs[i];

		ConversionEntry *e = chash[cert.realm.pubkey];
		if(!e)
		{
			e = new ConversionEntry;
			e->realmid = cert.realm.pubkey;
			chash.insert(e);
		}

		e->certs.push_back(cert);
	}

	for(int i = 0; i < _memcerts.size(); i++)
	{
		icp_memcert &memcert = _memcerts[i];
		
		ConversionEntry *e = chash[memcert.realm];
		assert(e);
		
		e->memcerts.push_back(memcert);
	}
}

vec<icp_pubkey>
SIRPReceiver::findImmediateNeighbors(icp_pubkey &realmid, conversionHash_t &chash)
{
	vec<icp_pubkey> neighbors;
	
	ConversionEntry *e = chash[realmid];
	assert(e);

	for(int i = 0; i < e->memcerts.size(); i++)
		neighbors.push_back(e->memcerts[i].member);

	return neighbors;
}

void
SIRPReceiver::getDestinations(vec<icp_pubkey> startPoints, partialHash_t &phash,
		conversionHash_t &chash)
{
	vec<icp_pubkey> next;

	for(int i = 0; i < startPoints.size(); i++)
	{
		icp_pubkey &realmid = startPoints[i];
		vec<icp_pubkey> neighbors = findImmediateNeighbors(realmid, chash);

		ConversionEntry *ce = chash[realmid];
		assert(ce);

		bool firstLevel = !(phash[realmid]);

		for(int j = 0; j < neighbors.size(); j++)
		{
			icp_pubkey &neighbor = neighbors[j];
			icp_realm realmvq;
			realmvq.pubkey = neighbor;

			PartialEntry *pe = phash[neighbor];
			if(!pe)
			{
				pe = new PartialEntry;
				pe->realmid = neighbor;
				pe->paths.realm = neighbor;
				phash.insert(pe);

				next.push_back(neighbor);
			}

			// Now start building our paths. In reverse order. So start at dest.

			icp_path path;
			
			// Find the vnode class necessary
			for(int k = 0; k < ce->memcerts.size(); k++)
			{
				if(ce->memcerts[k].member == neighbor)
				{
					// XXX: being cheeky. not checking the VQ mask.
					realmvq.vq = ce->memcerts[k].vq;
					path.memcerts.push_back(ce->memcerts[k]);
					break;
				}
			}

			path.certs.setsize(2);
			for(int k = 0; k < ce->certs.size(); k++)
			{
				if(ce->certs[k].next.vq == realmvq.vq)
					path.certs[0] = ce->certs[k];
				if(ce->certs[k].prev.vq == realmvq.vq)
					path.certs[1] = ce->certs[k];
			}

			uint32_t vnode = path.certs.back().realm.vq;
			if(firstLevel)
			{
				icp_realm temp;

				temp.pubkey = _realmid;
				temp.vq = vnode;
				path.waypoints.push_back(temp);

				realmvq.pubkey = realmid;
				path.waypoints.push_back(temp);
				path.waypoints.push_back(realmvq);

				for(int k = 0; k < ce->memcerts.size(); k++)
				{
					if(ce->memcerts[k].member == _realmid)
						path.memcerts.push_back(ce->memcerts[k]);
				}
			}
			else
			{
				path.waypoints.push_back(realmvq);
				PartialEntry *parentPE = phash[realmid];
				assert(parentPE);

				for(int k = 0; k < parentPE->paths.paths.size(); k++)
				{
					pe->paths.paths.push_back(parentPE->paths.paths[k]);
					pe->paths.paths.back().waypoints.back().vq = vnode;
					path_append(pe->paths.paths.back(), path);
				}
			}
		}
	}

	getDestinations(next, phash, chash);
}
