/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIRPRECEIVER_H_
#define SIRPRECEIVER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "async.h"
#include "arpc.h"
#include "util.h"
#include "icp_sirp.h"
#include "sirpreceiverdelegate.h"

#define RECEIVER_PORT 7476

/* This thing translates sirps output into something usable by say, a pathserver */

/**
  @class SIRPReceiver
  
  This class accepts connections on a socket and listens for IRP_UPDATEs.
  
  Users of this class must implement SIRPReceiverDelegate(by subclassing the SIRPReceiverDelegate abstract base class), which defines methods to deal with recipt of data from SIRP.
  
  MM: This class is not used anywhere.... How can I find an interface to it if it isn't used? (checked on icing{1,2,3}). Maybe Arun had something on nachtwey? It is also strange that updatecb is never used... this would be the mechanism for a pathserver to get a calblack... Additionally, the partials are never stored in an ivar... I'm not sure this class *HAS* a usable interface now.
*/
class SIRPReceiver
{
public:

	SIRPReceiver(SIRPReceiverDelegate &delegate) : _delegate(delegate), _port(RECEIVER_PORT) { };
  ~SIRPReceiver() { };

  /** Starts the SIRP reciever on port _port, and forwards connections to acceptConnection */
	void start();

  /** Sets the builtin certificates, which we distribute to clients whenever they connect to us */
  void set_builtin_certs(icp_cert_vec& certs) { _builtin_certs = certs; }

  /** Sets the builtin memnbership certificates, which we distribute to clients whenever they connect to us */
  void set_builtin_memcerts(icp_memcert_vec& certs) { _builtin_memcerts = certs; }

private:
  /** The callback to execute when we get an IRP_UPDATE */
  SIRPReceiverDelegate &_delegate;
  
  /** When a pathserver instantiates this class, it must provide a list of default certificates. This list is supplied to any SIRP client who connects to us. If we didn't have this, no client could reach any of the pathserver's clients(assuming the clients were realms)! Note: This might not be necessary in the future, when clients are subrealms of the pathserver. */
  icp_cert_vec _builtin_certs;
  
  /** Same as _builtin_certs, but for memcerts. MM: Need to add good headerdocs here.*/
  icp_memcert_vec _builtin_memcerts;
  
  /** The port to recieve updates on. MM: How can this be updated? It is private, and not set anywhere in sirpreceiver.{c,h}. */
	uint16_t _port;
	/** The FD which we recieve connections from icing neighbors on. Listens on port '_port'. */
	int _listenFD;
	/** The realmid of the current realm. MM: How can this be updated? It is private and not set anywhere in sirpreceiver.{c,h}. */
	icp_pubkey _realmid;

  /** The certificates received in the IRP_UPDATE. MM: Should this be an ivar? Seems like it would be better to pass this around as a param.*/
	icp_cert_vec _certs;
	/** The membership certificates received in the IRP_UPDATE.  MM: Should this be an ivar? Seems like it would be better to pass this around as a param.*/
	icp_memcert_vec _memcerts;
	/** MM: This is really sloppy. This ivar is never used! It probably should be though, to keep state. */
	ps_partialpath_vec _partials;
  
  /** public key + (list of certs and memcerts) + stuff for sfslite's hash */
	struct ConversionEntry
	{
		icp_pubkey realmid;
		icp_cert_vec certs;
		icp_memcert_vec memcerts;

		ihash_entry<ConversionEntry> hlink;
	};
	/** public key -> (list of certs and memcerts) hashtable */
	typedef ihash<icp_pubkey, ConversionEntry, 
		   &ConversionEntry::realmid, &ConversionEntry::hlink> conversionHash_t;

  /** public key + paths + stuff for sfslite's hash */
	struct PartialEntry
	{
		icp_pubkey realmid;
		ps_partialpath paths;

		ihash_entry<PartialEntry> hlink;
	};
	
	/** public key -> paths hashtable */
	typedef ihash<icp_pubkey, PartialEntry,
			&PartialEntry::realmid, &PartialEntry::hlink> partialHash_t;

  /**
    @brief Accepts connections on the listening file descriptor.
    
    Subsequent updates from the connection are delivered to dispatch(...).
  */
	void acceptConnection();

  /*
	  @brief Dispatches RPC callbacks to the appropriate handler method.
	  
	  The only accepted RPC method is IRP_UPDATE(dispatched to handleUpdate(...)).	  
	  This method stops accepting RPCs if it recieves an invalid RPC call.
	  
	  @param srv The server object accepting RPCs
	  
	  @param rpc The RPC result
	*/
	void dispatch(ptr<asrv>, svccb *);
	// XXX: MM: Add headerdocs
	void handleKeepAlive(svccb *rpc);
	/** @brief Handles an IRP_OPEN RPC.
	
	This method replies with an IRP_PS_OPEN_ACK to let the sIRP router know we're up and running, and provide it with a list of constant builtin paths which we have access to.
	
	@param rpc The RPC result */
	void handleOpen(svccb *);
  /**
    @brief Handles an IRP_UPDATE RPC.
    
    The method stores the certs and memcerts from the IRP_UPDATE into the _certs and _memcerts ivars, respectively, then dispatches an immediate callback to convertUpdate.
    
    @param srv The RPC result
  */
	void handleUpdate(svccb *);

	// This is where the magic happens
	
	void convertUpdate(svccb *);
	/**
	  @brief Builds up a ConversionHash(realmid -> (list of certs + memcerts)) dictionary
	  
	  Uses _certs and _memcerts (which are both private ivars) to build up the hash
	  
	  @param chash On return, the hash is populated with certs from _certs and memcerts from _memcerts	  
	*/
	void buildHash(conversionHash_t &);
	
	/**
	  @brief Finds a list of realms which can be reached through a given neighbor.
	  
    Uses chash to lookup the memcerts of a given node.
    
    MM: Shouldn't certs of a node also be considered? I thought memcerts were just an optimization for certs. Is this a bug?
    
    @param realmid The realm id which we want to find the neighbors of
    
    @param chash The conversion hash to use to lookup the neighbors in
    
    @returns A list of realmids accessible from the neighbor with realm id 'realmid'
	*/
	vec<icp_pubkey> findImmediateNeighbors(icp_pubkey &, conversionHash_t &);
	/**
	  @brief Builds up the topology of the network.
	  
	  Stores a (realmid -> path) mapping for all paths in 'phash'
	  
	  MM: This function confused me a bit(there is a lot going on). Need to go back and reread.
	  
	  @param startPoints A list of realmids which form the start of paths
	  
	  @param phash On return, contains a list of list of realmid -> path mappings which represent the topology
	  
	  @param chash A list of public key -> (certs and memcerts) mappings which are used to build the topology
	  
	*/
	void getDestinations(vec<icp_pubkey>, partialHash_t &, conversionHash_t &);
};

#endif /* SIRPRECEIVER_H_ */
