/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sirprouter.h"
#include "util.h"
#include "parseopt.h"
#include "icp_config.h"

SIRPRouter::SIRPRouter() :
	_port(SIRP_PORT), manager(NULL)
{
#ifdef DEBUG
	warn << "Welcome to a SIRP Router\n";
#endif
	_updateTimestamp = time(NULL);
}

void
SIRPRouter::start()
{
	_listenFD = inetsocket(SOCK_STREAM, _port);
	if(_listenFD < 0)
		fatal("Unable to initialize socket on port %d: %m\n", _port);

	int listenResult = listen(_listenFD, 100);
	if(listenResult < 0)
		fatal("Was not able to listen on socket:\n%m\n");
	
	fdcb(
			_listenFD,
			selread, 
			wrap(this, &SIRPRouter::acceptConnection)
			);
	_lastupdatecb = delaycb(UPDATE_INIT_DELAY_S, 0, 
			wrap(this, &SIRPRouter::updateCustomers, static_cast<timecb_t *>(NULL)));
}

/*
 * Get the RealmID, [port], and neighbor list
 * Currently neighbor list is based on IP address. Eventually want based on RealmID
 * when this flows over ICING (ie. SIRPRouter should install table entries for each
 * RealmID with PoC/PoCKey.
 */
void
SIRPRouter::config(str filename)
{
	str realmid;

	conftab ct;
	ct.add("realmid",		&realmid);
	ct.add("port",			&_port, 0, 65535);
	ct.add("neighbor",		wrap(this, &SIRPRouter::configAddNeighbor));
	ct.run(filename);

	convertFromHex(_realmid, realmid.cstr(), realmid.len()/2);
	manager = New UpdateManager(_realmid, _neighbors);
	delaycb(UpdateManager::EXPIRY_TIME, 0,
			wrap(this, &SIRPRouter::renewUpdate));
}

void
SIRPRouter::configAddNeighbor(vec<str> args, str loc, bool *err)
{
	if(args.size() != 5)
	{
		warn << loc << ": Usage: neighbor realmid <ip address> <port> <T | C | P>\n";
		*err = true;
		return;
	}

	SIRPNeighbor *neighbor = New SIRPNeighbor;

	convertFromHex(neighbor->realmid, args[1].cstr(), args[1].len()/2);
	neighbor->ipStr = args[2];
	neighbor->port = atoi(args[3]);

	switch(args[4].cstr()[0])
	{
		case 'T':
		case 't':
			neighbor->type = IRP_PROVIDER;
			break;
		case 'C':
		case 'c':
			neighbor->type = IRP_CUSTOMER;
			_customers.insert(neighbor);
			break;
		case 'P':
		case 'p':
			neighbor->type = IRP_PEER;
			break;
		default:
			warn << loc << ": Invalid neighbor type\n";
			warn << loc << ": Usage: neighbor <realmid> <ip address> <port> <T | C | P>\n";
			*err = true;
			delete neighbor;
			return;
	}

	_neighbors.insert(neighbor);
	neighbor->status = SIRPNeighbor::CONNECTING;
	delaycb(5, 0, 
			wrap(this, &SIRPRouter::openConnection, neighbor));
}

void
SIRPRouter::acceptConnection()
{
	int acceptFD = accept(_listenFD, NULL, NULL);
	if(acceptFD < 0)
	{
		warn("Could not accept request:\n%m\n");
		return;
	}

	ptr<asrv> srv = init_tcp_server(acceptFD, icing_sirp_1);
	if(!srv)
	{
		warn << "Failed to initialize a TCP server object\n";
		return;
	}
	srv->setcb(
			wrap(this, &SIRPRouter::dispatch, srv));
}

void
SIRPRouter::dispatch(ptr<asrv> srv, svccb *rpc)
{
	if(!rpc)
	{
		/* Note that this means that the client disconnected!
		 * Eg. A hang up occured.
		 */
		srv->setcb(NULL);
		return;
	}

	switch(rpc->proc())
	{
		case IRP_OPEN:
			handleOpen(srv, rpc);
			break;
		case IRP_UPDATE:
			handleUpdate(rpc);
			break;
		case IRP_KEEPALIVE:
			handleKeepAlive(rpc);
			break;
		default:
			warn("Procedure no. %d not handled by SIRP  v1\n", rpc->proc());
			rpc->reject(PROC_UNAVAIL);
	}
}

void
SIRPRouter::handleOpen(ptr<asrv> srv, svccb *rpc)
{
#ifdef DEBUG
	warn << "OPEN received\n";
#endif
	irp_reply reply;
	irp_open *openargs = rpc->getarg<irp_open>();

	irp_neighbor_type opposite = oppositeNeighborType(openargs->type);
#ifdef DEBUG
	warn << opposite << "\n";
#endif
	SIRPNeighbor *neighbor;
	if(! (neighbor = _neighbors[openargs->realmid]) || neighbor->type != opposite)
	{
		reply.set_status(IRP_UNKNOWN_NEIGHBOR);
#ifdef DEBUG
		warn << "Unknown neighbor\n";
#endif
	}
	else
	{
		assert(opposite == neighbor->type);
		reply.set_status(IRP_OPEN_ACK);
		reply.open->realmid = _realmid;
		reply.open->type = neighbor->type;
#ifdef DEBUG
		warn << "Sending OPENACK\n";
#endif
	}

	neighbor->server = srv;
	neighbor->keepAlive = time(NULL);
	neighbor->checkkacb = delaycb(KEEPALIVE_TIMEOUT_S, 0,
			wrap(this, &SIRPRouter::checkKeepAlive, neighbor));
	if(neighbor->status == SIRPNeighbor::DOWN)
	{
		neighbor->status = SIRPNeighbor::CONNECTING;
		IMMEDIATE_CB(
				wrap(this, &SIRPRouter::openConnection, neighbor));
	}
	rpc->reply(&reply);
}

void
SIRPRouter::handleUpdate(svccb *rpc)
{

	irp_update *update = rpc->getarg<irp_update>();
	assert(update);
	SIRPNeighbor *neighbor = _neighbors[update->realmid];

	assert(neighbor->type == IRP_PROVIDER);
	if(neighbor->lastUpdate < update->timestamp)
	{
#ifdef DEBUG
	warn << "Update!\n";
#endif
		neighbor->certs = update->certs;
		neighbor->memcerts = update->memcerts;
		neighbor->lastUpdate = update->timestamp;
		IMMEDIATE_CB(
				wrap(this, &SIRPRouter::updateCustomers, _lastupdatecb));
	}

	irp_reply reply;
	reply.set_status(IRP_ACK);
	rpc->reply(&reply);
}

void
SIRPRouter::handleKeepAlive(svccb *rpc)
{
#ifdef DEBUG
	warn << "KeepAlive!\n";
#endif
	irp_open *openargs = rpc->getarg<irp_open>();

	SIRPNeighbor *neighbor = _neighbors[openargs->realmid];
	//XXX: again, may not want to assert this
	assert(neighbor);
	assert(neighbor->type == oppositeNeighborType(openargs->type));
	neighbor->keepAlive = time(NULL);

	irp_reply reply;
	reply.set_status(IRP_ACK);
	rpc->reply(&reply);
}

void
SIRPRouter::openConnection(SIRPNeighbor *neighbor)
{
	assert(neighbor->status == SIRPNeighbor::CONNECTING);
#ifdef DEBUG
	warn << "Attempting neighbor connection\n";
#endif
	tcpconnect(neighbor->ipStr, neighbor->port, 
			wrap(this, &SIRPRouter::connectionMade, neighbor));
}

void
SIRPRouter::connectionMade(SIRPNeighbor *neighbor, int fd)
{
	assert(neighbor->status == SIRPNeighbor::CONNECTING);

	if(fd < 0)
	{
		warn("Could not open socket: %m\n");
		return;
	}

	ptr<aclnt> clnt = init_tcp_client(fd, icing_sirp_1);
	if(!clnt)
	{
		warn << "Failed to initialize a TCP client object\n";
		return;
	}

	neighbor->client = clnt;

	ref<irp_reply> reply = New refcounted<irp_reply>();
	irp_open openargs;
	openargs.realmid = _realmid;
	openargs.type = neighbor->type;

#ifdef DEBUG
	warn << "Sending OPEN\n";
#endif

	neighbor->client->call(
			IRP_OPEN,
			&openargs,
			reply,
			wrap(this, &SIRPRouter::handleReply, neighbor, reply));
}

void
SIRPRouter::handleReply(SIRPNeighbor *neighbor, ref<irp_reply> reply, clnt_stat err)
{
	switch(err)
	{
		case RPC_SUCCESS:
			parseReply(neighbor, reply);
			break;
		case RPC_CANTSEND:
		case RPC_CANTRECV:
		case RPC_TIMEDOUT:
			dropNeighbor(neighbor);
			break;
		default:
			warn("WARNING:\tReceived response other than RPC_SUCCESS\t%d", err);
	}
}

void
SIRPRouter::parseReply(SIRPNeighbor *neighbor, ref<irp_reply> reply)
{
	switch(reply->status)
	{
		case IRP_OPEN_ACK:
			parseOpenAck(neighbor, reply);
			break;
		case IRP_PS_OPEN_ACK:
		        parsePsOpenAck(neighbor, reply);
		        break;
		case IRP_ACK:
			break;
		case IRP_MALFORMED_MSG:
			break;
		case IRP_UNKNOWN_NEIGHBOR:
			break;
		default:
			assert(false);
	}
}

void
SIRPRouter::parsePsOpenAck(SIRPNeighbor *neighbor, ref<irp_reply> reply)
{
  assert(neighbor->type == IRP_CUSTOMER);
  assert(neighbor->realmid == reply->ps_open->realmid);
	assert(neighbor->type == oppositeNeighborType(reply->ps_open->type));
	
  warn << "Got PS Open Ack!\n";
  
	//Grab out the certs the client sent us, then stash them
  neighbor->certs = reply->ps_open->certs;
  neighbor->memcerts = reply->ps_open->memcerts;
  neighbor->lastUpdate = time(NULL); // XXX: MM: Is this OK?
  
  neighbor->status = SIRPNeighbor::UP;
	neighbor->sendkacb = delaycb(KEEPALIVE_DELAY_S, 0, 
			wrap(this, &SIRPRouter::sendKeepAlive, neighbor));
}
void
SIRPRouter::parseOpenAck(SIRPNeighbor *neighbor, ref<irp_reply> reply)
{
	//XXX: maybe not assert this, since it would cause system to crash
	assert(neighbor->realmid == reply->open->realmid);
	assert(neighbor->type == oppositeNeighborType(reply->open->type));
#ifdef DEBUG
	warn << "OpenAck received\n";
#endif
	neighbor->status = SIRPNeighbor::UP;
	neighbor->sendkacb = delaycb(KEEPALIVE_DELAY_S, 0, 
			wrap(this, &SIRPRouter::sendKeepAlive, neighbor));
}

void
SIRPRouter::sendKeepAlive(SIRPNeighbor *neighbor)
{
	assert(neighbor->status == SIRPNeighbor::UP);

	irp_open openargs;
	openargs.realmid = _realmid;
	openargs.type = neighbor->type;

	ref<irp_reply> reply = New refcounted<irp_reply>();
#ifdef DEBUG
	warn << "Sending KeepAlive\n";
#endif
	neighbor->sendkacb = delaycb(KEEPALIVE_DELAY_S, 0, 
			wrap(this, &SIRPRouter::sendKeepAlive, neighbor));
	neighbor->client->call(
			IRP_KEEPALIVE,
			&openargs,
			reply,
			wrap(this, &SIRPRouter::handleReply, neighbor, reply));
}

void
SIRPRouter::checkKeepAlive(SIRPNeighbor *neighbor)
{
#ifdef DEBUG
	warn << "Checking KeepAlive\n";
#endif

	time_t now = time(NULL);
	if(neighbor->keepAlive + KEEPALIVE_TIMEOUT_S < now)
		dropNeighbor(neighbor);
	else
		neighbor->checkkacb = delaycb(KEEPALIVE_TIMEOUT_S, 0, 
				wrap(this, &SIRPRouter::checkKeepAlive, neighbor));
}

void
SIRPRouter::updateCustomers(timecb_t *old)
{
	if(old)
	{
#ifdef DEBUG
		warn << "Removing old update callback\n";
#endif
		timecb_remove(old);
		manager->updateCerts();
		_updateTimestamp = time(NULL);
	}

	_customers.traverse(
			wrap(this, &SIRPRouter::sendUpdate));
	_lastupdatecb = delaycb(UPDATE_DELAY_S, 0,
			wrap(this, &SIRPRouter::updateCustomers, static_cast<timecb_t *>(NULL)));
}

void
SIRPRouter::sendUpdate(SIRPNeighbor *neighbor)
{
	if(neighbor->status != SIRPNeighbor::UP)
		return;

	assert(neighbor->type == IRP_CUSTOMER);
	irp_update update;
	update.realmid = _realmid;
	update.timestamp = _updateTimestamp;
	update.certs = manager->certs;
	update.memcerts = manager->memcerts;

	ref<irp_reply> reply = New refcounted<irp_reply>();
#ifdef DEBUG
	warn << "Sending Update\n";
	print_realmid(neighbor->realmid);
#endif
	neighbor->client->call(
			IRP_UPDATE,
			&update,
			reply,
			wrap(this, &SIRPRouter::handleReply, neighbor, reply));
}

void
SIRPRouter::renewUpdate()
{
	manager->renewCerts();
	_updateTimestamp = time(NULL);
	delaycb(UpdateManager::EXPIRY_TIME, 0,
			wrap(this, &SIRPRouter::renewUpdate));
}

void
SIRPRouter::dropNeighbor(SIRPNeighbor *neighbor)
{
#ifdef DEBUG
	warn << "Dropping neighbor\n";
#endif

#if 0
	if(neighbor->type == IRP_CUSTOMER)
		_customers.remove(neighbor);
	_neighbors.remove(neighbor);
#endif
	neighbor->status = SIRPNeighbor::DOWN;
	neighbor->server->setcb(NULL);
	neighbor->server = NULL;
	neighbor->client = NULL;

#ifdef DEBUG
	warn << "Removing sendkacb\n";
#endif
	timecb_remove(neighbor->sendkacb);
#ifdef DEBUG
	warn << "Removing checkkacb\n";
#endif
	timecb_remove(neighbor->checkkacb);

//	delete neighbor;

	// XXX: Make call to rebuild neighbor database and send updates to customers.
}
