/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIRPROUTER_H_
#define SIRPROUTER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "async.h"
#include "arpc.h"
#include "neighbor.h"
#include "icp_sirp.h"
#include "updatemanager.h"
#include "sirpreceiver.h"

#define SIRP_PORT 7477
#define KEEPALIVE_DELAY_S 30
#define KEEPALIVE_TIMEOUT_S 60
#define UPDATE_INIT_DELAY_S 30
#define UPDATE_DELAY_S 60

/**
  This class is responsible for routing SIRP traffic. Specifically, it carries traffic between SIRPNeighbors.
  
  The interface to the class is:
  
  1. Configure the router by calling config(...) with a conftab filename.
     The lines you can specify are:
     
     realmid <realmid> //Current realm id. Should only be one of these lines
     
     port <port> //Port to listen on. Should only be one of these lines.
     
     // List of neighbors delimited by newlines, in the following format:
     // T = provider
     // C = customer
     // P = peer
     neighbor <realmid> <ip address> <port> <T | C | P>
     
  2. Call start() to start the server listening for connections from neighbors
  
  3. Call amain() to start the event loop
*/
class SIRPRouter
{
public:
	SIRPRouter();
	~SIRPRouter()
	{
		delete manager;
	}

  /** Starts listening for connections on SIRP_PORT. Forwards connections to acceptConnection(). After UPDATE_INIT_DELAY_S time, calls updateCustomers(...).*/
	void start();
	/**
	@brief Attempts a connection to a SIRP neighbor.
	
	@param neighbor The neighbor to connect to.
	*/
	void openConnection(SIRPNeighbor *);
	/** 
	  @brief Configures the SIRPRouter.
	  
	  Sets the realmid, _port, and sets up _neighbors(in configAddNeighbor(...)).
	  
	  MM: Why is _port set in the constructor, and overridden as a config argument? Seems like this is confusing.
	  
	  @param filename The filename of the configuration file
	*/
	void config(str filename);

private:
  
  /** The port to listen for neighbor connections on. MM: This is set by both the constructor and config(...), and should only be set by one. */
	uint16_t _port;
	/** The FD which we listen for neighbor connections on */
	int _listenFD;
	/** Dictionary which maps public keys of neighbors to their corresponding SIRPNeighbor objects */
	neighborHash_t _neighbors;
	/** Dictionary which maps public keys of customers to their corresponding SIRPNeighbor objects(this is a subset of the _neighbors hash) */
	customerHash_t _customers;
	/** The realm id of the current node. Populated by config(...) */
	icp_pubkey _realmid;
	/** Callback object for customer updates(will invoke updateCustomers either immediately, or after some delay)*/
	timecb_t *_lastupdatecb;
	/** The last time we sent an update. */
	time_t _updateTimestamp;

  /** The update manager. This keeps track of all of the updates we recieve, and hands us a list of certificates/memcerts we can pass out to other realms. */
	UpdateManager *manager;

  /**
  @brief Dispatches out RPCs to the appropriate handles
  
  This method handles:
  
  IRP_OPEN: dispatched to handleOpen(...)
  IRP_UPDATE: dispatched to handleUpdate(...)
  IRP_KEEPALIVE: dispatched to handleKeepAlive(...)  
  
  @param srv The client which sent the request
  
  @param rpc The rpc which was just recieved
  */
	void dispatch(ptr<asrv>, svccb *);
	/** Accepts an incoming connection on SIRP_PORT, and forwards RPCs to dispatch(...) */
	void acceptConnection();
  /**
    @brief Handles an IRP_OPEN rpc
  
    Does some checks to make sure the neighbor is known and of the right type, then replies to the neighbor.
  
    Sets up a keepalive handler.
  
    Immediately calls openConnection(...) to open a TCP connection with the neighbor.
  
    @param srv The client which sent the request
  
    @param rpc The IRP_OPEN rpc which was just received
  */
	void handleOpen(ptr<asrv>, svccb *);
  /**
    @brief Handles an IRP_UPDATE rpc
  
    Updates the neighbor's SIRPNeighbor struct with updated certs, memcerts, and timestamp.
  
    Immediately calls updateCustomers(...) after updating the struct.
  
    @param rpc The IRP_UPDATE rpc which was just received
  */
	void handleUpdate(svccb *);
	/**
	  @brief Handles a IRP_KEEPALIVE rpc
	  
	  Sends an ACK to the neighbor; updates the keppAlive value of the neighbor.
	  
	  @param rpc The IRP_KEEPALIVE rpc which was just received
	*/
	void handleKeepAlive(svccb *);
  /**
    @brief Callback for when we achieve a successful connection to an IRP neighbor.
    
    After establishing the connection, we do an IRP_OPEN rpc on the neighbor.
    
    The callback is setup by openConnection(...).

    @param neighbor The neighbor which the router is now connected to

    @param fd The fd representing the connection to the neighbor
  */
	void connectionMade(SIRPNeighbor *, int);
  /**
    @brief Handles a reply RPC from an icing neighbor.
    
    Drops the neighbor if there are any errors with the RPC; forwards to parseReply(...) otherwise
    
    @param neighbor The neighbor which sent the RPC
  
    @param reply The RPC which the neighbor sent
  
    @param err The error of the RPC(if any)
    
  */
	void handleReply(SIRPNeighbor *, ref<irp_reply>, clnt_stat);
	/**
	  @brief Dispatches RPCs to the appropriate handler methods.
	  
	  If the RPC is a IRP_OPEN_ACK, it is forwarded to parseOpenAck(...). Otherwise, the RPC is ignored.
	  
  	@param neighbor The neighbor which sent the RPC
  	
  	@param reply The RPC which the neighbor sent
	*/
	void parseReply(SIRPNeighbor *, ref<irp_reply>);
	// XXX: MM: Need to add headerdocs for this function
  void parsePsOpenAck(SIRPNeighbor *neighbor, ref<irp_reply> reply);
	/**
	  @brief Handles an IRP_OPEN_ACK rpc
	  
	  Checks that the neighbor is of the appropriate type, then sets neighbor->status = SIRPNeighbor::UP, and sets up the callback handler.
	  
	  @param neighbor The neighbor which sent the RPC
	  
	  @param reply The RPC which the neighbor sent
	*/
	void parseOpenAck(SIRPNeighbor *, ref<irp_reply>);
  /**
    @brief Sends a keepalive message to a SIRP neighbor
    
    A keepalive message is sent every KEEPALIVE_DELAY_S seconds.
    
    @param neighbor The neighbor to send the keepalive message to
  */
	void sendKeepAlive(SIRPNeighbor *);
	/**
	  @brief Recieves a keepalive message from a SIRP neighbor.
	  
	  A keepalive message is expected every KEEPALIVE_TIMEOUT_S seconds. If a keepalive message is not recieved within this time, the neighbor is dropped.
	  
	  @param neighbor The neighbor we got(or expect to get) the keepalive message from
	*/
	void checkKeepAlive(SIRPNeighbor *);

  /**
    @brief Sends updates to each of the SIRP neighbors which are customers
    
    sendUpdate(...) actually sends out the update; this method just iterates over each customer and calls sendupdate(...) on it.
    
    Re-schedules a callback to updateCustomers(...) after another UPDATE_DELAY_S seconds.
    
    @param old The last callback. 
  */
	void updateCustomers(timecb_t *);
	/**
	  @brief Sends an IRP_UPDATE rpc to the neighbor
	  
	  The message is sent with the latest certificates from the UpdateManager
	  
	  @param neighbor The neighbor to send the update to.
	*/
	void sendUpdate(SIRPNeighbor *);
	/**
	  @brief Tells the UpdateManager to renew the certificates.
	  
    Schedules a callback to renewUpdate after UpdateManager::EXPIRY_TIME seconds.
	*/
	void renewUpdate();
  /**
    @brief Drops the SIRP neighbor
    
    Clears any state associated with the neighbor, removes any scheduled callbacks.
  */
	void dropNeighbor(SIRPNeighbor *);
  /**
    @brief Adds a SIRP neighbor from a conftab config file into _neighbors and _customers(where appropriate)
    
    @param args The arguments, in format "neighbor <realmid> <ip address> <port> <T | C | P>"(T == provider, C == customer, P == peer)
    
    @param loc The current output string. MM: Looks like this is just used for warn output, and nothing else. Is that its only purpose?
    
    @param err On return, true if an error was encountered parsing, false otherwise.
  */
	void configAddNeighbor(vec<str>, str, bool *);
  
  /** MM: This is sloppy. This function is never defined... Why is it here? */
	neighborHash_t &getHashTable(irp_neighbor_type &);
};

#endif /* SIRPROUTER_H_ */
