/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "updatemanager.h"
#include "util.h"

UpdateManager::UpdateManager(icp_pubkey &realm, neighborHash_t &neighbors) :
	_neighbors(neighbors), _realm(realm)
{
	icp_certificate cert;
	cert.realm.pubkey = _realm;

	cert.creation = time(NULL);
	cert.expiry = cert.creation + EXPIRY_TIME;

	memset(cert.prev.pubkey.base(), 0xFF, _realm.size());
	memset(cert.next.pubkey.base(), 0xFF, _realm.size());

	cert.prev.vq = htonl(CUSTOMER_VNODE);
	cert.next.vq = htonl(PROVIDER_VNODE);
	cert.realm.vq = htonl(C_P);
	_builtCerts.push_back(cert);

	cert.prev.vq = htonl(PROVIDER_VNODE);
	cert.next.vq = htonl(CUSTOMER_VNODE);
	cert.realm.vq = htonl(C_P);
	_builtCerts.push_back(cert);

	cert.prev.vq = htonl(CUSTOMER_VNODE);
	cert.next.vq = htonl(PEER_VNODE);
	cert.realm.vq = htonl(C_T);
	_builtCerts.push_back(cert);

	cert.prev.vq = htonl(PEER_VNODE);
	cert.next.vq = htonl(CUSTOMER_VNODE);
	cert.realm.vq = htonl(C_T);
	_builtCerts.push_back(cert);
}

void
UpdateManager::renewCerts()
{
	time_t creation, expiry;
	creation = time(NULL);
	expiry = creation + EXPIRY_TIME;

	for(int i = 0; i < _builtCerts.size(); i++)
	{
		icp_certificate &cert = _builtCerts[i];
		cert.creation = creation;
		cert.expiry = expiry;
	}

	for(int i = 0; i < _builtMemcerts.size(); i++)
	{
		icp_memcert &memcert = _builtMemcerts[i];
		memcert.creation = creation;
		memcert.expiry = expiry;
	}

	rebuildUpdate();
}

void
UpdateManager::updateCerts()
{
	_builtMemcerts.clear();
	_neighbors.traverse(
			wrap(this, &UpdateManager::buildMemcert));

	_receivedCerts.clear();
	_receivedMemcerts.clear();
	_neighbors.traverse(
			wrap(this, &UpdateManager::appendUpdate));

	rebuildUpdate();
}

void
UpdateManager::buildMemcert(SIRPNeighbor *neighbor)
{
	if(neighbor->status != SIRPNeighbor::UP)
		return;

	icp_memcert mc;
	mc.realm = _realm;
	mc.member = neighbor->realmid;
	mc.mask = MASK;

	switch(neighbor->type)
	{
		case IRP_CUSTOMER:
			mc.vq = CUSTOMER_VNODE;
			break;
		case IRP_PEER:
			mc.vq = PEER_VNODE;
			break;
		case IRP_PROVIDER:
			mc.vq = PROVIDER_VNODE;
			break;
		default:
			assert(false);
	}

	mc.creation = time(NULL);
	mc.expiry = mc.creation + EXPIRY_TIME;

	_builtMemcerts.push_back(mc);
}

void
UpdateManager::appendUpdate(SIRPNeighbor *neighbor)
{
	if(neighbor->status != SIRPNeighbor::UP)
		return;
	rpc_vec_append(certs, neighbor->certs);
	rpc_vec_append(memcerts, neighbor->memcerts);
}

void
UpdateManager::rebuildUpdate()
{
	certs.clear();
	rpc_vec_append(certs, _builtCerts);
	rpc_vec_append(certs, _receivedCerts);

	memcerts.clear();
	rpc_vec_append(memcerts, _builtMemcerts);
	rpc_vec_append(memcerts, _receivedMemcerts);
}
