/*
 * Copyright (C) 2009 Arun Seehra (aseehra@cs.utexas.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef UPDATEMANAGER_H_
#define UPDATEMANAGER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "async.h"
#include "arpc.h"
#include <time.h>
#include "icp_sirp.h"
#include "neighbor.h"

//MM: Need to find out where these constants come from and what exactly they mean.
#define MASK 0xFFFFFFFF
#define CUSTOMER_VNODE 26
#define PEER_VNODE 774
#define PROVIDER_VNODE 82
#define C_P 7264
#define C_T 1234 

/**
  This class manages keeping track of the updates we send out to neighbors(i.e. the certificates we are distributing), as well as the updates we receive from neighbors.
  
  The interface to the class is this:  
  - SIRPRouter invokes the constructor of this class, whcih takes in a reference to a neighborHash_t object(realmid -> SIRPNeighbor mapping).
  - This neighborHash_t hash is updated by SIRPRouter as it adds/removes neighbors
  - SIRPRouter calls updateCerts(...) when it gets an update. This triggers the arrays 'certs' and 'memcerts' to be updated. These arrays are concatenations of all of the available certs/memcerts from all realms
  - When constructing an update, SIRPRouter gets the contents of certs and memcerts, and sticks them in the update.
  

  The constructor also creates _builtCerts and _builtMemcerts, which contain certificates local to the current realm.
  
  MM: One (arguably sloppy) part of the design is that _neighbors is shared between SIRPRouter and UpdateManager, so updates are implicitly shared between the two. Better design: SIRPRouter always passes in its _neighbors array when calling updateCerts/renewCerts.
  
*/
class UpdateManager
{
public:
  /** The expiration time for certificates. */
	enum { EXPIRY_TIME = 120 };
	/**
	  @brief Constructs a new UpdateManager object
	  
	  @param realm The public key of the current realm(stored in _realm)
	  
	  @param neighbors A dictionary of realmid -> SIRPNeighbor mappings(stored in _neighbors)
	*/
	UpdateManager(icp_pubkey &, neighborHash_t &);
	/** A list of all of the certificates corresponding to _neighbors. MM: This is sloppy. Should be a getter!	*/
	icp_cert_vec certs;
	/** A list of all of the membership certificates corresponding to _neighbors. MM: This is sloppy. Should be a getter! */
	icp_memcert_vec memcerts;

  /**
    @brief Renews the (membership) certificates built by the update manager
    
    Sets the creation time to time(NULL), and the expiration to time(NULL) + EXPIRY_TIME. Then, calls rebuildUpdate().
  */
	void renewCerts();
	/**
	  @brief Rebuilds the certificates with all neighbors
	  
	  Clears all certificates, then calls buildMemcert(...) and appendUpdate(...) for each neighbor.	  
	*/
	void updateCerts();

private:
  /** The current realmid */
	icp_pubkey _realm;
	/** realmid -> SIRPNeighbor hash */
	neighborHash_t &_neighbors;

  /** Certificates which we are distributing to other realms */
	icp_cert_vec _builtCerts;
	/** MM: How is this written to? I don't see any code where items could be added to this vector. Presumably, this is the certificates we have recieved from other realms. This might be a bug. See comments in appendUpdate(...). */
	icp_cert_vec _receivedCerts;

  /** The membership certificates we are distributing to other realms. */
	icp_memcert_vec _builtMemcerts;
	/** MM: How is this written to? I don't see any code where items could be added to this vector. Presumably, this is the membership certificates we have recieved from other realms. This might be a bug. See comments in appendUpdate(...). */
	icp_memcert_vec _receivedMemcerts;

  /**
    @brief Build membership certificate which we are sharing with 'neighbor'
  
    @param neighbor The neighbor we are sharing the certificate with
  */
	void buildMemcert(SIRPNeighbor *neighbor);
	/**
	  @brief Append information from the update we have recieved from the neighbor
	  
	  @param neighbor The neighbor which sent the update
	  
	  Stores updates in 'certs' and 'memcerts' (ivars in this class)
	  
	  MM: Is it a bug that the certs are stored in 'certs' and 'memcerts'? It looks like they should be in '_receivedCerts' and '_receivedMemcerts'. The reason I suspect this is because updateCerts(...) - the only client of this method - calls this method, then calls rebuildUpdate() immediately after, which clears 'certs' and 'memcerts'.
	*/
	void appendUpdate(SIRPNeighbor *neighbor);
	/** @brief Clears certs, rebuilds it from appending _builtCerts and _receivedCerts. Then, clears memcerts, rebuilds it from _builtMemcerts and _receivedMemcerts */
	void rebuildUpdate();
};

#endif /* UPDATEMANAGER_H_ */
