#include "async.h"
#include "icinghdr.h"
#include "icingsec.h"
#include "tt.h"
#include <cstdlib>

FILE *out = NULL;

enum { NUM_PKTS = 3, NUM_LENS = 6, TIME_TO_RUN = 60, TIMES_TO_RUN = 10, NUM_REALMS = 100, PACKET_SIZE = 1500, NUM_PACKETS = 10000 };
uint32_t lens[] = {3, 7, 12, 20, 30, 37};
uint8_t *packets[NUM_PACKETS];

ecdh_priv me_p;
RealmSecret me_s;
RealmVoq me_r;

ecdh_priv all_realms[NUM_REALMS];

inline void
print_key(RealmIDWire &key)
{
	for(uint32_t i = 0 ; i < RealmIDWire::SIZE; i++)
		printf("%02X", key.bytes[i] & 0xFF);
	printf("\n");
	fflush(stdout);
}

void
generate_secret(RealmSecret &s)
{
	for(uint32_t i = 0; i < RealmSecret::SIZE; i++)
		s.bytes[i] = rand();
}

void
verify_pkt(uint8_t *packet)
{
	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(packet);

	advance_vtags(hdr1, me_p);

	if(hdr1->icing_ver == ICING_CURRENT_VERSION && hdr1->path_idx + 1 == hdr1->path_len)
		assert(validate_vtags(hdr1, &me_s, me_p));
}

void
generate_path(uint32_t len, ecdh_priv **privs, RealmVoq *realms)
{
	for(uint32_t i= 0; i < len - 1; i++)
	{
		realms[i].voq = 0;
		uint32_t cur = rand() % NUM_REALMS;
		ecdh_priv &p = all_realms[cur];

		privs[i] = &p;

		chopped_ecpk_raw raw;
		p.to_chopped_ecpk_raw(raw);
		memcpy(realms[i].realmid.bytes, raw, RealmIDWire::SIZE);
	}
	memcpy(&realms[len - 1], &me_r, sizeof(RealmVoq));
}

void
generate_pocs(uint32_t len, Path *path, POC *pocs)
{
	for(uint32_t i = 0; i < len - 2; i++)
	{
		RealmSecret s;
		generate_secret(s);
		poc_from_path(pocs[i], &s, path, len);
	}

	poc_from_path(pocs[len - 2], &me_s, path, len);
}

void
build_pkt(uint32_t len, uint8_t *&buf)
{
	buf = new uint8_t[PACKET_SIZE];

	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(buf);

	ecdh_priv **privs = new ecdh_priv*[len - 1];
	POC *pocs = new POC[len - 1];

	Path *path = get_path(hdr1);
	RealmVoq *realms = reinterpret_cast<RealmVoq *>(path);

	generate_path(len, privs, realms);
	generate_pocs(len, path, pocs);

	hdr1->icing_ver = ICING_CURRENT_VERSION;
	hdr1->path_len = static_cast<uint8_t>(len);
	hdr1->path_idx = 0;

	IcingHdr2 *hdr2 = get_icinghdr2(hdr1);
	hdr2->proto = 8;
	hdr2->pkt_len = htons(PACKET_SIZE);

	init_vtags(hdr1, pocs, *privs[0]);

	for(uint32_t i = 1; i < len - 1; i++)
	{
		advance_vtags(hdr1, *privs[i]);
	}

	delete[] privs;
	delete[] pocs;
}

void
build_all_pkts(uint32_t len)
{
	for(int i = 0; i < NUM_PACKETS; i++)
		build_pkt(len, packets[i]);
}

void
cleanup()
{
	for(int i = 0; i < NUM_PACKETS; i++)
		delete[] packets[i];
}

uint32_t
singleTrial()
{
	int i = 0;
	uint32_t count = 0;
	uint64_t start, end;

	start = get_now();
	end = start + (1000000 * TIME_TO_RUN);
	while(get_now() < end)
	{
		assert(i < NUM_PACKETS);
		verify_pkt(packets[i++]);
		count++;
	}

	return count;
}

int main(int argc, char **argv)
{
	srand((unsigned) time(NULL));
	if(argc > 1)
		out = fopen(argv[1], "w+");
	
	if(out)
	{
		fprintf(out, "# Pkt verify (throughput)\n#path len , throughput (packets/s)\n");
		fflush(out);
	}

	chopped_ecpk_raw raw;
	me_p.to_chopped_ecpk_raw(raw);
	memcpy(me_r.realmid.bytes, raw, RealmIDWire::SIZE);
	me_r.voq = 0;
	generate_secret(me_s);

	for(uint32_t i = 0; i < NUM_LENS; i++)
	{
		for(uint32_t j = 0; j < TIMES_TO_RUN; j++)
		{
			int len = lens[i];
			printf("Building packets of length %u ... ", len);
			fflush(stdout);
			build_all_pkts(len);
			printf("DONE\n");
			fflush(stdout);
			printf("Running trial for %d seconds ... ", TIME_TO_RUN);
			fflush(stdout);
			uint32_t count = singleTrial();
			double tp = count * 1.0;
			tp /= TIME_TO_RUN;

			if(out)
			{
				fprintf(out, "%u, %f\n", len, tp);
				fflush(out);
			}
			printf("DONE: %u, %f\n", count, tp);
			fflush(stdout);

			printf("cleaning up ...");
			fflush(stdout);
			cleanup();
			printf("DONE\n\n");
			fflush(stdout);
		}
	}
}
