#include "async.h"
#include "icinghdr.h"
#include "icingsec.h"
#include "tt.h"
#include <cstdlib>
#include "localkey.h"

FILE *out = NULL;

enum { NUM_PKTS = 3, NUM_LENS = 6, TIMES_TO_RUN = 100, NUM_REALMS = 100, PACKET_SIZE = 1500};
uint32_t lens[] = {3, 7, 12, 20, 30, 37};

ecdh_priv me_p;
RealmSecret me_s;
RealmVoq me_r;

ecdh_priv all_realms[NUM_REALMS];

keyhash_t hash;

void
key_calc(SharedKey &d, const RealmIDWire &rid, const RealmIDWire *l = NULL)
{
	LocalKey *k = hash[rid];
	assert(k);
	memcpy(d, k->k, sizeof(SharedKey));
}

inline void
print_key(RealmIDWire &key)
{
	for(uint32_t i = 0 ; i < RealmIDWire::SIZE; i++)
		printf("%02X", key.bytes[i] & 0xFF);
	printf("\n");
	fflush(stdout);
}

void
generate_secret(RealmSecret &s)
{
	for(uint32_t i = 0; i < RealmSecret::SIZE; i++)
		s.bytes[i] = rand();
}

void
verify_pkt(uint8_t *packet)
{
	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(packet);

	if(hdr1->icing_ver == ICING_CURRENT_VERSION && hdr1->path_idx + 1 == hdr1->path_len)
	{
		assert(validate_vtags(hdr1, &me_s, me_p, &key_calc));
	}
}

void
generate_path(uint32_t len, ecdh_priv **privs, RealmVoq *realms)
{
	for(uint32_t i= 0; i < len - 1; i++)
	{
		realms[i].voq = 0;
		uint32_t cur = rand() % NUM_REALMS;
		ecdh_priv &p = all_realms[cur];

		privs[i] = &p;

		chopped_ecpk_raw raw;
		p.to_chopped_ecpk_raw(raw);
		memcpy(realms[i].realmid.bytes, raw, RealmIDWire::SIZE);
	}
	memcpy(&realms[len - 1], &me_r, sizeof(RealmVoq));
}

void
generate_pocs(uint32_t len, Path *path, POC *pocs)
{
	for(uint32_t i = 0; i < len - 2; i++)
	{
		RealmSecret s;
		generate_secret(s);
		poc_from_path(pocs[i], &s, path, len);
	}

	poc_from_path(pocs[len - 2], &me_s, path, len);
}

void
build_pkt(uint32_t len, uint8_t *&buf)
{
	buf = new uint8_t[PACKET_SIZE];

	IcingHdr1 *hdr1 = reinterpret_cast<IcingHdr1 *>(buf);

	ecdh_priv **privs = new ecdh_priv*[len - 1];
	POC *pocs = new POC[len - 1];

	Path *path = get_path(hdr1);
	RealmVoq *realms = reinterpret_cast<RealmVoq *>(path);

	generate_path(len, privs, realms);
	generate_pocs(len, path, pocs);

	hdr1->icing_ver = ICING_CURRENT_VERSION;
	hdr1->path_len = static_cast<uint8_t>(len);
	hdr1->path_idx = 0;

	IcingHdr2 *hdr2 = get_icinghdr2(hdr1);
	hdr2->proto = 8;
	hdr2->pkt_len = htons(PACKET_SIZE);

	init_vtags(hdr1, pocs, *privs[0]);
	hdr1->path_idx++;

	for(uint32_t i = 1; i < len - 1; i++)
	{
		advance_vtags(hdr1, *privs[i]);
		hdr1->path_idx++;
	}
//	advance_vtags(hdr1, me_p);
//	hdr1->path_idx++;

	delete[] privs;
	delete[] pocs;
}

void
build_shared_keys()
{
    for(int i = 0; i < NUM_REALMS; i++)
    {
	RealmIDWire rid;
	LocalKey *k = new LocalKey;

	ecdh_priv &p = all_realms[i];
	chopped_ecpk_raw raw;
	p.to_chopped_ecpk_raw(raw);
	memcpy(k->rid.bytes, raw, RealmIDWire::SIZE);

	set_sharedkey(k->k, k->rid, me_p);
	hash.insert(k);

    }
}

int main(int argc, char **argv)
{
	srand((unsigned) time(NULL));
	if(argc > 1)
		out = fopen(argv[1], "w+");
	
	if(out)
	{
		fprintf(out, "# Pkt verify (latency)\n#path len , latency (us)");
		fflush(out);
	}

	chopped_ecpk_raw raw;
	me_p.to_chopped_ecpk_raw(raw);
	memcpy(me_r.realmid.bytes, raw, RealmIDWire::SIZE);
	me_r.voq = 0;
	generate_secret(me_s);

	uint8_t *buf;

	printf("Building shared keys ... ");
	fflush(stdout);
	build_shared_keys();
	printf("DONE\n\n");
	fflush(stdout);

	for(uint32_t i = 0; i < NUM_LENS; i++)
	{
		uint64_t min = 0;
		printf("Running experiment for path length %u ... ", lens[i]);
		fflush(stdout);

		for(uint32_t j = 0; j < TIMES_TO_RUN; j++)
		{
			build_pkt(lens[i], buf);
			uint64_t before = get_now();
			verify_pkt(buf);
			uint64_t after = get_now();

			uint64_t delta = after - before;
			if(min == 0 || delta < min)
				min = delta;
			if(out)
			{
				fprintf(out, "%u, %llu\n", lens[i], delta);
				fflush(out);
			}
			delete[] buf;
		}
		printf("min = %llu\n", min);
		fflush(stdout);
	}

	delete buf;
}
