#ifndef _ICING_HDR_H_
#define _ICING_HDR_H_

#include "bigint.h"

//#pragma pack(1)  XXX: MM: This line has caused me a rediculous amount of trouble!

const uint16_t ICING_ETHTYPE = 0x9090;

// Icing packet layout. 
// This file should not depend on any others.

const uint8_t ICING_CURRENT_VERSION = 1;
const uint8_t IV_SIZE = 7;

// Packet layout:
// | IcingHdr1    |
// | RealmList    |
// | vtags        |
// | IcingHdr2    |
//
//  Note that the length of the icing header includes the path
//  and vtag entries.
struct IcingHdr1 {
    uint8_t icing_ver;
    uint8_t  path_len;
    uint8_t  path_idx;
    char   init_vec[IV_SIZE];
} __attribute__((packed));

struct IcingHdr2 {
    uint8_t proto;     // what the packet is carrying
    uint16_t pkt_len;  // total length. network byte order.
} __attribute__((packed));

struct RealmIDWire {
    static const uint32_t SIZE = 20;
    uint8_t bytes[SIZE];
    RealmIDWire() { 
	memset(bytes, 0, SIZE); 
    }
private:
    RealmIDWire(const RealmIDWire&);
    RealmIDWire& operator=(const RealmIDWire&);
} __attribute__((packed));

struct RealmVoq {
    RealmIDWire realmid;
    uint32_t    voq;
} __attribute__((packed));

struct Vtag {
    char va[12];
    char vb[4];
} __attribute__((packed));

struct Path {
    RealmVoq realmvoqs[];
} __attribute__((packed));

struct VtagVector {
    Vtag vtags[];
} __attribute__((packed));

uint32_t icing_hdrlen(uint8_t nrealms);
uint32_t icing_hdrlen(const IcingHdr1* h);
uint16_t icing_pktlen(const IcingHdr1* h);

const Path* get_path(const IcingHdr1*);
Path* get_path(IcingHdr1*);

const VtagVector* get_vtagvector(const IcingHdr1*);
VtagVector* get_vtagvector(IcingHdr1*);

const IcingHdr2* get_icinghdr2(const IcingHdr1*);
IcingHdr2* get_icinghdr2(IcingHdr1*);

const uint8_t* get_payload(const IcingHdr1*);
uint8_t* get_payload(IcingHdr1*);

inline uint32_t
icing_hdrlen(const IcingHdr1* h)
{
    uint8_t nrealms = h->path_len;
    return icing_hdrlen(nrealms);
}

inline uint32_t
icing_hdrlen(uint8_t nrealms)
{
    uint32_t r = sizeof(IcingHdr1) + 
	   (nrealms*sizeof(RealmVoq)) +
	   ((nrealms-1)*sizeof(Vtag)) +
	   sizeof(IcingHdr2);
    return r;
}

inline uint16_t
icing_pktlen(const IcingHdr1* h)
{
    const IcingHdr2* h2 = get_icinghdr2(h);
    return ntohs(h2->pkt_len);
}


inline const Path*
get_path(const IcingHdr1* h)
{
    return 
	reinterpret_cast<const Path*>(
		reinterpret_cast<const char*>(h) + sizeof(IcingHdr1));
}

inline Path*
get_path(IcingHdr1* h)
{
    return
	reinterpret_cast<Path*>(
		reinterpret_cast<char*>(h) + sizeof(IcingHdr1));
}


inline const VtagVector*
get_vtagvector(const IcingHdr1* h)
{
    uint8_t nrealms = h->path_len;
    const char* vv = 
	reinterpret_cast<const char*>(h) + 
	    sizeof(IcingHdr1) + (nrealms*sizeof(RealmVoq));
    return reinterpret_cast<const VtagVector*>(vv);
}

inline VtagVector*
get_vtagvector(IcingHdr1* h)
{
    uint8_t nrealms = h->path_len;
    char* vv = 
	reinterpret_cast<char*>(h) + 
	    sizeof(IcingHdr1) + (nrealms*sizeof(RealmVoq));
    return reinterpret_cast<VtagVector*>(vv);
}

inline const IcingHdr2*
get_icinghdr2(const IcingHdr1* h)
{
    uint8_t nrealms = h->path_len;
    const char* hdr2 = 
	reinterpret_cast<const char*>(h) + 
	    sizeof(IcingHdr1) + 
		(nrealms*sizeof(RealmVoq)) + 
		    ((nrealms-1)*sizeof(Vtag));

    return reinterpret_cast<const IcingHdr2*>(hdr2);

}

inline IcingHdr2*
get_icinghdr2(IcingHdr1* h)
{
    uint8_t nrealms = h->path_len;
     char* hdr2 = 
	reinterpret_cast<char*>(h) + 
	    sizeof(IcingHdr1) + 
		(nrealms*sizeof(RealmVoq)) + 
		    ((nrealms-1)*sizeof(Vtag));

    return reinterpret_cast<IcingHdr2*>(hdr2);
}

inline const uint8_t*
get_payload(const IcingHdr1* h)
{
    const uint8_t* p;
    p = reinterpret_cast<const uint8_t*>(get_icinghdr2(h)) + sizeof(IcingHdr2) + 1;
    return p;
}

inline uint8_t*
get_payload(IcingHdr1* h)
{
    uint8_t* p;
    p = reinterpret_cast<uint8_t*>(get_icinghdr2(h)) + sizeof(IcingHdr2) + 1;
    return p;
}

inline bool
operator<(const RealmIDWire &r1, const RealmIDWire &r2)
{
	for(int i = 0; i < RealmIDWire::SIZE; i++)
	{
		if(r1.bytes[i] < r2.bytes[i])
			return true;
		else if(r1.bytes[i] > r2.bytes[i])
			return false;
	}

	return false;
}

#endif
