// implements a number of the security functions for ICING
#ifndef _ICING_SEC_HH_
#define _ICING_SEC_HH_

#include <string.h>
#include "ecdh.h"
#include "mac.h"
#include <map>

extern "C"
{
#include "chi.h"
}

typedef uint8_t MACKEY[16];
typedef MACKEY  SharedKey;
typedef MACKEY  POC;

#define HASHLEN 32

struct RealmSecret {
  enum { SIZE = sizeof(MACKEY) };
  //    static const uint32_t SIZE = sizeof(MACKEY);
    uint8_t bytes[SIZE];
    RealmSecret() { 
	memset(bytes, 0, SIZE); 
    }
private:
    RealmSecret(const RealmSecret&);
    RealmSecret& operator=(const RealmSecret&);
};

struct ReturnPocs {
	uint8_t num_rpocs;
	POC rpocs[];
};

// API
typedef void (*keycalc_ptr)(SharedKey &, const RealmIDWire &, const RealmIDWire *);
void poc_from_path(POC, const RealmSecret*, const Path*, uint32_t nrealms);
void derive_poc_key(const RealmSecret &, const uint32_t, const uint32_t, 
	const uint32_t, RealmSecret &);
void set_sharedkey(SharedKey &, const RealmIDWire &, ecdh_priv &);
void init_vtags(IcingHdr1*, const RealmSecret*, const uint32_t *, 
		const uint32_t *, ecdh_priv &, keycalc_ptr = NULL);
void init_vtags(IcingHdr1*, const POC*, ecdh_priv &, keycalc_ptr = NULL);
void advance_vtags(IcingHdr1*, ecdh_priv &, keycalc_ptr = NULL);
bool validate_vtags(const IcingHdr1* hdr1, const RealmSecret*, ecdh_priv &, keycalc_ptr = NULL);

inline void
poc_from_path(POC poc, const RealmSecret* k, 
	      const Path* path, uint32_t nrealms)
{
    pmac(k->bytes, poc, reinterpret_cast<const uint8_t *>(path->realmvoqs),
		    nrealms*sizeof(RealmVoq));
}

inline void
set_bit(uint8_t *buf, uint32_t bit_num)
{
    uint32_t byte = bit_num >> 3;
    uint32_t bit = bit_num % 8;
    buf[byte] |= (0x80 >> bit);
}

// assumes that the vnode is in network byte order
inline void
derive_poc_key(const RealmSecret &k, const uint32_t g, const uint32_t v, const uint32_t vnode, 
		RealmSecret &s)
{
#ifndef DEBUG
    aes_e ae;
    uint32_t vnode_size = sizeof(uint32_t) << 3;

    assert(g && (g <= vnode_size) && ((g & (g - 1)) == 0)); // sanity check on g
    assert(v % g == 0); // sanity check on v
    assert(v > g);

    uint8_t d = v / g;

    memcpy(s.bytes, k.bytes, RealmSecret::SIZE);

    MACKEY tag;
    bzero(tag, RealmSecret::SIZE);

    for(uint32_t t = 0; t < d; t++)
    {
	ae.setkey(s.bytes, RealmSecret::SIZE);

	// isolate the g bits that we care about
	// NOTE: the bit with index 0 is the *leftmost* bit, i.e., MSB, of vnode.
	//
	//    code below takes the g bits of vnode in the positions
	//    between start_idx and end_idx-1, inclusive, and puts them
	//    in the most significant bits of vnode_slice.
	uint32_t vnode_slice = ntohl(vnode);
	uint32_t start_idx = t*g;
	uint32_t end_idx = (t+1) * g;
	vnode_slice &= ~((1 <<( vnode_size - end_idx)) - 1);
	vnode_slice <<= start_idx; 

	// copy vnode_slice into temporary buffer
	vnode_slice = htonl(vnode_slice);
	memcpy(tag, &vnode_slice, sizeof(uint32_t));
	set_bit(tag, g);

	ae.encipher_bytes(s.bytes, tag);
    }
#else
  bzero(s.bytes, RealmSecret::SIZE);
#endif
}

// Used to store opaque bytes as a key in a map
class OpaqueBytes {
 public:
  OpaqueBytes(uint8_t *data, unsigned size) {
    _data = (uint8_t *)malloc(size);
    memcpy(_data, data, size);
    _size = size;
  };
  ~OpaqueBytes() { free(_data); } 

  OpaqueBytes(const OpaqueBytes& ob){
    _data = (uint8_t *)malloc(ob._size);
    _size = ob._size;
    memcpy(_data, ob._data, _size);
  }

  bool operator==(const OpaqueBytes& ob) const {
    return (ob._size == _size) && (memcmp(ob._data, _data, _size) == 0);
  };

  bool operator<(OpaqueBytes const& ob) const {
    int r = memcmp(_data, ob._data, MIN(_size, ob._size));
    return ( r < 0 ) || ( r == 0 && _size < ob._size);
  }
  
  uint8_t *data(){
    return _data;
  }
  unsigned size(){
    return _size;
  }
 private:
  uint8_t *_data;
  unsigned _size;
};

inline void
set_sharedkey(SharedKey &shkey, const RealmIDWire &r1, ecdh_priv &priv)
{
#ifndef DEBUG
  bzero(shkey, sizeof(SharedKey));

  //Caching logic
  bigint bi = priv.getraw();
  str rawInt = bi.getraw();
  
  uint8_t bytes[sizeof(SharedKey) + rawInt.len()];
  
  memcpy(bytes, r1.bytes, sizeof(SharedKey));
  memcpy(bytes+sizeof(SharedKey), (uint8_t *)rawInt.cstr(), rawInt.len());
  
  OpaqueBytes b(bytes, sizeof(SharedKey) + rawInt.len());
  static std::map<OpaqueBytes, OpaqueBytes> cache; // Maps r1 || priv -> SharedKey
  std::map<OpaqueBytes, OpaqueBytes>::iterator it = cache.find(b);
  if(it != cache.end()){
    //Hit in cache!
    //    printf("Cache hit!\n");
    uint8_t *sharedKey = it->second.data();
    for(int i=0; i < sizeof(SharedKey); i++)
      shkey[i] = sharedKey[i];
    return;
  }
  //  printf("Cache not hit..\n");

    chopped_ecpk_raw realmraw;
    memcpy(realmraw, &r1.bytes[0], RealmIDWire::SIZE);

    ecdh_pub pub(realmraw);


    ec_nidh(shkey, sizeof(SharedKey), &priv, &pub);
    
    OpaqueBytes sharedKeyBytes(shkey, sizeof(SharedKey));
    cache.insert( std::pair<OpaqueBytes, OpaqueBytes>(b, sharedKeyBytes) );
#else
    chopped_ecpk_raw raw;
    priv.to_chopped_ecpk_raw(raw);

    for(int i = 0; i < sizeof(SharedKey); i++)
	shkey[i] = r1.bytes[i] ^ ((uint8_t *)raw)[i];
#endif
}

inline void
set_realmid(RealmIDWire &wire, ecdh_priv &priv)
{
  bigint bi = priv.getraw();
  str rawInt = bi.getraw();
  
  uint8_t bytes[rawInt.len()];
  static std::map<OpaqueBytes, OpaqueBytes> cache; // Maps priv -> RealmIDWire
  OpaqueBytes b(bytes, rawInt.len());
  std::map<OpaqueBytes, OpaqueBytes>::iterator it = cache.find(b);
  if(it != cache.end()){
    //Hit in cache!
    //    printf("Cache hit!\n");
    uint8_t *wireBytes = it->second.data();
    memcpy(wire.bytes, wireBytes, RealmIDWire::SIZE);
    return;
  }
  
    chopped_ecpk_raw raw;
    priv.to_chopped_ecpk_raw(raw);
    memcpy(wire.bytes, raw, RealmIDWire::SIZE);
    
    OpaqueBytes wireBytes(wire.bytes, (unsigned)RealmIDWire::SIZE);
    cache.insert( std::pair<OpaqueBytes, OpaqueBytes>(b, wireBytes) );
}

inline uint32_t
construct_buf(uint8_t** ver_iv_payload, 
	      const IcingHdr1* hdr1)
{
    const IcingHdr2* hdr2 = get_icinghdr2(hdr1);
    uint32_t payloadlen = ntohs(hdr2->pkt_len) - icing_hdrlen(hdr1) +
	                  sizeof(hdr2->proto) + sizeof(hdr2->pkt_len);

    uint32_t pathlen = hdr1->path_len * sizeof(RealmVoq);

    uint32_t buflen = sizeof(hdr1->icing_ver) + IV_SIZE + pathlen + payloadlen;
    *ver_iv_payload = new uint8_t[buflen];
    bzero(*ver_iv_payload, buflen);

    **ver_iv_payload = hdr1->icing_ver;
    memcpy(*ver_iv_payload + 1, hdr1->init_vec, IV_SIZE);
    memcpy(*ver_iv_payload + 1 + IV_SIZE, get_path(hdr1), pathlen);
    memcpy(*ver_iv_payload + 1 + IV_SIZE + pathlen, hdr2, payloadlen);

    return buflen;
}

inline void
construct_hash_247(uint8_t H[], const IcingHdr1 *hdr1)
{
    uint8_t *buf;
    uint32_t buflen = construct_buf(&buf, hdr1);

#ifdef DEBUG
    printf("buf: ");
    for(int i = 0; i < buflen; i++)
	    printf("%02x:", buf[i]);
    printf("\n");
    printf("buflen: %d\n", buflen << 3);
#endif

    CHIHash(HASHLEN << 3, buf, buflen << 3, H);

#ifdef DEBUG
    printf("Hash: ");
    for(int i = 0; i  < HASHLEN; i++)
	printf("%02x:", H[i]);
    printf("\n");
#endif
   
    zero_top9(H);

    delete [] buf;
}

inline void
init_vtags(IcingHdr1* h1, const RealmSecret* k, const uint32_t *g, const uint32_t *v, 
		ecdh_priv &priv, keycalc_ptr func)
{
    Path* path = get_path(h1);
    uint32_t npocs = h1->path_len - 1;

    // calculate POC = PMAC1(k_j, path)
    POC* pocs = new POC[npocs];
    for (uint32_t j = 0; j < npocs; j++) 
    {
	RealmSecret key;
	uint32_t vnode;

	vnode = path->realmvoqs[j+1].voq;
	derive_poc_key(k[j+1], g[j+1], v[j+1], vnode, key);
	poc_from_path(pocs[j], &key, path, h1->path_len);
    }

    init_vtags(h1, pocs, priv, func);
    delete [] pocs;
}

inline void
init_vtags(IcingHdr1* h1, const POC* pocs, ecdh_priv &priv, keycalc_ptr func)
{
    uint32_t npocs = h1->path_len - 1;
    VtagVector* vv = get_vtagvector(h1);

    // for j = 1..n, calculate A[j]
    // H = last 247 bits of CHI(vers || IV || path || payload)
    // A[j] = CBCMAC-128(POC[j], 0^9 || H)
    uint8_t H[HASHLEN];
    construct_hash_247(H, h1);

    for (uint32_t j = 0; j < npocs; j++) 
    {
	Vtag *vt = &vv->vtags[j];
	cbcmac_128(pocs[j], H, reinterpret_cast<uint8_t *>(vt));
#ifdef DEBUG
	printf("A=");
	for(int i = 0; i < sizeof(Vtag); i++)
	    printf("%02x:", ((uint8_t *)vt)[i]);
	printf("\n");
#endif

	memcpy(reinterpret_cast<uint8_t *>(vt->vb),&(reinterpret_cast<uint8_t *>(vt->va))[8], 4);
    }
		

    // vtags[i] now holds A[i]. so we can just call the regular
    // "advance" function to output V0. (This is really regarding the
    // A[i] as V_{-1}.)
    advance_vtags(h1, priv, func);
}

inline bool
validate_vtags(const IcingHdr1* hdr1, const RealmSecret* k, ecdh_priv &priv, keycalc_ptr func)
{
    RealmIDWire id;
    set_realmid(id, priv);
    const Path* path = get_path(hdr1);
    const VtagVector* vv = get_vtagvector(hdr1);

    POC poc;
    poc_from_path(poc, k, path, hdr1->path_len);

    // calculate A[i]
    // H = last 247 bits of CHI(vers || IV || path || payload)
    // A[i] = CBCMAC-128(POC, 0^9 || H)
    uint8_t myA[16], H[HASHLEN];
    construct_hash_247(H, hdr1);

    cbcmac_128(poc, H, myA);

    // Check that Vb_{i-1}[i] == last 4 bytes of A[i].
    //   Note V_{i-1} is just the values of the vtags, as they were emitted 
    //   by our predecessor on the path.
    //   So Vb_{i-1}[i] is the b component of the vtag entry that
    //   corresopnds to us.
    const Vtag* vt = &vv->vtags[hdr1->path_idx-1];
    if (memcmp(vt->vb, &myA[8], 4))
	return false;

    SharedKey d;
    for (uint8_t j = 0; j < hdr1->path_idx; j++) {
	if(func)
	    (*func)(d, path->realmvoqs[j].realmid, &path->realmvoqs[hdr1->path_idx].realmid);
	else
	    set_sharedkey(d, path->realmvoqs[j].realmid, priv);

	concat_255(H, j);

	block96 temp;
	prf2_96(path->realmvoqs[j].realmid, id, d, H, temp);
	xor_96(myA, myA, temp);

	// Reduce H back to 247 bits
	zero_top9(H);
    }

    if (memcmp(vt->va, myA, 12))
	return false;

    return true;
}

inline void
advance_vtags(IcingHdr1* h1, ecdh_priv &priv, keycalc_ptr func)
{
    RealmIDWire id;
    set_realmid(id, priv);
    uint8_t path_idx = h1->path_idx;
    uint32_t npocs = h1->path_len - 1;
    const Path* path = get_path(h1);
    VtagVector* vv = get_vtagvector(h1);

    uint32_t j;

//    for (j = 0; j < path_idx; j++) {
//	Vtag* vt = &vv->vtags[j];
//	memset(vt, 0, sizeof(Vtag));
//    }

    uint8_t H[HASHLEN];
    construct_hash_247(H, h1);

    concat_255(H, path_idx);

    SharedKey d;
    for (j = path_idx; j < npocs; j++) {
	
	if(func)
	    (*func)(d, path->realmvoqs[j+1].realmid, &path->realmvoqs[path_idx].realmid);
	else
	    set_sharedkey(d, path->realmvoqs[j+1].realmid, priv);

#ifdef DEBUG
	printf("Key shared with %d: ", j);
	for(int i = 0; i < sizeof(SharedKey); i++)
	    printf("%02x:", d[i]);
	printf("\n");
#endif

    	Vtag* vt = &vv->vtags[j];

	block96 temp;
	uint8_t *va = reinterpret_cast<uint8_t *>(vt->va);
	prf2_96(id, path->realmvoqs[j+1].realmid, d, H, temp);
	
#ifdef DEBUG
	printf("Temp for %d: ", j);
	for(int i = 0; i < sizeof(temp); i++)
	    printf("%02x:", temp[i]);
	printf("\n");
#endif

	xor_96(va, temp, va);
    }
}

#endif
