/*
 * pmac implementation from:
 * ocb.c-- Implemented by Ted Krovetz (tdk@acm.org) -- Modified 2005.03.11
 * This implementation is in the public domain. No warranty applies.
 *
 */

#ifndef MAC_H_
#define MAC_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "aes.h"

typedef unsigned char byte;
#define BLOCKLEN 16                        /* AES block-length in bytes    */
typedef byte block[BLOCKLEN];              /* an AES block                 */
#define BLOCKLEN96 12
typedef byte block96[BLOCKLEN96];

/*************************************************************************** 
 * API
 ***************************************************************************/
void pmac(const byte *K, block out, const byte *M, unsigned mlen);
void cbcmac_128(const byte *K, const byte *in, block out);
void prf2_96(const RealmIDWire &r1, const RealmIDWire &r2, const byte *K,
		byte *in, block96 out);

void xor_block(block dst, const block src1, const block src2);
void xor_96(block96 dst, const block96 src1, const block96 src2);

void concat_255(uint8_t *dst, const uint8_t &src);
void zero_top9(uint8_t *dst);

/*************************************************************************** 
 * xor_block, xor_96  
 ***************************************************************************/
inline void
xor_block(block dst, const block src1, const block src2)
{
    unsigned i;
    for (i = 0; i < sizeof(block); i++) {
        dst[i] = src1[i] ^ src2[i];
    }
}

inline void
xor_96(block96 dst, const block96 src1, const block96 src2)
{
	for (unsigned i = 0; i < sizeof(block96); i++)
		dst[i] = src1[i] ^ src2[i];
}

/*************************************************************************** 
 * concat_255: assumes that the first 9 bits of dst are 0
 * zero_top9
 ***************************************************************************/
inline void
concat_255(uint8_t *dst, const uint8_t &src)
{
    // MW: XXX:this is backward, right?
    // the first bit in dst is the lower-order bit of dst[0], which is
    // the *rightmost* bit of dst[0]. thus, I think you want:
    //     dst[0] = (src << 1); // leaves rightmost bit clear
    //     dst[1] |= (src >> 7);
    //  no?
	dst[0] = (src >> 1);
	dst[1] |= (src << 7);
}

inline void
zero_top9(uint8_t *dst)
{
	dst[0] = 0;
	dst[1] &= 0x7F; // MW: XXX: should be dst[1] &= 0xfe, right?
			// MW: XXX: can you check to find out what Jad
			// is implementing. usually people pretend that
			// the first byte of an array is the low-order
			// byte, and it's an invariant that the lowest
			// order bit is the rightmost bit within the
			// byte
}

/*************************************************************************** 
 * cbcmac_128
 ***************************************************************************/
inline void
cbcmac_128(const byte *K, const byte *in, block out)
{
	const block *bin = reinterpret_cast<const block *>(in);
	aes_e ae;
	block x;
	
	ae.setkey(K, 16);
	ae.encipher_bytes(x, bin[0]);
	xor_block(x, x, bin[1]);
	ae.encipher_bytes(out, x);
}

/*************************************************************************** 
 * prf2_96
 * 
 * Requires that <in> is [0 | 255 bits of data]
 ***************************************************************************/
inline void
prf2_96(const RealmIDWire &r1, const RealmIDWire &r2, const byte *K,
		byte *in, block96 out)
{
	byte tag;
	block temp;

	// MW: XXX: I know we discussed this a while ago, but I think
	// the 1 is in the wrong place, no? I mean, it's not wrong if
	// we're consistent, but it's confusing. Essentially, if you're
	// prefixing the block by a tag, don't you want that tag in the
	// *lowest* bit position?
	assert( (in[0] & 0x80) == 0);
	tag = (r1 < r2) ? 0x80 : 0;
	
	in[0] |= tag;
	cbcmac_128(K, in, temp);
	in[0] &=0x7F; // MW: XXX: revist
	memcpy(out, temp, sizeof(block96));
}

/*************************************************************************** 
 * two_times, three_times
 *
 * These routines manipulate the offsets which are used for pre- and
 * post-whitening of blockcipher invocations. The offsets represent
 * polynomials, and these routines multiply the polynomials by other
 * constant polynomials. Note that as an optimization two consecutive
 * invocations of "three_times" can be efficiently replaced:
 *                     3(3(X)) == (2(2(X))) xor X
 *
 ***************************************************************************/
inline void
two_times(block dst, block src)
/* 128-bit shift-left by 1 bit: src <<= 1. Xor 0x87 if carry out.          */
{
    unsigned i;
    unsigned char carry = src[0] >> 7;      /* set carry = high bit of src */
    for (i = 0; i < sizeof(block)-1; i++) {
        dst[i] = (src[i] << 1) | (src[i+1] >> 7);
    }
    dst[sizeof(block)-1] = (src[sizeof(block)-1] << 1) ^ (carry * 0x87);
}

inline void
three_times(block dst, block src)
{
    block t;
    
    two_times(t,src);
    xor_block(dst, t, src);
}

/*************************************************************************** 
 * pmac_function  
 ***************************************************************************/
inline void
pmac(const byte *K, block T, const byte *M, unsigned mlen)
{
	aes_e ae;
	ae.setkey(K, 16);

    block Tmp,Offset,Checksum;
    block *mptr = (block *)M;
    
   /*
    * Initialize strings used for offsets and checksums
    */
    memset(Checksum,0,sizeof(block));
	ae.encipher_bytes(Offset, Checksum);
    three_times(Offset,Offset);
    three_times(Offset,Offset);

   /*
    * Accumulate all but the last block
    */
    while (mlen > sizeof(block)) {
        two_times(Offset,Offset);
        xor_block(Tmp,Offset,*mptr);
		ae.encipher_bytes(Tmp, Tmp);
        xor_block(Checksum,Tmp,Checksum);
        mlen = mlen - sizeof(block);
        mptr = mptr + 1;
    }

   /*
    * Accumulate the last block
    */
    two_times(Offset,Offset);
    if (mlen == sizeof(block)) {
        xor_block(Checksum,*mptr,Checksum);
        three_times(Offset,Offset);
    } else {
        memset(Tmp,0,sizeof(block));
        memcpy(Tmp,*mptr,mlen);
        Tmp[mlen] = 0x80;
        xor_block(Checksum,Tmp,Checksum);
        three_times(Offset,Offset);
        three_times(Offset,Offset);
    }

   /*
    * Compute result
    */
    xor_block(Tmp,Offset,Checksum);
	ae.encipher_bytes(T, Tmp);
}

#endif /* PMAC_H_ */
