/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: aes_decrypt.v
 * Project: ICING
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Does decryption without needing the last round key. Runs at 1.6Gbps.
 *              
 ******************************************************************************/
`timescale 1ns/1ps

module aes_decrypt_single
  #(parameter WIDTH = 128)
    (input [0:WIDTH-1]      block_in,
     input                  block_in_vld,

     output[0:WIDTH-1]      block_out,
     output                 block_out_vld,
     output                 rdy,

     input [0:WIDTH-1]      key,

     input                  clk,
     input                  reset);

   wire [0:WIDTH-1] block_in_delayed, last_round_key;

   aes_single #(.INVERSE(0), .ONLY_KEY_EXPANSION(1)) aes_key_expansion
     (// Inputs
      .key                              (key),
      .block_in                         (block_in),
      .block_in_vld                     (block_in_vld),

      // Outputs
      .block_out                        (block_in_delayed),
      .block_out_vld                    (block_in_delayed_vld),
      .last_round_key                   (last_round_key),
      .rdy                              (rdy),

      .clk                              (clk),
      .reset                            (reset));

   aes_single #(.INVERSE(1)) aes_decipher
     (// Inputs
      .key                              (last_round_key),
      .block_in                         (block_in_delayed),
      .block_in_vld                     (block_in_delayed_vld),

      // Outputs
      .block_out                        (block_out),
      .block_out_vld                    (block_out_vld),
      .last_round_key                   (),
      .rdy                              (),

      .clk                              (clk),
      .reset                            (reset));

endmodule

// synthesis translate_off
module aes_decrypt_single_test();

   localparam WIDTH = 128;
   
   reg [0:WIDTH-1]  block_in = 0;
   reg [0:WIDTH-1]  key = 0;
   reg              block_in_vld = 0;

   wire [0:WIDTH-1] block_out;
   wire             block_out_vld;

   reg              clk = 0;
   reg              reset = 0;

   integer          count = 0;

   wire             de_rdy;
   
   always #8  clk = ~clk;

`define BLOCK_0    128'h32_43_f6_a8_88_5a_30_8d_31_31_98_a2_e0_37_07_34
`define CIPHER_0   128'h39_25_84_1d_02_dc_09_fb_dc_11_85_97_19_6a_0b_32
`define KEY_0      128'h2b_7e_15_16_28_ae_d2_a6_ab_f7_15_88_09_cf_4f_3c
`define LAST_KEY_0 128'hd014f9a8_c9ee2589_e13f0cc8_b6630ca6
   
`define BLOCK_1    128'h00112233445566778899aabbccddeeff
`define CIPHER_1   128'h69c4e0d86a7b0430d8cdb78070b4c55a
`define KEY_1      128'h000102030405060708090a0b0c0d0e0f
`define LAST_KEY_1 128'h13111d7fe3944a17f307a78b4d2b30c5
   
   reg              t = 1;
   
   always @(posedge clk) begin
      count <= count < 3 ? count + 1'b1 : count;
      if(de_rdy) begin
         t <= !t;
      end
   end
   always @(*) begin
      case (count)
         1: begin
            reset = 1;
         end
         2: begin
            reset = 0;
         end
         3: begin
            block_in        = t ? `CIPHER_0 : `CIPHER_1;
            key             = t ? `KEY_0 : `KEY_1;
            if(de_rdy) begin
               block_in_vld    = 1;
            end
            else begin
               block_in_vld    = 0;
            end
         end
               
      endcase // case(count)
   end // always @ (posedge clk)

   always @(posedge clk) begin
      if(block_out_vld) begin
         if(block_out !== `BLOCK_0 && block_out !== `BLOCK_1)
           $display("AES Decryption failed!\n");
         else
           $display("AES Decryption Succeeded!\n");
      end
   end

   aes_decrypt_single aes_decrypt_single
     (// Inputs
      .key                              (key),
      .block_in                         (block_in),
      .block_in_vld                     (block_in_vld),

      // Outputs
      .block_out                        (block_out),
      .block_out_vld                    (block_out_vld),
      .rdy                              (de_rdy),

      .clk                              (clk),
      .reset                            (reset));
endmodule
// synthesis translate_on


