/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: aes_single.v
 * Project: ICING
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: brings together all parts of the aes-encoding/decoding algorithm
 * 
 * The INVERSE parameter is used to decide whether to do the encryption 
 * or decryption. For decryption, the input key is the key from the last
 * round of key expansion.
 * 
 * The block will also output the round key used for the last round.
 * 
 * This block runs at 1.6 Gbps
 *              
 ******************************************************************************/
`timescale 1ns/1ps

module aes_single
  #(parameter WIDTH = 128,
    parameter INVERSE = 0,
    parameter ONLY_KEY_EXPANSION = 0)
    (input [0:WIDTH-1]      block_in,
     input                  block_in_vld,

     output     [0:WIDTH-1] block_out,
     output                 block_out_vld,
     output                 block_in_ack,
     output                 rdy,

     input [0:WIDTH-1]      key,
     output     [0:WIDTH-1] last_round_key,

     input                  clk,
     input                  reset);

`include "aes_funcs.v.inc"

   /* number of physical rounds */
   localparam NUM_ROUNDS = 1;

   /* holds the state between each round */
   wire [0:WIDTH-1]     states[0:NUM_ROUNDS];

   /* holds the round keys between each round */
   wire [0:WIDTH-1]     round_keys[0:NUM_ROUNDS];

   /* keeps track of valid data going through the rounds */
   reg [9:0]            vld_reg;

   /* holds the round number for the round */
   reg [3:0]           round_number[0:NUM_ROUNDS-1];

   always @(posedge clk) begin
      /* shift the position of the vld data on every clock cycle */
      vld_reg     <= reset ? 0 : {vld_reg[8:0], block_in_ack};
   end
   
   always @(*) begin
      case(vld_reg)
         1: round_number[0] 	    = INVERSE ? 8 : 1;
         2: round_number[0] 	    = INVERSE ? 7 : 2;
         4: round_number[0] 	    = INVERSE ? 6 : 3;
         8: round_number[0] 	    = INVERSE ? 5 : 4;
         16: round_number[0] 	    = INVERSE ? 4 : 5;
         32: round_number[0] 	    = INVERSE ? 3 : 6;
         64: round_number[0] 	    = INVERSE ? 2 : 7;
         128: round_number[0] 	    = INVERSE ? 1 : 8;
         256: round_number[0] 	    = INVERSE ? 0 : 9;
         default: round_number[0]   = INVERSE ? 9 : 0;
      endcase // case(vld_reg)
   end // always @ (*)

   assign rdy             = !reset && ((!(|vld_reg[8:0])) || block_out_vld);
   assign block_in_ack    = rdy & block_in_vld;

   /* either get data from inputs or loop back the inputs from the
    * last round to get through the next set of rounds */
   assign states[0]       = !block_in_ack ? 
			    states[NUM_ROUNDS] : 
			    ONLY_KEY_EXPANSION ? 
			    block_in : aes_addroundkey(block_in, key);
   assign round_keys[0]   = !block_in_ack ? round_keys[NUM_ROUNDS] : key;
   
   /* put the rounds together */
   generate
      genvar i;
      for(i=0; i<NUM_ROUNDS; i=i+1) begin:aes_rounds
         /* instantiate a round */
         aes_round #(.INVERSE (INVERSE), .ONLY_KEY_EXPANSION(ONLY_KEY_EXPANSION))
           aes_round
           (.s_in           (states[i]),
            .s_out          (states[i+1]),
            .prev_round_key (round_keys[i]),
            .round_key      (round_keys[i+1]),
            .round_number   (round_number[i]),
            .clk            (clk)
            );
      end
   endgenerate

   assign block_out_vld     = vld_reg[9];
   assign block_out         = states[NUM_ROUNDS];
   assign last_round_key    = round_keys[NUM_ROUNDS];

endmodule // aes

module aes_single_test();

   localparam WIDTH = 128;
   
   reg [0:WIDTH-1]  block_in = 0;
   wire             block_in_ack;
   reg [0:WIDTH-1]  key = 0;
   wire             en_ack;
   reg              block_in_vld = 0;

   wire [0:WIDTH-1] block_out;
   wire             block_out_vld;
   wire [0:WIDTH-1] last_round_key;

   wire [0:WIDTH-1] de_block_out;
   wire             de_block_out_vld;
   wire [0:WIDTH-1] de_last_round_key;

   reg              clk = 0;
   reg              reset = 0;

   integer          count = 0;

   wire             en_rdy;
   
   always #8  clk = ~clk;

`define BLOCK_0    128'h32_43_f6_a8_88_5a_30_8d_31_31_98_a2_e0_37_07_34
`define CIPHER_0   128'h39_25_84_1d_02_dc_09_fb_dc_11_85_97_19_6a_0b_32
`define KEY_0      128'h2b_7e_15_16_28_ae_d2_a6_ab_f7_15_88_09_cf_4f_3c

`define BLOCK_1    128'h00112233445566778899aabbccddeeff
`define CIPHER_1   128'h69c4e0d86a7b0430d8cdb78070b4c55a
`define KEY_1      128'h000102030405060708090a0b0c0d0e0f

   reg              t = 1;
   
   always @(posedge clk) begin
      count <= count < 3 ? count + 1'b1 : count;
      if(en_ack) begin
         t <= !t;
      end
   end
   always @(*) begin
      case (count)
         1: begin
            reset = 1;
         end
         2: begin
            reset = 0;
         end
         3: begin
            block_in        = t ? `BLOCK_0 : `BLOCK_1;
            key             = t ? `KEY_0 : `KEY_1;
            block_in_vld    = 1;
         end
      endcase // case(count)
   end // always @ (posedge clk)

   always @(posedge clk) begin
      if(block_out_vld) begin
         if(block_out !== `CIPHER_0 && block_out !== `CIPHER_1)
           $display("AES Encryption failed!\n");
         else
           $display("AES Encryption Succeeded!\n");
      end
      if(de_block_out_vld) begin
         if(de_block_out !== `BLOCK_0 && de_block_out !== `BLOCK_1)
           $display("AES Decryption failed!\n");
         else
           $display("AES Decryption Succeeded!\n");
         if(de_last_round_key != `KEY_0 && de_last_round_key != `KEY_1)
           $display("AES Inverse key expansion failed!\n");
         else
           $display("AES Inverse key expansion Succeeded!\n");
      end
   end

   aes_single #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (key),
      .block_in                         (block_in),
      .block_in_vld                     (block_in_vld),

      // Outputs
      .block_out                        (block_out),
      .block_out_vld                    (block_out_vld),
      .last_round_key                   (last_round_key),
      .block_in_ack                     (en_ack),
      .rdy                              (en_rdy),

      .clk                              (clk),
      .reset                            (reset));

   aes_single #(.INVERSE(1)) aes_decipher
     (
      // Inputs
      .key                              (last_round_key),
      .block_in                         (block_out),
      .block_in_vld                     (block_out_vld),

      // Outputs
      .block_out                        (de_block_out),
      .block_out_vld                    (de_block_out_vld),
      .last_round_key                   (de_last_round_key),
      .block_in_ack                     (de_block_in_ack),
      .rdy                              (),

      .clk                              (clk),
      .reset                            (reset));
endmodule // aes_enblock_out_test

module aes_single_top
  (input [0:127] en_key,
   input [0:127] en_block_in,
   input         en_block_in_vld,
   output        en_block_in_ack,
   
   input [0:127] de_key,
   input [0:127] de_block_in,
   input         de_block_in_vld,
   output        de_block_in_ack,
   
   output [0:127] en_last_round_key,
   output [0:127] en_block_out,
   output         en_block_out_vld,
   
   output [0:127] de_last_round_key,
   output [0:127] de_block_out,
   output         de_block_out_vld,

   input clk,
   input reset);
   
   aes_single #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (en_key),
      .block_in                         (en_block_in),
      .block_in_vld                     (en_block_in_vld),

      // Outputs
      .block_out                        (en_block_out),
      .block_out_vld                    (en_block_out_vld),
      .last_round_key                   (en_last_round_key),
      .block_in_ack                     (block_in_ack),
      .rdy                              (),

      .clk                              (clk),
      .reset                            (reset));

endmodule // aes_top
