/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: aes_unrolled.v
 * Project: ICING
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: brings together all parts of the aes-encoding/decoding algorithm
 * 
 * The INVERSE parameter is used to decide whether to do the encryption 
 * or decryption. For decryption, the input key is the key from the last
 * round of key expansion.
 * 
 * The block will also output the round key used for the last round.
 *              
 * This implementation uses 10 physical rounds.
 ******************************************************************************/
`timescale 1ns/1ps

module aes_unrolled
  #(parameter WIDTH = 128,
    parameter INVERSE = 0)
    (input [0:WIDTH-1]      block_in,
     input                  block_in_vld,

     output reg [0:WIDTH-1] block_out,
     output reg             block_out_vld,
     output                 block_in_ack,
     output                 rdy,

     input [0:WIDTH-1]      key,
     output reg [0:WIDTH-1] last_round_key,

     input                  clk,
     input                  reset);

`include "aes_funcs.v.inc"

   /* number of physical rounds */
   localparam NUM_ROUNDS = 10;

   /* holds the state between each round */
   wire [0:WIDTH-1]     states[0:NUM_ROUNDS];
   reg [0:WIDTH-1]      states_0;

   /* holds the round keys between each round */
   wire [0:WIDTH-1]     round_keys[0:NUM_ROUNDS];
   reg  [0:WIDTH-1]     round_keys_0;

   /* keeps track of valid data going through the rounds */
   reg [9:0]           vld_reg;

   always @(posedge clk) begin
      /* shift the position of the vld data on every clock cycle */
      vld_reg     <= reset ? 0 : {vld_reg[8:0], block_in_ack};
   end

   assign states[0]       = aes_addroundkey(block_in, key);
   assign round_keys[0]   = key;
   assign block_in_ack    = block_in_vld;
   assign rdy             = !reset;

   /* put the rounds together */
   generate
      genvar i;
      for(i=0; i<NUM_ROUNDS; i=i+1) begin:aes_rounds
	 wire [3:0] round_number = INVERSE ? NUM_ROUNDS - 1 - i : i;
         /* instantiate a round */
         aes_round #(.INVERSE (INVERSE))
           aes_round
           (.s_in           (states[i]),
            .s_out          (states[i+1]),
            .prev_round_key (round_keys[i]),
            .round_key      (round_keys[i+1]),
            .round_number   (round_number),
            .clk            (clk)
            );
      end
   endgenerate

   /* latch the outputs */
   always @(posedge clk) begin
      if(reset) begin
         block_out_vld     <= 0;
      end
      else begin
         block_out_vld     <= vld_reg[9];
         block_out         <= states[NUM_ROUNDS];
         last_round_key    <= round_keys[NUM_ROUNDS];
      end
   end // always @ (posedge clk)
   
endmodule // aes

module aes_unrolled_test();

   localparam WIDTH = 128;
   
   reg [0:WIDTH-1]  block_in = 0;
   reg [0:WIDTH-1]  key = 0;
   reg              block_in_vld = 0;

   wire [0:WIDTH-1] block_out;
   wire             block_out_vld;
   wire [0:WIDTH-1] last_round_key;

   wire [0:WIDTH-1] de_block_out;
   wire             de_block_out_vld;
   wire [0:WIDTH-1] de_last_round_key;

   reg              clk = 0;
   reg              reset = 0;

   integer          count = 0;
   
   wire 	    en_rdy;
   
   always #8  clk = ~clk;

`define BLOCK_0    128'h32_43_f6_a8_88_5a_30_8d_31_31_98_a2_e0_37_07_34
`define CIPHER_0   128'h39_25_84_1d_02_dc_09_fb_dc_11_85_97_19_6a_0b_32
`define KEY_0      128'h2b_7e_15_16_28_ae_d2_a6_ab_f7_15_88_09_cf_4f_3c

`define BLOCK_1    128'h00112233445566778899aabbccddeeff
`define CIPHER_1   128'h69c4e0d86a7b0430d8cdb78070b4c55a
`define KEY_1      128'h000102030405060708090a0b0c0d0e0f

   reg 		    t = 0;
   
   always @(posedge clk) begin
      count <= count + 1'b1;

      case (count)
         1: begin
            reset <= 1;
         end
         2: begin
            reset <= 0;
         end
         3: begin
	    count 	    <= count;
            block_in 	    <= t ? `BLOCK_0 : `BLOCK_1;
            key 	    <= t ? `KEY_0 : `KEY_1;
	    if(en_rdy) begin
               block_in_vld    <= 1;
	       t <= !t;
	    end
         end
	       
      endcase // case(count)
   end // always @ (posedge clk)

   always @(posedge clk) begin
      if(block_out_vld) begin
         if(block_out !== `CIPHER_0 && block_out !== `CIPHER_1)
           $display("AES Encryption failed!\n");
         else
           $display("AES Encryption Succeeded!\n");
      end
      if(de_block_out_vld) begin
         if(de_block_out !== `BLOCK_0 && de_block_out !== `BLOCK_1)
           $display("AES Decryption failed!\n");
         else
           $display("AES Decryption Succeeded!\n");
         if(de_last_round_key != `KEY_0 && de_last_round_key != `KEY_1)
           $display("AES Inverse key expansion failed!\n");
         else
           $display("AES Inverse key expansion Succeeded!\n");
      end
   end

   aes_unrolled #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (key),
      .block_in                         (block_in),
      .block_in_vld                     (block_in_vld),

      // Outputs
      .block_out                        (block_out),
      .block_out_vld                    (block_out_vld),
      .last_round_key                   (last_round_key),
      .block_in_ack                     (block_in_ack),
      .rdy                              (rdy),

      .clk                              (clk),
      .reset                            (reset));

   aes_unrolled #(.INVERSE(1)) aes_decipher
     (
      // Inputs
      .key                              (last_round_key),
      .block_in                         (block_out),
      .block_in_vld                     (block_out_vld),

      // Outputs
      .block_out                        (de_block_out),
      .block_out_vld                    (de_block_out_vld),
      .last_round_key                   (de_last_round_key),
      .block_in_ack                     (de_block_in_ack),
      .rdy                              (de_rdy),

      .clk                              (clk),
      .reset                            (reset));
     
   
endmodule // aes_enblock_out_test

module aes_unrolled_top
  (
/* -----\/----- EXCLUDED -----\/-----
   input [0:127] en_key,
   input [0:127] en_block_in,
   input         en_block_in_vld,
   output        en_block_in_ack,
   
   output [0:127] en_last_round_key,
   output [0:127] en_block_out,
   output         en_block_out_vld,
 -----/\----- EXCLUDED -----/\----- */
   
   input [0:127] de_key,
   input [0:127] de_block_in,
   input         de_block_in_vld,
   output        de_block_in_ack,
   
   output [0:127] de_last_round_key,
   output [0:127] de_block_out,
   output         de_block_out_vld,

   input clk,
   input reset);
   
/* -----\/----- EXCLUDED -----\/-----
   aes_unrolled #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (en_key),
      .block_in                         (en_block_in),
      .block_in_vld                     (en_block_in_vld),

      // Outputs
      .block_out                        (en_block_out),
      .block_out_vld                    (en_block_out_vld),
      .last_round_key                   (en_last_round_key),
      .block_in_ack                     (en_block_in_ack),
      .rdy                              (),

      .clk                              (clk),
      .reset                            (reset));
 -----/\----- EXCLUDED -----/\----- */

   aes_unrolled #(.INVERSE(1)) aes_decipher
     (
      // Inputs
      .key                              (de_key),
      .block_in                         (de_block_in),
      .block_in_vld                     (de_block_in_vld),

      // Outputs
      .block_out                        (de_block_out),
      .block_out_vld                    (de_block_out_vld),
      .last_round_key                   (de_last_round_key),
      .block_in_ack                     (de_block_in_ack),
      .rdy                              (),

      .clk                              (clk),
      .reset                            (reset));
endmodule // aes_top
