/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: cmac.v
 * Project: ICING
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Implements the CMAC algorithm 
 * 
 * This module is designed according to the CMAC spec from NIST
 * 
 * To use, set <key> and <length>, and pulse <start> high. Keep feeding the module
 * <block>s and make sure it doesn't underrun. <ready> will go low indicating
 * the module can accept one more block only.
 * 
 * When all blocks have been processed, module will raise <result_vld> 
 * and set <result>.
 ******************************************************************************/
`timescale 1ns/1ps

module cmac
  #(parameter KEYLEN            = 128,
    parameter BLOCKLEN          = 128,
    parameter FIFO_DEPTH_BITS   = 2,
    parameter LATCH_OUTPUT      = 0,
    parameter ASSUME_PADDED     = 0,            /* If non-zero, no padding is done */
    parameter MAX_MSG_LEN_BITS  = 11)
    (/* interface to input message */
     input      [0:KEYLEN-1]   key,             /* MAC key */
     input      [0:BLOCKLEN-1] block,           /* block message to MAC */
     input                     block_vld,       /* indicates block is valid */
     input [MAX_MSG_LEN_BITS-1:0] length,       /* total length og msg in bytes */
     input                     start,           /* raise to start module */
     output reg                ready,           /* indicates that the module 
                                                 * can accept two more blocks */
     output                    full,            /* indicates the madule cannot
                                                 * accept any more blocks */

     /* result */
     output reg [0:BLOCKLEN-1] result,          /* Final result of CMAC */
     output reg                result_vld,      /* Indicates CMAC done */

     /* interface to block cipher */
     output reg [0:BLOCKLEN-1] cipher_data,        /* block input to block cipher */
     output reg                cipher_data_vld,    
     output reg [0:KEYLEN-1]   cipher_key,
     input                     cipher_rdy,

     input      [0:BLOCKLEN-1] cipher_result,
     input                     cipher_result_vld,

     /* misc */
     input                    clk,
     input                    reset);

   //-------------------- Internal Parameters ------------------------
   localparam R_b = {{(BLOCKLEN-8){1'b0}}, 8'h87};

   localparam WAIT_FOR_START = 1,
              WAIT_FOR_L     = 2,
              START_MAC      = 4,
              DO_BLOCKS      = 8,
              WAIT_FOR_BLOCKS = 16;
   
   //------------------------ Wires/Regs -----------------------------
   wire [0:BLOCKLEN-1]        block_dout;
   reg                        rd_block_fifo;

   wire [0:BLOCKLEN-1]        padded_block;
   wire [0:BLOCKLEN-1]        last_block_mask[0:BLOCKLEN/8 - 1];
   wire [0:BLOCKLEN-1]        last_block_set[0:BLOCKLEN/8 - 1];
   
   reg [0:KEYLEN-1]           key1, key1_nxt;
   reg [0:KEYLEN-1]           key2, key2_nxt;
   reg [MAX_MSG_LEN_BITS-1:0] bytes_left, bytes_left_nxt;

   reg [4:0]                  state, state_nxt;

   reg [0:BLOCKLEN-1]         cipher_result_latched;

   reg                        more_than_1_block_left;
   reg                        exactly_1_block_left;
   reg                        no_blocks_left;

   reg [0:BLOCKLEN-1]         result_nxt;
   reg                        result_vld_nxt;
   
   //-------------------------- Modules ------------------------------
   fallthrough_small_fifo 
     #(.WIDTH(BLOCKLEN), .MAX_DEPTH_BITS(FIFO_DEPTH_BITS))
      block_fifo
        (.din           (block),
         .dout          (block_dout),
         .wr_en         (block_vld),
         .rd_en         (rd_block_fifo),
         .full          (full),
         .prog_full     (),
         .nearly_full   (block_fifo_nearly_full),
         .empty         (block_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );
   //--------------------------- Logic -------------------------------
   /* generate the last block for when the total length is
    * not a multiple of the block size */
   generate
      genvar i;
      if(!ASSUME_PADDED) begin
         for(i=0; i<BLOCKLEN/8; i=i+1) begin:pad_mux
            assign last_block_mask[i] = {BLOCKLEN{1'b1}} << (BLOCKLEN - i*8);
            assign last_block_set[i] = {1'b1,{(BLOCKLEN-1){1'b0}}} >> (i*8);
         end
         assign    padded_block = (block_dout
                                   & last_block_mask[bytes_left[3:0]] 
                                   | last_block_set[bytes_left[3:0]]);
      end // if (!ASSUME_PADDED)
      else begin
         assign padded_block = block_dout;
      end
   endgenerate
   
   always @(*) begin
      cipher_key        = key;
      cipher_data       = padded_block ^ cipher_result ^ key2;
      cipher_data_vld   = 0;
      key1_nxt          = key1;
      key2_nxt          = key2;
      bytes_left_nxt    = bytes_left;
      state_nxt         = state;
      rd_block_fifo     = 0;
      result_nxt        = cipher_result;
      result_vld_nxt    = 0;
      ready             = !block_fifo_nearly_full;
      case(state)
         /* wait for the start signal, then generate subkeys */
         WAIT_FOR_START: begin
            ready = cipher_rdy & !block_fifo_nearly_full;
            if(start) begin
               cipher_data       = 0;
               cipher_data_vld   = 1;
               bytes_left_nxt    = length;
               state_nxt         = WAIT_FOR_L;
            end
         end // case: WAIT_FOR_START

         WAIT_FOR_L: begin
            if(cipher_result_vld) begin
               state_nxt   = START_MAC;
               if(!cipher_result[0]) begin
                  key1_nxt = (cipher_result << 1);
               end
               else begin
                  key1_nxt = (cipher_result << 1) ^ R_b;
               end
               if(!key1_nxt[0]) begin
                  key2_nxt = (key1_nxt << 1);
               end
               else begin
                  key2_nxt = (key1_nxt << 1) ^ R_b;
               end
            end // if (cipher_result_vld)
         end // case: WAIT_FOR_L

         /* request encoding the first block */
         START_MAC: begin
            if(cipher_rdy && (!block_fifo_empty || no_blocks_left)) begin
               cipher_data_vld   = 1;
               rd_block_fifo     = !block_fifo_empty;
               state_nxt         = DO_BLOCKS;
               if(more_than_1_block_left) begin
                  cipher_data       = block_dout;
                  bytes_left_nxt    = bytes_left - BLOCKLEN/8;
               end
               else if(exactly_1_block_left) begin
                  cipher_data       = block_dout ^ key1;
                  bytes_left_nxt    = 0;
               end
               else begin
                  cipher_data       = padded_block ^ key2;
                  bytes_left_nxt    = 0;
               end
            end // if (!block_fifo_empty)
         end // case: START_MAC

         /* wait for the result of the previous encoding 
          * to be done and request the next one */
         DO_BLOCKS: begin
            if(cipher_result_vld) begin
               if(!block_fifo_empty && !no_blocks_left) begin
                  /* not the last block */
                  if(more_than_1_block_left) begin
                     cipher_data_vld   = 1;
                     rd_block_fifo     = 1;
                     cipher_data       = block_dout ^ cipher_result;
                     bytes_left_nxt    = bytes_left - BLOCKLEN/8;
                  end
                  /* last block and size = BLOCKLEN */
                  else if(exactly_1_block_left) begin
                     cipher_data_vld   = 1;
                     rd_block_fifo     = 1;
                     cipher_data       = block_dout ^ cipher_result ^ key1;
                     bytes_left_nxt    = 0;
                  end
                  /* last block but not BLOCKLEN size */
                  else begin
                     cipher_data_vld   = 1;
                     rd_block_fifo     = 1;
                     cipher_data       = padded_block ^ cipher_result ^ key2;
                     bytes_left_nxt    = 0;
                  end // else: !if(bytes_left == BLOCKLEN/8)

                  // synthesis translate_off
                  if(!cipher_rdy) begin
                     $display("%t %m ERROR: cipher result out, but cipher not rdy. Lost a word!", $time);
                  end
                  // synthesis translate_on
               end // if (!block_fifo_empty && bytes != 0)

               else if(no_blocks_left) begin
                  /* we are done */
                  state_nxt        = WAIT_FOR_START;
                  result_nxt       = cipher_result;
                  result_vld_nxt   = 1;
               end

               else begin
                  state_nxt  = WAIT_FOR_BLOCKS;
               end
            end // if (cipher_result_vld)
         end // case: DO_BLOCKS

         /* Wait until we have results */
         WAIT_FOR_BLOCKS: begin
            if(!block_fifo_empty) begin
               /* not the last block */
               if(more_than_1_block_left) begin
                  cipher_data_vld   = 1;
                  rd_block_fifo     = 1;
                  cipher_data       = block_dout ^ cipher_result_latched;
                  bytes_left_nxt    = bytes_left - BLOCKLEN/8;
               end
               /* last block and size = BLOCKLEN */
               else if(exactly_1_block_left) begin
                  cipher_data_vld   = 1;
                  rd_block_fifo     = 1;
                  cipher_data       = block_dout ^ cipher_result_latched ^ key1;
                  bytes_left_nxt    = 0;
               end
               /* last block but not BLOCKLEN size */
               else begin
                  cipher_data_vld   = 1;
                  rd_block_fifo     = 1;
                  cipher_data       = padded_block ^ cipher_result_latched ^ key2;
                  bytes_left_nxt    = 0;
               end // else: !if(bytes_left == BLOCKLEN/8)

               state_nxt = DO_BLOCKS;

               // synthesis translate_off
               if(!cipher_rdy) begin
                  $display("%t %m ERROR: cipher result out, but cipher not rdy. Lost a word!", $time);
               end
               // synthesis translate_on
            end // if (!block_fifo_empty)
         end // case: WAIT_FOR_BLOCKS
         
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      state                     <= reset ? WAIT_FOR_START : state_nxt;
      key1                      <= key1_nxt;
      key2                      <= key2_nxt;
      bytes_left                <= bytes_left_nxt;
      more_than_1_block_left    <= bytes_left > BLOCKLEN/8;
      exactly_1_block_left      <= bytes_left == BLOCKLEN/8;
      no_blocks_left            <= bytes_left == 0;
      if(cipher_result_vld) begin
         cipher_result_latched <= cipher_result;
      end
   end // always @ (posedge clk)

   generate
      if(LATCH_OUTPUT) begin
         always @(posedge clk) begin
            result_vld <= result_vld_nxt;
            result     <= result_nxt;
         end
      end
      else begin
         always @(*) begin
            result_vld = result_vld_nxt;
            result     = result_nxt;
         end
      end // else: !if(LATCH_OUTPUT)
   endgenerate
endmodule // cmac

// synthesis translate_off
module cmac_test();
   localparam BLOCKLEN = 128;
   
   wire [0:BLOCKLEN-1]  cipher_data;            // From cmac of cmac.v
   wire [0:BLOCKLEN-1]  cipher_result;          // From cmac of cmac.v
   wire                 cipher_data_vld;        // From cmac of cmac.v
   wire [0:127]         cipher_key;             // From cmac of cmac.v
   wire                 ready;                  // From cmac of cmac.v
   wire [0:BLOCKLEN-1]  result;                 // From cmac of cmac.v
   wire                 result_vld;             // From cmac of cmac.v

   reg [0:127]          key, block;
   reg                  block_vld = 0;
   reg [11:0]           length;
   reg                  start = 0;

   reg                  clk = 0;
   reg                  reset = 0;

   integer              count = 0;
   integer              i;
   
   always #8  clk = ~clk;

   wire [0:127] KEY  = 128'h2b7e1516_28aed2a6_abf71588_09cf4f3c;

   wire [0:127] TAG0 = 128'hbb1d6929_e9593728_7fa37d12_9b756746;

   wire [0:127] MSG1 = 128'h6bc1bee2_2e409f96_e93d7e11_7393172a;
   wire [0:127] TAG1 = 128'h070a16b4_6b4d4144_f79bdd9d_d04a287c;

   wire [0:319] MSG2 = 320'h6bc1bee2_2e409f96_e93d7e11_7393172a_ae2d8a57_1e03ac9c_9eb76fac_45af8e51_30c81c46_a35ce411;
   wire [0:127] TAG2 = 128'hdfa66747_de9ae630_30ca3261_1497c827;

   wire [0:511] MSG3 = 512'h6bc1bee2_2e409f96_e93d7e11_7393172a_ae2d8a57_1e03ac9c_9eb76fac_45af8e51_30c81c46_a35ce411_e5fbc119_1a0a52ef_f69f2445_df4f9b17_ad2b417b_e66c3710;
   wire [0:127] TAG3 = 128'h51f0bebf_7e3b9d92_fc497417_79363cfe;

   wire [0:959] MSG4 = 960'hcafe00000000000700000007000000070000000741e25784cafe000000000003000000030000000300000003c9ab0e81cafe0000000000080000000800000008000000085f770dcd0102030405060708090a0b0c0d0e0f1011121314bbe9834ccafe0000000000010000000100000001000000012982130c;
   wire [0:127] TAG4 = 128'hb075d7c80b373d361e1655d22dd5b0a5;
   wire [0:127] KEY4 = 128'hd0650000010000000100000001000000;

   wire [0:127] MSG4_blocks[0:7];
   integer      count_2 = 0;
   
   generate
      genvar    ii;
      for(ii=0; ii<7; ii=ii+1) begin:gen_msg4
         assign MSG4_blocks[ii] = MSG4[ii*128 +: 128];
      end
      assign MSG4_blocks[7] = {MSG4[7*128 +: 64], 64'h0};
   endgenerate

   always @(posedge clk) begin
      case (count)
         0: begin
            key   <= KEY;
            reset <= 1;
            count <= count + 1;
         end
         1: begin
            reset <= 1;
            count <= count + 1;
         end
         2: begin
            reset <= 0;
            count <= count + 1;
         end
         3: begin
            block <= 0;
            length <= 0;
            block_vld <= 1'b0;
            start <= 1'b1;
            count <= count + 1;
         end
         4: begin
            block_vld <= 1'b0;
            start <= 1'b0;
            if(result_vld) begin
               if(result === TAG0) begin
                  $display("SUCCESS: 0 length msg passes.");
               end
               else begin
                  $display("ERROR: 0 length msg FAILS.");
               end
               count <= count + 1;
            end
         end // case: 4
         5: begin
            block <= MSG1;
            length <= 16;
            block_vld <= 1'b1;
            start <= 1'b1;
            count <= count + 1;
         end
         6: begin
            block_vld <= 1'b0;
            start <= 1'b0;
            if(result_vld) begin
               if(result === TAG1) begin
                  $display("SUCCESS: 128 length msg passes.");
               end
               else begin
                  $display("ERROR: 128 length msg FAILS.");
               end
               count <= count + 1;
            end
         end
         7: begin
            block <= MSG2[0:127];
            length <= 40;
            block_vld <= 1'b1;
            start <= 1'b1;
            count <= count + 1;
         end
         8: begin
            start <= 1'b0;
            if(ready) begin
               block <= MSG2[128:255];
               block_vld <= 1'b1;
               count <= count + 1;
            end
            else begin
               block_vld <= 1'b0;
            end
         end // case: 8
         9: begin
            start <= 1'b0;
            if(ready) begin
               block <= {MSG2[256:319], 64'h0};
               block_vld <= 1'b1;
               count <= count + 1;
            end
            else begin
               block_vld <= 1'b0;
            end
         end // case: 9
         10:begin
            block_vld <= 1'b0;
            if(result_vld) begin
               if(result === TAG2) begin
                  $display("SUCCESS: 320 length msg passes.");
               end
               else begin
                  $display("ERROR: 320 length msg FAILS.");
               end
               count <= count + 1;
            end
         end // case: 10

         11: begin
            block <= MSG3[0:127];
            length <= 64;
            block_vld <= 1'b1;
            start <= 1'b1;
            count <= count + 1;
         end
         12: begin
            start <= 1'b0;
            if(ready) begin
               block <= MSG3[128:255];
               block_vld <= 1'b1;
               count <= count + 1;
            end
            else begin
               block_vld <= 1'b0;
            end
         end
         13: begin
            start <= 1'b0;
            if(ready) begin
               block <= MSG3[256:383];
               block_vld <= 1'b1;
               count <= count + 1;
            end
            else begin
               block_vld <= 1'b0;
            end
         end
         14: begin
            start <= 1'b0;
            if(ready) begin
               block <= MSG3[384:511];
               block_vld <= 1'b1;
               count <= count + 1;
            end
            else begin
               block_vld <= 1'b0;
            end
         end
         15:begin
            block_vld <= 1'b0;
            if(result_vld) begin
               if(result === TAG3) begin
                  $display("SUCCESS: 512 length msg passes.");
               end
               else begin
                  $display("ERROR: 512 length msg FAILS.");
               end
               count <= count + 1;
            end
         end // case: 15
         16: begin
            if(ready) begin
               start <= 1'b1;
               length <= 120;
               key   <= KEY4;
               block <= MSG4_blocks[0];
               block_vld <= 1;
               count <= count + 1;
               count_2 <= 1;
            end
         end
         17: begin
            start <= 0;
            block_vld <= 0;
            if(count_2 >= 8) begin
               count <= count + 1;
            end
            else if(ready) begin
               count_2 <= count_2 + 1;
               block <= MSG4_blocks[count_2];
               block_vld <= 1;
            end
         end // case: 17
         18: begin
            block_vld <= 0;
            if(result_vld) begin
               if(result === TAG4) begin
                  $display("SUCCESS: 960 length msg passes.");
               end
               else begin
                  $display("ERROR: 960 length msg FAILS.");
               end
               count <= count + 1;
            end
         end
      endcase // case(count)
   end // always @ (posedge clk)

   cmac cmac
     (
      // Outputs
      .ready                            (ready),
      .result                           (result[0:BLOCKLEN-1]),
      .result_vld                       (result_vld),
      .cipher_data                      (cipher_data[0:BLOCKLEN-1]),
      .cipher_data_vld                  (cipher_data_vld),
      .cipher_key                       (cipher_key),
      // Inputs
      .key                              (key),
      .block                            (block[0:BLOCKLEN-1]),
      .block_vld                        (block_vld),
      .length                           (length),
      .start                            (start),
      .cipher_rdy                       (cipher_rdy),
      .cipher_result                    (cipher_result[0:BLOCKLEN-1]),
      .cipher_result_vld                (cipher_result_vld),
      .clk                              (clk),
      .reset                            (reset));

   aes_single #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (cipher_key),
      .block_in                         (cipher_data),
      .block_in_vld                     (cipher_data_vld),

      // Outputs
      .block_out                        (cipher_result),
      .block_out_vld                    (cipher_result_vld),
      .last_round_key                   (),
      .block_in_ack                     (cipher_block_in_ack),
      .rdy                              (cipher_rdy),

      .clk                              (clk),
      .reset                            (reset));


endmodule // cmac_test
// synthesis translate_on
