/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: pmac.v
 * Project: PMAC
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Implements the parallel MAC (PMAC) algorithm as described in
 *         http://www.cs.ucdavis.edu/~rogaway/papers/draft-krovetz-ocb-00.txt
 * 
 * WARNING: This MAC module does not allow empty messages! At least 1 byte is 
 *          needed.
 * 
 * Note on licensing: This code is protected by patents. P. Rogaway (the patent 
 * owner) licenses this code if it is distributed under GPL. Please do not 
 * distribute! The licensing might change...
 * 
 * Copyright 2009 Jad Naous
 * 
 * License:
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 ******************************************************************************/

`timescale 1ns/1ps

module pmac
  (input [0:127]      msg_block,
   input [0:127]      key,
   input              msg_block_vld,
   input [0:15]       msg_block_last_byte,
   output reg         msg_block_ack,

   output     [0:127] mac,
   output reg         mac_vld,

   output reg [0:127] aes_enc_plain,
   output     [0:127] aes_enc_key,
   output reg         aes_enc_plain_vld,
   input              aes_enc_plain_ack,
   input      [0:127] aes_enc_cipher,
   input              aes_enc_cipher_vld,

   input              clk,
   input              reset);

   function automatic [0:127] times2;
      input [0:127] s;
      times2 = (s << 1) ^ (s[0] ? 128'h87 : 128'h0);
   endfunction

   function automatic [0:127] times3;
      input [0:127] s;
      times3 = times2(s) ^ s;
   endfunction

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_START  = 0;
   localparam WAIT_FOR_OFFSET = 1;
   localparam DO_BLOCKS       = 2;
     
   //------------------------ Wires/Regs -----------------------------
   wire [0:127] pad_set, pad_mask, padded_block, last_enc_block;
   reg [3:0]    count;
   reg [0:1]    state, state_nxt;
   reg [0:127]  offset, offset_nxt;
   reg [0:127]  checksum, checksum_nxt;

   //-------------------------- Logic --------------------------------
   assign aes_enc_key = key;
   generate
      genvar i;
      assign pad_set[0:7]    = 8'h00;
      assign pad_mask[0:7]   = 8'hff;
      for(i=1; i<128/8; i=i+1) begin:gen_pad
         assign pad_set[i*8 : i*8+7] = {msg_block_last_byte[i-1], 7'h0};
         assign pad_mask[i*8 : i*8+7] = {8{~|msg_block_last_byte[0:i-1]}};
      end
   endgenerate
   assign padded_block = (msg_block & pad_mask) | pad_set;

   assign mac = aes_enc_cipher;
   assign last_enc_block = (checksum ^ padded_block
                            ^ (msg_block_last_byte[15]
                               ? times3(offset)
                               : times3(times3(offset))));

   /* Count the number of aes blocks still waiting */
   always @(posedge clk) begin
      if(reset) begin
         count <= 0;
      end else begin
         case({aes_enc_plain_ack, aes_enc_cipher_vld})
            2'b10: count <= count + 1'b1;
            2'b01: count <= count - 1'b1;
         endcase // case({aes_enc_plain_ack, aes_enc_cipher_vld})
      end
   end // always @ (posedge clk)

   always @(*) begin
      state_nxt           = state;
      mac_vld             = 0;
      aes_enc_plain_vld   = 0;
      aes_enc_plain       = last_enc_block;
      offset_nxt          = offset;
      checksum_nxt        = checksum;
      msg_block_ack       = 0;
      case(state)
         WAIT_FOR_START: begin
            /* if we are waiting for the 
             * tag to come out */
            if(count != 0) begin
               if(aes_enc_cipher_vld) begin
                  mac_vld = 1'b1;
               end
            end
            /* otherwise process new message */
            else if(msg_block_vld) begin
               aes_enc_plain       = 128'h0;
               aes_enc_plain_vld   = 1'b1;
               if(aes_enc_plain_ack) begin
                  state_nxt   = WAIT_FOR_OFFSET;
               end
            end
         end // case: WAIT_FOR_START

         WAIT_FOR_OFFSET: begin
            if(aes_enc_cipher_vld) begin
               offset_nxt     = times2(times3(times3(aes_enc_cipher)));
               state_nxt      = DO_BLOCKS;
               checksum_nxt   = 0;
            end
         end

         DO_BLOCKS: begin
            /* Request new encryptions */
            if(msg_block_vld) begin
               /* this is not the last block */
               if(msg_block_last_byte == 0) begin
                  aes_enc_plain_vld = 1'b1;
                  aes_enc_plain     = msg_block ^ offset;
                  if(aes_enc_plain_ack) begin
                     offset_nxt      = times2(offset);
                     msg_block_ack   = 1'b1;
                  end
               end
               /* if we are at the last block, then wait for
                * all the intermediate blocks to be done */
               else if(count == 0) begin
                  /* encrypt the padded block xored with the checksum
                   * and the last offset. This does the following part:
                   * 
                   *      if bitlength(H_m) = BLOCKLEN then
                   *         Offset = times3(Offset)
                   *         Checksum = Checksum xor H_m
                   *      else
                   *         Offset = times3(Offset)
                   *         Offset = times3(Offset)
                   *         Tmp = H_m || 1 || 
                   *               zeros(BLOCKLEN - (bitlength(H_m) + 1))
                   *         Checksum = Checksum xor Tmp
                   *      end if
                   */
                  aes_enc_plain_vld  = 1'b1;
                  aes_enc_plain      = last_enc_block;
                  if(aes_enc_plain_ack) begin
                     msg_block_ack = 1'b1;
                     state_nxt     = WAIT_FOR_START;
                  end
               end // if (count == 0 && msg_block_last_byte != 0)
            end // if (msg_block_vld)

            /* Consume encryptions and update the checksum */
            if(aes_enc_cipher_vld) begin
               checksum_nxt = checksum ^ aes_enc_cipher;
            end
         end // case: DO_BLOCKS
            
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      offset      <= offset_nxt;
      checksum    <= checksum_nxt;
      state       <= reset ? WAIT_FOR_START : state_nxt;
   end

endmodule // pmac

// synthesis translate_off
module pmac_tester();

   wire [0:127] mac, aes_enc_plain, aes_enc_key, aes_enc_cipher;
   reg [0:127]  msg_block, key;
   reg          msg_block_vld;
   reg [0:15]   msg_block_last_byte;
   reg          clk = 0;
   reg          reset = 1;
   reg          inc;
   integer      count = 0;

   always #8 clk = ~clk;
   
   pmac pmac
     (// Outputs
      .msg_block_ack                    (msg_block_ack),
      .mac                              (mac[0:127]),
      .mac_vld                          (mac_vld),
      .aes_enc_plain                    (aes_enc_plain[0:127]),
      .aes_enc_key                      (aes_enc_key[0:127]),
      .aes_enc_plain_vld                (aes_enc_plain_vld),
      // Inputs
      .msg_block                        (msg_block[0:127]),
      .key                              (key[0:127]),
      .msg_block_vld                    (msg_block_vld),
      .msg_block_last_byte              (msg_block_last_byte[0:15]),
      .aes_enc_plain_ack                (aes_enc_plain_ack),
      .aes_enc_cipher                   (aes_enc_cipher[0:127]),
      .aes_enc_cipher_vld               (aes_enc_cipher_vld),
      .clk                              (clk),
      .reset                            (reset));

   aes_single #(.INVERSE(0)) aes_encipher
     (
      // Inputs
      .key                              (aes_enc_key),
      .block_in                         (aes_enc_plain),
      .block_in_vld                     (aes_enc_plain_vld),

      // Outputs
      .block_out                        (aes_enc_cipher),
      .block_out_vld                    (aes_enc_cipher_vld),
      .last_round_key                   (),
      .rdy                              (cipher_rdy),
      .block_in_ack                     (aes_enc_plain_ack),

      .clk                              (clk),
      .reset                            (reset));

   initial key = 128'h000102030405060708090a0b0c0d0e0f;
   initial #32 reset = 0;

   wire [0:127]     msg0[0:2];
   wire [0:127]     tags[0:4];
   reg [0:9]        lengths[0:4];

   initial begin
      lengths[0] = 3;
      lengths[1] = 16;
      lengths[2] = 20;
      lengths[3] = 32;
      lengths[4] = 34;
   end

   assign           tags[0] = 128'h805099d05862a00f475902b23c214ee0;
   assign           tags[1] = 128'hba4a70b5b06628c1e82789ac45f1771f;
   assign           tags[2] = 128'h7e8ce03ee0d08d048a16c0eedf81e2d0;
   assign           tags[3] = 128'h48631bbaedd330e9e347b2291e670969;
   assign           tags[4] = 128'h78345fc25d597b1deee7b5ea2b1cfa83;

   generate
      genvar        i;
      for(i=0; i<48; i=i+1) begin
         assign msg0[i/16][(i%16)*8 +: 8] = i;
      end
   endgenerate

   integer tag_count=0;
   integer used = 0;
   always @(*) begin
      msg_block_vld   = 1;
      msg_block = msg0[used/16];
      if(used+16 >= lengths[count]) begin
         msg_block_last_byte = 16'h8000 >> ((lengths[count]-1)%16);
      end
      else begin
         msg_block_last_byte = 16'h00;
      end
   end // always @ (*)
   
   always @(posedge clk) begin
      if(!reset) begin
         if(msg_block_ack) begin
            used <= used + 16;
            if(msg_block_last_byte != 0) begin
               count <= count + 1'b1;
               used  <= 0;
            end
         end

         if(mac_vld) begin
            tag_count <= tag_count + 1'b1;
            if(mac === tags[tag_count]) begin
               $display("%t %m SUCCESS: PMAC of msg length %u correct.", $time, lengths[tag_count]);
            end
            else begin
               $display("%t %m ERROR: PMAC of msg length %u INCORRECT.", $time, lengths[tag_count]);
            end
            if(tag_count == 4) begin
               $display("PMAC test done!");
               $stop;
            end
         end // if (mac_vld)
      end // if (!reset)
   end // always @ (posedge clk)
   
endmodule  
// synthesis translate_on
