///////////////////////////////////////////////////////////////////////////////
// $Id: delay.v 1969 2007-07-18 21:59:27Z jnaous $
//
// Module: delay.v
// Project: delay
// Description: delays any pkts coming through by a programmable delay
//
///////////////////////////////////////////////////////////////////////////////
`include "udp_defines.v"
module delay
  #(parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH = DATA_WIDTH/8,
    parameter CPCI_NF2_DATA_WIDTH = 32)
    
   (output reg [DATA_WIDTH-1:0]        out_data,
    output reg [CTRL_WIDTH-1:0]        out_ctrl,
    output reg                         out_wr,
    input                              out_rdy,

    input  [DATA_WIDTH-1:0]            in_data,
    input  [CTRL_WIDTH-1:0]            in_ctrl,
    input                              in_wr,
    output reg                         in_rdy,

    // --- Register interface
    
    input                              delay_reg_req,
    input                              delay_reg_rd_wr_L,
    input  [`DELAY_REG_ADDR_WIDTH-1:0] delay_reg_addr,
    input  [CPCI_NF2_DATA_WIDTH-1:0]   delay_reg_wr_data,

    output [CPCI_NF2_DATA_WIDTH-1:0]   delay_reg_rd_data,
    output                             delay_reg_ack,

    // --- Misc
    input                              clk,
    input                              reset);



   //----------------------- local parameter ---------------------------
   parameter IN_WAIT_FOR_PKT = 0;
   parameter IN_WAIT_EOP     = 1;

   parameter OUT_WAIT_FOR_PKT = 0;
   parameter OUT_CHECK_TIMESTAMP = 1;
   parameter OUT_WAIT_EOP = 2;
   

   //----------------------- wires/regs---------------------------------
   wire [30:0]                         timestamp;
   wire [CTRL_WIDTH-1:0]               delay_fifo_out_ctrl;
   wire [DATA_WIDTH-1:0]               delay_fifo_out_data;

   wire [30:0]                         timecount_plus_1;
   reg [30:0]                          timecount;

   reg                                 timestamp_fifo_wr_en;
   reg                                 in_state, in_state_next;
   reg                                 in_ctrl_prev_is_0;

   reg [1:0]                           out_state_next, out_state;
   reg                                 out_wr_next;
   reg                                 timestamp_fifo_rd_en;
   reg                                 delay_fifo_rd_en;

   reg                                 delay_fifo_ctrl_prev_is_0;
   
   wire                                enable_delay;
   wire [29:0]                         delay_length;

   reg                                 delay_in_wr;
   wire                                delay_in_rdy;

   reg [DATA_WIDTH-1:0]                delay_out_data;
   reg [CTRL_WIDTH-1:0]                delay_out_ctrl;
   reg                                 delay_out_wr;

   //------------------------ Modules ----------------------------------
   syncfifo_2048x32 timestamp_fifo(
        .clk(clk),
        .din(timecount+delay_length), // Bus [30 : 0] 
        .rd_en(timestamp_fifo_rd_en),
        .rst(reset),
        .wr_en(timestamp_fifo_wr_en),
        .dout(timestamp), // Bus [30 : 0] 
        .empty(timestamp_fifo_empty),
        .almost_full(timestamp_fifo_almost_full),
        .full(timestamp_fifo_full));

   syncfifo_8192x72 delay_fifo
     (  .clk(clk),
        .din({in_ctrl, in_data}),
        .rd_en(delay_fifo_rd_en),
        .rst(reset),
        .wr_en(delay_in_wr),
        .dout({delay_fifo_out_ctrl, delay_fifo_out_data}),
        .empty(delay_fifo_empty),
        .almost_full(delay_fifo_almost_full),
        .full(delay_fifo_full));
                                
     
   delay_regs #(.CPCI_NF2_DATA_WIDTH(CPCI_NF2_DATA_WIDTH))
     delay_regs
       (
        // Outputs
        .delay_reg_rd_data              (delay_reg_rd_data[CPCI_NF2_DATA_WIDTH-1:0]),
        .delay_reg_ack                  (delay_reg_ack),
        .delay_length                   (delay_length[29:0]),
        .enable_delay                   (enable_delay),
        // Inputs
        .delay_reg_req                  (delay_reg_req),
        .delay_reg_rd_wr_L              (delay_reg_rd_wr_L),
        .delay_reg_addr                 (delay_reg_addr[`DELAY_REG_ADDR_WIDTH-1:0]),
        .delay_reg_wr_data              (delay_reg_wr_data[CPCI_NF2_DATA_WIDTH-1:0]),
        .out_data                       (out_data),
        .out_ctrl                       (out_ctrl),
        .out_wr                         (out_wr),
        .clk                            (clk),
        .reset                          (reset));
    
   //----------------------- Mux logic -----------------------
   always @(*) begin
      if(!enable_delay) begin
         /* mux the output */
         out_wr = in_wr;
         out_data = in_data;
         out_ctrl = in_ctrl;
         in_rdy = out_rdy;

         /* mux the input */
         delay_in_wr = 0;
      end
      else begin
         /* output */
         out_wr = delay_out_wr;
         out_data = delay_out_data;
         out_ctrl = delay_out_ctrl;
         in_rdy = delay_in_rdy;

         /* input */
         delay_in_wr = in_wr;
      end // else: !if(enable_delay)
   end // always @ (*)
   

   //----------------------- delay logic -----------------------
   assign delay_in_rdy = !delay_fifo_almost_full & !timestamp_fifo_almost_full;

   assign timecount_plus_1 = timecount + 1'b1;
   always @(posedge clk) begin
      if(reset) begin
         timecount <= 0;
      end
      else begin
         timecount <= timecount_plus_1;
      end
   end

   /*
    * Incoming state machine: wait for a packet to arrive,
    * add timestamp at which to remove packet
    * 
    */
   assign in_eop = (in_ctrl!=0) && in_ctrl_prev_is_0;
   always @(*) begin
      timestamp_fifo_wr_en  = 0;
      in_state_next         = in_state;
      case(in_state)
         /* wait for a packet to arrive
          * write the timestamp */
         IN_WAIT_FOR_PKT: begin 
            if(delay_in_wr) begin
               timestamp_fifo_wr_en = 1;
               in_state_next        = IN_WAIT_EOP;
            end
         end // case: IN_WAIT_FOR_PKT

         /* write the packet to the delay fifo */
         IN_WAIT_EOP: begin
            if(in_eop && delay_in_wr) begin
               in_state_next = IN_WAIT_FOR_PKT;
            end
         end
      endcase // case(in_state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         in_state          <= IN_WAIT_FOR_PKT;
         in_ctrl_prev_is_0 <= 0;
      end
      else begin
         in_state          <= in_state_next;
         if(delay_in_wr) begin
            in_ctrl_prev_is_0 <= (in_ctrl==0);
         end
      end // else: !if(reset)
   end // always @ (posedge clk)
   

   /*
    * Outgoing state machine: Wait for a timestamp
    * then check if the time has been reached
    */
   assign out_eop = delay_fifo_ctrl_prev_is_0 && (delay_fifo_out_ctrl!=0);
   always @(*) begin
      out_state_next         = out_state;
      out_wr_next            = 0;
      timestamp_fifo_rd_en   = 0;
      delay_fifo_rd_en       = 0;
      case(out_state)
         OUT_WAIT_FOR_PKT: begin
            if(!timestamp_fifo_empty) begin
               timestamp_fifo_rd_en = 1;
               delay_fifo_rd_en = 1;
               out_state_next   = OUT_CHECK_TIMESTAMP;
            end
         end // case: OUT_WAIT_FOR_PKT

         OUT_CHECK_TIMESTAMP: begin
            if(timestamp <= timecount || !enable_delay) begin
               out_state_next   = OUT_WAIT_EOP;
               delay_fifo_rd_en = out_rdy;
               out_wr_next      = out_rdy;
            end
         end

         OUT_WAIT_EOP: begin
            out_wr_next = out_rdy & (!delay_fifo_empty | out_eop);
            if(out_rdy) begin

               if(out_eop) begin
                  if(!timestamp_fifo_empty) begin
                     delay_fifo_rd_en       = 1;
                     timestamp_fifo_rd_en   = 1;
                     out_state_next         = OUT_CHECK_TIMESTAMP;
                  end
                  else begin
                     out_state_next = OUT_WAIT_FOR_PKT;
                  end
               end

               else begin
                  delay_fifo_rd_en = !delay_fifo_empty;
               end

            end
         end // case: OUT_WAIT_EOP

      endcase // case(out_state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         delay_fifo_ctrl_prev_is_0    <= 0;
         out_state                    <= OUT_WAIT_FOR_PKT;
         delay_out_data               <= 0;
         delay_out_ctrl               <= 0;
         delay_out_wr                 <= 0;
      end
      else begin
         delay_out_wr      <= out_wr_next;
         delay_out_data    <= delay_fifo_out_data;
         delay_out_ctrl    <= delay_fifo_out_ctrl;
         out_state         <= out_state_next;
         if(out_wr_next) begin
            delay_fifo_ctrl_prev_is_0 <= (delay_fifo_out_ctrl==0);
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

endmodule // delay
