///////////////////////////////////////////////////////////////////////////////
// $Id: delay_regs.v 1868 2007-06-08 21:32:31Z jnaous $
//
// Module: delay_regs.v
// Project: delay
// Description: Demultiplexes, stores and serves register requests
//
///////////////////////////////////////////////////////////////////////////////
`include "udp_defines.v"
`timescale 1ns/1ps

  module delay_regs
    #( parameter CPCI_NF2_DATA_WIDTH = 32,
       parameter DATA_WIDTH = 64,
       parameter CTRL_WIDTH = DATA_WIDTH/8)
   
   ( input                                delay_reg_req,
     input                                delay_reg_rd_wr_L,
     input [`DELAY_REG_ADDR_WIDTH-1:0]    delay_reg_addr,
     input [CPCI_NF2_DATA_WIDTH-1:0]      delay_reg_wr_data,
     
     output reg [CPCI_NF2_DATA_WIDTH-1:0] delay_reg_rd_data,
     output reg                           delay_reg_ack,

     output [29:0]                        delay_length,
     output                               enable_delay,

     input [DATA_WIDTH-1:0]               out_data,
     input [CTRL_WIDTH-1:0]               out_ctrl,
     input                                out_wr,

     input                                clk,
     input                                reset
     );

   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2

   // ------------- Internal parameters --------------
   parameter NUM_REGS_USED = 4; /* don't forget to update this when adding regs */
   parameter ADDR_WIDTH = log2(NUM_REGS_USED);

   // ------------- Wires/reg ------------------

   reg [CPCI_NF2_DATA_WIDTH-1:0]        reg_file [0:NUM_REGS_USED-1];

   wire [ADDR_WIDTH-1:0]                addr;

   wire                                 addr_good;

   reg                                  out_ctrl_prev_is_0;
   reg                                  seen_eop;

   // -------------- Logic --------------------

   assign enable_delay = reg_file[`DELAY_ENABLE];
   assign delay_length = reg_file[`DELAY_LENGTH];

   assign addr = delay_reg_addr[ADDR_WIDTH-1:0];
   assign addr_good = (delay_reg_addr<NUM_REGS_USED);

   always @(posedge clk) begin
      if( reset ) begin
         delay_reg_rd_data               <= 0;
         delay_reg_ack                   <= 0;

         reg_file[`DELAY_ENABLE]         <= 0;
         reg_file[`DELAY_LENGTH]         <= 0;
         reg_file[`DELAY_1ST_WORD_HI]    <= 0;
         reg_file[`DELAY_1ST_WORD_LO]    <= 0;

         out_ctrl_prev_is_0              <= 0;
         seen_eop                        <= 1;
      end
      else begin

         if(out_wr) begin
            out_ctrl_prev_is_0 <= (out_ctrl==0);
         end

         // end-of-packet
         if(out_wr && out_ctrl_prev_is_0 && (out_ctrl != 0)) begin
            seen_eop <= 1;
         end
         else if(out_wr) begin
            seen_eop <= 0;
         end

         if(delay_reg_req && !delay_reg_rd_wr_L && addr_good) begin // write
            reg_file[addr] <= delay_reg_wr_data;
         end

         else if(delay_reg_req && delay_reg_rd_wr_L) begin // read request
            if(addr_good) begin
               delay_reg_rd_data <= reg_file[addr];
            end
            else begin
               delay_reg_rd_data <= 32'hdead_beef;
            end
         end

         // latch the first word of an outgoing packet
         if(out_wr && seen_eop) begin
            reg_file[`DELAY_1ST_WORD_HI] <= out_data[DATA_WIDTH-1:DATA_WIDTH/2];
            reg_file[`DELAY_1ST_WORD_LO] <= out_data[DATA_WIDTH/2-1:0];
         end

         // requests complete after one cycle
         delay_reg_ack <= delay_reg_req;
         
      end // else: !if( reset )
   end // always @ (posedge clk)

endmodule // delay_regs


