//////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: ddr2_test_reg.v 1887 2007-06-19 21:33:32Z grg $
//
// Module: ddr2_test_reg.v
// Project: NetFPGA
// Description: Configuration, status and control registers for the DDR2 test
// 
// Note: Based heavily upon sram_test_reg.v
//       Clock-domain crossing is performed inside this module
//
///////////////////////////////////////////////////////////////////////////////

module ddr2_test_reg
   #( 
      parameter DRAM_ADDR_WIDTH = 24, // 2 - BA  +  22 - Addr
      parameter DRAM_DATA_WIDTH = 64,
      parameter CPCI_NF2_ADDR_WIDTH = 27,
      parameter CPCI_NF2_DATA_WIDTH = 32,
      parameter NF2_REG_ADDR_BITS = 10,
      parameter TOTAL_TEST_NUM = 5
    )
    (
      // Register interface
      input reg_req,
      input reg_rd_wr_L,
      input [NF2_REG_ADDR_BITS-1:0] reg_addr,
      input [CPCI_NF2_DATA_WIDTH -1:0] reg_wr_data,
      output reg [CPCI_NF2_DATA_WIDTH -1:0] reg_rd_data,
      output reg reg_ack, 

      // Self-test interface
      input log_wr,
      output reg log_rdy,
      input [DRAM_ADDR_WIDTH - 1: 0] log_addr,
      input [DRAM_DATA_WIDTH - 1 : 0] log_exp_data,
      input [DRAM_DATA_WIDTH - 1 : 0] log_rd_data,

      output reg [DRAM_DATA_WIDTH / 2 - 1 : 0] rand_seed,
      output reg test_start,
      output reg [TOTAL_TEST_NUM - 1 : 0] test_en, 

      input curr_test_done,
      input curr_test_pass, 
      input [2:0] curr_test_idx,

      // Clock/reset
      input clk_core_125,
      input clk_ddr_200,
      input reset_ddr,
      input reset_core
   );

   localparam 
             ERR_LOG_ARRAY_ADDR_BIT_NUM = 4,
             ERR_LOG_ARRAY_DEPTH = 1 << ERR_LOG_ARRAY_ADDR_BIT_NUM;

   localparam START_LEN = 5;

   localparam ERR_CNT_reg = 10'h 200;
   localparam RAND_SEED_reg = 10'h 208;
   localparam TEST_EN_reg = 10'h 20c;
   localparam TEST_STATUS_reg = 10'h 210;

   // ====================================
   // Log storage arrays
   // ====================================

   // Address array
   reg [DRAM_ADDR_WIDTH -1:0] err_log_addr_array [0 : ERR_LOG_ARRAY_DEPTH -1];

   // Expected and read data arrays
   reg [DRAM_DATA_WIDTH -1:0] err_log_exp_data_array [0 : ERR_LOG_ARRAY_DEPTH -1];
   reg [DRAM_DATA_WIDTH -1:0] err_log_rd_data_array [0 : ERR_LOG_ARRAY_DEPTH -1];

   // Error counter (note: should saturate not wrap)
   reg [ERR_LOG_ARRAY_ADDR_BIT_NUM :0] err_cnt;

   // Current entry being written
   reg [ERR_LOG_ARRAY_ADDR_BIT_NUM -1 :0] wrt_ptr;


   reg reg_ack_nxt;
   reg [CPCI_NF2_DATA_WIDTH -1:0] reg_rd_data_nxt;
    
   reg rand_seed_sel;
   reg test_en_sel;

   reg [START_LEN - 1:0] test_start_125;
   reg [START_LEN - 1:0] test_start_125_nxt;
   reg test_start_200;
   reg reset_d1;
      
   // Index to select which of the array elements to view
   wire [ERR_LOG_ARRAY_ADDR_BIT_NUM -1:0] array_idx = reg_addr[8:5];
   
   reg [DRAM_ADDR_WIDTH - 1 : 0] err_log_addr;
   reg [DRAM_DATA_WIDTH - 1 : 0] tmp_exp_data;
   reg [DRAM_DATA_WIDTH - 1 : 0] tmp_rd_data;

      
   reg log_wr_tgl_200;
   reg [DRAM_ADDR_WIDTH -1:0] log_addr_200_d1;
   reg [DRAM_DATA_WIDTH -1:0] log_exp_data_200_d1;
   reg [DRAM_DATA_WIDTH -1:0] log_rd_data_200_d1;

   reg log_wr_tgl_125;
   reg log_wr_tgl_125_d1;
   reg log_wr_tgl_125_d2;
   reg [DRAM_ADDR_WIDTH -1:0] log_addr_125;
   reg [DRAM_DATA_WIDTH -1:0] log_exp_data_125;
   reg [DRAM_DATA_WIDTH -1:0] log_rd_data_125;

   reg log_wr_125;

   reg log_rdy_tgl_125;
   reg log_rdy_tgl_200;
   reg log_rdy_tgl_200_d1;
   reg log_rdy_tgl_200_d2;

   reg curr_test_done_tgl_200;
   reg curr_test_pass_200_d1;
   reg [2:0] curr_test_idx_200_d1;

   reg curr_test_done_tgl_125;
   reg curr_test_done_tgl_125_d1;
   reg curr_test_done_tgl_125_d2;
   reg curr_test_done_125;
   reg curr_test_pass_125;
   reg [2:0] curr_test_idx_125;

   reg reset_ddr_125;
   reg reset_ddr_125_d1;

   // ===================================================
   // Clock domain crossing for log data to be written
   // ===================================================

   // DDR -> Core
   // Reset sigal
   always @(posedge clk_core_125)
   begin
      reset_ddr_125 <= reset_ddr;
      reset_ddr_125_d1 <= reset_ddr_125;
   end

   // DDR -> Core
   // DDR Clk, log and current test signals
   always @(posedge clk_ddr_200)
   begin
      if (reset_ddr) begin
         log_wr_tgl_200 <= 1'b0;
         log_addr_200_d1 <= 'h0;
         log_exp_data_200_d1 <= 'h0;
         log_rd_data_200_d1 <= 'h0;

         curr_test_done_tgl_200 <= 1'b0;
         curr_test_pass_200_d1 <= 1'b0;
         curr_test_idx_200_d1 <= curr_test_idx;
      end
      else begin
         if (log_wr) begin
            log_wr_tgl_200 <= !log_wr_tgl_200;
            log_addr_200_d1 <= log_addr;
            log_exp_data_200_d1 <= log_exp_data;
            log_rd_data_200_d1 <= log_rd_data;
         end

         if (curr_test_done) begin
            curr_test_done_tgl_200 <= !curr_test_done_tgl_200 ;
            curr_test_pass_200_d1 <= curr_test_pass;
            curr_test_idx_200_d1 <= curr_test_idx;
         end
      end
   end

   // DDR -> Core
   // Core Clk, log signals
   always @(posedge clk_core_125)
   begin
      log_wr_tgl_125 <= log_wr_tgl_200;
      log_wr_tgl_125_d1 <= log_wr_tgl_125;
      log_wr_tgl_125_d2 <= log_wr_tgl_125_d1;

      // Bring the 200 MHZ buses across to the 125 MHz domain
      //
      // Note: The bus signals have been stable for at least
      // a 125 MHz clock period
      if (log_wr_tgl_125_d2 != log_wr_tgl_125_d1) begin
         log_addr_125 <= log_addr_200_d1;
         log_exp_data_125 <= log_exp_data_200_d1;
         log_rd_data_125 <= log_rd_data_200_d1;

         log_wr_125 <= 1'b1;
      end
      else
         log_wr_125 <= 1'b0;
   end

   // DDR -> Core
   // Core Clk, current signals
   always @(posedge clk_core_125)
   begin
      curr_test_done_tgl_125 <= curr_test_done_tgl_200;
      curr_test_done_tgl_125_d1 <= curr_test_done_tgl_125;
      curr_test_done_tgl_125_d2 <= curr_test_done_tgl_125_d1;

      // Bring the 200 MHZ buses across to the 125 MHz domain
      //
      // Note: The bus signals have been stable for at least
      // a 125 MHz clock period
      if (curr_test_done_tgl_125_d2 != curr_test_done_tgl_125_d1) begin
         curr_test_pass_125 <= curr_test_pass_200_d1;
         curr_test_idx_125 <= curr_test_idx_200_d1;

         curr_test_done_125 <= 1'b1;
      end
      else
         curr_test_done_125 <= 1'b0;
   end

   // Core -> DDR
   // Core Clk, log ready signal
   always @(posedge clk_core_125)
   begin
      if (reset_core) begin
         log_rdy_tgl_125 <= 1'b0;
      end
      else if (log_wr_tgl_125_d2 != log_wr_tgl_125_d1) begin
         log_rdy_tgl_125 <= !log_rdy_tgl_125;
      end
   end
   
   // Core -> DDR
   // DDR clk, log ready signal
   always @(posedge clk_ddr_200)
   begin
      log_rdy_tgl_200 <= log_rdy_tgl_125;
      log_rdy_tgl_200_d1 <= log_rdy_tgl_200;
      log_rdy_tgl_200_d2 <= log_rdy_tgl_200_d1;

      if (reset_ddr || log_rdy_tgl_200_d2 != log_rdy_tgl_200_d1)
         log_rdy <= 1'b1;
      else if (log_wr)
         log_rdy <= 1'b0;
   end

   // Core -> DDR
   // DDR clock, test start signal
   always @(posedge clk_ddr_200)
   begin
      test_start_200 <= |test_start_125;
      test_start <= test_start_200;
   end


   // ===================================================
   // Host register access
   // ===================================================

   // Extract the array data being read
   //
   // Note: Could be reading another address but that's okay
   always @(posedge clk_core_125) begin
      err_log_addr <= err_log_addr_array[array_idx];
      tmp_exp_data <= err_log_exp_data_array[array_idx];
      tmp_rd_data <= err_log_rd_data_array[array_idx];
   end
   
   wire [31:0] err_log_exp_data_hi_word = tmp_exp_data[63:32];
   wire [31:0] err_log_exp_data_lo_word = tmp_exp_data[31: 0];

   wire [31:0] err_log_rd_data_hi_word = tmp_rd_data[63:32];
   wire [31:0] err_log_rd_data_lo_word = tmp_rd_data[31: 0];

   reg [TOTAL_TEST_NUM -1:0] test_done, test_done_nxt;
   reg [TOTAL_TEST_NUM -1:0] test_fail, test_fail_nxt;
      
   reg [1:0] state, state_nxt;

   parameter 
             IDLE_STATE = 2'h 0,
             BRAM_RD_STATE = 2'h 1,
             DONE_STATE = 2'h 2;
      
   //synthesis attribute SIGNAL_ENCODING of state is user; 

   // Register reads
   always @(*) begin
      // Default values
      rand_seed_sel = 1'b 0;
      test_en_sel = 1'b 0;
            
      reg_rd_data_nxt = 'h0;
      reg_ack_nxt = 1'b 0;
      state_nxt = state;
      
      case (state)
         IDLE_STATE: begin
            if (reg_req) begin
               reg_ack_nxt = 1'b 1;
               state_nxt = DONE_STATE;

               casez (reg_addr[NF2_REG_ADDR_BITS-1:0])
                  ERR_CNT_reg: begin
                     reg_rd_data_nxt[4:0] = err_cnt;
                  end

                  RAND_SEED_reg: begin
                     reg_rd_data_nxt = rand_seed[31:0];
                     rand_seed_sel = 1'b 1;
                  end

                  TEST_EN_reg: begin
                     reg_rd_data_nxt[TOTAL_TEST_NUM -1 : 0] = test_en;
                     test_en_sel = 1'b 1;
                  end

                  TEST_STATUS_reg: begin
                     reg_rd_data_nxt[TOTAL_TEST_NUM -1+ 8 :  8] = test_fail;
                     reg_rd_data_nxt[TOTAL_TEST_NUM -1    :  0] = test_done;
                  end

                  10'b 0?_????_????: begin
                     reg_ack_nxt = 1'b 0;
                     state_nxt = BRAM_RD_STATE;
                  end

                  default: begin
                     reg_rd_data_nxt = 32'h deadbeaf;
                  end
               endcase // casez(reg_addr[NF2_REG_ADDR_BITS-1:0])
            end // if (reg_req)
         end
        
         BRAM_RD_STATE: begin 
            // Access the log register array
            reg_ack_nxt = 1'b 1;
            state_nxt = DONE_STATE;

            casez (reg_addr[NF2_REG_ADDR_BITS-1:0])
               10'b 0_????_00000: begin
                  reg_rd_data_nxt = err_log_addr;
               end

               10'b 0_????_00100: begin
                  reg_rd_data_nxt = err_log_exp_data_hi_word;
               end

               10'b 0_????_01000: begin
                  reg_rd_data_nxt = err_log_exp_data_lo_word;
               end

               10'b 0_????_01100: begin
                  reg_rd_data_nxt = err_log_rd_data_hi_word;
               end

               10'b 0_????_10000: begin
                  reg_rd_data_nxt = err_log_rd_data_lo_word;
               end

               default: begin
                  reg_rd_data_nxt = 32'h deadbeef;
               end
            endcase // case(reg_addr[NF2_REG_ADDR_BITS-1:0])
        end // case: BRAM_RD_STATE
       
        DONE_STATE: begin
          state_nxt = IDLE_STATE;
       end
      endcase // case(state)
   end // always @ (*)

   // Test start signal
   //
   // Start when writing to the enable register
   // or when the reset signal is asserted
   //
   // Note: If the DDR reset is held then hold this signal
   always @*
   begin
      if (reset_ddr_125_d1) begin
         test_start_125_nxt[START_LEN - 1:1] = test_start_125[START_LEN - 1:1];
         test_start_125_nxt[0] = test_start_125[0] | 
                                 test_en_sel && !reg_rd_wr_L ||
                                 (reset_d1 && (! reset_core)) ||
                                 (!test_start_125[START_LEN - 1] && 
                                    |test_start_125[START_LEN - 2:0]);
      end
      else begin
         test_start_125_nxt[START_LEN - 1:1] = test_start_125[START_LEN - 2:0];
         test_start_125_nxt[0] = test_en_sel && !reg_rd_wr_L ||
                                 (reset_d1 && (! reset_core)) ||
                                 (!test_start_125[START_LEN - 1] && 
                                    |test_start_125[START_LEN - 2:0]);
      end
   end
   
   // Update the test done/test fail registers
   always @(*) begin
      test_done_nxt = test_done;
      test_fail_nxt = test_fail;
      
      if (curr_test_done_125) begin
         test_done_nxt[curr_test_idx_125] = 1'b1;
         test_fail_nxt[curr_test_idx_125] = !curr_test_pass_125;
      end // if (curr_test_done_125)
   end // always @ (*)
      
   // Update the done/test fail/pointer registers
   always @(posedge clk_core_125) begin
      if (reset_core | test_start) begin 
         err_cnt <= 'h0;
         wrt_ptr <= 'h0;

         test_done <= 'h0;
         test_fail <= 'h0;
      end
      else begin 
         if ( log_wr_125 && (err_cnt < ERR_LOG_ARRAY_DEPTH) )
           err_cnt <= err_cnt + 'h1;

         if ( log_wr_125 && (wrt_ptr < (ERR_LOG_ARRAY_DEPTH -1) ) )
           wrt_ptr <= wrt_ptr + 'h1;

         test_done <= test_done_nxt;
         test_fail <= test_fail_nxt;
      end // else: !if(reset_core | test_start)
   end // always @ (posedge clk_core_12)
   
   // Register writes
   always @(posedge clk_core_125) begin
      if (reset_core) begin
         reset_d1 <= 1'b 1;
         
         rand_seed <= 'h1;
         test_en <= {TOTAL_TEST_NUM {1'b 1}};

         test_start_125 <= 1'b 0;
                 
         reg_ack <= 1'b 0;
         reg_rd_data <= 'h0;

         state <= IDLE_STATE;
      end
      else begin
         reset_d1 <= reset_core;
                 
         if (rand_seed_sel && (! reg_rd_wr_L))
           rand_seed[31:0] <= reg_wr_data[31:0];

         if (test_en_sel && (! reg_rd_wr_L)) begin
            test_en <= reg_wr_data[TOTAL_TEST_NUM -1:0];
         end

         if ( log_wr_125 && ( err_cnt < ERR_LOG_ARRAY_DEPTH ) ) begin
            err_log_addr_array[wrt_ptr] <= log_addr_125;
            err_log_exp_data_array[wrt_ptr] <= log_exp_data_125;
            err_log_rd_data_array[wrt_ptr] <= log_rd_data_125;
         end

         test_start_125 <= test_start_125_nxt;
         
         reg_ack <= reg_ack_nxt;
         reg_rd_data <= reg_rd_data_nxt;

         state <= state_nxt;
      end // else: !if(reset_core)
   end // always @ (posedge clk_core_125)
endmodule // ddr2_test_reg
