///////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: generic_cntr_reg.v 2155 2007-08-16 01:23:02Z grg $
//
// Module: generic_cntr_reg.v
// Project: NF2.1
// Description: Implements a generic counter register block that uses RAM and
// temporarily stores updates in registers before committing them to RAM.
//
// This design is efficient in that the update registers are kept small so one
// large adder can be shared amongst all registers. The routing resources 
// associated with the RAM simplifies the routing of the registers.
// Demultiplexes, stores and serves register requests
//
// To use this block you should specify a number of parameters at
// instantiation:
//   TAG -- the tag to match against (probably defined in udp_defines.v)
//   REG_ADDR_WIDTH -- width of the address block allocated to this register
//                     group. It is important that this is specified correctly
//                     as this width is used to enable tag matching
//   NUM_REGS_USED -- how many registers in this block?
//
// Other parameter which may be useful
//   INPUT_WIDTH -- width of each update input
//   MIN_UPDATE_INTERVAL -- how many clock cycles between successive update
//                          inputs
//   RESET_ON_READ -- reset registers when read
//
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps

module generic_cntr_reg
   #( 
      parameter UDP_REG_SRC_WIDTH = 2,
      parameter TAG = 0,                  // Tag to match against
      parameter REG_ADDR_WIDTH = 5,       // Width of block addresses
      parameter NUM_REGS_USED = 8,        // How many registers
      parameter INPUT_WIDTH = 1,          // Width of each update request
      parameter MIN_UPDATE_INTERVAL = 8,  // Clocks between successive inputs
      parameter REG_WIDTH = `CPCI_NF2_DATA_WIDTH, // How wide should each counter be?
      parameter RESET_ON_READ = 0
   )
   (
      input                                  reg_req_in,
      input                                  reg_ack_in,
      input                                  reg_rd_wr_L_in,
      input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
      input  [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_in,
      input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,

      output reg                             reg_req_out,
      output reg                             reg_ack_out,
      output reg                             reg_rd_wr_L_out,
      output reg [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
      output reg [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_out,
      output reg [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,

      // --- update interface
      input [NUM_REGS_USED * INPUT_WIDTH - 1:0]   updates,

      input                                clk,
      input                                reset
    );

   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2

   function integer ceildiv;
      input integer num;
      input integer divisor;
      begin
         if (num <= divisor)
            ceildiv = 1;
         else begin
            ceildiv = num / divisor;
            if (ceildiv * divisor < num)
               ceildiv = ceildiv + 1;
         end
      end
   endfunction // ceildiv


   // ------------- Internal parameters --------------
   localparam ADDR_WIDTH = log2(NUM_REGS_USED);

   localparam MIN_CYCLE_TIME = NUM_REGS_USED + 1;
   localparam CNT_WIDTH = log2(MIN_CYCLE_TIME);

   // Calculate the number of updates we can see in a single cycle through the
   // RAM.
   //
   // This should be:
   //     ceil(MIN_CYCLE_TIME / MIN_UPDATE_INTERVAL)
   localparam UPDATES_PER_CYCLE = ceildiv(MIN_CYCLE_TIME, MIN_UPDATE_INTERVAL);
   localparam LOG_UPDATES_PER_CYCLE = log2(UPDATES_PER_CYCLE);

   // Calculate how much storage to allocate for each delta
   //
   // A single update requires INPUT_WIDTH bits of storage
   // In the worst case we would add the updates and get a total of
   //     (2^INPUT_WIDTH - 1) * UPDATES_PER_CYCLE
   // This can be represented in:
   //     log2( (2^INPUT_WIDTH - 1) * UPDATES_PER_CYCLE )
   //      = INPUT_WIDTH + log2(UPDATES_PER_CYCLE)
   localparam DELTA_WIDTH = INPUT_WIDTH + LOG_UPDATES_PER_CYCLE;

   // States
   localparam RESET = 0;
   localparam NORMAL = 1;


   // ------------- Wires/reg ------------------

   reg [REG_WIDTH-1:0]                    reg_file [0:NUM_REGS_USED-1];

   reg [REG_WIDTH-1:0]                    reg_file_in;
   wire [REG_WIDTH-1:0]                   reg_file_out;

   reg                                    reg_file_wr;
   reg [ADDR_WIDTH-1:0]                   reg_file_addr;

   wire [ADDR_WIDTH-1:0]                  addr;
   wire [REG_ADDR_WIDTH-1:0]              reg_addr;
   wire [`UDP_REG_ADDR_WIDTH-REG_ADDR_WIDTH-1:0] tag_addr;

   wire                                   addr_good;
   wire                                   tag_hit;

   wire [ADDR_WIDTH-1:0]                  addr_held;
   reg                                    addr_good_held;
   reg                                    tag_hit_held;

   reg [CNT_WIDTH-1:0]                    reg_cnt;
   reg [CNT_WIDTH-1:0]                    reg_cnt_nxt;


   reg                                    state;
   reg                                    state_nxt;

   reg [DELTA_WIDTH-1:0]                  deltas[NUM_REGS_USED-1:0];
   wire [DELTA_WIDTH-1:0]                 delta;

   wire [DELTA_WIDTH-1:0]                 update[NUM_REGS_USED-1:0];

   integer i;

   reg                                    reg_req_held;
   reg                                    reg_ack_held;
   reg                                    reg_rd_wr_L_held;
   reg [`UDP_REG_ADDR_WIDTH-1:0]          reg_addr_held;
   reg [`CPCI_NF2_DATA_WIDTH-1:0]         reg_data_held;
   reg [UDP_REG_SRC_WIDTH-1:0]            reg_src_held;


   // -------------------------------------------
   // Register file RAM
   // -------------------------------------------
  
   always @(posedge clk) begin
      if (reg_file_wr)
         reg_file[reg_file_addr] <= reg_file_in;
   end

   assign reg_file_out = reg_file[reg_file_addr];


   // -------------- Logic --------------------

   assign addr = reg_addr_in[ADDR_WIDTH-1:0];
   assign reg_addr = reg_addr_in[REG_ADDR_WIDTH-1:0];
   assign tag_addr = reg_addr_in[`UDP_REG_ADDR_WIDTH - 1:REG_ADDR_WIDTH];

   assign addr_good = reg_addr < NUM_REGS_USED;
   assign tag_hit = tag_addr == TAG;

   assign addr_held = reg_addr_held[ADDR_WIDTH-1:0];

   // Generate the individual update lines from the updates vector
   //
   // Note: I have the ugly bit selection because ModelSim doesn't seem to
   // like parameters used in :+ selects! :-(
   generate
      genvar j;
      for (j = 0; j < NUM_REGS_USED; j = j + 1) begin : update_gen
         assign update[j] = {{(DELTA_WIDTH - INPUT_WIDTH){1'b0}}, updates[j * INPUT_WIDTH + INPUT_WIDTH - 1 : j * INPUT_WIDTH]};
      end
   endgenerate

   // Track the deltas
   //
   // These are committed later to the register file
   always @(posedge clk) begin
      for (i = 0; i < NUM_REGS_USED; i = i + 1) begin
         if (reset)
            deltas[i] <= 'h0;
         else if (!(tag_hit_held && reg_req_held) && reg_cnt == i)
            // "Reset" the delta when the register counter is pointing to this
            // particular register (since the delta will be committed to RAM
            // in this cycle), except if a read is taking place since the
            // update to RAM will be delayed a cycle.
            //
            // Note: Any new update isn't included in the value recorded to
            // RAM so we record it here.
            deltas[i] <= update[i];
         else
            // Some other register is being updated so we should
            // add any new updates to the existing delta
            deltas[i] <= deltas[i] + update[i];
      end
   end
   
   // State machine that handles register file updates
   assign delta = deltas[reg_cnt];

   always @*
   begin
      // Set the defaults
      state_nxt = state;
      reg_file_in = reg_file_out;
      reg_cnt_nxt = reg_cnt;
      reg_file_addr = 'h0;
      reg_file_wr = 1'b0;

      // Default to the address specified by reg_cnt.
      //
      // This greatly reduces the maximum path delay
      reg_file_addr = reg_cnt;

      if (reset) begin
         state_nxt = RESET;
         reg_cnt_nxt = 'h0;
         reg_file_in = 'h0;
         reg_file_addr = 'h0;
      end
      else begin
         case (state)
            RESET : begin
               if (reg_cnt == NUM_REGS_USED - 1) begin
                  state_nxt = NORMAL;
                  reg_cnt_nxt = 'h0;
               end
               else
                  reg_cnt_nxt = reg_cnt + 'h1;

               reg_file_in = 'h0;
               reg_file_wr = 1'b1;
               reg_file_addr = reg_cnt;
            end

            NORMAL : begin
               // Handle register accesses
               //
               // This interrupts the normal circular updates of the registers
               if(tag_hit_held && reg_req_held) begin // read request
                  reg_file_addr = addr_held;
                  
                  // Reset the register on read or write a value depending
                  // upon the mode
                  if (RESET_ON_READ) begin
                     reg_file_wr = addr_good;
                     reg_file_in = 0;
                  end
                  else begin
                     reg_file_wr = addr_good_held && !reg_rd_wr_L_held;
                     reg_file_in = reg_data_in;
                  end
               end
               else begin
                  // Process the regular circular update of the registers
                  reg_file_wr = 1'b1;
                  reg_file_addr = reg_cnt;

                  if (reg_cnt == NUM_REGS_USED - 1)
                     reg_cnt_nxt = 'h0;
                  else
                     reg_cnt_nxt = reg_cnt + 'h1;

                  reg_file_in = reg_file_out + {{(REG_WIDTH - DELTA_WIDTH){1'b0}}, delta};
               end // if () else
            end // NORMAL

         endcase // case (state)
      end
   end

   // State machine that handles register access from the CPU
   always @(posedge clk) begin
      state <= state_nxt;
      reg_cnt <= reg_cnt_nxt;

      addr_good_held <= addr_good;
      tag_hit_held <= tag_hit;

      reg_req_held      <= reg_req_in;
      reg_ack_held      <= reg_ack_in;
      reg_rd_wr_L_held  <= reg_rd_wr_L_in;
      reg_addr_held     <= reg_addr_in;
      reg_data_held     <= reg_data_in;
      reg_src_held      <= reg_src_in;

      if(reset) begin
         reg_req_out                                    <= 0;
         reg_ack_out                                    <= 0;
         reg_rd_wr_L_out                                <= 0;
         reg_addr_out                                   <= 0;
         reg_data_out                                   <= 0;
         reg_src_out                                    <= 0;
      end // if (reset)
      else begin
         if (reg_req_held && tag_hit_held) begin
            reg_req_out       <= 1'b 1;
            reg_ack_out       <= 1'b 1;
            reg_rd_wr_L_out   <= reg_rd_wr_L_held;
            reg_addr_out      <= reg_addr_held;
            reg_data_out      <= addr_good_held ? reg_file_out : 32'hDEAD_BEEF;
            reg_src_out       <= reg_src_held;
         end
         else begin
            reg_req_out       <= reg_req_held;
            reg_ack_out       <= reg_ack_held;
            reg_rd_wr_L_out   <= reg_rd_wr_L_held;
            reg_addr_out      <= reg_addr_held;
            reg_data_out      <= reg_data_held;
            reg_src_out       <= reg_src_held;
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

endmodule
