/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id: opl_processor.v 3584 2008-04-10 19:54:57Z jnaous $
 *
 * Module: opl_processor.v
 * Project: decache
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: 
 * 
 ******************************************************************************/
`timescale 1ns/1ps

module opl_processor
  #(parameter NUM_OUTPUT_QUEUES = 8,
    parameter PKT_SIZE_WIDTH = 12,
    parameter NUM_RULE_BYTES = 72,
    parameter ENTRY_DATA_WIDTH = 72*2*8,
    parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH = DATA_WIDTH/8,
    parameter UDP_REG_SRC_WIDTH = 2)
  (// --- interface to lookups
   input                                   wildcard_hit,
   input                                   wildcard_miss,
   input[ENTRY_DATA_WIDTH-1:0]             wildcard_data,
   input                                   wildcard_data_vld,
   output reg                              wildcard_wins,
   output reg                              wildcard_loses,
   
   input [7:0]                             pkt_src_port,
   input                                   pkt_src_port_vld,
   
   // --- interface to input fifo
   input [CTRL_WIDTH-1:0]                  in_fifo_ctrl,
   input [DATA_WIDTH-1:0]                  in_fifo_data,
   output reg                              in_fifo_rd_en,
   input                                   in_fifo_empty,
   
   // --- interface to output
   output reg [DATA_WIDTH-1:0]             out_data,
   output reg [CTRL_WIDTH-1:0]             out_ctrl,
   output reg                              out_wr,
   input                                   out_rdy,
   
   // --- Misc
   input                                   clk,
   input                                   reset);

   //-------------------- Internal Parameters ------------------------
   localparam NUM_STATES              = 6;
   localparam WAIT_FOR_LOOKUPS        = 1,
              WRITE_FORWARD_BITMASK   = 2,
              WRITE_PKT               = 4,
              SEND_TO_CPU             = 8,
              SEND_FROM_CPU           = 16,
              WAIT_EOP                = 32;

   //------------------------ Wires/Regs -----------------------------
   reg [NUM_STATES-1:0]        state, state_nxt;
   reg                         out_wr_nxt;
   reg [DATA_WIDTH-1:0]        out_data_nxt;
   reg [CTRL_WIDTH-1:0]        out_ctrl_nxt;

   reg                         rd_data_fifo;
   wire [0:ENTRY_DATA_WIDTH-1] data_dout;

   reg                         rd_src_port_fifo;
   wire [3:0]                  dout_pkt_src_port;

   reg                         rd_hit_fifo;

   reg [NUM_OUTPUT_QUEUES-1:0] src_port_decoded;

   reg                         in_fifo_ctrl_prev_0;

   wire [63:0]                 data_sel;
   wire [63:0]                 data_mask_sel;
   wire [63:0]                 data[0:8];
   wire [63:0]                 data_mask[0:8];

   reg [3:0]                   counter, counter_nxt;

   //-------------------------- Modules ------------------------------

   small_fifo 
     #(.WIDTH(ENTRY_DATA_WIDTH), .MAX_DEPTH_BITS(2))
      data_fifo
        (.din           (wildcard_data),
         .wr_en         (wildcard_data_vld),
         .rd_en         (rd_data_fifo),
         .dout          (data_dout),
         .prog_full     (),
         .full          (),
         .nearly_full   (),
         .empty         (data_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(4), .MAX_DEPTH_BITS(2))
      src_port_fifo
        (.din           (pkt_src_port[3:0]),
         .wr_en         (pkt_src_port_vld),
         .rd_en         (rd_src_port_fifo),
         .dout          (dout_pkt_src_port),
         .prog_full     (),
         .full          (),
         .nearly_full   (),
         .empty         (src_port_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(1), .MAX_DEPTH_BITS(2))
      hit_fifo
        (.din           (wildcard_hit),
         .wr_en         (wildcard_hit || wildcard_miss),
         .rd_en         (rd_hit_fifo),
         .dout          (hit_dout),
         .prog_full     (),
         .full          (),
         .nearly_full   (),
         .empty         (hit_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //-------------------------- Logic --------------------------------
   /* decode source port */
   always @(*) begin
      src_port_decoded = 0;
      src_port_decoded[dout_pkt_src_port] = 1'b1;
   end

   generate
      genvar i;
      for(i=0; i<9; i=i+1) begin:gen_data_words
         assign data[i] = data_dout[i*64 +: 64];
      end
      for(i=9; i<18; i=i+1) begin:gen_mask_words
         assign data_mask[i-9] = data_dout[i*64 +: 64];
      end
   endgenerate
   assign data_sel = data[counter];
   assign data_mask_sel = data_mask[counter];
          
   /* even src ports are CPU ports */
   assign pkt_is_from_cpu = dout_pkt_src_port[0];

   always @(*) begin
      state_nxt           = state;
      counter_nxt         = counter;
      out_wr_nxt          = 0;
      out_data_nxt        = (in_fifo_data & ~data_mask_sel) | (data_sel & data_mask_sel);
      out_ctrl_nxt        = in_fifo_ctrl;
      in_fifo_rd_en       = 0;

      rd_data_fifo        = 1'b0;
      rd_hit_fifo         = 1'b0;
      rd_src_port_fifo    = 1'b0;

      wildcard_wins       = 1'b0;
      wildcard_loses      = 1'b0;

      case (state)
         WAIT_FOR_LOOKUPS: begin
            if(!hit_fifo_empty) begin
               /* If the pkt doesn't match in the table */
               if(!hit_dout) begin
                  /* and the pkt is not from the CPU */
                  if(!pkt_is_from_cpu) begin
                     /* send the pkt up to CPU for consideration */
                     state_nxt        = SEND_TO_CPU;
                     wildcard_loses   = 1;
                  end
                  /* otherwise the CPU sent it */
                  else begin
                     state_nxt = SEND_FROM_CPU;
                  end
                  rd_hit_fifo = 1;
               end
               else begin
                  /* if the pkt is not from the CPU then execute normally */
                  if(!pkt_is_from_cpu) begin
                     if(!data_fifo_empty) begin
                        state_nxt       = WRITE_FORWARD_BITMASK;
                        wildcard_wins   = 1;
                        rd_data_fifo    = 1'b1;
                        rd_hit_fifo     = 1;
                     end
                  end
                  /* otherwise, send directly */
                  else begin
                     state_nxt       = SEND_FROM_CPU;
                     rd_hit_fifo     = 1;
                  end
               end // if (wildcard_hit)
               counter_nxt = 0;
            end // if (!hit_fifo_empty)
         end // case: WAIT_FOR_LOOKUPS

         /* wait for the IOQ header to rewrite */
         SEND_TO_CPU: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               in_fifo_rd_en   = 1'b1;
               out_data_nxt    = in_fifo_data;
               if(in_fifo_ctrl == `IO_QUEUE_STAGE_NUM) begin
                  out_data_nxt[`IOQ_DST_PORT_POS + 7 : `IOQ_DST_PORT_POS] 
                                     = (src_port_decoded<<1);
                  rd_src_port_fifo   = 1'b1;
                  state_nxt          = WAIT_EOP;
               end
            end // if (out_rdy && !in_fifo_empty)
         end // case: SEND_TO_CPU

         SEND_FROM_CPU: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               in_fifo_rd_en   = 1'b1;
               out_data_nxt    = in_fifo_data;
               if(in_fifo_ctrl == `IO_QUEUE_STAGE_NUM) begin
                  out_data_nxt[`IOQ_DST_PORT_POS + 7 : `IOQ_DST_PORT_POS] 
                                     = (src_port_decoded>>1);
                  rd_src_port_fifo   = 1'b1;
                  state_nxt          = WAIT_EOP;
               end
            end // if (out_rdy && !in_fifo_empty)
         end // case: SEND_FROM_CPU

         WRITE_FORWARD_BITMASK: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               in_fifo_rd_en   = 1'b1;
               if(in_fifo_ctrl == `IO_QUEUE_STAGE_NUM) begin
                  rd_src_port_fifo   = 1'b1;
                  counter_nxt        = counter + 1'b1;
               end
               else if(in_fifo_ctrl == 0) begin
                  counter_nxt    = counter + 1'b1;
                  state_nxt      = WRITE_PKT;
               end
               else begin
                  out_data_nxt = in_fifo_data;
               end
            end // if (out_rdy && !in_fifo_empty)
         end // case: WRITE_FORWARD_BITMASK

         WRITE_PKT: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               in_fifo_rd_en   = 1'b1;
               counter_nxt     = counter + 1'b1;
               if(counter == 8) begin
                  state_nxt      = (in_fifo_ctrl != 0) ? WAIT_FOR_LOOKUPS : WAIT_EOP;
               end
            end
         end // case: WRITE_PKT

         WAIT_EOP: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt    = 1'b1;
               in_fifo_rd_en = 1'b1;
               out_data_nxt  = in_fifo_data;
               if(in_fifo_ctrl_prev_0 && in_fifo_ctrl != 0) begin
                  state_nxt = WAIT_FOR_LOOKUPS;
               end
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if (reset) begin
         state                  <= WAIT_FOR_LOOKUPS;
         counter                <= 0;
         out_wr                 <= 0;
         out_data               <= 0;
         out_ctrl               <= 1;
         in_fifo_ctrl_prev_0    <= 0;
      end
      else begin
         state              <= state_nxt;
         counter            <= counter_nxt;
         out_wr             <= out_wr_nxt;
         out_data           <= out_data_nxt;
         out_ctrl           <= out_ctrl_nxt;
         if(in_fifo_rd_en)  in_fifo_ctrl_prev_0 <= in_fifo_ctrl == 0;
      end // else: !if(reset)
   end // always @ (posedge clk)
   
endmodule // opl_processor

         
