/******************************************************************************
 * $Id$
 *
 * Module: rule_selector.v
 * Project: decache
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: concatenates parts of packet as rule words
 * 
 * Creates a rule from the first 64 bytes of the pkt and 
 * an additional ctrl word.
 *****************************************************************************/

`timescale 1 ns/1 ns

module rule_selector
  #(parameter NUM_RULE_BYTES = 64,               // rule length
    parameter TOTAL_LOOKUP_BYTES = 64,           // number of bytes to choose from
    parameter SELECT_BYTES_MASK = 64'h0,         // not used
    parameter INCLUDE_SRC_PORT = 1'b0,           // not used
    parameter INCLUDE_ADDITIONAL_WORD = 1'b0,    // not used
    parameter ADDITIONAL_WORD_BYTES_MASK = 8'h0, // not used
    parameter ADDITIONAL_WORD_CTRL = 8'hff,      // identifies module header to use
    parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH = DATA_WIDTH/8,
    parameter PKT_SIZE_WIDTH = 12)
    ( // --- Interface to datapath
      input  [DATA_WIDTH-1:0]                   in_data,
      input  [CTRL_WIDTH-1:0]                   in_ctrl,
      input                                     in_wr,

      // --- Interface to lookup modules
      output     [PKT_SIZE_WIDTH-1:0]           pkt_size,
      output reg                                pkt_size_vld,

      output [DATA_WIDTH-1:0]                   rule_word,
      output reg                                rule_word_vld,

      output     [7:0]                          pkt_src_port,
      output reg                                pkt_src_port_vld,
      
      // --- Misc
      input                                     reset,
      input                                     clk
      );

   `LOG2_FUNC

   //------------------ Internal Parameter ---------------------------
   localparam MODULE_HDRS          = 0;
   localparam PKT_WORDS            = 1;
   localparam WAIT_EOP             = 2;

   localparam TOTAL_LOOKUP_WORDS   = TOTAL_LOOKUP_BYTES/CTRL_WIDTH;

   //---------------------- Wires/Regs -------------------------------
   reg [1:0]                            state, state_next;
   reg [log2(TOTAL_LOOKUP_WORDS)-1:0]   counter, counter_next;

   //------------------------ Logic ----------------------------------
   assign rule_word      = in_data;
   assign pkt_size       = in_data[`IOQ_BYTE_LEN_POS + PKT_SIZE_WIDTH - 1 : `IOQ_BYTE_LEN_POS];
   assign pkt_src_port   = in_data[`IOQ_SRC_PORT_POS + 7 : `IOQ_SRC_PORT_POS];

   /*
    * Choose the correct mask for the bytes
    * at each word.
    */
   always @(*) begin
      pkt_size_vld       = 0;
      state_next         = state;
      counter_next       = counter;
      pkt_src_port_vld   = 0;
      rule_word_vld      = 0;
      case (state)
        MODULE_HDRS: begin
           if(in_wr) begin
	      if(in_ctrl == `IO_QUEUE_STAGE_NUM) begin
		 pkt_size_vld = 1;
		 pkt_src_port_vld = 1;
	      end
	      
              if(in_ctrl==ADDITIONAL_WORD_CTRL) begin
                 rule_word_vld = 1;
              end
              
              // pkt started
              else if(in_ctrl==0) begin
                 rule_word_vld   = 1;
                 state_next      = PKT_WORDS;
                 counter_next    = 1;
              end
           end // if (in_wr)
        end // case: MODULE_HDRS

        PKT_WORDS: begin
           if(in_wr) begin
              counter_next   = counter + 1'b1;
              rule_word_vld  = 1'b1;
              if(counter == TOTAL_LOOKUP_WORDS-1 || in_ctrl != 0) begin
                 state_next  = (in_ctrl != 0) ? MODULE_HDRS : WAIT_EOP;
              end
           end
        end // case: PKT_WORDS

        WAIT_EOP: begin
           if(in_wr && in_ctrl != 0) begin
              state_next = MODULE_HDRS;
           end
        end
      endcase // case(state)
   end // always @ (*)
               
   always @(posedge clk) begin
      if(reset) begin
         counter             <= 0;
         state               <= 0;
      end
      else begin
         counter             <= counter_next;
         state               <= state_next;
      end // else: !if(reset)
   end // always @ (posedge clk)

endmodule // rule_selector

