///////////////////////////////////////////////////////////////////////////////
// $Id: wildcard_match.v 3686 2008-05-05 18:32:57Z jnaous $
//
// Module: wildcard_match.v
// Project: decache
// Author: Jad Naous <jnaous@stanford.edu>
// Description: matches a rule allowing a wildcard
//   Uses a register block to maintain counters associated with the table
//
///////////////////////////////////////////////////////////////////////////////


  module wildcard_match
    #(parameter NUM_OUTPUT_QUEUES = 8,                  // obvious
      parameter PKT_SIZE_WIDTH = 12,                    // number of bits for pkt size
      parameter NUM_RULE_BYTES = 72,                    // number of bytes in the rule we are looking for
      parameter ENTRY_DATA_WIDTH = 72*2*9,              // length of the data read when rule is found in bits 
      parameter UDP_REG_SRC_WIDTH = 2,                  // identifies which module started this request
      parameter WILDCARD_TABLE_SIZE = 32,               // number of entries we can fit in the table
      parameter WILDCARD_REG_ADDR_WIDTH = 6,            // should be overridden
      parameter WILDCARD_LOOKUP_BLOCK_TAG = 3,          // should be overridden
      parameter DATA_WIDTH = 64
      )
   (// --- Interface for lookups
    input [DATA_WIDTH-1:0]                 rule_word,
    input                                  rule_word_vld,
    input [PKT_SIZE_WIDTH-1:0]             pkt_size,
    input                                  pkt_size_vld,

    // --- Interface to arbiter
    output reg                             wildcard_hit,
    output reg                             wildcard_miss,
    output [0:ENTRY_DATA_WIDTH-1]          wildcard_data,
    output reg                             wildcard_data_vld,
    
    // --- Interface to registers
    input                                  reg_req_in,
    input                                  reg_ack_in,
    input                                  reg_rd_wr_L_in,
    input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
    input  [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_in,
    input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,
    
    output                                 reg_req_out,
    output                                 reg_ack_out,
    output                                 reg_rd_wr_L_out,
    output     [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
    output     [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_out,
    output     [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,

    // --- Misc
    input [31:0]                           timer,
    input                                  reset,
    input                                  clk
   );

   `LOG2_FUNC
   `CEILDIV_FUNC

   //-------------------- Internal Parameters ------------------------
   localparam NUM_RULE_64BIT_WORDS     = NUM_RULE_BYTES/8;
   localparam TABLE_SIZE_BITS          = log2(WILDCARD_TABLE_SIZE);
   localparam COUNTER_WIDTH            = log2(NUM_RULE_64BIT_WORDS);
   localparam RAM_DATA_WIDTH           = WILDCARD_TABLE_SIZE*2;

   localparam NUM_STATES               = 10;
   localparam WAIT_FOR_RULE            = 1,
              MATCH_RULE               = 2,
              MATCH_B4_LAST_WORD       = 32,
              MATCH_B4_B4_LAST_WORD    = 64,
              MATCH_LAST_WORD          = 4,
              GET_DATA                 = 8,
              LATCH_OUTPUT             = 16,
              LATCH_MATCH_ADDR         = 128,
              WAIT_OUTPUT_1            = 256,
              WAIT_OUTPUT_2            = 512;
   
   //---------------------- Wires and regs----------------------------
   reg                                          rd_pkt_size_fifo;
   wire [PKT_SIZE_WIDTH-1:0]                    dout_pkt_size;
   reg                                          rd_rule;
/* -----\/----- EXCLUDED -----\/-----
   wire [DATA_WIDTH-1:0]                        dout_rule_word;
 -----/\----- EXCLUDED -----/\----- */
   wire [DATA_WIDTH-1:0]                        dout_rule_word_nxt;
   reg  [DATA_WIDTH-1:0]                        dout_rule_word;
/* -----\/----- EXCLUDED -----\/-----
   reg  [DATA_WIDTH-1:0]                        dout_rule_word_early1;
 -----/\----- EXCLUDED -----/\----- */
   reg [0:64*RAM_DATA_WIDTH-1]                  ram_data_in;
   wire [0:64*RAM_DATA_WIDTH-1]                 ram_data_out;
/* -----\/----- EXCLUDED -----\/-----
   wire [0:64*RAM_DATA_WIDTH-1]                 ram_data_out_nxt;
   reg [0:64*RAM_DATA_WIDTH-1]                  ram_data_out;
 -----/\----- EXCLUDED -----/\----- */
   reg [7:0]                                    ram_addr;
   reg                                          ram_wr_en;
   reg                                          ram_rd_en;

   reg [WILDCARD_TABLE_SIZE-1:0]                match_vector, match_vector_nxt;
   wire [WILDCARD_TABLE_SIZE-1:0]               match_vector_set;
   wire [WILDCARD_TABLE_SIZE-1:0]               current_match;
/* -----\/----- EXCLUDED -----\/-----
   wire [TABLE_SIZE_BITS-1:0]                   match_addr;
 -----/\----- EXCLUDED -----/\----- */
   wire [TABLE_SIZE_BITS-1:0]                   match_addr_nxt;
   reg [TABLE_SIZE_BITS-1:0]                    match_addr;

   wire [DATA_WIDTH-1:0]                        data[0:WILDCARD_TABLE_SIZE-1];
   wire [DATA_WIDTH-1:0]                        data_mask[0:WILDCARD_TABLE_SIZE-1];

   reg [COUNTER_WIDTH-1:0]                      counter, counter_nxt;
   reg [0:NUM_STATES-1]                         state, state_nxt;
   
   //------------------------- Modules -------------------------------

   /* we might receive four input packets simultaneously from ethernet. In addition,
    * we might receive a pkt from DMA. So we need at least 5 spots. */
   small_fifo 
     #(.WIDTH(PKT_SIZE_WIDTH),
       .MAX_DEPTH_BITS(3))
      pkt_size_fifo
        (.din           (pkt_size),
         .wr_en         (pkt_size_vld),
         .rd_en         (rd_pkt_size_fifo),
         .dout          (dout_pkt_size),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (pkt_size_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   small_fifo 
     #(.WIDTH(DATA_WIDTH),
       .MAX_DEPTH_BITS(4))
      rule_fifo
        (.din           (rule_word),
         .wr_en         (rule_word_vld),
         .rd_en         (rd_rule),
         .dout          (dout_rule_word_nxt),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (rule_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   mmapped_ram_256x64
     #(.WORD_WIDTH         (RAM_DATA_WIDTH),
       .UDP_REG_SRC_WIDTH  (UDP_REG_SRC_WIDTH),
       .REG_BLOCK_TAG      (WILDCARD_LOOKUP_BLOCK_TAG),
       .REG_ADDR_WIDTH     (WILDCARD_REG_ADDR_WIDTH))
     mmapped_ram
     (// --- register interface
      .reg_req_in         (reg_req_in),
      .reg_ack_in         (reg_ack_in),
      .reg_rd_wr_L_in     (reg_rd_wr_L_in),
      .reg_addr_in        (reg_addr_in),
      .reg_data_in        (reg_data_in),
      .reg_src_in         (reg_src_in),
      
      .reg_req_out        (reg_req_out),
      .reg_ack_out        (reg_ack_out),
      .reg_rd_wr_L_out    (reg_rd_wr_L_out),
      .reg_addr_out       (reg_addr_out),
      .reg_data_out       (reg_data_out),
      .reg_src_out        (reg_src_out),
      
      // --- Logic interface
      .ram_data_in        (ram_data_in),
/* -----\/----- EXCLUDED -----\/-----
      .ram_data_out       (ram_data_out_nxt),
 -----/\----- EXCLUDED -----/\----- */
      .ram_data_out       (ram_data_out),
      .ram_addr           (ram_addr),
      .ram_wr_en          (ram_wr_en),
      .ram_rd_en          (ram_rd_en),
      
      .clk                (clk),
      .reset              (reset)
      );

    priority_encoder  #(.OUTPUT_WIDTH (TABLE_SIZE_BITS),
                        .INPUT_WIDTH  (WILDCARD_TABLE_SIZE))
      priority_encoder
        (.unencoded_input (match_vector),
         .encoded_output  (match_addr_nxt));


   //-------------------------- Logic --------------------------------

   generate 
      genvar i;
      for(i=0; i<WILDCARD_TABLE_SIZE; i=i+1) begin:gen_cam
         assign data[i] = ram_data_out[i*DATA_WIDTH*2 : i*DATA_WIDTH*2 + DATA_WIDTH-1];
         assign data_mask[i] = ram_data_out[i*DATA_WIDTH*2 + DATA_WIDTH : i*DATA_WIDTH*2 + 2*DATA_WIDTH-1];
         assign current_match[i]    = ((data[i] & ~data_mask[i]) == (dout_rule_word & ~data_mask[i]));
         assign match_vector_set[i] = (current_match[i] && ((counter == 3) ? 1'b1 : match_vector[i]));
      end
   endgenerate

   assign wildcard_data = ram_data_out[0 +: ENTRY_DATA_WIDTH];
   
   always @(*) begin
      wildcard_miss        = 0;
      wildcard_hit         = 0;
      wildcard_data_vld    = 0;
      ram_addr             = {4'h0, counter};
      ram_rd_en            = 0;
      ram_wr_en            = 0;
      ram_data_in          = ram_data_out;
      rd_rule              = 0;
      counter_nxt          = counter;
      state_nxt            = state;
      match_vector_nxt     = match_vector;
      rd_pkt_size_fifo     = 0;
      
      case(state)
         WAIT_FOR_RULE: begin
            if(!rule_fifo_empty) begin
               ram_rd_en          = 1;
               rd_rule            = 1;
               counter_nxt        = 1;
               state_nxt          = MATCH_RULE;
            end
         end

         MATCH_RULE: begin
            if(!rule_fifo_empty) begin
               match_vector_nxt   = match_vector_set;
               counter_nxt        = counter + 1'b1;
               rd_rule            = 1;
               ram_rd_en          = 1;
               if(counter == NUM_RULE_64BIT_WORDS-1) begin
/* -----\/----- EXCLUDED -----\/-----
                  state_nxt      = MATCH_B4_B4_LAST_WORD;
 -----/\----- EXCLUDED -----/\----- */
                  state_nxt      = MATCH_B4_LAST_WORD;
               end
            end
         end // case: MATCH_RULE

         MATCH_B4_B4_LAST_WORD: begin
            match_vector_nxt = match_vector_set;
            state_nxt        = MATCH_B4_LAST_WORD;
            counter_nxt      = 0;
         end

         MATCH_B4_LAST_WORD: begin
            match_vector_nxt = match_vector_set;
            state_nxt        = MATCH_LAST_WORD;
            counter_nxt      = 0;
         end

         MATCH_LAST_WORD: begin
            match_vector_nxt = match_vector_set;
            state_nxt        = LATCH_MATCH_ADDR;
            counter_nxt      = 0;
         end

         LATCH_MATCH_ADDR: begin
            state_nxt = GET_DATA;
         end

         GET_DATA: begin
            ram_rd_en          = 1;
            ram_addr           = {3'h4, match_addr};
            rd_pkt_size_fifo   = 1;
            /* if an entry matches */
            if(|match_vector) begin
/* -----\/----- EXCLUDED -----\/-----
               state_nxt       = WAIT_OUTPUT_1;
 -----/\----- EXCLUDED -----/\----- */
               state_nxt       = WAIT_OUTPUT_2;
            end
            /* did not find entry */
            else begin
               wildcard_miss   = 1;
               state_nxt       = WAIT_FOR_RULE;
            end // else: !if(|match_vector)
         end // case: GET_DATA

         WAIT_OUTPUT_1: begin
            state_nxt = WAIT_OUTPUT_2;
         end

         WAIT_OUTPUT_2: begin
            state_nxt = LATCH_OUTPUT;
         end

         /* output the data, and update the counters */
         LATCH_OUTPUT: begin
            ram_addr                                 = {3'h4, match_addr};

            wildcard_hit                             = 1;
            wildcard_data_vld                        = 1;

            ram_wr_en                                = 1;
            /* update the pkt counter */
            ram_data_in[ENTRY_DATA_WIDTH    : 31 + ENTRY_DATA_WIDTH]   = ram_data_out[ENTRY_DATA_WIDTH    : 31 + ENTRY_DATA_WIDTH] + 1'b1;
            /* update the byte counter */
            ram_data_in[ENTRY_DATA_WIDTH+32 : 63 + ENTRY_DATA_WIDTH]   = ram_data_out[ENTRY_DATA_WIDTH+32 : 63 + ENTRY_DATA_WIDTH] + dout_pkt_size;
            /* update the timestamp */
            ram_data_in[ENTRY_DATA_WIDTH+64 : 95 + ENTRY_DATA_WIDTH]   = timer;

            state_nxt = WAIT_FOR_RULE;
         end // case: LATCH_OUTPUT
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      dout_rule_word           <= dout_rule_word_nxt;
/* -----\/----- EXCLUDED -----\/-----
      ram_data_out             <= ram_data_out_nxt;
      dout_rule_word_early1    <= dout_rule_word_nxt;
      dout_rule_word           <= dout_rule_word_early1;
 -----/\----- EXCLUDED -----/\----- */
      match_addr               <= match_addr_nxt;
      if(reset) begin
         state           <= WAIT_FOR_RULE;
         counter         <= 0;
         match_vector    <= 0;
      end
      else begin
         state           <= state_nxt;
         counter         <= counter_nxt;
         match_vector    <= match_vector_nxt;
      end
   end
   
endmodule // wildcard_match


