///////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: delay.v 3811 2008-05-30 18:26:35Z grg $
//
// Module: delay.v
// Project: Delay repeater
// Description: Delay packets passing through the device
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps
module delay
   #(
      parameter DATA_WIDTH = 64,
      parameter CTRL_WIDTH = DATA_WIDTH/8,
      parameter UDP_REG_SRC_WIDTH = 2,
      parameter INPUT_ARBITER_STAGE_NUM = 2,
      parameter IO_QUEUE_STAGE_NUM = `IO_QUEUE_STAGE_NUM,
      parameter NUM_OUTPUT_QUEUES = 8,
      parameter NUM_IQ_BITS = 3,
      parameter DELAY_BLOCK_TAG = `DELAY_BLOCK_TAG_0
   )
   (
      // --- data path interface
      output     [DATA_WIDTH-1:0]         out_data,
      output     [CTRL_WIDTH-1:0]         out_ctrl,
      output reg                          out_wr,
      input                               out_rdy,

      input  [DATA_WIDTH-1:0]             in_data,
      input  [CTRL_WIDTH-1:0]             in_ctrl,
      input                               in_wr,
      output                              in_rdy,

      // --- Register interface
      input                               reg_req_in,
      input                               reg_ack_in,
      input                               reg_rd_wr_L_in,
      input  [`UDP_REG_ADDR_WIDTH-1:0]    reg_addr_in,
      input  [`CPCI_NF2_DATA_WIDTH-1:0]   reg_data_in,
      input  [UDP_REG_SRC_WIDTH-1:0]      reg_src_in,

      output                              reg_req_out,
      output                              reg_ack_out,
      output                              reg_rd_wr_L_out,
      output [`UDP_REG_ADDR_WIDTH-1:0]    reg_addr_out,
      output [`CPCI_NF2_DATA_WIDTH-1:0]   reg_data_out,
      output [UDP_REG_SRC_WIDTH-1:0]      reg_src_out,

      // --- Misc
      input                               clk,
      input                               reset
   );

   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2
   
   //--------------------- Internal Parameter-------------------------
   localparam IN_MODULE_HDRS   = 0;
   localparam IN_PACKET        = 1;

   //---------------------- Wires/Regs -------------------------------
   reg [15:0]           decoded_src;
   reg                  state, state_nxt;

   wire [`CPCI_NF2_DATA_WIDTH-1:0] enable_reg;
   wire [`CPCI_NF2_DATA_WIDTH-1:0] delay_ms_reg;
   
   wire                 enable;

   // Delay in ms -- 16 bits should allow 64s delay
   wire [15:0]          delay_ms;
   reg [63:0]           delay_ns;

   reg [63:0]           curr_time;
   reg [63:0]           delayed_time;

   reg [63:0]           want_time;
   reg [63:0]           want_time_nxt;

   reg                  in_wr_modded;

   wire                 delay_good;
   reg                  delay_good_d1;

   // synthesis translate_off

   // Has the user been warned during simulation that the delay is
   // MICRO seconds and not MILLI seconds as it is when run in hardware
   reg                  user_warned;
   // synthesis translate_on

   //----------------------- Modules ---------------------------------
   generic_sw_regs #( 
      .UDP_REG_SRC_WIDTH      (UDP_REG_SRC_WIDTH),       // identifies which module started this request
      .TAG                    (DELAY_BLOCK_TAG),         // Tag to match against
      .REG_ADDR_WIDTH         (`DELAY_REG_ADDR_WIDTH),   // Width of block addresses
      .NUM_REGS_USED          (2)                        // How many sw regs
   ) delay_regs (
      .reg_req_in                            (reg_req_in),
      .reg_ack_in                            (reg_ack_in),
      .reg_rd_wr_L_in                        (reg_rd_wr_L_in),
      .reg_addr_in                           (reg_addr_in),
      .reg_data_in                           (reg_data_in),
      .reg_src_in                            (reg_src_in),

      .reg_req_out                           (reg_req_out),
      .reg_ack_out                           (reg_ack_out),
      .reg_rd_wr_L_out                       (reg_rd_wr_L_out),
      .reg_addr_out                          (reg_addr_out),
      .reg_data_out                          (reg_data_out),
      .reg_src_out                           (reg_src_out),

      // --- SW regs interface
      .software_regs                         ({delay_ms_reg, enable_reg}),

      .clk                                   (clk),
      .reset                                 (reset)
    );

   small_fifo #(.WIDTH(CTRL_WIDTH+DATA_WIDTH), .MAX_DEPTH_BITS(2), .NEARLY_FULL(3))
      input_fifo
        (.din           ({in_ctrl, in_data}),  // Data in
         .wr_en         (in_wr_modded),        // Write enable
         .rd_en         (in_fifo_rd_en),       // Read the next word 
         .dout          ({out_ctrl, out_data}),
         .full          (),
         .nearly_full   (in_fifo_nearly_full),
         .empty         (in_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //----------------------- Logic ---------------------------------

   assign enable = enable_reg[0];
   assign delay_ms = delay_ms_reg[15:0];
   assign in_rdy = !in_fifo_nearly_full && (delay_good_d1 || !enable);

   // When evaluating whether the delay is good we need to be careful of
   // wrap-around. This can happen when the 64-bit value overflows or when the
   // current time is low and the delay causes curr-delay to go negative.
   //
   // To overcome this problem we treat the two high bits differently. It is
   // assumed that the current time - delay will never affect the 2 MSBs. In
   // this case we have:
   //
   // (Note: Result means delay_good)
   //
   //       Time
   // Delayed   Desired   Result   Explanation 
   // =======   =======   ======   ===========
   //  0 0 X     0 0 Y    X >= Y
   //  0 1 X     0 0 Y    1
   //  0 0 X     0 1 Y    0
   //  0 1 X     0 1 Y    X >= Y
   //
   //  1 0 X     0 0 Y    ?        Should never see -- 1/2 phase difference
   //  1 1 X     0 0 Y    0        Delayed is "behind" by 1/2 phase
   //  1 0 X     0 1 Y    1        Delayed is "ahead" by 1/4 phase
   //  1 1 X     0 1 Y    ?        Should never see -- 1/2 phase difference
   //
   //  0 0 X     1 0 Y    ?        Should never see -- 1/2 phase difference
   //  0 1 X     1 0 Y    0        Delayed is "behind" by 1/4 phase
   //  0 0 X     1 1 Y    1        Delayed is "ahead" by 1/2 phase
   //  0 1 X     1 1 Y    ?        Should never see -- 1/2 phase difference
   //
   //  1 0 X     1 0 Y    X >= Y
   //  1 1 X     1 0 Y    1
   //  1 0 X     1 1 Y    0
   //  1 1 X     1 1 Y    X >= Y
   //
   // Summary:
   //   delay_good = (delayed[MSB] == desired[MSB] && 
   //                 ((delayed[MSB-1] && !desired[MSB-1]) ||
   //                  (delayed[MSB-1] == desired[MSB-1] && delayed[MSB-2:0] >= desired[MSB-2:0]))) ||
   //                (delayed[MSB] != desired[MSB] && desired[MSB-1])
   localparam MSB = 63;
   assign delay_good = (delayed_time[MSB] == want_time[MSB] && 
                        ((delayed_time[MSB-1] && !want_time[MSB-1]) ||
                         (delayed_time[MSB-1] == want_time[MSB-1] && delayed_time[MSB-2:0] >= want_time[MSB-2:0]))) ||
                       (delayed_time[MSB] != want_time[MSB] && want_time[MSB-1]);

// synthesis translate_off
   initial
      user_warned = 0;
// synthesis translate_on

   always @(posedge clk) begin
      delay_ns <= delay_ms * 1000000;
      // synthesis translate_off
      delay_ns <= delay_ms * 1000;
      if (enable === 1'b1 && !user_warned) begin
         $display("%t WARNING: Delay module %m enabled. In simulation, delays are interpreted as *microseconds* not *milliseconds*", $time);
         user_warned <= 1'b1;
      end
      // synthesis translate_on
   end

   /* Clock */
   always @(posedge clk)
   begin
      if (reset) begin
         curr_time <= 0;
         delayed_time <= 0;

         delay_good_d1 <= delay_good;
      end
      else begin
         curr_time <= curr_time + 64'h8;
         delayed_time <= curr_time - delay_ns;

         delay_good_d1 <= delay_good;
      end
   end

   /* modify the IOQ module header */
   always @(*) begin
      state_nxt        = state; 
      in_wr_modded     = in_wr;
      want_time_nxt    = want_time;

      case(state)
         IN_MODULE_HDRS: begin
            if (in_wr && in_ctrl == `TIMESTAMP_HDR) begin
               want_time_nxt = in_data;
               in_wr_modded = 1'b0;
            end

            if(in_wr && in_ctrl == 0) begin
               state_nxt = IN_PACKET;
            end
         end // case: IN_MODULE_HDRS

         IN_PACKET: begin
            if(in_wr && in_ctrl != 0) begin
               state_nxt = IN_MODULE_HDRS;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         state <= IN_MODULE_HDRS;
         want_time <= 'h0;
      end
      else begin
         state <= state_nxt;
         want_time <= want_time_nxt;
      end
   end

   /* handle outputs */
   assign in_fifo_rd_en = out_rdy && !in_fifo_empty;
   always @(posedge clk) begin
      out_wr <= reset ? 0 : in_fifo_rd_en;
   end

endmodule
