///////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: time_stamper.v 3811 2008-05-30 18:26:35Z grg $
//
// Module: time_stamper.v
// Project: Delay repeater
// Description: Stamp a 64-bit time stamp into each packet
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps
module time_stamper
   #(
      parameter DATA_WIDTH = 64,
      parameter CTRL_WIDTH = DATA_WIDTH/8,
      parameter UDP_REG_SRC_WIDTH = 2,
      parameter INPUT_ARBITER_STAGE_NUM = 2,
      parameter IO_QUEUE_STAGE_NUM = `IO_QUEUE_STAGE_NUM,
      parameter NUM_OUTPUT_QUEUES = 8,
      parameter NUM_IQ_BITS = 3
   )
   (
      // --- data path interface
      output reg [DATA_WIDTH-1:0]            out_data,
      output reg [CTRL_WIDTH-1:0]            out_ctrl,
      output reg                             out_wr,
      input                                  out_rdy,

      input  [DATA_WIDTH-1:0]                in_data,
      input  [CTRL_WIDTH-1:0]                in_ctrl,
      input                                  in_wr,
      output                                 in_rdy,

      // --- Register interface
      input                                  reg_req_in,
      input                                  reg_ack_in,
      input                                  reg_rd_wr_L_in,
      input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
      input  [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_in,
      input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,

      output reg                             reg_req_out,
      output reg                             reg_ack_out,
      output reg                             reg_rd_wr_L_out,
      output reg [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
      output reg [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_out,
      output reg [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,

      // --- Misc
      input                                  clk,
      input                                  reset
   );

   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2
   
   //--------------------- Internal Parameter-------------------------
   localparam IN_MODULE_HDRS   = 0;
   localparam IN_PACKET        = 1;

   //---------------------- Wires/Regs -------------------------------
   reg                   state, state_nxt;
   reg [63:0]            timestamp;
   wire [DATA_WIDTH-1:0] dout_data;
   wire [CTRL_WIDTH-1:0] dout_ctrl;
   reg                   in_fifo_rd_en;
   reg                   wr_timestamp;
   reg                   wr_timestamp_nxt;

   //----------------------- Modules ---------------------------------
   small_fifo #(.WIDTH(CTRL_WIDTH+DATA_WIDTH), .MAX_DEPTH_BITS(2), .NEARLY_FULL(3))
      input_fifo
        (.din           ({in_ctrl, in_data}),  // Data in
         .wr_en         (in_wr),             // Write enable
         .rd_en         (in_fifo_rd_en),    // Read the next word 
         .dout          ({dout_ctrl, dout_data}),
         .full          (),
         .nearly_full   (in_fifo_nearly_full),
         .empty         (in_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //----------------------- Logic ---------------------------------

   assign in_rdy = !in_fifo_nearly_full;

   /* Clock */
   always @(posedge clk)
   begin
      if (reset)
         timestamp <= 0;
      else 
         timestamp <= timestamp + 64'h8;
   end

   /* Output mux */
   always @* begin
      if (wr_timestamp) begin
         out_ctrl = `TIMESTAMP_HDR;
         out_data = timestamp;
      end
      else begin
         out_ctrl = dout_ctrl;
         out_data = dout_data;
      end
   end

   /* Insert the timestamp */
   always @(*)
   begin
      state_nxt         = state; 
      in_fifo_rd_en     = out_rdy && !in_fifo_empty && !wr_timestamp;
      wr_timestamp_nxt  = wr_timestamp;

      case(state)
         IN_MODULE_HDRS: begin
            // If we write the timestamp then attmept to read from the fifo
            // If we are writing the first header then don't read from the
            // fifo as we want to insert the timestamp header
            if (out_wr) begin
               if (wr_timestamp)
                  in_fifo_rd_en = out_rdy && !in_fifo_empty;
               else if (dout_ctrl == IO_QUEUE_STAGE_NUM)
                  in_fifo_rd_en = 1'b0;
            end

            // Attempt to write the timestamp if we see the first header
            if (out_wr) begin
               if (dout_ctrl == IO_QUEUE_STAGE_NUM && !wr_timestamp)
                  wr_timestamp_nxt = 1'b1;
               else
                  wr_timestamp_nxt = 1'b0;
            end

            if(out_wr && out_ctrl==0) begin
               state_nxt = IN_PACKET;
            end
         end // case: IN_MODULE_HDRS

         IN_PACKET: begin
            if(out_wr && out_ctrl!=0) begin
               state_nxt = IN_MODULE_HDRS;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         state <= IN_MODULE_HDRS;
         wr_timestamp <= 1'b0;
      end
      else begin
         state <= state_nxt;
         wr_timestamp <= wr_timestamp_nxt;
      end
   end

   /* handle outputs */
   always @(posedge clk) begin
      out_wr <= reset ? 0 : in_fifo_rd_en | (wr_timestamp_nxt & out_rdy);
   end

   /* registers unused */
   always @* begin
      reg_req_out        = reg_req_in;
      reg_ack_out        = reg_ack_in;
      reg_rd_wr_L_out    = reg_rd_wr_L_in;
      reg_addr_out       = reg_addr_in;
      reg_data_out       = reg_data_in;
      reg_src_out        = reg_src_in;
   end
   
endmodule
