#!/usr/bin/perl -w
# vim:set shiftwidth=3 softtabstop=3 expandtab:
# $id:$

# 
# Program to control the delay repeater
#

use strict;
use NF2::RegAccess;
use Getopt::Long;
require "reg_defines.ph";

# Addresses for controlling the delay module
my $DELAY_OFFSET = 0x100;
my $DELAY_ENABLE = 0x2001000;
my $DELAY_TIME   = 0x2001004;

# Maximum delay
my $MAX_DELAY = 0xffff;

# Parse the command line arguments
my $iface = 'nf2c0';   		# Default device
my $help = '';
my @delay = ('', '', '', '');

unless ( GetOptions ( 
		      "iface=s" => \$iface,
		      "help|?" => \$help,
                      "d0=i" => \$delay[0],
                      "d1=i" => \$delay[1],
                      "d2=i" => \$delay[2],
                      "d3=i" => \$delay[3],
		     )
	 and ($help eq '')
       ) { usage(); exit 1 }

# Verify the delays
for (my $i = 0; $i < 4; $i++) {
   if ($delay[$i] ne '') {
      if ($delay[$i] < 0 || $delay[$i] > $MAX_DELAY) {
         print STDERR "Error: Invalid delay for port $i (delay = $delay[$i]).\n";
         print STDERR "Must be between 0 and $MAX_DELAY\n";
         exit 1;
      }
   }
}

# Verify the delays
for (my $i = 0; $i < 4; $i++) {
   if ($delay[$i] ne '') {
      setDelay($i, $delay[$i]);
   }
}

# Output the delays
print "Delays:\n";
for (my $i = 0; $i < 4; $i++) {
   my $delay = getDelay($i);
   print "   Port $i: $delay ms\n";
}

exit (0);


#########################################################
# usage
#   print usage information
sub usage {
  (my $cmd = $0) =~ s/.*\///;
  print <<"HERE1";
NAME
   $cmd - Manage the repeater

SYNOPSIS
   $cmd [--iface <device>]  
        [--d<i> <delay>]

   $cmd --help  - show detailed help

HERE1

  return unless ($help);
  print <<"HERE";

DESCRIPTION

   Manage the repeater
   

OPTIONS
   --iface <device>
     Device to read registers from.

   --d<i> <delay>
     Set the delay on port <i> to <delay> ms

HERE
}


#
# setDelay -- Set the delay on a queue
#
sub setDelay {
   my $port = shift;
   my $delay = shift;

   if ($delay == 0) {
      nf_regwrite($iface, $DELAY_ENABLE + $port * $DELAY_OFFSET, 0);
      nf_regwrite($iface, $DELAY_TIME + $port * $DELAY_OFFSET, 0);
   }
   else {
      nf_regwrite($iface, $DELAY_TIME + $port * $DELAY_OFFSET, $delay);
      nf_regwrite($iface, $DELAY_ENABLE + $port * $DELAY_OFFSET, 1);
   }
}


#
# getDelay -- Get the delay on a queue
#
sub getDelay {
   my $port = shift;

   my $enabled = nf_regread($iface, $DELAY_ENABLE + $port * $DELAY_OFFSET);
   my $delay = nf_regread($iface, $DELAY_TIME + $port * $DELAY_OFFSET);

   if (!$enabled || $delay == 0) {
      return 0;
   }
   else {
      return $delay;
   }
}
