/********************************************************************
 * File: nf_icing3.c
 * Author: Jad Naous
 * Description: interface to hardware
 ********************************************************************/
#include <nf2util.h>
#include <reg_defines.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <arpa/inet.h>
#include "rijndael-alg-fst.h"
#include "crc32.h"
#include "nf_icing3.h"

#define DEBUG

static void pbuf(const void *buf, int n, char *s)
{
  int i;
  unsigned char *cp = (unsigned char *)buf;
    
  if (n < 0 || n > 64)
    n = 64;
    
  if (s)
    printf("%s: ", s);
        
  for (i = 0; i < n; i++)
    printf("%02X", (unsigned char)cp[i]);
  printf("\n");
}

static int fill_uhf(uint8_t uhf[ICING_UHF_NUM_BYTES], 
                    int deterministic) {
  FILE *rfd;
  uint32_t i, j;
  struct crc32 crc;
  uint32_t *u;

  if(deterministic) {
    crc32_init(&crc, POLY);
    u = (uint32_t *) uhf;
    for(i=0; i<ICING_UHF_NUM_BYTES; i+=4, u++) {
      j = i * 19571021;
      *u = crc32_calculate(&crc, &j, 4);
    }
    return 0;
  }

  if((rfd = fopen("/dev/urandom", "r")) == 0) {
    return errno;
  }

  if(fread(uhf, 1, ICING_UHF_NUM_BYTES, rfd) != ICING_UHF_NUM_BYTES) {
    return -1;
  }

  return 0;
}

int nf_icing3_init_dev(nf_icing_dev_t *icing_dev, 
                       const nf_icing_dev_config_t *config,
                       int deterministic, int full_uhf, int write_uhf) {
  int err;
  /* Open the NetFPGA device */
  icing_dev->nf2.device_name
    = config->dev_name;
  if(check_iface(&icing_dev->nf2) || openDescriptor(&icing_dev->nf2)) {
    printf("ERROR: Cannot find device or open descriptor\n");
    return -1;
  }

  /* set the realm id */
  memcpy(&icing_dev->realm_id, &config->realm_id, sizeof(realm_id_t));
  nf_icing3_write_realm_id(icing_dev->realm_id.id, &icing_dev->nf2);

  /* generate the UHF and write to HW */
  if((err = fill_uhf(icing_dev->uhf, deterministic))) {
    printf("ERROR: Could not fill the UHF.\n");
    return err;
  }
  if(write_uhf) {
    nf_icing3_write_uhf_table(icing_dev->uhf, full_uhf, &icing_dev->nf2);
  }

  /* Write the MAC addresses */
  nf_icing3_write_port_mac_addrs(config->macs, &icing_dev->nf2);

  return 0;
}

int nf_icing3_change_uhf(nf_icing_realm_info_t **realm_info,
                         int num_entries, int max_iters,
                         nf_icing_dev_t *icing_dev) {
  uint8_t temp_uhf[ICING_UHF_NUM_BYTES];
  uint8_t idx_used[(1<<SRAM_SKEY_NXTHOP_IDX_SIZE)];
  uint16_t sram_idx;
  nf_icing_realm_info_t **ptr;
  int collision, i, iter;

  if(num_entries > (1<<SRAM_SKEY_NXTHOP_IDX_SIZE)) {
    return -1;
  }

  for(iter = 0; iter<max_iters; iter++) {
    bzero(idx_used, (1<<SRAM_SKEY_NXTHOP_IDX_SIZE));
    collision = 0;
    /* get a new test UHF */
    fill_uhf(temp_uhf, 0);
    /* for each entry */
    for(ptr=realm_info, i=0; i<num_entries; ptr++, i++) {
      /* get and idx and check if it was already seen before */
      sram_idx = nf_icing3_uhf((*ptr)->realm_id.id, temp_uhf);
      sram_idx &= (1<<SRAM_SKEY_NXTHOP_IDX_SIZE) - 1;
      if(idx_used[sram_idx]) {
        collision = 1;
        break;
      }
      /* mark idx as seen */
      idx_used[sram_idx] = 1;
    }
    if(!collision) {
      /* delete all the old entries */
      for(ptr = realm_info, i=0; i<num_entries; ptr++, i++) {
        nf_icing3_delete_realm(*ptr, icing_dev);
      }
      
      /* install the new UHF */
      memcpy(icing_dev->uhf, temp_uhf, ICING_UHF_NUM_BYTES);
      nf_icing3_write_uhf_table(icing_dev->uhf, 1, &icing_dev->nf2);

      /* Write the new entries */
      for(ptr = realm_info, i=0; i<num_entries; ptr++, i++) {
        nf_icing3_write_realm(*ptr, icing_dev);
      }
      return 0;
    }
  }
  return -1;
}

void nf_icing3_close_dev(nf_icing_dev_t *icing_dev) {
  closeDescriptor(&icing_dev->nf2);
}

void nf_icing3_write_realm(nf_icing_realm_info_t *realm_info,
                           nf_icing_dev_t *icing_dev) {
  uint32_t idx;

  /* first get the hash value of the entry */
  idx = nf_icing3_uhf(realm_info->realm_id.id, icing_dev->uhf);
  idx &= (1<<SRAM_SKEY_NXTHOP_IDX_SIZE) - 1;
  printf("write_realm poc key idx = %04x\n", idx);

  /* zero the pad */
  bzero(realm_info->pad, 4);

  /* set/reset the greater than bit */
  realm_info->my_id_is_less = memcmp((uint8_t *)&icing_dev->realm_id, 
                                     (uint8_t *)&realm_info->realm_id, 20) < 0;

  /* write the realm info */
  nf_write_array((const uint8_t *) realm_info,
                 sizeof(nf_icing_realm_info_t),
                 SRAM_SKEY_NXTHOP_BASE_REG + (idx<<(SRAM_SKEY_NXTHOP_ENTRY_BITS+2)),
                 1, &icing_dev->nf2);
}

void nf_icing3_read_realm(nf_icing_realm_info_t *realm_info,
                          nf_icing_dev_t *icing_dev) {
  uint32_t idx;

  /* first get the hash value of the entry */
  idx = nf_icing3_uhf(realm_info->realm_id.id, icing_dev->uhf);
  idx &= (1<<SRAM_SKEY_NXTHOP_IDX_SIZE) - 1;

  /* read the realm info */
  nf_read_array((uint8_t *) realm_info,
                sizeof(nf_icing_realm_info_t),
                SRAM_SKEY_NXTHOP_BASE_REG + (idx<<(SRAM_SKEY_NXTHOP_ENTRY_BITS+2)),
                1, &icing_dev->nf2);
}

int nf_icing3_check_realm(const nf_icing_realm_info_t *realm_info,
                          nf_icing_dev_t *icing_dev) {
  nf_icing_realm_info_t temp;
  nf_icing_realm_info_t zero;
  memcpy(&temp.realm_id, &realm_info->realm_id, sizeof(realm_id_t));
  nf_icing3_read_realm(&temp, icing_dev);
  bzero(&zero, sizeof(nf_icing_realm_info_t));
  if(memcmp(&temp, &zero, sizeof(nf_icing_realm_info_t))) {
    return 1;
  }  
  return 0;
}

void nf_icing3_delete_realm(const nf_icing_realm_info_t *realm_info,
                            nf_icing_dev_t *icing_dev) {
  nf_icing_realm_info_t zero;
  uint32_t idx;

  bzero(&zero, sizeof(nf_icing_realm_info_t));

  /* first get the hash value of the entry */
  idx = nf_icing3_uhf(realm_info->realm_id.id, icing_dev->uhf);
  idx &= (1<<SRAM_SKEY_NXTHOP_IDX_SIZE) - 1;

  /* zero the realm info */
  nf_write_array((const uint8_t *) &zero,
                 sizeof(nf_icing_realm_info_t),
                 SRAM_SKEY_NXTHOP_BASE_REG + (idx<<(SRAM_SKEY_NXTHOP_ENTRY_BITS+2)),
                 1, &icing_dev->nf2);
}  

void nf_icing3_write_poc_key(const vnode_t *vnode, 
                             const poc_key_t *poc_key,
                             nf_icing_dev_t *icing_dev) {
  uint32_t idx;
  uint32_t key[4*(MAXNR+1)];
  uint8_t zero[16];
  poc_key_t poc_key0;
  
  /* get the index */
  memcpy(&idx, vnode, sizeof(uint32_t));
  idx = ntohl(idx);
  idx >>= (32 - VNODE_WIDTH);

  /* write the key */
  nf_write_array((const uint8_t *) poc_key->key,
                 sizeof(poc_key_t),
                 SRAM_POC_KEY_BASE_REG + (idx<<(POC_KEY_ENTRY_BITS+2)),
                 1, &icing_dev->nf2);
  /* write the E(key, 0) */
  bzero(&zero, 16);
  rijndaelKeySetupEnc(key, poc_key->key, 128);
  rijndaelEncrypt(key, 10, zero, poc_key0.key);
#ifdef DEBUG
  pbuf(poc_key->key, 16, "key");
  pbuf(poc_key0.key, 16, "key0");
#endif
  
  nf_write_array((const uint8_t *) poc_key0.key,
                 sizeof(poc_key_t),
                 SRAM_POC_KEY_BASE_REG 
                 + (idx<<(POC_KEY_ENTRY_BITS+2))
                 + sizeof(poc_key_t),
                 1, &icing_dev->nf2);
}

void nf_icing3_read_poc_key(const vnode_t *vnode,
                            poc_key_t *poc_key,
                            nf_icing_dev_t *icing_dev) {
  uint32_t idx;
  
  /* get the index */
  memcpy(&idx, vnode, sizeof(uint32_t));
  idx = ntohl(idx);
  idx >>= (32 - VNODE_WIDTH);
  
  /* read the key */
  nf_read_array((uint8_t *) poc_key,
                sizeof(poc_key_t),
                SRAM_POC_KEY_BASE_REG + (idx<<(POC_KEY_ENTRY_BITS+2)),
                1, &icing_dev->nf2);
}

void nf_icing3_delete_poc_key(const vnode_t *vnode,
                              nf_icing_dev_t *icing_dev) {
  poc_key_t poc_key;
  bzero(&poc_key, sizeof(poc_key_t));
  nf_icing3_write_poc_key(vnode, &poc_key, icing_dev);
}
