/********************************************************************
 * File: nf_icing3.h
 * Author: Jad Naous
 * Description: header file for interacting with the hardware
 ********************************************************************/
#ifndef _NF_ICING3_H_
#define _NF_ICING3_H_

#include "reg_defines_icing_forwarder.h"
#include "nf_icing3_types.h"
#include "nf_icing3_low_level.h"
  
/* Initialize/close the device. Return 0 on success. If deterministic
* is not zero, then the UHF is initialized deterministically. 
* full_uhf is used for simulation to decrease the number of writes.
* Always set to 1 in operation.
*/
int nf_icing3_init_dev(nf_icing_dev_t *icing_dev, 
                       const nf_icing_dev_config_t *config,
                       int deterministic, int full_uhf, int write_uhf);
void nf_icing3_close_dev(nf_icing_dev_t *icing_dev);

/* Write/read a mapping:
 * Realm ID -> {shared secret key, output mask, next_hop_mac} 
 * output_mask and next_hop_mask should be zeros if
 * a realm is not a neighbor, so that packets destined for it as
 * next hop will be dropped. Uses the preset UHF so make sure 
 * you write the uhf table before using these functions. */

/* This function will overwrite whatever is in the hardware, so make sure
 * you use the check function below to see if the entry is available */
void nf_icing3_write_realm(nf_icing_realm_info_t *realm_info,
                           nf_icing_dev_t *icing_dev);

/* reads the entry. */
void nf_icing3_read_realm(nf_icing_realm_info_t *realm_info,
                          nf_icing_dev_t *icing_dev);

/* Check if this realm would collide in the hardware hash table. */
int nf_icing3_check_realm(const nf_icing_realm_info_t *realm_info,
                          nf_icing_dev_t *icing_dev);

/* delete entry */
void nf_icing3_delete_realm(const nf_icing_realm_info_t *realm_info,
                            nf_icing_dev_t *icing_dev);

/* Change the UHF used to create the hashes for the HW hash table
 * such that none of the entries giving in realm_info[] collide.
 * This function will also delete the old entries and write the 
 * new ones. max_iters specifies the max number of tries. 
 * Returns 0 on succes, -1 if max_iters was reached or if 
 * num_entries > the size of the hash table. */
int nf_icing3_change_uhf(nf_icing_realm_info_t **realm_info,
                         int num_entries, int max_iters,
                         nf_icing_dev_t *icing_dev);

/* Write/read a vnode->poc_key mapping. Note that the hardware only looks
 * at the leftmost VNODE_WIDTH bits. It ignores the rest.*/
void nf_icing3_write_poc_key(const vnode_t *vnode, 
                             const poc_key_t *poc_key,
                             nf_icing_dev_t *icing_dev);
void nf_icing3_read_poc_key(const vnode_t *vnode,
                            poc_key_t *poc_key,
                            nf_icing_dev_t *icing_dev);
void nf_icing3_delete_poc_key(const vnode_t *vnode,
                              nf_icing_dev_t *icing_dev);

#endif /* _NF_ICING3_H_ */
