/********************************************************************
 * File: nf_icing3_low_level.c
 * Author: Jad Naous
 * Description: low-level interactions with the hardware
 ********************************************************************/
#include <nf2util.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <arpa/inet.h>

#ifdef ASSERT
#undef ASSERT
#endif
#define ASSERT assert

#ifndef DEBUG
#define DEBUG
#endif

#include "reg_defines_icing_forwarder.h"

/* write an array into the netfpga skipping stride
 * 32-bit words between each address */
void nf_write_array(const uint8_t *array, uint32_t len, 
                    uint32_t base_addr, int stride,
                    struct nf2device *nf2) {
  int i;
  uint32_t val;

  ASSERT(array);
  ASSERT(nf2);

  //  printf("nf_write_array to %08x\n", base_addr);

  /* write 32-bit words */
  for(i=0; i<len/4; i++) {
    //printf("write array using %02x, %02x, %02x, %02x\n", 
    //     array[0], array[1], array[2], array[3]);
    /* construct one 32-bit word */
    val = (array[0]<<24)
      + (array[1]<<16)
      + (array[2]<<8)
      + array[3];
    array += 4;

    /* write the word */
    writeReg(nf2, base_addr+i*4*stride, val);
  }
}  

/* reads an array from the netfpga skipping stride
 * 32-bit words between each address */
void nf_read_array(uint8_t *array, uint32_t len, 
                   uint32_t base_addr, int stride,
                   struct nf2device *nf2) {
  int i;
  uint32_t val;

  ASSERT(array);
  ASSERT(nf2);

  /* read 32-bit words */
  for(i=0; i<len/4; i++) {
    /* read the word */
    readReg(nf2, base_addr+i*4*stride, &val);
    array[3] = val & 0xff;
    val >>= 8;
    array[2] = val & 0xff;
    val >>= 8;
    array[1] = val & 0xff;
    val >>= 8;
    array[0] = val & 0xff;
    array += 4;
  }
}  

/* use these function to read/write the hash function for 
 * the shared keys. The functions either read or write 
 * arrays that are of length ICING_UHF_NUM_BYTES. These values
 * _MUST_ be randomly generated. The UHF must remain secret.*/
void nf_icing3_read_uhf_table(uint8_t *uhf, struct nf2device *dev) {
  nf_read_array(uhf, ICING_UHF_NUM_BYTES, ICING_UHF_BASE_REG, 1, dev);
}

void nf_icing3_write_uhf_table(const uint8_t *uhf, int full_uhf,
                               struct nf2device *dev) {
  int i;
  if(full_uhf) {
    nf_write_array(uhf, ICING_UHF_NUM_BYTES, ICING_UHF_BASE_REG, 1, dev);
  } else {
    for(i=0; i<8; i=i+1) {
      nf_write_array(uhf + 8*256*i, 64, ICING_UHF_BASE_REG + 8*256*i, 1, dev);
    }
  }
}

/* Returns a 16-bit hash using the preset UHF table. */
uint16_t nf_icing3_uhf(const uint8_t *realm_id, const uint8_t *uhf) {
  int i;
  uint16_t hash = 0;
  /* cast the unidimensional array uhf as a 3d array of shorts */
  uint16_t (*uhf_s)[256][4] = (uint16_t (*)[256][4])uhf;

  for(i = 0; i < 20; i++) {
/*     printf("Xoring: hash=%04x, uhf=%04x\n", hash, uhf_s[i%8][*realm_id][i/8]); */
    hash ^= uhf_s[i%8][*realm_id][i/8];
    realm_id++;
  }
  for(i = 20; i < 24; i++) {
/*     printf("Xoring: hash=%04x, uhf=%04x\n", hash, uhf_s[i%8][0][i/8]); */
    hash ^= uhf_s[i%8][0][i/8];
  }
/*   printf("********Result: hash=%04x\n", hash); */
  return ntohs(hash);
}

/* Set/read the mac addresses of all ports. mac must be uint8_t[4][6] */
void nf_icing3_write_port_mac_addrs(const uint8_t *mac, struct nf2device *dev) {
  nf_write_array(mac, 24, ICING_PORT_MAC_BASE_REG, 1, dev);
}

void nf_icing3_read_port_mac_addrs(uint8_t *mac, struct nf2device *dev) {
  nf_read_array(mac, 24, ICING_PORT_MAC_BASE_REG, 1, dev);
} 

/* Set/read the realm ID. Currently not needed. */
void nf_icing3_write_realm_id(const uint8_t *realm_id, struct nf2device *dev) {
  nf_write_array(realm_id, 20, ICING_MY_REALM_ID_BASE_REG, 1, dev);
}

void nf_icing3_read_realm_id(uint8_t *realm_id, struct nf2device *dev) {
  nf_read_array(realm_id, 20, ICING_MY_REALM_ID_BASE_REG, 1, dev);
}
