package ICING3Crypt;
use Carp;
use Crypt::Rijndael;

use constant CHI_TOOL => "$ENV{NF2_DESIGN_DIR}/lib/Perl5/chi/chi_tool";
use constant GROESTL_TOOL => "$ENV{NF2_DESIGN_DIR}/lib/Perl5/groestl/groestl_tool";
use constant PMAC_TOOL => "$ENV{NF2_DESIGN_DIR}/lib/Perl5/pmac/pmac_tool";

my $chi = CHI_TOOL;
my $pmac = PMAC_TOOL;
my $groestl = GROESTL_TOOL;

# Input: Array of bytes - data to hash
# Output: Array of 32 bytes - hash-256(data)
sub hash256 {
#  return groestl256(@_);
  return chi256(@_);
}

# Input: Array of bytes - data to hash
# Output: Array of 32 bytes - GROESTL-256(data)
sub groestl256 {
  # concatenate the bytes into a hex string
  my $data = join("", map(sprintf ("%02x", $_), @_));

#  printf "GROESTL'ing %s\n", join(":", map(sprintf ("%02x", $_), @_));
  # get result as a hex string
  $result = `$groestl $data`;
#  printf("GROESTL result: $result\n");

  # split into array of bytes
  my @res = ($result =~ /(.{2})/g);

  # return as decimal
  return map(hex($_), @res);
}

# Input: Array of bytes - data to hash
# Output: Array of 32 bytes - CHI-256(data)
sub chi256 {
  # concatenate the bytes into a hex string
  my $data = join("", map(sprintf ("%02x", $_), @_));

  printf "CHI'ing %s\n", join(":", map(sprintf ("%02x", $_), @_));
  # get result as a hex string
  $result = `$chi $data`;
#  printf("CHI result: $result\n");

  # split into array of bytes
  my @res = ($result =~ /(.{2})/g);

  # return as decimal
  return map(hex($_), @res);
}

# Input:  ref to Array of 16 bytes - key for PMAC
#         ref to Array of bytes - data to MAC
# Output: Array of 16 bytes - PMAC(key, data)
sub pmac {
  # concatenate the bytes into a hex string
  my $key = join("", map(sprintf ("%02x", $_), @{$_[0]}));
  my $data = join("", map(sprintf ("%02x", $_), @{$_[1]}));

  # get result as a hex string
  $result = `$pmac $key $data`;

  # split into array of bytes
  my @res = ($result =~ /(.{2})/g);

  # return as decimal
  return map(hex($_), @res);
}

# Input: ref to array of 16 bytes (128 bits) - key to use
#        ref to array of 32 bytes (256 bits) - block to mac
# Output: array of 16 bytes (128 bits) - MAC tag
# Algorithm:
#    CBC2-MAC-128(key[0..127], block[0..255]) {
#       X   <- AES-ENC(key, block[0..127])
#       Tag <- AES-ENC(key, X XOR block[128..255])
#        output Tag
#    }
sub cbc2_mac_128 {
  my @key = @{$_[0]};
  my @block = @{$_[1]};

  # count key length
  if(scalar @key != 16) {
    confess "Key needs to be 16 bytes for CBC2-MAC. Found " . scalar @key . "\n";
  }
  if(scalar @block != 32) {
    confess "block needs to be 32 bytes for CBC2-MAC. Found " . scalar @block . "\n";
  }

  my $aes = Crypt::Rijndael->new(pack("C*", @key));
  my @X = unpack "C*", $aes->encrypt(pack("C*", @block[0..15]));
#  printf "CBC-MAC2: K:(%s)\n", join (":", map (sprintf ("%02x", $_), @key));
#  printf "CBC-MAC2: block:(%s)\n", join (":", map (sprintf ("%02x", $_), @block));
#  printf "CBC-MAC2: plain0:(%s)\n", join (":", map (sprintf ("%02x", $_), @block[0..15]));
#  printf "CBC-MAC2: X:(%s)\n", join (":", map (sprintf ("%02x", $_), @X));

  # XOR the bytes
  my @nxt = map {$X[$_] ^ $block[$_+16]} 0..15;
#  printf "CBC-MAC2: X^block[16..31]:(%s)\n", join (":", map (sprintf ("%02x", $_), @nxt));

  my @tag = unpack "C*", $aes->encrypt(pack("C*", @nxt));
#  printf "CBC-MAC2: Tag:(%s)\n", join (":", map (sprintf ("%02x", $_), @tag));
  return @tag;
}

# Input: ref to array of 16 bytes (128 bits) - key to use
#        ref to array of 32 bytes (256 bits) - block to mac
# Output: array of 12 bytes (96 bits) - MAC tag
# Algorithm:
#    PRF1-96(key[0..127], block[0..255]) {
#        output first 12 bytes of (CBC2-MAC-128(key, block))
#    }
sub prf1_96 {
  my @tag = cbc2_mac_128(@_);
  return @tag[0..11];
}

# Input: ref to array of 20 bytes (160 bits) - a
#        ref to array of 20 bytes (160 bits) - b
sub less_than {
  my $a = shift;
  my $b = shift;

  foreach my $i (0..19) {
    if($a->[$i] > $b->[$i]) {
      return 0;
    } elsif($a->[$i] < $b->[$i]) {
      return 1;
    }
  }
  return 0;
}


# Input: ref to array of 20 bytes (160 bits) - R_sndr
#        ref to array of 20 bytes (160 bits) - R_rcvr
#        ref to array of 16 bytes (128 bits) - k_{sndr, rcvr}
#        ref to array of 32 bytes less one bit (255 bits) - block
# Output: array of 12 bytes (96 bits) - MAC tag
# Algorithm:
#    When calculating the MAC using a shared symmetric key k_{i,j} we
#    will assume k_{i,j} = k_{j,i}, and use the following function
#    where block is 255 bits:
#
#    PRF2-96(R_sndr, R_rcvr, k_{sndr,rcvr}, block[0..254]) {
#        tag = (R_sndr < R_rcvr) ? 1 : 0;
#        output PRF1-96(k_{sndr,rcvr}, tag || block)
#    }
sub prf2_96 {
  my $R_sndr = shift;
  my $R_rcvr = shift;
  my $k      = shift;
  my $block  = shift;

  # compare R_sndr and R_rcvr
  my $t = less_than($R_sndr, $R_rcvr);
  $block->[0] = ($block->[0] & 0x7f) | ($t << 7);

  printf "PRF2: Rsndr=(%s)\n", join (":", map (sprintf ("%02x", $_), @{$R_sndr}));
  printf "PRF2: Rrcvr=(%s)\n", join (":", map (sprintf ("%02x", $_), @{$R_rcvr}));
  printf "PRF2: K=(%s)\n", join (":", map (sprintf ("%02x", $_), @{$k}));
  printf "PRF2: block=(%s)\n", join (":", map (sprintf ("%02x", $_), @{$block}));

  my @ret = prf1_96($k, $block);

  $block->[0] = ($block->[0] & 0x7f);
  printf "PRF2: Tag=(%s)\n", join (":", map (sprintf ("%02x", $_), @ret));

  return @ret;
}

1;
