package ICING3Dev;
use strict;
use warnings;

use NF2::PacketGen;
use ICING3Realm;
use NF2::TestLib;

use constant ID => 'id';
use constant MACS => 'macs';
use constant HW => 'hw';

use constant SIM_TOOL => "$ENV{NF2_DESIGN_DIR}/lib/Perl5/sim_tool/sim_tool";
use constant LD_LIBRARY_PATH_SIM => "$ENV{NF2_DESIGN_DIR}/lib/C/:$ENV{NF2_DESIGN_DIR}/lib/Perl5/sim_tool/";
use constant LD_LIBRARY_PATH_HW => LD_LIBRARY_PATH_SIM();

# create a new device
# inputs: ID => ref to list of 20 bytes - my realm ID
#         MACS => ref to list of 24 bytes - MACS of 4 ports
#         HW => optional running real HW not sim.
sub new {
  my $class = shift;
  my %args = @_;
  my $hw = (defined $args{HW()}) ? $args{HW()} : 0;
  my $self = {ID() => $args{ID()},
              MACS() => $args{MACS()},
	      dev => "",
	      HW() => $hw,
             };
  bless $self, $class;
}

sub bytes2hexstr {
  my $bytes = shift; # ref to byte array
  return join("", map(sprintf ("%02x", $_), @{$bytes}));
}

sub bytes2words {
  return unpack("N*", pack("C*", @_));
}

sub checkSimTool {
  my $self = shift;
  my $cmd = sprintf('LD_LIBRARY_PATH="%s" %s',
		    $self->{HW()} ? LD_LIBRARY_PATH_HW() : LD_LIBRARY_PATH_SIM(),
		    SIM_TOOL());
  my $check = `$cmd`;
  chomp $check;
  if($check ne "Hello 3: libnf2sim!") {
    die "ERROR: Check Sim Tool failed: $check\n";
  }
  else{
    print "checkSimTool success!\n";
  }
}

# Takes a list of string args to pass to sim_tool and executes
# results on sim device
# sim_tool returns data as follows:
#   Writes: W:hex_addr:hex_val
#   Reads : R:hex_addr
#           E:hex_val_expected (maybe later in the output)
#   State : S:hex_string
#   Return: V:hex_string
# Inputs: $cmd  - cmd to pass to sim_tool
#         $args - ref to list of string args to cmd
# Outputs: string of whatever cmd returns
sub simCmd{
  my $self = shift;
  my $dev = $self->{dev};
  my $cmd = sprintf('LD_LIBRARY_PATH="%s" %s',
		    $self->{HW()} ? LD_LIBRARY_PATH_HW() : LD_LIBRARY_PATH_SIM(),
		    SIM_TOOL());
  my $unparsed;
  if($dev) {
#    print "dev: $dev\n";
    $cmd = $cmd . " $_[0] $dev @{$_[1]} 2>&1";
  } else {
#    print "dev: None\n";
    $cmd = $cmd . " $_[0] @{$_[1]} 2>&1";
  }

#  print("simCmd: $_[0] @{$_[1]}\n");
#  print "simCmd: $cmd\n";
  $unparsed = `$cmd | tee simCmd.log`;
#  print "Unparsed: $unparsed\n";

  if($unparsed =~ m/(ERROR)|(fault)/ or not $unparsed) {
    print "Unparsed: $unparsed\n";
    die "Error using sim_tool\n";
  }

  # parse result
  my %h = ();
  foreach("W", "R", "E", "S", "V") {
    $h{$_} = [$unparsed =~ /($_:[^\s]*)/g];
  }

#  print "Executing simCmd reads and writes\n";

  my @writes = @{$h{W}};
  foreach (@writes) {
    my @t = split /:/;
    #    print "parsing: $_ into @t\n";
    if($self->{HW()}) {
      nftest_regwrite("nf2c0", hex($t[1]), hex($t[2]));
    } else {
      nf_PCI_write32(hex($t[1]), hex($t[2]));
    }
  }

  my @reads = @{$h{R}};
  my @exp = @{$h{E}};
  foreach (@reads) {
    my @r = split /:/;
    my @e = split /:/, shift @exp;
    #    print "parsing: $_ into @r, @e\n";
    if($self->{HW()}) {
      my $val = nftest_regread("nf2c0", hex($r[1]));
      printf("Read addr %08x val 0x%08x = 0d%u", hex($r[1]), $val, $val);
    } else {
      nf_PCI_read32(hex($r[1]), hex($e[1]));
    }
  }

  my @state = split /:/, $h{S}->[0];
  $self->{dev} = $state[1];

  my @results = @{$h{V}};
  map {$_ =~ s/^V://} @results;

  #  print "simCmd done\n";
  return @results;
}

# Initialize the device
# Inputs: full_uhf - bool: use all bits of each byte in uhf
#                          or just least sig 3
#         write_uhf - bool: write the uhf to hw or use the preloaded uhf
# Outputs: None
sub initDevSim{
  my $self = shift;
  my $full_uhf = shift;
  my $write_uhf = shift;
  my $macs = bytes2hexstr($self->{MACS()});
  my $id = bytes2hexstr($self->{ID()});
  $self->simCmd("init", [$macs, $id, $full_uhf, $write_uhf]);
}

# Write an entry using the simulation API
# inputs: $realm_id    - ref to array of 20 bytes. Realm ID to write
#         $skey        - ref to array of 16 bytes. Shared symmetric key.
#         $nxt_hop_mac - ref to array of 6 bytes. Next Hop Mac.
#         $output_mask - Single byte. Output ports bitmask.
my %poc_keys = ();
sub writeRealmSim {
  my $self = shift;
  my ($realm_id, $skey, $nxt_hop_mac)
    = map(bytes2hexstr($_), @_[0..2]);
  my $output_mask = sprintf("%02x", $_[3]);

  print "write_realm:\n";
  print "    id: $realm_id\n";
  print "    skey:$skey\n";
  print "    out: $output_mask\n";
  print "    mac: $nxt_hop_mac\n";
  $self->simCmd("write_realm",
                [$realm_id, $skey, $output_mask, $nxt_hop_mac]);

  # check the poc indices for collisions
  my @idx = (`cat simCmd.log` =~ /(write_realm poc key idx = [^\s]*)/g);
  if(scalar @idx > 1) {
    die "Huh? Wrote more than one poc key idx?";
  }

  print `grep "write_realm poc key" simCmd.log`, "\n";
  die "PoC Key collision" if exists $poc_keys{$idx[0]};

  $poc_keys{$idx[0]} = 1;
}

# read an entry using the simulation API
# inputs: $realm_id    - ref to array of 20 bytes. Realm ID to read
#         $skey        - ref to array of 16 bytes. Expected Shared symmetric key.
#         $nxt_hop_mac - ref to array of 6 bytes. Expected Next Hop Mac.
#         $output_mask - Single byte. Expected Output ports bitmask.
sub readRealmSim {
  my $self = shift;
  my ($realm_id, $skey, $nxt_hop_mac)
    = map(bytes2hexstr($_), @_[0..2]);
  my $output_mask = sprintf("%02x", $_[3]);

  $self->simCmd("read_realm",
                [$realm_id, $skey, $output_mask, $nxt_hop_mac]);
}

# Delete an entry
# inputs: $realm_id    - ref to array of 20 bytes. Realm ID to delete
sub deleteRealmSim {
  my $self = shift;
  my $realm_id = bytes2hexstr(shift);

  $self->simCmd("delete_realm",[$realm_id]);
  my @idx = (`cat simCmd.log` =~ /(write_realm poc key idx = [^\s]*)/g);
  if(scalar @idx > 1) {
    die "Huh? Wrote more than one poc key idx?";
  }

  delete $poc_keys{$idx[0]};
}

# write a POC key into hw
# inputs: $vnode - ref to array of 4 bytes corresponding to vnode
#         $poc_key - ref to array of 16 bytes
sub writePocKey {
  my $self = shift;
  my @args = map(bytes2hexstr($_), @_);

  $self->simCmd("write_poc_key", \@args);
}

# read a POC key
# inputs: $vnode - ref to array of 4 bytes corresponding to vnode
#         $poc_key - expected poc key ref to array of 16 bytes
sub readPocKey {
  my $self = shift;
  my @args = map(bytes2hexstr($_), @_);

  $self->simCmd("read_poc_key", \@args);
}

# delete a POC key
# inputs: $vnode - ref to array of 4 bytes corresponding to vnode
sub deletePocKey {
  my $self = shift;
  my @args = map(bytes2hexstr($_), @_);

  $self->simCmd("delete_poc_key", \@args);
}

# Fill up the SRAM poc key table
# inputs: $fill_all: bool deciding whther to write all entries or 
#                    just the first 16.
sub fillPocKeyTable {
  my $self = shift;
  my $fill_all = shift;

  my $max = $fill_all ? 2**main::VNODE_WIDTH()-1 : 2**4-1;
  my $filename = "poc_key_cache_$max";
  # Use cache if available
  if(-r $filename) {
    my @writes = (`cat $filename` =~ /(W:[^\s]*)/g);
    foreach(@writes) {
      my @t = split /:/;
      if($self->{HW()}) {
	nftest_regwrite("nf2c0", hex($t[1]), hex($t[2]));
      } else {
	nf_PCI_write32(hex($t[1]), hex($t[2]));
      }
    }
  }
  else {
    foreach my $i (0..$max) {
      my $vnode = [unpack("C*", pack("N", ($i<<(32-main::VNODE_WIDTH()))))];
      $self->writePocKey($vnode, [Realm::getPocKey($vnode)]);
      # store data in cache
      system("cat simCmd.log >> $filename");
    }
  }
}

1;
