package VTag;
use strict;
use warnings;
use POSIX;
use Carp qw(cluck confess);

use constant VA => 'Va';
use constant VB  => 'Vb';

sub new {
  my ($class, %arg) = @_;

  confess "Need to define VA" unless defined $arg{VA()};
  confess "Need to define VB" unless defined $arg{VB()};

  confess "VA needs to be 12 bytes long" unless scalar @{$arg{VA()}} == 12;
  confess "VB needs to be 4 bytes long" unless scalar @{$arg{VB()}} == 4;

  my $tag =
    { VA() => $arg{VA()},
      VB() => $arg{VB()},
    };

  bless $tag, $class;
}

sub vabytes {
  my $self = shift;
  return @{$self->{VA()}};
}

sub vbbytes {
  my $self = shift;
  return @{$self->{VB()}};
}

sub bytes {
  my $self = shift;
#  cluck "Called bytes!\n";
  return ($self->vabytes, $self->vbbytes());
}

sub clear {
  my $self = shift;
  $self->{VA()} = [(0) x 12];
  $self->{VB()} = [(0) x 4];
}

sub setVaBytes {
  my $self = shift;
  my @bytes = @_[0..11];
  $self->{VA()} = \@bytes;
}

sub setVbBytes {
  my $self = shift;
  my @bytes = @_[0..3];
  $self->{VB()} = \@bytes;
}


package ICING3Packet;
use strict;
use warnings;
use POSIX;
use List::Util qw(min);

use NFUtils::GenericByteObject;
use ICING3Crypt;
use Carp qw(cluck);

use constant ETH_SRC => 'eth_src';
use constant ETH_DST => 'eth_dst';
use constant ETH_TYPE => 'eth_type';
use constant ICING_VER => 'icing_ver';
use constant ICING_PATH_LEN => 'icing_path_len';
use constant ICING_PATH_IDX => 'icing_path_idx';
use constant ICING_ERROR_IDX => 'icing_error_idx';
use constant ICING_IV => 'icing_iv';
use constant ICING_PATH => 'icing_path';
use constant ICING_VTAGS => 'icing_vtags';
use constant ICING_PROTO => 'icing_proto';
use constant ICING_PKT_LEN => 'icing_pkt_len';
use constant PAYLOAD_LEN => 'payload_len';
use constant PAYLOAD     => 'payload';

use constant ICING_ETH_TYPE => 0x9090;
use constant ICING_CURR_VER => 1;

use constant REALM_ID_SIZE => 20;
use constant REALM_VOQ_SIZE => 4;

use constant VTAG_VA_SIZE => 12;
use constant VTAG_VB_SIZE => 4;

use constant CTOKEN_SERVER => 'ctoken_server';

use constant MAX_PKT_LEN => 2040;

sub new {
  my ($class, %arg) = @_;

  my @bytes = ();
  $arg{ETH_SRC()} = [map(int(rand(256)), (1..6))] unless defined $arg{ETH_SRC()};
  $arg{ETH_DST()} = [map(int(rand(256)), (1..6))] unless defined $arg{ETH_DST()};
  $arg{ETH_TYPE()} = ICING_ETH_TYPE unless defined $arg{ETH_TYPE()};
  $arg{ICING_VER()} = ICING_CURR_VER unless defined $arg{ICING_VER()};
  if(!defined $arg{ICING_PATH_LEN()}) {
    if(defined $arg{ICING_PATH()}) {
      $arg{ICING_PATH_LEN()} = scalar @{$arg{ICING_PATH()}};
    } else {
      $arg{ICING_PATH_LEN()} = int(rand(46))+2;
    }
  }
  $arg{ICING_PATH_IDX()} = int(rand($arg{ICING_PATH_LEN()}-1))+1 unless defined $arg{ICING_PATH_IDX()};
  $arg{ICING_IV()} = [map(int(rand(256)), (1..7))] unless defined $arg{ICING_IV()};
  $arg{ICING_PATH()} = [map(RealmToken->new(RealmToken::ID() => map(int(rand(0x100)), (1..REALM_ID_SIZE)),
                                            RealmToken::VOQ() => map(int(rand(0x100)), (1..REALM_VOQ_SIZE))),
                            (1..$arg{ICING_PATH_LEN()}))] unless defined $arg{ICING_PATH()};
  $arg{ICING_VTAGS()} = [map(VTag->new(VTag::VA() => map(int(rand(0x100)), (1..VTAG_VA_SIZE)),
                                       VTag::VB() => map(int(rand(0x100)), (1..VTAG_VB_SIZE))),
                            (2..$arg{ICING_PATH_LEN()}))] unless defined $arg{ICING_VTAGS()};
  $arg{ICING_PROTO()} = int(rand(0x100)) unless defined $arg{ICING_PROTO()};
  $arg{ICING_PKT_LEN()} = int(rand(0x10000)) unless defined $arg{ICING_PKT_LEN()};
  $arg{ICING_ERROR_IDX()} = 0 unless defined $arg{ICING_ERROR_IDX()};

  if(defined $arg{PAYLOAD()}) {
    $arg{PAYLOAD_LEN()} = scalar @{$arg{PAYLOAD()}};
  } else {
    $arg{PAYLOAD_LEN()} = int(rand(MAX_PKT_LEN - ( 27 + ($arg{ICING_PATH_LEN()}-1)*40 + 24))) unless defined $arg{PAYLOAD_LEN()};
    if ($arg{PAYLOAD_LEN()}>0) {
      my @payload = map($_ & 0xff, (1..$arg{PAYLOAD_LEN()}));
      $arg{PAYLOAD()} = \@payload;
    } else {
      $arg{PAYLOAD()} = [] ;
    }
  }

  bless \%arg, $class;
}

sub pktLen {
  return scalar shift()->tobytes();
}

sub tobytes {
  my $self = shift;
  my @bytes = ();

  push @bytes, @{$self->{ETH_DST()}};
  push @bytes, @{$self->{ETH_SRC()}};
  push @bytes, ByteArrayUtils::int_to_bytes(2, $self->{ETH_TYPE()});
  push @bytes, ByteArrayUtils::int_to_bytes(1, $self->{ICING_VER()});
  push @bytes, ByteArrayUtils::int_to_bytes(1, $self->{ICING_PATH_LEN()});
  push @bytes, ByteArrayUtils::int_to_bytes(1, $self->{ICING_PATH_IDX()});
  push @bytes, @{$self->{ICING_IV()}};
  push @bytes, map($_->bytes(), @{$self->{ICING_PATH()}});
  push @bytes, map($_->bytes(), @{$self->{ICING_VTAGS()}});
  push @bytes, ByteArrayUtils::int_to_bytes(1, $self->{ICING_PROTO()});
  push @bytes, ByteArrayUtils::int_to_bytes(2, $self->{ICING_PKT_LEN()});
  push @bytes, ByteArrayUtils::int_to_bytes(1, $self->{ICING_ERROR_IDX()});
  push @bytes, @{$self->{PAYLOAD()}};
  return @bytes;
}

sub packed {
  return pack("C*", shift()->tobytes());
}

sub hexBytes {
  my @bytes = shift()->tobytes();

  if (scalar(@bytes) > 0) {
    my @tmp = map (sprintf ("%02x",$_), @bytes);
    return join(' ',@tmp).' ';
  }
  else {
    return "";
  }
}

# print pkt nicely
# input: offset from which to start printing
sub prettyPrint {
  my @hexb = shift->tobytes();
  @hexb = @hexb[shift..$#hexb];
  while ((scalar @hexb) % 16 != 0) {
    push @hexb, (0);
  }
  my $count=0;
  while(@hexb) {
    printf "0x%04x:  ", $count;
    foreach (0..7) {
      printf "%02x%02x ", shift @hexb, shift @hexb;
    }
    print "\n";
    $count += 16;
  }
}

sub newSenderPkt {
  my ($class, %args) = @_;

  my $ctokenserver = $args{CTOKEN_SERVER()};
  my @path = @{$args{ICING_PATH()}};
  my $payload_len = $args{PAYLOAD_LEN()};
  my $proto = $args{ICING_PROTO()};

  my $max_pkt_len = (defined $args{'max_pkt_len'}) ?  $args{'max_pkt_len'} : MAX_PKT_LEN;

  my $path_len = scalar @path;
  if(defined $payload_len) {
    $payload_len = min(($max_pkt_len - (28 + ($path_len-1)*40 + 24)), $payload_len);
  }
  my @payload;
  if(defined $args{PAYLOAD()}) {
    @payload = @{$args{PAYLOAD()}};
    $payload_len = scalar @payload;
  } else {
    $payload_len = int(rand($max_pkt_len - ( 27 + ($path_len-1)*40 + 24))) unless defined $payload_len;
    @payload = map($_&0xff, (1..$payload_len));
  }

  $proto = 0x11 unless defined $proto;

  printf "Creating pkt with path len=%u\n", scalar @path;

  my $icing_pkt_len = 14 + ($path_len-1)*40 + 24 + $payload_len;

  # get pocs
  my @ctokens = $ctokenserver->getCTokens(@path);

  print "Path PoCs: \n";
  for(my $i=0; $i<scalar @ctokens; $i++) {
    printf "    %u=(%s)\n", $i+1, join (":", map (sprintf ("%02x", $_), @{$ctokens[$i]}));
  }

  my @iv = (1..7);

  # get HASH256(version || IV || P || payload)
  my @hash = ICING3Crypt::hash256(ICING_CURR_VER, @iv,
                                  map($_->bytes(), @path), # path in bytes
                                  $proto, unpack("C*", pack("n", $icing_pkt_len)), 0,
                                  @payload);

  # use only last 247 bits of the hash
  shift @hash;
  $hash[0] = $hash[0] & 0x7f;
  printf "Hash: %s\n", join(":", map(sprintf("%02x", $_), @hash));

  # realm R_0, the source, computes, for 1 <= j <= n:
  #    A[j] = Va_{-1}[j] = PRF1-96(poc_{P,j}, 0^9 || H)
  #
  # Let i be 8 bits.
  #
  #  for 0 <= i <= n:
  #    for i < j <= n, let:
  #        temp    = PRF2-96(R_i, R_j, k_{i,j}, i || H) // remember i is 8 bits
  #        Va_i[j] = Va_{i-1}[j] XOR temp
  #        Vb_i[j] = last 4 bytes of Va{-1}[j]  // NOTE: DOES NOT DEPEND ON i
  # 
  # For the sender, i = 0
  my @vtags = ();
  for(my $i=0; $i<scalar @ctokens; $i++) {
    my $idx = $i + 1;
    my @key = Realm::getSharedKey($path[0]->{Realm::ID()},
                                  $path[$idx]->{Realm::ID()});
    printf "Key shared with %u=(%s)\n", $idx, join (":", map (sprintf ("%02x", $_), @key));
    my @A = ICING3Crypt::prf1_96($ctokens[$i], [0, @hash]);
    printf "A for %u=(%s)\n", $idx, join (":", map (sprintf ("%02x", $_), @A));
    my @vb = @A[8..11];
    my @temp = ICING3Crypt::prf2_96($path[0]->{Realm::ID()}, # my realm id
                                    $path[$idx]->{Realm::ID()}, # other realm id
                                    \@key,       # k_{sndr,rcvr}
                                   [0, @hash]); # i || H
    printf "Temp for %u=(%s)\n", $idx, join (":", map (sprintf ("%02x", $_), @temp));
    my @va = map {$A[$_] ^ $temp[$_]} 0..11;
    printf "Va for %u=(%s)\n", $idx, join (":", map (sprintf ("%02x", $_), @va));

    push @vtags, VTag->new(VTag::VA() => \@va,
                           VTag::VB() => \@vb);
  }

  return ICING3Packet->new(ETH_SRC() => $args{ETH_SRC()},
                           ETH_DST() => $args{ETH_DST()},
                           ICING_PATH_IDX() => 1,
                           ICING_IV() => \@iv,
                           ICING_PATH() => \@path,
                           ICING_VTAGS() => \@vtags,
                           ICING_PROTO() => $proto,
                           ICING_PKT_LEN() => $icing_pkt_len,
                           PAYLOAD() => \@payload,
                          );
}

sub defaultPayloadGen {
  return (0xff & shift);
}


1;
