package RealmToken;
use strict;
use warnings;
use POSIX;
use Carp;

use constant ID => 'id';
use constant VOQ  => 'voq';

sub new {
  my ($class, %arg) = @_;

  confess "Need to define ID" unless defined $arg{ID()};
  $arg{VOQ()} = [0, 0, 0, 0];
#  $arg{VOQ()} = [map(int(rand(0x100)), (1..2)), 0, 0]  unless defined $arg{VOQ()};

  confess "ID needs to be 20 bytes long. Found ".(scalar @{$arg{ID()}}) unless scalar @{$arg{ID()}} == 20;

  my $self = {};
  foreach (keys %arg) {
    $self->{$_} = $arg{$_};
  }

  bless $self, $class;
}

sub bytes {
  my $self = shift;
  return (@{$self->{ID()}}, @{$self->{VOQ()}});
}

package Realm;
use strict;
use warnings;
use POSIX;
use Carp qw(cluck confess);

use Crypt::Rijndael;
use ICING3Crypt;

use constant ID => 'id';
use constant NEXT_HOP_MAC => 'mac';
use constant OUTPUT_QUEUE => 'outqueue';

my $nxt_oq = 0;

# input: hash
# ID => ref to list of 20 bytes
# NEXT_HOP_MAC => ref to list of 6 bytes
# OUTPUT_QUEUE => byte
sub new {
  my ($class, %arg) = @_;

  confess "Need to define ID" unless defined $arg{ID()};
#  $arg{OUTPUT_QUEUE()} = 2*((int(rand(16))) >> 2) unless defined $arg{OUTPUT_QUEUE()};
  $arg{OUTPUT_QUEUE()} = 2*$nxt_oq unless defined $arg{OUTPUT_QUEUE()};
  $nxt_oq = ($nxt_oq + 1) & 0x3;
  $arg{NEXT_HOP_MAC()} = [0x00, 0x00, 0xca, map($arg{OUTPUT_QUEUE()}, (1..3))] unless defined $arg{NEXT_HOP_MAC()};

  confess "ID needs to be 20 bytes long. Found ".(scalar @{$arg{ID()}}) unless scalar @{$arg{ID()}} == 20;
  confess "NEXT_HOP_MAC needs to be 6 bytes long" unless scalar @{$arg{NEXT_HOP_MAC()}} == 6;

  my $self = {};
  foreach (keys %arg) {
    $self->{$_} = $arg{$_};
  }

  bless $self, $class;
}

sub prettyPrint{
  my $self = shift;
  my $id = join(":", map(sprintf("%02x", $_), @{$self->{ID()}}));
  printf ("%s\n", $id);
}

sub getPocKey {
  my $vnode = shift; # vnode
  return map(@{$vnode}, (1..4));
}

# input: path as a list of RealmTokens - P
# output: list of bytes = PMAC(s_i, P) where s_i is the PoC key
sub createCToken {
  my $self = shift;
  my @realm_tokens = @_;

  my @msg_bytes = ();
  my $my_realm_token;
  my $id_str = pack('C*', @{$self->{ID()}});
  foreach (@realm_tokens) {
    # if this realm token is me, then store it for later use.
    if(pack('C*', @{$_->{RealmToken::ID()}}) eq $id_str) {
      $my_realm_token = $_;
    }
    push @msg_bytes, $_->bytes();
  }
  print "createCToken: bytes PMAC'ed:\n";
  printf "      %s\n", join ("_", map (sprintf ("%02x", $_), @msg_bytes));

  my @key = getPocKey($my_realm_token->{RealmToken::VOQ()});

  my @poc = ICING3Crypt::pmac(\@key, \@msg_bytes);
  printf "Key used: %s\n", join ("_", map (sprintf ("%02x", $_), @key));
  printf "PMAC: %s\n", join ("_", map (sprintf ("%02x", $_), @poc));
  return @poc;
}

# only does ICING specific things
# input: ICING3Packet
# output: modified packet or undef on error
sub forwardPkt {
  my ($self, $pkt) = @_;

  return undef unless defined $pkt;

  if($pkt->{ICING3Packet::ICING_VER()}
     != ICING3Packet::ICING_CURR_VER) {
    printf "Rcvd wrong pkt version %u!\n", $pkt->{ICING3Packet::ICING_VER()};
    return undef;
  } elsif ($pkt->{ICING3Packet::ICING_PATH_IDX()} >= $pkt->{ICING3Packet::ICING_PATH_LEN()} - 1) {
    printf("Path index too large: idx=%u, pathlen=%u!\n",
           $pkt->{ICING3Packet::ICING_PATH_IDX()},
           $pkt->{ICING3Packet::ICING_PATH_LEN()});
    return undef;
  } elsif ($pkt->{ICING3Packet::ICING_PATH_IDX()} <= 0) {
    printf "Path idx too small %u!\n", $pkt->{ICING3Packet::ICING_PATH_IDX()};
    return undef;
  }

  printf "Rcvd pkt. Path Idx=%u\n", $pkt->{ICING3Packet::ICING_PATH_IDX()};

  # get Hash256(version || IV || P || payload)
  my @hash = ICING3Crypt::hash256($pkt->{ICING3Packet::ICING_VER()},
                                  @{$pkt->{ICING3Packet::ICING_IV()}},
                                  map($_->bytes(), @{$pkt->{ICING3Packet::ICING_PATH()}}),
                                  $pkt->{ICING3Packet::ICING_PROTO()},
                                  unpack("C*", pack("n", $pkt->{ICING3Packet::ICING_PKT_LEN()})),
                                  0,
                                  @{$pkt->{ICING3Packet::PAYLOAD()}});

  # use only last 247 bits of the hash
  shift @hash;
  $hash[0] = $hash[0] & 0x7f;
  printf "Hash: %s\n", join("", map(sprintf("%02x", $_), @hash));

  # get the PoC
  my @c_i = $self->createCToken(@{$pkt->{ICING3Packet::ICING_PATH()}});
  printf "Calc PoC=(%s)\n",join (":", map (sprintf ("%02x", $_), @c_i));

  my @A = ICING3Crypt::prf1_96(\@c_i, [0, @hash]);
  printf "Calc A=(%s)\n",join (":", map (sprintf ("%02x", $_), @A));

  my $ci_low_4 = pack('C*', @A[8..11]);

  # Check that the sender has the required capability
  my $vtags = $pkt->{ICING3Packet::ICING_VTAGS()};
  my $idx = $pkt->{ICING3Packet::ICING_PATH_IDX()};
  my $pathlen = $pkt->{ICING3Packet::ICING_PATH_LEN()};
  my $my_v = $vtags->[$idx-1];
  my $vb = pack('C*', $my_v->vbbytes());
  printf "Pkt Va=(%s)\n",join (":", map (sprintf ("%02x", $_), $my_v->vabytes()));
  printf "Pkt Vb=(%s)\n",join (":", map (sprintf ("%02x", $_), $my_v->vbbytes()));

  if ($vb ne $ci_low_4) {
    printf("Last 4 bytes of Pkt A(%s) doesn't match my Vb(%s)!\n",
           unpack("H*", $ci_low_4), unpack("H*", $vb));
    return undef;
  }

  # Check validation
  my @va = @A;
  for(my $i=0; $i<$idx; $i++) {
    my $realm_token = $pkt->{ICING3Packet::ICING_PATH()}->[$i];
    my @realm_id = @{$realm_token->{RealmToken::ID()}};
    my @my_id = @{$self->{ID()}};
    my @key = getSharedKey(\@my_id, \@realm_id);
    printf "Key Shared with $i=(%s)\n",join (":", map (sprintf ("%02x", $_), @key));
    my @temp = ICING3Crypt::prf2_96(\@realm_id, \@my_id, \@key, 
                                   [$i >> 1, (0x80 & ($i << 7)) | $hash[0], @hash[1..$#hash]]);
    printf "Calc tmp_$i=(%s)\n",join (":", map (sprintf ("%02x", $_), @temp));
    @va = map {$va[$_] ^ $temp[$_]} 0..11;
    printf "Calc Va_$i=(%s)\n",join (":", map (sprintf ("%02x", $_), @va));
  }

  my @va_bytes = $my_v->vabytes();
  printf "My pkt Va=(%s)\n",join (":", map (sprintf ("%02x", $_), @va_bytes));
  for(my $i=0; $i<12; $i++) {
    if($va[$i] != $va_bytes[$i]) {
      print "Byte mismatch between calculated Va and Pkt Va!\n";
      return undef;
    }
  }

  # get the rest of the keys and encrypt the rest of the vtags
  for(my $i=$idx+1; $i<$pathlen; $i++) {
    my $realm_token = $pkt->{ICING3Packet::ICING_PATH()}->[$i];
    my @realm_id = @{$realm_token->{RealmToken::ID()}};
    my @my_id = @{$self->{ID()}};
    my @key = getSharedKey(\@my_id, \@realm_id);
    printf "Key Shared with $i=(%s)\n",join (":", map (sprintf ("%02x", $_), @key));

    my @temp = ICING3Crypt::prf2_96(\@my_id, \@realm_id, \@key, 
                                   [$idx >> 1, (0x80 & ($idx << 7)) | $hash[0], @hash[1..30]]);
    my @va = $vtags->[$i-1]->vabytes();
    my @new_bytes = map {$va[$_] ^ $temp[$_]} (0..11);
    $vtags->[$i-1]->setVaBytes(@new_bytes);
    printf "Re-encrypted Va$i=(%s)\n",join (":", map (sprintf ("%02x", $_), @new_bytes));
  }

  # increment index
  $pkt->{ICING3Packet::ICING_PATH_IDX()} = $pkt->{ICING3Packet::ICING_PATH_IDX()} + 1;

  return $pkt;
}

sub getMySharedKey {
  my $self = shift;
  my $other_realm = shift;

  return getSharedKey($self->{ID()}, $other_realm->{ID()});
}

sub getSharedKey {
  my $my_id = shift;
  my $other_id = shift;

  my @key=();
  for(my $j=0; $j<16; $j++) {
    push @key, $other_id->[$j] ^ $my_id->[$j];
  }

  return @key;
}

package CTokenServer;
use strict;
use warnings;
use POSIX;
use List::Util qw(shuffle min);
use Carp;

use constant NUM_REALMS => 'num_realms';
use constant NUM_REALMS_START => 'num_realms_start';
use constant FULL_VNODE => 'full_vnode';
use constant FULL_UHF => 'full_uhf';

my $full_vnode = 0;
my $full_uhf = 0;

sub new {
  my ($class, %args) = @_;

  my $num = $args{NUM_REALMS()};
  confess "Need num realms" unless defined $num;
  confess "Num realms should be less than 255" unless $num<255;

  $full_vnode = $args{FULL_VNODE()} if defined $args{FULL_VNODE()};

  $full_uhf = $args{FULL_UHF()} if defined $args{FULL_UHF()};

  my $self = {};

  my $start = 1;
  $start = $args{NUM_REALMS_START()} if defined $args{NUM_REALMS_START()};

  my $finish = $start+$num-1;

  my $max_b = $full_uhf ? 256 : 8;

  my @id = map(int(rand($max_b)), (1..20));
  foreach ($start..$finish){
    # add special cases to test key lookups
    my $id_sel;
    if($_ == $start + 4) {
      $id_sel = [@id[0..15], map(int(rand($max_b)), (1..4))];
    } elsif ($_ == $start + 5) {
      $id_sel = [map(int(rand($max_b)), (1..4)), @id[4..19]];
    } else {
      $id_sel = [$_ & ($max_b - 1), $_ & ($max_b - 1),
                 map(int(rand($max_b)), (1..18))];
    }
    my $realm = Realm->new(Realm::ID()  => $id_sel,
                          );
    print("Adding new realm: @{$id_sel}\n");
    $realm->prettyPrint();
    addRealm($self, $realm);
  }

  bless $self, $class;
}

sub addRealm {
  my ($self, $realm) = @_;
  my $id = genRealmID($realm);
  $self->{$id} = $realm;
}

sub getCTokens {
  my $self = shift;
  my $src_token = shift;
  my @realmtokens = @_;

  my @ctokens = ();
  foreach (@realmtokens) {
    my $id = pack "C*", @{$_->{Realm::ID()}};
    my $realm = $self->{$id};
    my $ctoken = [$realm->createCToken($src_token, @realmtokens)];
    push @ctokens, $ctoken;
  }
  return @ctokens;
}

# creates a random path through a random number
# of realms with the given realm at the given idx 
# in the path
sub generateRandomPath {
  my ($self, $realm, $path_idx, $path_len) = @_;
  $self->addRealm($realm);
  my $id = genRealmID($realm);

  # remove the id of the given realm
  my @ids = grep({$_ ne $id} $self->getRealmsIDs());

  # make the list large enough to accomodate requests
  while(($path_idx > scalar @ids - 1)
       || ($path_len > scalar @ids +1)) {
#    confess "Path idx larger than number of available realms";
    push @ids, @ids;
  }

  # randomize list
  @ids = shuffle @ids;

  # select a path len
  $path_len = $path_idx + 2 + int(rand(min((scalar @ids), 50) - $path_idx - 1)) unless defined $path_len;

  # select elements
  my @path = (@ids[0..$path_idx-1], $id, @ids[$path_idx..$path_len-2]);
  print "RandomPath: \n";
  foreach my $id (@path) {
    printf "    (%s)\n", join(":", map(sprintf("%02x", $_), @{$self->{$id}->{Realm::ID()}}));
  }

  # create realm tokens for all
  if($full_vnode) {
    return map(RealmToken->new(RealmToken::ID() => $self->{$_}->{Realm::ID()}), @path);
  } else {
    return map(RealmToken->new(RealmToken::ID() => $self->{$_}->{Realm::ID()},
                               RealmToken::VOQ() => [0,int(rand(16)),0,0]),
               @path);
  }
}

# input: instance of RealmToken
# returns: the Realm corresponding to the RealmToken
sub getRealm {
  my $self = shift;
  my $realmToken = shift;

  confess "Realm ID ref is undefined." unless defined $realmToken->{RealmToken::ID()};
  my $id = pack "C*", @{$realmToken->{RealmToken::ID()}};
  return $self->{$id};
}

# input: instance of Realm
# returns: id used to lookup in hash ($self)
sub genRealmID {
  my $realm = shift;
  return pack "C*", @{$realm->{Realm::ID()}};
}

sub getRealmsIDs {
  return keys %{shift()};
}

sub getOtherRealms {
  my $self = shift;
  my $not_realm = shift;
  my $not_id = genRealmID($not_realm);

  my @realms = ();
  foreach (keys %{$self}){
    if($_ ne $not_id) {
      push @realms, $self->{$_};
    }
  }
  return @realms;
}

1;
