/********************************************************************
 * File: sim_tool.c
 * Author: Jad Naous
 * Description: Provides glue between Perl and C
 ********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "../../C/nf_icing3.h"
#include <netfpga/nf2util.h>

#define DEBUG

static uint8_t *string2bytes(char *str, uint8_t *buf) {
  uint8_t *buf_ptr;
  int len = strlen(str);

  if(buf == NULL) {
    buf = (uint8_t *) malloc(len/2);
  }

  if(!buf) {
    printf("ERROR:_Out_of_memory ");
    return NULL;
  }

  buf_ptr = buf;
  while(*str){
    sscanf(str, "%2hhx", buf_ptr);
    str+=2; buf_ptr++;
  }

  return buf;
}

static void print_array(uint8_t *array, int len) {
  printf("%02x", *array);
  array++; len--;
  while(len--) {
    printf(":%02x", *array);
    array++;
  }
  printf("\n");
}

static void parse_realm_info(nf_icing_realm_info_t *realm_info, 
                             char *argv[], 
                             nf_icing_dev_t *dev) {
  string2bytes(argv[3], (uint8_t *)&realm_info->realm_id);
  string2bytes(argv[4], realm_info->shared_key);
  realm_info->my_id_is_less = memcmp(&dev->realm_id,
                                     &realm_info->realm_id,
                                     sizeof(realm_id_t)) < 0;
  string2bytes(argv[5], &realm_info->output_mask);
  string2bytes(argv[6], realm_info->next_hop_mac);
#ifdef DEBUG
  printf("realm_id:\n");
  print_array((uint8_t *)&realm_info->realm_id, 20);
  printf("skey:\n");
  print_array(realm_info->shared_key, 16);
  printf("out_mask: %02x\n", realm_info->output_mask);
  printf("nxthop:\n");
  print_array(realm_info->next_hop_mac, 6);
#endif
}

static void printExpWords(uint8_t *list, int len){
  int i;
  if(len % 4) {
    printf("ERROR:_expected_byte_array_len_not_multiple_of_4 ");
  }
  for(i=0; len > 0; i+=4, len -= 4) {
    printf("E:%02x%02x%02x%02x ", list[i], list[i+1], list[i+2], list[i+3]);
  }
}

static void parse_state(nf_icing_dev_t *dev, char *state_str){
  string2bytes(state_str, (uint8_t *)dev);
}

static void print_state(nf_icing_dev_t *dev) {
  uint8_t *dev_bytes = (uint8_t *) dev;
  int i;
  printf("S:");
  for(i=0; i<sizeof(nf_icing_dev_t); i++) {
    printf("%02x", dev_bytes[i]);
  }
  printf(" ");
}

int main(int argc, char *argv[]) {
  nf_icing_dev_t dev;
  nf_icing_dev_config_t config;
  nf_icing_realm_info_t realm_info;
  nf_icing_realm_info_t realm_info_tmp;
  poc_key_t poc_key;
  vnode_t vnode;
  int full_uhf, write_uhf;

  int val = 3;

  if(argc == 1) {
    printHello(&dev.nf2, &val);
    return 0;
  }

  /* check the first argument */
  if(strcmp(argv[1], "init") == 0) {
    config.dev_name = "sim_dev";
    /* next arg is mac addresses */
    string2bytes(argv[2], config.macs);
    /* next is realm id */
    string2bytes(argv[3], (uint8_t *)&config.realm_id);
    /* next is full_uhf */
    if(*argv[4] == '0') {
      full_uhf = 0;
    } else {
      full_uhf = 1;
    }
    /* next is write_uhf */
    if(*argv[5] == '0') {
      write_uhf = 0;
    } else {
      write_uhf = 1;
    }
#ifdef DEBUG
    printf("Doing INIT\n");
    printf("MACs:\n");
    print_array(config.macs, 24);
    printf("realm_id:\n");
    print_array((uint8_t *)&config.realm_id, 20);
#endif

    /* init */
    nf_icing3_init_dev(&dev, &config, 1, full_uhf, write_uhf);
  }

  else {
    /* The second argument is the device state */
    parse_state(&dev, argv[2]);
    
    /* parse the command now */
    if(strcmp(argv[1], "write_realm") == 0) {
#ifdef DEBUG
      printf("Doing WRITE_REALM\n");
#endif
      parse_realm_info(&realm_info, argv, &dev);
      nf_icing3_write_realm(&realm_info, &dev);
    }
    
    else if(strcmp(argv[1], "read_realm") == 0) {
      printf("Doing READ_REALM\n");
      parse_realm_info(&realm_info, argv, &dev);
      memcpy(&realm_info_tmp, &realm_info, sizeof realm_info);
      nf_icing3_read_realm(&realm_info_tmp, &dev);
      printExpWords((uint8_t *)&realm_info, sizeof realm_info);
    }

    else if(strcmp(argv[1], "delete_realm") == 0) {
      string2bytes(argv[3], (uint8_t *)&realm_info.realm_id);
      nf_icing3_delete_realm(&realm_info, &dev);
    }

    else if(strcmp(argv[1], "write_poc_key") == 0) {
      printf("Doing WRITE_POC_KEY\n");
      string2bytes(argv[3], (uint8_t *)&vnode);
      string2bytes(argv[4], (uint8_t *)&poc_key);
      nf_icing3_write_poc_key(&vnode, &poc_key, &dev);
    }

    else if(strcmp(argv[1], "read_poc_key") == 0) {
      string2bytes(argv[3], (uint8_t *)&vnode);
      nf_icing3_read_poc_key(&vnode, &poc_key, &dev);
      string2bytes(argv[4], (uint8_t *)&poc_key);
      printExpWords((uint8_t *)&poc_key, sizeof poc_key);
    }

    else if(strcmp(argv[1], "delete_poc_key") == 0) {
      string2bytes(argv[3], (uint8_t *)&vnode);
      nf_icing3_delete_poc_key(&vnode, &dev);
    }

    else {
      fprintf(stderr, "Unknown Command %s ", argv[1]);
      printf("ERROR:_Unknown_Command_%s ", argv[1]);
    }
  }

  print_state(&dev);
  return 0;
}
