/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: calculate_nxt_vtags.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *         Va_i[j] = Va_{i-1}[j] XOR PRF2-96(R_i, R_j, k_{i,j}, i[7:0] || H)
 *
 ******************************************************************************/

module calculate_nxt_vtags
  (// --- Interface to parse_pkt
   input      [7:0]   path_idx,
   input              path_idx_vld,
   
   // --- Interface to hash_pkt
   input      [0:246] hash,
   input              hash_vld,

   // --- Interface to key lookup
   input      [0:127] shared_dh_key,
   input              my_realm_id_is_smaller,
   input              shared_dh_key_vld,
   input              shared_dh_key_last,
   input              shared_dh_key_b4_me,
   input              rd_shared_key_chk_vtag,
   output reg         rd_shared_key,

   // --- Interface to shared 2 block mac core
   output     [0:127] mac2_key,
   output     [0:255] mac2_block,
   output reg         mac2_block_vld,
   input              mac2_block_ack,
   input      [0:95]  mac2_tag,
   input              mac2_tag_vld,

   // --- Interface to results
   output     [0:95]  prf_result,
   output             prf_result_vld,
   output             prf_result_last,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ localparams ----------------------------
   localparam WAIT_FOR_B4_ME_KEY   = 0,
              WAIT_FOR_KEY         = 1;
   
   //------------------------ Wires/Regs -----------------------------
   wire [5:0]         count_dout;
   wire [0:246]       hash_dout;
   wire [7:0]         path_idx_dout;

   reg                state, state_nxt;
   reg [5:0]          count, count_nxt;
   reg                hash_fifo_rd_en;

   reg [5:0]          check_count;
   
   //------------------------- Modules -------------------------------
   fallthrough_small_fifo 
     #(.WIDTH(6),
       .MAX_DEPTH_BITS(1))
      count_fifo
        (.din           (count[5:0]),
         .wr_en         (hash_fifo_rd_en),
         .rd_en         (count_rd_en),
         .dout          (count_dout),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (count_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(8),
       .MAX_DEPTH_BITS(3))
      path_idx_fifo
        (.din           (path_idx),
         .wr_en         (path_idx_vld),
         .rd_en         (hash_fifo_rd_en),
         .dout          (path_idx_dout),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (path_idx_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(247),
       .MAX_DEPTH_BITS(1))
      hash_fifo
        (.din           (hash),
         .wr_en         (hash_vld),
         .rd_en         (hash_fifo_rd_en),
         .dout          (hash_dout),
         .full          (hash_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (hash_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );
      
   //-------------------------- Logic --------------------------------
   assign mac2_key = shared_dh_key;
   assign mac2_block = {my_realm_id_is_smaller, path_idx_dout, hash_dout};
   
   /******************************************************************
    * As the keys arrive, call the mac function for second set of keys
    ******************************************************************/
   always @(*) begin
      mac2_block_vld   = 0;
      state_nxt        = state;
      count_nxt        = count;
      hash_fifo_rd_en  = 0;
      rd_shared_key    = 0;
      case(state)
         /* Wait for the second set of keys to arrive */
         WAIT_FOR_B4_ME_KEY: begin
            count_nxt = 0;
            if(rd_shared_key_chk_vtag && shared_dh_key_b4_me) begin
               /* if there are no more hops then stay here */
               if(shared_dh_key_last) begin
		  /* This will only happen if path idx == path_len-1
		   * i.e. the fwdr is the dst */
                  hash_fifo_rd_en = 1;
               end
               /* otherwise go do the macs */
               else begin
                  state_nxt = WAIT_FOR_KEY;
               end
            end
         end

         /* Wait for the key, then mac the hash. If we are done
          * then repeat */
         WAIT_FOR_KEY: begin
            if(shared_dh_key_vld && !hash_fifo_empty) begin
               mac2_block_vld = 1;
               /* input accepted into mac so move on */
               if(mac2_block_ack) begin
                  count_nxt      = count + 1'b1;
                  rd_shared_key  = 1;
                  if(shared_dh_key_last) begin
                     hash_fifo_rd_en   = 1;
                     state_nxt         = WAIT_FOR_B4_ME_KEY;
                  end
               end
            end // if (shared_dh_key_vld && !hash_fifo_empty)
         end // case: WAIT_FOR_KEY
      endcase // case(state)
   end // always @ (*)
   
   always @(posedge clk) begin
      if(reset) begin
         state <= WAIT_FOR_B4_ME_KEY;
         count <= 0;
      end else begin
         state <= state_nxt;
         count <= count_nxt;
      end
   end

   /******************************************************************
    * Output the results as they come in. Keep track of 
    * which is the last.
    ******************************************************************/
   assign prf_result      = mac2_tag[0:95];
   assign prf_result_vld  = mac2_tag_vld;
   assign prf_result_last = !count_fifo_empty && (check_count == count_dout);
   assign count_rd_en     = prf_result_vld && prf_result_last;

   always @(posedge clk) begin
      if(reset) begin
         check_count <= 0;
      end else begin
         if(mac2_tag_vld) begin
            if(prf_result_last) begin
               check_count <= 0;
            end else begin
               check_count <= check_count + 1'b1;
            end
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

   // synthesis translate_off
   integer num_nxt_vtags = 0;
   integer num_last_vtags = 0;
   reg 	   prf_result_vld_d1 = 0;
   always @(posedge clk) begin
      prf_result_vld_d1 <= prf_result_vld;
      if(prf_result_vld === 1) begin
         num_nxt_vtags <= num_nxt_vtags + 1;
         num_last_vtags <= num_last_vtags + prf_result_last;
      end
      if(prf_result_vld_d1 === 1) begin
         $display($time, " %m Num nxt V's/last V's calc'ed  :  %u/%u", num_nxt_vtags, num_last_vtags);
      end
   end
   // synthesis translate_on

endmodule // calculate_nxt_vtags

