/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: cbc_mac2_128.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *
 *    CBC-MAC2-128(key[0..127], block[0..255]) {
 *        X   <- AES-ENC(key, block[0..127])
 *        Tag <- AES-ENC(key, X XOR block[128..255])
 *        output Tag
 *    }
 *******************************************************************************/

module cbc_mac2_128
  (input [0:127]  key,
   input [0:255]  block,
   input          block_vld,
   output         block_ack,
   
   output [0:127] tag,
   output         tag_vld,

   // --- Interface to shared AES core
   output     [0:127] aes_enc_plain_0,
   output     [0:127] aes_enc_key_0,
   output             aes_enc_plain_vld_0,
   input              aes_enc_plain_ack_0,
   input      [0:127] aes_enc_cipher_0,
   input              aes_enc_cipher_vld_0,

   // --- Interface to shared AES core
   output     [0:127] aes_enc_plain_1,
   output     [0:127] aes_enc_key_1,
   output             aes_enc_plain_vld_1,
   input              aes_enc_plain_ack_1,
   input      [0:127] aes_enc_cipher_1,
   input              aes_enc_cipher_vld_1,

   input              clk,
   input              reset
   );

   wire [0:127]       block_dout, x, key_dout;

   fallthrough_small_fifo 
     #(.WIDTH(128),
       .MAX_DEPTH_BITS(4))
      block_fifo
        (.din           (block[128:255]),
         .wr_en         (block_ack),
         .rd_en         (aes_enc_plain_ack_1),
         .dout          (block_dout),
         .full          (full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(128),
       .MAX_DEPTH_BITS(3))
      X_fifo
        (.din           (aes_enc_cipher_0),
         .wr_en         (aes_enc_cipher_vld_0),
         .rd_en         (aes_enc_plain_ack_1),
         .dout          (x),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (x_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(128),
       .MAX_DEPTH_BITS(4))
      key_fifo
        (.din           (key),
         .wr_en         (block_ack),
         .rd_en         (aes_enc_plain_ack_1),
         .dout          (key_dout),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (),
         .reset         (reset),
         .clk           (clk)
         );

   assign aes_enc_plain_0       = block[0:127];
   assign aes_enc_key_0         = key;
   assign aes_enc_plain_vld_0   = block_vld & !full;
   assign block_ack             = aes_enc_plain_ack_0;

   assign aes_enc_plain_1       = block_dout ^ x;
   assign aes_enc_key_1         = key_dout;
   assign aes_enc_plain_vld_1   = !x_empty;
   assign tag                   = aes_enc_cipher_1;
   assign tag_vld               = aes_enc_cipher_vld_1;
   
endmodule

/******************************
    cbc_mac2_128 cbc_mac2_128
     (.key                    (key),
      .block                  (block),
      .block_vld              (block_vld),
      .block_ack              (block_ack),
      .tag                    (tag),
      .tag_vld                (tag_vld),
      .aes_enc_rdy_0          (aes_enc_rdy_0),
      .aes_enc_cipher_0       (aes_enc_cipher_0),
      .aes_enc_cipher_vld_0   (aes_enc_cipher_vld_0),
      .aes_enc_plain_0        (aes_enc_plain_0),
      .aes_enc_key_0          (aes_enc_key_0),
      .aes_enc_plain_vld_0    (aes_enc_plain_vld_0),
      .aes_enc_rdy_1          (aes_enc_rdy_1),
      .aes_enc_cipher_1       (aes_enc_cipher_1),
      .aes_enc_cipher_vld_1   (aes_enc_cipher_vld_1),
      .aes_enc_plain_1        (aes_enc_plain_1),
      .aes_enc_key_1          (aes_enc_key_1),
      .aes_enc_plain_vld_1    (aes_enc_plain_vld_1),
      .clk                    (clk),
      .reset                  (reset));
********************************/