/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: check_hdr.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 * Checks that the dst mac address, the lengths, path_idx, version, ... are
 * all valid.
 ******************************************************************************/

module check_hdr
  (// --- Interface to parser
   input [3:0]        input_port,
   input              input_port_vld,
   input [10:0]       byte_length,
   input              byte_length_vld,
   input [0:47]       dst_mac,
   input              dst_mac_vld,
   input [0:15]       eth_type,
   input              eth_type_vld,
   input [7:0]        icing_version,
   input              icing_version_vld,
   input [7:0]        icing_path_len,
   input              icing_path_len_vld,
   input [7:0]        icing_path_idx,
   input              icing_path_idx_vld,

   // --- Interface to registers
   input  [0:48*4-1]  port_macs,

   // --- Interface to results
   output reg         dst_mac_not_ok,
   output reg         ethertype_not_ok,
   output reg         path_idx_not_ok,
   output reg         path_len_not_ok,
   output reg         pkt_len_not_ok,
   output reg         version_not_ok,
   output             pkt_not_ok,
   output reg         hdr_check_done,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ Wires/Regs -----------------------------
   wire [10:0]        hdr_len;
   wire [0:47]        port_mac[0:3];
   reg [2:0]          input_port_reg;
   reg [10:0]         len_reg;
   reg [5:0]          path_len_reg;

   reg                pkt_not_ok_reg;
   //-------------------------- Logic --------------------------------
   assign pkt_not_ok_nxt = (dst_mac_not_ok || ethertype_not_ok || path_idx_not_ok
                            || path_len_not_ok || pkt_len_not_ok || version_not_ok);
   assign pkt_not_ok = pkt_not_ok_reg | path_idx_not_ok;
          
   // overhead = 13 + path_len * 40 bytes - 16 bytes
   assign hdr_len =   4'd13 
                    + ({icing_path_len[5:0] , 5'h0} + {icing_path_len[5:0], 3'h0}) 
                    - 5'd16;

   generate
      genvar i;
      /* split up port macs */
      for(i=0; i<4; i=i+1) begin:gen_mac_words
         assign port_mac[i] = port_macs[i*48 : i*48 + 47];
      end
   endgenerate
   
   always @(posedge clk) begin
      dst_mac_not_ok      <= 0;
      ethertype_not_ok    <= 0;
      path_idx_not_ok     <= 0;
      path_len_not_ok     <= 0;
      pkt_len_not_ok      <= 0;
      version_not_ok      <= 0;
      hdr_check_done      <= 0;

      if(input_port_vld) input_port_reg <= input_port[2:0];

      if(byte_length_vld) len_reg <= byte_length - 4'd14; // remove eth hdr

      if(dst_mac_vld 
         && dst_mac != port_mac[input_port_reg[2:1]] // check the addr
         && !input_port_reg[0]) begin                // don't care if it's from CPU
         dst_mac_not_ok <= 1;
      end

      if(eth_type_vld && eth_type != `ICING_ETHTYPE) begin
         ethertype_not_ok <= 1;
      end

      if(icing_version_vld && icing_version != `ICING_CURR_VER) begin
         version_not_ok <= 1;
      end

      if(icing_path_len_vld) begin
         path_len_reg <= icing_path_len[5:0] - 1'b1;
         if(icing_path_len > `ICING_MAX_PATH_LEN 
            || icing_path_len < 2) begin
            
            path_len_not_ok <= 1;
         end

         if(len_reg < hdr_len) pkt_len_not_ok <= 1;
      end

      if(icing_path_idx_vld 
         && (icing_path_idx == 0 
             || icing_path_idx >= {2'h0, path_len_reg})) begin
         path_idx_not_ok <= 1;
      end

      if(icing_path_idx_vld) hdr_check_done <= 1;

      pkt_not_ok_reg <= (reset || hdr_check_done) ? 0 : pkt_not_ok | pkt_not_ok_nxt;
      
   end // always @ (posedge clk)

   // synthesis translate_off
   integer num_checks = 0;
   always @(posedge clk) begin
      if(hdr_check_done === 1) begin
         num_checks = num_checks + 1;
         $display($time, " %m Num Hdr checks:  %u", num_checks);
      end
   end
   // synthesis translate_on
            
endmodule // check_hdr

