/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: check_vtag.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Checks my Va and Vb in pkt are correct:
 * 
 * calculates Va_{i-1}[i] as:
 *       for 0 <= j < i:
 *           temp      = PRF2-96(R_j, R_i, k_{j,i}, j[7:0] || H)
 *           Va_{j}[i] = Va_{j-1}[i] XOR temp
 ******************************************************************************/

module check_vtag
  (// --- Interface to parser
   input      [0:63]  icing_va,
   input              icing_va_0_vld,
   input              icing_va_1_vld,
   input      [0:31]  icing_vb,
   input              icing_vb_vld,
   input              icing_vtag_is_me,
   input              icing_vtag_is_last,

   // --- Interface to hash_pkt
   input      [0:246] hash,
   input              hash_vld,

   // --- Interface to key lookup
   input      [0:127] shared_dh_key,
   input              my_realm_id_is_smaller,
   input              shared_dh_key_vld,
   input              shared_dh_key_last,
   input              shared_dh_key_b4_me,
   output             rd_shared_key,

   // --- Interface to calculate_A
   input      [0:95]  A,
   input              A_vld,
   
   // --- Interface to shared 2 block mac core
   output     [0:127] mac2_key,
   output     [0:255] mac2_block,
   output reg         mac2_block_vld,
   input              mac2_block_ack,
   input      [0:95]  mac2_tag,
   input              mac2_tag_vld,

   // --- Interface to results
   output reg         vtag_check_done,
   output reg         va_invld,
   output reg         vb_invld,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ localparams ----------------------------
   localparam WAIT_FOR_KEY             = 0,
              WAIT_FOR_LAST_KEY        = 1;

   localparam WAIT_FOR_FIRST_MAC_TAG   = 0,
              AGGR_MACS                = 1;
   //------------------------ Wires/Regs -----------------------------
   wire [0:95]        va_dout;
   wire [0:31]        vb_dout;
   wire [5:0]         count_dout;
   wire [0:95]        A_dout;
   wire [0:246]       hash_dout;

   reg                state, state_nxt;
   reg [5:0]          count, count_nxt;
   reg                hash_fifo_rd_en;
   reg                vtemp_vld;

   reg [0:95]         va_temp_nxt, va_temp;
   reg                check_state_nxt, check_state;
   reg [5:0]          check_count_nxt, check_count;
   reg                a_fifo_rd_en;
   reg                count_rd_en;

   wire [0:95]        xored_vatemp, vtemp_dout, calc_va;
   reg                vtemp_fifo_rd_en;
   reg                va_fifo_rd_en;
   
   //------------------------- Modules -------------------------------
   fallthrough_small_fifo 
     #(.WIDTH(6),
       .MAX_DEPTH_BITS(1))
      count_fifo
        (.din           (count[5:0] ),
         .wr_en         (hash_fifo_rd_en),
         .rd_en         (count_rd_en),
         .dout          (count_dout),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (count_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(96),
       .MAX_DEPTH_BITS(1))
      A_fifo
        (.din           (A),
         .wr_en         (A_vld),
         .rd_en         (a_fifo_rd_en),
         .dout          (A_dout),
         .full          (a_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (a_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(96),
       .MAX_DEPTH_BITS(1))
      vtemp_fifo
        (.din           (xored_vatemp),
         .wr_en         (vtemp_vld),
         .rd_en         (vtemp_fifo_rd_en),
         .dout          (vtemp_dout),
         .full          (vtemp_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (vtemp_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(64),
       .MAX_DEPTH_BITS(3))
      va_fifo_0
        (.din           (icing_va),
         .wr_en         (icing_va_0_vld && icing_vtag_is_me),
         .rd_en         (va_fifo_rd_en),
         .dout          (va_dout[0:63]),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(64),
       .MAX_DEPTH_BITS(3))
      va_fifo_1
        (.din           ({icing_va[0:31], icing_vb}),
         .wr_en         (icing_va_1_vld && icing_vtag_is_me),
         .rd_en         (va_fifo_rd_en),
         .dout          ({va_dout[64:95], vb_dout}),
         .full          (va_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (va_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(247),
       .MAX_DEPTH_BITS(1))
      hash_fifo
        (.din           (hash),
         .wr_en         (hash_vld),
         .rd_en         (hash_fifo_rd_en),
         .dout          (hash_dout),
         .full          (hash_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (hash_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );
      
   //-------------------------- Logic --------------------------------
   assign rd_shared_key = mac2_block_ack;
   assign mac2_key = shared_dh_key;
   assign mac2_block = {~my_realm_id_is_smaller, 2'h0, count, hash_dout};
   
   /******************************************************************
    * As the keys arrive, call the mac function for first set of keys
    ******************************************************************/
   always @(*) begin
      mac2_block_vld   = 0;
      state_nxt        = state;
      count_nxt        = count;
      hash_fifo_rd_en  = 0;
      case(state)
         WAIT_FOR_KEY: begin
            /* as the keys arrive, request mac */
            if(shared_dh_key_vld && !hash_fifo_empty && !shared_dh_key_last) begin
               mac2_block_vld = 1;
               if(mac2_block_ack) begin
                  count_nxt      = count + 1'b1;
                  /* if this is the last key shared with realms upstream */
                  if(shared_dh_key_b4_me) begin
                     hash_fifo_rd_en   = 1;
                     count_nxt         = 0;
                     /* If the last hop has been reached, then there
                      * are no more keys (no downstream hops). Otherwise,
                      * wait for the last key. */
                     if(!shared_dh_key_last) begin
                        state_nxt   = WAIT_FOR_LAST_KEY;
                     end
                  end // if (shared_dh_key_b4_me)
               end // if (mac2_block_ack)
            end // if (shared_dh_key_vld && !hash_fifo_empty)
         end // case: WAIT_FOR_KEY

         WAIT_FOR_LAST_KEY: begin
            if(shared_dh_key_vld && shared_dh_key_last) begin
               state_nxt = WAIT_FOR_KEY;
               count_nxt = 0;
            end
         end
      endcase // case(state)
   end // always @ (*)
   
   always @(posedge clk) begin
      if(reset) begin
         state <= WAIT_FOR_KEY;
         count <= 0;
      end else begin
         state <= state_nxt;
         count <= count_nxt;
      end
   end

   /******************************************************************
    * Aggregate the MACs together
    ******************************************************************/
   assign xored_vatemp = va_temp ^ mac2_tag[0:95];
   always @(*) begin
      va_temp_nxt       = va_temp;
      check_state_nxt   = check_state;
      check_count_nxt   = check_count;
      count_rd_en       = 0;
      vtemp_vld         = 0;
      case(check_state)
         WAIT_FOR_FIRST_MAC_TAG: begin
            if(mac2_tag_vld) begin
               va_temp_nxt       = mac2_tag[0:95];
               check_count_nxt   = 1;
               if(!count_fifo_empty && count_dout == 0) begin
                  count_rd_en = 1;
                  vtemp_vld   = 1;
                  va_temp_nxt = 0;
               end
               else begin
                  check_state_nxt   = AGGR_MACS;
               end
            end
         end

         AGGR_MACS: begin
            if(mac2_tag_vld) begin
               va_temp_nxt = xored_vatemp;
               check_count_nxt = check_count + 1'b1;
               // Aggregate as many as we sent to mac2
               if(!count_fifo_empty && check_count == count_dout) begin
                  check_state_nxt    = WAIT_FOR_FIRST_MAC_TAG;
                  count_rd_en        = 1;
                  vtemp_vld          = 1;
                  va_temp_nxt        = 0;
               end
            end
         end // case: AGGR_MACS
      endcase // case(check_state)
   end // always @ (*)

   always @(posedge clk) begin
      check_state    <= reset ? WAIT_FOR_FIRST_MAC_TAG : check_state_nxt;
      va_temp        <= reset ? 96'h0 : va_temp_nxt;
      check_count    <= check_count_nxt;
   end

   /**************************************************
    * Check the result when all are ready
    **************************************************/
   assign calc_va = vtemp_dout ^ A_dout;
   always @(*) begin
      if(!a_fifo_empty && !vtemp_fifo_empty) begin
         vtag_check_done    = 1;
         va_invld           = calc_va != va_dout;
         vb_invld           = A_dout[64:95] != vb_dout;
         a_fifo_rd_en       = 1;
         vtemp_fifo_rd_en   = 1;
         va_fifo_rd_en      = 1;
      end else begin
         a_fifo_rd_en       = 0;
         vtemp_fifo_rd_en   = 0;
         vtag_check_done    = 0;
         va_invld           = 0;
         vb_invld           = 0;
         va_fifo_rd_en      = 0;
      end
   end

   // synthesis translate_off
   integer num_checks = 0;
   integer num_va_miss = 0;
   integer num_vb_miss = 0;
   always @(posedge clk) begin
      if(vtag_check_done === 1) begin
         num_checks = num_checks + 1;
         num_va_miss = num_va_miss + va_invld;
         num_vb_miss = num_vb_miss + vb_invld;
         $display($time, " %m Num Vtag chk/va miss/vb miss:  %u/%u/%u", 
                  num_checks, num_va_miss, num_vb_miss);
      end
   end
   // synthesis translate_on

endmodule // check_vtag
