/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: chi_hash.v
 * Project: CHI
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Implements the CHI-256 hash function (SHA-3 candidate)
 * 
 * Assumes that input blocks are already padded.
 * 
 ******************************************************************************/
`timescale 1ns/1ps

module chi_hash
  (input  [0:511] msg_block,
   input          msg_block_vld,
   input          last_block,
   output reg     msg_block_ack,

   output [0:255] hash,
   output reg     hash_vld,

   input          clk,
   input          reset);

`include "chi.v.inc"

   //------------------ Internal Parameter ---------------------------
   localparam NUM_ROUNDS       = 20;
   localparam WAIT_FOR_START   = 0;
   localparam DO_ROUNDS        = 1;

   //---------------------- Wires/Regs -------------------------------
   wire [0 : 6*64-1] new_abcdef;
   reg [0 : 6*64-1]  abcdef, abcdef_nxt;
   wire [0:63]       W_2t, W_2t1;
   wire [0:63]       K_2t, K_2t1;
   reg [1:0]         tmod3, tmod3_nxt;
   reg [4:0]         round, round_nxt;
   reg               ld_msg_block;
   reg               next_W;
   wire [0 : 6*64-1] rotate_H0, H0, rotated_hash_reg, rotated_abcdef;
   wire [0 : 6*64-1] new_hash_reg;
   
   reg [0 : 6*64-1]  hash_reg, hash_reg_nxt;
   reg               state, state_nxt;
   reg               last_block_nxt, last_block_reg;
   reg               hash_vld_nxt;
   
   //----------------------- Modules ---------------------------------
   chi_round chi_round
     (// Outputs
      .new_abcdef                       (new_abcdef[0+:6*64] ),
      // Inputs
      .abcdef                           (abcdef[0+:6*64]),
      .W_2t                             (W_2t[0:63]),
      .W_2t1                            (W_2t1[0:63]),
      .K_2t                             (K_2t),
      .K_2t1                            (K_2t1),
      .tmod3                            (tmod3));

   chi_cst chi_cst
     (.kt    (K_2t),
      .kt1   (K_2t1),
      .t     ({1'b0, round_nxt}),
      .clk   (clk));

   chi_stepinput chi_stepinput
     (.msg_block       (msg_block),
      .ld_msg_block    (ld_msg_block),
      .next_W          (next_W),
      .W_2t            (W_2t),
      .W_2t1           (W_2t1),
      .clk             (clk),
      .reset           (reset)
      );

   generate
      genvar  i;
      for(i=0; i<6; i=i+1) begin:rotate
         chi_rotr64 chi_rotr64_0 (chi256_init(i), 1, rotate_H0[i*64 +: 64]);
         assign H0[i*64 +: 64] = chi256_init(i);

         chi_rotr64 chi_rotr64_1 (hash_reg[i*64 +: 64], 1, rotated_hash_reg[i*64 +: 64]); 

         chi_rotr64 chi_rotr64_2 (new_hash_reg[i*64 +: 64],
                                1, rotated_abcdef[i*64 +: 64]);
      end
   endgenerate
   
   //------------------------ Logic ----------------------------------
   
   assign hash = {hash_reg[0:127], hash_reg[64*3 +: 128]};
   assign new_hash_reg = rotated_hash_reg ^ new_abcdef;
   
   always @(*) begin
      msg_block_ack    = 0;
      ld_msg_block     = 0;
      round_nxt        = round;
      tmod3_nxt        = tmod3;
      next_W           = 0;
      hash_reg_nxt     = hash_reg;
      abcdef_nxt       = abcdef;
      state_nxt        = state;
      last_block_nxt   = last_block_reg;
      hash_vld_nxt     = 0;
      case(state)
        WAIT_FOR_START: begin
           if(msg_block_vld) begin
              msg_block_ack   = ~reset;
              ld_msg_block    = 1'b1;
              round_nxt       = 0;
              tmod3_nxt       = 0;
              hash_reg_nxt    = H0;
              if(last_block) begin
                 abcdef_nxt   = rotate_H0;
              end else begin
                 abcdef_nxt   = H0;
              end
              state_nxt       = DO_ROUNDS;
              last_block_nxt  = last_block;
           end // if (msg_block_vld)
        end // case: WAIT_FOR_START

        DO_ROUNDS: begin
           if(round == NUM_ROUNDS - 1'b1) begin
              /* if this is the last block then emit the hash the next
               * clock cycle and go to start. We can save a cycle here
               * by not going to start if we need to. */
              if(last_block_reg) begin
                 hash_vld_nxt  = 1'b1;
                 state_nxt     = WAIT_FOR_START;
                 /* latch the intermediate value */
                 hash_reg_nxt    = new_hash_reg;
              end
              /* If there are more blocks coming, then wait for a new
               * block to be available. */
              else begin
                 if(msg_block_vld) begin
                    /* initialize everything */
                    msg_block_ack   = 1'b1;
                    ld_msg_block    = 1'b1;
                    round_nxt       = 0;
                    tmod3_nxt       = 0;
                    last_block_nxt  = last_block;
                    /* latch the intermediate value */
                    hash_reg_nxt    = new_hash_reg;
                    
                    /* If this new block is the last, then we need to rotate
                     * the new hash_reg. */
                    if(last_block) begin
                       abcdef_nxt   = rotated_abcdef;
                    end
                    /* otherwise just use the new hash_reg */
                    else begin
                       abcdef_nxt   = new_hash_reg;
                    end
                 end // if (msg_block_vld)
              end // else: !if(last_block_reg)
           end // if (round == NUM_ROUNDS - 1'b1)
           else begin
              /* do a round */
              round_nxt     = round + 1'b1;
              tmod3_nxt     = tmod3 == 2 ? 0 : tmod3 + 1'b1;
              next_W        = 1'b1;
              abcdef_nxt    = new_abcdef;
           end // else: !if(round == NUM_ROUNDS - 1'b1)
        end // case: DO_ROUNDS
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      round             <= round_nxt;
      tmod3             <= tmod3_nxt;
      abcdef            <= abcdef_nxt;
      hash_reg          <= hash_reg_nxt;
      last_block_reg    <= last_block_nxt;
      if(reset) begin
         hash_vld    <= 0;
         state       <= WAIT_FOR_START;
      end
      else begin
         hash_vld    <= hash_vld_nxt;
         state       <= state_nxt;
      end
   end // always @ (posedge clk)

endmodule // chi_hash

         
// synthesis translate_off
/*****************************************************************************
 * Nonfunctional. Use chi_hash_wrapper_tester in chi_hash_wrapper.v instead.
 *****************************************************************************/
module chi_hash_tester ();

   reg [0:511] msg_block_test[0:2];
   reg [0:511] msg_block;
   reg         msg_block_vld;
   reg         last_block_test[0:2];
   reg         last_block;
   wire        msg_block_ack;
   reg [0:255] hash_test[0:1];
   wire [0:255] hash;
   wire         hash_vld;
   reg          clk = 0;
   reg          reset = 0;

   always #4 clk = ~clk;
   initial begin
      msg_block_test[0] = {24'b01100001_01100010_01100011, // "abc"
                           1'b1, 423'h0,                   // pad
                           64'd24};                        // length
      last_block_test[0] = 1;

      msg_block_test[1] = 512'h16E8B3D8F988E9BB04DE9C96F2627811C973CE4A5296B4772CA3EEFEB80A652BDF21F50DF79F32DB23F9F73D393B2D57D9A0297F7A2F2E79CFDA39FA393DF1AC;
      msg_block_test[2] = {8'h00, 440'h0, 64'd520};
      last_block_test[1] = 0;
      last_block_test[2] = 1;

      hash_test[0] = 256'hfe36843faadfa08a_dd876f951337614c_fbb2b64d986731d8_fc981607f374d45e;
      hash_test[1] = 256'h06EB685059D1C00603BFFE8C802D21CFA625557B8F759929913E3AD5AFD97894;
   end

   chi_hash chi_hash
     (// Outputs
      .msg_block_ack                    (msg_block_ack),
      .hash                             (hash[0:255]),
      .hash_vld                         (hash_vld),
      // Inputs
      .msg_block                        (msg_block[0:511]),
      .msg_block_vld                    (msg_block_vld),
      .last_block                       (last_block),
      .clk                              (clk),
      .reset                            (reset));

   integer      count = 0;
   integer      index = 0;
   always @(posedge clk) begin
      count <= count + 1'b1;
      if(msg_block_ack) begin
         index <= index + 1;
         if(index == 2) index <= 0;
      end
   end

   integer test_index = 0;
   always @(*) begin
      reset = 0;
      msg_block = msg_block_test[index];
      msg_block_vld = ~reset;
      last_block = last_block_test[index];
      if(count < 3) begin
         reset = 1;
      end else if (hash_vld) begin
         if(hash !== hash_test[test_index]) begin
            $display("%t %m ERROR: didn't get correct hash %u", $time, test_index);
         end else begin
            $display("%t SUCCESS %u!", $time, test_index);
         end
         test_index = test_index + 1;
         if(test_index === 2) $finish;
      end
   end // always @ (*)
endmodule // chi_hash_tester
// synthesis translate_on
