/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: chi_hash.v
 * Project: CHI
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Implements the CHI-256 hash function (SHA-3 candidate)
 * 
 * Assumes that input blocks are already padded. It is pipelined so that it can
 * calculate two hashes at the same time.
 ******************************************************************************/
//`timescale 1ns/1ps

module chi_hash_pipelined
  (input  [0:2*512-1] msg_block_in,
   input  [0:1]       msg_block_vld,
   input  [0:1]       last_block,
   output reg [0:1]   msg_block_ack,

   output [0:2*256-1] hash_out,
   output reg [0:1]   hash_vld,

   input              clk,
   input              reset);

`include "chi.v.inc"

   //------------------ Internal Parameter ---------------------------
   localparam NUM_ROUNDS       = 20;
   localparam WAIT_FOR_START   = 0;
   localparam DO_ROUNDS        = 1;

   //---------------------- Wires/Regs -------------------------------
   wire [0 : 6*64-1]  new_abcdef;
   reg [0 : 6*64-1]   abcdef[0:1];
   reg  [0 : 6*64-1]  abcdef_nxt[0:1];
   wire [0 : 2*6*64-1] abcdef_exp;
   wire [0:63]        W_2t[0:1];
   wire [0:63]        W_2t1[0:1];
   wire [0:63]        K_2t, K_2t1;
   reg [1:0]          tmod3[0:1];
   reg [1:0]          tmod3_nxt[0:1];
   reg [4:0]          round[0:1];
   reg [4:0]          round_nxt[0:1];
   wire [5*2-1:0]     round_exp;
   reg [0:1]          ld_msg_block;
   reg [0:1]          next_W;
   wire [0 : 6*64-1]  rotate_H0, H0, rotated_hash_reg, rotated_abcdef;
   wire [0 : 6*64-1]  new_hash_reg;
   
   reg [0 : 6*64-1]  hash_reg[0:1];
   wire [0 : 2*6*64-1] hash_reg_exp;
   reg [0 : 6*64-1]   hash_reg_nxt[0:1];
   reg [0:1]          state, state_nxt;
   reg [0:1]          last_block_reg, last_block_reg_nxt;
   reg [0:1]          hash_vld_nxt;
   
   wire [0:511]       msg_block[0:1];
   wire [0:255]       hash[0:1];
   reg                toggle, toggle_d1, toggle_d2;
   
   wire [4:0]         round_sel;
   wire [1:0]         tmod3_plus1;
   wire [3:0]         tmod3_exp;

   wire [0 : 6*64-1]  abcdef_sel;
   wire [0:63]        W_2t_sel, W_2t1_sel;
   wire [1:0]         tmod3_sel;
   
   //----------------------- Modules ---------------------------------
   assign             abcdef_sel = abcdef[~toggle];
   assign             W_2t_sel = W_2t[~toggle];
   assign             W_2t1_sel = W_2t1[~toggle];
   assign             tmod3_sel = tmod3[~toggle];
   
   chi_round_pipelined chi_round
     (// Outputs
      .new_abcdef                       (new_abcdef),
      // Inputs
      .abcdef                           (abcdef_sel),
      .W_2t                             (W_2t_sel),
      .W_2t1                            (W_2t1_sel),
      .K_2t                             (K_2t),
      .K_2t1                            (K_2t1),
      .tmod3                            (tmod3_sel),
      .clk                              (clk)
      );

   chi_cst chi_cst
     (.kt    (K_2t),
      .kt1   (K_2t1),
      .t     ({1'b0, round_nxt[toggle]}),
      .clk   (clk));

   chi_stepinput chi_stepinput_0
     (.msg_block       (msg_block[0]),
      .ld_msg_block    (ld_msg_block[0]),
      .next_W          (next_W[0]),
      .W_2t            (W_2t[0]),
      .W_2t1           (W_2t1[0]),
      .clk             (clk),
      .reset           (reset)
      );
   chi_stepinput chi_stepinput_1
     (.msg_block       (msg_block[1]),
      .ld_msg_block    (ld_msg_block[1]),
      .next_W          (next_W[1]),
      .W_2t            (W_2t[1]),
      .W_2t1           (W_2t1[1]),
      .clk             (clk),
      .reset           (reset)
      );

   generate
      genvar  i;
      for(i=0; i<6; i=i+1) begin:rotate
         wire [0:63] hash_reg_sel = hash_reg[toggle][i*64 +: 64];
         
         chi_rotr64 chi_rotr64_0 (chi256_init(i), 1, rotate_H0[i*64 +: 64]);
         assign H0[i*64 +: 64] = chi256_init(i);

         chi_rotr64 chi_rotr64_1 (hash_reg_sel, 1, 
                                  rotated_hash_reg[i*64 +: 64]); 

         chi_rotr64 chi_rotr64_2 (new_hash_reg[i*64 +: 64],
                                1, rotated_abcdef[i*64 +: 64]);
      end
   endgenerate
   
   //------------------------ Logic ----------------------------------

   generate
      for(i=0; i<2; i=i+1) begin: gen_wires
         assign hash[i] = {hash_reg[i][0:127], hash_reg[i][64*3 +: 128]};
         assign msg_block[i] = msg_block_in[i*512 : i*512 + 511];
         assign hash_out[256*i:256*i+255] = hash[i];
      end
   endgenerate

   assign new_hash_reg = rotated_hash_reg ^ new_abcdef;
   assign round_sel = round[toggle];   
   assign round_exp = {round[1], round[0]};
   assign tmod3_exp = {tmod3[1], tmod3[0]};
   assign tmod3_plus1 = tmod3[toggle] == 2 ? 0 : tmod3[toggle] + 1'b1;
   assign hash_reg_exp = {hash_reg[1], hash_reg[0]};
   assign abcdef_exp = {abcdef[1], abcdef[0]};
   assign state_sel = state[toggle];
   
   always @(*) begin
      msg_block_ack    = 0;
      ld_msg_block     = 0;
      {round_nxt[1], 
       round_nxt[0]}   = round_exp;
      next_W           = 0;
      {tmod3_nxt[1],
       tmod3_nxt[0]}       = tmod3_exp;
      {hash_reg_nxt[1],
       hash_reg_nxt[0]}    = hash_reg_exp;
      {abcdef_nxt[1],
       abcdef_nxt[0]}      = abcdef_exp;
      state_nxt            = state;
      last_block_reg_nxt   = last_block_reg;
      hash_vld_nxt         = 0;
      
      case(state_sel)
        WAIT_FOR_START: begin
           if(msg_block_vld[toggle]) begin
              msg_block_ack[toggle]   = ~reset;
              ld_msg_block[toggle]    = 1'b1;
              round_nxt[toggle]       = 0;
              tmod3_nxt[toggle]       = 0;
              hash_reg_nxt[toggle]    = H0;
              if(last_block[toggle]) begin
                 abcdef_nxt[toggle_d2]   = rotate_H0;
              end else begin
                 abcdef_nxt[toggle_d2]   = H0;
              end
              state_nxt[toggle]            = DO_ROUNDS;
              last_block_reg_nxt[toggle]   = last_block[toggle];
           end // if (msg_block_vld)
        end // case: WAIT_FOR_START

        DO_ROUNDS: begin
           if(round_sel == NUM_ROUNDS - 1'b1) begin
              /* if this is the last block then emit the hash the next
               * clock cycle and go to start. We can save a cycle here
               * by not going to start if we need to. */
              if(last_block_reg[toggle]) begin
                 hash_vld_nxt[toggle]  = 1'b1;
                 state_nxt[toggle]     = WAIT_FOR_START;
                 /* latch the intermediate value */
                 hash_reg_nxt[toggle]  = new_hash_reg;
              end
              /* If there are more blocks coming, then wait for a new
               * block to be available. */
              else begin
                 if(msg_block_vld[toggle]) begin
                    /* initialize everything */
                    msg_block_ack[toggle]        = 1'b1;
                    ld_msg_block[toggle]         = 1'b1;
                    round_nxt[toggle]            = 0;
                    tmod3_nxt[toggle]            = 0;
                    last_block_reg_nxt[toggle]   = last_block[toggle];
                    /* latch the intermediate value */
                    hash_reg_nxt[toggle]         = new_hash_reg;
                    
                    /* If this new block is the last, then we need to rotate
                     * the new hash_reg. */
                    if(last_block[toggle]) begin
                       abcdef_nxt[toggle_d2]   = rotated_abcdef;
                    end
                    /* otherwise just use the new hash_reg */
                    else begin
                       abcdef_nxt[toggle_d2]   = new_hash_reg;
                    end
                 end // if (msg_block_vld)
              end
           end // if (round == NUM_ROUNDS - 1'b1)
           else begin
              /* do a round */
              round_nxt[toggle]    = 1'b1 + round_sel;
              next_W[toggle]       = 1'b1;
              tmod3_nxt[toggle]    = tmod3_plus1;
              abcdef_nxt[toggle_d2]   = new_abcdef;
           end // else: !if(round == NUM_ROUNDS - 1'b1)
        end // case: DO_ROUNDS
      endcase // case(state)
   end // always @ (*)
            
   always @(posedge clk) begin
      {round[1], round[0]}          <= {round_nxt[1], round_nxt[0]};
      hash_vld                      <= 0;
      {tmod3[1], tmod3[0]}          <= {tmod3_nxt[1], tmod3_nxt[0]};
      {hash_reg[1], hash_reg[0]}    <= {hash_reg_nxt[1], hash_reg_nxt[0]};
      {abcdef[1], abcdef[0]}        <= {abcdef_nxt[1], abcdef_nxt[0]};
      last_block_reg                <= last_block_reg_nxt;
      if(reset) begin
         toggle    <= 0;
         state[0]  <= WAIT_FOR_START;
         state[1]  <= WAIT_FOR_START;
         hash_vld  <= 0;
      end
      else begin
         toggle      <= ~toggle;
	 toggle_d1   <= toggle;
	 toggle_d2   <= toggle_d1;
         state       <= state_nxt;
         hash_vld    <= hash_vld_nxt;
      end // else: !if(reset)
   end // always @ (posedge clk)
endmodule // chi_hash_pipelined
