/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: chi_hash_wrapper
 * Project: CHI
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Wrapper around the CHI hash to pad blocks and to read input as
 *              64-bit words.
 * 
 * Assumes messages are multiples of 8-bits in length (i.e. a byte stream).
 * 
 ******************************************************************************/
`timescale 1ns/1ps

module chi_hash_wrapper
  #(parameter MAX_LENGTH_BITS = 12) // bit length of the maximum msg size
  (input  [0:63]  in_data,
   input  [0:7]   in_ctrl,
   input          in_wr,
   output         in_rdy,

   output [0:255] hash,
   output         hash_vld,

   input          clk,
   input          reset);

   //-------------------- Internal Parameters ------------------------
   localparam     NUM_STATES           = 4;
   localparam     DO_BLOCKS            = 1,
                  ADD_LENGTH           = 2,
                  ADD_PAD_AND_LENGTH   = 4,
                  DO_LAST_BLOCK        = 8;
   
   //------------------------ Wires/Regs -----------------------------
   wire [MAX_LENGTH_BITS-1:0] msg_length_added;
   reg [MAX_LENGTH_BITS-1:0]  msg_length, msg_length_nxt;
   reg [0:511]                msg_block, msg_block_nxt;
   reg                        msg_block_vld;
   reg                        last_block;
   wire [3:0]                 len;
   wire [0:64]                padded_word;
   wire [0:64]                padded_words[0:7];
   reg [3:0]                  count, count_nxt;
   reg [0:NUM_STATES-1]       state, state_nxt;
   wire [0:7]                 cout;
   wire [0:63]                dout;
   wire [0:511]               last_blocks[0:7];
   wire [0:2]                 num_bytes_in_word;
   wire [0:511]               last_block_sel;
   reg                        fifo_rd_en;

   //------------------------- Modules ------------------------------
   fallthrough_small_fifo 
     #(.WIDTH(72),
       .MAX_DEPTH_BITS(2))
      in_fifo
        (.din           ({in_ctrl, in_data}),
         .wr_en         (in_wr),
         .rd_en         (fifo_rd_en),
         .dout          ({cout, dout}),
         .full          (),
         .prog_full     (),
         .nearly_full   (nearly_full),
         .empty         (fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   chi_hash chi_hash
     (// Outputs
      .msg_block_ack                    (msg_block_ack),
      .hash                             (hash[0:255]),
      .hash_vld                         (hash_vld),
      // Inputs
      .msg_block                        (msg_block[0:511]),
      .msg_block_vld                    (msg_block_vld),
      .last_block                       (last_block),
      .clk                              (clk),
      .reset                            (reset));

   //-------------------------- Logic --------------------------------
   assign in_rdy = !nearly_full;
   assign msg_length_added = msg_length + len;

   /* Get the number of bytes in a word */
   generate
      genvar i;
      for(i=0; i<7; i=i+1) begin: gen_pads
         assign padded_words[i] = {dout[0:8*i+7], 1'b1, {(56-i*8){1'b0}}};
      end
      assign padded_words[7] = {dout, 1'b1};
      
      assign last_blocks[0] = {padded_word,
                               {7*64-MAX_LENGTH_BITS-4{1'b0}},
                               msg_length_added, 3'h0};
      for(i=1; i<6; i=i+1) begin: gen_last_blocks
         assign last_blocks[i] = {msg_block[(8-i)*64:511],
                                  padded_word,
                                  {(7-i)*64-MAX_LENGTH_BITS-4{1'b0}},
                                  msg_length_added, 3'h0};
      end
      assign last_blocks[6] = {msg_block[2*64:511],
                               padded_word,
                               {1*64-MAX_LENGTH_BITS-4{1'b0}},
                               ( cout[7] 
                                 ? {MAX_LENGTH_BITS{1'b0}}
                                 : msg_length_added ),
                               3'h0
                                };
      assign last_blocks[7] = {msg_block[64:511],
                               padded_word[0:63]};
   endgenerate
   priority_encoder #(.OUTPUT_WIDTH(3)) penc(cout, num_bytes_in_word, num_bytes_vld);

   assign padded_word = num_bytes_vld ? padded_words[num_bytes_in_word] : dout;
   assign len = num_bytes_vld ? num_bytes_in_word + 4'h1 : 4'h8;
   assign last_block_sel = last_blocks[count[2:0]];

   always @(*) begin
      count_nxt        = count;
      msg_length_nxt   = msg_length;
      msg_block_nxt    = msg_block;
      msg_block_vld    = 0;
      state_nxt        = state;
      last_block       = 0;
      fifo_rd_en       = 0;
      
      case(state)
         DO_BLOCKS: begin
            /* if buffer full then send out */
            if(count == 8) begin
               msg_block_vld   = 1;
	       if(msg_block_ack) begin
		  count_nxt = 0;
	       end
            end

            /* Fill up the buffer */
            else if(!fifo_empty) begin
               fifo_rd_en = 1;
               count_nxt = (count == 8) ? 4'h1 : count + 1'b1;

               /* increment msg byte length */
               msg_length_nxt = msg_length_added;
               
               /* if not the last block */
               if(cout == 0) begin
                  /* then shift the msg block and add the new word */
                  msg_block_nxt  = {msg_block[64:511], dout};
               end
               /* Else this is the last block */
               else begin
                  msg_block_nxt = last_block_sel;
                  /* If no space for pad or for length */
                  if(count == 7 && cout[7]) begin
                     state_nxt = ADD_PAD_AND_LENGTH;
                  end
                  /* If no space for length only */
                  else if((count == 7 && !cout[7]) 
                          || (count == 6 && cout[7] == 1)) begin
                     state_nxt = ADD_LENGTH;
                  end
                  /* Otherwise, pad and length fit */
                  else begin
                     state_nxt = DO_LAST_BLOCK;
                  end
               end // else: !if(cout == 0)
            end // if (!fifo_empty)
         end // case: DO_BLOCKS

         /* Do an additional block with the padding and length */
         ADD_PAD_AND_LENGTH: begin
            msg_block_vld   = 1'b1;
            if(msg_block_ack) begin
               msg_block_nxt   = {1'b1, {511-MAX_LENGTH_BITS-3{1'b0}}, msg_length, 3'h0};
               state_nxt       = DO_LAST_BLOCK;
            end
         end

         /* Do an additional block with length only */
         ADD_LENGTH: begin
            msg_block_vld   = 1'b1;
            if(msg_block_ack) begin
               msg_block_nxt   = {{512-MAX_LENGTH_BITS-3{1'b0}}, msg_length, 3'h0};
               state_nxt       = DO_LAST_BLOCK;
            end
         end

         /* Do the last block */
         DO_LAST_BLOCK: begin
            last_block      = 1'b1;
            msg_block_vld   = 1'b1;
            count_nxt       = 0;
            msg_length_nxt  = 0;
            if(msg_block_ack) begin
               state_nxt = DO_BLOCKS;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      msg_block     <= msg_block_nxt;
      if(reset) begin
         msg_length    <= 0;
         state         <= DO_BLOCKS;
         count         <= 0;
      end else begin
         msg_length    <= msg_length_nxt;
         state         <= state_nxt;
         count         <= count_nxt;
      end
   end // always @ (posedge clk)

   // synthesis translate_off
   always @(posedge clk) begin
      if(msg_block_ack) begin
         $display($time, " %m Inserting block");
         $display($time, " %m                  %x", msg_block[128*0:128*1-1]);
         $display($time, " %m                  %x", msg_block[128*1:128*2-1]);
         $display($time, " %m                  %x", msg_block[128*2:128*3-1]);
         $display($time, " %m                  %x", msg_block[128*3:128*4-1]);
         if(last_block) begin
            $display($time, " %m --------------------------------------");
         end
      end
      if(hash_vld) begin
         $display($time, " %m Hash: %x", hash);
      end
   end
   // synthesis translate_on
   
endmodule // chi_hash_wrapper

// synthesis translate_off
module chi_hash_wrapper_tester ();

   reg clk = 0;
   reg reset = 1;
   wire [0:255] hash;
   wire         hash_vld;
   integer      count = 0;
   integer      test_index;
   integer      words_index;

   reg [0:63]   in_data = 0;
   reg [0:7]    in_ctrl = 0;
   reg          in_wr = 0;
   wire         in_rdy;

   always #4 clk = ~clk;

/* -----\/----- EXCLUDED -----\/-----
`define STARTING_LENGTH 1696
 -----/\----- EXCLUDED -----/\----- */
`define STARTING_LENGTH 8
`define TOTAL_WORDS 22707
`define NUM_MSGS 321
   
   reg [0:63]   in_data_test[0:`TOTAL_WORDS-1];
   reg [0:63]   in_ctrl_test[0:`TOTAL_WORDS-1];
   reg [31:0]   lengths[0:`NUM_MSGS-1];
   reg [0:255]  mds[0:`NUM_MSGS-1];

   initial begin
      $readmemh("msgs.vtxt", in_data_test);
      $readmemh("ctrl.vtxt", in_ctrl_test);
      $readmemh("lengths.vtxt", lengths);
      $readmemh("mds.vtxt", mds);
   end

   integer i;
   initial begin
      words_index = 0;
      test_index  = 0;
      for(test_index=0; 
          test_index<`NUM_MSGS && lengths[test_index] != `STARTING_LENGTH;
          test_index=test_index+1) begin
      end
      for(i=0;
          i<test_index; 
          words_index=words_index+1) begin
         
         if(in_ctrl_test[words_index] != 0) begin
            i = i+1;
         end
      end
   end // initial begin

   always @(posedge clk) begin
      count <= count + 1'b1;
      if(count > 2) begin
         reset <= 0;
         if(in_rdy) begin
            if(words_index == `TOTAL_WORDS-1) begin
               words_index <= 0;
            end else begin
               words_index <= words_index + 1;
            end
         end
      end
   end

   always @(posedge clk) begin
      in_data <= in_data_test[words_index];
      in_ctrl <= in_ctrl_test[words_index];
      in_wr <= in_rdy;
   end

   always @(*) begin
      if(hash_vld) begin
         if(hash !== mds[test_index]) begin
            $display("ERROR: CHI msg length %u failed. Expected %064x. Found %064x.",
                     lengths[test_index], mds[test_index], hash);
            $stop;
         end else begin
            $display("SUCCESS: msg length %u", lengths[test_index]);
         end
         test_index = test_index + 1;
         if(test_index == `NUM_MSGS) #32 $finish;
      end
   end
   
   chi_hash_wrapper chi_hash_wrapper
     (// Outputs
      .in_rdy                           (in_rdy),
      .hash                             (hash[0:255]),
      .hash_vld                         (hash_vld),
      // Inputs
      .in_data                          (in_data),
      .in_ctrl                          (in_ctrl[0:7]),
      .in_wr                            (in_wr),
      .clk                              (clk),
      .reset                            (reset));

endmodule // chi_hash_wrapper_tester
// synthesis translate_on
