/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: chi_hash_wrapper_pipelined
 * Project: CHI
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: Wrapper around the CHI hash to pad blocks and to read input as
 *              64-bit words.
 * 
 * Assumes messages are multiples of 8-bits in length (i.e. a byte stream).
 * 
 ******************************************************************************/
`timescale 1ns/1ps
module chi_hash_wrapper_ext2
  #(parameter MAX_LENGTH_BITS = 12) // bit length of the maximum msg size
  (input  [0:63]      in_data,
   input  [0:7]       in_ctrl,
   input              in_wr,
   output             in_rdy,
   output             prog_full,

   output     [0:255] hash,
   output             hash_vld,

   output reg [0:511] msg_block,
   output reg         msg_block_vld,
   output reg         last_block,
   input              msg_block_ack,

   input  [0:255]     hash_in,
   input              hash_vld_in,
   
   input          clk,
   input          reset);

   //-------------------- Internal Parameters ------------------------
   localparam WRITE_WORDS = 1,
              ADD_PAD     = 2,
              ADD_LENGTH  = 4;
   
   //------------------------ Wires/Regs -----------------------------
   wire [MAX_LENGTH_BITS-1:0] msg_length_added;
   reg [0:63]                 msg_word;
   reg                        msg_word_wr_en;
   reg                        last_block_in;

   reg [2:0]                  count, count_nxt;
   reg [MAX_LENGTH_BITS-1:0]  msg_length_nxt, msg_length;
   reg [2:0]                  state, state_nxt;

   wire [0:63]                padded_words[0:7];
   wire [0:63]                padded_data;
   wire [2:0]                 sel_pad;
   wire [3:0]                 len;

   wire [0:7]                 in_ctrl_dout;
   wire [0:63]                in_data_dout;
   reg                        in_fifo_rd_en;

   //------------------------- Modules ------------------------------
   /* need to fit at least 1 max sized packet so we can
    * simultaneously hash two packets. (second starts after first
    * is stored here). */
   fallthrough_fifo_prog256_512x72 in_fifo
     (.din           ({in_ctrl, in_data}),
      .wr_en         (in_wr),
      .rd_en         (in_fifo_rd_en),
      .dout          ({in_ctrl_dout, in_data_dout}),
      .full          (),
      .prog_full     (prog_full),
      .almost_full   (in_fifo_nearly_full),
      .empty         (in_fifo_empty),
      .rst           (reset),
      .clk           (clk)
      );

   //-------------------------- Logic --------------------------------
   assign hash_vld = hash_vld_in;
   assign hash = hash_in;
   assign in_rdy = !in_fifo_nearly_full;

   /********************************************
    * Control input into fifos
    ********************************************/

   /* Generate the pad for the last word */
   generate
      genvar i;
      for(i=0; i<6; i=i+1) begin:gen_pad
         assign padded_words[i] = {in_data_dout[0:i*8+7], 8'h80, {(6-i){8'h00}}};
      end
      assign    padded_words[6] = {in_data_dout[0:6*8+7], 8'h80};
      assign    padded_words[7] = in_data_dout;
   endgenerate
   priority_encoder #(.OUTPUT_WIDTH(3)) penc(in_ctrl_dout, sel_pad, sel_pad_vld);
   assign padded_data = padded_words[sel_pad];

   /* get the number of bytes written in */
   assign len = sel_pad_vld ? sel_pad + 1'b1 : 8;
   assign msg_length_added = msg_length + len;

   /* Select what fifo to write to, and add the 
    * pad and length when needed */
   always @(*) begin
      count_nxt        = count;
      msg_word_wr_en   = 0;
      msg_length_nxt   = msg_length;
      state_nxt        = state;
      msg_word         = in_data_dout;
      last_block_in    = 0;
      in_fifo_rd_en    = 0;
      case(state)
         WRITE_WORDS: begin
            if(!in_fifo_empty && !msg_block_vld) begin
               count_nxt           = count + 1'b1;
               in_fifo_rd_en       = 1;
               msg_word_wr_en      = 1;
               msg_length_nxt      = msg_length_added;
               if(in_ctrl_dout != 0) begin
                  /* if we need to write the pad in a new word */
                  if(in_ctrl_dout[7]) begin
                     state_nxt   = ADD_PAD;
                  end
                  else begin
                     state_nxt   = ADD_LENGTH;
                     msg_word    = padded_data;
                  end
               end // if (in_ctrl_dout != 0)
            end // if (in_wr)
         end // case: WRITE_WORDS

         ADD_PAD: begin
            if(!msg_block_vld) begin
               count_nxt        = count + 1'b1;
               msg_word_wr_en   = 1;
               msg_word         = {1'b1, 63'h0};
               state_nxt        = ADD_LENGTH;
            end
         end

         ADD_LENGTH: begin
            if(!msg_block_vld) begin
               count_nxt           = count + 1'b1;
               msg_word_wr_en      = 1;
               if(count == 7) begin
                  msg_word         = {{(64-MAX_LENGTH_BITS-3){1'b0}}, 
                                      msg_length, 3'h0};
                  state_nxt        = WRITE_WORDS;
                  msg_length_nxt   = 0;
                  last_block_in    = 1;
               end
               else begin
                  msg_word = 64'h0;
               end // else: !if(count == 7)
            end // if (!msg_block_vld)
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(msg_word_wr_en) begin
         msg_block     <= {msg_block[64:511], msg_word};
         last_block    <= last_block_in;
      end
      if(reset) begin
         msg_length       <= 0;
         state            <= WRITE_WORDS;
         count            <= 0;
         msg_block_vld    <= 0;
      end
      else begin
         msg_length    <= msg_length_nxt;
         state         <= state_nxt;
         count         <= count_nxt;

         if(msg_word_wr_en && count == 7) begin
            msg_block_vld <= 1'b1;
         end
         else if(msg_block_ack) begin
            msg_block_vld <= 1'b0;
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

/* -----\/----- EXCLUDED -----\/-----
   // synthesis translate_off
   always @(posedge clk) begin
      if(msg_block_ack) begin
         $display($time, " %m Inserting block");
         $display($time, " %m                  %x", msg_block[128*0:128*1-1]);
         $display($time, " %m                  %x", msg_block[128*1:128*2-1]);
         $display($time, " %m                  %x", msg_block[128*2:128*3-1]);
         $display($time, " %m                  %x", msg_block[128*3:128*4-1]);
         if(last_block) begin
            $display($time, " %m --------------------------------------");
         end
      end
      if(hash_vld) begin
         $display($time, " %m Hash: %x", hash);
      end
   end
   // synthesis translate_on
 -----/\----- EXCLUDED -----/\----- */

endmodule // chi_hash_wrapper_ext2

// synthesis translate_off
module chi_hash_wp_tester2 ();

   reg clk = 0;
   reg reset = 1;
   wire [0:255] hash, hash_out;
   wire         hash_vld, hash_vld_out;
   integer      count = 0;
   integer      test_index;
   integer      words_index;

   reg [0:63]   in_data = 0;
   reg [0:7]    in_ctrl = 0;
   reg          in_wr = 0;
   wire         in_rdy;
   wire [0:511] msg_block;

   always #4 clk = ~clk;

`define STARTING_LENGTH 'h18
`define TOTAL_WORDS 22707
`define NUM_MSGS 321
   
   reg [0:63]   in_data_test[0:`TOTAL_WORDS-1];
   reg [0:63]   in_ctrl_test[0:`TOTAL_WORDS-1];
   reg [31:0]   lengths[0:`NUM_MSGS-1];
   reg [0:255]  mds[0:`NUM_MSGS-1];

   initial begin
      $readmemh("msgs.vtxt", in_data_test);
      $readmemh("ctrl.vtxt", in_ctrl_test);
      $readmemh("lengths.vtxt", lengths);
      $readmemh("mds.vtxt", mds);
   end

   integer i;
   initial begin
      words_index = 0;
      test_index  = 0;
      for(test_index=0; 
          test_index<`NUM_MSGS && lengths[test_index] != `STARTING_LENGTH;
          test_index=test_index+1) begin
      end
      for(i=0;
          i<test_index; 
          words_index=words_index+1) begin
         
         if(in_ctrl_test[words_index] != 0) begin
            i = i+1;
         end
      end
   end // initial begin

   always @(posedge clk) begin
      count <= count + 1'b1;
      if(count > 2) begin
         reset <= 0;
         if(in_rdy) begin
            if(words_index == `TOTAL_WORDS-1) begin
               words_index <= 0;
            end else begin
               words_index <= words_index + 1;
            end
         end
      end
   end

   always @(posedge clk) begin
      in_data <= in_data_test[words_index];
      in_ctrl <= in_ctrl_test[words_index];
      in_wr <= in_rdy;
   end

   always @(*) begin
      if(hash_vld) begin
         if(hash !== mds[test_index]) begin
            $display("ERROR: CHI msg length %u failed. Expected %064x. Found %064x.",
                     lengths[test_index], mds[test_index], hash);
            $stop;
         end else begin
            $display("SUCCESS: msg length %u", lengths[test_index]);
         end
         test_index = test_index + 1;
         if(test_index == `NUM_MSGS) #32 $finish;
      end
   end

   chi_hash_wrapper_ext2 chi_hash_wrapper
     (// Outputs
      .in_rdy                           (in_rdy),
      .prog_full                        (prog_full),
      .hash                             (hash[0:255]),
      .hash_vld                         (hash_vld),
      .msg_block                        (msg_block[0:511]),
      .msg_block_vld                    (msg_block_vld),
      .last_block                       (last_block),
      // Inputs
      .in_data                          (in_data),
      .in_ctrl                          (in_ctrl[0:7]),
      .in_wr                            (in_wr),
      .msg_block_ack                    (msg_block_ack),
      .hash_in                          (hash_out[0:255]),
      .hash_vld_in                      (hash_vld_out),
      .clk                              (clk),
      .reset                            (reset));

   wire [0:255] temp1;
   wire [0:511] temp2;
   chi_hash_pipelined2 chi_hash
     (// Outputs
      .msg_block_ack                    ({msg_block_ack, temp}),
      .hash_out                         ({hash_out[0:255], temp1}),
      .hash_vld                         ({hash_vld_out, temp3}),
      // Inputs
      .msg_block_in                     ({msg_block[0:511], {512{1'bX}}}),
      .msg_block_vld                    ({msg_block_vld, 1'bx}),
      .last_block                       ({last_block, 1'bx}),
      .clk                              (clk),
      .reset                            (reset));

endmodule // chi_hash_wrapper_tester
// synthesis translate_on
