/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: grostl.v
 * Project: Grostl
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: implements the Grostl hash function.
 * 
 ******************************************************************************/
`timescale 1ns/1ps

`ifndef GROSTL_BLOCK_SIZE
 `include "grostl_def.v"
`endif

module grostl
  (input [0:`GROSTL_BLOCK_SIZE-1]  block,
   input                           block_vld,
   input                           last_block,
   output reg                      block_ack,
   
   output [0:255]                  hash,
   output reg                      hash_vld,
   
   input                           clk,
   input                           reset);

   //------------------ Internal Parameter ---------------------------
   localparam WAIT_FOR_NEW_MSG = 1,
              WAIT_FOR_PERMUTE = 2,
              DO_NXT_BLOCK     = 4,
              DO_LAST_P        = 8,
              WAIT_FOR_OUTPUT  = 16;

   //---------------------- Wires/Regs -------------------------------
   reg [0:511]  p_in, q_in;
   reg          p_in_vld, q_in_vld;
   wire [0:511] p_out, q_out, p_q_h;
   reg [0:511]  h_reg, h_reg_nxt;
   reg [0:4]    state, state_nxt;

   //----------------------- Modules ---------------------------------
   grostl_permute P
     (
      .s_in              (p_in),
      .s_in_vld          (p_in_vld),
      .s_in_ack          (p_in_ack),
                         
      .s_out             (p_out),
      .s_out_vld         (p_out_vld),

      .Q                 (1'b0),
                         
      .clk               (clk),
      .reset             (reset));
   
   grostl_permute Q
     (
      .s_in              (q_in),
      .s_in_vld          (q_in_vld),
      .s_in_ack          (q_in_ack),
                         
      .s_out             (q_out),
      .s_out_vld         (q_out_vld),
                         
      .Q                 (1'b1),

      .clk               (clk),
      .reset             (reset));
      
   //------------------------ Logic ----------------------------------
   assign hash = h_reg[256:511] ^ p_out[256:511];
   assign p_q_h = p_out ^ q_out ^ h_reg;

   always @(*) begin
      block_ack   = 0;
      p_in        = block ^ p_q_h;
      p_in_vld    = 0;
      q_in        = block;
      q_in_vld    = 0;
      h_reg_nxt   = h_reg;
      state_nxt   = state;
      hash_vld    = 0;
      case(state)
         WAIT_FOR_NEW_MSG: begin
            if(block_vld) begin
               block_ack   = 1;
               p_in        = {block[0:502], block[503:511] ^ `GROSTL_BLOCK_SIZE'd256};
               p_in_vld    = 1;
               q_in_vld    = 1;
               h_reg_nxt   = `GROSTL_BLOCK_SIZE'd256;
               if(last_block) begin
                  state_nxt = DO_LAST_P;
               end else begin
                  state_nxt   = WAIT_FOR_PERMUTE;
               end
            end
         end // case: WAIT_FOR_NEW_MSG

         WAIT_FOR_PERMUTE: begin
            if(p_out_vld) begin
               h_reg_nxt  = p_q_h;
/* -----\/----- EXCLUDED -----\/-----
               if(block_vld) begin
                  block_ack   = 1;
                  p_in        = block ^ p_q_h;
                  p_in_vld    = 1;
                  q_in_vld    = 1;
                  if(last_block) begin
                     state_nxt = DO_LAST_P;
                  end
               end // if (block_vld)
               else begin
 -----/\----- EXCLUDED -----/\----- */
                  state_nxt  = DO_NXT_BLOCK;
/* -----\/----- EXCLUDED -----\/-----
               end
 -----/\----- EXCLUDED -----/\----- */
            end
         end

         DO_NXT_BLOCK: begin
            if(block_vld) begin
               block_ack   = 1;
               p_in        = block ^ h_reg;
               p_in_vld    = 1;
               q_in_vld    = 1;
               if(last_block) begin
                  state_nxt = DO_LAST_P;
               end else begin
                  state_nxt   = WAIT_FOR_PERMUTE;
               end
            end // if (block_vld)
         end // case: DO_NXT_BLOCK

         DO_LAST_P: begin
            if(p_out_vld) begin
               h_reg_nxt  = p_q_h;
               p_in       = p_q_h;
               p_in_vld   = 1;
               state_nxt  = WAIT_FOR_OUTPUT;
            end
         end

         WAIT_FOR_OUTPUT: begin
            if(p_out_vld) begin
               hash_vld = 1'b1;
               state_nxt = WAIT_FOR_NEW_MSG;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      h_reg <= h_reg_nxt;
      state <= reset ? WAIT_FOR_NEW_MSG : state_nxt;
   end
endmodule

module grostl_serial
  (input [0:`GROSTL_BLOCK_SIZE-1]  block,
   input                           block_vld,
   input                           last_block,
   output reg                      block_ack,
   
   output [0:255]                  hash,
   output reg                      hash_vld,
   
   input                           clk,
   input                           reset);

   //------------------ Internal Parameter ---------------------------
   localparam WAIT_FOR_NEW_MSG = 1,
              WAIT_FOR_PERMUTE = 2,
              DO_NXT_BLOCK     = 4,
              DO_LAST_P        = 8,
              WAIT_FOR_OUTPUT  = 16,
              DO_Q             = 32;

   //---------------------- Wires/Regs -------------------------------
   reg [0:511]  pq_in;
   reg          pq_in_vld;
   wire [0:511] pq_out, p_q_h;
   reg [0:511]  h_reg, h_reg_nxt;
   reg [0:5]    state, state_nxt;
   reg          do_q;
   reg [0:511]  p, p_nxt;

   //----------------------- Modules ---------------------------------
   grostl_permute PQ
     (
      .s_in              (pq_in),
      .s_in_vld          (pq_in_vld),
      .s_in_ack          (pq_in_ack),
                         
      .s_out             (pq_out),
      .s_out_vld         (pq_out_vld),

      .Q                 (do_q),
                         
      .clk               (clk),
      .reset             (reset));
      
   //------------------------ Logic ----------------------------------
   assign hash = h_reg[256:511] ^ pq_out[256:511];
   assign p_q_h = p ^ pq_out ^ h_reg;

   always @(*) begin
      block_ack   = 0;
      pq_in       = block ^ p_q_h;
      pq_in_vld   = 0;
      p_nxt       = p;
      h_reg_nxt   = h_reg;
      state_nxt   = state;
      do_q        = 0;
      hash_vld    = 0;
      case(state)
         WAIT_FOR_NEW_MSG: begin
            if(block_vld) begin
               pq_in       = block ^ `GROSTL_BLOCK_SIZE'd256;
               pq_in_vld   = 1;
               do_q        = 0;
               h_reg_nxt   = `GROSTL_BLOCK_SIZE'd256;
               state_nxt   = DO_Q;
            end
         end // case: WAIT_FOR_NEW_MSG

         DO_Q: begin
            if(pq_out_vld) begin
               p_nxt       = pq_out;
               pq_in       = block;
               pq_in_vld   = 1;
               do_q        = 1;
               block_ack   = 1;
               if(last_block) begin
                  state_nxt   = DO_LAST_P;
               end else begin
                  state_nxt   = WAIT_FOR_PERMUTE;
               end
            end
         end // case: WAIT_FOR_NEW_MSG

         WAIT_FOR_PERMUTE: begin
            if(pq_out_vld) begin
               h_reg_nxt  = p_q_h;
               if(block_vld) begin
                  pq_in       = block ^ p_q_h;
                  pq_in_vld   = 1;
                  do_q        = 0;
                  state_nxt   = DO_Q;
               end // if (block_vld)
               else begin
                  state_nxt  = DO_NXT_BLOCK;
               end
            end
         end

         DO_NXT_BLOCK: begin
            if(block_vld) begin
               block_ack   = 1;
               pq_in       = block ^ h_reg;
               pq_in_vld   = 1;
               do_q        = 0;
               state_nxt   = DO_Q;
            end // if (block_vld)
         end // case: DO_NXT_BLOCK

         DO_LAST_P: begin
            if(pq_out_vld) begin
               h_reg_nxt   = p_q_h;
               pq_in       = p_q_h;
               pq_in_vld   = 1;
               do_q        = 0;
               state_nxt   = WAIT_FOR_OUTPUT;
            end
         end

         WAIT_FOR_OUTPUT: begin
            if(pq_out_vld) begin
               hash_vld    = 1'b1;
               state_nxt   = WAIT_FOR_NEW_MSG;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      h_reg    <= h_reg_nxt;
      state    <= reset ? WAIT_FOR_NEW_MSG : state_nxt;
      p        <= p_nxt;
   end
endmodule // grostl_serial

