/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: grostl_round.v
 * Project: Grostl
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description: implements one round the P or Q permutations of grostl
 * 
 ******************************************************************************/
`timescale 1ns/1ps

module grostl_round
  (input [0:511]  s_in,
   input          s_in_vld,
   output [0:511] s_out,
   output reg     s_out_vld,
   input [3:0]    round_num,
   input          Q_in,
   output reg     Q_out,

   input          clk,
   input          reset);

`include "grostl.v.inc"

   wire [0:511]   s_addroundcst, s_shiftbytes;
   reg [0:511]    s_subbytes;

   assign s_addroundcst = grostl_addroundconstant({4'h0, round_num}, Q_in, s_in);
   
   generate
      genvar  i;
      for (i=0; i<512/8; i=i+1) begin:subgen
         always @(posedge clk) 
           s_subbytes[i*8 +: 8] <= sbox(s_addroundcst[i*8 +: 8]);
      end
   endgenerate

   assign s_shiftbytes = grostl_shiftbytes(s_subbytes);

   grostl_mixbytes grostl_mixbytes(s_shiftbytes, s_out);
   always @(posedge clk) s_out_vld <= s_in_vld;
   always @(posedge clk) Q_out <= Q_in;

   // synthesis translate_off
   wire [0:7] s_in_m[0:7][0:7];
   wire [0:7] s_addroundcst_m[0:7][0:7];
   wire [0:7] s_subbytes_m[0:7][0:7];
   wire [0:7] s_shiftbytes_m[0:7][0:7];
   wire [0:7] s_out_m[0:7][0:7];
   generate
      genvar  j;
      for(i=0; i<8; i=i+1) begin
         for(j=0; j<8; j=j+1) begin
            assign s_in_m[j][i] = s_in[8 * (j + i*8) +: 8];
            assign s_addroundcst_m[j][i] = s_addroundcst[8 * (j + i*8) +: 8];
            assign s_shiftbytes_m[j][i] = s_shiftbytes[8 * (j + i*8) +: 8];
            assign s_subbytes_m[j][i] = s_subbytes[8 * (j + i*8) +: 8];
            assign s_out_m[j][i] = s_out[8 * (j + i*8) +: 8];
         end
      end // for (i=0; i<8; i=i+1)
   endgenerate
   // synthesis translate_on
  
endmodule // grostl_round

module grostl_round_23457
  (input [0:511]  s_in,
   input          s_in_vld,
   output [0:511] s_out,
   output reg     s_out_vld,
   input [3:0]    round_num,
   input          Q_in,
   output reg     Q_out,

   input          clk,
   input          reset);

`include "grostl.v.inc"

   wire [0:511]   s_addroundcst;
   wire [0:8*8*40-1] s_subbytes, s_shiftbytes;

   assign s_addroundcst = grostl_addroundconstant({4'h0, round_num}, Q_in, s_in);

   grostl_subtimes23457 grostl_subtimes23457(s_addroundcst, s_subbytes, clk);

   grostl_shiftbytes23457 grostl_shiftbytes23457(s_subbytes, s_shiftbytes);

   grostl_mixbytes23457 grostl_mixbytes23457(s_shiftbytes, s_out);
   
   always @(posedge clk) s_out_vld <= s_in_vld;
   always @(posedge clk) Q_out <= Q_in;

   // synthesis translate_off
   wire [0:7] s_in_m[0:7][0:7];
   wire [0:7] s_addroundcst_m[0:7][0:7];
   wire [0:39] s_subbytes_m[0:7][0:7];
   reg [0:7] s_subbytes1_m[0:7][0:7];
   wire [0:39] s_shiftbytes_m[0:7][0:7];
   wire [0:7] s_out_m[0:7][0:7];
   generate
      genvar  i,j;
      for(i=0; i<8; i=i+1) begin
         for(j=0; j<8; j=j+1) begin
            assign s_in_m[j][i] = s_in[8 * (j + i*8) +: 8];
            assign s_addroundcst_m[j][i] = s_addroundcst[8 * (j + i*8) +: 8];
            assign s_subbytes_m[j][i] = s_subbytes[40 * (j + i*8) +: 40];
            always @(posedge clk) s_subbytes1_m[j][i] <= sbox(s_addroundcst[8 * (j + i*8) +: 8]);
            assign s_shiftbytes_m[j][i] = s_shiftbytes[40 * (j + i*8) +: 40];
            assign s_out_m[j][i] = s_out[8 * (j + i*8) +: 8];
         end
      end // for (i=0; i<8; i=i+1)
   endgenerate
   // synthesis translate_on
  
endmodule // grostl_round_23457
/* -----\/----- EXCLUDED -----\/-----

module grostl_round_64
  (input [0:63]   s_in,
   input          s_in_vld,
   output [0:511] s_out,
   output reg     s_out_vld,
   input [3:0]    round_num,
   input          Q_in,

   input          clk,
   input          reset);

`include "grostl.v.inc"

   wire [0:63] 	     s_addroundcst;
   wire [0:8*40-1]   s_subbytes_word;
   reg [0:8*8*40-1]  s_subbytes;
   wire [0:8*8*40-1] s_shiftbytes;
   reg [2:0] 	     word_num;

   always @(posedge clk) word_num <= reset ? 0 : word_num + s_in_vld;

   assign s_addroundcst = grostl_addroundconstant_64({4'h0, round_num}, Q_in, s_in, word_num);

   grostl_subtimes23457_64 grostl_subtimes23457(s_addroundcst, s_subbytes_word, clk);
 
   always @(posedge clk) begin
      if(s_in_vld) begin
	 s_subbytes <= {s_subbytes[8*40 : 8*8*40 - 1], s_subbytes_word};
      end
   end

   grostl_shiftbytes23457 grostl_shiftbytes23457(s_subbytes, s_shiftbytes);

   grostl_mixbytes23457 grostl_mixbytes23457(s_shiftbytes, s_out);
   
   always @(posedge clk) s_out_vld <= (word_num == 7) && s_in_vld;

   // synthesis translate_off
   wire [0:39] s_shiftbytes_m[0:7][0:7];
   wire [0:7] s_out_m[0:7][0:7];
   generate
      genvar  i,j;
      for(i=0; i<8; i=i+1) begin
         for(j=0; j<8; j=j+1) begin
            assign s_shiftbytes_m[j][i] = s_shiftbytes[40 * (j + i*8) +: 40];
            assign s_out_m[j][i] = s_out[8 * (j + i*8) +: 8];
         end
      end // for (i=0; i<8; i=i+1)
   endgenerate
   // synthesis translate_on
  
endmodule // grostl_round_23457
 -----/\----- EXCLUDED -----/\----- */
