`timescale 1ns/1ps

module grostl_wrapper
  #(parameter MAX_LENGTH_BITS = 6, // block length of the maximum msg size
    parameter SERIAL          = 1)
  (input  [0:63]      in_data,
   input  [0:7]       in_ctrl,
   input              in_wr,
   output             in_rdy,

   output     [0:255] hash,
   output             hash_vld,

   input              clk,
   input              reset);

   wire  msg_block_vld, last_block, 
         msg_block_ack, hash_vld_out;
   wire [0:511] msg_block;
   wire [0:255] hash_out;

   grostl_wrapper_ext #(MAX_LENGTH_BITS) grostl_wrapper_ext
     (// Outputs
      .in_rdy                           (in_rdy),
      .hash                             (hash),
      .hash_vld                         (hash_vld),
      .msg_block                        (msg_block),
      .msg_block_vld                    (msg_block_vld),
      .last_block                       (last_block),
      // Inputs
      .in_data                          (in_data),
      .in_ctrl                          (in_ctrl),
      .in_wr                            (in_wr),
      .msg_block_ack                    (msg_block_ack),
      .hash_in                          (hash_out),
      .hash_vld_in                      (hash_vld_out),
      .clk                              (clk),
      .reset                            (reset));

   generate
      if(SERIAL) begin:gen_grostl_s
         grostl_serial grostl
           (// Outputs
            .block_ack                        (msg_block_ack),
            .hash                             (hash_out),
            .hash_vld                         (hash_vld_out),
            
            // Inputs
            .block                            (msg_block),
            .block_vld                        (msg_block_vld),
            .last_block                       (last_block),
            .clk                              (clk),
            .reset                            (reset));
      end // if (SERIAL)
      else begin:gen_grostl_p
         grostl grostl
           (// Outputs
            .block_ack                        (msg_block_ack),
            .hash                             (hash_out),
            .hash_vld                         (hash_vld_out),
            
            // Inputs
            .block                            (msg_block),
            .block_vld                        (msg_block_vld),
            .last_block                       (last_block),
            .clk                              (clk),
            .reset                            (reset));
      end // else: !if(SERIAL)
   endgenerate
endmodule

module grostl_wrapper_ext
  #(parameter MAX_LENGTH_BITS = 6) // block length of the maximum msg size
  (input  [0:63]      in_data,
   input  [0:7]       in_ctrl,
   input              in_wr,
   output             in_rdy,

   output reg [0:255] hash,
   output reg         hash_vld,

   output reg [0:511] msg_block,
   output reg         msg_block_vld,
   output reg         last_block,
   input              msg_block_ack,

   input  [0:255]     hash_in,
   input              hash_vld_in,
   
   input          clk,
   input          reset);

   //-------------------- Internal Parameters ------------------------
   localparam WRITE_WORDS = 1,
              ADD_PAD     = 2,
              ADD_LENGTH  = 4;
   
   //------------------------ Wires/Regs -----------------------------
   wire [MAX_LENGTH_BITS-1:0] msg_length_added;
   reg [0:63]                 msg_word;
   reg                        msg_word_wr_en;
   reg                        last_block_in;

   reg [2:0]                  count, count_nxt;
   reg [MAX_LENGTH_BITS-1:0]  msg_length;
   reg [2:0]                  state, state_nxt;

   wire [0:63]                padded_words[0:7];
   wire [0:63]                padded_data;
   wire [2:0]                 sel_pad;

   wire [0:7]                 in_ctrl_dout;
   wire [0:63]                in_data_dout;
   reg                        in_fifo_rd_en;

   //------------------------- Modules ------------------------------
   
   fallthrough_small_fifo 
     #(.WIDTH(72),
       .MAX_DEPTH_BITS(3),
       .PROG_FULL_THRESHOLD(5))
      in_fifo
        (.din           ({in_ctrl, in_data}),
         .wr_en         (in_wr),
         .rd_en         (in_fifo_rd_en),
         .dout          ({in_ctrl_dout, in_data_dout}),
         .full          (),
         .prog_full     (in_fifo_nearly_full),
         .nearly_full   (),
         .empty         (in_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //-------------------------- Logic --------------------------------
/* -----\/----- EXCLUDED -----\/-----
   assign hash_vld = hash_vld_in;
   assign hash = hash_in;
 -----/\----- EXCLUDED -----/\----- */
   always @(posedge clk) begin
      hash_vld    <= reset ? 0 : hash_vld_in;
      hash        <= hash_in;
   end

   assign in_rdy = !in_fifo_nearly_full;

   /********************************************
    * Control input into fifos
    ********************************************/

   /* Generate the pad for the last word */
   generate
      genvar i;
      for(i=0; i<6; i=i+1) begin:gen_pad
         assign padded_words[i] = {in_data_dout[0:i*8+7], 8'h80, {(6-i){8'h00}}};
      end
      assign    padded_words[6] = {in_data_dout[0:6*8+7], 8'h80};
      assign    padded_words[7] = in_data_dout;
   endgenerate
   priority_encoder #(.OUTPUT_WIDTH(3)) penc(in_ctrl_dout, sel_pad, sel_pad_vld);
   assign padded_data = padded_words[sel_pad];

   /* Select what fifo to write to, and add the 
    * pad and length when needed */
   always @(*) begin
      count_nxt        = count;
      msg_word_wr_en   = 0;
      state_nxt        = state;
      msg_word         = in_data_dout;
      last_block_in    = 0;
      in_fifo_rd_en    = 0;
      case(state)
         WRITE_WORDS: begin
            if(!reset && !in_fifo_empty && !msg_block_vld) begin
               count_nxt           = count + 1'b1;
               in_fifo_rd_en       = 1;
               msg_word_wr_en      = 1;
               if(in_ctrl_dout != 0) begin
                  /* if we need to write the pad in a new word */
                  if(in_ctrl_dout[7]) begin
                     state_nxt   = ADD_PAD;
                  end
                  else begin
                     state_nxt   = ADD_LENGTH;
                     msg_word    = padded_data;
                  end
               end // if (in_ctrl_dout != 0)
            end // if (in_wr)
         end // case: WRITE_WORDS

         ADD_PAD: begin
            if(!msg_block_vld) begin
               count_nxt        = count + 1'b1;
               msg_word_wr_en   = 1;
               msg_word         = {1'b1, 63'h0};
               state_nxt        = ADD_LENGTH;
            end
         end

         ADD_LENGTH: begin
            if(!msg_block_vld) begin
               count_nxt           = count + 1'b1;
               msg_word_wr_en      = 1;
               if(count == 7) begin
                  msg_word         = {{(64-MAX_LENGTH_BITS){1'b0}}, 
                                      msg_length};
                  state_nxt        = WRITE_WORDS;
                  last_block_in    = 1;
               end
               else begin
                  msg_word = 64'h0;
               end // else: !if(count == 7)
            end // if (!msg_block_vld)
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(msg_word_wr_en) begin
         msg_block     <= {msg_block[64:511], msg_word};
         last_block    <= last_block_in;
      end
      if(reset) begin
         msg_length       <= 1;
         state            <= WRITE_WORDS;
         count            <= 0;
         msg_block_vld    <= 0;
      end
      else begin
         state         <= state_nxt;
         count         <= count_nxt;

         if(msg_word_wr_en && count == 7) begin
            msg_block_vld <= 1'b1;
         end
         else if(msg_block_ack) begin
            msg_block_vld <= 1'b0;
            msg_length    <= last_block ? 1 : msg_length + 1'b1;
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

   // synthesis translate_off
   always @(posedge clk) begin
      if(msg_block_ack) begin
         $display($time, " %m Inserting block");
         $display($time, " %m                  %x", msg_block[128*0:128*1-1]);
         $display($time, " %m                  %x", msg_block[128*1:128*2-1]);
         $display($time, " %m                  %x", msg_block[128*2:128*3-1]);
         $display($time, " %m                  %x", msg_block[128*3:128*4-1]);
         if(last_block) begin
            $display($time, " %m --------------------------------------");
         end
      end
      if(hash_vld) begin
         $display($time, " %m Hash: %x", hash);
      end
   end
   // synthesis translate_on

endmodule // chi_hash_wrapper_ext2

// synthesis translate_off
module grostl_wrapper_tester ();

   reg clk = 0;
   reg reset = 1;
   wire [0:255] hash;
   wire         hash_vld;
   integer      count = 0;
   integer      test_index;
   integer      words_index;

   reg [0:63]   in_data = 0;
   reg [0:7]    in_ctrl = 0;
   reg          in_wr = 0;
   wire         in_rdy;

   always #4 clk = ~clk;

`define STARTING_LENGTH 'h18
`define TOTAL_WORDS 22681
`define NUM_MSGS 321
   
   reg [0:63]   in_data_test[0:`TOTAL_WORDS-1];
   reg [0:63]   in_ctrl_test[0:`TOTAL_WORDS-1];
   reg [31:0]   lengths[0:`NUM_MSGS-1];
   reg [0:255]  mds[0:`NUM_MSGS-1];

   initial begin
      $readmemh("msgs.vtxt", in_data_test);
      $readmemh("ctrl.vtxt", in_ctrl_test);
      $readmemh("lengths.vtxt", lengths);
      $readmemh("mds.vtxt", mds);
   end

   integer i;
   initial begin
      words_index = 0;
      test_index  = 0;
      for(test_index=0; 
          test_index<`NUM_MSGS && lengths[test_index] != `STARTING_LENGTH;
          test_index=test_index+1) begin
      end
      for(i=0;
          i<test_index; 
          words_index=words_index+1) begin
         
         if(in_ctrl_test[words_index] != 0) begin
            i = i+1;
         end
      end
   end // initial begin

   always @(posedge clk) begin
      count <= count + 1'b1;
      if(count > 2) begin
         reset <= 0;
         if(in_rdy) begin
            if(words_index == `TOTAL_WORDS-1) begin
               words_index <= 0;
            end else begin
               words_index <= words_index + 1;
            end
         end
      end
   end

   always @(posedge clk) begin
      in_data <= in_data_test[words_index];
      in_ctrl <= in_ctrl_test[words_index];
      in_wr <= in_rdy;
   end

   always @(*) begin
      if(hash_vld) begin
         if(hash !== mds[test_index]) begin
            $display("ERROR: msg length %u failed. Expected %064x. Found %064x.",
                     lengths[test_index], mds[test_index], hash);
            $stop;
         end else begin
            $display("SUCCESS: msg length %u", lengths[test_index]);
         end
         test_index = test_index + 1;
         if(test_index == `NUM_MSGS) #32 $finish;
      end
   end

   grostl_wrapper grostl_wrapper
     (// Outputs
      .in_rdy                           (in_rdy),
      .hash                             (hash[0:255]),
      .hash_vld                         (hash_vld),
      // Inputs
      .in_data                          (in_data),
      .in_ctrl                          (in_ctrl[0:7]),
      .in_wr                            (in_wr),
      .clk                              (clk),
      .reset                            (reset));

endmodule // chi_hash_wrapper_tester
// synthesis translate_on
