/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: hash_pkt.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Finds CHI(version || IV || path || proto || pkt_len || 0x00 || payload)
 *   This module instantiates two CHI modules, and selects one to use based on
 *   the input port of the packet.
 ******************************************************************************/

module hash_pkt
  (input      [0:7]   in_ctrl,
   // --- Interface to packet parser
   input      [7:0]   icing_version,
   input              icing_version_vld,
   input      [0:55]  icing_iv,
   input              icing_iv_vld,
   input      [0:63]  icing_realm_id,
   input              icing_realm_id_0_vld,
   input              icing_realm_id_1_vld,
   input              icing_realm_id_2_vld,
   input      [0:63]  icing_payload,
   input              icing_payload_vld,
   input              icing_payload_is_last,
   input              icing_error_idx_vld,

   // --- Interface to flow control
   output             pause,

   // --- Results
   output     [0:246] hash,
   output             hash_vld,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ Wires/Regs -----------------------------
   wire [0:1]         hash_out_vld;
   wire [0:1]         hash_in_rdy;
   wire [0:1]         prog_full;
   wire [0:246]       hash_out[0:1];

   reg                hash_sel;
   reg [0:7]          ver;
   reg                hashing_pkt;
   reg [0:63]         hash_in_data;
   reg [0:7]          hash_in_ctrl;
   reg [0:1]          hash_in_wr;

   reg [0:7] 	      in_ctrl_d1;

   wire [0:63]        icing_payload_masked;
   reg                toggle;

   wire [0:246]       hash_fifo_dout[0:1];
   wire [0:1] 	      rd_hash;
   wire [0:1] 	      hash_fifo_empty;

   //------------------------- Modules -------------------------------

   fallthrough_small_fifo 
     #(.WIDTH(1),
       .MAX_DEPTH_BITS(2))
      order_fifo
        (.din           (hash_sel),
         .wr_en         (icing_payload_is_last && icing_payload_vld),
         .rd_en         (hash_vld),
         .dout          (hash_sel_out),
         .full          (),
         .prog_full     (),
         .nearly_full   (order_fifo_almost_full),
         .empty         (),
         .reset         (reset),
         .clk           (clk)
         );

/* -----\/----- EXCLUDED -----\/-----
   generate 
      genvar   i;
      for(i=0; i<2; i=i+1) begin:gen_hashes
         wire [0:8] unused;
         chi_hash_wrapper chi_hash_wrapper
           (.in_data (hash_in_data),
            .in_ctrl (hash_in_ctrl),
            .in_wr   (hash_in_wr[i]),
            .in_rdy  (hash_in_rdy[i]),
            .hash    ({unused, hash_out[i]}),
            .hash_vld(hash_out_vld[i]),
            .clk     (clk),
            .reset   (reset));
      end // block: gen_hashes
   endgenerate
 -----/\----- EXCLUDED -----/\----- */

   wire [0:8] 	      unused, unused2;
   chi_hash_wrapper_pipelined chi_hash_wrapper
     (.in_data ({hash_in_data, hash_in_data}),
      .in_ctrl ({hash_in_ctrl, hash_in_ctrl}),
      .in_wr   (hash_in_wr),
      .prog_full (prog_full),
      .in_rdy  (hash_in_rdy),
      .hash    ({unused, hash_out[0],
		 unused2, hash_out[1]}),
      .hash_vld(hash_out_vld),
      .clk     (clk),
      .reset   (reset));

   fallthrough_small_fifo 
     #(.WIDTH(247),
       .MAX_DEPTH_BITS(2))
      hash_fifo_0
        (.din           (hash_out[0]),
         .wr_en         (hash_out_vld[0]),
         .rd_en         (rd_hash[0]),
         .dout          (hash_fifo_dout[0]),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (hash_fifo_empty[0]),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(247),
       .MAX_DEPTH_BITS(2))
      hash_fifo_1
        (.din           (hash_out[1]),
         .wr_en         (hash_out_vld[1]),
         .rd_en         (rd_hash[1]),
         .dout          (hash_fifo_dout[1]),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (hash_fifo_empty[1]),
         .reset         (reset),
         .clk           (clk)
         );
   //-------------------------- Logic --------------------------------

   /*********************************************************************
    * Latch data
    ********************************************************************/
   always @(posedge clk) begin
      if(reset) begin
	 hash_sel 	<= 0;
	 hashing_pkt 	<= 0;
	 toggle         <= 0;
      end
      else if(hashing_pkt) begin
	 if(icing_payload_vld && icing_payload_is_last) begin
	    hashing_pkt    <= 0;
	    toggle         <= ~toggle;
	 end
      end
      else begin
	 /* when not hashing a pkt, update hash_sel
	  * so we can be ready */
	 hash_sel    <= prog_full[toggle] ? ~toggle : toggle;
	 if(icing_version_vld) begin
	    hashing_pkt    <= 1;
	 end
      end // else: !if(hashing_pkt)

      in_ctrl_d1 <= reset ? 1 : in_ctrl;
      
      if(icing_version_vld) ver <= icing_version;
   end

   /*********************************************************************
    * Flow control
    ********************************************************************/
   assign pause = !hash_in_rdy[hash_sel] || order_fifo_almost_full;

   /*********************************************************************
    * Choose input to hash
    ********************************************************************/
   assign icing_payload_masked 
          = icing_error_idx_vld
          ? {icing_payload[0:23], 8'h0, icing_payload[32:63]} 
          : icing_payload;
   
   always @(*) begin
      hash_in_data   = icing_realm_id;
      hash_in_ctrl   = 0;
      hash_in_wr     = 0;
      if(icing_iv_vld) begin
         hash_in_data           = {ver, icing_iv};
         hash_in_wr[hash_sel]   = 1'b1;
      end else if(icing_realm_id_0_vld 
                  || icing_realm_id_1_vld 
                  || icing_realm_id_2_vld) begin
         hash_in_data           = icing_realm_id;
         hash_in_wr[hash_sel]   = 1'b1;
      end else if(icing_payload_vld && !icing_payload_is_last) begin
         hash_in_data           = icing_payload_masked;
         hash_in_wr[hash_sel]   = 1'b1;
      end else if(icing_payload_vld && icing_payload_is_last) begin
         hash_in_ctrl           = in_ctrl_d1;
         hash_in_data           = icing_payload_masked;
         hash_in_wr[hash_sel]   = 1'b1;
      end
   end // always @ (*)
   
   /*********************************************************************
    * Generate Output
    ********************************************************************/
   assign hash = hash_fifo_dout[hash_sel_out];
   assign hash_vld = !hash_fifo_empty[hash_sel_out];
   assign rd_hash[0] = hash_vld && (hash_sel_out == 0);
   assign rd_hash[1] = hash_vld && (hash_sel_out == 1);

   // synthesis translate_off
   integer num_hashes = 0;
   always @(posedge clk) begin
      if(hash_vld === 1) begin
         num_hashes = num_hashes + 1;
         $display($time, " %m Num hashes calc'ed:  %u", num_hashes);
      end
   end
   // synthesis translate_on
endmodule // hash_pkt

module hash_pkt_grostl
  (input      [0:7]   in_ctrl,
   // --- Interface to packet parser
   input      [7:0]   icing_version,
   input              icing_version_vld,
   input      [0:55]  icing_iv,
   input              icing_iv_vld,
   input      [0:63]  icing_realm_id,
   input              icing_realm_id_0_vld,
   input              icing_realm_id_1_vld,
   input              icing_realm_id_2_vld,
   input      [0:63]  icing_payload,
   input              icing_payload_vld,
   input              icing_payload_is_last,
   input              icing_error_idx_vld,

   // --- Interface to flow control
   output             pause,

   // --- Results
   output     [0:246] hash,
   output             hash_vld,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ Wires/Regs -----------------------------
   reg [0:7]          ver;
   reg [0:63]         hash_in_data;
   reg [0:7]          hash_in_ctrl;
   reg                hash_in_wr;

   wire [0:63]        icing_payload_masked;
   
   //------------------------- Modules -------------------------------

   wire [0:8] 	      unused;
   grostl_wrapper grostl_wrapper
     (.in_data (hash_in_data),
      .in_ctrl (hash_in_ctrl),
      .in_wr   (hash_in_wr),
      .in_rdy  (hash_in_rdy),
      .hash    ({unused, hash}),
      .hash_vld(hash_vld),
      .clk     (clk),
      .reset   (reset));

   //-------------------------- Logic --------------------------------

   /*********************************************************************
    * Latch data
    ********************************************************************/
   always @(posedge clk) begin
      
      if(icing_version_vld) ver <= icing_version;
   end

   /*********************************************************************
    * Flow control
    ********************************************************************/
   assign pause = !hash_in_rdy;

   /*********************************************************************
    * Choose input to hash
    ********************************************************************/
   assign icing_payload_masked 
          = icing_error_idx_vld
          ? {icing_payload[0:23], 8'h0, icing_payload[32:63]} 
          : icing_payload;
   
   always @(*) begin
      hash_in_data   = icing_realm_id;
      hash_in_ctrl   = 0;
      hash_in_wr     = 0;
      if(icing_iv_vld) begin
         hash_in_data           = {ver, icing_iv};
         hash_in_wr             = 1'b1;
      end else if(icing_realm_id_0_vld 
                  || icing_realm_id_1_vld 
                  || icing_realm_id_2_vld) begin
         hash_in_data           = icing_realm_id;
         hash_in_wr             = 1'b1;
      end else if(icing_payload_vld && !icing_payload_is_last) begin
         hash_in_data           = icing_payload_masked;
         hash_in_wr             = 1'b1;
      end else if(icing_payload_vld && icing_payload_is_last) begin
         hash_in_ctrl           = in_ctrl;
         hash_in_data           = icing_payload_masked;
         hash_in_wr             = 1'b1;
      end
   end // always @ (*)
   
   /*********************************************************************
    * Generate Output
    ********************************************************************/

   // synthesis translate_off
   integer num_hashes = 0;
   always @(posedge clk) begin
      if(hash_vld === 1) begin
         num_hashes = num_hashes + 1;
         $display($time, " %m Num hashes calc'ed:  %u", num_hashes);
      end
   end
   // synthesis translate_on
endmodule // hash_pkt

