/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: icing_check_hdr.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 *
 * Checks that the packet is well-formed before allowing it to pass
 ******************************************************************************/
`timescale 1ns/1ps

module icing_check_hdr
  (// --- data path interface
   input [0:63]            in_data,
   input [0:7]             in_ctrl,
   input                   in_wr,
   output                  in_rdy,

   output reg [0:63]       out_data,
   output reg [0:7]        out_ctrl,
   output reg              out_wr,
   input                   out_rdy,
   
   output                  dst_mac_not_ok,
   output                  ethertype_not_ok,
   output                  path_idx_not_ok,
   output                  path_len_not_ok,
   output                  pkt_len_not_ok,
   output                  version_not_ok,
   output                  pkt_not_ok,
   output                  hdr_check_done,

   input  [0:48*4-1]       port_macs,

   input                   clk,
   input                   reset);

   localparam WAIT_FOR_CHK = 1,
              DISCARD_PKT  = 2,
              OUTPUT_PKT   = 4;

   reg         in_fifo_rd_en;
   wire [0:7]  out_ctrl_nxt;
   wire [0:63] out_data_nxt;
   wire [3:0]  input_port;
   wire [10:0] byte_length;
   wire [0:47] dst_mac;
   wire [0:15] eth_type;
   wire [7:0]  icing_version;
   wire [7:0]  icing_path_len;
   wire [7:0]  icing_path_idx;
   reg         pkt_stat_fifo_rd_en;
   reg [2:0]   state, state_nxt;
   reg         out_wr_nxt;
   reg         out_ctrl_prev_is_0;
   
   fallthrough_small_fifo 
     #(.WIDTH(72),
       .MAX_DEPTH_BITS(3)) in_fifo
     (.din           ({in_ctrl, in_data}),
      .wr_en         (in_wr),
      .rd_en         (in_fifo_rd_en),
      .dout          ({out_ctrl_nxt, out_data_nxt}),
      .full          (in_fifo_full),
      .prog_full     (),
      .nearly_full   (in_fifo_almost_full),
      .empty         (in_fifo_empty),
      .reset         (reset),
      .clk           (clk)
      );

   icing_pkt_parser
     #(.DATA_WIDTH        (64),
       .CTRL_WIDTH        (8)) icing_pkt_parser
       (// --- interface to data path
        .in_data                  (in_data),
        .in_ctrl                  (in_ctrl),
        .in_wr                    (in_wr),

        .wait_for_eop             (pkt_not_ok),
        
        // --- parser results
        .input_port               (input_port),
        .input_port_vld           (input_port_vld),
        .word_length              (),
        .word_length_vld          (),
        .byte_length              (byte_length),
        .byte_length_vld          (byte_length_vld),
        .dst_mac                  (dst_mac),
        .dst_mac_vld              (dst_mac_vld),
        .src_mac                  (),
        .src_mac_vld              (),
        .eth_type                 (eth_type),
        .eth_type_vld             (eth_type_vld),
        .icing_version            (icing_version),
        .icing_version_vld        (icing_version_vld),
        .icing_path_len           (icing_path_len),
        .icing_path_len_vld       (icing_path_len_vld),
        .icing_path_idx           (icing_path_idx),
        .icing_path_idx_vld       (icing_path_idx_vld),
        .icing_last_hop_reached   (),
        .icing_iv                 (),
        .icing_iv_vld             (),
        .icing_realm_id           (),
        .icing_realm_id_0_vld     (),
        .icing_realm_id_1_vld     (),
        .icing_realm_id_2_vld     (),
        .icing_subrealm           (),
        .icing_subrealm_vld       (),
        .icing_realm_is_first     (),
        .icing_realm_is_last      (),
        .icing_realm_is_me        (),
        .icing_realm_is_b4_me     (),
        .icing_realm_is_nxt_hop   (),
        .icing_va                 (),
        .icing_va_0_vld           (),
        .icing_va_1_vld           (),
        .icing_vb                 (),
        .icing_vb_vld             (),
        .icing_vtag_is_me         (),
        .icing_vtag_is_last       (),
        .icing_proto              (),
        .icing_proto_vld          (),
        .icing_pkt_len            (),
        .icing_pkt_len_vld        (),
        .icing_error_idx          (),
        .icing_error_idx_vld      (),
        .icing_payload            (),
        .icing_payload_is_last    (),
        .icing_payload_vld        (),
        .word_count               (),
        .word_count_is_my_realm_id(),
        .word_count_is_my_vtag    (),
        
        // --- Misc
        .clk                      (clk),
        .reset                    (reset));

   check_hdr check_hdr
     (// Outputs
      .dst_mac_not_ok                   (dst_mac_not_ok),
      .ethertype_not_ok                 (ethertype_not_ok),
      .path_idx_not_ok                  (path_idx_not_ok),
      .path_len_not_ok                  (path_len_not_ok),
      .pkt_len_not_ok                   (pkt_len_not_ok),
      .version_not_ok                   (version_not_ok),
      .pkt_not_ok                       (pkt_not_ok),
      .hdr_check_done                   (hdr_check_done),
      // Inputs
      .input_port                       (input_port[3:0]),
      .input_port_vld                   (input_port_vld),
      .byte_length                      (byte_length[10:0]),
      .byte_length_vld                  (byte_length_vld),
      .dst_mac                          (dst_mac[0:47]),
      .dst_mac_vld                      (dst_mac_vld),
      .eth_type                         (eth_type[0:15]),
      .eth_type_vld                     (eth_type_vld),
      .icing_version                    (icing_version[7:0]),
      .icing_version_vld                (icing_version_vld),
      .icing_path_len                   (icing_path_len[7:0]),
      .icing_path_len_vld               (icing_path_len_vld),
      .icing_path_idx                   (icing_path_idx[7:0]),
      .icing_path_idx_vld               (icing_path_idx_vld),
      .port_macs                        (port_macs[0:48*4-1]),
      .clk                              (clk),
      .reset                            (reset));

   fallthrough_small_fifo 
     #(.WIDTH(1),
       .MAX_DEPTH_BITS(2))
      pkt_stat_fifo
        (.din           (pkt_not_ok),
         .wr_en         (hdr_check_done),
         .rd_en         (pkt_stat_fifo_rd_en),
         .dout          (pkt_not_ok_dout),
         .full          (),
         .prog_full     (),
         .nearly_full   (pkt_stat_fifo_almost_full),
         .empty         (pkt_stat_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   assign in_rdy = !pkt_stat_fifo_almost_full && !in_fifo_almost_full;

   always @(*) begin
      state_nxt             = state;
      pkt_stat_fifo_rd_en   = 0;
      in_fifo_rd_en         = 0;
      out_wr_nxt            = 0;
      case(state)
         WAIT_FOR_CHK: begin
            if(!pkt_stat_fifo_empty) begin
               if(pkt_not_ok_dout) begin
                  state_nxt     = DISCARD_PKT;
               end
               else begin
                  state_nxt     = OUTPUT_PKT;
               end
               pkt_stat_fifo_rd_en = 1;
            end // if (!pkt_stat_fifo_empty)
         end // case: WAIT_FOR_CHK
         
         DISCARD_PKT: begin
            if(!in_fifo_empty) begin
               in_fifo_rd_en = 1;
               /* eop */
               if(out_ctrl_nxt != 0 && out_ctrl_prev_is_0) begin
                  state_nxt = WAIT_FOR_CHK;
               end
            end
         end // case: DISCARD_PKT

         OUTPUT_PKT: begin
            if(!in_fifo_empty && out_rdy) begin
               in_fifo_rd_en = 1;
               out_wr_nxt    = 1;
               /* eop */
               if(out_ctrl_nxt != 0 && out_ctrl_prev_is_0) begin
                  state_nxt = WAIT_FOR_CHK;
               end
            end
         end // case: OUTPUT_PKT
      endcase // case(state)
   end // always @ (*)
   
   always @(posedge clk) begin
      out_data <= out_data_nxt;
      out_ctrl <= out_ctrl_nxt;
      if(reset) begin
         out_wr               <= 0;
         state                <= WAIT_FOR_CHK;
         out_ctrl_prev_is_0   <= 0;
      end
      else begin
         out_wr    <= out_wr_nxt;
         state     <= state_nxt;
         if(in_fifo_rd_en) out_ctrl_prev_is_0 <= (out_ctrl_nxt == 0);
      end // else: !if(reset)
   end // always @ (posedge clk)
endmodule // icing_check_hdr
