/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: icing_pkt_parser.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   This module parses a packet as it comes in and 
 *   sets signals indicating fields
 * 
 * deftype struct realm {
 *    uint8_t id[20];
 *    uint8_t voq[4];
 * } realm_t;
 * 
 * deftype struct vtag {
 *    uint8_t va[12];
 *    uint8_t vb[4];
 * } vtag_t;
 * 
 * struct icing_hdr {
 *    uint8_t version;
 *    uint8_t path_len;
 *    uint8_t path_idx;
 *    uint56_t init_vec;
 *    realm_t *path;
 *    vtag_t *vtag_vector;
 *    uint8_t proto;
 *    uint16_t pkt_len;
 * }
 * 
 ******************************************************************************/

module icing_pkt_parser
  #(parameter DATA_WIDTH           = 64,
    parameter CTRL_WIDTH           = DATA_WIDTH/8)
   (// --- interface to data path
    input [0:DATA_WIDTH-1] in_data,
    input [0:CTRL_WIDTH-1] in_ctrl,
    input                  in_wr,

    // --- interface to error
    input                  wait_for_eop,

    // --- parser results
    output reg [3:0]       input_port,
    output reg             input_port_vld,
    output reg [7:0]       word_length,
    output reg             word_length_vld,
    output reg [10:0]      byte_length,
    output reg             byte_length_vld,
    output reg [0:47]      dst_mac,
    output reg             dst_mac_vld,
    output reg [0:47]      src_mac,
    output reg             src_mac_vld,
    output reg [0:15]      eth_type,
    output reg             eth_type_vld,
    output reg [7:0]       icing_version,
    output reg             icing_version_vld,
    output reg [7:0]       icing_path_len,
    output reg             icing_path_len_vld,
    output reg [7:0]       icing_path_idx,
    output reg             icing_path_idx_vld,
    output reg             icing_last_hop_reached,
    output reg [0:55]      icing_iv,
    output reg             icing_iv_vld,
    output reg [0:63]      icing_realm_id,
    output reg             icing_realm_id_0_vld,
    output reg             icing_realm_id_1_vld,
    output reg             icing_realm_id_2_vld,
    output reg [0:31]      icing_subrealm,
    output reg             icing_subrealm_vld,
    output reg             icing_realm_is_first,
    output reg             icing_realm_is_last,
    output reg             icing_realm_is_me,
    output reg             icing_realm_is_b4_me,
    output reg             icing_realm_is_nxt_hop,
    output reg [0:63]      icing_va,
    output reg             icing_va_0_vld,
    output reg             icing_va_1_vld,
    output reg [0:31]      icing_vb,
    output reg             icing_vb_vld,
    output reg             icing_vtag_is_me,
    output reg             icing_vtag_is_last,
    output reg [0:7]       icing_proto,
    output reg             icing_proto_vld,
    output reg [0:15]      icing_pkt_len,
    output reg             icing_pkt_len_vld,
    output reg [0:7]       icing_error_idx,
    output reg             icing_error_idx_vld,
    output reg [0:63]      icing_payload,
    output reg             icing_payload_is_last,
    output reg             icing_payload_vld,

    output reg [7:0]       word_count,
    output                 word_count_is_my_realm_id,
    output                 word_count_is_my_vtag,
    
    // --- Misc
    input                  clk,
    input                  reset);

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_PKT      = 1,
              ETH_HDR_2         = 2**1,
              ICING_INIT_VEC    = 2**2,
              READ_PATH_1       = 2**3,
              READ_PATH_2       = 2**4,
              READ_PATH_3       = 2**5,
              SKIP_VTAGS        = 2**6,
              SKIP_VTAGS_1      = 2**7,
              LATCH_MY_VTAG     = 2**8,
              LATCH_MY_VTAG_1   = 2**9,
              OTHER_VTAG_1      = 2**10,
              OTHER_VTAG_2      = 2**11,
              PROTO_LEN         = 2**12,
              PAYLOAD           = 2**13,
              WAIT_FOR_EOP      = 2**14;
   localparam NUM_STATES = 15;
   //------------------------ Wires/Regs -----------------------------
   reg [0:15]              temp_reg;
   reg [0:5]               count;
   reg [0:NUM_STATES-1]    state;
   
   //-------------------------- Logic --------------------------------
   assign word_count_is_my_vtag     = icing_vtag_is_me && icing_va_0_vld;
   assign word_count_is_my_realm_id = icing_realm_is_me && icing_realm_id_0_vld;

   always @(posedge clk) begin
      byte_length                 <= in_data[53:63];
      byte_length_vld             <= 0;
      input_port                  <= in_data[45:47];
      input_port_vld              <= 0;
      word_length                 <= in_data[24:31];
      word_length_vld             <= 0;
      dst_mac                     <= in_data[0:47];
      dst_mac_vld                 <= 0;
      src_mac                     <= {temp_reg[0:15], in_data[0:31]};
      src_mac_vld                 <= 0;
      eth_type                    <= in_data[32:47];
      eth_type_vld                <= 0;
      icing_version               <= in_data[48:55];
      icing_version_vld           <= 0;
      icing_path_len_vld          <= 0;
      icing_path_idx_vld          <= 0;
      icing_last_hop_reached      <= (icing_path_idx == icing_path_len - 1'b1);
      icing_iv                    <= in_data[8:63];
      icing_iv_vld                <= 0;
      icing_realm_id              <= in_data;
      icing_realm_id_0_vld        <= 0;
      icing_realm_id_1_vld        <= 0;
      icing_realm_id_2_vld        <= 0;
      icing_subrealm              <= in_data[32:63];
      icing_subrealm_vld          <= 0;
      icing_realm_is_first        <= 0;
      icing_realm_is_b4_me        <= 0;
      icing_realm_is_me           <= 0;
      icing_realm_is_nxt_hop      <= 0;
      icing_realm_is_last         <= 0;
      icing_va                    <= in_data;
      icing_va_0_vld              <= 0;
      icing_va_1_vld              <= 0;
      icing_vb                    <= in_data[32:63];
      icing_vb_vld                <= 0;
      icing_vtag_is_me            <= 0;
      icing_vtag_is_last          <= 0;
      icing_proto                 <= in_data[0:7];
      icing_proto_vld             <= 0;
      icing_pkt_len               <= in_data[8:23];
      icing_pkt_len_vld           <= 0;
      icing_error_idx             <= in_data[24:31];
      icing_error_idx_vld         <= 0;
      icing_payload               <= in_data;
      icing_payload_vld           <= 0;
      icing_payload_is_last       <= 0;

      if(reset) begin
         state         <= WAIT_FOR_PKT;
         word_count    <= 0;
      end else if(wait_for_eop) begin
         state <= WAIT_FOR_EOP;
      end else begin
         case (state)
            /* skip module headers,
             * latch len from IOQ hdr,
             * rd dst_mac
             */
            WAIT_FOR_PKT: begin
               if(in_wr) begin
                  word_count <= word_count + 1'b1;
                  if(in_ctrl == 8'hFF) begin
                     byte_length_vld   <= 1;
                     input_port_vld    <= 1;
                     word_length_vld   <= 1;
                  end
                  /* {DstMac, SrcMac[0:15]} */
                  else if(in_ctrl == 0) begin
                     dst_mac_vld    <= 1;
                     temp_reg       <= in_data[48:63];
                     state          <= ETH_HDR_2;
                  end
               end // if (in_wr)
            end // case: WAIT_FOR_PKT

            /* src mac, eth type,
             * ICING version, path len */
            ETH_HDR_2: begin
               if(in_wr) begin
                  word_count            <= word_count + 1'b1;
                  src_mac_vld           <= 1;
                  eth_type_vld          <= 1;
                  icing_version_vld     <= 1;
                  icing_path_len_vld    <= 1;
                  icing_path_len        <= in_data[56:63];
                  state                 <= ICING_INIT_VEC;
               end // if (in_wr)
            end // case: ETH_HDR_2

            /* path idx, init vec */
            ICING_INIT_VEC: begin
               if(in_wr) begin
                  word_count            <= word_count + 1'b1;
                  icing_path_idx_vld    <= 1;
                  icing_path_idx        <= in_data[0:7];
                  icing_iv_vld          <= 1;

                  count                 <= 0;
                  state                 <= READ_PATH_1;
               end // if (in_wr)
            end // case: ICING_INIT_VEC
            
            /* Latch first 64 bits of realm_id */
            READ_PATH_1: begin
               if(in_wr) begin
                  word_count              <= word_count + 1'b1;
                  icing_realm_id_0_vld    <= 1'b1;
                  state                   <= READ_PATH_2;

                  if(count == 0) begin
                     icing_realm_is_first   <= 1'b1;
                  end
                  if(count == icing_path_idx - 1'b1) begin
                     icing_realm_is_b4_me   <= 1'b1;
                  end 
                  if(count == icing_path_idx) begin
                     icing_realm_is_me      <= 1'b1;
                  end 
                  if(count == icing_path_idx + 1'b1) begin
                     icing_realm_is_nxt_hop <= 1'b1;
                  end
                  if(count == icing_path_len - 1'b1) begin
                     icing_realm_is_last    <= 1'b1;
                  end
               end
            end

            READ_PATH_2: begin
               if(in_wr) begin
                  word_count              <= word_count + 1'b1;
                  icing_realm_id_1_vld    <= 1'b1;
                  state                   <= READ_PATH_3;

                  if(count == 0) begin
                     icing_realm_is_first <= 1'b1;
                  end
                  if(count == icing_path_idx - 1'b1) begin
                     icing_realm_is_b4_me <= 1'b1;
                  end 
                  if(count == icing_path_idx) begin
                     icing_realm_is_me <= 1'b1;
                  end 
                  if(count == icing_path_idx + 1'b1) begin
                     icing_realm_is_nxt_hop <= 1'b1;
                  end
                  if(count == icing_path_len - 1'b1) begin
                     icing_realm_is_last <= 1'b1;
                  end
               end
            end // case: READ_PATH_2

            READ_PATH_3: begin
               if(in_wr) begin
                  word_count              <= word_count + 1'b1;
                  icing_realm_id_2_vld    <= 1'b1;
                  icing_subrealm_vld      <= 1'b1;
                  
                  if(count == 0) begin
                     icing_realm_is_first <= 1'b1;
                  end
                  if(count == icing_path_idx - 1'b1) begin
                     icing_realm_is_b4_me <= 1'b1;
                  end 
                  if(count == icing_path_idx) begin
                     icing_realm_is_me <= 1'b1;
                  end 
                  if(count == icing_path_idx + 1'b1) begin
                     icing_realm_is_nxt_hop <= 1'b1;
                  end
                  
                  if(count == icing_path_len - 1'b1) begin
                     state                  <= (icing_path_idx == 1) ? LATCH_MY_VTAG : SKIP_VTAGS;
                     icing_realm_is_last    <= 1'b1;
                     count                  <= 1;
                  end
                  else begin
                     state <= READ_PATH_1;
                     count <= count + 1'b1;
                  end
               end // if (in_wr)
            end // case: READ_PATH_3
            
            /* Skip 1st 64bits of old VTAG */
            SKIP_VTAGS: begin
               if(in_wr) begin
                  word_count   <= word_count + 1'b1;
                  count        <= count + 1'b1;
                  state        <= SKIP_VTAGS_1;
               end
            end

            /* Skip 2nd 64bits of old VTAG */
            SKIP_VTAGS_1: begin
               if(in_wr) begin
                  word_count <= word_count + 1'b1;
                  state <= ((count == icing_path_idx) 
                                ? LATCH_MY_VTAG : SKIP_VTAGS);
               end
            end // case: SKIP_VTAGS_1

            LATCH_MY_VTAG: begin
               if(in_wr) begin
                  word_count          <= word_count + 1'b1;
                  icing_vtag_is_me    <= 1;
                  icing_va_0_vld      <= 1;
                  state               <= LATCH_MY_VTAG_1;
               end
            end

            LATCH_MY_VTAG_1: begin
               if(in_wr) begin
                  icing_va_1_vld      <= 1;
                  icing_vb_vld        <= 1;
                  icing_vtag_is_me    <= 1;
                  count               <= count + 1'b1;
                  state               <= OTHER_VTAG_1;
               end
            end // case: LATCH_MY_VTAG_1

            OTHER_VTAG_1: begin
               if(in_wr) begin
                  icing_va_0_vld    <= 1;
                  state             <= OTHER_VTAG_2;
                  count             <= count + 1'b1;
                  if(count == icing_path_len-1'b1) begin
                     icing_vtag_is_last   <= 1'b1;
                  end
               end
            end

            OTHER_VTAG_2: begin
               if(in_wr) begin
                  icing_va_1_vld    <= 1;
                  icing_vb_vld      <= 1;
                  state             <= OTHER_VTAG_1;
                  if(count == icing_path_len) begin
                     state                 <= PROTO_LEN;
                     icing_vtag_is_last    <= 1'b1;
                  end
               end
            end // case: OTHER_VTAG_2

            PROTO_LEN: begin
               if(in_wr) begin
                  icing_payload_vld   <= 1;
                  icing_proto_vld     <= 1;
                  icing_error_idx_vld <= 1;
                  icing_pkt_len_vld   <= 1;
                  if(in_ctrl != 0) begin
                     icing_payload_is_last    <= 1;
                     word_count               <= 0;
                     state                    <= WAIT_FOR_PKT;
                  end
                  else begin
                     state <= PAYLOAD;
                  end
               end
            end

            PAYLOAD: begin
               if(in_wr) begin
                  icing_payload_vld   <= 1;
                  if(in_ctrl != 0) begin
                     icing_payload_is_last    <= 1;
                     word_count               <= 0;
                     state                    <= WAIT_FOR_PKT;
                  end
               end
            end // case: PAYLOAD

            WAIT_FOR_EOP: begin
               if(in_wr && in_ctrl != 0) begin
                  word_count   <= 0;
                  state        <= WAIT_FOR_PKT;
               end
            end
         endcase // case(state)
      end // else: !if(wait_for_eop)
   end // always @ (posedge clk)
endmodule

module icing_pkt_parser_comb
  #(parameter DATA_WIDTH           = 64,
    parameter CTRL_WIDTH           = DATA_WIDTH/8)
   (// --- interface to data path
    input [0:DATA_WIDTH-1] in_data,
    input [0:CTRL_WIDTH-1] in_ctrl,
    input                  in_wr,

    // --- interface to error
    input                  wait_for_eop,

    // --- parser results
    output reg [3:0]       input_port,
    output reg             input_port_vld,
    output reg [7:0]       word_length,
    output reg             word_length_vld,
    output reg [10:0]      byte_length,
    output reg             byte_length_vld,
    output reg [0:47]      dst_mac,
    output reg             dst_mac_vld,
    output reg [0:47]      src_mac,
    output reg             src_mac_vld,
    output reg [0:15]      eth_type,
    output reg             eth_type_vld,
    output reg [7:0]       icing_version,
    output reg             icing_version_vld,
    output reg [7:0]       icing_path_len,
    output reg             icing_path_len_vld,
    output reg [7:0]       icing_path_idx,
    output reg             icing_path_idx_vld,
    output reg             icing_last_hop_reached,
    output reg [0:55]      icing_iv,
    output reg             icing_iv_vld,
    output reg [0:63]      icing_realm_id,
    output reg             icing_realm_id_0_vld,
    output reg             icing_realm_id_1_vld,
    output reg             icing_realm_id_2_vld,
    output reg [0:31]      icing_subrealm,
    output reg             icing_subrealm_vld,
    output reg             icing_realm_is_first,
    output reg             icing_realm_is_last,
    output reg             icing_realm_is_me,
    output reg             icing_realm_is_b4_me,
    output reg             icing_realm_is_nxt_hop,
    output reg [0:63]      icing_va,
    output reg             icing_va_0_vld,
    output reg             icing_va_1_vld,
    output reg [0:31]      icing_vb,
    output reg             icing_vb_vld,
    output reg             icing_vtag_is_me,
    output reg             icing_vtag_is_last,
    output reg [0:7]       icing_proto,
    output reg             icing_proto_vld,
    output reg [0:15]      icing_pkt_len,
    output reg             icing_pkt_len_vld,
    output reg [0:7]       icing_error_idx,
    output reg             icing_error_idx_vld,
    output reg [0:63]      icing_payload,
    output reg             icing_payload_is_last,
    output reg             icing_payload_vld,

    output reg [7:0]       word_count,
    output                 word_count_is_my_realm_id,
    output                 word_count_is_my_vtag,
    
    // --- Misc
    input                  clk,
    input                  reset);

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_PKT      = 1,
              ETH_HDR_2         = 2**1,
              ICING_INIT_VEC    = 2**2,
              READ_PATH_1       = 2**3,
              READ_PATH_2       = 2**4,
              READ_PATH_3       = 2**5,
              SKIP_VTAGS        = 2**6,
              SKIP_VTAGS_1      = 2**7,
              LATCH_MY_VTAG     = 2**8,
              LATCH_MY_VTAG_1   = 2**9,
              OTHER_VTAG_1      = 2**10,
              OTHER_VTAG_2      = 2**11,
              PROTO_LEN         = 2**12,
              PAYLOAD           = 2**13,
              WAIT_FOR_EOP      = 2**14;
   localparam NUM_STATES = 15;
   //------------------------ Wires/Regs -----------------------------
   reg [0:15]              temp_reg, temp_reg_nxt;
   reg [0:5]               count, count_nxt;
   reg [0:5]               path_idx_reg, path_len_reg;
   reg [0:NUM_STATES-1]    state, state_nxt;
   reg [7:0]               word_count_nxt;
   
   //-------------------------- Logic --------------------------------
   assign word_count_is_my_vtag     = icing_vtag_is_me && icing_va_0_vld;
   assign word_count_is_my_realm_id = icing_realm_is_me && icing_realm_id_0_vld;

   always @(*) begin
      byte_length                 = in_data[53:63];
      byte_length_vld             = 0;
      input_port                  = in_data[45:47];
      input_port_vld              = 0;
      word_length                 = in_data[24:31];
      word_length_vld             = 0;
      dst_mac                     = in_data[0:47];
      dst_mac_vld                 = 0;
      src_mac                     = {temp_reg[0:15], in_data[0:31]};
      src_mac_vld                 = 0;
      eth_type                    = in_data[32:47];
      eth_type_vld                = 0;
      icing_version               = in_data[48:55];
      icing_version_vld           = 0;
      icing_path_len              = in_data[56:63];
      icing_path_len_vld          = 0;
      icing_path_idx              = in_data[0:7];
      icing_path_idx_vld          = 0;
      icing_last_hop_reached      = (path_idx_reg == path_len_reg - 1'b1);
      icing_iv                    = in_data[8:63];
      icing_iv_vld                = 0;
      icing_realm_id              = in_data;
      icing_realm_id_0_vld        = 0;
      icing_realm_id_1_vld        = 0;
      icing_realm_id_2_vld        = 0;
      icing_subrealm              = in_data[32:63];
      icing_subrealm_vld          = 0;
      icing_realm_is_first        = 0;
      icing_realm_is_b4_me        = 0;
      icing_realm_is_me           = 0;
      icing_realm_is_nxt_hop      = 0;
      icing_realm_is_last         = 0;
      icing_va                    = in_data;
      icing_va_0_vld              = 0;
      icing_va_1_vld              = 0;
      icing_vb                    = in_data[32:63];
      icing_vb_vld                = 0;
      icing_vtag_is_me            = 0;
      icing_vtag_is_last          = 0;
      icing_proto                 = in_data[0:7];
      icing_proto_vld             = 0;
      icing_pkt_len               = in_data[8:23];
      icing_pkt_len_vld           = 0;
      icing_error_idx             = in_data[24:31];
      icing_error_idx_vld         = 0;
      icing_payload               = in_data;
      icing_payload_vld           = 0;
      icing_payload_is_last       = 0;

      word_count_nxt              = word_count;
      temp_reg_nxt                = temp_reg;
      state_nxt                   = state;
      count_nxt                   = count;
      
      case (state)
         /* skip module headers,
          * latch len from IOQ hdr,
          * rd dst_mac
          */
         WAIT_FOR_PKT: begin
            if(in_wr) begin
               word_count_nxt = word_count + 1'b1;
               if(in_ctrl == 8'hFF) begin
                  byte_length_vld   = 1;
                  input_port_vld    = 1;
                  word_length_vld   = 1;
               end
               /* {DstMac, SrcMac[0:15]} */
               else if(in_ctrl == 0) begin
                  dst_mac_vld    = 1;
                  temp_reg_nxt   = in_data[48:63];
                  state_nxt      = ETH_HDR_2;
               end
            end // if (in_wr)
         end // case: WAIT_FOR_PKT

         /* src mac, eth type,
          * ICING version, path len */
         ETH_HDR_2: begin
            if(in_wr) begin
               word_count_nxt       = word_count + 1'b1;
               src_mac_vld          = 1;
               eth_type_vld         = 1;
               icing_version_vld    = 1;
               icing_path_len_vld   = 1;
               state_nxt            = ICING_INIT_VEC;
            end // if (in_wr)
         end // case: ETH_HDR_2

         /* path idx, init vec */
         ICING_INIT_VEC: begin
            if(in_wr) begin
               word_count_nxt       = word_count + 1'b1;
               icing_path_idx_vld   = 1;
               icing_iv_vld         = 1;

               count_nxt            = 0;
               state_nxt            = READ_PATH_1;
            end // if (in_wr)
         end // case: ICING_INIT_VEC
         
         /* Latch first 64 bits of realm_id */
         READ_PATH_1: begin
            if(in_wr) begin
               word_count_nxt         = word_count + 1'b1;
               icing_realm_id_0_vld   = 1'b1;
               state_nxt              = READ_PATH_2;

               if(count == 0) begin
                  icing_realm_is_first   = 1'b1;
               end
               if(count == path_idx_reg - 1'b1) begin
                  icing_realm_is_b4_me   = 1'b1;
               end 
               if(count == path_idx_reg) begin
                  icing_realm_is_me      = 1'b1;
               end 
               if(count == path_idx_reg + 1'b1) begin
                  icing_realm_is_nxt_hop = 1'b1;
               end
               if(count == path_len_reg - 1'b1) begin
                  icing_realm_is_last    = 1'b1;
               end
            end
         end

         READ_PATH_2: begin
            if(in_wr) begin
               word_count_nxt         = word_count + 1'b1;
               icing_realm_id_1_vld   = 1'b1;
               state_nxt              = READ_PATH_3;

               if(count == 0) begin
                  icing_realm_is_first = 1'b1;
               end
               if(count == path_idx_reg - 1'b1) begin
                  icing_realm_is_b4_me = 1'b1;
               end 
               if(count == path_idx_reg) begin
                  icing_realm_is_me = 1'b1;
               end 
               if(count == path_idx_reg + 1'b1) begin
                  icing_realm_is_nxt_hop = 1'b1;
               end
               if(count == path_len_reg - 1'b1) begin
                  icing_realm_is_last = 1'b1;
               end
            end
         end // case: READ_PATH_2

         READ_PATH_3: begin
            if(in_wr) begin
               word_count_nxt         = word_count + 1'b1;
               icing_realm_id_2_vld   = 1'b1;
               icing_subrealm_vld     = 1'b1;
               
               if(count == 0) begin
                  icing_realm_is_first = 1'b1;
               end
               if(count == path_idx_reg - 1'b1) begin
                  icing_realm_is_b4_me = 1'b1;
               end 
               if(count == path_idx_reg) begin
                  icing_realm_is_me = 1'b1;
               end 
               if(count == path_idx_reg + 1'b1) begin
                  icing_realm_is_nxt_hop = 1'b1;
               end
               
               if(count == path_len_reg - 1'b1) begin
                  state_nxt = (path_idx_reg == 1) ? LATCH_MY_VTAG : SKIP_VTAGS;
                  icing_realm_is_last = 1'b1;
                  count_nxt = 1;
               end
               else begin
                  state_nxt = READ_PATH_1;
                  count_nxt = count + 1'b1;
               end
            end // if (in_wr)
         end // case: READ_PATH_3
         
         /* Skip 1st 64bits of old VTAG */
         SKIP_VTAGS: begin
            if(in_wr) begin
               word_count_nxt   = word_count + 1'b1;
               count_nxt        = count + 1'b1;
               state_nxt        = SKIP_VTAGS_1;
            end
         end

         /* Skip 2nd 64bits of old VTAG */
         SKIP_VTAGS_1: begin
            if(in_wr) begin
               word_count_nxt = word_count + 1'b1;
               state_nxt = ((count == path_idx_reg) 
                            ? LATCH_MY_VTAG : SKIP_VTAGS);
            end
         end // case: SKIP_VTAGS_1

         LATCH_MY_VTAG: begin
            if(in_wr) begin
               word_count_nxt   = word_count + 1'b1;
               icing_vtag_is_me = 1;
               icing_va_0_vld   = 1;
               state_nxt        = LATCH_MY_VTAG_1;
            end
         end

         LATCH_MY_VTAG_1: begin
            if(in_wr) begin
               icing_va_1_vld     = 1;
               icing_vb_vld       = 1;
               icing_vtag_is_me   = 1;
               count_nxt          = count + 1'b1;
               state_nxt          = OTHER_VTAG_1;
            end
         end // case: LATCH_MY_VTAG_1

         OTHER_VTAG_1: begin
            if(in_wr) begin
               icing_va_0_vld   = 1;
               state_nxt        = OTHER_VTAG_2;
               count_nxt        = count + 1'b1;
               if(count == path_len_reg-1'b1) begin
                  icing_vtag_is_last   = 1'b1;
               end
            end
         end

         OTHER_VTAG_2: begin
            if(in_wr) begin
               icing_va_1_vld   = 1;
               icing_vb_vld     = 1;
               state_nxt        = OTHER_VTAG_1;
               if(count == path_len_reg) begin
                  state_nxt            = PROTO_LEN;
                  icing_vtag_is_last   = 1'b1;
               end
            end
         end // case: OTHER_VTAG_2

         PROTO_LEN: begin
            if(in_wr) begin
               icing_payload_vld   = 1;
               icing_proto_vld     = 1;
               icing_error_idx_vld = 1;
               icing_pkt_len_vld   = 1;
               if(in_ctrl != 0) begin
                  icing_payload_is_last   = 1;
                  word_count_nxt          = 0;
                  state_nxt               = WAIT_FOR_PKT;
               end
               else begin
                  state_nxt = PAYLOAD;
               end
            end
         end

         PAYLOAD: begin
            if(in_wr) begin
               icing_payload_vld   = 1;
               if(in_ctrl != 0) begin
                  icing_payload_is_last   = 1;
                  word_count_nxt          = 0;
                  state_nxt               = WAIT_FOR_PKT;
               end
            end
         end // case: PAYLOAD

         WAIT_FOR_EOP: begin
            if(in_wr && in_ctrl != 0) begin
               word_count_nxt   = 0;
               state_nxt        = WAIT_FOR_PKT;
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         state <= WAIT_FOR_PKT;
      end else if(wait_for_eop) begin
         state <= WAIT_FOR_EOP;
      end else begin
         state <= state_nxt;
      end
      count       <= count_nxt;
      word_count  <= reset ? 0 : word_count_nxt;
      temp_reg    <= temp_reg_nxt;
      if(icing_path_len_vld) path_len_reg <= icing_path_len;
      if(icing_path_idx_vld) path_idx_reg <= icing_path_idx;
   end
   
endmodule

// synthesis translate_off
module icing_pkt_parser_tester();

   localparam DATA_WIDTH = 64,
              CTRL_WIDTH = 8;
   
    // --- interface to data path
   reg [0:DATA_WIDTH-1] in_data = 0;
   reg [0:CTRL_WIDTH-1] in_ctrl = 1;
   reg                  in_wr   = 0;

   // --- parser results
   wire [3:0]           input_port;
   wire                 input_port_vld;
   wire [7:0]           word_length;
   wire                 word_length_vld;
   wire [10:0]          byte_length;
   wire                 byte_length_vld;
   wire [0:47]          dst_mac;
   wire                 dst_mac_vld;
   wire [0:47]          src_mac;
   wire                 src_mac_vld;
   wire [0:15]          eth_type;
   wire                 eth_type_vld;
   wire [7:0]           icing_version;
   wire                 icing_version_vld;
   wire [7:0]           icing_path_len;
   wire                 icing_path_len_vld;
   wire [7:0]           icing_path_idx;
   wire                 icing_path_idx_vld;
   wire [0:55]          icing_iv;
   wire                 icing_iv_vld;
   wire [0:63]          icing_realm_id;
   wire                 icing_realm_id_0_vld;
   wire                 icing_realm_id_1_vld;
   wire                 icing_realm_id_2_vld;
   wire [0:31]          icing_subrealm;
   wire                 icing_subrealm_vld;
   wire                 icing_realm_is_last;
   wire                 icing_realm_is_me;
   wire                 icing_realm_is_nxt_hop;
   wire [0:63]          icing_va;
   wire                 icing_va_0_vld;
   wire                 icing_va_1_vld;
   wire [0:31]          icing_vb;
   wire                 icing_vb_vld;
   wire                 icing_vtag_is_me;
   wire                 icing_vtag_is_last;
   wire [0:7]           icing_proto;
   wire                 icing_proto_vld;
   wire [0:15]          icing_pkt_len;
   wire                 icing_pkt_len_vld;
   wire [0:63]          icing_payload;
   wire                 icing_payload_vld;
   wire                 icing_payload_is_last;
   
   // --- Misc
   reg                  clk = 0;
   reg                  reset = 1;

   always #4 clk = !clk;

   localparam           PATH_LEN = 3;
   localparam           PATH_WORDS = PATH_LEN * 3;
   localparam           PATH_BITS = PATH_WORDS * 64;
   localparam           V_WORDS = (PATH_LEN - 1) * 2;
   localparam           V_BITS = V_WORDS * 64;
   localparam           OVERHEAD_BITS = (8 + 14 + 3 + 7)*8 + PATH_BITS + V_BITS;
   localparam           PAYLOAD_WORDS = 2;
   localparam           PAYLOAD_BITS = PAYLOAD_WORDS * 64;
   localparam           PKT_BITS = OVERHEAD_BITS + PAYLOAD_BITS;
   localparam           PKT_WORDS = PKT_BITS/DATA_WIDTH;

   wire [3:0]           in_input_port = 1;
   wire [7:0]           in_word_length = 2;
   wire [10:0]          in_byte_length = 3;
   wire [0:47]          in_dst_mac = 48'h4;
   wire [0:47]          in_src_mac = 48'h5;
   wire [0:15]          in_eth_type = 6;
   wire [7:0]           in_icing_version = 7;
   wire [7:0]           in_icing_path_len = 3;
   wire [7:0]           in_icing_path_idx = 1;
   wire [0:55]          in_icing_iv = 56'h10;
   wire [0:PATH_BITS-1] in_icing_path = {64'h11, 64'h12, 32'h13, 32'h20,
                                         64'h14, 64'h15, 32'h16, 32'h21,
                                         64'h17, 64'h18, 32'h19, 32'h22};
   wire [0:V_BITS-1]    in_icing_v = {64'h23, 32'h24, 32'h25,
                                      64'h26, 32'h27, 32'h28};
   wire [0:7]           in_icing_proto = 29;
   wire [0:15]          in_icing_pkt_len = 30;
   wire [0:PAYLOAD_BITS-1] in_icing_payload = {in_icing_proto, in_icing_pkt_len, 40'h31,
                                            64'h 32};

   wire [0:PKT_BITS - 1]   packet = {16'hAA, 8'hBB, in_word_length, 12'hCCC, in_input_port, 5'h0, in_byte_length,
                                  in_dst_mac, in_src_mac, 
                                  in_eth_type, in_icing_version, in_icing_path_len,
                                  in_icing_path_idx, in_icing_iv,
                                  in_icing_path,
                                  in_icing_v,
                                  in_icing_payload};

   wire [0:63]          packet_words[0:PKT_WORDS-1];
   wire [0:63]          in_icing_path_words[0:PATH_WORDS-1];
   wire [0:63]          in_icing_v_words[0:V_WORDS-1];
   wire [0:63]          in_icing_payload_words[0:PAYLOAD_WORDS-1];
   generate
      genvar            i;
      for(i=0; i<PKT_WORDS; i=i+1) begin
         assign packet_words[i] = packet[i*64 +: 64];
      end
      for(i=0; i<PATH_WORDS; i=i+1) begin
         assign in_icing_path_words[i] = in_icing_path[i*64 +: 64];
      end
      for(i=0; i<V_WORDS; i=i+1) begin
         assign in_icing_v_words[i] = in_icing_v[i*64 +: 64];
      end
      for(i=0; i<PAYLOAD_WORDS; i=i+1) begin
         assign in_icing_payload_words[i] = in_icing_payload[i*64 +: 64];
      end
   endgenerate

   /* Do the input side */
   integer count = 0;
   initial begin
      repeat(3) begin
         @(posedge clk) begin end
      end
      @(posedge clk) begin
         reset = 0;
      end
      repeat(PKT_WORDS) begin
         @(posedge clk) begin
            in_wr   = 1;
            in_data = packet_words[count];
            if(count == 0) in_ctrl = 8'hff;
            else if(count == PKT_WORDS-1) in_ctrl = 8'h1;
            else in_ctrl = 0;
            count = count + 1;
         end
      end
      @(posedge clk) begin
         in_wr = 0;
      end
      @(posedge clk) begin
         $display("Simulation Done. If no ERROR then success");
      end
   end // initial begin

   /* Do the output side */
   integer out_count_p = 0;
   integer out_count_v = 0;
   integer out_count_pld = 0;
   always @(posedge clk) begin
      out_count_p <= out_count_p + (icing_realm_id_0_vld
                                    || icing_realm_id_1_vld
                                    || icing_realm_id_2_vld);
      out_count_v <= out_count_v + (icing_va_0_vld || icing_va_1_vld);
      out_count_pld <= out_count_pld + icing_payload_vld;
      
      if(input_port_vld && (in_input_port !== input_port)) begin
         $display("ERROR parsing input_port");
      end
      if(byte_length_vld && (in_byte_length !== byte_length)) begin
         $display("ERROR parsing byte_length");
      end
      if(word_length_vld && (in_word_length !== word_length)) begin
         $display("ERROR parsing word_length");
      end
      if(dst_mac_vld && (in_dst_mac !== dst_mac)) begin
         $display("ERROR parsing dst_mac");
      end
      if(src_mac_vld && (in_src_mac !== src_mac)) begin
         $display("ERROR parsing src_mac");
      end
      if(eth_type_vld && (in_eth_type !== eth_type)) begin
         $display("ERROR parsing eth_type");
      end
      if(icing_version_vld && (in_icing_version !== icing_version)) begin
         $display("ERROR parsing icing_version");
      end
      if(icing_path_len_vld && (in_icing_path_len !== icing_path_len)) begin
         $display("ERROR parsing icing_path_len");
      end
      if(icing_path_idx_vld && (in_icing_path_idx !== icing_path_idx)) begin
         $display("ERROR parsing icing_path_idx");
      end
      if(icing_iv_vld && (in_icing_iv !== icing_iv)) begin
         $display("ERROR parsing icing_iv");
      end

      if(icing_realm_id_0_vld && (in_icing_path_words[out_count_p] !== icing_realm_id)) begin
         $display("ERROR parsing realm_id_0");
      end
      if(icing_realm_id_1_vld && (in_icing_path_words[out_count_p] !== icing_realm_id)) begin
         $display("ERROR parsing realm_id_1");
      end
      if(icing_realm_id_2_vld && (in_icing_path_words[out_count_p] !== icing_realm_id)) begin
         $display("ERROR parsing realm_id_2");
      end
      if(icing_subrealm_vld && (in_icing_path_words[out_count_p][32:63] !== icing_subrealm)) begin
         $display("ERROR parsing subrealm");
      end

      if(icing_va_0_vld && (in_icing_v_words[out_count_v] !== icing_va)) begin
         $display("ERROR parsing va_0");
      end
      if(icing_va_1_vld && (in_icing_v_words[out_count_v] !== icing_va)) begin
         $display("ERROR parsing va_1");
      end
      if(icing_vb_vld && (in_icing_v_words[out_count_v][32:63] !== icing_vb)) begin
         $display("ERROR parsing va_0");
      end

      if(icing_proto_vld && (in_icing_proto !== icing_proto)) begin
         $display("ERROR parsing icing_proto");
      end
      if(icing_pkt_len_vld && (in_icing_pkt_len !== icing_pkt_len)) begin
         $display("ERROR parsing icing_pkt_len");
      end

      if(icing_payload_vld && (in_icing_payload_words[out_count_pld] !== icing_payload)) begin
         $display("ERROR parsing icing_payload");
      end

      if(icing_subrealm_vld && (out_count_p/3 == in_icing_path_idx) && !icing_realm_is_me) begin
         $display("ERROR parsing icing_realm_is_me");
      end
      if(icing_subrealm_vld && (out_count_p/3 == in_icing_path_idx+1) && !icing_realm_is_nxt_hop) begin
         $display("ERROR parsing icing_realm_is_nxt_hop");
      end
      if(icing_subrealm_vld && (out_count_p/3 == in_icing_path_len-1) && !icing_realm_is_last) begin
         $display("ERROR parsing icing_realm_is_last");
      end
      
      if(icing_vb_vld && (out_count_v/2 == in_icing_path_idx-1) && !icing_vtag_is_me) begin
         $display("ERROR parsing icing_vtag_is_me");
      end
      if(icing_vb_vld && (out_count_v/2 == in_icing_path_len-2) && !icing_vtag_is_last) begin
         $display("ERROR parsing icing_vtag_is_last");
      end

      if(icing_payload_vld && (out_count_pld == 1) && !icing_payload_is_last) begin
         $display("ERROR parsing icing_payload_is_last");
      end
   end // always @ (posedge clk)
   
   icing_pkt_parser parser
     (/*AUTOINST*/
      // Outputs
      .input_port                       (input_port[3:0]),
      .input_port_vld                   (input_port_vld),
      .word_length                      (word_length[7:0]),
      .word_length_vld                  (word_length_vld),
      .byte_length                      (byte_length[10:0]),
      .byte_length_vld                  (byte_length_vld),
      .dst_mac                          (dst_mac[0:47]),
      .dst_mac_vld                      (dst_mac_vld),
      .src_mac                          (src_mac[0:47]),
      .src_mac_vld                      (src_mac_vld),
      .eth_type                         (eth_type[0:15]),
      .eth_type_vld                     (eth_type_vld),
      .icing_version                    (icing_version[7:0]),
      .icing_version_vld                (icing_version_vld),
      .icing_path_len                   (icing_path_len[7:0]),
      .icing_path_len_vld               (icing_path_len_vld),
      .icing_path_idx                   (icing_path_idx[7:0]),
      .icing_path_idx_vld               (icing_path_idx_vld),
      .icing_iv                         (icing_iv[0:55]),
      .icing_iv_vld                     (icing_iv_vld),
      .icing_realm_id                   (icing_realm_id[0:63]),
      .icing_realm_id_0_vld             (icing_realm_id_0_vld),
      .icing_realm_id_1_vld             (icing_realm_id_1_vld),
      .icing_realm_id_2_vld             (icing_realm_id_2_vld),
      .icing_subrealm                   (icing_subrealm[0:31]),
      .icing_subrealm_vld               (icing_subrealm_vld),
      .icing_realm_is_last              (icing_realm_is_last),
      .icing_realm_is_me                (icing_realm_is_me),
      .icing_realm_is_nxt_hop           (icing_realm_is_nxt_hop),
      .icing_va                         (icing_va[0:63]),
      .icing_va_0_vld                   (icing_va_0_vld),
      .icing_va_1_vld                   (icing_va_1_vld),
      .icing_vb                         (icing_vb[0:31]),
      .icing_vb_vld                     (icing_vb_vld),
      .icing_vtag_is_me                 (icing_vtag_is_me),
      .icing_vtag_is_last               (icing_vtag_is_last),
      .icing_proto                      (icing_proto[0:7]),
      .icing_proto_vld                  (icing_proto_vld),
      .icing_pkt_len                    (icing_pkt_len[0:15]),
      .icing_pkt_len_vld                (icing_pkt_len_vld),
      .icing_payload                    (icing_payload[0:63]),
      .icing_payload_vld                (icing_payload_vld),
      .icing_payload_is_last            (icing_payload_is_last),
      // Inputs
      .in_data                          (in_data),
      .in_ctrl                          (in_ctrl[0:CTRL_WIDTH-1]),
      .in_wr                            (in_wr),
      .clk                              (clk),
      .reset                            (reset));
endmodule // icing_pkt_parser_tester
// synthesis translate_on
