/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: lookup_shared_keys.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Lookup the realms in the SRAM and gets the shared keys and next hop info
 * 
 ******************************************************************************/

`timescale 1ns/1ps

`ifndef CPCI_NF2_DATA_WIDTH
 `include "global_defines.v"
 `include "NF_2.1_defines.v"
 `include "udp_defines.v"
 `include "icing_include.v"
`endif

module lookup_shared_keys
  #(parameter UDP_REG_SRC_WIDTH = 3)
  (// --- Interface to packet parser
   input      [0:63]  icing_realm_id,
   input              icing_realm_id_0_vld,
   input              icing_realm_id_1_vld,
   input              icing_realm_id_2_vld,
   input              icing_realm_is_b4_me,
   input              icing_realm_is_first,
   input              icing_realm_is_me,
   input              icing_realm_is_nxt_hop,
   input              icing_realm_is_last,
   input              icing_last_hop_reached,

   // --- Interface to flow control
   output             pause,

   // --- Interface to SRAM
   output reg [18:0]  sram_rd_addr,
   output reg         sram_rd_req,
   input      [0:63]  sram_rd_data,
   input              sram_rd_vld,
   input              sram_rd_ack,

   // --- Results
   output     [0:127] shared_dh_key,
   output             my_realm_id_is_smaller,
   output             shared_dh_key_missing,
   output             shared_dh_key_vld,
   input              rd_shared_key,
   output             shared_dh_key_first,
   output             shared_dh_key_last,
   output             shared_dh_key_b4_me,

   output     [0:47]  next_hop_mac,
   output     [7:0]   output_port,
   output             next_hop_info_vld,
   input              rd_nxthop_info,

   // --- register interface
   input                                  reg_req_in,
   input                                  reg_ack_in,
   input                                  reg_rd_wr_L_in,
   input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
   input  [0:`CPCI_NF2_DATA_WIDTH-1]      reg_data_in,
   input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,
   
   output                                 reg_req_out,
   output                                 reg_ack_out,
   output                                 reg_rd_wr_L_out,
   output     [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
   output     [0:`CPCI_NF2_DATA_WIDTH-1]  reg_data_out,
   output     [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,
   
   // --- Misc
   input              clk,
   input              reset);

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_HASH = 0,
              RD_ALL_WORDS  = 1;

   localparam CHECK_ID_0 = 1,
              CHECK_ID_1 = 2,
              CHECK_ID_2 = 4,
              RD_SHARED_KEY_0 = 8,
              RD_SHARED_KEY_1 = 16,
              RD_SHARED_KEY_2 = 32,
              RD_SHARED_KEY_3 = 64,
              RD_NXT_HOP_INFO = 128;
   
   //------------------------ Wires/Regs -----------------------------
   wire [0:63]                            id, icing_realm_id_masked;
   wire [0:`SRAM_SKEY_NXTHOP_IDX_SIZE-1]  hash_in, hash_out;
   wire [0:`SRAM_SKEY_NXTHOP_IDX_SIZE-1]  hash_nxt;
   reg [0:`SRAM_SKEY_NXTHOP_IDX_SIZE-1]   hash_reg;
   wire [0:511]                           hash_vector;
   reg                                    vld_0_d1, vld_1_d1, hash_fifo_wr_en;
   
   reg                                    req_state, req_state_nxt;
   reg [`SRAM_SKEY_NXTHOP_ENTRY_BITS-1:0] count_nxt, count;
   reg                                    hash_fifo_rd_en;
   
   reg                                    id_fifo_rd_en;
   reg [7:0]                              state, state_nxt;
   
   reg                                    key0_fifo_wr_en, key1_fifo_wr_en, 
                                          gt_fifo_wr_en, nxthop_fifo_wr_en;
   reg                                    match, match_nxt;

   wire [0:`SRAM_SKEY_NXTHOP_IDX_SIZE-1]  words[0:7][0:3];
   //------------------------- Modules -------------------------------

   /* This fifo should store the id until the SRAM outputs the id
    * so we can check it. The latency for a check is the actual latency (4 cycles)
    * + latency if the PoC key lookup happens at the same time (4 cycles). For a min size pkt,
    * there will be a poc key lookup for two skey lookups. Add 5 for uhf, total = 13 cycles.
    * Fifo should be at least 13 deep. */
   fallthrough_small_fifo 
     #(.WIDTH(65),
       .MAX_DEPTH_BITS(4),
       .PROG_FULL_THRESHOLD(14))
      realm_id_fifo
        (.din           ({icing_realm_id, icing_realm_is_nxt_hop}),
         .wr_en         (id_fifo_wr_en),
         .rd_en         (id_fifo_rd_en),
         .dout          ({id, is_nxt_hop}),
         .full          (id_fifo_full),
         .prog_full     (id_fifo_almost_full),
         .nearly_full   (),
         .empty         (id_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   /* This needs to be big enough to handle a lookup interruption
    * due to poc key lookup, so for every 2 skey (min path), there might
    * be a pkey lookup which take 4 cycles. The fifo is read after the lookup
    * is done which is 6 cycles. In 10=6+4 cycles, we might get 4 ids
    * id. So this should be at least 4 entries. CALCULATION WRONG.*/
   fallthrough_small_fifo 
     #(.WIDTH(`SRAM_SKEY_NXTHOP_IDX_SIZE),
       .MAX_DEPTH_BITS(3))
      hash_fifo
        (.din           (hash_in),
         .wr_en         (hash_fifo_wr_en),
         .rd_en         (hash_fifo_rd_en),
         .dout          (hash_out),
         .full          (),
         .prog_full     (),
         .nearly_full   (hash_fifo_nearly_full),
         .empty         (hash_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   /* This needs to be large enough to store status info
    * for each key until it is read. The keys are read by the check_vtag and
    * calc_nxt_vtags, which read them after each PRF2 op. It should be big enough
    * to absorb all keys so as not to stop the input to the hash. i.e. it needs to
    * store around 2 max pkts path len.*/
   fallthrough_small_fifo 
     #(.WIDTH(3),
       .MAX_DEPTH_BITS(6),
       .PROG_FULL_THRESHOLD(62))
      status_fifo
        (.din           ({icing_realm_is_first, 
                          icing_realm_is_b4_me,
                          last_key_lookup}),
         .wr_en         (icing_realm_id_0_vld 
                         && !icing_realm_is_me),
         .rd_en         (rd_shared_key),
         .dout          ({shared_dh_key_first,
                          shared_dh_key_b4_me,
                          shared_dh_key_last}),
         .full          (),
         .prog_full     (status_fifo_almost_full),
         .nearly_full   (),
         .empty         (),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(64),
       .MAX_DEPTH_BITS(6))
      key_fifo_0
        (.din           (sram_rd_data),
         .wr_en         (key0_fifo_wr_en),
         .rd_en         (rd_shared_key),
         .dout          (shared_dh_key[0:63]),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (key0_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(64),
       .MAX_DEPTH_BITS(6))
      key_fifo_1
        (.din           (sram_rd_data),
         .wr_en         (key1_fifo_wr_en),
         .rd_en         (rd_shared_key),
         .dout          (shared_dh_key[64:127]),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (key1_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(2),
       .MAX_DEPTH_BITS(6))
      gt_fifo
        (.din           ({sram_rd_data[7], !match}),
         .wr_en         (gt_fifo_wr_en),
         .rd_en         (rd_shared_key),
         .dout          ({my_realm_id_is_smaller, shared_dh_key_missing}),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (gt_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   fallthrough_small_fifo 
     #(.WIDTH(56),
       .MAX_DEPTH_BITS(5))
      nxthop_fifo
        (.din           (sram_rd_data[8:63]),
         .wr_en         (nxthop_fifo_wr_en),
         .rd_en         (rd_nxthop_info),
         .dout          ({output_port, next_hop_mac}),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (nxthop_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   /* We implement Carter and Wegman's H_3 class of Universal hashing 
    * functions. See "Efficient Hardware Hashing Functions for High Performance
    * Computers" by Ramakrishna, M.V.; Fu, E.; Bahcekapili, E.; in IEEE
    * Transaction on Computers vol 46, issue 12, Dec 1997 p. 1378-1381
    * for a succinct description and more references.
    * 
    * Here we generalize the implementation so that XORs are not done on
    * a bit-by-bit basis but rather on a byte by byte basis.
    * 
    * Basically each byte position in input message has its own random mapping
    * from 8-bits->16 bits. We take all bytes in the input message to hash,
    * apply the mappings on them, and XOR all the results.
    * 
    * In our case, the input is 160bits = 20bytes, arriving as 8, 8, 4 bytes.
    * Our mappings are stored in BRAM. At each cycle, we read the mappings
    * from BRAM.
    * The output from BRAM is 64 bits of which we need 16 bits in different
    * locations for every set of input bytes. */
   mmapped_ram_8x256x64
     #(.UDP_REG_SRC_WIDTH  (UDP_REG_SRC_WIDTH),
       .REG_BLOCK_TAG      (`ICING_UHF_BLOCK_TAG),
       .REG_ADDR_WIDTH     (`ICING_UHF_REG_ADDR_WIDTH))
     mmapped_ram
     (// --- register interface
      .reg_req_in         (reg_req_in),
      .reg_ack_in         (reg_ack_in),
      .reg_rd_wr_L_in     (reg_rd_wr_L_in),
      .reg_addr_in        (reg_addr_in),
      .reg_data_in        (reg_data_in),
      .reg_src_in         (reg_src_in),
      
      .reg_req_out        (reg_req_out),
      .reg_ack_out        (reg_ack_out),
      .reg_rd_wr_L_out    (reg_rd_wr_L_out),
      .reg_addr_out       (reg_addr_out),
      .reg_data_out       (reg_data_out),
      .reg_src_out        (reg_src_out),
      
      // --- Logic interface
      .ram_data_out       (hash_vector),
      .ram_addr           (icing_realm_id_masked),
      .ram_rd_en          (ram_rd_en),
      
      .clk                (clk),
      .reset              (reset)
      );
   //-------------------------- Logic --------------------------------
   /*********************************************************************
    * info is available
    ********************************************************************/
   assign shared_dh_key_vld = !gt_fifo_empty;
   assign next_hop_info_vld = !nxthop_fifo_empty;
   
   /*********************************************************************
    * Set the last 32 bits of the icing_realm id in the last word
    ********************************************************************/
   assign icing_realm_id_masked = (icing_realm_id_2_vld
                                   ? {icing_realm_id[0:31],32'h0}
                                   : icing_realm_id);

   /*********************************************************************
    * Check if doing last lookup
    ********************************************************************/
   assign last_key_lookup = (icing_realm_is_last 
                             || (icing_realm_is_b4_me 
                                 && icing_last_hop_reached));
          
   /*********************************************************************
    * Handle flow control
    ********************************************************************/
   assign pause = id_fifo_almost_full | status_fifo_almost_full | hash_fifo_nearly_full;

   /*********************************************************************
    * Write input into id fifo
    ********************************************************************/
   assign id_fifo_wr_en = (icing_realm_id_0_vld | icing_realm_id_1_vld 
                           | icing_realm_id_2_vld) && !icing_realm_is_me;

   /*********************************************************************
    *  generate the hash and write it in the fifo
    ********************************************************************/
   generate
      genvar i,j;
      for(i=0; i<8; i=i+1) begin:gen_word_sel
         wire [0:63] word = hash_vector[i*64 : i*64+63];
         for(j=0; j<4; j=j+1) begin:gen_byte_sel
            assign words[i][j] = word[j*16:j*16+15];
         end
      end
   endgenerate
      
   assign hash_nxt 
     = (vld_0_d1) ? (words[0][0] ^ words[1][0] ^ words[2][0] ^ words[3][0]
                     ^ words[4][0] ^ words[5][0] ^ words[6][0] ^ words[7][0])
       : (vld_1_d1) ? (words[0][1] ^ words[1][1] ^ words[2][1] ^ words[3][1]
                       ^ words[4][1] ^ words[5][1] ^ words[6][1] ^ words[7][1])
         : (words[0][2] ^ words[1][2] ^ words[2][2] ^ words[3][2]
            ^ words[4][2] ^ words[5][2] ^ words[6][2] ^ words[7][2]);

/* -----\/----- EXCLUDED -----\/-----
   // synthesis translate_off
   always @(posedge clk) begin
      if(vld_0_d1) begin
         $display("Hashing 0: %x %x %x %x %x %x %x %x = %x",
                  words[0][0], words[1][0], words[2][0], words[3][0],
                  words[4][0], words[5][0], words[6][0], words[7][0],
                  hash_nxt);
      end
      else if(vld_1_d1) begin
         $display("Hashing 1: %x %x %x %x %x %x %x %x = %x",
                  words[0][1], words[1][1], words[2][1], words[3][1],
                  words[4][1], words[5][1], words[6][1], words[7][1],
                  hash_in);
      end
      else if(hash_fifo_wr_en) begin
         $display("Hashing 2: %x %x %x %x %x %x %x %x = %x",
                  words[0][2], words[1][2], words[2][2], words[3][2],
                  words[4][2], words[5][2], words[6][2], words[7][2],
                  hash_in);
      end
   end // always @ (*)
   // synthesis translate_on
 -----/\----- EXCLUDED -----/\----- */
   
   assign ram_rd_en = (icing_realm_id_0_vld | icing_realm_id_1_vld 
                       | icing_realm_id_2_vld) && !icing_realm_is_me;
   assign hash_in = hash_reg ^ hash_nxt;
   always @(posedge clk) begin
/* -----\/----- EXCLUDED -----\/-----
      // synthesis translate_off
      if (icing_realm_id_0_vld && !icing_realm_is_me) begin
         $display($time, "Doing ID %x", icing_realm_id);
      end
      // synthesis translate_on
 -----/\----- EXCLUDED -----/\----- */
      vld_0_d1 <= reset ? 0 : icing_realm_id_0_vld && !icing_realm_is_me;
      vld_1_d1 <= reset ? 0 : icing_realm_id_1_vld && !icing_realm_is_me;
      hash_fifo_wr_en <= reset ? 0 : icing_realm_id_2_vld && !icing_realm_is_me;
      if(vld_0_d1) hash_reg <= hash_nxt;
      else if(vld_1_d1) hash_reg <= hash_in;
   end // always @ (posedge clk)

   
   /*********************************************************************
    * Generate SRAM lookups
    ********************************************************************/
   always @(*) begin
      sram_rd_addr    = `SRAM_SKEY_NXTHOP_BASE 
                        | {hash_out, `SRAM_SKEY_NXTHOP_ENTRY_BITS'h0};
      req_state_nxt   = req_state;
      sram_rd_req     = 0;
      count_nxt       = count;
      hash_fifo_rd_en = 0;
      case(req_state)
         WAIT_FOR_HASH: begin
            count_nxt      = 0;
            if(!hash_fifo_empty) begin
               req_state_nxt   = RD_ALL_WORDS;
               sram_rd_req     = 1;
               if(sram_rd_ack) begin
                  count_nxt = 1'b1;
               end
            end
         end
         RD_ALL_WORDS: begin
            sram_rd_req    = 1;
            sram_rd_addr[`SRAM_SKEY_NXTHOP_ENTRY_BITS-1:0]
              = count;
            if(sram_rd_ack) begin
               if(count == 5) begin
                  count_nxt = 0;
                  req_state_nxt = WAIT_FOR_HASH;
                  hash_fifo_rd_en = 1;
               end else begin
                  count_nxt = count + 1'b1;
               end
            end
         end // case: RD_ALL_WORDS
      endcase // case(req_state)
   end // always @ (*)

   always @(posedge clk) begin
      req_state    <= reset ? WAIT_FOR_HASH : req_state_nxt;
      count        <= count_nxt;
   end

   
   /*********************************************************************
    * Check the match and output results
    ********************************************************************/
   always @(*) begin
      id_fifo_rd_en       = 0;
      state_nxt           = state;
      match_nxt           = match;
      key0_fifo_wr_en     = 0;
      key1_fifo_wr_en     = 0;
      gt_fifo_wr_en       = 0;
      nxthop_fifo_wr_en   = 0;
      case(state)
         CHECK_ID_0: begin
            if(sram_rd_vld) begin
               id_fifo_rd_en    = 1;
               state_nxt        = CHECK_ID_1;
               match_nxt        = (sram_rd_data === id);
            end // if (sram_rd_vld)
         end
         
         CHECK_ID_1: begin
            if(sram_rd_vld) begin
               id_fifo_rd_en    = 1;
               state_nxt        = CHECK_ID_2;
               match_nxt        = match & (sram_rd_data === id);
            end // if (sram_rd_vld)
         end

         CHECK_ID_2: begin
            if(sram_rd_vld) begin
               state_nxt        = RD_SHARED_KEY_0;
               match_nxt        = match & (sram_rd_data[0:31] === id[0:31]);
            end // if (sram_rd_vld)
         end

         RD_SHARED_KEY_0: begin
            if(sram_rd_vld) begin
               key0_fifo_wr_en   = 1;
               state_nxt         = RD_SHARED_KEY_1;
            end
         end
         
         RD_SHARED_KEY_1: begin
            if(sram_rd_vld) begin
               key1_fifo_wr_en   = 1;
               state_nxt         = RD_NXT_HOP_INFO;
            end
         end

         RD_NXT_HOP_INFO: begin
            if(sram_rd_vld) begin
               /* rd id fifo here instead of earlier
                * because we want to read is_nxt_hop here */
               id_fifo_rd_en           = 1;
               state_nxt               = CHECK_ID_0;
               gt_fifo_wr_en           = 1;
               nxthop_fifo_wr_en       = is_nxt_hop;
            end
         end // case: RD_NXT_HOP_INFO
      endcase // case(state)
   end // always @ (*)
   
   always @(posedge clk) begin
      if(reset) begin
         state            <= CHECK_ID_0;
         match            <= 0;
      end
      else begin
         state            <= state_nxt;
         match            <= match_nxt;
      end // else: !if(reset)
   end // always @ (posedge clk)

   // synthesis translate_off
   integer keys_found_count = 0;
   integer keys_read_count = 0;
   integer nxthop_found_count = 0;
   integer nxthop_read_count = 0;
   integer keys_missing_count = 0;
   reg [7:0] state_d1;
   always @(posedge clk) begin
      if(!reset) begin
         keys_found_count <= keys_found_count + gt_fifo_wr_en;
         keys_read_count <= keys_read_count + rd_shared_key;
         nxthop_found_count <= nxthop_found_count + nxthop_fifo_wr_en;
         nxthop_read_count <= nxthop_read_count + rd_nxthop_info;
         keys_missing_count <= keys_missing_count + (gt_fifo_wr_en & !match);
         state_d1 <= state;
      end
      if(state_d1 != CHECK_ID_0 && state == CHECK_ID_0) begin
         $display($time, " %m Num skeys  found :  %u", keys_found_count);
         $display($time, " %m Num skeys  read  :  %u", keys_read_count);
         $display($time, " %m Num skeys  missed:  %u", keys_missing_count);
         $display($time, " %m Num nxthop found :  %u", nxthop_found_count);
         $display($time, " %m Num nxthop read  :  %u", nxthop_read_count);
      end
   end
   // synthesis translate_on

endmodule // lookup_shared_keys

// synthesis translate_off

module lookup_skey_tester();
   localparam REG_ADDR_WIDTH = `ICING_UHF_REG_ADDR_WIDTH;
   localparam REG_BLOCK_TAG = `ICING_UHF_BLOCK_TAG;
   localparam NUM_REALMS = 32;

   reg                                  reg_req_in = 0;
   reg                                  reg_ack_in = 0;
   reg                                  reg_rd_wr_L_in = 1;
   reg [`UDP_REG_ADDR_WIDTH-1:0]        reg_addr_in = 0;
   reg [`CPCI_NF2_DATA_WIDTH-1:0]       reg_data_in = 0;
   reg [2:0]                            reg_src_in = 0;
   
   wire                                 reg_req_out;
   wire                                 reg_ack_out;
   wire                                 reg_rd_wr_L_out;
   wire [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_out;
   wire [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_out;
   wire [2:0]                           reg_src_out;

   reg [0:63]                           icing_realm_id=0;
   reg                                  icing_realm_id_0_vld=0;
   reg                                  icing_realm_id_1_vld=0;
   reg                                  icing_realm_id_2_vld=0;
   reg                                  icing_realm_is_b4_me=0;
   reg                                  icing_realm_is_first=0;
   reg                                  icing_realm_is_me=0;
   reg                                  icing_realm_is_nxt_hop=0;
   reg                                  icing_realm_is_last=0;
   reg                                  icing_last_hop_reached=0;

   // --- Interface to flow control
   wire                                 pause;

   // --- Interface to SRAM
   wire [18:0]                          sram_rd_addr;
   wire                                 sram_rd_req;
   reg [0:63]                           sram_rd_data = 0;
   reg                                  sram_rd_vld = 0;
   reg                                  sram_rd_ack = 0;

   // --- Results
   wire [0:127]                         shared_dh_key;
   wire                                 my_realm_id_is_smaller;
   wire                                 shared_dh_key_missing;
   wire                                 shared_dh_key_vld;
   reg                                  rd_shared_key = 0;
   wire                                 shared_dh_key_first;
   wire                                 shared_dh_key_last;
   wire                                 shared_dh_key_b4_me;

   wire [0:47]                          next_hop_mac;
   wire [7:0]                           output_port;
   wire                                 next_hop_info_vld;
   reg                                  rd_nxthop_info = 0;


   reg                                  clk = 0;
   reg                                  reset = 1;

   wire [0:63]                          mem64[0:7][0:255];
   wire [0:31]                          mem32[0:7][0:1][0:255];
   wire [0:15]                          mem16[0:7][0:3][0:255];

   reg [0:63]                           sram[0:2**19-1];

   wire [0:159]                         realm_ids[0:NUM_REALMS];
   wire [0:127]                         skey[0:NUM_REALMS];
   wire [0:47]                          nxt_hops[0:NUM_REALMS];
   wire [0:7]                           output_ports[0:NUM_REALMS];
   wire                                 i_am_smaller[0:NUM_REALMS];

   always #4 clk = ~clk;

   generate
      genvar k, l, m;
      for(k=0; k<NUM_REALMS+1; k=k+1) begin
         assign realm_ids[k] = {16'habcd, k[15:0], 
                                32'hd065f00d, {3{k[31:0]}}};
         assign skey[k] = {4{4'hf, k[27:0]}};
         assign nxt_hops[k] = {8'hff, 8'h33, k[31:0]};
         assign output_ports[k] = k;
         assign i_am_smaller[k] = k[1];
      end
      
      for(k=0; k<8; k=k+1) begin
         for(m=0; m<256; m=m+1) begin
            for(l=0; l<4; l=l+1) begin
               assign mem16[k][l][m] = $random();
            end
         end
      end
      for(k=0; k<8; k=k+1) begin
         for(m=0; m<256; m=m+1) begin
            for(l=0; l<2; l=l+1) begin
               assign mem32[k][l][m] = {mem16[k][2*l][m], mem16[k][2*l+1][m]};
            end
         end
      end
      for(k=0; k<8; k=k+1) begin
         for(l=0; l<256; l=l+1) begin
            assign mem64[k][l] = {mem32[k][0][l], mem32[k][1][l]};
         end
      end
   endgenerate

   function [0:`SRAM_SKEY_NXTHOP_IDX_SIZE-1] hash;
      input [0:159] realm_id;

      begin
         hash = 0;
         hash = (mem16[0][0][realm_id[0*8 +: 8]] ^ 
                 mem16[1][0][realm_id[1*8 +: 8]] ^ 
                 mem16[2][0][realm_id[2*8 +: 8]] ^ 
                 mem16[3][0][realm_id[3*8 +: 8]] ^ 
                 mem16[4][0][realm_id[4*8 +: 8]] ^ 
                 mem16[5][0][realm_id[5*8 +: 8]] ^ 
                 mem16[6][0][realm_id[6*8 +: 8]] ^ 
                 mem16[7][0][realm_id[7*8 +: 8]]);
         $display("Sim hash 0: %x %x %x %x %x %x %x %x = %x",
                  mem16[0][0][realm_id[0*8 +: 8]] , 
                  mem16[1][0][realm_id[1*8 +: 8]] , 
                  mem16[2][0][realm_id[2*8 +: 8]] , 
                  mem16[3][0][realm_id[3*8 +: 8]] , 
                  mem16[4][0][realm_id[4*8 +: 8]] , 
                  mem16[5][0][realm_id[5*8 +: 8]] , 
                  mem16[6][0][realm_id[6*8 +: 8]] , 
                  mem16[7][0][realm_id[7*8 +: 8]], hash);
         hash = (mem16[0][1][realm_id[8*8 +: 8]] ^ 
                 mem16[1][1][realm_id[9*8 +: 8]] ^ 
                 mem16[2][1][realm_id[10*8 +: 8]] ^ 
                 mem16[3][1][realm_id[11*8 +: 8]] ^ 
                 mem16[4][1][realm_id[12*8 +: 8]] ^ 
                 mem16[5][1][realm_id[13*8 +: 8]] ^ 
                 mem16[6][1][realm_id[14*8 +: 8]] ^ 
                 mem16[7][1][realm_id[15*8 +: 8]] ^ hash);
         $display("Sim hash 1: %x %x %x %x %x %x %x %x = %x",
                  mem16[0][1][realm_id[8*8 +: 8]] , 
                  mem16[1][1][realm_id[9*8 +: 8]] , 
                  mem16[2][1][realm_id[10*8 +: 8]] , 
                  mem16[3][1][realm_id[11*8 +: 8]] , 
                  mem16[4][1][realm_id[12*8 +: 8]] , 
                  mem16[5][1][realm_id[13*8 +: 8]] , 
                  mem16[6][1][realm_id[14*8 +: 8]] , 
                  mem16[7][1][realm_id[15*8 +: 8]], hash);
         hash = (mem16[0][2][realm_id[16*8 +: 8]] ^ 
                 mem16[1][2][realm_id[17*8 +: 8]] ^ 
                 mem16[2][2][realm_id[18*8 +: 8]] ^ 
                 mem16[3][2][realm_id[19*8 +: 8]] ^ 
                 mem16[4][2][0] ^ 
                 mem16[5][2][0] ^ 
                 mem16[6][2][0] ^ 
                 mem16[7][2][0] ^ hash);
         $display("Sim hash 2: %x %x %x %x %x %x %x %x = %x",
                  mem16[0][2][realm_id[16*8 +: 8]] , 
                  mem16[1][2][realm_id[17*8 +: 8]] , 
                  mem16[2][2][realm_id[18*8 +: 8]] , 
                  mem16[3][2][realm_id[19*8 +: 8]] , 
                  mem16[4][2][0] , 
                  mem16[5][2][0] , 
                  mem16[6][2][0] , 
                  mem16[7][2][0], hash);
      end
   endfunction // hash

   task fill_sram;
      integer i;
      reg [18:0] idx;
      begin
         for(i=0; i<NUM_REALMS; i=i+1) begin
            $display("Writing sram entry for realm %u", i);
            idx = {`SRAM_SKEY_NXTHOP_BASE 
                   | {hash(realm_ids[i]), `SRAM_SKEY_NXTHOP_ENTRY_BITS'h0}};
            $display("        at idx %x", idx);
            sram[idx] = realm_ids[i][0:63];
            idx = idx + 1'b1;
            sram[idx] = realm_ids[i][64:127];
            idx = idx + 1'b1;
            sram[idx] = {realm_ids[i][128:159], 32'h0};
            idx = idx + 1'b1;
            sram[idx] = skey[i][0:63];
            idx = idx + 1'b1;
            sram[idx] = skey[i][64:127];
            idx = idx + 1'b1;
            sram[idx] = {7'h0, i_am_smaller[i], output_ports[i], nxt_hops[i]};
         end // for (i=0; i<NUM_REALMS; i=i+1)
      end
   endtask // fill_sram

   task send_realm;
      input integer i;
      input integer path_len;
      input integer path_idx;
      begin
         icing_realm_is_b4_me = (i == path_idx - 1);
         icing_realm_is_first = (i == 0);
         icing_realm_is_me = (i == path_idx);
         icing_realm_is_nxt_hop = (i == path_idx + 1);
         icing_realm_is_last = (i == path_len - 1);
         icing_last_hop_reached = (path_idx == path_len - 1);
         icing_realm_id = realm_ids[i][0:63];
         icing_realm_id_0_vld = 1;
         icing_realm_id_1_vld = 0;
         icing_realm_id_2_vld = 0;
         @(posedge clk) begin
            icing_realm_id = realm_ids[i][64:127];
            icing_realm_id_0_vld = 0;
            icing_realm_id_1_vld = 1;
            icing_realm_id_2_vld = 0;
         end
         @(posedge clk) begin
            icing_realm_id = {realm_ids[i][128:159], 32'hDEADBEEF};
            icing_realm_id_0_vld = 0;
            icing_realm_id_1_vld = 0;
            icing_realm_id_2_vld = 1;
         end
         @(posedge clk) begin
            icing_realm_id = 0;
            icing_realm_id_0_vld = 0;
            icing_realm_id_1_vld = 0;
            icing_realm_id_2_vld = 0;
         end
      end
   endtask // send_realm
   
   task send_path;
      input integer path_len;
      input integer path_idx;
      integer i;
      begin
         @(posedge clk) begin
            for(i=0; i<path_len; i=i+1) begin
               send_realm(i, path_len, path_idx);
            end // for (i=0; i<path_len; i=i+1)
         end
      end
   endtask // send_path

   task read_realm_info;
      input integer path_len;
      input integer path_idx;
      output error;
      
      integer i;
      begin
         error = 0;
         for(i=0; i<path_len; i=i+1) begin
            wait(shared_dh_key_vld && next_hop_info_vld);
            #1 wait(shared_dh_key_vld && next_hop_info_vld);
            if(i != path_idx) begin
               @(posedge clk) begin
                  if(shared_dh_key !== skey[i]) begin
                     $display("%t ERROR: skey exp %x, got %x",
                              $time, skey[i], shared_dh_key);
                     error = 1;
                  end
                  if(my_realm_id_is_smaller != i_am_smaller[i]) begin
                     $display("%t ERROR: smaller exp %x, got %x",
                              $time, i_am_smaller[i], my_realm_id_is_smaller);
                     error = 1;
                  end
                  if(shared_dh_key_first != (i == 0)) begin
                     $display("%t ERROR: first exp %x, got %x",
                              $time, i==0, shared_dh_key_first);
                     error = 1;
                  end
                  if(shared_dh_key_last != (i == path_len-1)) begin
                     $display("%t ERROR: last exp %x, got %x",
                              $time, i==path_len-1, shared_dh_key_last);
                     error = 1;
                  end
                  if(shared_dh_key_b4_me != (i == path_idx-1)) begin
                     $display("%t ERROR: first exp %x, got %x",
                              $time, i==path_idx-1, shared_dh_key_b4_me);
                     error = 1;
                  end
                  if(next_hop_mac != nxt_hops[i]) begin
                     $display("%t ERROR: mac exp %x, got %x",
                              $time, nxt_hops[i], next_hop_mac);
                     error = 1;
                  end
                  if(output_port != output_ports[i]) begin
                     $display("%t ERROR: out exp %x, got %x",
                              $time, output_ports[i], output_port);
                     error = 1;
                  end
                  rd_nxthop_info = 1'b1;
                  rd_shared_key = 1'b1;
               end // @(posedge clk)
               @(posedge clk) begin 
                  rd_nxthop_info = 1'b0;
                  rd_shared_key = 1'b0;
               end
            end
         end // for (i=0; i<path_len; i=i+1)
      end
   endtask // read_realm_info
      
   task reset_task;
      begin
         reset = 1;
         repeat(3) begin
            @(posedge clk) begin end
         end
         @(posedge clk) begin
            reset = 0;
         end
      end
   endtask // reset_task
   
   task write_reg;
      input [11:0] reg_num;
      input [31:0] reg_val;

      reg [REG_ADDR_WIDTH-1:0] reg_addr;
      begin
         @(posedge clk) begin
            reg_req_in = 1;
            reg_ack_in = 0;
            reg_rd_wr_L_in = 0;
            reg_addr = {4'h0, reg_num};
            reg_addr_in = (`UDP_REG_ADDR_WIDTH'h0 // tag
                           | {REG_BLOCK_TAG, reg_addr});
            reg_data_in = reg_val;
            reg_src_in = 0;
         end
         @(posedge clk) begin
            reg_req_in = 0;
         end
      end
   endtask // write_reg

   task read_reg;
      input [11:0] reg_num;
      input [31:0] exp_val;
      output error;

      reg [REG_ADDR_WIDTH-1:0] reg_addr;
      begin
         error = 0;
         @(posedge clk) begin
            reg_req_in = 1;
            reg_ack_in = 0;
            reg_rd_wr_L_in = 1;
            reg_addr = {4'h0, reg_num};
            reg_addr_in = (`UDP_REG_ADDR_WIDTH'h0 // tag
                           | {REG_BLOCK_TAG, reg_addr});
            reg_data_in = 0;
            reg_src_in = 0;
         end

         while(!reg_ack_out) begin
            @(posedge clk) begin
               reg_req_in = 0;
            end
         end

         // verify address
         if(reg_addr_out != reg_addr_in) begin
            $display("%t ERROR: address out %x not same as address in %x",
                     $time, reg_addr_out, reg_addr_in);
            error = 1;
         end
         if(reg_data_out != exp_val) begin
            $display("%t ERROR: data out %x not same as expected %x",
                     $time, reg_data_out, exp_val);
            error = 1;
         end
      end
   endtask // read_reg

   task fill_mem;
      integer k,m,l;
      begin
         for(k=0; k<8; k=k+1) begin
            for(m=0; m<256; m=m+1) begin
               for(l=0; l<2; l=l+1) begin
                  write_reg(k*256*2 + m*2 + l, mem32[k][l][m]);
               end
            end
         end
      end
   endtask // fill_mem

   /******************
    * Implement SRAM
    *****************/
   reg [2:0] sram_rd_req_d;
   reg [0:63] sram_rd_addr_d1, sram_rd_addr_d2, sram_rd_addr_d3;
   always @(*) sram_rd_ack = sram_rd_req;
   always @(posedge clk) begin
      sram_rd_req_d <= {sram_rd_req_d[1:0], sram_rd_req};
      sram_rd_addr_d1 <= sram_rd_addr;
      sram_rd_addr_d2 <= sram_rd_addr_d1;
      sram_rd_addr_d3 <= sram_rd_addr_d2;
      if(sram_rd_req_d[2]) begin
         sram_rd_data <= sram[sram_rd_addr_d3];
         sram_rd_vld <= 1;
      end
      else begin
         sram_rd_vld <= 0;
      end
   end
   
   reg error, error2, error_gb;
   initial begin
      error = 0;
      error_gb = 0;

      $display("%t Executing Reset", $time);
      reset_task();

      $display("%t Filling SRAM", $time);
      fill_sram;
      
      $display("%t Filling UHF mem", $time);
      fill_mem();

      #80

      $display("%t Sending path len = 3, idx = 1", $time);
      fork
         send_path(3, 1);
         read_realm_info(3, 1, error);
      join
      if(error) begin
         $display("%t ERROR l=3,i=1", $time);
         error_gb = 1;
      end

      #256
      $display("%t Sending path len = 3, idx = 2", $time);
      fork
         send_path(3, 1);
         read_realm_info(3, 1, error);
      join
      if(error) begin
         $display("%t ERROR l=3,i=2", $time);
         error_gb = 1;
      end

      #256
      $display("%t Sending path len = 6, idx = 5", $time);
      fork
         send_path(3, 1);
         read_realm_info(3, 1, error);
      join
      if(error) begin
         $display("%t ERROR l=6,i=5", $time);
         error_gb = 1;
      end

      #512
      $display("%t Sending missing realm", $time);
      fork
         send_realm(NUM_REALMS, NUM_REALMS+5, 2);
         begin
            wait(shared_dh_key_vld);
            @(posedge clk) begin
               if(!shared_dh_key_missing) begin
                  $display("%t ERROR missing realm", $time);
                  error_gb = 1;
               end
               rd_shared_key = 1;
            end
            @(posedge clk) begin
               rd_shared_key = 0;
            end
         end
      join
      

      #32
      $display("%t Sending path len = 3, idx = 1 again", $time);
      fork
         send_path(3, 1);
         read_realm_info(3, 1, error);
      join
      if(error) begin
         $display("%t ERROR l=3,i=1 again", $time);
         error_gb = 1;
      end

      if(error_gb) begin
         $display("%t Tests FAILED", $time);
      end else begin
         $display("%t Tests SUCCESS", $time);
      end
   end // initial begin

   lookup_shared_keys lookup_shared_keys
     (// Outputs
      .pause                            (pause),
      .sram_rd_addr                     (sram_rd_addr[18:0]),
      .sram_rd_req                      (sram_rd_req),
      .shared_dh_key                    (shared_dh_key[0:127]),
      .my_realm_id_is_smaller           (my_realm_id_is_smaller),
      .shared_dh_key_missing            (shared_dh_key_missing),
      .shared_dh_key_vld                (shared_dh_key_vld),
      .shared_dh_key_first              (shared_dh_key_first),
      .shared_dh_key_last               (shared_dh_key_last),
      .shared_dh_key_b4_me              (shared_dh_key_b4_me),
      .next_hop_mac                     (next_hop_mac[0:47]),
      .output_port                      (output_port[7:0]),
      .next_hop_info_vld                (next_hop_info_vld),
      .reg_req_out                      (reg_req_out),
      .reg_ack_out                      (reg_ack_out),
      .reg_rd_wr_L_out                  (reg_rd_wr_L_out),
      .reg_addr_out                     (reg_addr_out),
      .reg_data_out                     (reg_data_out),
      .reg_src_out                      (reg_src_out),
      // Inputs
      .icing_realm_id                   (icing_realm_id),
      .icing_realm_id_0_vld             (icing_realm_id_0_vld),
      .icing_realm_id_1_vld             (icing_realm_id_1_vld),
      .icing_realm_id_2_vld             (icing_realm_id_2_vld),
      .icing_realm_is_b4_me             (icing_realm_is_b4_me),
      .icing_realm_is_first             (icing_realm_is_first),
      .icing_realm_is_me                (icing_realm_is_me),
      .icing_realm_is_nxt_hop           (icing_realm_is_nxt_hop),
      .icing_realm_is_last              (icing_realm_is_last),
      .icing_last_hop_reached           (icing_last_hop_reached),
      .sram_rd_data                     (sram_rd_data[0:63]),
      .sram_rd_vld                      (sram_rd_vld),
      .sram_rd_ack                      (sram_rd_ack),
      .rd_shared_key                    (rd_shared_key),
      .rd_nxthop_info                   (rd_nxthop_info),
      .reg_req_in                       (reg_req_in),
      .reg_ack_in                       (reg_ack_in),
      .reg_rd_wr_L_in                   (reg_rd_wr_L_in),
      .reg_addr_in                      (reg_addr_in),
      .reg_data_in                      (reg_data_in),
      .reg_src_in                       (reg_src_in),
      .clk                              (clk),
      .reset                            (reset));
endmodule

// synthesis translate_on
