/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: merge.v
 * Project: ICING v2
 * Author: Jad Naous <jnaous@stanford.edu>
 * Merges two input streams.
 ******************************************************************************/
`timescale 1ns/1ps

`ifndef LOG2_FUNC
 `include "NF_2.1_defines.v"
`endif

module merge
  #(parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH=DATA_WIDTH/8,
    parameter NUM_QUEUES = 4)

    (// --- data path interface
     output reg [0:DATA_WIDTH-1]        out_data,
     output reg [0:CTRL_WIDTH-1]        out_ctrl,
     output reg                         out_wr,
     input                              out_rdy,

     input  [0:DATA_WIDTH*NUM_QUEUES-1] in_data,
     input  [0:CTRL_WIDTH*NUM_QUEUES-1] in_ctrl,
     input  [0:NUM_QUEUES-1]            in_wr,
     output [0:NUM_QUEUES-1]            in_rdy,

     // --- Misc
     input                              clk,
     input                              reset);

   `LOG2_FUNC

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_INPUT = 0,
              READ_PKT       = 1;

   localparam NUM_QUEUES_BITS = NUM_QUEUES == 1 ? 1 : log2(NUM_QUEUES);
   
   //------------------------ Wires/Regs -----------------------------
   reg [NUM_QUEUES_BITS-1:0]  current_q, current_q_nxt;
   reg                        state, state_nxt;
   reg [0:NUM_QUEUES-1]       in_fifo_rd_en;
   reg                        in_fifo_ctrl_prev_0;

   wire [0:DATA_WIDTH-1]      in_fifo_data[0:NUM_QUEUES-1];
   wire [0:CTRL_WIDTH-1]      in_fifo_ctrl[0:NUM_QUEUES-1];
   wire [0:NUM_QUEUES-1]      in_fifo_nearly_full;
   wire [0:NUM_QUEUES-1]      in_fifo_empty;

   reg                        out_wr_nxt;
   wire [0:CTRL_WIDTH-1]      in_fifo_ctrl_sel;
   
   //------------------------- Modules -------------------------------

   generate
      genvar                  i;
      for(i=0; i<NUM_QUEUES; i=i+1) begin:gen_fifos
         /* store pkt as it comes in */
         fallthrough_small_fifo #(.WIDTH(CTRL_WIDTH+DATA_WIDTH), .MAX_DEPTH_BITS(2))
           input_fifo
             (.din           ({in_ctrl[i*CTRL_WIDTH +: CTRL_WIDTH], in_data[i*DATA_WIDTH +: DATA_WIDTH]}),
              .wr_en         (in_wr[i]),
              .rd_en         (in_fifo_rd_en[i]),
              .dout          ({in_fifo_ctrl[i], in_fifo_data[i]}),
              .full          (),
              .prog_full     (),
              .nearly_full   (in_fifo_nearly_full[i]),
              .empty         (in_fifo_empty[i]),
              .reset         (reset),
              .clk           (clk)
              );
         assign in_rdy[i] = !in_fifo_nearly_full[i];
      end // block: gen_fifos
   endgenerate

   //--------------------------- Logic -------------------------------
/* -----\/----- EXCLUDED -----\/-----
   assign out_data = in_fifo_data[current_q];
   assign out_ctrl = in_fifo_ctrl[current_q];
 -----/\----- EXCLUDED -----/\----- */

   assign in_fifo_ctrl_sel = in_fifo_ctrl[current_q];
   always @(posedge clk) begin
      out_data    <= in_fifo_data[current_q];
      out_ctrl    <= in_fifo_ctrl_sel;
      out_wr      <= reset ? 0 : out_wr_nxt;
   end
   
   always @(*) begin
      state_nxt         = state;
      out_wr_nxt        = 0;
      in_fifo_rd_en     = 0;
      current_q_nxt     = current_q;
      
      case(state)
         WAIT_FOR_INPUT: begin
            if(!in_fifo_empty[current_q]) begin
               state_nxt                  = READ_PKT;
            end
            else begin
               current_q_nxt = (current_q == NUM_QUEUES-1)? 0 : current_q + 1'b1;
            end
         end // case: WAIT_FOR_INPUT

         READ_PKT: begin
            if(out_rdy && !in_fifo_empty[current_q]) begin
               out_wr_nxt                 = 1;
               in_fifo_rd_en[current_q]   = 1;
               if(in_fifo_ctrl_prev_0 && in_fifo_ctrl_sel != 0) begin
                  state_nxt   = WAIT_FOR_INPUT;
               end
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         state                    <= WAIT_FOR_INPUT;
         current_q                <= 0;
         in_fifo_ctrl_prev_0      <= 0;
      end
      else begin
         state                    <= state_nxt;
         in_fifo_ctrl_prev_0      <= in_fifo_rd_en[current_q] ? 
                                     (in_fifo_ctrl_sel == 0) : 
                                     in_fifo_ctrl_prev_0;
         current_q                <= current_q_nxt;
      end // else: !if(reset)
   end // always @ (posedge clk)
endmodule // merge


