/*****************************************************
 * module: mmapped_ram_256x32.v
 * author: Jad Naous
 * description: memory mapped ram that is accessible
 *              via registers from software
 * Each 32bit words can be accessed with its own addr
 * bus from hw.
 *****************************************************/

`timescale 1ns/1ps

module mmapped_ram_256x32
  #(parameter WORD_WIDTH = 4,             /* width in number of 64-bit words */
    parameter WIDTH = WORD_WIDTH*64,
    parameter DEPTH_BITS = 8,
    parameter UDP_REG_SRC_WIDTH = 3,
    parameter REG_BLOCK_TAG = 0,
    parameter REG_ADDR_WIDTH = log2(WORD_WIDTH*2*256))

    (// --- register interface
     input                                  reg_req_in,
     input                                  reg_ack_in,
     input                                  reg_rd_wr_L_in,
     input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
     input  [0:`CPCI_NF2_DATA_WIDTH-1]      reg_data_in,
     input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,
    
     output reg                             reg_req_out,
     output reg                             reg_ack_out,
     output                                 reg_rd_wr_L_out,
     output     [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
     output reg [0:`CPCI_NF2_DATA_WIDTH-1]  reg_data_out,
     output     [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,

     // --- Logic interface
     input [0:WIDTH-1]                      ram_data_in,
     output [0:WIDTH-1]                     ram_data_out,
     input [DEPTH_BITS*WORD_WIDTH*2-1:0]    ram_addr,
     input                                  ram_wr_en,
     input                                  ram_rd_en,

     input                                  clk,
     input                                  reset);

   `LOG2_FUNC

   //-------------------- Internal Parameters ------------------------
   localparam WORD_WIDTH_BITS = log2(WORD_WIDTH);
   localparam NUM_REGS_USED   = WORD_WIDTH*2*256;

   //---------------------- Wires and regs----------------------------
   wire [0:31]                     dout_words[0:2*WORD_WIDTH-1];
   reg [0:31]                      dout_words_reg[0:2*WORD_WIDTH-1];
   wire [0:31]                     din_words[0:2*WORD_WIDTH-1];
   wire [7:0]                      addr_word[0:2*WORD_WIDTH-1];
   wire [0:2*WORD_WIDTH-1]         wr_en;
   wire [0:`CPCI_NF2_DATA_WIDTH-1] reg_data_fifo;

   wire [REG_ADDR_WIDTH-1:0]       reg_addr;
   wire [`UDP_REG_ADDR_WIDTH - 1 : REG_ADDR_WIDTH] tag_addr;
   wire [7:0]                      reg_line_addr;
   wire [WORD_WIDTH_BITS:0]        reg_word_addr;
   reg [0:31]                      dout_words_sel;
   reg  [0:2*WORD_WIDTH-1]         reg_wr_req;
   reg  [2:0]                      ram_not_busy;
   
   //------------------------- Modules -------------------------------
   generate
      genvar i;
      for(i=0; i<WORD_WIDTH; i=i+1) begin:gen_ram
         RAMB16_S36_S36 ram (.DOA    (dout_words[2*i]),
                             .DOPA   (),
                             .ADDRA  ({1'b0,addr_word[2*i]}),
                             .CLKA   (clk),
                             .DIA    (din_words[2*i]),
                             .DIPA   (4'h0),
                             .ENA    (1'b1),
                             .SSRA   (reset),
                             .WEA    (wr_en[2*i]),
                             
                             .DOB    (dout_words[2*i+1]),
                             .DOPB   (),
                             .ADDRB  ({1'b1, addr_word[2*i]}),
                             .CLKB   (clk),
                             .DIB    (din_words[2*i+1]),
                             .DIPB   (4'h0),
                             .ENB    (1'b1),
                             .SSRB   (reset),
                             .WEB    (wr_en[2*i+1]));
      end // block: gen_ram
   endgenerate

   fallthrough_small_fifo 
     #(.WIDTH(2+`UDP_REG_ADDR_WIDTH+`CPCI_NF2_DATA_WIDTH+UDP_REG_SRC_WIDTH),
       .MAX_DEPTH_BITS(2))
      reg_req_fifo
        (.din           ({reg_ack_in, reg_rd_wr_L_in, 
                          reg_addr_in, reg_data_in, 
                          reg_src_in}),
         .wr_en         (reg_req_in),
         .rd_en         (reg_req_out),
         .dout          ({reg_ack_fifo, reg_rd_wr_L_out, 
                          reg_addr_out, reg_data_fifo, 
                          reg_src_out}),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (reg_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //-------------------------- Logic --------------------------------

   generate
      for(i=0; i<WORD_WIDTH*2; i=i+1) begin:gen_assigns
	 assign addr_word[i] = ((ram_rd_en | ram_wr_en) 
				? ram_addr[DEPTH_BITS*i-1 +: DEPTH_BITS]
				: reg_line_addr);
         always @(posedge clk) dout_words_reg[i] <= dout_words[i];
         assign ram_data_out[i*32 +: 32] = dout_words_reg[i];
         assign din_words[i] = ram_wr_en ? ram_data_in[i*32 +: 32] : reg_data_fifo;
         assign wr_en[i] = ram_wr_en || reg_wr_req[i];
      end
   endgenerate

   /* parse register requests */
   assign reg_addr         = reg_addr_out;                                              // addresses in this module
   assign tag_addr         = reg_addr_out[`UDP_REG_ADDR_WIDTH - 1:REG_ADDR_WIDTH];      // block adress
   assign tag_hit          = tag_addr == REG_BLOCK_TAG;                                 // address is in this block
   assign addr_good        = !reg_fifo_empty && reg_addr < NUM_REGS_USED && tag_hit;    // address is used in this module
   assign reg_line_addr    = reg_addr[WORD_WIDTH_BITS+8 : WORD_WIDTH_BITS+1];           // each line has 2*WORD_WIDTH 32-bit words
   assign reg_word_addr    = reg_addr[WORD_WIDTH_BITS   : 0];                           // selects word from line
     
   always @(*) begin
      reg_wr_req     = 0;
      reg_req_out    = 0;
      reg_ack_out    = reg_ack_fifo;
      reg_data_out   = dout_words_sel;
      /* wait until there's a register request */
      if(!reg_fifo_empty) begin
         /* check if the request is ours to process */
         if(addr_good) begin
            /* if the request is a read,
             * then it should have gone through
             * only if the ram wasn't busy two cycles ago */
            if(reg_rd_wr_L_out) begin
               if(ram_not_busy[2]) begin
                  reg_req_out     = 1;
                  reg_ack_out     = 1;
               end
               /* otherwise we have to wait */
            end
            /* else if it is a write and the
             * RAM is not busy right now, then
             * execute it */
            else begin
               /* wait until the ram is not in use */
               if(!ram_rd_en && !ram_wr_en) begin
                  reg_req_out                  = 1;
                  reg_ack_out                  = 1;
                  reg_wr_req[reg_word_addr]    = 1;
                  reg_data_out                 = reg_data_fifo;
               end
               /* otherwise, wait */
            end // else: !if(reg_rd_wr_L_out)
         end // if (addr_good)

         /* otherwise, the address isn't ours
          * so send it on its way */
         else begin
            reg_req_out     = 1;
            reg_data_out    = reg_data_fifo;
         end // else: !if(addr_good)
      end // if (!reg_fifo_empty)
   end // always @ (*)

   always @(posedge clk) begin
      ram_not_busy           <= {ram_not_busy[1:0], (!ram_rd_en && !ram_wr_en && !reg_fifo_empty)};
      dout_words_sel         <= dout_words_reg[reg_word_addr];
   end
endmodule // mmapped_ram_256x64
