/*****************************************************
 * module: mmapped_ram.v
 * author: Jad Naous
 * description: memory mapped ram that is accessible
 *              via registers from software
 * This instantiates 8 memories with 8 bit
 * addresses and 32 bit outputs. The RAMs are rd/wr by 
 * sw and rd by hw.
 *****************************************************/

`timescale 1ns/1ps

`ifndef UDP_REG_ADDR_WIDTH
 `define UDP_REG_ADDR_WIDTH 23
`endif
`ifndef CPCI_NF2_DATA_WIDTH
 `define CPCI_NF2_DATA_WIDTH 32
`endif

module mmapped_ram_8x256x64
  #(parameter UDP_REG_SRC_WIDTH = 2,
    parameter REG_BLOCK_TAG = 0,
    parameter REG_ADDR_WIDTH = 12)

    (// --- register interface
     input                                  reg_req_in,
     input                                  reg_ack_in,
     input                                  reg_rd_wr_L_in,
     input  [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_in,
     input  [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_in,
     input  [UDP_REG_SRC_WIDTH-1:0]         reg_src_in,
    
     output reg                             reg_req_out,
     output reg                             reg_ack_out,
     output                                 reg_rd_wr_L_out,
     output     [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_out,
     output reg [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_out,
     output     [UDP_REG_SRC_WIDTH-1:0]     reg_src_out,

     // --- Logic interface
     output [0:8*64-1]                      ram_data_out,
     input [0:8*8-1]                        ram_addr,
     input                                  ram_rd_en,

     input                                  clk,
     input                                  reset);

   //-------------------- Internal Parameters ------------------------
   localparam NUM_REGS_USED   = 2*8*256;

   //---------------------- Wires and regs----------------------------
   wire [0:31]                     dout_words[0:15];
   wire [0:63]                     addr;
   reg  [0:15]                     wr_en;
   wire [0:`CPCI_NF2_DATA_WIDTH-1] reg_data_fifo;

   wire [`UDP_REG_ADDR_WIDTH - 1 : REG_ADDR_WIDTH] tag_addr;
   wire [0:31]                     dout_words_sel;
   reg                             ram_not_busy;
   
   wire [3:0]                      mem_sel;

   //------------------------- Modules -------------------------------

   generate
      genvar i;
      for(i=0; i<8; i=i+1) begin: gen_ram
         assign ram_data_out[i*64:i*64+31] = dout_words[2*i];
         assign ram_data_out[i*64+32:i*64+63] = dout_words[2*i+1];
         /* Make the output of the RAMs look 64-bits wide */
         RAMB16_S36_S36 ram (.DOA    (dout_words[2*i]),
                             .DOPA   (),
                             .ADDRA  ({1'b0,addr[i*8:i*8+7]}),
                             .CLKA   (clk),
                             .DIA    (reg_data_fifo),
                             .DIPA   (4'h0),
                             .ENA    (1'b1),
                             .SSRA   (reset),
                             .WEA    (wr_en[2*i]),
                             
                             .DOB    (dout_words[2*i+1]),
                             .DOPB   (),
                             .ADDRB  ({1'b1, addr[i*8:i*8+7]}),
                             .CLKB   (clk),
                             .DIB    (reg_data_fifo),
                             .DIPB   (4'h0),
                             .ENB    (1'b1),
                             .SSRB   (reset),
                             .WEB    (wr_en[2*i+1]));
	 defparam ram.INIT_00 = "000000010004000500080009000c000d001000110014001500180019001c001d";
	 defparam ram.INIT_20 = "0002000300060007000a000b000e000f0012001300160017001a001b001e001f";
	 defparam ram.INIT_01 = "002000210024002500280029002c002d003000310034003500380039003c003d";
	 defparam ram.INIT_21 = "0022002300260027002a002b002e002f0032003300360037003a003b003e003f";
	 defparam ram.INIT_02 = "004000410044004500480049004c004d005000510054005500580059005c005d";
	 defparam ram.INIT_22 = "0042004300460047004a004b004e004f0052005300560057005a005b005e005f";
	 defparam ram.INIT_03 = "006000610064006500680069006c006d007000710074007500780079007c007d";
	 defparam ram.INIT_23 = "0062006300660067006a006b006e006f0072007300760077007a007b007e007f";
	 defparam ram.INIT_04 = "008000810084008500880089008c008d009000910094009500980099009c009d";
	 defparam ram.INIT_24 = "0082008300860087008a008b008e008f0092009300960097009a009b009e009f";
	 defparam ram.INIT_05 = "00a000a100a400a500a800a900ac00ad00b000b100b400b500b800b900bc00bd";
	 defparam ram.INIT_25 = "00a200a300a600a700aa00ab00ae00af00b200b300b600b700ba00bb00be00bf";
	 defparam ram.INIT_06 = "00c000c100c400c500c800c900cc00cd00d000d100d400d500d800d900dc00dd";
	 defparam ram.INIT_26 = "00c200c300c600c700ca00cb00ce00cf00d200d300d600d700da00db00de00df";
	 defparam ram.INIT_07 = "00e000e100e400e500e800e900ec00ed00f000f100f400f500f800f900fc00fd";
	 defparam ram.INIT_27 = "00e200e300e600e700ea00eb00ee00ef00f200f300f600f700fa00fb00fe00ff";
	 defparam ram.INIT_08 = "010001010104010501080109010c010d011001110114011501180119011c011d";
	 defparam ram.INIT_28 = "0102010301060107010a010b010e010f0112011301160117011a011b011e011f";
	 defparam ram.INIT_09 = "012001210124012501280129012c012d013001310134013501380139013c013d";
	 defparam ram.INIT_29 = "0122012301260127012a012b012e012f0132013301360137013a013b013e013f";
	 defparam ram.INIT_0A = "014001410144014501480149014c014d015001510154015501580159015c015d";
	 defparam ram.INIT_2A = "0142014301460147014a014b014e014f0152015301560157015a015b015e015f";
	 defparam ram.INIT_0B = "016001610164016501680169016c016d017001710174017501780179017c017d";
	 defparam ram.INIT_2B = "0162016301660167016a016b016e016f0172017301760177017a017b017e017f";
	 defparam ram.INIT_0C = "018001810184018501880189018c018d019001910194019501980199019c019d";
	 defparam ram.INIT_2C = "0182018301860187018a018b018e018f0192019301960197019a019b019e019f";
	 defparam ram.INIT_0D = "01a001a101a401a501a801a901ac01ad01b001b101b401b501b801b901bc01bd";
	 defparam ram.INIT_2D = "01a201a301a601a701aa01ab01ae01af01b201b301b601b701ba01bb01be01bf";
	 defparam ram.INIT_0E = "01c001c101c401c501c801c901cc01cd01d001d101d401d501d801d901dc01dd";
	 defparam ram.INIT_2E = "01c201c301c601c701ca01cb01ce01cf01d201d301d601d701da01db01de01df";
	 defparam ram.INIT_0F = "01e001e101e401e501e801e901ec01ed01f001f101f401f501f801f901fc01fd";
	 defparam ram.INIT_2F = "01e201e301e601e701ea01eb01ee01ef01f201f301f601f701fa01fb01fe01ff";
	 defparam ram.INIT_10 = "020002010204020502080209020c020d021002110214021502180219021c021d";
	 defparam ram.INIT_30 = "0202020302060207020a020b020e020f0212021302160217021a021b021e021f";
	 defparam ram.INIT_11 = "022002210224022502280229022c022d023002310234023502380239023c023d";
	 defparam ram.INIT_31 = "0222022302260227022a022b022e022f0232023302360237023a023b023e023f";
	 defparam ram.INIT_12 = "024002410244024502480249024c024d025002510254025502580259025c025d";
	 defparam ram.INIT_32 = "0242024302460247024a024b024e024f0252025302560257025a025b025e025f";
	 defparam ram.INIT_13 = "026002610264026502680269026c026d027002710274027502780279027c027d";
	 defparam ram.INIT_33 = "0262026302660267026a026b026e026f0272027302760277027a027b027e027f";
	 defparam ram.INIT_14 = "028002810284028502880289028c028d029002910294029502980299029c029d";
	 defparam ram.INIT_34 = "0282028302860287028a028b028e028f0292029302960297029a029b029e029f";
	 defparam ram.INIT_15 = "02a002a102a402a502a802a902ac02ad02b002b102b402b502b802b902bc02bd";
	 defparam ram.INIT_35 = "02a202a302a602a702aa02ab02ae02af02b202b302b602b702ba02bb02be02bf";
	 defparam ram.INIT_16 = "02c002c102c402c502c802c902cc02cd02d002d102d402d502d802d902dc02dd";
	 defparam ram.INIT_36 = "02c202c302c602c702ca02cb02ce02cf02d202d302d602d702da02db02de02df";
	 defparam ram.INIT_17 = "02e002e102e402e502e802e902ec02ed02f002f102f402f502f802f902fc02fd";
	 defparam ram.INIT_37 = "02e202e302e602e702ea02eb02ee02ef02f202f302f602f702fa02fb02fe02ff";
	 defparam ram.INIT_18 = "030003010304030503080309030c030d031003110314031503180319031c031d";
	 defparam ram.INIT_38 = "0302030303060307030a030b030e030f0312031303160317031a031b031e031f";
	 defparam ram.INIT_19 = "032003210324032503280329032c032d033003310334033503380339033c033d";
	 defparam ram.INIT_39 = "0322032303260327032a032b032e032f0332033303360337033a033b033e033f";
	 defparam ram.INIT_1A = "034003410344034503480349034c034d035003510354035503580359035c035d";
	 defparam ram.INIT_3A = "0342034303460347034a034b034e034f0352035303560357035a035b035e035f";
	 defparam ram.INIT_1B = "036003610364036503680369036c036d037003710374037503780379037c037d";
	 defparam ram.INIT_3B = "0362036303660367036a036b036e036f0372037303760377037a037b037e037f";
	 defparam ram.INIT_1C = "038003810384038503880389038c038d039003910394039503980399039c039d";
	 defparam ram.INIT_3C = "0382038303860387038a038b038e038f0392039303960397039a039b039e039f";
	 defparam ram.INIT_1D = "03a003a103a403a503a803a903ac03ad03b003b103b403b503b803b903bc03bd";
	 defparam ram.INIT_3D = "03a203a303a603a703aa03ab03ae03af03b203b303b603b703ba03bb03be03bf";
	 defparam ram.INIT_1E = "03c003c103c403c503c803c903cc03cd03d003d103d403d503d803d903dc03dd";
	 defparam ram.INIT_3E = "03c203c303c603c703ca03cb03ce03cf03d203d303d603d703da03db03de03df";
	 defparam ram.INIT_1F = "03e003e103e403e503e803e903ec03ed03f003f103f403f503f803f903fc03fd";
	 defparam ram.INIT_3F = "03e203e303e603e703ea03eb03ee03ef03f203f303f603f703fa03fb03fe03ff";
      end // block: gen_ram
   endgenerate

   fallthrough_small_fifo 
     #(.WIDTH(2+`UDP_REG_ADDR_WIDTH+`CPCI_NF2_DATA_WIDTH+UDP_REG_SRC_WIDTH),
       .MAX_DEPTH_BITS(2))
      reg_req_fifo
        (.din           ({reg_ack_in, reg_rd_wr_L_in, 
                          reg_addr_in, reg_data_in, 
                          reg_src_in}),
         .wr_en         (reg_req_in),
         .rd_en         (reg_req_out),
         .dout          ({reg_ack_fifo, reg_rd_wr_L_out, 
                          reg_addr_out, reg_data_fifo, 
                          reg_src_out}),
         .full          (),
         .prog_full     (),
         .nearly_full   (),
         .empty         (reg_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //-------------------------- Logic --------------------------------

   /* parse register requests */
   assign tag_addr         = reg_addr_out[`UDP_REG_ADDR_WIDTH - 1:REG_ADDR_WIDTH];      // block adress
   assign tag_hit          = tag_addr == REG_BLOCK_TAG;                                 // address is in this block
   assign addr_good        = (!reg_fifo_empty
                              && reg_addr_out[REG_ADDR_WIDTH-1:0] < NUM_REGS_USED
                              && tag_hit);            // address is used in this module
   /* Assign the addresses that the 2 32-bit words of every 64-bit word
    * are consecutive in addresses. */
   assign addr             = ram_rd_en ? ram_addr : {8{reg_addr_out[8:1]}};
   assign mem_sel          = {reg_addr_out[11:9], reg_addr_out[0]};

   assign dout_words_sel   = dout_words[mem_sel];
   
   always @(*) begin
      wr_en          = 0;
      reg_req_out    = 0;
      reg_ack_out    = reg_ack_fifo;
      reg_data_out   = dout_words_sel;

      /* wait until there's a register request */
      if(!reg_fifo_empty) begin
         /* check if the request is ours to process */
         if(addr_good) begin
            /* if the request is a read,
             * then it should have gone through
             * only if the ram wasn't busy a cycle ago */
            if(reg_rd_wr_L_out) begin
               if(ram_not_busy) begin
                  reg_req_out     = 1;
                  reg_ack_out     = 1;
               end
               /* otherwise we have to wait */
            end
            /* else if it is a write and the
             * RAM is not busy right now, then
             * execute it */
            else begin
               /* wait until the ram is not in use */
               if(!ram_rd_en) begin
                  reg_req_out      = 1;
                  reg_ack_out      = 1;
                  wr_en[mem_sel]   = 1'b1;
                  reg_data_out     = reg_data_fifo;
               end
               /* otherwise, wait */
            end // else: !if(reg_rd_wr_L_out)
         end // if (addr_good)

         /* otherwise, the address isn't ours
          * so send it on its way */
         else begin
            reg_req_out     = 1;
            reg_data_out    = reg_data_fifo;
         end // else: !if(addr_good)
      end // if (!reg_fifo_empty)
   end // always @ (*)

   always @(posedge clk) begin
      ram_not_busy           <= !ram_rd_en & !reg_fifo_empty;
   end
endmodule // mmapped_ram_8x256x64

// synthesis translate_off

module mmapped_ram_tester();
   reg                                  reg_req_in = 0;
   reg                                  reg_ack_in = 0;
   reg                                  reg_rd_wr_L_in = 1;
   reg [`UDP_REG_ADDR_WIDTH-1:0]        reg_addr_in = 0;
   reg [`CPCI_NF2_DATA_WIDTH-1:0]       reg_data_in = 0;
   reg [2-1:0]                          reg_src_in = 0;
   
   wire                                 reg_req_out;
   wire                                 reg_ack_out;
   wire                                 reg_rd_wr_L_out;
   wire [`UDP_REG_ADDR_WIDTH-1:0]       reg_addr_out;
   wire [`CPCI_NF2_DATA_WIDTH-1:0]      reg_data_out;
   wire [2-1:0]                         reg_src_out;

   wire [0:8*64-1]                      ram_data_out;
   reg [8*8-1:0]                        ram_addr = 0;
   reg                                  ram_rd_en = 0;

   reg                                  clk = 0;
   reg                                  reset = 1;

   wire [0:63]                          mem64[0:7][0:255];
   wire [0:31]                          mem32[0:7][0:1][0:255];
   wire [0:15]                          mem16[0:7][0:3][0:255];

   localparam REG_ADDR_WIDTH = 16;
   localparam REG_BLOCK_TAG = 7'b0001001;
   
   always #8 clk = ~clk;

   generate
      genvar k, l, m;
      for(k=0; k<8; k=k+1) begin
         for(m=0; m<256; m=m+1) begin
            for(l=0; l<4; l=l+1) begin
               assign mem16[k][l][m] = k*256*4 + m*4 + l;
            end
         end
      end
      for(k=0; k<8; k=k+1) begin
         for(m=0; m<256; m=m+1) begin
            for(l=0; l<2; l=l+1) begin
               assign mem32[k][l][m] = {mem16[k][2*l][m], mem16[k][2*l+1][m]};
            end
         end
      end
      for(k=0; k<8; k=k+1) begin
         for(l=0; l<256; l=l+1) begin
            assign mem64[k][l] = {mem32[k][0][l], mem32[k][1][l]};
         end
      end
   endgenerate

   task reset_task;
      begin
         reset = 1;
         repeat(3) begin
            @(posedge clk) begin end
         end
         @(posedge clk) begin
            reset = 0;
         end
      end
   endtask // reset_task
   
   task write_reg;
      input [11:0] reg_num;
      input [31:0] reg_val;

      reg [REG_ADDR_WIDTH-1:0] reg_addr;
      begin
         @(posedge clk) begin
            reg_req_in = 1;
            reg_ack_in = 0;
            reg_rd_wr_L_in = 0;
            reg_addr = {4'h0, reg_num};
            reg_addr_in = (`UDP_REG_ADDR_WIDTH'h0 // tag
                           | {REG_BLOCK_TAG, reg_addr});
            reg_data_in = reg_val;
            reg_src_in = 0;
         end
         @(posedge clk) begin
            reg_req_in = 0;
         end
      end
   endtask // write_reg

   task read_reg;
      input [11:0] reg_num;
      input [31:0] exp_val;
      output error;

      reg [REG_ADDR_WIDTH-1:0] reg_addr;
      begin
         error = 0;
         @(posedge clk) begin
            reg_req_in = 1;
            reg_ack_in = 0;
            reg_rd_wr_L_in = 1;
            reg_addr = {4'h0, reg_num};
            reg_addr_in = (`UDP_REG_ADDR_WIDTH'h0 // tag
                           | {REG_BLOCK_TAG, reg_addr});
            reg_data_in = 0;
            reg_src_in = 0;
         end

         while(!reg_ack_out) begin
            @(posedge clk) begin
               reg_req_in = 0;
            end
         end

         // verify address
         if(reg_addr_out != reg_addr_in) begin
            $display("%t ERROR: address out %x not same as address in %x",
                     $time, reg_addr_out, reg_addr_in);
            error = 1;
         end
         if(reg_data_out != exp_val) begin
            $display("%t ERROR: data out %x not same as expected %x",
                     $time, reg_data_out, exp_val);
            error = 1;
         end
      end
   endtask // read_reg

   task fill_mem;
      integer k,m,l;
      begin
         for(k=0; k<8; k=k+1) begin
            for(m=0; m<256; m=m+1) begin
               for(l=0; l<2; l=l+1) begin
                  write_reg(k*256*2 + m*2 + l, mem32[k][l][m]);
               end
            end
         end
      end
   endtask // fill_mem

   task read_mem;
      output error_out;

      integer k,m,l;
      reg error;

      begin
         error_out = 0;
         error = 0;
         for(k=0; k<8; k=k+1) begin
            for(m=0; m<256; m=m+1) begin
               for(l=0; l<2; l=l+1) begin
                  read_reg(k*256*2 + m*2 + l, mem32[k][l][m], error);
                  if(error) begin
                     $display("%t ERROR: Reading value from mem for k=%u, m=%u, l=%u",
                              $time, k,m,k);
                     error_out = 1;
                  end
               end
            end
         end
      end
   endtask // read_mem
         
   task check_read;
      input [0:63] start_addrs;
      input [8:0] num_reads;
      output error;

      integer i,j;
      reg [0:8*64-1] exp;
      reg [0:7] addr;
      begin
         i = 0;
         error = 0;
         @(posedge clk) #1 begin
            ram_addr = start_addrs;
            ram_rd_en = 1;
            exp = {mem64[0][start_addrs[0 +: 8]],
                   mem64[1][start_addrs[1*8 +: 8]],
                   mem64[2][start_addrs[2*8 +: 8]],
                   mem64[3][start_addrs[3*8 +: 8]],
                   mem64[4][start_addrs[4*8 +: 8]],
                   mem64[5][start_addrs[5*8 +: 8]],
                   mem64[6][start_addrs[6*8 +: 8]],
                   mem64[7][start_addrs[7*8 +: 8]]};
            start_addrs[0*8 +: 8] = start_addrs[0*8 +: 8] + 1'b1;
            start_addrs[1*8 +: 8] = start_addrs[1*8 +: 8] + 1'b1;
            start_addrs[2*8 +: 8] = start_addrs[2*8 +: 8] + 1'b1;
            start_addrs[3*8 +: 8] = start_addrs[3*8 +: 8] + 1'b1;
            start_addrs[4*8 +: 8] = start_addrs[4*8 +: 8] + 1'b1;
            start_addrs[5*8 +: 8] = start_addrs[5*8 +: 8] + 1'b1;
            start_addrs[6*8 +: 8] = start_addrs[6*8 +: 8] + 1'b1;
            start_addrs[7*8 +: 8] = start_addrs[7*8 +: 8] + 1'b1;
         end
         for(i=0; i < num_reads-1; i=i+1) begin
            @(posedge clk) #1 begin
               if(ram_data_out !== exp) begin
                  $display("%t ERROR: Wide Data not as expected. Exp %x, saw %x",
                           $time, exp, ram_data_out);
                  error = 1;
               end
               ram_addr = start_addrs;
               ram_rd_en = 1;
               exp = {mem64[0][start_addrs[0 +: 8]],
                      mem64[1][start_addrs[1*8 +: 8]],
                      mem64[2][start_addrs[2*8 +: 8]],
                      mem64[3][start_addrs[3*8 +: 8]],
                      mem64[4][start_addrs[4*8 +: 8]],
                      mem64[5][start_addrs[5*8 +: 8]],
                      mem64[6][start_addrs[6*8 +: 8]],
                      mem64[7][start_addrs[7*8 +: 8]]};
               start_addrs[0*8 +: 8] = start_addrs[0*8 +: 8] + 1'b1;
               start_addrs[1*8 +: 8] = start_addrs[1*8 +: 8] + 1'b1;
               start_addrs[2*8 +: 8] = start_addrs[2*8 +: 8] + 1'b1;
               start_addrs[3*8 +: 8] = start_addrs[3*8 +: 8] + 1'b1;
               start_addrs[4*8 +: 8] = start_addrs[4*8 +: 8] + 1'b1;
               start_addrs[5*8 +: 8] = start_addrs[5*8 +: 8] + 1'b1;
               start_addrs[6*8 +: 8] = start_addrs[6*8 +: 8] + 1'b1;
               start_addrs[7*8 +: 8] = start_addrs[7*8 +: 8] + 1'b1;
            end
         end // while (i < num_reads-1)
         @(posedge clk) #1 begin
            if(ram_data_out !== exp) begin
               $display("%t ERROR: Wide Data not as expected. Exp %x, saw %x",
                        $time, exp, ram_data_out);
               error = 1;
            end
            ram_rd_en = 0;
         end
      end
   endtask // check_read

   reg error, error2, error_gb;
   initial begin
      error = 0;
      error_gb = 0;

      $display("%t Executing Reset", $time);
      reset_task();

      $display("%t Filling mem", $time);
      fill_mem();

      #80

      $display("%t Reading mem", $time);
      read_mem(error);
      if(error) begin
         $display("%t ERROR: reading memory through reg iface", $time);
         error_gb = 1;
      end

      $display("%t Checking read", $time);
      check_read({8{8'h0}}, 256, error);
      if(error) begin
         $display("%t ERROR: reading memory trial 1 through dp iface", $time);
         error_gb = 1;
      end

      $display("%t Checking read again", $time);
      check_read({8'h02, 8'h52, 8'h12, 8'h0, 
                  8'hfa, 8'ha0, 8'h00, 8'h31}, 6, error);
      if(error) begin
         $display("%t ERROR: reading memory trial 2 through dp iface", $time);
         error_gb = 1;
      end

      $display("%t Checking conflicting read and reg 1", $time);
      fork
         check_read({8{8'h20}}, 4, error);
         #16 read_reg(0, {16'h0, 16'h1}, error2);
      join
      if(error || error2) begin
         $display("%t ERROR: Conflict check1", $time);
         error_gb = 1;
      end
      
      $display("%t Checking conflicting read and reg 2", $time);
      fork
         #32 check_read({8{8'h20}}, 4, error);
         read_reg(1, {16'h2, 16'h3}, error2);
      join
      if(error || error2) begin
         $display("%t ERROR: Conflict check2", $time);
         error_gb = 1;
      end

      $display("%t Checking conflicting read and reg 3", $time);
      fork
         #48 check_read({8{8'h20}}, 4, error);
         read_reg(2, {16'h4, 16'h5}, error2);
      join
      if(error || error2) begin
         $display("%t ERROR: Conflict check3", $time);
         error_gb = 1;
      end

      $display("%t Checking conflicting read and reg 4", $time);
      fork
         #64 check_read({8{8'h20}}, 4, error);
         read_reg(3, {16'h6, 16'h7}, error2);
      join
      if(error || error2) begin
         $display("%t ERROR: Conflict check4", $time);
         error_gb = 1;
      end

      if(error_gb) begin
         $display("%t Tests FAILED", $time);
      end else begin
         $display("%t Tests SUCCESS", $time);
      end
   end // initial begin

   mmapped_ram_8x256x64
     #(.UDP_REG_SRC_WIDTH  (2),
       .REG_BLOCK_TAG      (REG_BLOCK_TAG),
       .REG_ADDR_WIDTH     (REG_ADDR_WIDTH))
       mmapped_ram_8x256x64
         (// Outputs
          .reg_req_out                      (reg_req_out),
          .reg_ack_out                      (reg_ack_out),
          .reg_rd_wr_L_out                  (reg_rd_wr_L_out),
          .reg_addr_out                     (reg_addr_out[`UDP_REG_ADDR_WIDTH-1:0]),
          .reg_data_out                     (reg_data_out[`CPCI_NF2_DATA_WIDTH-1:0]),
          .reg_src_out                      (reg_src_out),
          .ram_data_out                     (ram_data_out[0:8*64-1]),
          // Inputs
          .reg_req_in                       (reg_req_in),
          .reg_ack_in                       (reg_ack_in),
          .reg_rd_wr_L_in                   (reg_rd_wr_L_in),
          .reg_addr_in                      (reg_addr_in[`UDP_REG_ADDR_WIDTH-1:0]),
          .reg_data_in                      (reg_data_in[`CPCI_NF2_DATA_WIDTH-1:0]),
          .reg_src_in                       (reg_src_in),
          .ram_addr                         (ram_addr[8*8-1:0]),
          .ram_rd_en                        (ram_rd_en),
          .clk                              (clk),
          .reset                            (reset));
endmodule // mmapped_ram_tester

/* -----\/----- EXCLUDED -----\/-----
module RAMB16_S36_S36 (DOA, DOB, DOPA, DOPB, ADDRA, ADDRB, CLKA, CLKB, DIA, DIB, DIPA, DIPB, ENA, ENB, SSRA, SSRB, WEA, WEB);

   output reg [31:0] DOA;
   output reg [3:0]  DOPA;
   output reg [31:0] DOB;
   output reg [3:0]  DOPB;

   input [8:0]   ADDRA;
   input [31:0]  DIA;
   input [3:0]   DIPA;
   input         ENA, CLKA, WEA, SSRA;
   input [8:0]   ADDRB;
   input [31:0]  DIB;
   input [3:0]   DIPB;
   input         ENB, CLKB, WEB, SSRB;

   reg [0:35]    mem[0:511];
   integer       i;
   
   always @(posedge CLKA) begin
      if(SSRA) begin
         for(i=0; i<512; i=i+1) begin
            mem[i] <= 36'h0;
         end
      end
      else begin
         if(WEA) begin
            mem[ADDRA] <= {DIPA, DIA};
         end
         if(WEB) begin
            mem[ADDRB] <= {DIPB, DIB};
         end
         {DOPA, DOA} <= mem[ADDRA];
         {DOPB, DOB} <= mem[ADDRB];
      end // else: !if(SSRA)
   end // always @ (posedge CLKA)
endmodule // RAMB16_S36_S36
 -----/\----- EXCLUDED -----/\----- */

// synthesis translate_on
