/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: output_port_lookup.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 *
 ******************************************************************************/
`timescale 1ns/1ps
module output_port_lookup
  #(parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH = DATA_WIDTH/8,
    parameter UDP_REG_SRC_WIDTH = 2,
    parameter SRAM_ADDR_WIDTH = 19)

   (// --- data path interface
    input [0:DATA_WIDTH-1]             in_data,
    input [0:CTRL_WIDTH-1]             in_ctrl,
    input                              in_wr,
    output                             in_rdy,

    output  [0:DATA_WIDTH-1]           out_data,
    output  [0:CTRL_WIDTH-1]           out_ctrl,
    output                             out_wr,
    input                              out_rdy,

    // --- Register interface
    input                              reg_req_in,
    input                              reg_ack_in,
    input                              reg_rd_wr_L_in,
    input  [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_in,
    input  [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_in,
    input  [UDP_REG_SRC_WIDTH-1:0]     reg_src_in,

    output                             reg_req_out,
    output                             reg_ack_out,
    output                             reg_rd_wr_L_out,
    output  [`UDP_REG_ADDR_WIDTH-1:0]  reg_addr_out,
    output  [`CPCI_NF2_DATA_WIDTH-1:0] reg_data_out,
    output  [UDP_REG_SRC_WIDTH-1:0]    reg_src_out,

    // --- SRAM Interface
    input                              rd_0_ack,
    input  [DATA_WIDTH+CTRL_WIDTH-1:0] rd_0_data,
    input                              rd_0_vld,
    output [SRAM_ADDR_WIDTH-1:0]       rd_0_addr,
    output                             rd_0_req,

    // --- Misc
    input                              clk,
    input                              reset);

   `LOG2_FUNC
   `CEILDIV_FUNC

   //-------------------- Internal Parameters ------------------------
   //------------------------ Wires/Regs -----------------------------
   wire [`UDP_REG_ADDR_WIDTH-1:0]  lookup_keys_reg_addr_out;
   wire [0:`CPCI_NF2_DATA_WIDTH-1] lookup_keys_reg_data_out;
   wire [UDP_REG_SRC_WIDTH-1:0]    lookup_keys_reg_src_out;
   wire [3:0]                      input_port;
   wire [7:0]                      word_length;
   wire [10:0]                     byte_length;
   wire [0:47]                     dst_mac;
   wire [0:47]                     src_mac;
   wire [0:15]                     eth_type;
   wire [7:0]                      icing_version;
   wire [7:0]                      icing_path_len;
   wire [7:0]                      icing_path_idx;
   wire [7:0]                      icing_error_idx;
   wire [0:55]                     icing_iv;
   wire [0:63]                     icing_realm_id;
   wire [0:31]                     icing_subrealm;
   wire [0:63]                     icing_va;
   wire [0:31]                     icing_vb;
   wire [0:7]                      icing_proto;
   wire [0:15]                     icing_pkt_len;
   wire [0:63]                     icing_payload;
   wire [7:0]                      word_count;
   wire [0:4*48-1]                 port_macs;
   wire [18:0]                     sram_rd_addr_0;
   wire [18:0]                     sram_rd_addr_1;
   wire [0:63]                     sram_rd_data;
   wire [0:127]                    shared_dh_key;
   wire [0:47]                     next_hop_mac;
   wire [7:0]                      output_port;
   wire [0:127]                    poc_key, poc_key0;
   wire [0:246]                    hash;
   wire [0:127]                    aes_enc_plain_0, aes_enc_plain_1, aes_enc_plain_2;
   wire [0:127]                    aes_enc_key_0, aes_enc_key_1, aes_enc_key_2;
   wire [0:127]                    aes_enc_cipher_0, aes_enc_cipher_1, aes_enc_cipher_2;
   wire [0:127]                    poc;
   wire [0:127]                    mac2_key_0, mac2_key_1, mac2_key_2;
   wire [0:255]                    mac2_block_0, mac2_block_1, mac2_block_2;
   wire [0:95]                     mac2_tag_0, mac2_tag_1, mac2_tag_2;
   wire [0:95]                     A;
   wire [0:95]                     prf_result;
   wire [0:159]                    my_realm_id;
   wire [0:7] 			   out_ctrl_chk;
   wire [0:63] 			   out_data_chk;

   //------------------------- Modules -------------------------------

   icing_check_hdr icing_check_hdr
     (// Outputs
      .in_rdy				(in_rdy),
      .out_data				(out_data_chk[0:63]),
      .out_ctrl				(out_ctrl_chk[0:7]),
      .out_wr				(out_wr_chk),
      .dst_mac_not_ok			(dst_mac_not_ok),
      .ethertype_not_ok			(ethertype_not_ok),
      .path_idx_not_ok			(path_idx_not_ok),
      .path_len_not_ok			(path_len_not_ok),
      .pkt_len_not_ok			(pkt_len_not_ok),
      .version_not_ok			(version_not_ok),
      .pkt_not_ok			(pkt_not_ok),
      .hdr_check_done			(hdr_check_done),
      // Inputs
      .in_data				(in_data[0:63]),
      .in_ctrl				(in_ctrl[0:7]),
      .in_wr				(in_wr),
      .out_rdy				(out_rdy_chk),
      .port_macs			(port_macs[0:48*4-1]),
      .clk				(clk),
      .reset				(reset));

   icing_pkt_parser
     #(.DATA_WIDTH        (DATA_WIDTH),
       .CTRL_WIDTH        (CTRL_WIDTH)) icing_pkt_parser
       (// --- interface to data path
        .in_data                  (out_data_chk),
        .in_ctrl                  (out_ctrl_chk),
        .in_wr                    (out_wr_chk),

        .wait_for_eop             (1'b0),
        
        // --- parser results
        .input_port               (input_port),
        .input_port_vld           (input_port_vld),
        .word_length              (word_length),
        .word_length_vld          (word_length_vld),
        .byte_length              (byte_length),
        .byte_length_vld          (byte_length_vld),
        .dst_mac                  (dst_mac),
        .dst_mac_vld              (dst_mac_vld),
        .src_mac                  (src_mac),
        .src_mac_vld              (src_mac_vld),
        .eth_type                 (eth_type),
        .eth_type_vld             (eth_type_vld),
        .icing_version            (icing_version),
        .icing_version_vld        (icing_version_vld),
        .icing_path_len           (icing_path_len),
        .icing_path_len_vld       (icing_path_len_vld),
        .icing_path_idx           (icing_path_idx),
        .icing_path_idx_vld       (icing_path_idx_vld),
        .icing_last_hop_reached   (icing_last_hop_reached),
        .icing_iv                 (icing_iv),
        .icing_iv_vld             (icing_iv_vld),
        .icing_realm_id           (icing_realm_id),
        .icing_realm_id_0_vld     (icing_realm_id_0_vld),
        .icing_realm_id_1_vld     (icing_realm_id_1_vld),
        .icing_realm_id_2_vld     (icing_realm_id_2_vld),
        .icing_subrealm           (icing_subrealm),
        .icing_subrealm_vld       (icing_subrealm_vld),
        .icing_realm_is_first     (icing_realm_is_first),
        .icing_realm_is_last      (icing_realm_is_last),
        .icing_realm_is_me        (icing_realm_is_me),
        .icing_realm_is_b4_me     (icing_realm_is_b4_me),
        .icing_realm_is_nxt_hop   (icing_realm_is_nxt_hop),
        .icing_va                 (icing_va),
        .icing_va_0_vld           (icing_va_0_vld),
        .icing_va_1_vld           (icing_va_1_vld),
        .icing_vb                 (icing_vb),
        .icing_vb_vld             (icing_vb_vld),
        .icing_vtag_is_me         (icing_vtag_is_me),
        .icing_vtag_is_last       (icing_vtag_is_last),
        .icing_proto              (icing_proto),
        .icing_proto_vld          (icing_proto_vld),
        .icing_pkt_len            (icing_pkt_len),
        .icing_pkt_len_vld        (icing_pkt_len_vld),
        .icing_error_idx          (icing_error_idx),
        .icing_error_idx_vld      (icing_error_idx_vld),
        .icing_payload            (icing_payload),
        .icing_payload_is_last    (icing_payload_is_last),
        .icing_payload_vld        (icing_payload_vld),
        .word_count               (word_count[7:0]),
        .word_count_is_my_realm_id(word_count_is_my_realm_id),
        .word_count_is_my_vtag    (word_count_is_my_vtag),
        
        // --- Misc
        .clk                      (clk),
        .reset                    (reset));

   lookup_shared_keys
     #(.UDP_REG_SRC_WIDTH   (UDP_REG_SRC_WIDTH))
       lookup_shared_keys
       (// --- Interface to packet parser
        .icing_realm_id             (icing_realm_id),
        .icing_realm_id_0_vld       (icing_realm_id_0_vld),
        .icing_realm_id_1_vld       (icing_realm_id_1_vld),
        .icing_realm_id_2_vld       (icing_realm_id_2_vld),
        .icing_realm_is_b4_me       (icing_realm_is_b4_me),
        .icing_realm_is_first       (icing_realm_is_first),
        .icing_realm_is_me          (icing_realm_is_me),
        .icing_realm_is_nxt_hop     (icing_realm_is_nxt_hop),
        .icing_realm_is_last        (icing_realm_is_last),
        .icing_last_hop_reached     (icing_last_hop_reached),
        
        // --- Interface to flow control
        .pause                      (lookup_keys_pause),
        
        // --- Interface to SRAM
        .sram_rd_addr               (sram_rd_addr_0),
        .sram_rd_req                (sram_rd_req_0),
        .sram_rd_data               (sram_rd_data),
        .sram_rd_vld                (sram_rd_vld_0),
        .sram_rd_ack                (sram_rd_ack_0),
        
        // --- Results
        .shared_dh_key              (shared_dh_key),
        .my_realm_id_is_smaller     (my_realm_id_is_smaller),
        .shared_dh_key_missing      (shared_dh_key_missing),
        .shared_dh_key_vld          (shared_dh_key_vld),
        .rd_shared_key              (rd_shared_key),
        .shared_dh_key_first        (shared_dh_key_first),
        .shared_dh_key_last         (shared_dh_key_last),
        .shared_dh_key_b4_me        (shared_dh_key_b4_me),
        
        .next_hop_mac               (next_hop_mac),
        .output_port                (output_port),
        .next_hop_info_vld          (next_hop_info_vld),
        .rd_nxthop_info             (rd_nxthop_info),
        
        // --- register interface
        .reg_req_in                 (reg_req_in),
        .reg_ack_in                 (reg_ack_in),
        .reg_rd_wr_L_in             (reg_rd_wr_L_in),
        .reg_addr_in                (reg_addr_in),
        .reg_data_in                (reg_data_in),
        .reg_src_in                 (reg_src_in),
        
        .reg_req_out                (lookup_keys_reg_req_out),
        .reg_ack_out                (lookup_keys_reg_ack_out),
        .reg_rd_wr_L_out            (lookup_keys_reg_rd_wr_L_out),
        .reg_addr_out               (lookup_keys_reg_addr_out),
        .reg_data_out               (lookup_keys_reg_data_out),
        .reg_src_out                (lookup_keys_reg_src_out),
        
        // --- Misc
        .clk                        (clk),
        .reset                      (reset));

   lookup_poc_key lookup_poc_key
     (// --- Interface to packet parser
      .icing_subrealm         (icing_subrealm),
      .icing_subrealm_vld     (icing_subrealm_vld),
      .icing_realm_is_me      (icing_realm_is_me),
      
      // --- Interface to flow control
      .pause                  (poc_key_pause),
      
      // --- Interface to SRAM
      .sram_rd_addr           (sram_rd_addr_1),
      .sram_rd_req            (sram_rd_req_1),
      .sram_rd_data           (sram_rd_data),
      .sram_rd_vld            (sram_rd_vld_1),
      .sram_rd_ack            (sram_rd_ack_1),
      
      // --- Results
      .poc_key                (poc_key),
      .poc_key0               (poc_key0),
      .poc_key_vld            (poc_key_vld),
      
      // --- Misc
      .clk                    (clk),
      .reset                  (reset));

   hash_pkt hash_pkt
     (.in_ctrl                (out_ctrl_chk),
      // --- Interface to packet parser
      .icing_version          (icing_version),
      .icing_version_vld      (icing_version_vld),
      .icing_iv               (icing_iv),
      .icing_iv_vld           (icing_iv_vld),
      .icing_realm_id         (icing_realm_id),
      .icing_realm_id_0_vld   (icing_realm_id_0_vld),
      .icing_realm_id_1_vld   (icing_realm_id_1_vld),
      .icing_realm_id_2_vld   (icing_realm_id_2_vld),
      .icing_payload          (icing_payload),
      .icing_payload_is_last  (icing_payload_is_last),
      .icing_payload_vld      (icing_payload_vld),
      .icing_error_idx_vld    (icing_error_idx_vld),

      // --- Interface to flow control
      .pause                  (hash_pause),
      
      // --- Results
      .hash                   (hash),
      .hash_vld               (hash_vld),
      
      // --- Misc
      .clk                    (clk),
      .reset                  (reset));

   calculate_poc calculate_poc
     (// Outputs
      .pause                            (calc_poc_pause),
      .aes_enc_plain                    (aes_enc_plain_0[0:127]),
      .aes_enc_key                      (aes_enc_key_0[0:127]),
      .aes_enc_plain_vld                (aes_enc_plain_vld_0),
      .poc                              (poc[0:127]),
      .poc_vld                          (poc_vld),
      // Inputs
      .icing_realm_id                   (icing_realm_id[0:63]),
      .icing_realm_id_0_vld             (icing_realm_id_0_vld),
      .icing_realm_id_1_vld             (icing_realm_id_1_vld),
      .icing_realm_id_2_vld             (icing_realm_id_2_vld),
      .icing_realm_is_last              (icing_realm_is_last),
      .poc_key                          (poc_key[0:127]),
      .poc_key0                         (poc_key0[0:127]),
      .poc_key_vld                      (poc_key_vld),
      .aes_enc_plain_ack                (aes_enc_plain_ack_0),
      .aes_enc_cipher                   (aes_enc_cipher_0[0:127]),
      .aes_enc_cipher_vld               (aes_enc_cipher_vld_0),
      .clk                              (clk),
      .reset                            (reset));

   calculate_A calculate_A
     (// Outputs
      .mac2_key                         (mac2_key_0[0:127]),
      .mac2_block                       (mac2_block_0),
      .mac2_block_vld                   (mac2_block_vld_0),
      .A                                (A),
      .A_vld                            (A_vld),
      // Inputs
      .hash                             (hash[0:246]),
      .hash_vld                         (hash_vld),
      .poc                              (poc[0:127]),
      .poc_vld                          (poc_vld),
      .mac2_block_ack                   (mac2_block_ack_0),
      .mac2_tag                         (mac2_tag_0),
      .mac2_tag_vld                     (mac2_tag_vld_0),
      .clk                              (clk),
      .reset                            (reset));

   check_vtag check_vtag
     (// Outputs
      .rd_shared_key                    (rd_shared_key_0),
      .mac2_key                         (mac2_key_1[0:127]),
      .mac2_block                       (mac2_block_1),
      .mac2_block_vld                   (mac2_block_vld_1),
      .vtag_check_done                  (vtag_check_done),
      .va_invld                         (va_invld),
      .vb_invld                         (vb_invld),
      // Inputs
      .icing_va                         (icing_va[0:63]),
      .icing_va_0_vld                   (icing_va_0_vld),
      .icing_va_1_vld                   (icing_va_1_vld),
      .icing_vb                         (icing_vb[0:31]),
      .icing_vb_vld                     (icing_vb_vld),
      .icing_vtag_is_me                 (icing_vtag_is_me),
      .icing_vtag_is_last               (icing_vtag_is_last),
      .hash                             (hash[0:246]),
      .hash_vld                         (hash_vld),
      .shared_dh_key                    (shared_dh_key[0:127]),
      .my_realm_id_is_smaller           (my_realm_id_is_smaller),
      .shared_dh_key_vld                (shared_dh_key_vld),
      .shared_dh_key_last               (shared_dh_key_last),
      .shared_dh_key_b4_me              (shared_dh_key_b4_me),
      .A                                (A),
      .A_vld                            (A_vld),
      .mac2_block_ack                   (mac2_block_ack_1),
      .mac2_tag                         (mac2_tag_1),
      .mac2_tag_vld                     (mac2_tag_vld_1),
      .clk                              (clk),
      .reset                            (reset));

   calculate_nxt_vtags calculate_nxt_vtags
     (// Outputs
      .rd_shared_key                    (rd_shared_key_1),
      .mac2_key                         (mac2_key_2[0:127]),
      .mac2_block                       (mac2_block_2),
      .mac2_block_vld                   (mac2_block_vld_2),
      .prf_result                       (prf_result[0:95]),
      .prf_result_vld                   (prf_result_vld),
      .prf_result_last                  (prf_result_last),
      // Inputs
      .path_idx                         (icing_path_idx),
      .path_idx_vld                     (icing_path_idx_vld),
      .hash                             (hash[0:246]),
      .hash_vld                         (hash_vld),
      .shared_dh_key                    (shared_dh_key[0:127]),
      .my_realm_id_is_smaller           (my_realm_id_is_smaller),
      .shared_dh_key_vld                (shared_dh_key_vld),
      .shared_dh_key_last               (shared_dh_key_last),
      .shared_dh_key_b4_me              (shared_dh_key_b4_me),
      .rd_shared_key_chk_vtag           (rd_shared_key_0),
      .mac2_block_ack                   (mac2_block_ack_2),
      .mac2_tag                         (mac2_tag_2),
      .mac2_tag_vld                     (mac2_tag_vld_2),
      .clk                              (clk),
      .reset                            (reset));

   process_pkt process_pkt
     (// Outputs
      .out_rdy                          (out_rdy),
      .out_data                         (out_data[0:63]),
      .out_ctrl                         (out_ctrl[0:7]),
      .out_wr                           (out_wr),
      .rd_nxthop_info                   (rd_nxthop_info),
      // Inputs
      .in_data                          (out_data_chk),
      .in_ctrl                          (out_ctrl_chk),
      .in_wr                            (out_wr_chk),
      .in_rdy                           (in_rdy_process),
      .input_port                       (input_port[2:0]),
      .input_port_vld                   (input_port_vld),
      .word_count                       (word_count[7:0]),
      .word_count_is_my_realm_id        (word_count_is_my_realm_id),
      .word_count_is_my_vtag            (word_count_is_my_vtag),
      .icing_last_hop_reached           (icing_last_hop_reached),
      .vtag_check_done                  (vtag_check_done),
      .va_invld                         (va_invld),
      .vb_invld                         (vb_invld),
      .prf_result                       (prf_result[0:95]),
      .prf_result_vld                   (prf_result_vld),
      .prf_result_last                  (prf_result_last),
      .shared_dh_key_missing            (shared_dh_key_missing),
      .shared_dh_key_vld                (shared_dh_key_vld),
      .shared_dh_key_last               (shared_dh_key_last),
      .rd_shared_key                    (rd_shared_key),
      .next_hop_mac                     (next_hop_mac[0:47]),
      .output_port                      (output_port[7:0]),
      .next_hop_info_vld                (next_hop_info_vld),
      .port_macs                        (port_macs[0:48*4-1]),
      .my_realm_id                      (my_realm_id[0:159]),
      .clk                              (clk),
      .reset                            (reset));

   arbitrated_aes arbitrated_aes
     (// Outputs
      .block_out                        ({aes_enc_cipher_0,
                                          aes_enc_cipher_1,
                                          aes_enc_cipher_2}),
      .block_out_vld                    ({aes_enc_cipher_vld_0,
                                          aes_enc_cipher_vld_1,
                                          aes_enc_cipher_vld_2}),
      .block_in_ack                     ({aes_enc_plain_ack_0,
                                          aes_enc_plain_ack_1,
                                          aes_enc_plain_ack_2}),
      // Inputs
      .block_in                         ({aes_enc_plain_0,
                                          aes_enc_plain_1,
                                          aes_enc_plain_2}),
      .block_in_vld                     ({aes_enc_plain_vld_0,
                                          aes_enc_plain_vld_1,
                                          aes_enc_plain_vld_2}),
      .key                              ({aes_enc_key_0,
                                          aes_enc_key_1,
                                          aes_enc_key_2}),
      .clk                              (clk),
      .reset                            (reset));

   arbitrated_mac2 arbitrated_mac2
     (// Outputs
      .block_ack                        ({mac2_block_ack_0,
                                          mac2_block_ack_1,
                                          mac2_block_ack_2}),
      .tag                              ({mac2_tag_0,
                                          mac2_tag_1,
                                          mac2_tag_2}),
      .tag_vld                          ({mac2_tag_vld_0,
                                          mac2_tag_vld_1,
                                          mac2_tag_vld_2}),
      .aes_enc_plain_0                  (aes_enc_plain_1[0:127]),
      .aes_enc_key_0                    (aes_enc_key_1[0:127]),
      .aes_enc_plain_vld_0              (aes_enc_plain_vld_1),
      .aes_enc_plain_1                  (aes_enc_plain_2[0:127]),
      .aes_enc_key_1                    (aes_enc_key_2[0:127]),
      .aes_enc_plain_vld_1              (aes_enc_plain_vld_2),
      // Inputs
      .block                            ({mac2_block_0,
                                          mac2_block_1,
                                          mac2_block_2}),
      .block_vld                        ({mac2_block_vld_0,
                                          mac2_block_vld_1,
                                          mac2_block_vld_2}),
      .key                              ({mac2_key_0,
                                          mac2_key_1,
                                          mac2_key_2}),
      .aes_enc_plain_ack_0              (aes_enc_plain_ack_1),
      .aes_enc_cipher_0                 (aes_enc_cipher_1[0:127]),
      .aes_enc_cipher_vld_0             (aes_enc_cipher_vld_1),
      .aes_enc_plain_ack_1              (aes_enc_plain_ack_2),
      .aes_enc_cipher_1                 (aes_enc_cipher_2[0:127]),
      .aes_enc_cipher_vld_1             (aes_enc_cipher_vld_2),
      .clk                              (clk),
      .reset                            (reset));
   
   arbitrator 
     #(.SERV_DATA_WIDTH(64), 
       .SUPP_DATA_WIDTH(SRAM_ADDR_WIDTH),
       .FIFO_DEPTH_BITS(3), 
       .NUM_CLIENTS(2), 
       .USE_RESULTS(1))
     sram_rd_arbitrator
       (.client_supp_data   ({sram_rd_addr_0,
                              sram_rd_addr_1}),
        .client_serv_data   (128'h0),
        .client_req         ({sram_rd_req_0,
                              sram_rd_req_1}),
        .client_ack         ({sram_rd_ack_0,
                              sram_rd_ack_1}),
        .client_vld         ({sram_rd_vld_0,
                              sram_rd_vld_1}),
        .client_rslt_data   (sram_rd_data),
                            
        .server_supp_data   (rd_0_addr),
        .server_serv_data   (),
        .server_req         (rd_0_req),
        .server_ack         (rd_0_ack),
        .server_vld         (rd_0_vld),
        .server_rslt_data   (rd_0_data[63:0]),
                            
        .clk                (clk),
        .reset              (reset)
        );

   generic_regs
     #(.UDP_REG_SRC_WIDTH (UDP_REG_SRC_WIDTH),
       .TAG (`ICING_LOOKUP_BLOCK_TAG),
       .REG_ADDR_WIDTH (`ICING_LOOKUP_REG_ADDR_WIDTH),
       .RESET_ON_READ (0),
       .NUM_COUNTERS (10),
       .NUM_SOFTWARE_REGS ((160+4*48)/32),
       .NUM_HARDWARE_REGS (0),
       .COUNTER_INPUT_WIDTH (1),
       .REVERSE_WORD_ORDER (1)
       )
   generic_regs
     (
      .reg_req_in        (lookup_keys_reg_req_out),
      .reg_ack_in        (lookup_keys_reg_ack_out),
      .reg_rd_wr_L_in    (lookup_keys_reg_rd_wr_L_out),
      .reg_addr_in       (lookup_keys_reg_addr_out),
      .reg_data_in       (lookup_keys_reg_data_out),
      .reg_src_in        (lookup_keys_reg_src_out),
                         
      .reg_req_out       (reg_req_out),
      .reg_ack_out       (reg_ack_out),
      .reg_rd_wr_L_out   (reg_rd_wr_L_out),
      .reg_addr_out      (reg_addr_out),
      .reg_data_out      (reg_data_out),
      .reg_src_out       (reg_src_out),
                         
      // --- counters interface
      .counter_updates   ({dst_mac_not_ok,
                           ethertype_not_ok,
                           path_idx_not_ok,
                           path_len_not_ok,
                           pkt_len_not_ok,
                           version_not_ok,
                           shared_dh_key_missing,
                           vtag_check_done & !va_invld,
                           vtag_check_done & va_invld,
                           vtag_check_done & vb_invld}),
      .counter_decrement (10'h0),

      // --- SW regs interface
      .software_regs     ({my_realm_id,
                           port_macs}),
                         
      // --- HW regs interface
      .hardware_regs     (1'b0),

      .clk               (clk),
      .reset             (reset));

   assign rd_shared_key = rd_shared_key_0 | rd_shared_key_1;
   assign out_rdy_chk = !hash_pause & in_rdy_process && !lookup_keys_pause && !poc_key_pause && !calc_poc_pause;

endmodule // output_port_lookup
