/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: poly_mac.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *
 *    POLY-MAC-128(key1[0..127], key2[0..127], block[0..254]) {
 *        // '.' is poly. mult. in GF(2^{128})
 *        X <- bits [0..126] of (block[0..127] . key2)
 *        X <- X XOR block[128..254]
 *        Tag <- AES-ENC(key1, 1 || X)
 *        output Tag
 *    }
 *******************************************************************************/

module poly_mac
  (input [0:127]  key1,
   input [0:127]  key2,
   input [0:254]  block,
   input          block_vld,
   output         block_ack,
   
   output [0:127] tag,
   output         tag_vld,

   // --- Interface to shared AES core
   output     [0:127] aes_enc_plain,
   output     [0:127] aes_enc_key,
   output             aes_enc_plain_vld,
   input              aes_enc_plain_ack,
   input      [0:127] aes_enc_cipher,
   input              aes_enc_cipher_vld,

   input              clk,
   input              reset
   );

   wire [0:127]   Xtemp;
   
   mult128 mult(block[0:127], key2, Xtemp);
   assign aes_enc_plain       = Xtemp[0:126] ^ block[128:254];
   assign aes_enc_key         = key1;
   assign aes_enc_plain_vld   = block_vld;

   assign tag                 = aes_enc_cipher;
   assign tag_vld             = aes_enc_cipher_vld;
endmodule // poly_mac

module mult128
  (input [0:127] a,
   input [0:127] b,
   output [0:127] p);

   generate 
      genvar      i;
      for(i=0; i<128; i=i+1) begin:gen_temp
         wire [i:127] temp;
         assign temp = a[i:127] & b[0:127-i];
      end

      for(i=0; i<128; i=i+1) begin:gen_out
         assign p[i] = ^gen_temp[i].temp;
      end
   endgenerate

   // synthesis translate_off
   reg [0:127]  temp_a, verif_p, temp_b;
   always @(*) begin
      temp_a = a;
      temp_b = b;
      verif_p = 0;
      repeat(128) begin
         if(temp_b[0]) verif_p = verif_p ^ temp_a;
         temp_a = temp_a << 1;
         temp_b = temp_b >> 1;
      end
      if(verif_p != p) begin
         $display("%t %m ERROR:calculating multiplication.", $time);
      end
   end // always @ (*)
   // synthesis translate_on   
endmodule // mult128

