/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: process_pkt.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Does one of the following:
 *   - Drop the pkt on error or if credentials are wrong
 *   - Send pkt to software if key missing
 *   - update the vtags and forward the pkt if all checks out
 * 
 * WARNING: Assumes there is only one module hdr---the IOQ hdr.
 ******************************************************************************/

module process_pkt
  (// Interface to datapath
   input [0:63]       in_data,
   input [0:7]        in_ctrl,
   input              in_wr,
   output reg         in_rdy,
  
   output reg [0:63]  out_data,
   output     [0:7]   out_ctrl,
   output reg         out_wr,
   input              out_rdy,
  
   // --- Interface to parser
   input [2:0]        input_port,
   input              input_port_vld,
   input [7:0]        word_count,
   input              word_count_is_my_realm_id,
   input              word_count_is_my_vtag,
   input              icing_last_hop_reached,
  
   // --- Interface to check_vtag
   input              vtag_check_done,
   input              va_invld,
   input              vb_invld,

   // --- Interface to next vtags
   input      [0:95]  prf_result,
   input              prf_result_vld,
   input              prf_result_last,

   // --- Interface to key lookup
   input              shared_dh_key_missing,
   input              shared_dh_key_vld,
   input              rd_shared_key,
   input              shared_dh_key_last,

   // --- Interface to nxt hop lookup
   input      [0:47]  next_hop_mac,
   input      [7:0]   output_port,
   input              next_hop_info_vld,
   output reg         rd_nxthop_info,

   // --- Interface to registers
   input  [0:48*4-1]  port_macs,
   input  [0:159]     my_realm_id,

   // --- Misc
   input              clk,
   input              reset);

   //------------------------ localparams ----------------------------
   localparam         CHECK_AUTH = 2,
                      WRITE_ETH_DST_SRC = 4,
                      WRITE_SRC_MAC = 8,
                      INCREMENT_PATH_IDX = 16, //0x10
                      SKIP_TO_MY_VTAG = 32, //0x20
                      UPDATE_VTAG_0 = 64, //0x40
                      UPDATE_VTAG_1 = 128, //0x80
                      WAIT_EOP = 256, //0x100
                      DO_DISCARD = 512, //0x200
                      SEND_TO_CPU = 1; //0x1

   //------------------------ Wires/Regs -----------------------------
   wire [0:63]        in_fifo_data;
   wire [1:0]         inport;
   wire [7:0]         word_count_my_vtag;
   wire [0:95]        nxt_vtag;
   reg [7:0]          cpu_port;
   reg                key_missing;
   wire [0:47]        port_mac[0:3];
   reg [0:47]         src_mac;
   reg [10:0]         state, state_nxt;
   reg                in_fifo_rd_en, vtag_chk_fifo_rd_en,
                      wc_vtag_fifo_rd_en, key_miss_fifo_rd_en;
   reg [7:0]          count, count_nxt;
   reg                nxt_vtag_fifo_rd_en;
   reg                inport_fifo_rd_en;

   //------------------------- Modules -------------------------------
   fallthrough_fifo_512x72 in_fifo
     (.din           ({in_ctrl, in_data}),
      .wr_en         (in_wr),
      .rd_en         (in_fifo_rd_en),
      .dout          ({out_ctrl, in_fifo_data}),
      .full          (in_fifo_full),
      .almost_full   (in_fifo_almost_full),
      .empty         (in_fifo_empty),
      .rst           (reset),
      .clk           (clk)
      );
   
   fallthrough_small_fifo 
     #(.WIDTH(2),
       .MAX_DEPTH_BITS(5))
       inport_fifo
         (.din           (input_port[2:1]),
          .wr_en         (input_port_vld),
          .rd_en         (inport_fifo_rd_en),
          .dout          (inport),
          .full          (inport_fifo_full),
          .prog_full     (),
          .nearly_full   (inport_fifo_almost_full),
          .empty         (inport_fifo_empty),
          .reset         (reset),
          .clk           (clk)
          );

   fallthrough_small_fifo 
     #(.WIDTH(1),
       .MAX_DEPTH_BITS(5))
       key_miss_fifo
         (.din           (key_missing | shared_dh_key_missing),
          .wr_en         (key_miss_fifo_wr_en),
          .rd_en         (key_miss_fifo_rd_en),
          .dout          (key_missing_dout),
          .full          (key_miss_fifo_full),
          .prog_full     (),
          .nearly_full   (),
          .empty         (key_miss_fifo_empty),
          .reset         (reset),
          .clk           (clk)
          );

   fallthrough_small_fifo 
     #(.WIDTH(2),
       .MAX_DEPTH_BITS(7))
       vtag_chk_fifo
         (.din           ({va_invld, vb_invld}),
          .wr_en         (vtag_check_done),
          .rd_en         (vtag_chk_fifo_rd_en),
          .dout          ({va_invld_dout, vb_invld_dout}),
          .full          (vtag_chk_fifo_full),
          .prog_full     (),
          .nearly_full   (),
          .empty         (vtag_chk_fifo_empty),
          .reset         (reset),
          .clk           (clk)
          );
   
   /* -----\/----- EXCLUDED -----\/-----
    fallthrough_small_fifo 
    #(.WIDTH(8),
    .MAX_DEPTH_BITS(1))
    wc_id_fifo
    (.din           (word_count),
    .wr_en         (word_count_is_my_realm_id),
    .rd_en         (wc_id_fifo_rd_en),
    .dout          (word_count_my_id),
    .full          (wc_id_fifo_full),
    .prog_full     (),
    .nearly_full   (),
    .empty         (wc_id_fifo_empty),
    .reset         (reset),
    .clk           (clk)
    );
    -----/\----- EXCLUDED -----/\----- */

   fallthrough_small_fifo 
     #(.WIDTH(9),
       .MAX_DEPTH_BITS(5))
       wc_vtag_fifo
         (.din           ({icing_last_hop_reached, word_count}),
          .wr_en         (word_count_is_my_vtag),
          .rd_en         (wc_vtag_fifo_rd_en),
          .dout          ({last_hop, word_count_my_vtag}),
          .full          (),
          .prog_full     (),
          .nearly_full   (wc_vtag_fifo_almost_full),
          .empty         (wc_vtag_fifo_empty),
          .reset         (reset),
          .clk           (clk)
          );

   fallthrough_small_fifo 
     #(.WIDTH(97),
       .MAX_DEPTH_BITS(7))
       nxt_vtag_fifo
         (.din           ({prf_result_last, prf_result}),
          .wr_en         (prf_result_vld),
          .rd_en         (nxt_vtag_fifo_rd_en),
          .dout          ({nxt_vtag_last, nxt_vtag}),
          .full          (nxt_vtag_fifo_full),
          .prog_full     (),
          .nearly_full   (),
          .empty         (nxt_vtag_fifo_empty),
          .reset         (reset),
          .clk           (clk)
          );

   //-------------------------- Logic --------------------------------
   always @(*) begin
      in_rdy = ~in_fifo_almost_full;
   end
   
   always @(*) begin
      cpu_port                 = 0;
      cpu_port[{inport, 1'b1}] = 1'b1;
   end

   /* latch stat signals to synchronize them */
   always @(posedge clk) begin
      if(reset) begin
         key_missing           <= 0;
      end else begin
         if(shared_dh_key_vld & shared_dh_key_last) begin
            key_missing <= 0;
         end else if(shared_dh_key_vld) begin
            key_missing <= key_missing | shared_dh_key_missing;
         end
      end // else: !if(reset)
   end // always @ (posedge clk)

   assign key_miss_fifo_wr_en = (shared_dh_key_last && rd_shared_key);

   generate
      genvar i;
      /* split up port macs */
      for(i=0; i<4; i=i+1) begin:gen_mac_words
         assign port_mac[i] = port_macs[i*48 +: 48];
      end
   endgenerate

   always @(port_mac[0], port_mac[1], port_mac[2], 
            port_mac[3], in_data, output_port) begin
      if(output_port[0]) begin
         src_mac = port_mac[0];
      end else if(output_port[2]) begin
         src_mac = port_mac[1];
      end else if(output_port[4]) begin
         src_mac = port_mac[2];
      end else if(output_port[6]) begin
         src_mac = port_mac[3];
      end else begin
         /* keep the same src mac */
         src_mac = {in_data[48:63], in_data[0:31]};
      end
   end // always @ (*)

   always @(*) begin
      state_nxt                    = state;
      in_fifo_rd_en                = 0;
      key_miss_fifo_rd_en          = 0;
      vtag_chk_fifo_rd_en          = 0;
      wc_vtag_fifo_rd_en           = 0;
      out_data                     = in_fifo_data;
      out_wr                       = 0;
      count_nxt                    = count;
      nxt_vtag_fifo_rd_en          = 0;
      inport_fifo_rd_en            = 0;
      rd_nxthop_info               = 0;
      
      case(state)
         /* Wait for authorization */
         CHECK_AUTH: begin
            if(!vtag_chk_fifo_empty && !key_miss_fifo_empty) begin
               /* If the authorization fails */
               if(!vb_invld_dout && key_missing_dout) begin
                  /* send the packet to the CPU for slow path processing */
                  state_nxt                 = SEND_TO_CPU;
                  // synthesis translate_off
                  $display($time, " WARNING: Sending packet to CPU because key missing.");
                  // synthesis translate_on
               end // if (!vb_invld_dout && va_invld_dout && key_missing_dout)

               /* Otherwise, packet is bad, so drop packet data and
                * preprocessing results */
               else if(vb_invld_dout || va_invld_dout) begin
                  state_nxt       = DO_DISCARD;
                  in_fifo_rd_en   = 1; // This is necessary to throw out IOQ hdr
                  // synthesis translate_off
                  if(va_invld_dout) $display($time, " WARNING: Dropping pkt at %m because Va is wrong.");
                  if(vb_invld_dout) $display($time, " WARNING: Dropping pkt at %m because Vb is wrong.");
                  // synthesis translate_on
               end

               /* Authorization succeeded, so wait for the next hop info
                * and update packet. */
               else if(out_rdy && next_hop_info_vld) begin
                  out_data[0:15]        = {8'h0, output_port};
                  in_fifo_rd_en         = 1;
                  out_wr                = 1;
                  state_nxt             = WRITE_ETH_DST_SRC;
               end // else: !if(va_invld_dout)
            end // if (!vtag_chk_fifo_empty)
         end // case: CHECK_AUTH

         WRITE_ETH_DST_SRC: begin
            if(out_rdy) begin
               out_data        = {next_hop_mac, src_mac[0:15]};
               in_fifo_rd_en   = 1;
               out_wr          = 1;
               state_nxt       = WRITE_SRC_MAC;
            end
         end

         WRITE_SRC_MAC: begin
            if(out_rdy) begin
               out_data[0:31]      = src_mac[16:47];
               in_fifo_rd_en       = 1;
               out_wr              = 1;
               // TODO: We should check the realm ID for a match!
               state_nxt           = INCREMENT_PATH_IDX;
            end
         end

         INCREMENT_PATH_IDX: begin
            if(out_rdy) begin
               out_data[0:7]    = in_fifo_data[0:7] + 1'b1;
               in_fifo_rd_en    = 1;
               out_wr           = 1;
               count_nxt        = 4;
               state_nxt        = SKIP_TO_MY_VTAG;
            end
         end

         /* Count until my vtag arrives */
         SKIP_TO_MY_VTAG: begin
            if(out_rdy) begin
               in_fifo_rd_en   = 1;
               out_wr          = 1;
               count_nxt       = count + 1'b1;
               if(count == word_count_my_vtag) begin
                  state_nxt            = last_hop ? WAIT_EOP : UPDATE_VTAG_0;
               end
            end
         end // case: SKIP_TO_MY_VTAG

         /* XOR the first 64 bits of va */
         UPDATE_VTAG_0: begin
            if(out_rdy & !nxt_vtag_fifo_empty) begin
               in_fifo_rd_en   = 1;
               out_wr          = 1;
               out_data        = in_fifo_data ^ nxt_vtag[0:63];
               state_nxt       = UPDATE_VTAG_1;
            end
         end

         /* XOR the next 32 bits of the vtag and
          * wait for eop if we are done */
         UPDATE_VTAG_1: begin
            if(out_rdy) begin
               in_fifo_rd_en         = 1;
               out_wr                = 1;
               out_data[0:31]        = in_fifo_data[0:31] ^ nxt_vtag[64:95];
               nxt_vtag_fifo_rd_en   = 1;
               if(nxt_vtag_last) begin
                  state_nxt = WAIT_EOP;
               end else begin
                  state_nxt = UPDATE_VTAG_0;
               end
            end // if (out_rdy)
         end // case: UPDATE_VTAG_1

         /* continue reading until packet done */
         WAIT_EOP: begin
            if(out_rdy) begin
               in_fifo_rd_en   = 1;
               out_wr          = 1;
               if(out_ctrl != 0) begin
                  /* read all the FIFOs */
                  wc_vtag_fifo_rd_en    = 1;
                  vtag_chk_fifo_rd_en   = 1;
                  key_miss_fifo_rd_en   = 1;
                  inport_fifo_rd_en     = 1;
                  rd_nxthop_info        = 1;
                  state_nxt             = CHECK_AUTH;
               end
            end
         end

         /* Do the actual discards */
         DO_DISCARD: begin
            /* if we are at the last word,
             * then wait till everythinge else can be discarded */
            if(out_ctrl != 0 && out_ctrl != 8'hff) begin
               if(!nxt_vtag_fifo_empty && nxt_vtag_last) begin
                  wc_vtag_fifo_rd_en    = 1;
                  vtag_chk_fifo_rd_en   = 1;
                  rd_nxthop_info        = 1;
                  key_miss_fifo_rd_en   = 1;
                  inport_fifo_rd_en     = 1;
                  nxt_vtag_fifo_rd_en   = 1;
                  in_fifo_rd_en         = 1;
                  state_nxt             = CHECK_AUTH;
               end
               else if(!nxt_vtag_fifo_empty) begin
                  nxt_vtag_fifo_rd_en   = 1;
               end
            end
            /* otherwise, just empty the input fifo */
            else begin
               in_fifo_rd_en = 1;
               if(!nxt_vtag_fifo_empty && !nxt_vtag_last) begin
                  nxt_vtag_fifo_rd_en   = 1;
               end
            end
         end // case: DO_DISCARD

         /* send the packet unmodified to the CPU */
         SEND_TO_CPU: begin
            /* If we are at the IOQ hdr then write the CPU port */
            if(out_ctrl == 8'hff) begin
               in_fifo_rd_en       = 1;
               out_wr              = 1;
               out_data[0:15]      = {8'h0, cpu_port};
            end

            /* If we are at the last word*/
            else if(out_ctrl != 0) begin
               /* wait until we have read the last nxt_vtag */
               if(!nxt_vtag_fifo_empty && nxt_vtag_last) begin
                  wc_vtag_fifo_rd_en    = 1;
                  vtag_chk_fifo_rd_en   = 1;
                  rd_nxthop_info        = 1;
                  key_miss_fifo_rd_en   = 1;
                  inport_fifo_rd_en     = 1;
                  nxt_vtag_fifo_rd_en   = 1;
                  in_fifo_rd_en         = 1;
                  out_wr                = 1;
                  state_nxt             = CHECK_AUTH;
               end
            end // if (out_ctrl != 0)

            /* otherwise, send the unmodified packet */
            else begin
               in_fifo_rd_en   = 1;
               out_wr          = 1;
               if(!nxt_vtag_fifo_empty && !nxt_vtag_last) begin
                  nxt_vtag_fifo_rd_en   = 1;
               end
            end // else: !if(out_ctrl != 0)
         end // case: SEND_TO_CPU
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if(reset) begin
         state                     <= CHECK_AUTH;
         count                     <= 3;
      end 
      else begin
         state                     <= state_nxt;
         count                     <= count_nxt;
      end // else: !if(reset)
   end // always @ (posedge clk)

   // synthesis translate_off
   integer pkt_done_count = 0;
   always @(posedge clk) begin
      if(state != CHECK_AUTH && state_nxt == CHECK_AUTH) begin
         pkt_done_count <= pkt_done_count + 1'b1;
         $display($time, " %m Finished packet #%u", pkt_done_count);
      end
   end
   // synthesis translate_on
endmodule // process_pkt
