/* ****************************************************************************
 * $Id$
 *
 * Module: cli.c
 * Project: ICING2 Forwarder
 * Author: Jad Naous
 * Description: interface to read and write entries to the ICING hardware
 *
 * Change history:
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <net/if.h>

#include <time.h>

#include <reg_defines.h>
#include "../lib/C/reg_defines_icing_forwarder.h"
#include <nf2util.h>
#include "../lib/C/nf_icing3.h"
#include "../lib/C/nf_icing3_types.h"
#include "../lib/C/nf_icing3_low_level.h"

#define PATHLEN         80

#define DEFAULT_IFACE   "nf2c0"

/* Global vars */
static nf_icing_dev_config_t config;

static uint8_t src_mac_port[24] = { 0, 0, 0x55, 0, 0, 3,
				    0, 0, 0x55, 0, 0, 3,
				    0, 0, 0x55, 0, 0, 3,
				    0, 0, 0x55, 0, 0, 3};

static uint8_t next_hop_mac[6] = {0, 0, 0xdd, 0, 0, 0};
static uint8_t realm_ids[20] = {0x00, 0xca, 0xfe, 0x00,
				0x00, 0x00, 0x00, 0x01,
				0x00, 0x00, 0x00, 0x01,
				0x00, 0x00, 0x00, 0x01,
				0x00, 0x00, 0x00, 0x00};
static uint8_t my_realm_id[20] = {20, 21, 22, 23, 24, 25,26,27,28,29,30,31,32,33,34,35};
static uint8_t shared_key[16];
static int write_entries = 0;
static int dump_entries  = 0;
static int write_config  = 0;
static int dump_config  = 0;
static int write_poc_keys = 0;
static int read_poc_keys = 0;

static void print_buff(uint8_t *buff, unsigned len) {
  while(len-- > 1) {
    printf("%02x:", *(buff++));
  }
  printf("%02x", *buff);
}

static void usage();
static void processArgs(int, char **);

int main(int argc, char *argv[]) {
  int i,j;
  nf_icing_dev_t dev;
  nf_icing_realm_info_t info;
  uint8_t poc_key[16];
  uint8_t vnode[4];

  config.dev_name = DEFAULT_IFACE;
  memcpy(config.macs, src_mac_port, 24);
  memcpy(&config.realm_id, my_realm_id, 20);

  processArgs(argc, argv);

  nf_icing3_init_dev(&dev, &config, 1, 1, write_config);

  if(write_poc_keys) {
    for(i=0; i<(1<<VNODE_WIDTH); i=i+1) {
      vnode[0] = i>>8;
      vnode[1] = i & 0xff;
      vnode[2] = 0;
      vnode[3] = 0;
      memcpy(poc_key + 0, vnode, 4);
      memcpy(poc_key + 4, vnode, 4);
      memcpy(poc_key + 8, vnode, 4);
      memcpy(poc_key + 12, vnode, 4);
      nf_icing3_write_poc_key((vnode_t *)&vnode, (poc_key_t *)poc_key, &dev);
    }
  }

  if(read_poc_keys) {
    for(i=0; i<512; i=i+8) {
      vnode[0] = i>>8;
      vnode[1] = i & 0xff;
      vnode[2] = 0;
      vnode[3] = 0;
      nf_icing3_read_poc_key((vnode_t *)vnode, (poc_key_t *)poc_key, &dev);
      printf("Vnode: "); print_buff(vnode, 4); 
      printf(" PoC Key: "); print_buff(poc_key, 16);
      printf("\n");
    }
  }

  for(i=0; i<write_entries; i++) {
    realm_ids[19] = realm_ids[15] = realm_ids[11] = realm_ids[7] = i;
    next_hop_mac[5] = i;
    for(j=0; j<16; j++) {
      shared_key[j] = my_realm_id[j]^realm_ids[j];
    }
    bzero(&info, sizeof info);
    memcpy(&info.realm_id, realm_ids, 20);
    memcpy(info.shared_key, shared_key, 16);
    info.output_mask = 1<<((i&0x3)*2);
    memcpy(info.next_hop_mac, next_hop_mac, 6);
    nf_icing3_write_realm(&info, &dev);
  }

  for(i=0; i<dump_entries; i++) {
    bzero(&info, sizeof info);
    realm_ids[19] = realm_ids[15] = realm_ids[11] = realm_ids[7] = i;
    memcpy(&info.realm_id, realm_ids, 20);
    nf_icing3_read_realm(&info, &dev);
    printf("%02u: ", i);
    print_buff(info.realm_id.id, 20);
    printf(", ");
    print_buff(info.shared_key, 16);
    printf(", ");
    print_buff(info.next_hop_mac, 6);
    printf(", 0x%02x\n", info.output_mask);
  }
  
  if(dump_config) {
    nf_icing3_read_port_mac_addrs(src_mac_port, &dev.nf2);
    nf_icing3_read_realm_id(my_realm_id, &dev.nf2);

    printf("mac 0: "); print_buff(src_mac_port, 6); printf("\n");
    printf("mac 1: "); print_buff(src_mac_port+6, 6); printf("\n");
    printf("mac 2: "); print_buff(src_mac_port+12, 6); printf("\n");
    printf("mac 3: "); print_buff(src_mac_port+18, 6); printf("\n");
    printf("my realm id: "); print_buff(my_realm_id, 20); printf("\n");
  }

  nf_icing3_close_dev(&dev);

  return 0;
}

/* 
 *  Process the arguments.
 */
static void processArgs(int argc, char **argv) {
  char c;

  /* don't want getopt to moan - I can do that just fine thanks! */
  opterr = 0;

  while ((c = getopt(argc, argv, "i:w:d:kurph")) != -1) {
    switch (c) {
    case 'i': /* interface name */
      config.dev_name = optarg;
      break;
    case 'w': /* write entries */
      write_entries = atoi(optarg);
      break;
    case 'd': /* dump entries */
      dump_entries = atoi(optarg);
      break;
    case 'u': /* write uhf */
      write_config = 1;
      break;
    case 'r': /* read config */
      dump_config = 1;
      break;
    case 'p': /* write poc keys */
      write_poc_keys = 1;
      break;
    case 'k': /* read poc keys */
      read_poc_keys = 1;
      break;
    case 'h':
      usage();
      exit(1);
    default:
      fprintf(stderr,
              "Unknown option character `\\x%x'.\n",
              optopt);
      usage();
      exit(1);
    }
  }
}

/*
 *  Describe usage of this program.
 */
static void usage(void) {
  printf("Usage: ./cli <options> \n\n");
  printf("Options: -i <iface> : interface name (default nf2c0)\n");
  printf("         -w <num>: Write num entries.\n");
  printf("         -d <num>: Dump num entries.\n");
  printf("         -u : Write UHF.\n");
  printf("         -r : Dump config.\n");
  printf("         -p : Write poc key table.\n");
  printf("         -k : Read poc key table.\n");
  printf("         -h : Print this message and exit.\n");
}
