/**
 * Filename: client.c
 * Purpose: defines a client which continuously monitors the NetFPGA and reports
 *          its observations to a server.
 */

/* so we can use rw lock */
#define _GNU_SOURCE

#include <arpa/inet.h> /* inet_aton */
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>    /* strtoul, system */
#include <string.h>    /* strlen */
#include <time.h>      /* time */
#include <unistd.h>    /* sleep */
#include "common.h"
#include "debug.h"
#include "nf2util.h"
#include "protocol.h"

/** default number of seconds between updates */
#define DEFAULT_UPDATE_INTERVAL 2
#define MAX_UPDATE_INTERVAL 3600

/** number of seconds between pings */
#define PING_INTERVAL 2

/** default name of the software */
#define DEFAULT_SR_NAME "sr"
#define MAX_SR_NAME_LEN 32

/** name of the ping target file */
#define DEFAULT_PING_TARGET_FILE "ping_list"

/** encapsulates ping target info */
typedef struct {
    nf2_ping_t stats;    /* stats about pings to this target */
    uint32_t   last;     /* information about which of the last MAX_PING_COUNT
                            (up to 32) pings succeeded */
} ping_target_t;

/** encapsulates a route and who to ping to see if it works */
typedef struct {
    addr_ip_t subnet;
    addr_ip_t mask;
    unsigned num_targets;
    ping_target_t target[MAX_PING_TARGETS_PER_SUBNET];
} ping_info_t;

/** Encapsulates information for a single client. */
typedef struct {
    nf2_device_t nf2;               /* contains info about the NetFPGA */
    int fd;                         /* socket descriptor for UDP to server */
    struct sockaddr_in server_addr; /* address of the server */
    unsigned interval;              /* how often to send updates */
    const char* sr_name;            /* name of the sr program */
    const char* group;              /* group name */

    nf2_rt_t rt_prev[ROUTER_RT_SIZE]; /* last routing table */

    pthread_rwlock_t ping_lock;     /* protects shared access to ping */
    ping_info_t ping[MAX_PING_TARGETS]; /* info about what to ping */
} client_t;

typedef struct {
    client_t* client;
    int ping_index;   /* the index into ping info array which we're resp. for */
    int target_index; /* target index we're responsible for (-1 if all) */
} ping_handler_info_t;

static void ping_fail( nf2_ping_t* p, uint32_t* last ) {
    *last = ((*last << 1) + 0) & MAX_PING_COUNT_MASK;
    p->num_ok = ones(*last);

    if( ((int16_t)ntohs(p->streak)) > 0 )
        p->streak = 0; /* lost the success streak */

    if( ((int16_t)ntohs(p->streak)) > INT16_MIN )
        p->streak = htons( ntohs(p->streak) - 1 );
}

static void ping_ok( nf2_ping_t* p, uint32_t* last ) {
    *last = ((*last << 1) + 1) & MAX_PING_COUNT_MASK;
    p->num_ok = ones(*last);

    if( ((int16_t)ntohs(p->streak)) < 0 )
        p->streak = 0; /* lost the failure streak */

    if( ((int16_t)ntohs(p->streak)) < INT16_MAX )
        p->streak = htons( ntohs(p->streak) + 1 );
}

/** Will regularly ping the target specified by the pphi->index field. */
static void* ping_handler_main( void* pphi ) {
    /* get a pointer to the route we'll be pinging */
    ping_handler_info_t* phi = (ping_handler_info_t*)pphi;
    client_t* client = phi->client;
    ping_info_t* ping = &client->ping[phi->ping_index];
    int index = phi->target_index;

    /* initialize our thread */
    char name[32];
    my_snprintf( name, 32, "Ping Thread %u:%u", phi->ping_index, phi->target_index );
    debug_pthread_init( "Ping Thread", "Ping Thread" );
    free( phi );

    char ping_cmd[64];
    char str_ip[STRLEN_IP];
    if( index != -1 ) {
        /* build the IP we want to ping */
        ip_to_string( str_ip, ping->target[index].stats.ip );

        /* create the ping cmd */
        my_snprintf( ping_cmd, 64, "ping -c 1 -w 1 %s > /dev/null", str_ip );
    }

    /* initialize stats */
    unsigned start, end, at;
    if( index == -1 ) {
        start = 0;
        end = ping->num_targets;
    }
    else {
        start = end = index;
    }

    for( at=start; at<end; at++ ) {
        ping->target[at].stats.num_ok = 0;
        ping->target[at].stats.streak = 0;
        ping->target[at].last = 0;
    }
    at = (index == - 1) ? 0 : index;

    /* send a ping at a given interval */
    while( 1 ) {
        sleep( PING_INTERVAL );

        /* ping our target */
        bool ok;
        if( system( ping_cmd ) == 0 )
            ok = TRUE;
        else
            ok = FALSE;

        /* we will only write our own entry (no other readers will touch it) */
        pthread_rwlock_rdlock( &client->ping_lock );

        /* update the ping stats */
        if( !ok )
            ping_fail( &ping->target[at].stats, &ping->target[at].last );
        else
            ping_ok(   &ping->target[at].stats, &ping->target[at].last );

        /* done */
        pthread_rwlock_unlock( &client->ping_lock );

        /* round-robin if we're responsible for pinging multiple targets */
        if( index == -1 ) {
            at = (at + 1) % ping->num_targets;
            ip_to_string( str_ip, ping->target[at].stats.ip );
            my_snprintf( ping_cmd, 64, "ping -c 1 -w 1 %s > /dev/null", str_ip );
        }
    }

    return NULL;
}

typedef struct {
    char str_ip[32];
} str_target_ip_t;

/**
 * Parses the ping target file to determine what IPs to ping to
 * determine subnet connectivity.
 */
void parse_ping_target_file( bool count_per_subnet, ping_info_t* ping, const char* filename ) {
    FILE* fp;
    const char* err;
    char  line[512];
    char  str_subnet[32];
    char  str_mask[32];
    str_target_ip_t str_target[MAX_PING_TARGETS_PER_SUBNET];
    struct in_addr subnet, mask, target;
    unsigned i, num_targets;

    err = "Error loading target ping table,";
    debug_println( "Loading target ping table from %s", filename );

    fp = fopen( filename, "r" );
    if( !fp )
        die( "Could  not access the target ping file named %s", filename );

    i = 0;
    while( fgets(line,512,fp) != 0) {
        if( i == MAX_PING_TARGETS )
            die( "Error: too many ping targets (max = %u)", MAX_PING_TARGETS );

        int matches = sscanf( line, "%s %s %s %s %s %s %s",
                              str_subnet,
                              str_mask,
                              str_target[0].str_ip, str_target[1].str_ip,
                              str_target[2].str_ip, str_target[3].str_ip, str_target[4].str_ip );

        if( matches < 3 )
            die( "Error: bad entry in ping target file: %s (format: <subnet> <mask> <target_ip1> [<target_ip2> [... [<target_ip5>] ...]])", line );

        /* all other matches are target IPs */
        num_targets = matches - 2;

        if( inet_aton(str_subnet,&subnet) == 0 )
            die( "%s cannot convert subnet (%s) to valid IP", err, str_subnet );

        if( inet_aton(str_mask,&mask) == 0 )
            die( "%s cannot convert mask (%s) to valid IP", err, str_mask );

        unsigned j;
        for( j=0; j<num_targets; j++ ) {
            if( inet_aton(str_target[j].str_ip,&target) == 0 )
                die( "%s cannot convert target IP (%s) to valid IP",
                     err, str_target[j].str_ip );
            else
                ping[i].target[j].stats.ip = target.s_addr;
        }

        ping[i].subnet      = subnet.s_addr;
        ping[i].mask        = mask.s_addr;
        ping[i].num_targets = num_targets;
        i += 1;
    }
}

/** Parse the command-line arguments to populate client. */
static void parse_cli_args( int argc, char** argv, client_t* client ) {
    bool got_req_args = FALSE;
    bool do_pings     = TRUE;
    bool count_by_subnet = FALSE;
    unsigned advance;
    char ping_target_file[MAX_SR_NAME_LEN];

    /* set the default parameter values */
    client->interval = DEFAULT_UPDATE_INTERVAL;
    client->sr_name  = DEFAULT_SR_NAME;
    strncpy( ping_target_file, DEFAULT_PING_TARGET_FILE, MAX_SR_NAME_LEN );

    /* read values from the command-line */
    int i;
    for( i=1; i<argc; i+=advance ) {
        /* server ip must be the last argument */
        if( i == argc - 2 ) {
            /* get the group name */
            client->group = argv[i];
            if( strlen( client->group ) > MAX_GROUP_NAME_LEN )
                die( "Error: group name '%s' too long (max length is %u)",
                     client->group, MAX_GROUP_NAME_LEN );

            /* server ip must be the last argument */
            i += 1;

            /* create the desintation for our UDP packets */
            client->server_addr.sin_family = AF_INET;
            client->server_addr.sin_port = htons( SR_MONITOR_PORT );
            client->server_addr.sin_addr.s_addr = make_ip_addr( argv[i] );
            got_req_args = TRUE;
            break;
        }

        /* all command-line switcdhes are 2-characters long  */
        if( strlen(argv[i]) != 2 )
            die( "Invalid command-line option: %s", argv[i] );

        /* handle help, the only arg with a value */
        if( argv[i][1] == '?' )
            die( "\
usage: %s [-i <update_interval_sec>] [-n sr_name] [-p ping_file] [-q] [-s] <group name> <server_ip>\n\
  -i <update_interval_sec>: how often to send an update to the server (default=%usec)\n\
  -n <sr_name>: name of your software (default=%s)\n\
  -p <ping_file>: name of the ping targets file (default=%s)\n\
  -q: quiet (do not do ping checks)\n\
  -s: one ping thread per subnet and do round-robin over targets (instead of thead per target)\n\
  <group_name>: name to display for your group\n\
  <server_ip>: where to send updates",
                 argv[0], DEFAULT_UPDATE_INTERVAL, DEFAULT_PING_TARGET_FILE, DEFAULT_SR_NAME );

        /* handle switches without values */
        advance = 1;
        if( argv[i][1] == 'q' ) {
            do_pings = FALSE;
            continue;
        }
        else if( argv[i][1] == 's' ) {
            count_by_subnet = TRUE;
            continue;
        }

        /* make sure we have the value to assign to the switch's option */
        if( i+1 >= argc )
            die( "Error: expected value after %s", argv[i] );

        /* figure out which switch with a value we were given */
        advance = 2;
        switch( argv[i][1] ) {
        case 'i':
            client->interval = strtoul( argv[i+1], NULL, 10 );
            if( client->interval > MAX_UPDATE_INTERVAL )
                die( "Error: interval %s is too big or invalid (max is %u)",
                     argv[i+1], MAX_UPDATE_INTERVAL );
            break;

        case 'n':
            if( strlen(argv[i+1]) > MAX_SR_NAME_LEN )
                die( "Error: %s is too long (max length is %u)",
                     argv[i+1], MAX_SR_NAME_LEN );
            client->sr_name = argv[i+1];
            break;

        case 'p':
            if( strlen(argv[i+1]) > MAX_SR_NAME_LEN )
                die( "Error: %s is too long (max length is %u)",
                     argv[i+1], MAX_SR_NAME_LEN );
            strncpy( ping_target_file, argv[i+1], 32 );
            break;

        default:
            die( "Error: unknown command line argument %s", argv[i] );
        }
    }
    if( !got_req_args )
        die( "Error: missing required command-line argument(s) <group name> <server_ip>" );

    if( do_pings ) {
        parse_ping_target_file( count_by_subnet, client->ping, ping_target_file );

        /* start a thread for each ping test */
        pthread_t tid;
        unsigned i;
        for( i=0; i<MAX_PING_TARGETS; i++ ) {
            if( client->ping[i].num_targets == 0 )
                break;

            if( count_by_subnet ) {
                ping_handler_info_t* phi = malloc_or_die( sizeof(*phi) );
                phi->client = client;
                phi->ping_index = i;
                phi->target_index = -1;

                if( 0 != pthread_create( &tid, NULL, ping_handler_main, phi ) )
                    die( "Error: failed to create Ping Handler %u thread", i );
            }
            else {
                unsigned j;
                for( j=0; j<client->ping[i].num_targets; j++ ){
                    ping_handler_info_t* phi = malloc_or_die( sizeof(*phi) );
                    phi->client = client;
                    phi->ping_index = i;
                    phi->target_index = j;

                    if( 0 != pthread_create( &tid, NULL, ping_handler_main, phi ) )
                        die( "Error: failed to create Ping Handler %u-%u thread", i, j );
                }
            }
        }
    }
}

#ifdef USE_ALL_DATA
/**
 * Sorts an array of n nf2_arp_t objects by decreasing IP.
 */
static void sort_arp_entries( nf2_arp_t* A, int n ) {
    nf2_arp_t tmp;
    int i, j;

    /* use insertion sort to sort the entries */
    for( i=1; i<n-1; i++ ) {
        tmp = A[i];
        j = i - 1;
        while( j>=0 && A[j].ip > tmp.ip ) {
            A[j + 1] = A[j];
            j -= 1;
        }
        A[j+1] = tmp;
    }
}

static void get_arp_table( nf2_device_t* nf2, nf2_info_t* info ) {
    unsigned i;
    uint32_t val;
    uint16_t val16;

    /* read the arp translations from hardware */
    for( i=0; i<ROUTER_ARP_SIZE; i++ ) {
        /* tell the hardware we want to read row i */
        writeReg( nf2, ROUTER_OP_LUT_ARP_LUT_RD_ADDR_REG, i );

        /* get the contents of row i */
        readReg( nf2, ROUTER_OP_LUT_ARP_MAC_HI_REG, &val ); val16 = val;
        readReg( nf2, ROUTER_OP_LUT_ARP_MAC_LO_REG, &val );
        readReg( nf2, ROUTER_OP_LUT_ARP_NEXT_HOP_IP_REG, &info->arp[i].ip );
        mac_set_hi( &info->arp[i].mac, val16 );
        mac_set_lo( &info->arp[i].mac, val );

        /* convert hw host values to network */
        info->arp[i].ip = htonl( info->arp[i].ip );
    }

    /* sort the entries */
    sort_arp_entries( info->arp, ROUTER_ARP_SIZE );
}

/**
 * Sorts an array of n nf2_filter_t objects by decreasing IP.
 */
static void sort_ip_entries( addr_ip_t* A, int n ) {
    addr_ip_t tmp;
    int i, j;

    /* use insertion sort to sort the entries */
    for( i=1; i<n-1; i++ ) {
        tmp = A[i];
        j = i - 1;
        while( j>=0 && A[j] > tmp ) {
            A[j + 1] = A[j];
            j -= 1;
        }
        A[j+1] = tmp;
    }
}

static void get_filter_table( nf2_device_t* nf2, nf2_info_t* info ) {
    unsigned i;

    /* read the filter translations from hardware */
    for( i=0; i<FILTER_DEPTH; i++ ) {
        /* tell the hardware we want to read row i */
        writeReg( nf2, ROUTER_OP_LUT_DST_IP_FILTER_RD_ADDR_REG, i );

        /* get the contents of row i */
        readReg( nf2, ROUTER_OP_LUT_DST_IP_FILTER_IP_REG, &info->filter[i] );

        /* convert hw host values to network */
        info->filter[i] = htonl( info->filter[i] );
    }

    /* sort the entries */
    sort_ip_entries( info->filter, FILTER_DEPTH );
}
#endif /* USE_ALL_DATA */

static void get_route_table( nf2_device_t* nf2,
                             nf2_info_t* info, nf2_rt_t* rt_prev ) {
    unsigned i, j;
    uint32_t intf_hw_id;

    /* read the routes from hardware */
    for( i=0; i<ROUTER_RT_SIZE; i++ ) {
        /* tell the hardware we want to read row i */
        writeReg( nf2, ROUTER_OP_LUT_RT_LUT_RD_ADDR_REG, i );

        /* get the contents of row i */
        readReg( nf2, ROUTER_OP_LUT_RT_IP_REG, &info->rt[i].subnet );
        readReg( nf2, ROUTER_OP_LUT_RT_MASK_REG, &info->rt[i].mask );
        readReg( nf2, ROUTER_OP_LUT_RT_NEXT_HOP_IP_REG, &info->rt[i].next_hop );
        readReg( nf2, ROUTER_OP_LUT_RT_OUTPUT_PORT_REG, &intf_hw_id );

        /* swap host ordered hw values to network values */
        info->rt[i].subnet     = htonl( info->rt[i].subnet   );
        info->rt[i].mask       = htonl( info->rt[i].mask     );
        info->rt[i].next_hop   = htonl( info->rt[i].next_hop );
        info->rt[i].intf_hw_id = intf_hw_id; /* just copy; lhs only one byte */
    }

    /* update the timestamps for new routes */
    unsigned start_j = 0;
    unsigned now = htonl( time( NULL ) );
    for( i=0; i<ROUTER_RT_SIZE; i++ ) {
        nf2_rt_t* cur = &info->rt[i];

        for( j=start_j; j<ROUTER_RT_SIZE; j++ ) {
            nf2_rt_t* prev = &rt_prev[j];

            /* sorted, so if mask of prev is smaller, we didn't have it before */
            if( prev->mask < cur->mask ) {
                cur->timestamp = now;
                break;
            }

            /* if we find a match, then timestamp stays the same */
            if( cur->subnet==prev->subnet &&
                cur->next_hop==prev->next_hop &&
                cur->mask==prev->mask &&
                cur->intf_hw_id==prev->intf_hw_id ) {
                start_j = j; /* earlier entries won't match later masks in cur */
                break;
            }
        }
    }

    /* copy current routes to rt_prev */
    memcpy( rt_prev, info->rt, sizeof(nf2_rt_t)*ROUTER_RT_SIZE );
}

#ifdef USE_ALL_DATA
static void get_one_intf_info( nf2_device_t* nf2, nf2_intf_t* intf, unsigned index ) {
    uint16_t val16;
    unsigned i;
    byte b[9];
    uint32_t n[12];
    uint32_t base;
    uint32_t val;
    uint32_t addr[12] = {
        RX_QUEUE_0_NUM_PKTS_STORED_REG,
        RX_QUEUE_0_NUM_PKTS_DROPPED_FULL_REG,
        RX_QUEUE_0_NUM_PKTS_DROPPED_BAD_REG,
        RX_QUEUE_0_NUM_WORDS_PUSHED_REG,
        RX_QUEUE_0_NUM_BYTES_PUSHED_REG,
        RX_QUEUE_0_NUM_PKTS_DEQUEUED_REG,
        RX_QUEUE_0_NUM_PKTS_IN_QUEUE_REG,
        TX_QUEUE_0_NUM_PKTS_IN_QUEUE_REG,
        TX_QUEUE_0_NUM_PKTS_SENT_REG,
        TX_QUEUE_0_NUM_WORDS_PUSHED_REG,
        TX_QUEUE_0_NUM_BYTES_PUSHED_REG,
        TX_QUEUE_0_NUM_PKTS_ENQUEUED_REG
    };

    /* determine the base address for this interface */
    switch( index ) {
    case 0:
        base = MAC_GRP_0_CONTROL_REG;
        readReg( nf2, ROUTER_OP_LUT_MAC_0_HI_REG, &val ); val16 = val;
        readReg( nf2, ROUTER_OP_LUT_MAC_0_LO_REG, &val );
        break;

    case 1:
        base = MAC_GRP_1_CONTROL_REG;
        readReg( nf2, ROUTER_OP_LUT_MAC_1_HI_REG, &val ); val16 = val;
        readReg( nf2, ROUTER_OP_LUT_MAC_1_LO_REG, &val );
        break;

    case 2:
        base = MAC_GRP_2_CONTROL_REG;
        readReg( nf2, ROUTER_OP_LUT_MAC_2_HI_REG, &val ); val16 = val;
        readReg( nf2, ROUTER_OP_LUT_MAC_2_LO_REG, &val );
        break;

    case 3:
        base = MAC_GRP_3_CONTROL_REG;
        readReg( nf2, ROUTER_OP_LUT_MAC_3_HI_REG, &val ); val16 = val;
        readReg( nf2, ROUTER_OP_LUT_MAC_3_LO_REG, &val );
        break;

    default: die( "bad case in get_one_intf_info: %u", index );
    }
    mac_set_hi( &intf->mac, val16 );
    mac_set_lo( &intf->mac, val );

    /* get the control bitfield */
    readReg( nf2, base, &val );
    for( i=0; i<9; i++ )
        b[i] = (1 << i) & val; /* 0 => enabled */

    /* get the other values (assumes all interfaces have same reg offsets) */
    for( i=0; i<12; i++ ) {
        readReg( nf2, base + (addr[i] - MAC_GRP_0_CONTROL_REG), &val );
        n[i] = val;
    }

    /* set all the values we read */
    intf->flags.tx_queue  = b[0];
    intf->flags.rx_queue  = b[1];
    intf->flags.reset_mac = b[2];
    intf->flags.tx_mac    = b[3];
    intf->flags.rx_mac    = b[4];
    intf->flags.tx_jumbo  = b[5];
    intf->flags.rx_jumbo  = b[6];
    intf->flags.crc_check = b[7];
    intf->flags.crc_gen   = b[8];

    intf->rx_stored           = htonl( n[0] );
    intf->rx_dropped          = htonl( n[1] );
    intf->rx_dropped_bad      = htonl( n[2] );
    intf->rx_words_pushed     = htonl( n[3] );
    intf->rx_bytes_pushed     = htonl( n[4] );
    intf->rx_packets_dequeued = htonl( n[5] );
    intf->rx_packets_in_queue = htonl( n[6] );
    intf->tx_packets_in_queue = htonl( n[7] );
    intf->tx_packets_sent     = htonl( n[8] );
    intf->tx_words_pushed     = htonl( n[9] );
    intf->tx_bytes_pushed     = htonl( n[10] );
    intf->tx_packets_enqueued = htonl( n[11] );
}

static void get_intf_info( nf2_device_t* nf2, nf2_info_t* info ) {
    unsigned i;

    for( i=0; i<MAX_INTERFACES; i++ )
        get_one_intf_info( nf2, &info->intf[i], i );
}
#endif

/** Periodically sends an update with the NetFPGA's info to the server. */
static void inform_server_loop( client_t* client ) {
#define CMD_BUF_LEN (MAX_SR_NAME_LEN + 19)
    nf2_info_t info;
    char cmd[CMD_BUF_LEN];

    /* make the command we'll use to check up on the sr program */
    if( !my_snprintf( cmd, CMD_BUF_LEN, "ps -C %s > /dev/null", client->sr_name ) )
        die( "Error: could not create command (sr program name too long)" );

    /* put the group's name in the info packet (zero extra bytes) */
    memset( info.group, 0, MAX_GROUP_NAME_LEN + 1 );
    strncpy( info.group, client->group, MAX_GROUP_NAME_LEN + 1 );

    /* zero the routing table structures */
    memset( info.rt, 0, sizeof(nf2_rt_t)*ROUTER_RT_SIZE );

    while( 1 ) {
        /* is sr running? */
        info.sr_is_alive = (0 == system(cmd));

        /* get the current time */
        info.timestamp = htonl( time(NULL) );

#ifdef USE_ALL_DATA
        /* get the tables' values */
        get_arp_table( &client->nf2, &info );
        get_filter_table( &client->nf2, &info );
#endif
        get_route_table( &client->nf2, &info, client->rt_prev );

        /* write lock to lock out all the ping handler threads */
        pthread_rwlock_wrlock( &client->ping_lock );

        unsigned i, j, count = 0;
        for( i=0; i<MAX_PING_TARGETS; i++ ) {
            for( j=0; j<client->ping[i].num_targets; j++ ) {
                info.ping[count++] = client->ping[i].target[j].stats;
                if( count == MAX_PING_TARGETS )
                    goto no_more_room;
            }
        }
    no_more_room:
        pthread_rwlock_unlock( &client->ping_lock );

#ifdef USE_ALL_DATA
        /* get info about the interfaces */
        get_intf_info( &client->nf2, &info );
#endif

        /* send the update to the server */
        if( -1 == sendto( client->fd, (char*)&info, sizeof(info), 0,
                          (struct sockaddr*)&client->server_addr,
                          sizeof(client->server_addr) ) ) {
            if( errno != EINTR ) {
                perror( "Error: sendto failed" );
                exit( 1 );
            }
        }
        else
            debug_println( "Sent a packet to the server" );

        /* wait a while before compiling the next packet */
        sleep( client->interval );
    }
}

/**
 * Starts the client and then periodically sends updates to a server specified
 * by the command-line arguments.
 */
int main( int argc, char** argv ) {
    client_t client;

    debug_pthread_init_init();
    debug_pthread_init( "Client", "Client" );

    /* clear the client's routing table */
    memset( client.rt_prev, 0, sizeof(nf2_rt_t)*ROUTER_RT_SIZE );

    /* clear the ping data */
    memset( client.ping, 0, sizeof(ping_info_t) * MAX_PING_TARGETS );

    /* initialize the client data structure with the cli args */
    pthread_rwlock_init( &client.ping_lock, NULL );
    parse_cli_args( argc, argv, &client );

    /* connect to the hardware and UDP socket */
    hw_init( &client.nf2 );
    client.fd = monitor_init();

    /* tell the server what is going on from time to time */
    inform_server_loop( &client );

    /* cleanup */
    closeDescriptor( &client.nf2 );
    close( client.fd );
    return 0;
}
