/* Filename: debug.c */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "debug.h"

/** file-scope lock to provide mutual exclusion to print statements */
static pthread_mutex_t print_mutex = PTHREAD_MUTEX_INITIALIZER;

#ifdef _DEBUG_
/** returns the user-defined thread name */
char* debug_tname();
#endif

/** prints a message about what thread this is */
void debug_print_tname() {
#ifdef _DEBUG_
    char str_time[STRLEN_TIME];

    fprintf( stderr, "%s %15s: ",
             time_to_string_short(str_time, time(0)), debug_tname() );
#endif
}

void debug_print( const char* format, ... ) {
#ifdef _DEBUG_
    va_list args;
    va_start( args, format );

    pthread_mutex_lock( &print_mutex );
    debug_print_tname();
    vfprintf( stderr, format, args );
    pthread_mutex_unlock( &print_mutex );

    va_end( args );
#endif
}

void debug_print_more( const char* format, ... ) {
#ifdef _DEBUG_
    va_list args;
    va_start( args, format );

    pthread_mutex_lock( &print_mutex );
    vfprintf( stderr, format, args );
    pthread_mutex_unlock( &print_mutex );

    va_end( args );
#endif
}

void debug_println( const char* format, ... ) {
#ifdef _DEBUG_
    va_list args;
    va_start( args, format );

    pthread_mutex_lock( &print_mutex );
    debug_print_tname();
    vfprintf( stderr, format, args );
    fprintf( stderr, "\n" );
    pthread_mutex_unlock( &print_mutex );

    va_end( args );
#endif
}

#ifdef _DEBUG_
static pthread_key_t key_name;
#endif

void debug_pthread_init_init() {
#ifdef _DEBUG_
    pthread_key_create( &key_name, free );
#endif
}

void debug_pthread_init( const char* shortName, const char* longName ) {
#ifdef _DEBUG_
    char* name;

    name = malloc_or_die( (strlen(shortName)+1) * sizeof(char) );
    strcpy( name, shortName );

    pthread_setspecific( key_name, name );
    debug_println( "%s is now alive", longName );
#endif
}

#ifdef _DEBUG_
char* debug_tname() {
    return (char*)pthread_getspecific( key_name );
}
#endif

void die( const char* format, ... ) {
    va_list args;
    va_start( args, format );

    pthread_mutex_lock( &print_mutex );
    debug_print_tname();
    vfprintf( stderr, format, args );
    fprintf( stderr, "\n" );
    pthread_mutex_unlock( &print_mutex );

    va_end( args );
    exit( 1 );
}

#ifndef _DEBUG_
void debug_true_or_die( int bool, const char* format, ... ) { }
#endif

void true_or_die( int bool, const char* format, ... ) {
    if( !bool ) {
        va_list args;
        va_start( args, format );

        pthread_mutex_lock( &print_mutex );
        vfprintf( stderr, format, args );
        fprintf( stderr, "\n" );
        pthread_mutex_unlock( &print_mutex );

        va_end( args );
        assert( 0 );
        exit( 1 );
    }
}
