/* Filename: server.c */

#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>    /* strcmp */
#include <time.h>      /* time */
#include <unistd.h>    /* sleep */
#include "debug.h"
#include "server.h"
#include "server_pres.h"

/** leave a client in the list for up to 60 seconds of no updates */
#define CLIENT_EXPIRE_TIME 60

#define MAX_UINT -1
#define DEFAULT_HOST_FILE "hosts"

/**
 * Expires clients which have not sent an update for some period of time.
 */
void* scrubber_main( void* pserver ) {
    server_t* server;
    unsigned earliest;
    unsigned now;

    debug_pthread_init( "Scrubber", "Scrubber" );
    pthread_detach( pthread_self() );
    server = (server_t*)pserver;

    /* look for expired clients */
    while( 1 ) {
        while( !server->client_list_head ) { /* ok b/c only I remove clients */
            /* nothing to do: no clients */
            debug_println( "No clients, performing empty refresh" );
            sleep( CLIENT_EXPIRE_TIME );
        }

        earliest = MAX_UINT;

        pthread_mutex_lock( &server->lock );

        /* check for expired clients */
        now = time( NULL );
        bool expired_any = FALSE;
        client_state_t* prev_c = NULL;
        client_state_t* c = server->client_list_head;
        while( c ) {
            /* client expired? */
            if( now - c->timestamp > CLIENT_EXPIRE_TIME ) {
                debug_println( "Scrubbing %s", c->info.group );

                /* update next pointer from previous node in the list */
                if( prev_c )
                    prev_c->next = c->next;
                else
                    server->client_list_head = c->next;

                /* delete the old client state */
                client_state_t* c_to_free = c;
                c = c->next;
                myfree( c_to_free );

                expired_any = TRUE;
                server->num_clients -= 1;
            }
            else {
                /* track oldest timestamp from non-expired client */
                if( c->timestamp < earliest )
                    earliest = c->timestamp;

                prev_c = c;
                c = c->next;
            }
        }

        /* re-create the presentation if any client was dropped */
        if( expired_any )
            if( server->refresh_on_change )
                pthread_cond_signal( &server->cond );

        pthread_mutex_unlock( &server->lock );

        /* sleep until the next client could possibly expire */
        if( earliest != MAX_UINT )
            sleep( CLIENT_EXPIRE_TIME - (now - earliest) );
    }

    return NULL;
}

void server_recv_loop( server_t* server ) {
    nf2_info_t tmp;
    struct sockaddr_in client_addr;
    socklen_t client_addr_len;
    unsigned now, len;

    /* very weird, but OS won't seem to give us any packets until I send one for
       some reason */
    struct sockaddr_in server_addr;
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons( SR_MONITOR_PORT );
    server_addr.sin_addr.s_addr = make_ip_addr( "172.24.89.136" );
    if( -1 == sendto( server->fd, (char*)&tmp, sizeof(tmp), 0,
                      (struct sockaddr*)&server_addr,
                      sizeof(server_addr) ) ) {
        debug_println( "send wasn't happy" );
    }

    while( 1 ) {
        client_addr_len = sizeof(client_addr);
        if( -1 == (len=recvfrom( server->fd, (char*)&tmp, sizeof(tmp), 0,
                                 (struct sockaddr*)&client_addr,
                                 &client_addr_len )) ) {
            if( errno == EINTR )
                continue; /* try again */

            perror( "Error: recvfrom failed" );
            exit( 1 );
        }
        now = time( NULL );

        /* make sure we got the packet we were expecting */
        if( len != sizeof(nf2_info_t) ) {
            debug_println( "Dropping packet of length %uB (expected %uB)",
                           len, sizeof(nf2_info_t) );
            continue;
        }

        /* make sure the group name is nul-terminated (assume all 0s after name) */
        if( tmp.group[MAX_GROUP_NAME_LEN] ) {
            tmp.group[MAX_GROUP_NAME_LEN] = '\0';
            debug_println( "Dropping packet with non-null terminated group name: %s",
                           tmp.group );
            continue;
        }

        debug_println( "Received packet from %s", tmp.group );

        /* handle the received packet */
        pthread_mutex_lock( &server->lock );

        /* search for the client's existing state */
        client_state_t* c = server->client_list_head;
        while( c ) {
            if( strcmp(tmp.group,c->info.group)==0 )
                break;
            else
                c = c->next;
        }

        if( c ) {
            /* make sure the update is newer than the last one from the group */
            if( c->info.timestamp > ntohl(tmp.timestamp) ) {
                debug_println( "Received older update from %s (ignoring it)",
                               tmp.group );
                pthread_mutex_unlock( &server->lock );
                continue;
            }
        }
        else {
            /* we didn't find the client so add a new one */
            c = malloc_or_die( sizeof(client_state_t) );
            server->num_clients += 1;

            /* insert into the list in lexicographic order */
            client_state_t* prev = NULL;
            client_state_t* citr = server->client_list_head;
            while( citr ) {
                /* look for first elt this is smaller than */
                if( strcmp(tmp.group, citr->info.group) < 0 ) {
                    /* insert c into list before citr */
                    c->next = citr;
                    if( prev )
                        prev->next = c;
                    else
                        server->client_list_head = c;

                    /* done, we've inserted it into the list */
                    break;
                }

                prev = citr;
                citr = citr->next;
            }

            /* put at end of list if it didn't come before any other */
            if( !citr ) {
                if( prev )
                    prev->next = c;
                else
                    server->client_list_head = c;

                c->next = NULL;
            }
        }

        /* convert numbers (not IPs though) to host-order */
        unsigned i;
#ifdef USE_ALL_DATA
        for( i=0; i<MAX_INTERFACES; i++ ) {
            tmp.intf[i].rx_stored           = ntohl( tmp.intf[i].rx_stored );
            tmp.intf[i].rx_dropped          = ntohl( tmp.intf[i].rx_dropped );
            tmp.intf[i].rx_dropped_bad      = ntohl( tmp.intf[i].rx_dropped_bad );
            tmp.intf[i].rx_words_pushed     = ntohl( tmp.intf[i].rx_words_pushed );
            tmp.intf[i].rx_bytes_pushed     = ntohl( tmp.intf[i].rx_bytes_pushed );
            tmp.intf[i].rx_packets_dequeued = ntohl( tmp.intf[i].rx_packets_dequeued );
            tmp.intf[i].rx_packets_in_queue = ntohl( tmp.intf[i].rx_packets_in_queue );
            tmp.intf[i].tx_packets_in_queue = ntohl( tmp.intf[i].tx_packets_in_queue );
            tmp.intf[i].tx_packets_sent     = ntohl( tmp.intf[i].tx_packets_sent );
            tmp.intf[i].tx_words_pushed     = ntohl( tmp.intf[i].tx_words_pushed );
            tmp.intf[i].tx_bytes_pushed     = ntohl( tmp.intf[i].tx_bytes_pushed );
            tmp.intf[i].tx_packets_enqueued = ntohl( tmp.intf[i].tx_packets_enqueued );
        }
#endif
        tmp.timestamp = ntohl( tmp.timestamp );
        for( i=0; i<ROUTER_RT_SIZE; i++ ) {
            nf2_rt_t* rt = &tmp.rt[i];

            /* convert route timestamps to time elapsed */
            rt->timestamp = ntohl( rt->timestamp );
            if( rt->timestamp > tmp.timestamp || rt->timestamp == 0 )
                rt->timestamp = 0;
            else
                rt->timestamp = tmp.timestamp - rt->timestamp;
        }

        for( i=0; i<MAX_PING_TARGETS; i++ ) {
            nf2_ping_t* ping = &tmp.ping[i];
            ping->streak = ntohs( ping->streak );
        }

        /* update the client's data and timestamp */
        c->info = tmp;
        c->timestamp = now;

        /* update the count of updates received */
        server->info.update_count += 1;

        /* tell the refresh thread to wake up and refresh the presentation */
        if( server->refresh_on_change )
            pthread_cond_signal( &server->cond );

        pthread_mutex_unlock( &server->lock );
    }
}

/** periodically refresh the presentation */
void* refresh_thread_main( void* pserver ) {
    server_t* server = (server_t*)pserver;
    struct timespec wait;

    debug_pthread_init( "Refresher", "Refresher" );
    pthread_detach( pthread_self() );

    wait.tv_nsec = 0;

    pthread_mutex_lock( &server->lock );
    while( 1 ) {
        debug_println( "refreshing!" );
        refresh_presentation( server );

        /* wait until someone tells us to refresh or a timer expires */
        if( wait.tv_sec ) {
            wait.tv_sec = server->refresh_interval + time( NULL );
            pthread_cond_timedwait( &server->cond, &server->lock, &wait );
        }
        else
            pthread_cond_wait( &server->cond, &server->lock );
    }

    return NULL;
}

/**
 * Parses the host file to determine what owner names to assign to each IP.
 */
void parse_host_file( server_t* server, const char* filename ) {
    FILE* fp;
    const char* err;
    char  line[512];
    char  str_ip[STRLEN_IP];
    char  str_name[MAX_NAME_LEN];
    struct in_addr ip;

    err = "Error loading host table,";
    debug_println( "Loading target ping table from %s", filename );

    fp = fopen( filename, "r" );
    if( !fp )
        die( "Could  not access the host file named %s", filename );

    server->num_hosts = 0;
    while( fgets(line,512,fp) != 0) {
        if( server->num_hosts == MAX_PING_TARGETS )
            die( "Error: too many hosts (max = %u)", MAX_PING_TARGETS );

        sscanf( line, "%s %s", str_ip, str_name );

        if( inet_aton(str_ip,&ip) == 0 )
            die( "%s cannot convert %s to valid IP", err, str_ip );

        server->host[server->num_hosts].ip = ip.s_addr;
        strncpy( server->host[server->num_hosts].name, str_name, MAX_NAME_LEN );
        server->num_hosts += 1;
    }
}

/** Parse the command-line arguments for the server. */
static void parse_cli_args( int argc, char** argv, server_t* server ) {
    bool got_req_args = FALSE;
    unsigned advance;

    /* set the default parameter values */
    server->refresh_interval  = DEFAULT_REFRESH_INTERVAL;
    server->refresh_on_change = DEFAULT_REFRESH_ON_CHANGE;

    /* read values from the command-line */
    int i;
    for( i=1; i<argc; i+=advance ) {
        /* handle help, the only arg with a value */
        if( argv[i][1] == '?' )
            die( "\
usage: %s [-c] [-r <refresh_interval_sec>] <presentation output path>\n\
  -c: refresh presentation on every update (default=off)\n\
  -r <refresh_interval_sec>: minimum refresh interval (default=%usec) (0=only on change)",
                 argv[0], DEFAULT_REFRESH_INTERVAL );

        /* server ip must be the last argument */
        if( i == argc - 1 ) {
            if( argv[i][0] == '-' )
                break; /* go to the missing required argument print stmt */

            /* get the output folder */
            server->info.output_path = argv[i];
            if( 0 == my_snprintf( server->info.mv_cmd, 512, "mv %s %s/index.htm",
                                  TMP_FILE, server->info.output_path ) )
                die( "Error: unable to create mv command (path too long?): %s", server->info.mv_cmd );

            debug_println( "present command: %s", server->info.mv_cmd );
            got_req_args = TRUE;
            break;
        }

        /* handle switches without values */
        advance = 1;
        if( argv[i][1] == 'c' ) {
            server->refresh_on_change = TRUE;
            continue;
        }

        /* all command-line switcdhes are 2-characters long  */
        if( strlen(argv[i]) != 2 )
            die( "Invalid command-line option: %s", argv[i] );

        /* make sure we have the value to assign to the switch's option */
        if( i+1 >= argc )
            die( "Error: expected value after %s", argv[i] );

        /* figure out which switch we were given */
        advance = 2;
        switch( argv[i][1] ) {
        case 'r':
            server->refresh_interval = strtoul( argv[i+1], NULL, 10 );
            if( server->refresh_interval > 60 )
                die( "Refresh interval too long" );
            break;

        default:
            die( "Error: unknown command line argument %s", argv[i] );
        }
    }
    if( !got_req_args )
        die( "Error: missing required command-line argument(s) <presentation output path>" );

    if( !server->refresh_on_change && !server->refresh_interval )
        die( "Error: refresh interval is 0 (e.g. off) and refresh on change is also off" );

    parse_host_file( server, DEFAULT_HOST_FILE );
}

int main( int argc, char** argv ) {
    server_t server;
    pthread_t tid;

    debug_pthread_init_init();
    debug_pthread_init( "Server", "Server" );

    /* zero the server info */
    memset( &server, 0, sizeof(server_t) );

    /* intialize the basic server fields */
    pthread_cond_init(  &server.cond, NULL );
    pthread_mutex_init( &server.lock, NULL );
    server.info.start_time = time( NULL );

    /* initialize server fields based on cli args */
    parse_cli_args( argc, argv, &server );

    /* initalize the other pres info fields */

    /* spawn a new thread to handle client expiration */
    if( 0 != pthread_create( &tid, NULL, scrubber_main, &server ) )
        die( "Error: failed to create Scrubber thread" );

    /* connect to the UDP socket */
    server.fd = monitor_init();

    /* start the refresh thread */
    if( 0 != pthread_create( &tid, NULL, refresh_thread_main, &server ) )
        die( "Error: failed to create refresh thread" );

    /* process incoming UDP packets */
    server_recv_loop( &server );

    /* cleanup */
    close( server.fd );
    return 0;
}
