/* Filename: server_pres.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>       /* memset */
#include "debug.h"
#include "reg_defines.h"
#include "server.h"

#define GOOD_THRESH 90
#define BAD_THRESH 50
#define GOOD_STYLE "good"
#define SOSO_STYLE "soso"
#define BAD_STYLE "bad"

/** status of targets */
typedef struct {
    bool live[MAX_CLIENTS];
    nf2_ping_t status[MAX_CLIENTS];
} subnet_targets_t;

/** encapsulates info about how each group routes to a given subnet and mask */
typedef struct {
    addr_ip_t subnet;
    addr_ip_t mask;
    byte intf[MAX_CLIENTS];  /* intf each group sends from to get to this s/m */
    unsigned time_up[MAX_CLIENTS];
    unsigned num_targets;
    addr_ip_t target_ip[MAX_PING_TARGETS_PER_SUBNET];
    subnet_targets_t target[MAX_PING_TARGETS_PER_SUBNET];
} subnet_mask_t;

void print_sm( server_t* server, subnet_mask_t* sm ) {
    char str_subnet[STRLEN_SUBNET];
    subnet_to_string( str_subnet, sm->subnet, sm->mask );
    debug_print( "%s => ", str_subnet );
    int i;
    for( i=0; i<server->num_clients; i++ )
        debug_print_more( "%0x ", sm->intf[i] );
    debug_print_more( "\n" );
}

/**
 * Sorts an array of n subnet_mask_t objects by decreasing mask.
 */
static void sort_sm_entries( subnet_mask_t* A, int n ) {
    subnet_mask_t tmp;
    int i, j;

    /* use insertion sort to sort the entries */
    for( i=1; i<n-1; i++ ) {
        tmp = A[i];
        j = i - 1;
        while( j>=0 && (ntohl(A[j].mask) > ntohl(tmp.mask) ||
                        (A[j].mask==tmp.mask && ntohl(A[j].subnet & A[j].mask)>ntohl(tmp.subnet & tmp.mask))) ) {
            A[j + 1] = A[j];
            j -= 1;
        }
        A[j+1] = tmp;
    }
}

void refresh_presentation( server_t* server ) {
    subnet_mask_t dest[ROUTER_RT_SIZE];
    unsigned num_dest = 0;
    unsigned i, j, c_on;
    nf2_rt_t* rt;

    /* determine a list of all subnets and how each group is routing to it */
    memset( dest, 0, sizeof(subnet_mask_t) * ROUTER_RT_SIZE ); /* zero it */
    c_on = 0;
    client_state_t* c = server->client_list_head;
    while( c ) {
        /* loop over each route the client has in their rtable */
        for( i=0; i<ROUTER_RT_SIZE; i++ ) {
            rt = &c->info.rt[i];

            /* ignore 224.0.0.5 */
            if( rt->subnet == htonl(0xE0000005) )
                continue;

            /* stop if we get to an empty entry */
            if( rt->subnet == 0 )
                break;

            /* scan the list of subnets for this one */
            for( j=0; j<num_dest; j++ ) {
                if( (dest[j].subnet & dest[j].mask)==(rt->subnet & rt->mask) && dest[j].mask==rt->mask ) {
                    dest[j].intf[c_on] = rt->intf_hw_id;
                    dest[j].time_up[c_on] = rt->timestamp;
                    break;
                }
            }

            /* if it wasn't in our list, then add it */
            if( j == num_dest ) {
                dest[j].subnet = rt->subnet;
                dest[j].mask   = rt->mask;
                dest[j].intf[c_on] = rt->intf_hw_id;
                dest[j].time_up[c_on] = rt->timestamp;
                dest[j].num_targets = 0;
                num_dest += 1;
                if( num_dest == ROUTER_RT_SIZE )
                    goto no_more_room;
            }

            /* stop after we hit an item with mask = 0.0.0.0 even if ip != 0 */
            if( rt->mask == 0 )
                break;
        }

        c = c->next;
        c_on += 1;
        if( c_on == MAX_CLIENTS ) {
            debug_println( "Warning: not enough space for any more clients! (%u active, %u max)",
                           server->num_clients, MAX_CLIENTS );
            break;
        }
    }
 no_more_room:

    /* sort the subnet and mask pairs */
    sort_sm_entries( dest, num_dest );

    /* add ping target data */
    c_on = 0;
    c = server->client_list_head;
    while( c ) {
        for( i=0; i<MAX_PING_TARGETS; i++ ) {
            nf2_ping_t* ping = &c->info.ping[i];

            /* stop when we get to an empty ping target */
            if( ping->ip == 0 )
                break;

            /* search for the right subnet to put this ping on */
            for( j=0; j<num_dest; j++ ) {
                if( (dest[j].subnet & dest[j].mask) == (ping->ip & dest[j].mask) ) {
                    /* search for the right target to update */
                    unsigned k;
                    for( k=0; k<dest[j].num_targets; k++ ) {
                        if( dest[j].target_ip[k] == ping->ip ) {
                            /* set the status data for this target on this client */
                            dest[j].target[k].status[c_on] = *ping;
                            dest[j].target[k].live[c_on] = TRUE;
                            break;
                        }
                    }

                    /* if we didn't find it, add the target ip */
                    if( k == dest[j].num_targets ) {
                        if( k == MAX_PING_TARGETS_PER_SUBNET ) {
                            debug_println( "Warning: not enough room for all targets" );
                            break;
                        }

                        unsigned x;
                        for( x=0; x<server->num_clients; x++ )
                            dest[j].target[k].live[x] = FALSE;

                        dest[j].num_targets += 1;
                        dest[j].target_ip[k] = ping->ip;
                        dest[j].target[k].status[c_on] = *ping;
                        dest[j].target[k].live[c_on] = TRUE;
                    }

                    /* found the right subnet */
                    break;
                }
            }
        }

        c = c->next;
        c_on += 1;
        if( c_on == MAX_CLIENTS )
            break;
    }

    /* get the current time (hh:mm:ss only) */
    unsigned now = time( NULL );
    char str_now_full[STRLEN_TIME];
    char* str_now = time_to_string_short( str_now_full, now );

    /* compute total uptime */
    unsigned uptime_s = now - server->info.start_time;
    unsigned uptime_h = uptime_s / 3600; uptime_s -= (uptime_h * 3600);
    unsigned uptime_m = uptime_s /   60; uptime_s -= (uptime_m *   60);

    /* open the temporary file which we'll populate */
    FILE* fp = fopen( TMP_FILE, "w" );
    if( !fp ) {
        debug_println( "Warning: failed to open %s, skipping presentation generation",
                       TMP_FILE );
        return;
    }

#define IJ(str) ,fp); fputs(str,fp); fputs(
#define IJFU(unum) ,fp); fprintf(fp,"%u",unum); fputs(
#define IJFT(hh,mm,ss) ,fp); fprintf(fp,"%uhr, %umin, %usec",hh,mm,ss); fputs(
    /* generate the HTML for the main page */
/* HTML header */
    fputs( "\
<html><head>\n\
<meta http-equiv=\"refresh\" content=\"1\">\n\
<title>CS344 NetFPGA Monitoring Tool (" IJ(str_now) ")</title>\n\
<link href=\"sr_monitor.css\" rel=\"stylesheet\" type=\"text/css\">\n\
</head>\n\
<body>\n\
<table width=\"100%\" border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n\
  <tr bgcolor=\"#006600\" class=\"hdr\">\n\
    <td width=\"33%\">Last Update: " IJ(str_now) "</td>\n\
    <td width=\"33%\"># of Clients: " IJFU(server->num_clients) "</td>\n\
    <td width=\"33%\"><i>CS344 NetFPGA Monitor Tool v" IJ(SERVER_VERSION) "</i></td>\n\
  </tr>\n\
</table>\n\
\n\
<table width=\"100%\" border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n\
  <tr bgcolor=\"#666666\" class=\"subhdr\">\n", fp );

/* header row for main table */
unsigned width = 100 / (server->num_clients + 1);
fputs( "    <td>Subnet</td>\n", fp );
fputs( "    <td>Owner</td>\n", fp );
c = server->client_list_head;
while( c ) {
    /* use local timestamp */
    str_now = time_to_string_short( str_now_full, c->timestamp );
    const char* str_alive = (c->info.sr_is_alive ? "" : "<span class=\"bad\"> (SW DEAD!)</span>");
    fprintf( fp, "    <td width=\"%u%%\">Group %s<br/>%s%s</td>", width, c->info.group, str_now, str_alive );
    c = c->next;
}
fputs( "\n  </tr>\n", fp );

/* fill in the main table (row=subnet, subrow=host, col=group) */
char str_subnet[STRLEN_SUBNET];
char str_ip[STRLEN_IP];
const char* str_ip_ptr;
const char* cell;
unsigned t;
for( i=0; i<num_dest; i++ ) {
    subnet_mask_t* sm = &dest[i];

    /* print subnet in first column */
    subnet_to_string( str_subnet, sm->subnet, sm->mask );
    fprintf( fp, "\
  <tr>\n\
    <td bgcolor=\"#666666\" class=\"subhdr\">%s</td>", str_subnet );

    /* print owner information in second column */
    fputs( "    <td>Route (Uptime)", fp );
    unsigned count = 0;
    for( j=0; j<server->num_hosts; j++ ) {
        host_info_t* host = &server->host[j];

        for( t=0; t<sm->num_targets; t++ ) {
            /* does host match this target ip? */
            if( host->ip==sm->target_ip[t] ) {
                ip_to_string( str_ip, host->ip );

                /* determine how many host octets we need to show */
                unsigned num_octets;
                if(      (ntohl(sm->mask) & 0xFF000000) != 0xFF000000 )
                    num_octets = 4;
                else if( (ntohl(sm->mask) & 0x00FF0000) != 0x00FF0000 )
                    num_octets = 3;
                else if( (ntohl(sm->mask) & 0x0000FF00) != 0x0000FF00 )
                    num_octets = 2;
                else
                    num_octets = 1;

                /* find ptr to period before first octet we want to keep */
                str_ip_ptr = str_ip;
                while( num_octets < 4 && *str_ip_ptr!='\0' ) {
                    str_ip_ptr += 1;

                    if( *str_ip_ptr == '.' )
                        num_octets += 1;
                }

                count += 1;
                fputs( "<br/>\n        ", fp );

                /* print the owner info */
                fprintf(fp, "%s (%s)", str_ip_ptr, host->name );
            }
        }
        if( sm->num_targets == 0 )
            fputs( "---", fp );
    }
    if( count > 0 )
        fputs( "\n    </td>\n", fp );
    else
        fputs( "</td>\n", fp );

    /* each groups route / stats for this subnet */
    for( j=0; j<server->num_clients; j++ ) {
        fputs( "    <td>", fp );

        /* print route interface first */
        cell = "";
        switch( dest[i].intf[j] ) {
            case INTF_NF2_0: cell = "eth0"; break;
            case INTF_NF2_1: cell = "eth1"; break;
            case INTF_NF2_2: cell = "eth2"; break;
            case INTF_NF2_3: cell = "eth3"; break;
            case INTF_CPU_0: cell = "cpu0"; break;
            case INTF_CPU_1: cell = "cpu1"; break;
            case INTF_CPU_2: cell = "cpu2"; break;
            case INTF_CPU_3: cell = "cpu3"; break;
            case 0:  fprintf( fp, "<span class=\"bad\">NO ROUTE</span>" ); break;
            default: fprintf( fp, "??? %u", dest[i].intf[j] ); break;
        }

        /* compute uptime breakdown for this route */
        unsigned uptime_s = dest[i].time_up[j];
        unsigned uptime_m = uptime_s /   60; uptime_s -= (uptime_m *   60);
        fprintf( fp, "%s (%02u:%02u)<br/>\n        ", cell, uptime_m, uptime_s );

        /* add ping target results */
        count = 0;
        for( t=0; t<sm->num_targets; t++ ) {
            nf2_ping_t* st = &dest[i].target[t].status[j];

            if( count++ )
                fputs( "<br/>\n        ", fp );

            if( dest[i].target[t].live[j] ) {
                /* compute % success of last pings */
                unsigned percent_ok = (st->num_ok * 100) / MAX_PING_COUNT;
                const char* str_percent_style = (percent_ok>=GOOD_THRESH) ? GOOD_STYLE : ((percent_ok<BAD_THRESH) ? BAD_STYLE : SOSO_STYLE);
                const char* str_percent_pad = (percent_ok < 10) ? "&nbsp;&nbsp;" : ((percent_ok < 100) ? "&nbsp;" : "" );

                /* choose style for streak */
                const char* str_streak_style = (st->streak>0) ? GOOD_STYLE : BAD_STYLE;

                /* print target's info */
                fprintf( fp, "<span class=\"%s\">%s%u%%</span> / <span class=\"%s\">%d</span>",
                         str_percent_style, str_percent_pad, percent_ok,
                         str_streak_style, st->streak );
            }
            else {
                /* no ping stats for this route yet */
                fprintf( fp, "---" );
            }
        }
        if( count > 1 )
            fputs( "\n    </td>\n", fp );
        else
            fputs( "</td>\n", fp );
    }
    fputs( "  </tr>\n", fp );
}

/* HTML footer */
fputs( "\
</table>\n\
\n\
<table width=\"100%\" border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n\
  <tr bgcolor=\"#006600\" class=\"hdr\">\n\
    <td width=\"33%\">Total Updates: " IJFU(server->info.update_count) "</td>\n\
    <td width=\"33%\">&nbsp;</td>\n\
    <td width=\"33%\">Server Uptime: " IJFT(uptime_h,uptime_m,uptime_s) "</td>\n \
  </tr>\n\
</table>\n\
\n\
</body></html>", fp );

    /* flush to disk */
    fclose( fp );

    /* overwrite the real output file with the newly updated tmp file */
    if( system( server->info.mv_cmd ) != 0 )
        debug_println( "Warning: unable to execute move command to put new presentation in the output folder: %s",
                       server->info.mv_cmd );
}
