/* VLAN defines */
`define VLAN_CTRL_WORD 8'h42
`define VLAN_ETHERTYPE 16'h8100

/* Other Constants */
`define ETH_TYPE_IP 16'h0800
`define IP_PROTO_TCP 8'h6
`define IP_PROTO_UDP 8'h11
`define IP_PROTO_ICMP 8'h1

/* Each flow entry is specified by
 * input port     : 8  bits
 * MAC src port   : 48 bits
 * MAC dst port   : 48 bits
 * Ethernet type  : 16 bits
 * VLAN id        : 12 bits - 16 bits used
 * IP src address : 32 bits
 * IP dst address : 32 bits
 * IP protocol    : 8  bits
 * Transport src  : 16 bits
 * Transport dst  : 16 bits
 * 
 * TOTAL          : 240 bits
 */
`define OPENFLOW_ENTRY_TRANSP_DST_WIDTH       16
`define OPENFLOW_ENTRY_TRANSP_DST_POS         0
`define OPENFLOW_ENTRY_TRANSP_SRC_WIDTH       16
`define OPENFLOW_ENTRY_TRANSP_SRC_POS         `OPENFLOW_ENTRY_TRANSP_DST_POS+`OPENFLOW_ENTRY_TRANSP_DST_WIDTH
`define OPENFLOW_ENTRY_IP_PROTO_WIDTH         8
`define OPENFLOW_ENTRY_IP_PROTO_POS           `OPENFLOW_ENTRY_TRANSP_SRC_POS+`OPENFLOW_ENTRY_TRANSP_SRC_WIDTH
`define OPENFLOW_ENTRY_IP_DST_WIDTH           32
`define OPENFLOW_ENTRY_IP_DST_POS             `OPENFLOW_ENTRY_IP_PROTO_POS+`OPENFLOW_ENTRY_IP_PROTO_WIDTH
`define OPENFLOW_ENTRY_IP_SRC_WIDTH           32
`define OPENFLOW_ENTRY_IP_SRC_POS             `OPENFLOW_ENTRY_IP_DST_POS+`OPENFLOW_ENTRY_IP_DST_WIDTH
`define OPENFLOW_ENTRY_ETH_TYPE_WIDTH         16
`define OPENFLOW_ENTRY_ETH_TYPE_POS           `OPENFLOW_ENTRY_IP_SRC_POS+`OPENFLOW_ENTRY_IP_SRC_WIDTH
`define OPENFLOW_ENTRY_ETH_DST_WIDTH          48
`define OPENFLOW_ENTRY_ETH_DST_POS            `OPENFLOW_ENTRY_ETH_TYPE_POS+`OPENFLOW_ENTRY_ETH_TYPE_WIDTH
`define OPENFLOW_ENTRY_ETH_SRC_WIDTH          48
`define OPENFLOW_ENTRY_ETH_SRC_POS            `OPENFLOW_ENTRY_ETH_DST_POS+`OPENFLOW_ENTRY_ETH_DST_WIDTH
`define OPENFLOW_ENTRY_SRC_PORT_WIDTH         8
`define OPENFLOW_ENTRY_SRC_PORT_POS           `OPENFLOW_ENTRY_ETH_SRC_POS+`OPENFLOW_ENTRY_ETH_SRC_WIDTH
`define OPENFLOW_ENTRY_VLAN_ID_WIDTH          16
`define OPENFLOW_ENTRY_VLAN_ID_POS            `OPENFLOW_ENTRY_SRC_PORT_POS+`OPENFLOW_ENTRY_SRC_PORT_WIDTH
`define OPENFLOW_ENTRY_WIDTH                  10'd240

/* The action field is composed of a bitmask 
 * specifying actions to take and arguments.
 * 
 * Action bitmask is composed of
 * - 1x16-bit word specifying ports to forward on.
 * - 1x16-bit word indicating packet trimming
 * - 1x16-bit word indicating VLAN id modifications
 * - 6x24-bit words indicating which 16-bit words/bytes to replace
 * - 2x40-bit words indicating which 16-bit words to replace with bitmasks
 * - 2x24-bit words indicating which 16-bit words to increment
 * TOTAL: 320 bits = 5*64-bit words
 */
`define OPENFLOW_ACTION_WIDTH                     10'd320
`define OPENFLOW_FORWARD_BITMASK_WIDTH            16
`define OPENFLOW_FORWARD_BITMASK_POS              0
`define OPENFLOW_PKT_TRIM_ACTION_WIDTH            16
`define OPENFLOW_PKT_TRIM_ACTION_POS              `OPENFLOW_FORWARD_BITMASK_POS+`OPENFLOW_FORWARD_BITMASK_WIDTH
`define OPENFLOW_VLAN_MOD_ACTION_WIDTH            16
`define OPENFLOW_VLAN_MOD_ACTION_POS              `OPENFLOW_PKT_TRIM_ACTION_POS+`OPENFLOW_PKT_TRIM_ACTION_WIDTH
`define OPENFLOW_REPLACEMENT_ACTION_WIDTH         24
`define OPENFLOW_NUM_REPLACEMENT_ACTIONS          6
`define OPENFLOW_REPLACEMENT_ACTION_POS           `OPENFLOW_VLAN_MOD_ACTION_WIDTH+`OPENFLOW_VLAN_MOD_ACTION_POS
`define OPENFLOW_REPLACEMENT_BITMASK_ACTION_WIDTH 40
`define OPENFLOW_NUM_REPLACEMENT_BITMASK_ACTIONS  2
`define OPENFLOW_REPLACEMENT_BITMASK_ACTION_POS   `OPENFLOW_REPLACEMENT_ACTION_POS+`OPENFLOW_NUM_REPLACEMENT_ACTIONS*`OPENFLOW_REPLACEMENT_ACTION_WIDTH
`define OPENFLOW_INCREMENT_ACTION_WIDTH           24
`define OPENFLOW_NUM_INCREMENT_ACTIONS            2
`define OPENFLOW_INCREMENT_ACTION_POS             `OPENFLOW_REPLACEMENT_ACTION_POS+`OPENFLOW_NUM_REPLACEMENT_ACTIONS*`OPENFLOW_REPLACEMENT_ACTION_WIDTH

/* Each 32-bit register read takes about 2us. If we have 32000 flow entries,
 * it will take 64ms to read 1 register out of each entry. If we want CPU usage 
 * to remain below 10%, then the CPU can read a 32-bit register out of each
 * entry every 640ms.
 * 
 * In the worst case, a min-sized packet arrives every 16 cycles = 128ns.
 * If all packets arrive for the same entry, the pkt counter increments by
 * 640ms/128ns=5 million every 640ms. This needs 22bits.
 * 
 * We will allow CPU utilization to be up to 4%. This means the CPU will
 * read 32000 32-bit registers every 1.6s, or 32000 64-bit registers every 3.2s.
 * 
 * In 3.2s, we can get 25M packets, and 3.2s*4Gbps/8 approx 1.6GB.
 * This requires 25 bits for the packet counter and 31 bits for the byte counter.
 * 
 * This leaves 8 bits for the last-seen timestamp out of the 64bits read by the CPU
 * for each entry. If we count seconds, then this will overflow once every
 * 256s. Since the CPU is reading the counters every 3.2s, it will notice the 
 * overflow and update its internal counter for that flow entry.
 * 
 * So that each counter can be read in a single read, we set the timestamp
 * width to 7 so it would fit with the pkt counter.
 */
`define OPENFLOW_EXACT_ENTRY_PKT_COUNTER_WIDTH      25
`define OPENFLOW_EXACT_ENTRY_PKT_COUNTER_POS         0
`define OPENFLOW_EXACT_ENTRY_LAST_SEEN_WIDTH         7
`define OPENFLOW_EXACT_ENTRY_LAST_SEEN_POS          25
`define OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_WIDTH     32
`define OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_POS       32

/* location of words in the SRAM for each entry.
 * Entries in the SRAM are aligned to 16 words. That is
 * each entry takes up 16 words, although only 9 of 
 * them are used. */
`define OPENFLOW_EXACT_ENTRY_HDR_BASE_POS     4'h0
`define OPENFLOW_EXACT_ENTRY_COUNTERS_POS     4'h8
`define OPENFLOW_EXACT_ENTRY_ACTION_BASE_POS  4'ha

/* a replace action is specified as follows:
 * +------------------------+
 * |1| 3 |1|       16       | 
 * |V|WRD|B|    new data    |
 * +------------------------+
 */
`define OPENFLOW_REPLACEMENT_ACTION_VLD_BIT           23
`define OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_START    20
`define OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_LEN      3
`define OPENFLOW_REPLACEMENT_ACTION_2B_H_1B_L_BIT     16
`define OPENFLOW_REPLACEMENT_ACTION_NEW_DATA_WIDTH    16
`define OPENFLOW_REPLACEMENT_ACTION_CTRL              8'h10

/* an increment action is as follows:
 * +------------------------+
 * |1| 3 |1|       16       |
 * |V|WRD|0|     addend     |
 * +------------------------+
 */
`define OPENFLOW_INCREMENT_ACTION_VLD_BIT           23
`define OPENFLOW_INCREMENT_ACTION_WORD_SEL_START    20
`define OPENFLOW_INCREMENT_ACTION_WORD_SEL_LEN      3
`define OPENFLOW_INCREMENT_ACTION_ADDEND_WIDTH      16
`define OPENFLOW_INCREMENT_ACTION_CTRL              8'h20

/* VLAN Modification Action word
 */
`define OPENFLOW_VLAN_MOD_ACTION_SEL_POS            12
`define OPENFLOW_VLAN_MOD_ACTION_SEL_WIDTH          4
`define OPENFLOW_VLAN_MOD_ACTION_CTRL               8'h40
`define VLAN_MOD_DO_NOTHING                         4'h0
`define VLAN_MOD_ADD                                4'h1
`define VLAN_MOD_MODIFY                             4'h2
`define VLAN_MOD_ADD_MODIFY                         4'h3
`define VLAN_MOD_REMOVE                             4'h4

/* Register macros */
`define OPENFLOW_LOOKUP_REG_ADDR_WIDTH       `UDP_BLOCK_SIZE_64_REG_ADDR_WIDTH
`define OPENFLOW_LOOKUP_BLOCK_ADDR_WIDTH     `UDP_BLOCK_SIZE_64_BLOCK_ADDR_WIDTH
`define OPENFLOW_LOOKUP_BLOCK_ADDR           `OPENFLOW_LOOKUP_BLOCK_ADDR_WIDTH'h8
`define OPENFLOW_LOOKUP_BLOCK_TAG            ({`UDP_BLOCK_SIZE_64_TAG, `OPENFLOW_LOOKUP_BLOCK_ADDR})

`define OPENFLOW_WILDCARD_LOOKUP_REG_ADDR_WIDTH       `UDP_BLOCK_SIZE_1k_REG_ADDR_WIDTH
`define OPENFLOW_WILDCARD_LOOKUP_BLOCK_ADDR_WIDTH     `UDP_BLOCK_SIZE_1k_BLOCK_ADDR_WIDTH
`define OPENFLOW_WILDCARD_LOOKUP_BLOCK_ADDR           `OPENFLOW_WILDCARD_LOOKUP_BLOCK_ADDR_WIDTH'h3
`define OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG            ({`UDP_BLOCK_SIZE_1k_TAG, `OPENFLOW_WILDCARD_LOOKUP_BLOCK_ADDR})

`define OPENFLOW_LOOKUP_WILDCARD_MISSES      `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h0
`define OPENFLOW_LOOKUP_WILDCARD_HITS        `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h1
`define OPENFLOW_LOOKUP_EXACT_MISSES         `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h2
`define OPENFLOW_LOOKUP_EXACT_HITS           `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h3
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_0   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h4
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_1   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h5
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_2   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h6
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_3   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h7
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_4   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h8
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_5   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'h9
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_6   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'ha
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_7   `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'hb
// skip the two unused sw regs
`define OPENFLOW_LOOKUP_TIMER                `OPENFLOW_LOOKUP_REG_ADDR_WIDTH'he

`define OPENFLOW_WILDCARD_TABLE_SIZE            10'd32
`define OPENFLOW_WILDCARD_NUM_DATA_WORDS_USED   `OPENFLOW_ACTION_WIDTH/10'd32
`define OPENFLOW_WILDCARD_NUM_CMP_WORDS_USED    ((`OPENFLOW_ENTRY_WIDTH/10'd32)+1'b1)

`define OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE    `OPENFLOW_WILDCARD_LOOKUP_REG_ADDR_WIDTH'h0
`define OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE  (`OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE + `OPENFLOW_WILDCARD_NUM_DATA_WORDS_USED)
`define OPENFLOW_WILDCARD_LOOKUP_CMP_BASE       (`OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE + `OPENFLOW_WILDCARD_NUM_CMP_WORDS_USED)
`define OPENFLOW_WILDCARD_LOOKUP_READ_ADDR      (`OPENFLOW_WILDCARD_LOOKUP_CMP_BASE + `OPENFLOW_WILDCARD_NUM_CMP_WORDS_USED)
`define OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR     `OPENFLOW_WILDCARD_LOOKUP_READ_ADDR+`OPENFLOW_WILDCARD_LOOKUP_REG_ADDR_WIDTH'h1

`define OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE (`OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR + `OPENFLOW_WILDCARD_LOOKUP_REG_ADDR_WIDTH'h1)
`define OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE  (`OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE + `OPENFLOW_WILDCARD_TABLE_SIZE)
// skip the two unused sw regs
`define OPENFLOW_WILDCARD_LOOKUP_LAST_SEEN_TS   (`OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE + `OPENFLOW_WILDCARD_TABLE_SIZE + `OPENFLOW_WILDCARD_LOOKUP_REG_ADDR_WIDTH'h2)

`define OPENFLOW_LOOKUP_WILDCARD_MISSES_REG    (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_WILDCARD_MISSES})
`define OPENFLOW_LOOKUP_WILDCARD_HITS_REG      (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_WILDCARD_HITS})
`define OPENFLOW_LOOKUP_EXACT_MISSES_REG       (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_EXACT_MISSES})
`define OPENFLOW_LOOKUP_EXACT_HITS_REG         (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_EXACT_HITS})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_0_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_0})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_1_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_1})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_2_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_2})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_3_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_3})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_4_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_4})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_5_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_5})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_6_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_6})
`define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_7_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_7})
`define OPENFLOW_LOOKUP_TIMER_REG              (`UDP_BASE_ADDRESS | {`OPENFLOW_LOOKUP_BLOCK_TAG, `OPENFLOW_LOOKUP_TIMER})

`define OPENFLOW_WILDCARD_LOOKUP_READ_ADDR_REG     (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_READ_ADDR})
`define OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR_REG    (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR})
`define OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE_REG   (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE})
`define OPENFLOW_WILDCARD_LOOKUP_CMP_BASE_REG      (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_CMP_BASE})
`define OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE})

`define OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE_REG (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE})
`define OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE_REG  (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE})
`define OPENFLOW_WILDCARD_LOOKUP_LAST_SEEN_TS_REG   (`UDP_BASE_ADDRESS | {`OPENFLOW_WILDCARD_LOOKUP_BLOCK_TAG, `OPENFLOW_WILDCARD_LOOKUP_LAST_SEEN_TS})

`define PRINT_USER_REG_ADDRESSES                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_TABLE_SIZE                   %07d\n", `OPENFLOW_WILDCARD_TABLE_SIZE);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_NUM_DATA_WORDS_USED          %07d\n", `OPENFLOW_WILDCARD_NUM_DATA_WORDS_USED);              \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_NUM_CMP_WORDS_USED           %07d\n\n", `OPENFLOW_WILDCARD_NUM_CMP_WORDS_USED);             \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_WIDTH                           %07d\n", `OPENFLOW_ENTRY_WIDTH);                               \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_TRANSP_DST_WIDTH                %07d\n", `OPENFLOW_ENTRY_TRANSP_DST_WIDTH);                    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_TRANSP_DST_POS                  %07d\n", `OPENFLOW_ENTRY_TRANSP_DST_POS);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_TRANSP_SRC_WIDTH                %07d\n", `OPENFLOW_ENTRY_TRANSP_SRC_WIDTH);                    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_TRANSP_SRC_POS                  %07d\n", `OPENFLOW_ENTRY_TRANSP_SRC_POS);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_PROTO_WIDTH                  %07d\n", `OPENFLOW_ENTRY_IP_PROTO_WIDTH);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_PROTO_POS                    %07d\n", `OPENFLOW_ENTRY_IP_PROTO_POS);                        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_DST_WIDTH                    %07d\n", `OPENFLOW_ENTRY_IP_DST_WIDTH);                        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_DST_POS                      %07d\n", `OPENFLOW_ENTRY_IP_DST_POS);                          \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_SRC_WIDTH                    %07d\n", `OPENFLOW_ENTRY_IP_SRC_WIDTH);                        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_IP_SRC_POS                      %07d\n", `OPENFLOW_ENTRY_IP_SRC_POS);                          \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_VLAN_ID_WIDTH                   %07d\n", `OPENFLOW_ENTRY_VLAN_ID_WIDTH);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_VLAN_ID_POS                     %07d\n", `OPENFLOW_ENTRY_VLAN_ID_POS);                         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_TYPE_WIDTH                  %07d\n", `OPENFLOW_ENTRY_ETH_TYPE_WIDTH);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_TYPE_POS                    %07d\n", `OPENFLOW_ENTRY_ETH_TYPE_POS);                        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_DST_WIDTH                   %07d\n", `OPENFLOW_ENTRY_ETH_DST_WIDTH);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_DST_POS                     %07d\n", `OPENFLOW_ENTRY_ETH_DST_POS);                         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_SRC_WIDTH                   %07d\n", `OPENFLOW_ENTRY_ETH_SRC_WIDTH);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_ETH_SRC_POS                     %07d\n", `OPENFLOW_ENTRY_ETH_SRC_POS);                         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_SRC_PORT_WIDTH                  %07d\n", `OPENFLOW_ENTRY_SRC_PORT_WIDTH);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ENTRY_SRC_PORT_POS                    %07d\n\n", `OPENFLOW_ENTRY_SRC_PORT_POS);                      \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_ACTION_WIDTH                          %07d\n", `OPENFLOW_ACTION_WIDTH);                              \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_FORWARD_BITMASK_WIDTH                 %07d\n", `OPENFLOW_FORWARD_BITMASK_WIDTH);                     \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_FORWARD_BITMASK_POS                   %07d\n", `OPENFLOW_FORWARD_BITMASK_POS);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_PKT_TRIM_ACTION_WIDTH                 %07d\n", `OPENFLOW_PKT_TRIM_ACTION_WIDTH);                     \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_PKT_TRIM_ACTION_POS                   %07d\n", `OPENFLOW_PKT_TRIM_ACTION_POS);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_WIDTH              %07d\n", `OPENFLOW_REPLACEMENT_ACTION_WIDTH);                  \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_NUM_REPLACEMENT_ACTIONS               %07d\n", `OPENFLOW_NUM_REPLACEMENT_ACTIONS);                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_POS                %07d\n", `OPENFLOW_REPLACEMENT_ACTION_POS);                    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_BITMASK_ACTION_WIDTH      %07d\n", `OPENFLOW_REPLACEMENT_BITMASK_ACTION_WIDTH);          \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_NUM_REPLACEMENT_BITMASK_ACTIONS       %07d\n", `OPENFLOW_NUM_REPLACEMENT_BITMASK_ACTIONS);           \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_BITMASK_ACTION_POS        %07d\n", `OPENFLOW_REPLACEMENT_BITMASK_ACTION_POS);            \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_WIDTH                %07d\n", `OPENFLOW_INCREMENT_ACTION_WIDTH);                    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_NUM_INCREMENT_ACTIONS                 %07d\n", `OPENFLOW_NUM_INCREMENT_ACTIONS);                     \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_POS                  %07d\n", `OPENFLOW_INCREMENT_ACTION_POS);                      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_VLAN_MOD_ACTION_POS                   %07d\n", `OPENFLOW_VLAN_MOD_ACTION_POS);                       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_VLAN_MOD_ACTION_WIDTH                 %07d\n\n", `OPENFLOW_VLAN_MOD_ACTION_WIDTH);                   \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_VLD_BIT            %07d\n", `OPENFLOW_REPLACEMENT_ACTION_VLD_BIT);                \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_START     %07d\n", `OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_START);         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_LEN       %07d\n", `OPENFLOW_REPLACEMENT_ACTION_WORD_SEL_LEN);           \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_2B_H_1B_L_BIT      %07d\n", `OPENFLOW_REPLACEMENT_ACTION_2B_H_1B_L_BIT);          \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_NEW_DATA_WIDTH     %07d\n", `OPENFLOW_REPLACEMENT_ACTION_NEW_DATA_WIDTH);         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_REPLACEMENT_ACTION_CTRL               0x%02x\n\n", `OPENFLOW_REPLACEMENT_ACTION_CTRL);               \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_VLD_BIT              %07d\n", `OPENFLOW_INCREMENT_ACTION_VLD_BIT);                  \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_WORD_SEL_START       %07d\n", `OPENFLOW_INCREMENT_ACTION_WORD_SEL_START);           \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_WORD_SEL_LEN         %07d\n", `OPENFLOW_INCREMENT_ACTION_WORD_SEL_LEN);             \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_ADDEND_WIDTH         %07d\n", `OPENFLOW_INCREMENT_ACTION_ADDEND_WIDTH);             \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_INCREMENT_ACTION_CTRL                 0x%02x\n\n", `OPENFLOW_INCREMENT_ACTION_CTRL);                 \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_VLAN_MOD_ACTION_SEL_POS               %07d\n", `OPENFLOW_VLAN_MOD_ACTION_SEL_POS);                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_VLAN_MOD_ACTION_SEL_WIDTH             %07d\n", `OPENFLOW_VLAN_MOD_ACTION_SEL_WIDTH);                 \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_VLAN_MOD_ACTION_CTRL                  0x%02x\n\n", `OPENFLOW_VLAN_MOD_ACTION_CTRL);                  \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define VLAN_MOD_DO_NOTHING                            0x%x\n", `VLAN_MOD_DO_NOTHING);                                \
  $fwrite(c_reg_defines_fd, "#define VLAN_MOD_ADD                                   0x%x\n", `VLAN_MOD_ADD);                                       \
  $fwrite(c_reg_defines_fd, "#define VLAN_MOD_MODIFY                                0x%x\n", `VLAN_MOD_MODIFY);                                    \
  $fwrite(c_reg_defines_fd, "#define VLAN_MOD_ADD_MODIFY                            0x%x\n", `VLAN_MOD_ADD_MODIFY);                                \
  $fwrite(c_reg_defines_fd, "#define VLAN_MOD_REMOVE                                0x%x\n\n", `VLAN_MOD_REMOVE);                                  \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_HDR_BASE_POS              %07d\n", `OPENFLOW_EXACT_ENTRY_HDR_BASE_POS);                  \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_COUNTERS_POS              %07d\n", `OPENFLOW_EXACT_ENTRY_COUNTERS_POS);                  \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_ACTION_BASE_POS           %07d\n\n", `OPENFLOW_EXACT_ENTRY_ACTION_BASE_POS);             \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_PKT_COUNTER_WIDTH         %07d\n", `OPENFLOW_EXACT_ENTRY_PKT_COUNTER_WIDTH);             \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_WIDTH        %07d\n", `OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_WIDTH);            \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_LAST_SEEN_WIDTH           %07d\n", `OPENFLOW_EXACT_ENTRY_LAST_SEEN_WIDTH);               \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_PKT_COUNTER_POS           %07d\n", `OPENFLOW_EXACT_ENTRY_PKT_COUNTER_POS);               \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_POS          %07d\n", `OPENFLOW_EXACT_ENTRY_BYTE_COUNTER_POS);              \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_EXACT_ENTRY_LAST_SEEN_POS             %07d\n\n", `OPENFLOW_EXACT_ENTRY_LAST_SEEN_POS);               \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_WILDCARD_MISSES_REG            0x%07x\n", `OPENFLOW_LOOKUP_WILDCARD_MISSES_REG<<2);           \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_WILDCARD_HITS_REG              0x%07x\n", `OPENFLOW_LOOKUP_WILDCARD_HITS_REG<<2);             \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_EXACT_MISSES_REG               0x%07x\n", `OPENFLOW_LOOKUP_EXACT_MISSES_REG<<2);              \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_EXACT_HITS_REG                 0x%07x\n", `OPENFLOW_LOOKUP_EXACT_HITS_REG<<2);                \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_0_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_0_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_1_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_1_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_2_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_2_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_3_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_3_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_4_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_4_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_5_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_5_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_6_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_6_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_7_REG         0x%07x\n", `OPENFLOW_LOOKUP_NUM_PKTS_DROPPED_7_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_LOOKUP_TIMER_REG                      0x%07x\n\n", `OPENFLOW_LOOKUP_TIMER_REG<<2);                   \
                                                                                                                                                   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_READ_ADDR_REG         0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_READ_ADDR_REG<<2);        \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR_REG        0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_WRITE_ADDR_REG<<2);       \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE_REG       0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_ACTION_BASE_REG<<2);      \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_CMP_BASE_REG          0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_CMP_BASE_REG<<2);         \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE_REG     0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_CMP_MASK_BASE_REG<<2);    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE_REG    0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_BYTES_HIT_BASE_REG<<2);   \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE_REG     0x%07x\n", `OPENFLOW_WILDCARD_LOOKUP_PKTS_HIT_BASE_REG<<2);    \
  $fwrite(c_reg_defines_fd, "#define OPENFLOW_WILDCARD_LOOKUP_LAST_SEEN_TS_REG      0x%07x\n\n", `OPENFLOW_WILDCARD_LOOKUP_LAST_SEEN_TS_REG<<2)
