///////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: pkt_gen_ctrl_reg_full.v 4104 2008-06-17 22:20:05Z grg $
//
// Module: pkt_gen_ctrl_reg_full.v
// Project: Packet generator ctrl registers
// Description: Full implementation of the packet gen control registers
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps
module pkt_gen_ctrl_reg_full
   #(
      parameter UDP_REG_SRC_WIDTH = 2,
      parameter BYTE_CNT_WIDTH = 40,
      parameter TIME_WIDTH = 64
   )
   (
      // --- data path interface
      output reg [3:0]                      enable,

      input [`CPCI_NF2_DATA_WIDTH - 1:0]    pkt_cnt_0, 
      input [BYTE_CNT_WIDTH - 1:0]          byte_cnt_0,
      input [TIME_WIDTH - 1:0]              time_first_0,
      input [TIME_WIDTH - 1:0]              time_last_0,
   
      input [`CPCI_NF2_DATA_WIDTH - 1:0]    pkt_cnt_1, 
      input [BYTE_CNT_WIDTH - 1:0]          byte_cnt_1,
      input [TIME_WIDTH - 1:0]              time_first_1,
      input [TIME_WIDTH - 1:0]              time_last_1,
   
      input [`CPCI_NF2_DATA_WIDTH - 1:0]    pkt_cnt_2, 
      input [BYTE_CNT_WIDTH - 1:0]          byte_cnt_2,
      input [TIME_WIDTH - 1:0]              time_first_2,
      input [TIME_WIDTH - 1:0]              time_last_2,
   
      input [`CPCI_NF2_DATA_WIDTH - 1:0]    pkt_cnt_3, 
      input [BYTE_CNT_WIDTH - 1:0]          byte_cnt_3,
      input [TIME_WIDTH - 1:0]              time_first_3,
      input [TIME_WIDTH - 1:0]              time_last_3,

      // --- Register interface
      input                                 reg_req_in,
      input                                 reg_ack_in,
      input                                 reg_rd_wr_L_in,
      input  [`UDP_REG_ADDR_WIDTH-1:0]      reg_addr_in,
      input  [`CPCI_NF2_DATA_WIDTH-1:0]     reg_data_in,
      input  [UDP_REG_SRC_WIDTH-1:0]        reg_src_in,

      output reg                            reg_req_out,
      output reg                            reg_ack_out,
      output reg                            reg_rd_wr_L_out,
      output reg [`UDP_REG_ADDR_WIDTH-1:0]  reg_addr_out,
      output reg [`CPCI_NF2_DATA_WIDTH-1:0] reg_data_out,
      output reg [UDP_REG_SRC_WIDTH-1:0]    reg_src_out,


      // --- Misc
      input                                 clk,
      input                                 reset
   );

   function integer log2;
      input integer number;
      begin
         log2=0;
         while(2**log2<number) begin
            log2=log2+1;
         end
      end
   endfunction // log2
   
   // ------------- Internal parameters --------------
   localparam NUM_REGS_USED = 29;
   localparam ADDR_WIDTH = log2(NUM_REGS_USED);

   // ------------- Wires/reg ------------------

   wire [ADDR_WIDTH-1:0]                  addr;
   wire [`PKT_GEN_CTRL_REG_ADDR_WIDTH - 1:0] reg_addr;
   wire [`UDP_REG_ADDR_WIDTH-`PKT_GEN_CTRL_REG_ADDR_WIDTH - 1:0] tag_addr;

   wire                                   addr_good;
   wire                                   tag_hit;


   // -------------- Logic --------------------

   assign addr = reg_addr_in[ADDR_WIDTH-1:0];
   assign reg_addr = reg_addr_in[`PKT_GEN_CTRL_REG_ADDR_WIDTH-1:0];
   assign tag_addr = reg_addr_in[`UDP_REG_ADDR_WIDTH - 1:`PKT_GEN_CTRL_REG_ADDR_WIDTH];

   assign addr_good = reg_addr < NUM_REGS_USED;
   assign tag_hit = tag_addr == `PKT_GEN_CTRL_BLOCK_TAG;


   /* run the counters and mux between write and update */
   always @(posedge clk) begin
      if(reset) begin
         enable <= 'h0;
         
         reg_req_out                                    <= 0;
         reg_ack_out                                    <= 0;
         reg_rd_wr_L_out                                <= 0;
         reg_addr_out                                   <= 0;
         reg_data_out                                   <= 0;
         reg_src_out                                    <= 0;

      end // if (reset)
      else begin
         reg_req_out       <= reg_req_in;
         reg_rd_wr_L_out   <= reg_rd_wr_L_in;
         reg_addr_out      <= reg_addr_in;
         reg_src_out       <= reg_src_in;

         // Check if we see a request and the tags match (we need to
         // respond)
         if (reg_req_in && tag_hit) begin

            if (!reg_rd_wr_L_in) begin // write
               // Update the appropriate register
               case (addr)
                  `PKT_GEN_CTRL_ENABLE    : enable <= reg_data_in;
               endcase
               reg_data_out      <= reg_data_in;
            end
            else begin // read
               if (addr_good) begin
                  case (addr)
                     `PKT_GEN_CTRL_ENABLE          : reg_data_out <= enable;

                     `PKT_GEN_CTRL_PKT_CNT_0       : reg_data_out <= pkt_cnt_0;
                     `PKT_GEN_CTRL_BYTE_CNT_HI_0   : reg_data_out <= byte_cnt_0[BYTE_CNT_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_BYTE_CNT_LO_0   : reg_data_out <= byte_cnt_0[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_FIRST_HI_0 : reg_data_out <= time_first_0[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_FIRST_LO_0 : reg_data_out <= time_first_0[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_LAST_HI_0  : reg_data_out <= time_last_0[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_LAST_LO_0  : reg_data_out <= time_last_0[`CPCI_NF2_DATA_WIDTH-1:0];

                     `PKT_GEN_CTRL_PKT_CNT_1       : reg_data_out <= pkt_cnt_1;
                     `PKT_GEN_CTRL_BYTE_CNT_HI_1   : reg_data_out <= byte_cnt_1[BYTE_CNT_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_BYTE_CNT_LO_1   : reg_data_out <= byte_cnt_1[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_FIRST_HI_1 : reg_data_out <= time_first_1[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_FIRST_LO_1 : reg_data_out <= time_first_1[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_LAST_HI_1  : reg_data_out <= time_last_1[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_LAST_LO_1  : reg_data_out <= time_last_1[`CPCI_NF2_DATA_WIDTH-1:0];

                     `PKT_GEN_CTRL_PKT_CNT_2       : reg_data_out <= pkt_cnt_2;
                     `PKT_GEN_CTRL_BYTE_CNT_HI_2   : reg_data_out <= byte_cnt_2[BYTE_CNT_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_BYTE_CNT_LO_2   : reg_data_out <= byte_cnt_2[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_FIRST_HI_2 : reg_data_out <= time_first_2[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_FIRST_LO_2 : reg_data_out <= time_first_2[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_LAST_HI_2  : reg_data_out <= time_last_2[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_LAST_LO_2  : reg_data_out <= time_last_2[`CPCI_NF2_DATA_WIDTH-1:0];

                     `PKT_GEN_CTRL_PKT_CNT_3       : reg_data_out <= pkt_cnt_3;
                     `PKT_GEN_CTRL_BYTE_CNT_HI_3   : reg_data_out <= byte_cnt_3[BYTE_CNT_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_BYTE_CNT_LO_3   : reg_data_out <= byte_cnt_3[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_FIRST_HI_3 : reg_data_out <= time_first_3[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_FIRST_LO_3 : reg_data_out <= time_first_3[`CPCI_NF2_DATA_WIDTH-1:0];
                     `PKT_GEN_CTRL_TIME_LAST_HI_3  : reg_data_out <= time_last_3[TIME_WIDTH-1:`CPCI_NF2_DATA_WIDTH];
                     `PKT_GEN_CTRL_TIME_LAST_LO_3  : reg_data_out <= time_last_3[`CPCI_NF2_DATA_WIDTH-1:0];

                     default                       : reg_data_out <= 32'h DEAD_BEEF;
                  endcase
               end
               else
                  reg_data_out <= 32'h DEAD_BEEF;
            end
            reg_ack_out       <= 1'b1;
         end
         else begin
            reg_ack_out       <= reg_ack_in;
            reg_data_out      <= reg_data_in;
         end
      end // if else
   end // always @ (posedge clk)
   
endmodule
