   /*
    * top decapsulation module, used in user_data_path for IP-IP tuneling. 
    * This module is mounted to input_arbiter in the user_data_path.
    */
  
 
module decap
    #(parameter DATA_WIDTH = 64,
      parameter CTRL_WIDTH=DATA_WIDTH/8,
      parameter UDP_REG_SRC_WIDTH = 2)

   (// --- data path interface
    output  [DATA_WIDTH-1:0]           out_data,
    output  [CTRL_WIDTH-1:0]           out_ctrl,
    output                             out_wr,
    input                              out_rdy,
    
    input  [DATA_WIDTH-1:0]            in_data,
    input  [CTRL_WIDTH-1:0]            in_ctrl,
    input                              in_wr,
    output                             in_rdy,

    // --- Register interface
    input                              reg_req_in,
    input                              reg_ack_in,
    input                              reg_rd_wr_L_in,
    input  [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_in,
    input  [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_in,
    input  [UDP_REG_SRC_WIDTH-1:0]     reg_src_in,

    output                             reg_req_out,
    output                             reg_ack_out,
    output                             reg_rd_wr_L_out,
    output  [`UDP_REG_ADDR_WIDTH-1:0]  reg_addr_out,
    output  [`CPCI_NF2_DATA_WIDTH-1:0] reg_data_out,
    output  [UDP_REG_SRC_WIDTH-1:0]    reg_src_out,

    // --- Misc
    
    input                              clk,
    input                              reset);


   // --- wire/regs


   wire [31:0] 			       soft_reg1;
   wire [31:0] 			       soft_reg2;
   wire [31:0] 			       hard_reg1;
   wire [31:0] 			       hard_reg2;
   
   


   // --- regs/wires
   
   wire [DATA_WIDTH-1:0] 	       mid_out_data;
   reg [2:0] 			       drop_cnt;
   reg [DATA_WIDTH-1:0] 	       user_header;
   wire [DATA_WIDTH-1:0] 	       user_header_up;
   wire 			       decap_data_nearly_full;
   wire 			       fifo_rd;
   wire 			       decap_data_empty;
   reg 				       valid_tuneling;
   reg 				       mid_out_wr;
   wire [15:0] 			       word_len;
   wire [15:0]			       byte_len;
   reg 				       start_reading_fifo;
   reg [2:0] 			       counter;

   reg [31:0] 			       debug_pkt_counter_out;
   reg [31:0] 			       debug_pkt_counter_in;
   reg [31:0] 			       debug_word_counter_out;
   reg [31:0] 			       debug_word_counter_in;
   reg [31:0] 			       proto1;
   reg [31:0] 			       proto2;
   reg [31:0] 			       proto3;
   reg [31:0] 			       proto4;

   reg 				       last_word_in;
   
   
   // --- logic
   // --- handshaking signals
   
   assign out_wr = (mid_out_wr & ~valid_tuneling) | (mid_out_wr & valid_tuneling & ~((drop_cnt < 5) & (drop_cnt > 1))); 
      
  
   assign in_rdy = ~decap_data_nearly_full & ~last_word_in;
   
   always @(posedge clk) 
   begin
      if (reset)
	mid_out_wr <= 1'b0;
      else
	mid_out_wr <= fifo_rd;
   end
   
   assign fifo_rd = (~decap_data_empty) & (out_rdy) & (start_reading_fifo);
   
   assign fifo_wr = in_wr;


   always @(posedge clk) begin
      if (reset)
	last_word_in <= 0;
      else begin
	 if (in_wr && ((in_ctrl != 8'hFF) && (in_ctrl != 8'h00)))
	   last_word_in <= 1'b1;
	 else if (out_wr && ((out_ctrl != 8'hFF) && (out_ctrl != 8'h00)))
	   last_word_in <= 1'b0;	 
      end
   end   
    // --- debugging resgiters
   
   always @(posedge clk) begin
      if (reset)
	debug_word_counter_out <= 0;
      else if(out_wr)
	debug_word_counter_out <= debug_word_counter_out + 1;
   end

   always @(posedge clk) begin
      if (reset)
	debug_word_counter_in <= 0;
      else if(in_wr)
	debug_word_counter_in <= debug_word_counter_in + 1;
   end
   
   always @(posedge clk) begin
      if(reset)
	debug_pkt_counter_in <= 0;
      else if(in_ctrl == 8'hFF && in_wr == 1)
	debug_pkt_counter_in <= debug_pkt_counter_in + 1;
   end
   
   always @(posedge clk) begin
      if(reset)
	debug_pkt_counter_out <= 0;
      else if(out_ctrl == 8'hFF && out_wr == 1)
	debug_pkt_counter_out <= debug_pkt_counter_out + 1;
   end
   
	
   
   small_fifo
     #(.WIDTH    (DATA_WIDTH+CTRL_WIDTH),
       .MAX_DEPTH_BITS (3))decap_fifo_data 
       
       (
        .din         ({in_data, in_ctrl}),       // Data in
        .wr_en       (fifo_wr),       // Write enable
	
        .rd_en       (fifo_rd),       // Read the next word
        
        .dout         ({mid_out_data, out_ctrl}),     // Data out  
        .full         (decap_data_full),
        .nearly_full  (decap_data_nearly_full),
        .empty        (decap_data_empty),
        
        .reset        (reset),
        .clk          (clk)
        );

   
   
   // check if it is valid tuneling data


   always @(posedge clk)
     begin
	if (reset) begin
	   valid_tuneling <= 1'b0;
	   start_reading_fifo <= 1'b0;
	   proto1 <= 32'b0;
	   proto2 <= 32'b0;
	   proto3 <= 32'b0;
	  
	end
	else if (in_wr)  begin
	   if (in_ctrl == 8'hFF) begin
	      valid_tuneling <= 1'b0;
	      start_reading_fifo <= 1'b0;
	   end
	   else begin 
	      if(counter == 3'd2) begin
		 start_reading_fifo <= 1'b1;
		 if (in_data[7:0] == 8'd244) begin
		    valid_tuneling <= 1'b1;
		 end		
	      end
	   end
	end // if (in_wr)
     end // always @ (posedge clk)
   
   
   //-- check the ip of the data in incoming line
   always @(posedge clk)
   begin
      if (reset)
	counter <= 3'b0;
      else if (in_wr)
	if (in_ctrl == 8'hFF)
	  counter <= 3'b0;
	else if (counter < 3'd4)
	  counter <= counter + 1;
   end // always @ (posedge clk)
   

   always @(posedge clk)
   begin
      if (reset || (mid_out_wr && (out_ctrl == 8'hFF)))
	drop_cnt <= 3'b0;
      else if (mid_out_wr && (drop_cnt < 5))
	drop_cnt <= drop_cnt +1;
   end

   //--- handling the user_data_path header

   always @(posedge clk) begin
      //-- save the first word of data
      if(reset)
	user_header <= {DATA_WIDTH{1'b0}};
      else if(in_wr && (in_ctrl == 8'hFF))
	user_header <= in_data;    
   end

   assign word_len = user_header[47:32]-3;
   assign byte_len = user_header[15:0]-24;
   
   assign user_header_up = {user_header[63:48],word_len, user_header[31:16], byte_len};
   
   assign out_data = (valid_tuneling && mid_out_wr && (out_ctrl == 8'hFF)) ? user_header_up : mid_out_data;
   
  

     generic_regs
   #( 
      .UDP_REG_SRC_WIDTH    (UDP_REG_SRC_WIDTH),
      .TAG                  (`DECAP_BLOCK_TAG),
      .REG_ADDR_WIDTH       (`DECAP_REG_ADDR_WIDTH),  
      .NUM_COUNTERS         (2),
      .NUM_SOFTWARE_REGS    (2),
      .NUM_HARDWARE_REGS    (6),
      .COUNTER_INPUT_WIDTH  (2)
   ) generic_regs
   (
    .reg_req_in      (reg_req_in),
    .reg_ack_in      (reg_ack_in),
    .reg_rd_wr_L_in  (reg_rd_wr_L_in),
    .reg_addr_in     (reg_addr_in),
    .reg_data_in     (reg_data_in),
    .reg_src_in      (reg_src_in),
    
    .reg_req_out     (reg_req_out),
    .reg_ack_out     (reg_ack_out),
    .reg_rd_wr_L_out (reg_rd_wr_L_out),
    .reg_addr_out    (reg_addr_out),
    .reg_data_out    (reg_data_out),
    .reg_src_out     (reg_src_out),

    // --- counters interface
    .counter_updates  (4'b0),   // all the counter updates are concatenated
    .counter_decrement(2'b0), // if 1 then subtract the update, else add.
    
    // --- SW regs interface
    .software_regs    ({temp, soft_reg}), // signals from the software
    
    // --- HW regs interface
    .hardware_regs  ({debug_pkt_counter_out, debug_pkt_counter_in, user_header_up,  valid_tuneling, debug_word_counter_out, debug_word_counter_in}),
    
    .clk           (clk),
    .reset         (reset));


   assign hard_reg1 = 32'b0;
   assign hard_reg2 = 32'b0;
   

endmodule // decap



