 module encap
    #(parameter DATA_WIDTH = 64,
      parameter CTRL_WIDTH=DATA_WIDTH/8,
      parameter UDP_REG_SRC_WIDTH = 2)

   (// --- data path interface
    output  [DATA_WIDTH-1:0]            out_data,
    output  [CTRL_WIDTH-1:0]            out_ctrl,
    output reg                         out_wr,
    input                              out_rdy,
    
    input  [DATA_WIDTH-1:0]            in_data,
    input  [CTRL_WIDTH-1:0]            in_ctrl,
    input                              in_wr,
    output                             in_rdy,

    // --- Register interface
    input                              reg_req_in,
    input                              reg_ack_in,
    input                              reg_rd_wr_L_in,
    input  [`UDP_REG_ADDR_WIDTH-1:0]   reg_addr_in,
    input  [`CPCI_NF2_DATA_WIDTH-1:0]  reg_data_in,
    input  [UDP_REG_SRC_WIDTH-1:0]     reg_src_in,

    output                             reg_req_out,
    output                             reg_ack_out,
    output                             reg_rd_wr_L_out,
    output  [`UDP_REG_ADDR_WIDTH-1:0]  reg_addr_out,
    output  [`CPCI_NF2_DATA_WIDTH-1:0] reg_data_out,
    output  [UDP_REG_SRC_WIDTH-1:0]    reg_src_out,

    // --- Misc
    input                              clk,
    input                              reset);

   
   //--- wires/ registers
   wire 			       encap_data_nearly_full;
   wire 			       encap_data_empty;
   wire  			       fifo_wr;
     
   wire [1:0] 			       sel;
   wire 			       mask_fifo_rd;
   wire				       mask_fifo_rd_2;
   
   wire [31:0]			       encap_enable;
   wire [31:0] 			       ip_data_0;
   wire [31:0] 			       ip_data_1;
   wire [31:0] 			       ip_data_2;
   wire [31:0] 			       ip_data_3;
   wire [31:0] 			       ip_data_4;
   wire [31:0] 			       ip_data_5;

   reg [DATA_WIDTH-1:0] 	       updated_header;
   wire [DATA_WIDTH-1:0] 	       fifo_out_data;
   reg [DATA_WIDTH-1:0] 	       mid_out_data;

   wire [CTRL_WIDTH-1:0] 	       fifo_out_ctrl;
   reg [CTRL_WIDTH-1:0] 	       mid_out_ctrl;

   reg 				       packet_in;
   reg 				       enable_packet_in;
   reg 				       disable_packet_in;
   
   wire [31:0] 			       ip;
   reg  [31:0] 			       ip_reg;
   reg   			       ip_match;
   
   wire 			       ip_header_enable_1;
   wire 			       ip_header_enable_2;



   //--- debug registers
   reg [31:0] 			       num_packet_in;
   reg [31:0] 			       num_packet_out;
   reg [31:0] 			       num_bytes_in;
   reg [31:0] 			       num_bytes_out;
   wire [7:0] 			       out_state;
   reg [31:0] 			       com_reg;
   
   

   
   // --- Debug registers
   //--- Counting the number of packets come into the encap module
   always @(posedge clk)begin
      if(reset)
	num_packet_in <= 32'b0;
      else if ((in_ctrl == 8'hFF) & in_wr)
	num_packet_in <= num_packet_in + 1;
   end

   //--- Counting the number of packets come go out of encap module
   always @(posedge clk)begin
      if(reset)
	num_packet_out <= 32'b0;
      else if ((out_ctrl == 8'hFF) & out_wr)
	num_packet_out <= num_packet_out + 1;
   end

   //--- Counting the number of bytes come into the encap module
   always @(posedge clk)begin
      if(reset)
	num_bytes_in <= 32'b0;
      else if (in_wr)
	num_bytes_in <= num_bytes_in + 1;
  end

   //--- Counting the number of bytes go out of encap module
   always @(posedge clk)begin
      if(reset)
	num_bytes_out <= 32'b0;
      else if (out_wr)
	num_bytes_out <= num_bytes_out + 1;
  end

   always @(posedge clk) begin
     if(reset) 
        com_reg <= 32'b0;
    else
       com_reg <= {19'b0, out_state, sel, mask_fifo_rd, in_wr, out_wr, in_rdy, out_rdy};
   end

   // --- handshaking signals
   assign  fifo_rd = out_rdy & !encap_data_empty & mask_fifo_rd & mask_fifo_rd_2;
   assign  in_rdy = !encap_data_nearly_full & !packet_in;
   assign  fifo_wr = in_wr;
   
   always @(posedge clk) begin
      if(reset)
	out_wr <= 1'b0;      
      else if(sel == 2'b00)
	out_wr <= fifo_rd;
      else // this is insert mode
	out_wr <= out_rdy;
   end


  
   //while end of packet is in the fifo don't read the new packet mask the in_rdy

   always @(posedge clk) begin
      if(reset)
	packet_in <= 1'b0;
      else if(in_ctrl !=8'hFF & in_ctrl !=8'h00 & in_wr == 1'b1)
	packet_in <= 1'b1;
      else if(out_ctrl !=8'hFF & out_ctrl !=8'h00 & out_wr == 1'b1)
	packet_in <= 1'b0;
   end
   
   //--- Update user_data_path header
   always @(posedge clk) begin
      if(reset)
	updated_header <= {DATA_WIDTH{1'b0}};
      else if((in_ctrl == 8'hFF) && in_wr) begin
	 updated_header[DATA_WIDTH-1:DATA_WIDTH-16]    <=  in_data[DATA_WIDTH-1:DATA_WIDTH-16];
	 updated_header[DATA_WIDTH-1-16:DATA_WIDTH-32] <=  in_data[DATA_WIDTH-1-16:DATA_WIDTH-32] + 3;
	 updated_header[DATA_WIDTH-1-32:DATA_WIDTH-48] <=  in_data[DATA_WIDTH-1-32:DATA_WIDTH-48];
	 updated_header[DATA_WIDTH-1-48:0]             <=  in_data[DATA_WIDTH-1-48:0] + 24;
      end	
   end

   //--- Detect the IP destination
   always @(posedge clk) begin
      if(reset)
	ip_reg <= 32'h0;
      else if(ip_header_enable_1)
	ip_reg[31:16] <= in_data[DATA_WIDTH-1-48:0];
      else if(ip_header_enable_2)
	ip_reg[15:0] <= in_data[DATA_WIDTH-1:DATA_WIDTH-16];
      else if(in_ctrl != 8'hFF & in_ctrl != 8'h00)
	ip_reg <= 32'b0;      
   end

   always @(posedge clk) begin
      if (reset)
	ip_match <= 1'b0;
      else
	ip_match <= &(ip ^~ ip_reg);
   end
   

 
   //assign ip_match = &(ip ^~ ip_reg);
   
   //-----Adding the encap IP header to the packet, IP header is prepared by the software
   always @(*)
     begin
	if (sel == 2'b00 & fifo_out_ctrl ==8'hFF & encap_enable[0] & ip_match) begin
	   mid_out_data = updated_header;
	   mid_out_ctrl = fifo_out_ctrl;
	end
	else if(sel == 2'b00) begin
	   mid_out_data = fifo_out_data;
	   mid_out_ctrl = fifo_out_ctrl;
	end
	else if (fifo_out_ctrl==8'hFF) begin
	      mid_out_data = updated_header;
	      mid_out_ctrl = fifo_out_ctrl;
	end
	else if (sel == 2'b01) begin
	   mid_out_data = {ip_data_0,ip_data_1};
	   mid_out_ctrl = 8'b00;
	end
	else if (sel == 2'b10) begin
	   mid_out_data = {ip_data_2,ip_data_3};
	   mid_out_ctrl = 8'b00;
	end
	else if (sel == 2'b11) begin
	   mid_out_data = {ip_data_4,ip_data_5};
	   mid_out_ctrl = 8'b00;
	end
	else begin
	   mid_out_data = 64'b0;
	   mid_out_ctrl = 8'b00;
	end
	   
     end

   assign out_ctrl = mid_out_ctrl;
   assign out_data = mid_out_data;
   
		     

   small_fifo
     #(.WIDTH    (DATA_WIDTH+CTRL_WIDTH),
       .MAX_DEPTH_BITS (4))encap_fifo_data
       
       (
        
        .din         ({in_data, in_ctrl}),        // Data in
        .wr_en       (fifo_wr),                    // Write enable
        .rd_en       (fifo_rd),                    // Read the next word
        
        .dout         ({fifo_out_data, fifo_out_ctrl}), // Data out  
        .full         (encap_data_full),
        .nearly_full  (encap_data_nearly_full),
        .empty        (encap_data_empty),
        
        .reset        (reset),
        .clk          (clk)
        );


   //--- controller
   encap_control
     #(.DATA_WIDTH   (DATA_WIDTH),
       .CTRL_WIDTH   (CTRL_WIDTH)
       ) encap_control
       (
	.out_ctrl      (out_ctrl),
	.out_rdy       (out_rdy),
	.encap_enable  (encap_enable[0] & ip_match),
	.out_we        (out_wr),
	
	.sel           (sel),
	.mask_fifo_rd  (mask_fifo_rd),
	.out_state     (out_state),
	
	// --- Misc   
	.reset        (reset),
	.clk          (clk)
	);


   //--- ip controller
   encap_ip_control
    #(.DATA_WIDTH    (DATA_WIDTH),
      .CTRL_WIDTH    (CTRL_WIDTH)
      )encap_ip_control
   (
    .in_ctrl                 (in_ctrl),
    .in_we                   (in_wr),
    .packet_in               (packet_in),
    
    .ip_header_enable_1      (ip_header_enable_1),
    .ip_header_enable_2      (ip_header_enable_2),
    .mask_fifo_rd_2          (mask_fifo_rd_2),
    
    // --- Misc   
    .reset                   (reset),
    .clk                     (clk)
   );

   //--- registers
   generic_regs
   #( 
      .UDP_REG_SRC_WIDTH    (UDP_REG_SRC_WIDTH),
      .TAG                  (`ENCAP_BLOCK_TAG),
      .REG_ADDR_WIDTH       (`ENCAP_REG_ADDR_WIDTH),  
      .NUM_COUNTERS         (2),
      .NUM_SOFTWARE_REGS    (8),  //remember to change the number if you add new registers
      .NUM_HARDWARE_REGS    (4),
      .COUNTER_INPUT_WIDTH  (2)
   ) generic_regs
   (
    .reg_req_in      (reg_req_in),
    .reg_ack_in      (reg_ack_in),
    .reg_rd_wr_L_in  (reg_rd_wr_L_in),
    .reg_addr_in     (reg_addr_in),
    .reg_data_in     (reg_data_in),
    .reg_src_in      (reg_src_in),
    
    .reg_req_out     (reg_req_out),
    .reg_ack_out     (reg_ack_out),
    .reg_rd_wr_L_out (reg_rd_wr_L_out),
    .reg_addr_out    (reg_addr_out),
    .reg_data_out    (reg_data_out),
    .reg_src_out     (reg_src_out),

    // --- counters interface
    .counter_updates  (4'b0),   // all the counter updates are concatenated
    .counter_decrement(2'b0),   // if 1 then subtract the update, else add.
    
    // --- SW regs interface
    .software_regs    ({ip, ip_data_5, ip_data_4, ip_data_3, ip_data_2, ip_data_1, ip_data_0, encap_enable}), // signals from the software
 
    // --- HW regs interface
    .hardware_regs  ({num_bytes_out, num_bytes_in, num_packet_out, num_packet_in}),
    
    .clk           (clk),
    .reset         (reset));



endmodule // encap
