///////////////////////////////////////////////////////////////////////////////
// vim:set shiftwidth=3 softtabstop=3 expandtab:
// $Id: opl_processor.v 3584 2008-04-10 19:54:57Z jnaous $
//
// Module: opl_processor.v
// Project: NF2.1 OpenFlow Switch
// Author: Jad Naous <jnaous@stanford.edu>
// Description: Appends the actions to take on a packet to the beginning of
// a packet then pushes the packet out to the next module.
//              
// Licensing: In addition to the NetFPGA license, the following license applies
//            to the source code in the OpenFlow Switch implementation on NetFPGA.
//
// Copyright (c) 2008 The Board of Trustees of The Leland Stanford Junior University
//
// We are making the OpenFlow specification and associated documentation (Software)
// available for public use and benefit with the expectation that others will use, 
// modify and enhance the Software and contribute those enhancements back to the
// community. However, since we would like to make the Software available for 
// broadest use, with as few restrictions as possible permission is hereby granted, 
// free of charge, to any person obtaining a copy of this Software to deal in the 
// Software under the copyrights without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
// sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
// SOFTWARE.
//
// The name and trademarks of copyright holder(s) may NOT be used in advertising 
// or publicity pertaining to the Software or any derivatives without specific, 
// written prior permission.
///////////////////////////////////////////////////////////////////////////////
`timescale 1ns/1ps

module opl_processor
  #(parameter NUM_OUTPUT_QUEUES = 8,
    parameter DATA_WIDTH = 64,
    parameter CTRL_WIDTH = DATA_WIDTH/8,
    parameter RESULT_WIDTH = `OPENFLOW_ACTION_WIDTH + `OPENFLOW_ENTRY_SRC_PORT_WIDTH)
  (// --- interface to results fifo
   input      [RESULT_WIDTH-1:0]           result_fifo_dout,
   output reg                              result_fifo_rd_en,
   input                                   result_fifo_empty,

   // --- interface to input fifo
   input [CTRL_WIDTH-1:0]                  in_fifo_ctrl,
   input [DATA_WIDTH-1:0]                  in_fifo_data,
   output reg                              in_fifo_rd_en,
   input                                   in_fifo_empty,
   
   // --- interface to output
   output reg [DATA_WIDTH-1:0]             out_data,
   output reg [CTRL_WIDTH-1:0]             out_ctrl,
   output reg                              out_wr,
   input                                   out_rdy,
   
   // --- interface to registers
   output reg [NUM_OUTPUT_QUEUES-1:0]      pkts_dropped,

   // --- Misc
   input                                   clk,
   input                                   reset);

   `LOG2_FUNC
   `CEILDIV_FUNC

   //-------------------- Internal Parameters ------------------------
   localparam TOTAL_NUM_ACTION_WORDS = (`OPENFLOW_NUM_REPLACEMENT_ACTIONS/2
                                        + `OPENFLOW_NUM_INCREMENT_ACTIONS/2
                                        + 1);

   localparam NUM_STATES = 7;
   localparam WAIT_FOR_INPUT           = 1,
              WRITE_REPLACEMENTS       = 2,
              WRITE_INCREMENTS         = 4,
              WRITE_VLAN_MODS          = 8,
              WRITE_OUTPUT_DESTINATION = 16,
              WRITE_PACKET             = 32,
              DROP_PKT                 = 64;
   
   localparam CNTR_SIZE = 3;

   //------------------------ Wires/Regs -----------------------------
   wire [`OPENFLOW_ENTRY_SRC_PORT_WIDTH-1:0]       src_port;
   wire [`OPENFLOW_FORWARD_BITMASK_WIDTH-1:0]      forward_bitmask;
   wire [2*`OPENFLOW_REPLACEMENT_ACTION_WIDTH-1:0] replacements[`OPENFLOW_NUM_REPLACEMENT_ACTIONS/2-1:0];
   wire [2*`OPENFLOW_INCREMENT_ACTION_WIDTH-1:0]   increments  [`OPENFLOW_NUM_INCREMENT_ACTIONS/2-1:0];
   wire [`OPENFLOW_VLAN_MOD_ACTION_WIDTH-1:0]      vlan_mod;

   wire [2*`OPENFLOW_REPLACEMENT_ACTION_WIDTH-1:0] replacement;
   wire [2*`OPENFLOW_INCREMENT_ACTION_WIDTH-1:0]   increment;

   reg [NUM_STATES-1:0]                            state, state_nxt;
   reg                                             out_wr_nxt;
   reg [DATA_WIDTH-1:0]                            out_data_nxt;
   reg [CTRL_WIDTH-1:0]                            out_ctrl_nxt;
   reg [CNTR_SIZE-1:0]                             cntr, cntr_nxt;
   reg [NUM_OUTPUT_QUEUES-1:0]                     pkts_dropped_nxt;

   reg [15:0]                                      src_port_decoded;

   //-------------------------- Logic --------------------------------
   assign src_port        = result_fifo_dout[RESULT_WIDTH-1 -: `OPENFLOW_ENTRY_SRC_PORT_WIDTH];

   /* decode source port */
   always @(*) begin
      src_port_decoded = 0;
      src_port_decoded[src_port[3:0]] = 1'b1;
   end

   /* always take out the src port from the output destinations */
/* -----\/----- EXCLUDED -----\/-----
   assign forward_bitmask = (result_fifo_dout[`OPENFLOW_FORWARD_BITMASK_POS +: `OPENFLOW_FORWARD_BITMASK_WIDTH]
                             & ~src_port_decoded);
 -----/\----- EXCLUDED -----/\----- */
   assign forward_bitmask = (result_fifo_dout[`OPENFLOW_FORWARD_BITMASK_POS +: `OPENFLOW_FORWARD_BITMASK_WIDTH]);

   generate
      genvar i;
      for (i=0; i<`OPENFLOW_NUM_REPLACEMENT_ACTIONS/2; i=i+1) begin:replacements_gen
         assign replacements[i] = result_fifo_dout[`OPENFLOW_REPLACEMENT_ACTION_POS
                                                   + 2*i*`OPENFLOW_REPLACEMENT_ACTION_WIDTH +: 
                                                   2*`OPENFLOW_REPLACEMENT_ACTION_WIDTH];
      end
      for (i=0; i<`OPENFLOW_NUM_INCREMENT_ACTIONS/2; i=i+1) begin:increments_gen
         assign increments[i] = result_fifo_dout[(`OPENFLOW_INCREMENT_ACTION_POS
                                                  + 2*i*`OPENFLOW_INCREMENT_ACTION_WIDTH) +: 
                                                 2*`OPENFLOW_INCREMENT_ACTION_WIDTH];
      end
   endgenerate

   assign vlan_mod = result_fifo_dout[`OPENFLOW_VLAN_MOD_ACTION_POS +: `OPENFLOW_VLAN_MOD_ACTION_WIDTH];

   assign replacement = replacements[cntr];
   assign increment   = increments[cntr];

   /* This state machine parses the action data from the flow table,
    * then stores the actions as header modules prepended to the
    * pkt */
   always @(*) begin
      state_nxt           = state;
      result_fifo_rd_en   = 0;
      out_wr_nxt          = 0;
      out_data_nxt        = 0;
      out_ctrl_nxt        = out_ctrl;
      cntr_nxt            = cntr;
      in_fifo_rd_en       = 0;
      pkts_dropped_nxt    = 0;
      case (state)
         /* wait until the lookup is done and we have the actions we
          * need to do. */
         WAIT_FOR_INPUT: begin
            if(!result_fifo_empty) begin
               result_fifo_rd_en   = 1;
               state_nxt           = WRITE_REPLACEMENTS;
            end
         end

         /* check if an output is specified */
         WRITE_REPLACEMENTS: begin
            /* Let the packet through if the src port is set for that */
            if (src_port >= `DONT_TOUCH_SRC_PORT) begin
               state_nxt   = WRITE_PACKET;
            end
            /* no destination so drop packet */
            else if(forward_bitmask == 0) begin
               state_nxt   = DROP_PKT;
            end
            else if (out_rdy) begin
               state_nxt   = WRITE_OUTPUT_DESTINATION;
            end
/* -----\/----- EXCLUDED -----\/-----
            /-* we should forward the packet, so write out the first set of 
             * actions *-/
            else if (out_rdy) begin
               out_wr_nxt                                               = 1'b1;
               out_data_nxt[0 +: `OPENFLOW_REPLACEMENT_ACTION_WIDTH*2]  = replacement;
               out_ctrl_nxt                                             = `OPENFLOW_REPLACEMENT_ACTION_CTRL;
               if(cntr == `OPENFLOW_NUM_REPLACEMENT_ACTIONS/2-1) begin
                  state_nxt   = WRITE_INCREMENTS;
                  cntr_nxt    = 0;
               end
               else begin
                  cntr_nxt = cntr + 1'b1;
               end
            end // if (out_rdy)
 -----/\----- EXCLUDED -----/\----- */
         end // case: WRITE_REPLACEMENTS

         /* write second set of actions (increment actions) */
         WRITE_INCREMENTS: begin
            if (out_rdy) begin
               out_wr_nxt     = 1'b1;
               out_data_nxt   = {{(DATA_WIDTH-`OPENFLOW_INCREMENT_ACTION_WIDTH*2){1'b0}}, increment};
               out_ctrl_nxt   = `OPENFLOW_INCREMENT_ACTION_CTRL;
               cntr_nxt       = cntr + 1'b1;
               if(cntr == `OPENFLOW_NUM_INCREMENT_ACTIONS/2-1) begin
                  state_nxt = WRITE_VLAN_MODS;
                  cntr_nxt  = 0;
               end
            end // if (out_rdy)
         end // case: WRITE_INCREMENTS

         /* write third set of actions (vlan modification actions) */
         WRITE_VLAN_MODS: begin
            if (out_rdy) begin
               out_wr_nxt     = 1'b1;
               out_data_nxt   = {{(DATA_WIDTH-`OPENFLOW_VLAN_MOD_ACTION_WIDTH*2){1'b0}}, vlan_mod};
               out_ctrl_nxt   = `OPENFLOW_VLAN_MOD_ACTION_CTRL;
               state_nxt      = WRITE_OUTPUT_DESTINATION;
            end // if (out_rdy)
         end

         /* write out all the module headers and search for the
          * I/O queue header to insert the destination output queues */
         WRITE_OUTPUT_DESTINATION: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               out_ctrl_nxt    = in_fifo_ctrl;
               in_fifo_rd_en   = 1'b1;
               out_data_nxt    = in_fifo_data;
               if(in_fifo_ctrl == `IO_QUEUE_STAGE_NUM) begin
                  out_data_nxt[`IOQ_DST_PORT_POS+15:`IOQ_DST_PORT_POS] = forward_bitmask;
                  state_nxt = WRITE_PACKET;
               end
               // synthesis translate_off
               else if (in_fifo_ctrl==0 || in_fifo_empty) begin
                  $display ("%t %m ERROR: Could not find IOQ module header", $time);
                  $stop;
               end
               // synthesis translate_on
            end // if (out_rdy && !in_fifo_empty)
         end // case: WRITE_OUTPUT_DESTINATION

         /* write the rest of the module headers and the packet data */
         WRITE_PACKET: begin
            if(out_rdy && !in_fifo_empty) begin
               out_wr_nxt      = 1'b1;
               out_data_nxt    = in_fifo_data;
               out_ctrl_nxt    = in_fifo_ctrl;
               in_fifo_rd_en   = 1'b1;
               if(in_fifo_ctrl !=0 && out_ctrl == 0) begin // eop
                  if(!result_fifo_empty) begin
                     result_fifo_rd_en   = 1;
                     state_nxt           = WRITE_REPLACEMENTS;
                  end
                  else begin
                     state_nxt = WAIT_FOR_INPUT;
                  end
               end
            end // if (out_rdy)
         end // case: WRITE_PACKET

         /* drop the packet */
         DROP_PKT: begin
            if(!in_fifo_empty) begin
               out_ctrl_nxt    = in_fifo_ctrl;
               in_fifo_rd_en   = 1'b1;
               if(in_fifo_ctrl !=0 && out_ctrl == 0) begin // eop
                  pkts_dropped_nxt = src_port_decoded;
                  if(!result_fifo_empty) begin
                     result_fifo_rd_en   = 1;
                     state_nxt           = WRITE_REPLACEMENTS;
                  end
                  else begin
                     state_nxt = WAIT_FOR_INPUT;
                  end
               end
            end // if (!in_fifo_empty)
         end // case: DROP_PKT
         
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      if (reset) begin
         state           <= WAIT_FOR_INPUT;
         out_wr          <= 0;
         out_data        <= 0;
         out_ctrl        <= 1;
         cntr            <= 0;
         pkts_dropped    <= 0;
      end
      else begin
         state           <= state_nxt;
         out_wr          <= out_wr_nxt;
         out_data        <= out_data_nxt;
         out_ctrl        <= out_ctrl_nxt;
         cntr            <= cntr_nxt;
         pkts_dropped    <= pkts_dropped_nxt;
      end // else: !if(reset)
   end // always @ (posedge clk)
   
endmodule // opl_processor

         
