#!/bin/bash

# Script to set up an OpenFlow switch with tunneling
#
# Tunneling is enabled individually per port. If tunneling is enabled 
# for a port then all traffic will be tunneled and non-tunneled input 
# traffic will be dropped.

# File locations
OPENFLOW_DIR=../openflow
BITFILE_DIR=../../../bitfiles
HEADER_DIR=../lib/C

# Tunneling configuration
#
# For each port the following items need to be configured:
#   * Src IP: IP address to be used for the interface
#   * Dst IP: IP address of the far-end of the tunnel
#   * Src MAC: MAC address to use for the interface
#   * GW MAC: MAC address of the gateway (next-hop)
#   * Tunnel enable: enable/disable tunnel on this interface

# nf2c0 configuration
ENABLETUN0=0
SRCIP0=192.168.0.1
SRCMAC0=00:aa:bb:cc:dd:00
DSTIP0=192.168.128.1
GWMAC0=00:ca:fe:00:00:01

# nf2c1 configuration
ENABLETUN1=0
SRCIP1=192.168.0.2
SRCMAC1=00:aa:bb:cc:dd:01
DSTIP1=192.168.129.1
GWMAC1=00:ca:fe:00:00:01

# nf2c2 configuration
ENABLETUN2=0
SRCIP2=192.168.0.3
SRCMAC2=00:aa:bb:cc:dd:02
DSTIP2=192.168.130.1
GWMAC2=00:ca:fe:00:00:01

# nf2c3 configuration
ENABLETUN3=0
SRCIP3=192.168.0.4
SRCMAC3=00:aa:bb:cc:dd:03
DSTIP3=192.168.131.1
GWMAC3=00:ca:fe:00:00:01


# OpenFlow controller
OF_CONTROLLER=192.168.254.1



## useful func definitions;
dotdec2hex (){
      echo $1 | awk -F"." '{printf("0x%.2x%.2x%.2x%.2x",$1,$2,$3,$4);}'
}
macupper (){
      echo $1 | awk -F":" '{printf("0x%s%s",$1,$2);}'
}
maclower (){
      echo $1 | awk -F":" '{printf("0x%s%s%s%s",$3,$4,$5,$6);}'
}
setreg (){
	REGNAME=$1;
	val=$2;
	echo "$REGNAME=$val"
	regwrite `grep  $REGNAME $HEADER_DIR/reg_defines_tunneling_openflow_switch.h  |awk '{print $3}'`  $val;
	return 1;
}
getreg (){
	REGNAME=$1;
	echo $REGNAME;
	regread `grep  $REGNAME $HEADER_DIR/reg_defines_tunneling_openflow_switch.h  |awk '{print $3}'`;
	return 1;
}
pow (){
	VALUE=$1
	EXPONENT=$2
	return `perl -e 'print $VALUE ** $EXPONENT'`
}


for iface in 0 1 2 3 ; do
	echo $val
	
done


# Begin by downloading the bit file
nf2_download $BITFILE_DIR/tunneling_openflow_switch.bit

# Install OpenFlow kernel modules
/sbin/insmod $OPENFLOW_DIR/datapath/linux-2.6/openflow_mod.ko
/sbin/insmod $OPENFLOW_DIR/datapath/linux-2.6/hwtable_nf2_mod.ko

# Set up OpenFlow
$OPENFLOW_DIR/utilities/dpctl adddp nl:0
$OPENFLOW_DIR/utilities/dpctl addif nl:0 nf2c0
$OPENFLOW_DIR/utilities/dpctl addif nl:0 nf2c1
$OPENFLOW_DIR/utilities/dpctl addif nl:0 nf2c2
$OPENFLOW_DIR/utilities/dpctl addif nl:0 nf2c3

# Process each interface
DECAP=0
for iface in 0 1 2 3 ; do
	# Painful shell script acrobatics to get the variables 
	# associated with a single interface
	ENABLETUN=\$ENABLETUN$iface
	ENABLETUN=`eval echo \$ENABLETUN`
	SRCIP=\$SRCIP$iface
	SRCIP=`eval echo \$SRCIP`
	SRCMAC=\$SRCMAC$iface
	SRCMAC=`eval echo \$SRCMAC`
	DSTIP=\$DSTIP$iface
	DSTIP=`eval echo \$DSTIP`
	GWMAC=\$GWMAC$iface
	GWMAC=`eval echo \$GWMAC`
	IFACE=nf2c$iface


	# Internal Vars
	h_SRCIP=`dotdec2hex $SRCIP`
	h_DSTIP=`dotdec2hex $DSTIP`
	U_SRCMAC=`macupper $SRCMAC`
	L_SRCMAC=`maclower $SRCMAC`
	U_GWMAC=`macupper $GWMAC`
	L_GWMAC=`maclower $GWMAC`

	# Remove the IP address associated with the port
	/sbin/ifconfig $IFACE 0.0.0.0

	if [ "$ENABLETUN" -ne "0" ] ; then
		echo "Enabling tunneling on nf2c$iface..."

		# Set up the tunnel 1 (for nf2c1)
		setreg ARP_REPLY_ETH_ADDR_PORT_${iface}_LO_REG $L_SRCMAC
		setreg ARP_REPLY_ETH_ADDR_PORT_${iface}_HI_REG $U_SRCMAC
		setreg ARP_REPLY_IP_ADDR_PORT_${iface}_REG $h_SRCIP
		setreg ARP_REPLY_ENABLE_REG 1

		#setreg ENCAP_PORTS_PORT_2_REG 0x1f
		setreg ENCAP_PORTS_PORT_${iface}_REG 0x11
		setreg ENCAP_SRC_IP_PORT_${iface}_REG $h_SRCIP
		setreg ENCAP_DST_IP_PORT_${iface}_REG $h_DSTIP
		setreg ENCAP_TTL_PROTO_PORT_${iface}_REG 0xff52
		setreg ENCAP_TOS_PORT_${iface}_REG 0x0
		setreg ENCAP_SRC_MAC_HI_PORT_${iface}_REG $U_SRCMAC
		setreg ENCAP_SRC_MAC_LO_PORT_${iface}_REG $L_SRCMAC
		setreg ENCAP_DST_MAC_HI_PORT_${iface}_REG $U_GWMAC
		setreg ENCAP_DST_MAC_LO_PORT_${iface}_REG $L_GWMAC
		setreg ENCAP_ENABLE_PORT_${iface}_REG 1

		DECAP_INC=`perl -e "print 4 ** $iface"`
		DECAP=`expr $DECAP + $DECAP_INC`
	fi
done

# Update the decap register
setreg DECAP_ENABLE_REG $DECAP

# Run secchan
echo "Capsulator interfaces and routes ready!"
$OPENFLOW_DIR/secchan/secchan --fail=open nl:0 tcp:$OF_CONTROLLER:975
