#include "async.h"
#include "arpc.h"
#include "sha1.h"
#include "ps_iface.h"
#include "pathclient.h"
#include "icinghdr.h"
#include "tt.h"

void
common_init(IcingCHdr* chdr, IcingPktType type, 
	    uint16_t hdrlen, uint32_t totallen)
{
    chdr->icing_ver = ICING_CURRENT_VERSION;
    chdr->pkt_type = type;
    chdr->hdr_len = htons(hdrlen);
    chdr->proto = 8;
    chdr->total_len = htonl(totallen);
}

void
caplist::put(const cap_rpc_vec& path) 
{
    for (size_t i = 0; i < path.size(); i++) {
	const xdr_capmac* x = &path[i];
	const CapabilityMac* cm = reinterpret_cast<const CapabilityMac*>
				    (x->base());
	_list_suio.copy(cm, sizeof(CapabilityMac));
    }
    _len = path.size() * sizeof(CapabilityMac);
    assert(_len == _list_suio.resid());
}

int
caplist::copyout(void* dest, ssize_t size) const 
{
    if (size >= 0 && static_cast<size_t>(size) < _len)
	return -1;
    _list_suio.copyout(dest, _len);
    return _len;
}

icing_packet::icing_packet(size_t ncaps, 
			   size_t payloadlen, 
			   const caplist* cl) :
  packet(sizeof(IcingSlowHdr) + ncaps*sizeof(CapabilityMac) + 
	 ncaps*sizeof(LogEntry) + payloadlen),
  _ncaps(ncaps),
  _hdrlen(sizeof(IcingSlowHdr) + _ncaps*sizeof(CapabilityMac) +
	  _ncaps*sizeof(LogEntry))
{
   init_hdr(_ncaps, payloadlen, cl);
}

void
icing_packet::init_hdr(size_t ncaps, size_t payloadlen, const caplist* cl) 
{
    size_t totallen = _hdrlen + payloadlen;

    IcingSlowHdr* ice_slow_hdr = icing_slow_hdr();

    common_init(&ice_slow_hdr->chdr, ICING_TYPE_SLOW, _hdrlen, totallen);
    ice_slow_hdr->num_caps = htonl(ncaps);
    ice_slow_hdr->curr_cap = htonl(0);

    if (cl != NULL)
	set_caplist(*cl);

    init_log();
}

void 
icing_packet::set_caplist(const caplist& cl) 
{
    CapList* caplist = get_caplist();
    assert(cl.len() == _ncaps*sizeof(CapabilityMac));
    cl.copyout(&caplist->cap_list);
}

void
icing_packet::init_log()
{
    Log* log = icing_log();
    memset(&log->entry_list, 0, _ncaps*sizeof(LogEntry));
}

icing_flow_packet::icing_flow_packet(const IcingSlowHdr* ice_hdr,
				     size_t payloadlen)
: packet(sizeof(IcingFlowHdr) + payloadlen),
  _hdrlen(sizeof(IcingFlowHdr))
{
   init_hdr(ice_hdr, payloadlen);
}

void
icing_flow_packet::init_hdr(const IcingSlowHdr* hdr, size_t payloadlen)
{
    size_t totallen = _hdrlen + payloadlen;

    IcingFlowHdr* flow_hdr = icing_flow_hdr();
    common_init(&flow_hdr->chdr, ICING_TYPE_FLOW, _hdrlen, totallen);
    construct_flowid(hdr, &flow_hdr->flowid);
}

void 
reachability::set_caplist(const cap_rpc_vec& path) 
{ 
    _cl.put(path); 
}

void 
reachability::set_ncaps(const cap_rpc_vec& path) 
{ 
    _ncaps = path.size(); 
}

void
reachability::set_expiry(const cap_rpc_vec& path)
{
    // the expiry of the path is the earliest of the elements in the path
    const CapabilityMac* cm = 
	    reinterpret_cast<const CapabilityMac*>(path[0].base());
    _expiry = icingts2us(cm->cap.expiry);

    for (size_t i = 1; i < path.size(); i++) {
	cm = reinterpret_cast<const CapabilityMac*>(path[i].base());
	uint64_t candidate = icingts2us(cm->cap.expiry);
	if (candidate < _expiry) 
	    _expiry = candidate;
    }
}

void
byte_print(const precord_ent& entry)
{
    strbuf sb;
    size_t nnodes = entry.path_from_waypoint.size();

    for (size_t i = 0; i < nnodes; i++) {

        const CapabilityMac* cm = 
	    reinterpret_cast<const CapabilityMac*>
		    (entry.path_from_waypoint[i].base());
	const uint8_t* cp = reinterpret_cast<const uint8_t*>(cm);

        for (size_t j = 0; j < sizeof(CapabilityMac); j++) 
	    sb.fmt(" %02x", cp[j] & 0xff);
    }

    sb << "\n";

    int len_caplist = sizeof(CapabilityMac) * nnodes;

    warnx("capability list has length %d bytes. capability is: \n", len_caplist);
    warnx << sb;
}

uint64_t icingts2us(const TimestampWire& ts)
{
    return (ntohl(ts.sec) * MILLION_LL  + ntohl(ts.usec));
}

