#ifndef _PATHCLIENT_H_
#define _PATHCLIENT_H_

#include "suio++.h"
#include "str.h"
#include "icinghdr.h"

typedef rpc_vec<xdr_capmac, RPC_INFINITY> cap_rpc_vec;

// Store the list of capabilities, for flattening.
struct caplist {

    public:
	caplist() {}
	caplist(cap_rpc_vec& path) { put(path); }
	void put(const cap_rpc_vec&);
	int copyout(void* dest, ssize_t size = -1) const; 
	str tostr() const { return str(_list_suio); }
	size_t len() const { return _len; }

    private:
	//no copying
	caplist(const caplist&);
	caplist& operator=(const caplist&);

	suio _list_suio;
	size_t _len;
};

class packet {

    public:
	packet(size_t nbytes) : data(new uint8_t[nbytes]), len(nbytes) {}
	virtual ~packet() { delete[] data; }
	size_t length() const { return len; }
	const uint8_t* const_dat() const { return data; }
	uint8_t* dat() const { return data; }

    private:
	// no copying
	packet(const packet&);
	packet& operator=(const packet&);

	uint8_t* data;
	size_t   len;
};

class icing_packet : public packet {

    public:
	icing_packet(size_t ncaps, size_t payloadlen, const caplist* cl = NULL);
	virtual ~icing_packet() {}
	void set_caplist(const caplist&); 
	IcingSlowHdr* icing_slow_hdr() const; 
	const IcingSlowHdr* const_icing_slow_hdr() const;
	Log* icing_log() const;
	const Log* const_icing_log() const;
	const uint8_t* const_icing_payload() const { return dat() + _hdrlen; }
	uint8_t* icing_payload() const { return dat() + _hdrlen; }
	const CapList* const_get_caplist() const;
	CapList* get_caplist() const;
    private:
	void init_hdr(size_t ncaps, size_t payloadlen, const caplist*); 
	void init_log();

	// no copying
	icing_packet(const icing_packet&);
	icing_packet& operator=(const icing_packet&);

	size_t _ncaps;
	size_t _hdrlen;
};

class icing_flow_packet : public packet {

    public:
	icing_flow_packet(const IcingSlowHdr*, size_t payloadlen);
	virtual ~icing_flow_packet() {}
	const IcingFlowHdr* const_icing_flow_hdr() const;
	IcingFlowHdr* icing_flow_hdr() const;
	const uint8_t* const_icing_payload() const { return dat() + _hdrlen;}
	uint8_t* icing_payload() const { return dat() + _hdrlen;}
	    
    private:
	void init_hdr(const IcingSlowHdr*, size_t payloadlen);

	// no copying
	icing_flow_packet(const icing_flow_packet&);
	icing_flow_packet& operator=(const icing_flow_packet&);

	size_t _hdrlen;
};

struct reachability {

    void set_waypoint_name(const str& s) { _waypoint_name = s; }
    void set_waypoint_port(uint32_t p) { 
	_waypoint_port = htons(static_cast<uint16_t>(ntohl(p)));
    }
    void set_waypoint_port(uint16_t p) { _waypoint_port = p; }
    void set_waypoint_ip(uint32_t a) { _waypoint_ip = a; }
    void set_caplist(const cap_rpc_vec&); 
    void set_ncaps(const cap_rpc_vec&); 
    void set_expiry(const cap_rpc_vec&); 

    str _waypoint_name;
    uint16_t _waypoint_port;
    uint32_t _waypoint_ip;
    caplist _cl;
    size_t _ncaps;

    uint64_t _expiry;

    reachability() : _waypoint_name(""), _waypoint_port(0), 
		      _waypoint_ip(0), _ncaps(0) {}

    private:
	// no copying
	reachability(const reachability&);
	reachability& operator=(const reachability&);
};

class precord_ent;
class click_timestamp;

// helper functions
void byte_print(const precord_ent&);
uint64_t icingts2us(const TimestampWire&);
void common_init(IcingCHdr* chdr, IcingPktType type, 
		 uint16_t hdrlen, uint32_t totallen);

inline IcingSlowHdr*
icing_packet::icing_slow_hdr() const
{ 
    return reinterpret_cast<IcingSlowHdr*>(dat()); 
}

inline const IcingSlowHdr* 
icing_packet::const_icing_slow_hdr() const
{ 
    return reinterpret_cast<const IcingSlowHdr*>(const_dat()); 
}

inline const CapList*
icing_packet::const_get_caplist() const
{
    return ::get_caplist(const_icing_slow_hdr());
}

inline CapList*
icing_packet::get_caplist() const
{
    return ::get_caplist(icing_slow_hdr());
}

inline Log*
icing_packet::icing_log() const
{
    return get_log(icing_slow_hdr(), _ncaps);
}

inline const Log*
icing_packet::const_icing_log() const
{
    return get_log(const_icing_slow_hdr(), _ncaps);
}

inline IcingFlowHdr*
icing_flow_packet::icing_flow_hdr() const
{
    return reinterpret_cast<IcingFlowHdr*>(dat());
}

inline const IcingFlowHdr* 
icing_flow_packet::const_icing_flow_hdr() const
{
    return reinterpret_cast<const IcingFlowHdr*>(const_dat());
}

#endif
