// $Id: tt.h,v 1.8 2006/03/17 15:51:21 mwalfish Exp $
//   Implements timing utilities. Copied 
//   from net-work/src/thinner/okws/websrv and 
//   from dqm/src/storer
#ifndef _TT_H_
#define _TT_H_

#include <sys/time.h>

#define BILLION       1000000000
#define BILLION_LL    1000000000LL
#define BILLION_D     1000000000.

#define MILLION          1000000
#define MILLION_LL       1000000LL
#define MILLION_D        1000000.


/* us is the number of microseconds since the epoch. */
inline void us2tv(u_int64_t us, struct timeval* tv)
{
    tv->tv_sec  = us / MILLION;
    tv->tv_usec = us - tv->tv_sec * MILLION_LL;
}

inline void us2tv(u_int32_t us, struct timeval* tv)
{
    tv->tv_sec  = us / MILLION;
    tv->tv_usec = us % MILLION;
}

inline uint64_t tv2us(const struct timeval& tv)
{
    return tv.tv_sec*MILLION_LL + tv.tv_usec;
}

inline uint64_t ts2us(const struct timespec& ts)
{
    return ts.tv_sec*MILLION_LL + ts.tv_nsec / 1000;
}

inline uint64_t get_now()
{
#if 0
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv2us(tv);
#else
    struct timespec ts;
    sfs_get_tsnow(&ts);
    return ts2us(ts);
#endif
}

inline timespec 
get_now_ts()
{
    struct timespec ts;
    sfs_get_tsnow(&ts);
    return ts;
}

// Copying DM's logic for subtracting timespecs. the only tricky part is
// that a and b are references, which means we can't twiddle their
// fields.
inline struct timeval
operator-(const struct timeval& a, const struct timeval& b)
{
    struct timeval ret;

    ret.tv_sec  = a.tv_sec - b.tv_sec;

    if (a.tv_usec < b.tv_usec) {
	ret.tv_sec--;
	ret.tv_usec = a.tv_usec + MILLION - b.tv_usec;
    } else {
	ret.tv_usec = a.tv_usec - b.tv_usec;
    }

    return ret;
}

// keep track of the avg. amount of time to do a particular task.
// callers should call begin() and end() before and after the task,
// respectively. avg() gives the average amount of time the task has been
// taking.
class time_it {

    public:
	time_it () : total(0), cntr(0), curr_min(0) {}
	~time_it() {}
	
	void begin()   {
	    time_at_begin = get_now();
	}
	void end() {
	    u_int64_t delta = get_now() - time_at_begin;
	    total += delta;
	    if (cntr == 0 || delta < curr_min) 
		curr_min = delta;
	    ++cntr;
	}
	double avg_us() {
	    if (cntr == 0) return -1.;
	    return (double)total / cntr;
	}

	double min_us() {
	    if (cntr == 0) return -1.;
	    return curr_min;
	}

    private:
	u_int64_t total;	
	u_int32_t cntr;
	u_int64_t time_at_begin;
	u_int64_t curr_min;
};

// keep track of the rate at which a counter is being incremented
class tt {

    public:
	tt() : last_time(get_now()), last_counter(0), curr_counter(0) {}
	~tt() {}

	u_int32_t counter()     { return curr_counter; }
	void inc()              { ++curr_counter; }
	void inc_by(uint32_t n) { curr_counter += n; }

	double num_per_sec() {

	    u_int64_t curr_time = get_now();

	    // calculate the gaps
	    u_int32_t num_in_interval = curr_counter - last_counter;
	    u_int64_t delta           = curr_time - last_time;
	   
	    // update our internal state
	    last_time    = curr_time;
	    last_counter = curr_counter;

	    // and do the calculation
	    double rate = (num_in_interval * 1000000.) / delta;

	    return rate;
	}

    private:

	u_int64_t last_time;  // in usec
	u_int64_t last_counter;
	u_int64_t curr_counter;

};

class avg_it {

    public:
	avg_it() : cntr(0), total(0), cntr_intvl(0), total_intvl(0) {}
	~avg_it() {}

	void add(uint32_t n) { 
	    total       += n; 
	    total_intvl += n;
	    ++cntr; 
	    ++cntr_intvl;
	}

	double avg() { 
	    return static_cast<double>(total) / cntr; 
	}

	double avg_intvl() {
	    double ret = static_cast<double>(total_intvl) / cntr_intvl;
	    total_intvl = cntr_intvl = 0;
	    return ret;
	}


    private:
	u_int64_t cntr;
	u_int64_t total;
	u_int64_t cntr_intvl;
	u_int64_t total_intvl;
};

#endif
