/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/**********************************
 * Hardwiring B-163               *
 *                                *
 * Irreducible polynomial:        *
 *                                *
 *   x^163 + x^7 + x^6 + x^3 + 1  *
 *                                *
 **********************************/

#ifndef _B163_H_
#define _B163_H_

/* XXX: should link against sfslite; use dcrypt for now */
#ifndef HAVE_SFS
#include "dcrypt.h"             
	
/* needed for mpz_sizeinbase2, random_bigint, random_zn, (not part of gmp) */
#include "dcinternal.h"

/* dummy typedef and macros to give illusion of bigints without sfslite */
typedef MP_INT bigint;
#define SFS_RANDOM_BIGINT(bint,bits)   random_bigint(&(bint),bits)
#define SFS_RANDOM_ZN(bint,mod)        random_zn(&(bint),mod)
#define SFS_SHA1HASHSIZE               sha1_hashsize
#define SFS_SHA1CTX(s)                 sha1_ctx s; sha1_init (&(s))
#define SFS_SHA1UPDATE(s,b,l)          sha1_update(&(s),b,l)
#define SFS_SHA1FINAL(s,b)             sha1_final(&(s),b)
#else  /* HAVE_SFS */

#define SFS_RANDOM_BIGINT(bint,bits)   bint = random_bigint(bits)
#define SFS_RANDOM_ZN(bint,mod)        bint = random_zn(mod)
#define SFS_SHA1HASHSIZE               sha1::hashsize
#define SFS_SHA1CTX(s)                 sha1ctx s
#define SFS_SHA1UPDATE(s,b,l)          s.update(b,l)
#define SFS_SHA1FINAL(s,b)             s.final(b)

#endif /* HAVE_SFS */

#include <assert.h>
#include <stdio.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif
  
#include "gf2m.h"
  
/* Store max degree by itself   */
#define B163_MAX_DEGREE              163

/* Store low terms (x^7+x^6+x^3+1) by itself */
#define B163_MOD_MOD                  0x000000c9   /* "modified modulo" */

/* Max u_int32_t array size necessary for holding B-163 elements */
#define B163_MAX_LEN                   6
/* Max u_int32_t array size necessary for double precision */
#define B163_MAX_LEN_DP               11

  typedef u_int32_t b163_elem[B163_MAX_LEN]; 

/* Double precision (so can do modular reduction after multiplication) */
typedef u_int32_t b163_elem_dp[B163_MAX_LEN_DP];

#define B163_IS_ZERO(sp)     (!((sp)[5] & 0x00000007) \
			      && !((sp)[4]) && !((sp)[3]) && !((sp)[2]) \
			      && !((sp)[1]) && !((sp)[0]))

#define B163_IS_ZERO_DP(dp)  (B163_IS_ZERO(dp) \
			      && !((dp)[5] & 0xfffffff8) \
			      && !((dp)[6]) && !((dp)[7]) && !((dp)[8]) \
			      && !((dp)[9]) && !((dp)[10]))

#define B163_ARE_EQUAL(sp1,sp2)   ((((sp1)[5] & 0x7) == ((sp2)[5] & 0x7)) \
				   && ((sp1)[4] == (sp2)[4]) \
				   && ((sp1)[3] == (sp2)[3]) \
				   && ((sp1)[2] == (sp2)[2]) \
				   && ((sp1)[1] == (sp2)[1]) \
				   && ((sp1)[0] == (sp2)[0]))

/* Test if double-precision value needs modular reduction */
#define B163_NEEDS_MOD(dp)     (((dp)[5] & 0xfffffff8) \
				|| ((dp)[6]) || ((dp)[7]) || ((dp)[8]) \
				|| ((dp)[9]) || ((dp)[10]))

/* Test the lsb, ie, return the constant term (0 or 1) of the polynomial */
#define B163_CONSTANT_TERM(sp) ((sp)[0] & 0x00000001)

void b163_clear_constant_term (b163_elem sp);

#define B163_DEGREE(sp) \
  (((sp)[5]) ? 160 + gf4g_degree((sp)[5]) : \
   ((sp)[4]) ? 128 + gf4g_degree((sp)[4]) : \
   ((sp)[3]) ?  96 + gf4g_degree((sp)[3]) : \
   ((sp)[2]) ?  64 + gf4g_degree((sp)[2]) : \
   ((sp)[1]) ?  32 + gf4g_degree((sp)[1]) : \
   gf4g_degree((sp)[0]))

/* Trace computation is especially easy for B163: just xor bits 0 and 157 */
#define B163_TRACE(sp)      ((((sp)[4] & 0x20000000) >> 29) \
			     ^((sp)[0] & 0x00000001))

void b163_set_to_zero (b163_elem sp);

void b163_set_to_zero_dp (b163_elem_dp dp);

void b163_copy (b163_elem sp1, b163_elem sp2);

void  b163_copy_dp (b163_elem_dp dp1, b163_elem_dp dp2);
void b163_set_to_one (b163_elem sp);

void b163_set_to_one_dp (b163_elem_dp dp);

void b163_invert (b163_elem inv, b163_elem sp);

/* Splits dp as sp1<<163 + sp2 */
void b163_get_sp_pair (b163_elem sp1, b163_elem sp2, b163_elem_dp dp);
  
void b163_dp_from_sp (b163_elem_dp dp, b163_elem sp);

void b163_sp_from_dp (b163_elem sp, b163_elem_dp dp);

void b163_add_into (b163_elem sp1, b163_elem sp2);

void b163_add_into_dp (b163_elem_dp dp1, b163_elem_dp dp2);

void b163_add4g_into (b163_elem sp,u_int32_t v);

void b163_set_to_mod (b163_elem sp);

void b163_cat (b163_elem sp);

void b163_cat_dp (b163_elem_dp dp);

/* Polynomial division with reminder for B163 (no mod. red.) */
/* Note:   Destroys pre-existing values of Q, R              */
void b163_div_rem (b163_elem Q, b163_elem R, b163_elem U, b163_elem V);

/* Sets sp to sp1/sp2 modulo (x^163+x^7+x^6+x^3+1) 
 * Note: overwrites pre-existing value of sp     
 */
void b163_div(b163_elem sp, b163_elem sp1, b163_elem sp2);

/* Sets sp to sp1*sp2 modulo (x^163+x^7+x^6+x^3+1) 
 * Note: overwrites pre-existing value of sp     
 */
void b163_mul(b163_elem sp, b163_elem sp1, b163_elem sp2);


/* Sets sp to sp1 squared modulo (x^163+x^7+x^6+x^3+1) 
 * Note: overwrites pre-existing value of sp
 */
/* Note: defined in terms of b163_mul as it did not seem to affect
         the running time of the major crypto functions much 
*/
#define b163_square(sp,sp1)    b163_mul(sp, sp1, sp1)
  
void b163_sqrt(b163_elem sp1, b163_elem sp2);

void b163_halftrace(b163_elem sp1, b163_elem sp2);

/* relies on dcrypt.h for randomness */
void b163_set_random (b163_elem sp);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* !_B163_H_ */
