// -*-c++-*-
/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* 
 * Sources
 * 
 *   [list sources here]
 */

#ifndef _ECDH_H_
#define _ECDH_H_

#include <assert.h>

#include "ecc.h"

class ec_ephem_key_pair {
protected:
  bigint *k;
  b163_point *R; 

public:
  bigint *private_half() { return k; }
  b163_point *public_half() { return R; }

  ec_ephem_key_pair (bigint &kk, b163_point &RR) 
    : k (new bigint ()), R (new b163_point (RR)) { mpz_set(k, &kk); }

  ~ec_ephem_key_pair () { mpz_clear (k); delete k; delete R; }
};


/* forward declaration needed for friend definition */
class ecdh_priv;

class ecdh_pub {
protected:
  b163_point *P;
  // underlying curve stored as static data member b163_curve in the 
  // class b163_point, and is also accessible from ecdh_pub via
  // convenience static member function curve()

  bool is_group_elem (b163_point *Q)
  { return Q && Q->is_on_curve (); }
  
  static void random_group_log (bigint &log) 
  { SFS_RANDOM_BIGINT(log, b163_point::b163_curve->m); }

  static void elem_from_log (b163_point &elem, bigint &log) 
  { b163_point::mul_s (elem, *b163_point::b163_curve->base_point, log); }

  static b163_curve_s *curve () { return b163_point::b163_curve;  }
  static bigint &order () { return curve()->order; }
  b163_point *generator () { return curve()->base_point; }
public:
  ecdh_pub (chopped_ecpk_raw raw) 
    : P (b163_point::make_point_from_chopped(raw)) {}
  ecdh_pub (b163_point &PP) : P (new b163_point (PP)) {}
  ecdh_pub (ecdh_pub &other) : P (new b163_point (*other.P)) {}
  ecdh_pub (bigint &k) { P = new b163_point(); elem_from_log (*P, k); }
  ecdh_pub () {
    bigint k;
    mpz_init (&k);
    random_group_log (k);
    P = new b163_point();
    elem_from_log (*P, k);
    mpz_clear(&k);
  }
  virtual ~ecdh_pub () { if (P) delete P; } 

  bool operator== (const ecdh_pub &other) const { 
    return *P == *other.P;
  }
  
  static bool to_chopped_ecpk_raw (chopped_ecpk_raw raw, ecdh_pub &pub) {
    return pub.P->chop_into (raw);
  }
  
  bool to_chopped_ecpk_raw (chopped_ecpk_raw raw) {
    return to_chopped_ecpk_raw(raw, *this);
  }

  bool verify (u_char *msg, size_t len, bigint &e, bigint &s);

  static ec_ephem_key_pair *make_ec_ephem_key_pair () {
    bigint k;
    mpz_init (&k);
    random_group_log (k);
    b163_point elem;
    elem_from_log (elem, k);
    
    ec_ephem_key_pair *res = new ec_ephem_key_pair (k, elem);
    mpz_clear (&k);
    return res;
  }
  friend b163_point *ec_nidh_premaster (ecdh_priv *priv, 
						      ecdh_pub *pub);

  friend bool ec_point_label_kdf(u_char *buf, size_t len, b163_point *P,
					 u_char *label, size_t llen);

  friend bool ec_nidh (u_char *buf, size_t len, 
				     ecdh_priv *priv, ecdh_pub *pub);

};

class ecdh_priv : public ecdh_pub {
protected:
  bigint x;
  ec_ephem_key_pair *ekp;

  void make_ekp  ()
  {
    if (ekp) return;
    ekp = make_ec_ephem_key_pair ();
  }

public:
  ecdh_priv (bigint &xx) 
    : ecdh_pub (xx), x (xx), ekp (make_ec_ephem_key_pair ()) { }

  ecdh_priv () : ecdh_pub (), ekp (make_ec_ephem_key_pair ()) {
    mpz_init (&x);
    P = new b163_point ();
    /* generate only keys that are choppable to exactly 20 bytes */
    do {
      random_group_log (x);
      elem_from_log (*P, x);
    } while (!P->choppable ());
  }

  bool operator== (const ecdh_priv &other) const { 
    return (0 == mpz_cmp (&x, &other.x));
  }

  ~ecdh_priv () { mpz_clear (&x); if (ekp) delete ekp; }

  ecdh_pub *public_key ()  { return new ecdh_pub (*this); }

  bool sign (bigint &e, bigint &s, u_char *msg, size_t len);

  friend b163_point *ec_nidh_premaster (ecdh_priv *priv, 
						      ecdh_pub *pub);

  friend bool ec_point_label_kdf (u_char *buf, size_t len, b163_point *P,
					 u_char *label, size_t llen);

  friend bool ec_nidh (u_char *buf, size_t len, 
  				     ecdh_priv *priv, ecdh_pub *pub);
};

extern const char *ecdh_suffix_str;

b163_point *ec_nidh_premaster (ecdh_priv *priv, ecdh_pub *pub);

bool ec_point_label_kdf (u_char *buf, size_t len, b163_point *P,
			 u_char *label=NULL, size_t llen=0);

bool ec_nidh (u_char *buf, size_t len, ecdh_priv *priv, ecdh_pub *pub);


#endif /* !_ECDH_H_ */
