/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/* 
 * Sources
 * 
 *   [list sources here]
 */

#ifndef _GF2M_H_
#define _GF2M_H_

/* Used only for assert and sys/type */
#include <dcrypt.h>

#ifdef __cplusplus
extern "C" {
#endif
  
/* Multiplication by monomials (i.e., left shifts)  */
/* Special case: GF_2^32                            */
/* Input:  32-bit O, A, size_t k (0<=k<=32)         */
/* Output: (O||0) ^ (A << k), stored in (O, A)      */
/*                                                  */
/* Note: Zero out O before this to get A << k       */
void gf64k_shl (u_int32_t *OO, u_int32_t *AA, size_t k);

/* Check that a 32-byte array has got enough space for a 
   left shift to go through without bits falling on the floor */
int gf_can_do_shift(size_t U_len, u_int32_t U[], size_t k);

/* Multiplication by monomials (i.e., left shifts)  */
/* Input:  32-bit array U, size_t k >=0             */ 
/* Output: (U << k), stored in U                    */
/*                                                  */
/* Note: Aborts if space is not enough              */
void gf_shl (size_t U_len, u_int32_t U[], size_t k);

/* Polynomial multiplication                              */
/* Special case: (U||A)*B                                 */
/* Input:  32-bit array U; 32-bit A, B                    */
/* Output  32-bit array W; 32-bit Z, Y (Z high, Y low)    */
/*                                                        */
/*              (W||Z||Y) ^= (U||A)*B       (note: ^=)    */
/*                                                        */
void gf_mul_by4g (u_int32_t W[], u_int32_t *ZZ, u_int32_t *YY, 
		  size_t U_len, u_int32_t U[], u_int32_t A, u_int32_t B);

/* Polynomial multiplication                              */
/* General case: (U||A)*(V||B), U as long as V or longer  */
/* Input:  32-bit arrays U, V; 32-bit A, B                */
/* Output  32-bit array S; 32-bit T, Y (T high, Y low)    */
/*                                                        */
/*              (S||T||Y) ^= (U||A)*(V||B)  (note: ^=)    */
/*                                                        */
void gf_mul (u_int32_t S[], u_int32_t *TT, u_int32_t *YY, 
	     size_t U_len, u_int32_t U[], u_int32_t A, 
	     size_t V_len, u_int32_t V[], u_int32_t B);

/* Convenience method for computing S=U*V */
void gf_multiply (u_int32_t S[], size_t U_len, u_int32_t U[],
		  size_t V_len, u_int32_t V[]);

/* degree of a polynomial in GF_2m */
u_int16_t gf_degree (size_t U_len, u_int32_t U[]);

/* degree of a polynomial in GF_2^32 */
u_int16_t gf4g_degree (u_int32_t);

/* degree of a polynomial in GF_256, ie, index of leftmost 1-bit in a byte */
#define GF256_DEGREE(b) \
  (((b) & 0x80) ? 7 : \
   ((b) & 0x40) ? 6 : \
   ((b) & 0x20) ? 5 : \
   ((b) & 0x10) ? 4 : \
   ((b) & 0x08) ? 3 : \
   ((b) & 0x04) ? 2 : \
   ((b) & 0x02) ? 1 : \
   0)

#define GF2M_IS_ZERO(L, U)   ((0 == (L)) \
			       || ((0 == (U)[0]) && (0 == gf2m_degree (L, U))))

#define GF2M_IS_ZERO(L, U)   ((0 == (L)) \
			       || ((0 == (U)[0]) && (0 == gf2m_degree (L, U))))

#define GF2M_ZERO(L, U)       bzero(U, sizeof(u_int32_t)*(L))
#define GF2M_COPY(L, U, V)    memcpy(U, V, sizeof(u_int32_t)*(L))


#ifdef __cplusplus
}
#endif

#endif /* !_GF2M_H_ */
