/* $Id: $ */

/*
 *
 * Copyright (C) 2009 Antonio Nicolosi (nicolosi@cs.stevens.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "ecdh.h"

#include <sys/time.h>

int _ts;

timeval tv;

inline void startt () { gettimeofday (&tv, NULL); }
inline int stopt () 
{ 
  timeval tv2; 
  gettimeofday (&tv2, NULL); 
  return ( (tv2.tv_sec - tv.tv_sec) * 1000000 +
	   (tv2.tv_usec - tv.tv_usec)) ;
}

void
test_ecc_1 () {
  b163_point O, P;

  /* Let P be the curve generator */
  P.set_to (*b163_point::b163_curve->base_point);


  /* Verifying that P satisfies the curve equation */
  printf("Checking that curve generator satisfies curve eq: %s",
	 P.is_on_curve() ? "OK\n" : "XXX---Something is wrong\n");

  /* Let Q be P times the curve order */
  b163_point::mul_s (O, P, b163_point::b163_curve->order);

  /* Result shoud be the point at infinity */
  printf("Verifying scalar multiplication by group order: %s",
	 O.is_infinity() ? "OK\n" : "XXX---Something is wrong\n");

  /* Verifying that O satisfies the curve equation */
  printf("Checking that point at infinity lies on the curve: %s",
	 O.is_on_curve() ? "OK\n" : "XXX---Something is wrong\n");

}

/* Test the y-compressed-coordinate constructor */
void
test_ecc_2a () {
  mpz_t r0;   
  mpz_init (r0); 
  
  b163_point G, Y0, O;
  
  /* Let G be the curve generator */
  G.set_to (*b163_point::b163_curve->base_point);

  int i, count=0;

  for (i=0; i < 10; i++) {

    /* Let Y0 be a point of random order */  
    SFS_RANDOM_BIGINT(*r0, b163_point::b163_curve->m);
  
    b163_point::mul_s (Y0, G, *r0);
    
    /* Now construct W0 from Y0 via y-coord compression */
    b163_elem x0, y0, z0;
    Y0.clone_x(x0);
    Y0.clone_y(y0);
    
    /* y-bit is the lsb of z=y/x, unless x == 0, but then it doesn't matter  */
    if (!B163_IS_ZERO (x0)) {
      b163_div (z0, y0, x0);
    } else {
      b163_set_to_zero (z0);
    }
    
    b163_point *W0 = b163_point::make_point (x0, B163_CONSTANT_TERM(z0));
    assert (W0);
    b163_point::set_to_neg_p (O, Y0);
    O.plus_eq (*W0);

    delete W0;

    /* Result shoud be the point at infinity */
    
    if (O.is_infinity()) {
      printf("+");
    } else {
      count++;
      printf("-");
    }
  }  

  printf("Testing constructor with y-compression: %d/%d\n",
	 count, 10);
  mpz_clear (r0);
}

/* Test constructor w/ both coordinate compression */
void
test_ecc_2b () {
  mpz_t r0;   
  mpz_init (r0); 
  
  b163_point G, Y0, O;
  
  /* Let G be the curve generator */
  G.set_to (*b163_point::b163_curve->base_point);

  int i, count=0;

  for (i=0; i < 10; i++) {

    /* Let Y0 be a point of random order */  
    SFS_RANDOM_BIGINT(*r0, b163_point::b163_curve->m);
  
    b163_point::mul_s (Y0, G, *r0);
    
    /* Now construct W0 from Y0 via compression of both coordinates */
    b163_elem x0, y0, z0;
    Y0.clone_x(x0);
    Y0.clone_y(y0);
    
    /* y-bit is the lsb of z=y/x, unless x == 0, but then it doesn't matter  */
    if (!B163_IS_ZERO (x0)) {
      b163_div (z0, y0, x0);
    } else {
      b163_set_to_zero (z0);
    }
    
    /* store the lsb of z0 in x0 */
    b163_clear_constant_term (x0);
    b163_add4g_into (x0, B163_CONSTANT_TERM(z0));
    b163_point *W0 = b163_point::make_point (x0);
    assert(W0);
    
    b163_point::set_to_neg_p (O, Y0);
    O.plus_eq (*W0);
    delete W0;

    if (O.is_infinity()) {
      printf("+");
    } else {
      count++;
      printf("-");
    }
  }
  
  printf("Testing constructor with compression of both coordinate: %d/%d\n",
	 count, 10);

    mpz_clear(r0);

}


/* Verify the Diffie-Hellman relation */
void
test_ecc_3 () {
  mpz_t r0, r1; mpz_init (r0); mpz_init (r1);

  b163_point G, Y0, Y1, W0, W1, O;

  /* Let G be the curve generator */
  G.set_to (*b163_point::b163_curve->base_point);

  /* Let Y0 and Y1 be points of (known) random order */

  SFS_RANDOM_BIGINT(*r0, b163_point::b163_curve->m);
  SFS_RANDOM_BIGINT(*r1, b163_point::b163_curve->m);

  b163_point::mul_s (Y0, G, *r0);
  b163_point::mul_s (Y1, G, *r1);

  printf ("First point:\n");
  Y0.cat();

  printf ("\nSecond point:\n");
  Y1.cat();

  b163_point::mul_s (W0, Y1, *r0);
  b163_point::mul_s (W1, Y0, *r1);

  printf ("\nFirst DH value:\n");
  W0.cat();

  printf ("\nSecond DH value:\n");
  W1.cat();

  b163_point::set_to_neg_p (O, W0);
  O.plus_eq (W1);

  /* Result shoud be the point at infinity */
  printf("Verifying Diffie-Hellman relation: %s",
	 O.is_infinity() ? "OK\n" : "XXX---Something is wrong\n");

  mpz_clear (r0);     mpz_clear (r1); 
}

/* Verify the Diffie-Hellman relation, via ICING-visible functions */
void
test_ecc_4 () {
  u_char    my_key[128];
  u_char other_key[128];

  /* pick two random private keys */
  int running_timer = 0;

  for (int i=0; i<5; i++) {
    ecdh_priv    me;
    ecdh_priv other;
    ecdh_pub    me_pub (   me);
    ecdh_pub other_pub (other);

    
    startt ();
    ec_nidh (   my_key, 128,    &me, &other_pub);
    ec_nidh (other_key, 128, &other,    &me_pub);
    running_timer += stopt ();

    /* Result shoud be the point at infinity */
    printf("Verifying Diffie-Hellman relation (ICING-visible functions): %s",
	   !memcmp (my_key, other_key, 128) ? 
	   "OK\n" : "XXX---Something is wrong\n");
  }

  printf ("Average time (out of %d trials): %f\n", 
	  2*5, float (running_timer)/ (2*5));
}

/* Exercise EC signatures */
void
test_ecc_5 () {
  char icing_msg[] = "00\
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules \
ICING rules; ICING rules; ICING rules; ICING rules" ;

  int i, count=0;

  int sign_time=0, ver_time=0;

  bool check;

  mpz_t e, s;     mpz_init (e);     mpz_init (s);

  for (i=0; i < 10; i++) {
    /* pick a random private key */
    ecdh_priv        me;
    ecdh_pub me_pub (me);
    
    /* vary the msg at each iteration */
    icing_msg[0] = ('0' + (i / 10));
    icing_msg[1] = ('0' + (i % 10));

    startt();
    me.sign (*e, *s, (u_char *) icing_msg, strlen ( icing_msg));
    sign_time += stopt();

    startt();
    check = me_pub.verify ((u_char *) icing_msg, strlen (icing_msg), *e, *s);
    ver_time += stopt();

    if (check)      printf("+");
    else {
      count++;
      printf("-");
    }
  }

  mpz_clear (e);
  mpz_clear (s);

  printf("Testing signature verification: %d/%d\n",
	 count, 10);

  printf ("Average time (out of %d trials): %f (sign), %f (ver)\n", 
	  10, float (sign_time)/ 10, float (ver_time)/ 10);
  
}

/* Exercise Chopped PKs */
void
test_ecc_6 () {
  int i, count=0;

  for (i=0; i < 10; i++) {
    /* pick a random private key */
    ecdh_priv priv;
    
    /* serialize it */
    chopped_ecpk_raw raw;
    assert (priv.to_chopped_ecpk_raw (raw));

    /* cat it out */
    printf ("0x%08x\t", raw[4]);
    printf ("0x%08x\t", raw[3]);
    printf ("0x%08x\t", raw[2]);
    printf ("0x%08x\t", raw[1]);
    printf ("0x%08x\t", raw[0]);

    /* deserialize it */
    ecdh_pub check(raw);

    /* compare it against original key */
    if (*priv.public_key () == check) {
      printf("+");
    } else {
      count++;
      printf("-");
    }
  }

  printf("Testing chopped public keys: %d/%d\n",
	 count, 10);
  
}

/* quick'n dirty way to inialize the pseudorandom number generator */
void
ri (void)
{
  struct {
    int pid;
    int time;
  } rid;
  rid.pid = getpid ();
  rid.time = time (NULL);
  prng_seed (&rid, sizeof (rid));
}

int main (int argc, char**argv) {
  
  ri();

  test_ecc_1();
  test_ecc_2a();
  test_ecc_2b();
  test_ecc_3();
  test_ecc_4();
  test_ecc_5();
  test_ecc_6();
  return 0;
}
