/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: calculate_poc.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Finds PMAC(poc_key, path)
 ******************************************************************************/

module calculate_poc
  (// --- Interface to packet parser
   input      [0:63]  icing_realm_id,
   input              icing_realm_id_0_vld,
   input              icing_realm_id_1_vld,
   input              icing_realm_id_2_vld,
   input              icing_realm_is_last,

   // --- Interface to poc key
   input      [0:127] poc_key,
   input      [0:127] poc_key0,
   input              poc_key_vld,
   
   // --- Interface to flow control
   output             pause,

   // --- Interface to shared AES core
   output     [0:127] aes_enc_plain,
   output     [0:127] aes_enc_key,
   output             aes_enc_plain_vld,
   input              aes_enc_plain_ack,
   input      [0:127] aes_enc_cipher,
   input              aes_enc_cipher_vld,

   // --- Results
   output     [0:127] poc,
   output             poc_vld,

   // --- Misc
   input              clk,
   input              reset);

   //----------------------- local params ----------------------------
   localparam         WAIT_MSG_BLOCK_START   = 1,
                      DO_WORD0               = 4,
                      DO_WORD1               = 8,
                      DO_LAST_WORD_ODD0      = 16,
                      DO_LAST_WORD_ODD1      = 32,
                      DO_LAST_WORD_EVEN0     = 64,
                      DO_LAST_WORD_EVEN1     = 2;
   
   //------------------------ Wires/Regs -----------------------------
   wire [0:63]        word;
   reg [0:127] 	      msg_block;
   wire [0:127]       msg_block_buf;

   reg                fifo_rd_en;
   reg [0:6]          state, state_nxt;
   reg                msg_block_vld;
   reg [0:15]         msg_block_last_byte;
   wire [0:15]        msg_block_last_byte_buf;

   reg [0:63]         word_prev;

   wire [0:127]        poc_key_reg, poc_key0_reg;

   reg 		       poc_key_fifo_rd_en;

   reg 		       id_fifo_almost_full;
   
   //------------------------- Modules -------------------------------

   fallthrough_small_fifo 
     #(.WIDTH(256),
       .MAX_DEPTH_BITS(1))
      poc_key_fifo
        (.din           ({poc_key, poc_key0}),
         .wr_en         (poc_key_vld),
         .rd_en         (poc_key_fifo_rd_en),
         .dout          ({poc_key_reg, poc_key0_reg}),
         .full          (poc_key_fifo_full),
         .prog_full     (),
         .nearly_full   (),
         .empty         (poc_key_fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   /* Need to store max path len before poc key arrives
    * which might be at the end of pkt (max path idx) */
   fallthrough_small_fifo 
     #(.WIDTH(65),
       .MAX_DEPTH_BITS(8),
       .PROG_FULL_THRESHOLD(252))
      id_fifo
        (.din           ({icing_realm_id, icing_realm_is_last}),
         .wr_en         (icing_realm_id_0_vld
                         || icing_realm_id_1_vld
                         || icing_realm_id_2_vld),
         .rd_en         (fifo_rd_en),
         .dout          ({word, is_last}),
         .full          (),
         .prog_full     (id_fifo_almost_full_nxt),
         .nearly_full   (),
         .empty         (fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   buffer #(.WIDTH(128+16)) buffer
     (.din ({msg_block, msg_block_last_byte}),
      .din_vld (msg_block_vld),
      .din_ack (msg_block_ack),
      .dout ({msg_block_buf, msg_block_last_byte_buf}),
      .dout_vld (msg_block_vld_buf),
      .dout_ack (msg_block_ack_buf),
      .clk (clk),
      .reset (reset));
      
   pmac #(.ENABLE_KEY0(1), .DO_PADDING(0)) pmac
     (// Outputs
      .msg_block_ack                    (msg_block_ack_buf),
      .mac                              (poc),
      .mac_vld                          (poc_vld),
      .aes_enc_plain                    (aes_enc_plain[0:127]),
      .aes_enc_key                      (aes_enc_key[0:127]),
      .aes_enc_plain_vld                (aes_enc_plain_vld),
      // Inputs
      .msg_block                        (msg_block_buf),
      .key                              (poc_key_reg),
      .key0                             (poc_key0_reg),
      .msg_block_vld                    (msg_block_vld_buf),
      .msg_block_last_byte              (msg_block_last_byte_buf),
      .aes_enc_plain_ack                (aes_enc_plain_ack),
      .aes_enc_cipher                   (aes_enc_cipher[0:127]),
      .aes_enc_cipher_vld               (aes_enc_cipher_vld),
      .clk                              (clk),
      .reset                            (reset));
   
   //-------------------------- Logic --------------------------------
   always @(posedge clk) begin
      id_fifo_almost_full <= id_fifo_almost_full_nxt;
   end
   assign pause = id_fifo_almost_full || poc_key_fifo_full;

   always @(*) begin
      fifo_rd_en            = 0;
      state_nxt             = state;
      msg_block_vld         = 0;
      msg_block_last_byte   = 0;
      msg_block             = {word_prev, word};      
      poc_key_fifo_rd_en    = 1'b0;
      case(state)
         /* Assemble the first word of the msg block input */
         WAIT_MSG_BLOCK_START: begin
            if(!poc_key_fifo_empty) begin
               /* by now we should at least have the 
                * first word of the path. So no need to
                * check if !fifo_empty */
               fifo_rd_en = 1'b1;
               state_nxt  = DO_WORD1;
            end
         end

         DO_WORD0: begin
            if(!fifo_empty) begin
               fifo_rd_en   = 1;
               if(is_last) begin
                  state_nxt  = DO_LAST_WORD_ODD0;
               end else begin
                  state_nxt  = DO_WORD1;
               end
            end
         end // case: DO_WORD0

         DO_WORD1: begin
            if(!fifo_empty) begin
               msg_block_vld   = 1'b1;
               if(msg_block_ack) begin
                  fifo_rd_en      = 1;
                  if(is_last) begin
                     state_nxt  = DO_LAST_WORD_EVEN0;
                  end else begin
                     state_nxt  = DO_WORD0;
                  end
               end
            end
         end // case: DO_WORD1

         DO_LAST_WORD_ODD0: begin
            if(!fifo_empty) begin
               msg_block_vld   = 1'b1;
               if(msg_block_ack) begin
                  fifo_rd_en      = 1;
                  state_nxt       = DO_LAST_WORD_ODD1;
               end
            end
         end

         DO_LAST_WORD_ODD1: begin
            if(!fifo_empty) begin
               msg_block_vld         = 1'b1;
               msg_block_last_byte   = 16'h0100;
               msg_block             = {word, 1'b1, 63'h0};
               if(msg_block_ack) begin
                  fifo_rd_en   = 1'b1;
                  state_nxt    = WAIT_MSG_BLOCK_START;
		  poc_key_fifo_rd_en = 1'b1;
               end
            end
         end

         DO_LAST_WORD_EVEN0: begin
            if(!fifo_empty) begin
               fifo_rd_en   = 1;
               state_nxt    = DO_LAST_WORD_EVEN1;
            end
         end

         DO_LAST_WORD_EVEN1: begin
            if(!fifo_empty) begin
               msg_block_last_byte   = 16'h0001;
               msg_block_vld         = 1;
               if(msg_block_ack) begin
                  fifo_rd_en   = 1;
                  state_nxt    = WAIT_MSG_BLOCK_START;
		  poc_key_fifo_rd_en = 1'b1;
               end
            end
         end
      endcase // case(state)
   end // always @ (*)

   always @(posedge clk) begin
      state <= reset ? WAIT_MSG_BLOCK_START : state_nxt;
      if(fifo_rd_en) word_prev <= word;
   end

   // synthesis translate_off
   integer num_pocs = 0;
   reg 	   poc_vld_d1 = 0;
   always @(posedge clk) begin
      poc_vld_d1 <= poc_vld;
      if(poc_vld === 1) begin
         num_pocs <= num_pocs + 1;
      end
      if(poc_vld_d1 === 1) begin
         $display($time, " %m Num PoCs calc'ed  :  %u", num_pocs);
      end
   end
   // synthesis translate_on

endmodule // calculate_poc
