/*******************************************************************************
 * vim:set shiftwidth=3 softtabstop=3 expandtab:
 * $Id$
 *
 * Module: lookup_poc_key.v
 * Project: ICING v3
 * Author: Jad Naous <jnaous@stanford.edu>
 * Description:
 *   Looks up my subrealm in the SRAM and gets the PoC key
 * 
 ******************************************************************************/

module lookup_poc_key
  (// --- Interface to packet parser
   input      [0:31]  icing_subrealm,
   input              icing_subrealm_vld,
   input              icing_realm_is_me,

   // --- Interface to flow control
   output             pause,

   // --- Interface to SRAM
   output reg [18:0]  sram_rd_addr,
   output reg         sram_rd_req,
   input      [0:63]  sram_rd_data,
   input              sram_rd_vld,
   input              sram_rd_ack,

   // --- Results
   output     [0:127] poc_key,
   output     [0:127] poc_key0, // E(poc_key, 0)
   output             poc_key_vld,

   // --- Misc
   input              clk,
   input              reset);

   //-------------------- Internal Parameters ------------------------
   localparam WAIT_FOR_SUBREALM   = 0,
              RD_KEY_0            = 1,
              RD_KEY_1            = 2,
              RD_KEY_2            = 3;
   localparam V_WIDTH = `VNODE_WIDTH;
   
   //------------------------ Wires/Regs -----------------------------
   wire [0:V_WIDTH-1] subrealm;
   reg [0:1]          req_state_nxt, req_state;
   reg                fifo_rd_en;
   reg [0:1]          poc_word_cnt;
   reg [0:191]        poc_word_0;
   
   //------------------------- Modules -------------------------------
   fallthrough_small_fifo 
     #(.WIDTH(V_WIDTH),
       .MAX_DEPTH_BITS(2))
      subrealm_fifo
        (.din           (icing_subrealm[0:V_WIDTH-1]),
         .wr_en         (icing_subrealm_vld & icing_realm_is_me),
         .rd_en         (fifo_rd_en),
         .dout          (subrealm),
         .full          (),
         .prog_full     (),
         .nearly_full   (pause),
         .empty         (fifo_empty),
         .reset         (reset),
         .clk           (clk)
         );

   //-------------------------- Logic --------------------------------

   /*********************************************************************
    * Generate SRAM lookups
    ********************************************************************/
   always @(*) begin
      sram_rd_addr    = `SRAM_POC_KEY_BASE | {subrealm, `POC_KEY_ENTRY_BITS'h0};
      req_state_nxt   = req_state;
      sram_rd_req     = 0;
      fifo_rd_en      = 0;
      
      case(req_state)
         WAIT_FOR_SUBREALM: begin
            sram_rd_req = !fifo_empty;
            if(sram_rd_ack) begin
               req_state_nxt   = RD_KEY_0;
            end
         end // case: WAIT_FOR_SUBREALM
         
         RD_KEY_0: begin
            sram_rd_req       = 1;
            sram_rd_addr[`POC_KEY_ENTRY_BITS-1:0] = `POC_KEY_ENTRY_BITS'h1;
            if(sram_rd_ack) begin
               req_state_nxt   = RD_KEY_1;
            end
         end

         RD_KEY_1: begin
            sram_rd_req       = 1;
            sram_rd_addr[`POC_KEY_ENTRY_BITS-1:0] = `POC_KEY_ENTRY_BITS'h2;
            if(sram_rd_ack) begin
               req_state_nxt   = RD_KEY_2;
            end
         end

         RD_KEY_2: begin
            sram_rd_req       = 1;
            sram_rd_addr[`POC_KEY_ENTRY_BITS-1:0] = `POC_KEY_ENTRY_BITS'h3;
            if(sram_rd_ack) begin
               fifo_rd_en      = 1'b1;
               req_state_nxt   = WAIT_FOR_SUBREALM;
            end
         end
      endcase // case(req_state)
   end // always @ (*)

   always @(posedge clk) begin
      req_state    <= reset ? WAIT_FOR_SUBREALM : req_state_nxt;
   end
   
   /*********************************************************************
    * Generate Output
    ********************************************************************/
   assign poc_key_vld   = sram_rd_vld & (poc_word_cnt==2'h3);
   assign poc_key       = poc_word_0[0:127];
   assign poc_key0      = {poc_word_0[128:191], sram_rd_data};
   
   always @(posedge clk) begin
      if(reset) poc_word_cnt <= 2'h0;
      else if(sram_rd_vld) poc_word_cnt <= poc_word_cnt + 1'b1;

      if(sram_rd_vld) poc_word_0 <= {poc_word_0[64:191], sram_rd_data};
   end

   // synthesis translate_off
   integer num_keys = 0;
   always @(posedge clk) begin
      if(poc_key_vld === 1) begin
         num_keys = num_keys + 1;
         $display($time, " %m Num poc keys found:  %u", num_keys);
      end
   end
   // synthesis translate_on
endmodule // lookup_poc_key


